(*  Title     : A Shallow Encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Fuse operator\<close>   

theory Fuse
imports Semantics 
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="
  cvc4 z3 e spass vampire
"]
*)
text \<open>
This theory introduces the fuse operator.
 \<close>

subsection \<open>Definitions\<close>

primrec fuse :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> 'a interval"
 where  fuse_INil   : "fuse \<langle>x\<rangle> ys = ys"
      | fuse_ICons : "fuse (x \<odot> xs) ys = x \<odot> (fuse xs ys)"

primrec lfuse :: "'a interval interval \<Rightarrow> 'a interval"
where   lfuse_INil   : "lfuse \<langle>xs\<rangle> = xs"
      | lfuse_ICons : "lfuse (x \<odot> xs) = fuse x (lfuse xs)"


primrec lastfirst :: "'a interval interval \<Rightarrow> bool"
 where "lastfirst \<langle>xs\<rangle> = True"
      | "lastfirst (xs \<odot> xxs) = 
         ( ((ilast xs) = (ifirst (ifirst xxs))) \<and> (lastfirst xxs))"


subsection \<open>Lemmas\<close>

lemma fuse_ilen :
  assumes  "ilast xs = ifirst ys"
  shows    " ilen (fuse xs ys) = (ilen xs) + (ilen ys)"
using assms by (induct xs) simp_all

lemma fuse_ilen_a:
 "ilen(fuse xs ys) = ilen xs + ilen ys"
proof
 (induct xs arbitrary: ys)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case by simp
qed

lemma fuse_inth:
assumes "i\<le> ilen (fuse xs ys)"
        "ilast xs = ifirst ys"
shows   "(i \<le> ilen xs \<longrightarrow> inth (fuse xs ys) i = inth xs i)
         \<and> 
         (ilen xs \<le> i \<and> i\<le> ilen (fuse xs ys) \<longrightarrow> inth (fuse xs ys) i = inth ys (i - ilen xs))
        "
using assms
proof
  (induct xs arbitrary: i)
  case (INil x)
  then show ?case by simp 
  next
  case (ICons x1a xs)
  then show ?case 
  using less_Suc_eq_0_disj less_Suc_eq_le by fastforce
qed

lemma fuse_inth_a:
 assumes " j \<le> ilen ys"
         " ilast xs = ifirst ys "
 shows   " inth (fuse xs ys) (ilen xs +j) = (inth ys j)"
using assms 
by (simp add: fuse_ilen_a fuse_inth)

lemma fuse_leftneutral : 
    "fuse (INil (ifirst xs)) xs = xs"
by simp

lemma fuse_rightneutral : 
    "fuse xs (INil (ilast xs)) = xs"
by (induct xs) simp_all

lemma ifirst_fuse :
 assumes " ilast xs = ifirst ys"
 shows   " ifirst (fuse xs ys) = ifirst xs"
using assms by (induct xs) simp_all

lemma ilast_fuse :
 assumes "ilast xs = ifirst ys"
 shows   " ilast (fuse xs ys) = ilast ys"
using assms by (induct xs) simp_all

lemma FusionAssoc :
 assumes " (ilast xs) = (ifirst ys)"
         " (ilast ys) = (ifirst zs)"
 shows   " (fuse xs (fuse ys zs)) = (fuse (fuse xs ys) zs)"
using assms by (induct xs) simp_all


lemma ilast_ifirst:
    " (ilast (prefix i xs)) = (ifirst (suffix i xs))"
using ilast_ifirst by blast
  
lemma prefix_fuse :
  assumes  " ilast xs = ifirst ys "
  shows    " (prefix (ilen xs) (fuse xs ys)) = xs"
using assms by (induct xs arbitrary: ys) simp_all

lemma suffix_fuse :
 assumes   "ilast xs = ifirst ys "
 shows     "(suffix (ilen xs) (fuse xs ys)) = ys"
using assms by (induct xs arbitrary: ys) simp_all
 
lemma fuse_prefix_suffix_ilen :
assumes "n \<le> ilen xs"
shows   "ilen (fuse (prefix n xs) (suffix n xs)) = ilen xs "
using assms 
by (metis fuse_ilen_a prefix_ilen_good suffix_ilen_good le_add_diff_inverse)

lemma fuse_prefix_suffix_inth :
assumes "n \<le> ilen xs"
        "i \<le> ilen xs"
shows   "   inth (fuse (prefix n xs) (suffix n xs)) i = inth xs i "
using assms fuse_inth[of "i" "(prefix n xs)" "(suffix n xs)"]  
    ilast_ifirst[of "n" "xs"] 
    fuse_prefix_suffix_ilen[of "n" "xs"] 
    by (metis diff_add inth_prefix inth_suffix prefix_ilen_good le_cases 
        le_diff_conv ordered_cancel_comm_monoid_diff_class.add_diff_inverse)

lemma fuse_prefix_suffix:
assumes "n\<le> ilen xs"
shows   "fuse (prefix n xs) (suffix n xs) = xs"
using assms 
by (simp add: fuse_prefix_suffix_ilen fuse_prefix_suffix_inth interval_eq_inth_eq)

lemma chop_fuse_1 : 
  " (\<exists> \<sigma>1 \<sigma>2. \<sigma> = fuse \<sigma>1 \<sigma>2 \<and>
       (\<sigma>1  \<Turnstile> f) \<and> (\<sigma>2  \<Turnstile> g) \<and> 
       (ilast \<sigma>1 = ifirst \<sigma>2)) =
      (\<exists> i. 0 \<le> i \<and> i\<le> ilen \<sigma> \<and> ( prefix i \<sigma>  \<Turnstile> f) \<and> (suffix i \<sigma>  \<Turnstile> g)) "
by (metis fuse_ilen fuse_prefix_suffix ilast_ifirst 
          ilen_gr_zero prefix_fuse suffix_fuse le_add_same_cancel1)

lemma chop_fuse_2 : 
  " (\<exists> \<sigma>1 \<sigma>2. \<sigma> = fuse \<sigma>1 \<sigma>2 \<and>
       (\<sigma>1 \<in> X) \<and> (\<sigma>2 \<in> Y) \<and> 
       (ilast \<sigma>1 = ifirst \<sigma>2)) =
      (\<exists> i\<le> ilen \<sigma>. (prefix i \<sigma>)\<in>X \<and> (suffix i \<sigma>)\<in>Y) "
by (metis fuse_ilen fuse_prefix_suffix ilast_ifirst prefix_fuse suffix_fuse le_add1)
  
lemma chop_fuse:
  " (\<exists> \<sigma>1 \<sigma>2. \<sigma> = fuse \<sigma>1 \<sigma>2 \<and>
       (\<sigma>1  \<Turnstile> f) \<and> (\<sigma>2  \<Turnstile> g) \<and> 
       (ilast \<sigma>1 = ifirst \<sigma>2)) =
      (\<sigma>  \<Turnstile> f;g) "
by (metis chop_defs fuse_ilen_a fuse_prefix_suffix ilast_ifirst 
          prefix_fuse suffix_fuse le_add1)

lemma sub_fuse:
 assumes " k \<le> n "
         " n \<le> m"
         " m \<le> ilen xs"
 shows   "fuse (sub k n xs) (sub n m xs) = (sub k m xs)"
proof -
 have 1: " ilast(sub k n xs) = (inth xs n)"
    using assms ilast_sub le_trans less_imp_le_nat by blast
 have 2: " ifirst(sub n m xs) = (inth xs n)"
    using assms ifirst_sub less_imp_le_nat by blast
 have 3: " ilen(fuse (sub k n xs) (sub n m xs)) = ilen(sub k m xs)"
    by (metis Nat.add_diff_assoc2 assms(1) assms(2) assms(3) fuse_ilen_a 
        ilen_sub le_trans ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 4: " (\<forall> i. i \<le> ilen(sub k m xs) \<longrightarrow>
              (inth (fuse (sub k n xs) (sub n m xs)) i) = (inth xs (k+i)))"
    proof 
      fix i
      show "i \<le> ilen (sub k m xs) \<longrightarrow> inth (fuse (sub k n xs) (sub n m xs)) i = inth xs (k + i)  "
      proof -
       have 41: "ilen (sub k m xs) = (m-k) "
          using assms ilen_sub le_trans less_imp_le_nat by metis
       have 42: "i \<le> ilen (sub k m xs) \<longrightarrow> inth (fuse (sub k n xs) (sub n m xs)) i =
                  (if i \<le> ilen(sub k n xs) then (inth (sub k n xs) i) 
                                             else (inth (sub n m xs) (i-ilen(sub k n xs))))"
       by (metis "1" "2" "3" fuse_inth le_cases)
       have 43: " i \<le> (m-k) \<longrightarrow> inth (fuse (sub k n xs) (sub n m xs)) i = 
                  (if i\<le> (n-k) then (inth xs (k+i)) else (inth xs (n+(i-(n-k)))))"
            using "42" Nat.le_diff_conv2 assms(1) assms(2) assms(3) by auto
       have 44: "i \<le> (m-k) \<longrightarrow> inth (fuse (sub k n xs) (sub n m xs)) i = 
                  (inth xs (k+i))  "
              by (simp add: "43" add.commute assms less_imp_le_nat)
        show ?thesis 
        by (simp add: "41" "44")
     qed
   qed
  have 5: "(\<forall> i. i \<le> ilen(sub k m xs) \<longrightarrow> (inth (sub k m xs) i) = (inth xs (k+i))) "
      using assms(1) assms(2) assms(3) by auto
  show ?thesis 
  by (simp add: "3" "4" "5" interval_eq_inth_eq)
qed

lemma sub_fuse_idx:
 assumes "index_sequence 0 l"
         " inth l (ilen l) = ilen \<sigma>"
         " (Suc i) < ilen l"
 shows   " fuse (sub (inth l i) (inth l (Suc i)) \<sigma>) (sub (inth l (Suc i)) (inth l (ilen l)) \<sigma>) = 
           (sub (inth l i) (inth l (ilen l)) \<sigma>)"
proof -
 have 1: " ilast(sub (inth l i) (inth l (Suc i)) \<sigma>) = (inth \<sigma> (inth l (Suc i)))"
    by (metis assms idx_less_equal idx_less_last_1 ilast_sub le_add2 less_imp_le_nat plus_1_eq_Suc)
 have 2: " ifirst(sub (inth l (Suc i)) (inth l (ilen l)) \<sigma> ) = (inth \<sigma> (inth l (Suc i))) "
    by (metis assms dual_order.strict_iff_order eq_imp_le idx_less_last_1 ifirst_sub)
 have 3: " (inth l i) < (inth l (Suc i)) " 
    using Suc_lessD assms idx_less_than less_imp_le_nat by blast
 have 4: "(inth l (Suc i)) < (inth l (ilen l)) " 
    using assms idx_less_last_1 by blast
 show ?thesis using "3" "4" assms by (simp add: sub_fuse)
qed


lemma idx_fuse_ifirst_ilast:
 assumes "index_sequence 0 l1"
         "index_sequence 0 l2" 
         "inth l1 (ilen l1) = cp"
         "inth l2 (ilen l2) = ilen \<sigma> -cp"
         "cp \<le> ilen \<sigma>"
         "l = fuse l1 (imap (shift cp) l2) "
 shows   "ilast l1 = ifirst (imap (shift cp) l2)"
using assms 
 by (metis  shift_def add.left_neutral index_sequence_def inth_imap )

lemma idx_fuse_inth_cp:
 assumes "index_sequence 0 l1"
         "index_sequence 0 l2"
         "inth l1 (ilen l1) = cp"
         "inth l2 (ilen l2) = ilen \<sigma> -cp"
         "cp \<le> ilen \<sigma>"
         "l = fuse l1 (imap (shift cp) l2)"
         "i \<le> ilen l2"
 shows   "inth l (ilen l1 + i) = cp + inth l2 i"
proof -
 have 1: "ilast l1 = ifirst (imap (shift cp) l2)"
    using assms idx_fuse_ifirst_ilast by blast
 have 2: "inth l (ilen l1 + i) = inth (imap (shift cp) l2) i"
    using assms by (metis "1" fuse_inth_a ilen_imap)
 have 3: "inth (imap (shift cp) l2) i = inth l2 i +cp"
    by (simp add:  shift_def inth_imap)
  show ?thesis using "2" "3" by auto 
qed


lemma idx_fuse_idx:
 assumes "index_sequence 0 l1"
         "index_sequence 0 l2" 
         "inth l1 (ilen l1) = cp"
         "inth l2 (ilen l2) = ilen \<sigma> -cp"
         "cp \<le> ilen \<sigma>"
         "l = fuse l1 (imap (shift cp) l2)"
         "i \<le> ilen l2"
 shows   "index_sequence 0 l"
proof -
 have 1: "ilast l1 = ifirst (imap (shift cp) l2)"
    using assms idx_fuse_ifirst_ilast by blast
 have 2: " inth (fuse l1 (imap (shift cp) l2)) 0 = inth l1 0"
     using "1" fuse_inth by blast
 have 3: " ilen (fuse l1 (imap (shift cp) l2)) = ilen l1 + ilen l2"
    by (simp add: fuse_ilen_a)
 have 4: "\<forall>i. 0\<le>i  \<and> i\<le>ilen l1 \<longrightarrow> 
           inth (fuse l1 (imap (shift cp) l2)) i  =
           inth l1 i"
    by (metis "1"  inth_prefix prefix_fuse)
 have 5: "\<forall>i. ilen l1 \<le>i  \<and> i\<le>ilen l1 + ilen l2 \<longrightarrow> 
           inth (fuse l1 (imap (shift cp) l2)) i  =
           cp + inth l2 (i -ilen l1) " 
    by (metis (no_types, lifting) "1" "3"  shift_def add.commute fuse_inth inth_imap)
 have 6: "\<forall>i. 0\<le>i  \<and> i<ilen l1+ilen l2 \<longrightarrow> 
            inth (fuse l1 (imap (shift cp) l2)) i < inth (fuse l1 (imap (shift cp) l2)) (Suc i)"    
    using assms by (metis "1" "3" "4" add.right_neutral index_sequence_def idx_link 
        idx_split ilen_gr_zero prefix_fuse suffix_fuse le_add1)
 have 7: "index_sequence 0 l =
          ((inth l1 0) = 0 \<and>
          (\<forall>i. 0\<le>i  \<and> i<ilen l1+ilen l2 \<longrightarrow> 
            inth (fuse l1 (imap (shift cp) l2)) i < inth (fuse l1 (imap (shift cp) l2)) (Suc i)))  "
    by (simp add: "2" "3" assms index_sequence_def)
 from 7 6 2 show ?thesis 
 using assms index_sequence_def by auto
qed


lemma idx_fuse_ilen:
 assumes "index_sequence 0 l1"
         "index_sequence 0 l2" 
         "inth l1 (ilen l1) = cp"
         "inth l2 (ilen l2) = ilen \<sigma> -cp"
         "cp \<le> ilen \<sigma>"
         "l = fuse l1 (imap (shift cp) l2)"
         "i \<le> ilen l2"
 shows   "inth l (ilen l) = ilen \<sigma>"
using assms idx_fuse_inth_cp[of "l1" "l2" "cp" "\<sigma>"] by (simp add: fuse_ilen_a)

lemma ifirst_lfuse_ifirst:
assumes "lastfirst (xs \<odot> xxs)"
shows  "ifirst(lfuse xxs) = ifirst(ifirst xxs)"
using assms 
proof 
 (induct xxs)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xxs)
 then show ?case 
      proof (cases x1a)
      case (INil x1)
      then show ?thesis using ICons.hyps ICons.prems by auto
      next
      case (ICons x21 x22)
      then show ?thesis by simp
      qed
qed 

lemma ifirst_lfuse:
assumes "lastfirst (xs \<odot> xxs)"
shows "(ifirst (lfuse (xs \<odot> xxs)) ) = (ifirst xs)"
proof -
 have 1: "lastfirst (xs \<odot> xxs)"
  using assms by auto
 have 2: "( ((ilast xs) = (ifirst (ifirst xxs))) \<and> (lastfirst xxs))  "
   using "1" by simp
 have 3: "(ilast xs) = (ifirst (ifirst xxs))  "
   using "2" by auto
 have 4: "ifirst (lfuse (xs \<odot> xxs)) = ifirst(fuse xs (lfuse xxs))  "
   by simp
 have 5: "ifirst(lfuse xxs) = ifirst(ifirst xxs)"
   using assms ifirst_lfuse_ifirst by blast
 have 6: "ifirst(fuse xs (lfuse xxs)) = ifirst xs "
   by (metis "3" "5" ifirst_fuse)
 show ?thesis using "6" by auto 
qed 

lemma lastfirst_lfuse_ilast:
 assumes "lastfirst xxs"
 shows   " ilast(lfuse xxs) = ilast(ilast xxs)"
using assms
 proof 
  (induct xxs)
  case (INil x)
  then show ?case by simp
  next
  case (ICons x1a xxs)
  then show ?case
  by (metis fuse_ilen_a fuse_inth_a ifirst_lfuse_ifirst 
            inth_last lastfirst.simps(2) lfuse_ICons order.order_iff_strict)
qed

lemma lastfirst_lfuse:
assumes "lastfirst xxs"
shows "ifirst (lfuse xxs) = ifirst(ifirst(xxs))"
using assms 
proof
 (cases xxs)
 case (INil x1)
 then show ?thesis by simp
 next
 case (ICons x21 x22)
 then show ?thesis 
 using assms ifirst_lfuse by auto
 qed

lemma lfuse_ilen :
 assumes " lastfirst xxs "
 shows   " ilen (lfuse xxs) = (\<Sum>k::nat= 0..(ilen xxs). ilen(inth xxs k))"
using assms
proof
  (induct xxs)
  case (INil x)
  then show ?case by simp
  next
  case (ICons x1a xxs)
  then show ?case proof -
    have 1 : "ilen (lfuse (x1a \<odot> xxs)) = ilen (fuse x1a (lfuse xxs))" 
     by simp
    have 2 : "lastfirst (x1a \<odot> xxs)" using ICons.prems by auto
    have 3 : "ilast x1a = ifirst(lfuse xxs)" 
       using ICons.prems ifirst_lfuse_ifirst by fastforce
    have 4 : "ilen (fuse x1a (lfuse xxs)) = (ilen x1a) + ilen(lfuse xxs)  "
     using "3" fuse_ilen by blast
    have 5: "(ilen x1a) + ilen(lfuse xxs) =
             (ilen x1a) + (\<Sum>k::nat= 0..(ilen xxs). ilen(inth xxs k))  "
     using ICons.hyps ICons.prems by auto
    have 6: "(\<Sum>k = 0..ilen (x1a \<odot> xxs). ilen ( inth (x1a \<odot> xxs) k)) =
             (ilen(inth (x1a \<odot> xxs) 0)   ) + 
             (\<Sum>k = 1..1+ilen (xxs). ilen ( inth (x1a \<odot> xxs) k)) "
        by (simp add: sum.atLeast_Suc_atMost)
    have 7: "(ilen(inth (x1a \<odot> xxs) 0)   ) = ilen(x1a) "
     by simp
    have 8: "(\<Sum>k = 1..1+ilen (xxs). ilen ( inth (x1a \<odot> xxs) k)) =
             (\<Sum>k = 0..ilen (xxs). ilen ( inth (x1a \<odot> xxs) (k+1))) "
      using sum.shift_bounds_cl_nat_ivl[of "\<lambda>k. ilen ( inth (x1a \<odot> xxs) k)"   0 1 "ilen(xxs)"]
          by simp   
    have 9: "(\<Sum>k = 0..ilen (xxs). ilen ( inth (x1a \<odot> xxs) (k+1))) =
              (\<Sum>k = 0..ilen (xxs). ilen ( inth (xxs) (k))) "
     by auto
    have 10: "(ilen x1a) + (\<Sum>k::nat= 0..(ilen xxs). ilen(inth xxs k)) =
             (\<Sum>k = 0..ilen (x1a \<odot> xxs). ilen ( inth (x1a \<odot> xxs) k)) "
    using "6" "7" "8" "9" by linarith
    show ?thesis 
    by (simp add: "10" "4" "5")       
  qed
qed

lemma idx_fuse:
 assumes "ilast l1 = ifirst l2" 
 shows  
 " (index_sequence (ifirst l1) (fuse l1 l2)) =
     ( index_sequence (ifirst l1) l1 \<and> index_sequence (ifirst l2) l2 ) "
using assms 
 proof
  (induct l1 arbitrary: l2)
  case (INil x)
  then show ?case by (simp add: index_sequence_def)
  next
  case (ICons x1a l1)
  then show ?case 
   using idx_expand1 ifirst_fuse by force
qed

lemma idx_lfuse_help1:
assumes " (\<forall>k. k<ilen (lfuse l) \<longrightarrow> 
       inth (fuse x1a (lfuse l)) (ilen x1a + k) < 
       inth (fuse x1a (lfuse l)) (ilen x1a + Suc k)) 
        "
        "ilen x1a \<le> n"
        "n< ilen x1a +ilen (lfuse l)"
shows " inth (fuse x1a (lfuse l)) n < inth (fuse x1a (lfuse l)) (Suc n)"
using assms 
by (metis add_Suc_right add_less_imp_less_left le_Suc_ex)

lemma idx_lfuse:
 assumes "lastfirst l"
 shows   "(index_sequence (ifirst (lfuse l)) (lfuse l)) = 
          (\<forall> i\<le>ilen l. index_sequence (ifirst (inth l i)) (inth l i))"
using assms 
 proof 
   (induct l)
   case (INil x)
   then show ?case by (simp add: index_sequence_def)
   next
   case (ICons x1a l)
   then show ?case 
    proof -
     have 0: "lastfirst l" 
         using ICons.prems lastfirst.simps(2) by blast
     have 1: "index_sequence (ifirst (lfuse (x1a \<odot> l))) (lfuse (x1a \<odot> l)) =
              (inth (fuse x1a (lfuse l)) 0 = ifirst (fuse x1a (lfuse l)) \<and>
              (\<forall>n<ilen (fuse x1a (lfuse l)). 
                  inth (fuse x1a (lfuse l)) n < inth (fuse x1a (lfuse l)) (Suc n))) "
        by (simp add: index_sequence_def)
     have 2: "(inth (fuse x1a (lfuse l)) 0) = ifirst (fuse x1a (lfuse l)) "
        by simp
     have 3: "ilen (fuse x1a (lfuse l)) = ilen x1a + ilen(lfuse l) "
        by (simp add: fuse_ilen_a)
     have 4: " (\<forall>n<ilen (fuse x1a (lfuse l)). 
                  inth (fuse x1a (lfuse l)) n < inth (fuse x1a (lfuse l)) (Suc n)) =
               (\<forall>n<ilen x1a + ilen (lfuse l). 
                  inth (fuse x1a (lfuse l)) n < inth (fuse x1a (lfuse l)) (Suc n))"
        by (simp add: "3")
     have 5: "(\<forall>n<ilen x1a + ilen (lfuse l). 
                  inth (fuse x1a (lfuse l)) n < inth (fuse x1a (lfuse l)) (Suc n)) =
              ((\<forall>n<ilen x1a. inth (fuse x1a (lfuse l)) n < inth (fuse x1a (lfuse l)) (Suc n))  \<and>
               (\<forall>n. 0 \<le> n-ilen x1a \<and> n - ilen x1a<ilen (lfuse l) \<longrightarrow> 
                       inth (fuse x1a (lfuse l)) n < inth (fuse x1a (lfuse l)) (Suc n)))   "
         by auto
           (metis add.commute less_diff_conv2 not_less)
     have 6: "(\<forall>n<ilen x1a. inth (fuse x1a (lfuse l)) n < inth (fuse x1a (lfuse l)) (Suc n)) =
              index_sequence (ifirst x1a) x1a"
          proof (simp add: index_sequence_def)
          have "ilast x1a = ifirst (lfuse l)"
          using ICons.prems ifirst_lfuse_ifirst by force
          then show "(\<forall>n<ilen x1a. inth (fuse x1a (lfuse l)) n < inth (fuse x1a (lfuse l)) (Suc n)) = (\<forall>n<ilen x1a. inth x1a n < inth x1a (Suc n))"
          by (metis (no_types) Suc_leI inth_prefix less_or_eq_imp_le prefix_fuse)
          qed
     have 7: "(\<forall>n. ilen x1a \<le> n \<and> n < ilen x1a + ilen (lfuse l) \<longrightarrow> 
                       inth (fuse x1a (lfuse l)) n < inth (fuse x1a (lfuse l)) (Suc n)) =
              (\<forall>k.  k<ilen (lfuse l) \<longrightarrow> 
                       inth (fuse x1a (lfuse l)) (ilen x1a + k) < 
                       inth (fuse x1a (lfuse l)) (ilen x1a + Suc k)) "    
            using idx_lfuse_help1 by auto
     have 8: "(\<forall>k.  k<ilen (lfuse l) \<longrightarrow> 
                       inth (fuse x1a (lfuse l)) (ilen x1a + k) < 
                       inth (fuse x1a (lfuse l)) (ilen x1a + Suc k)) =
              (\<forall>k.  k<ilen (lfuse l) \<longrightarrow> 
                       inth ( (lfuse l)) (k) < 
                       inth ( (lfuse l)) (Suc k)) "  (is "?L=?R")
            proof 
             show "?L \<Longrightarrow> ?R"
             by (metis ICons.prems Suc_leI fuse_inth_a ifirst_lfuse_ifirst lastfirst.simps(2) 
                  le_simps(1))
             show "?R \<Longrightarrow> ?L"
              by (metis ICons.prems Suc_leI add_Suc_right fuse_inth_a 
                  ifirst_lfuse_ifirst lastfirst.simps(2) less_imp_le_nat)      
           qed
     have 9: "(\<forall>k.  k<ilen (lfuse l) \<longrightarrow> 
                       inth ( (lfuse l)) (k) < 
                       inth ( (lfuse l)) (Suc k)) = index_sequence (ifirst (lfuse l)) (lfuse l)"
        by (simp add: index_sequence_def)   
     have 91: "index_sequence (ifirst (lfuse (x1a \<odot> l))) (lfuse (x1a \<odot> l)) =
               (index_sequence (ifirst x1a) x1a \<and> index_sequence (ifirst (lfuse l)) (lfuse l))"  
         using 1 2  4 5 6 7 8 9 
         by (metis ICons.prems idx_fuse ifirst_lfuse 
             ifirst_lfuse_ifirst lastfirst.simps(2) lfuse_ICons)
     have 10: "index_sequence (ifirst (lfuse l)) (lfuse l) =
               (\<forall> i\<le>ilen l. index_sequence (ifirst (inth l i)) (inth l i)) "
          using "0" ICons.hyps by blast
     have 11: "(\<forall> i\<le> ilen(x1a\<odot>l). index_sequence (ifirst (inth (x1a\<odot>l) i)) (inth (x1a\<odot>l) i)) =
               (\<forall> i\<le> 1  +ilen l. index_sequence (ifirst (inth (x1a\<odot>l) i)) (inth (x1a\<odot>l) i)) "
        by auto
     have 12: "(\<forall> i\<le> 1  +ilen l. index_sequence (ifirst (inth (x1a\<odot>l) i)) (inth (x1a\<odot>l) i)) =
               ( index_sequence (ifirst (x1a)) (x1a) \<and> 
               (\<forall> i. 1 \<le> i \<and> i\<le> 1  +ilen l \<longrightarrow> 
                 index_sequence (ifirst (inth (x1a \<odot> l) i)) (inth (x1a \<odot> l) i))) "
         by (metis One_nat_def Suc_leI ilen_gr_zero inth_zero 
                prefix_ilen_good ilen.simps(2) order.strict_iff_order)
     have 13: "(\<forall> i. 1 \<le> i \<and> i\<le> 1  +ilen l \<longrightarrow> 
                index_sequence (ifirst (inth (x1a \<odot> l) i)) (inth (x1a \<odot> l) i)) =
               (\<forall> j.  j\<le> ilen l \<longrightarrow> 
                index_sequence (ifirst (inth (x1a \<odot> l) (1+j))) (inth (x1a \<odot> l) (1+j))) "  
         by auto
           (simp add: Nitpick.case_nat_unfold)         
     have 14: "(\<forall> j.  j\<le> ilen l \<longrightarrow> 
                 index_sequence (ifirst (inth (x1a \<odot> l) (1+j))) (inth (x1a \<odot> l) (1+j))) =
               (\<forall> j.  j\<le> ilen l \<longrightarrow> 
                 index_sequence (ifirst (inth (l) (j))) (inth (l) (j))) "
         by simp
     show ?thesis 
     using "10" "12" "13" "91" by auto
   qed
qed



lemma lfuse_ilen_a:
 assumes "lastfirst xxs"
  shows  " ( \<forall>i. i \<le>ilen (xxs) \<longrightarrow>
              (\<forall> j\<le>ilen (inth (xxs) (i)) . j\<le> ilen(lfuse xxs)) )"
using assms
 proof (induct xxs)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xxs)
 then show ?case 
 proof -
  have 0: "ilast x1a = ifirst(ifirst xxs)"
     using ICons.prems lastfirst.simps(2) by blast
  have 1: "(\<forall>i. i \<le> ilen (x1a \<odot> xxs) \<longrightarrow> 
            (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) i). j \<le> ilen (lfuse (x1a \<odot> xxs))))
           =
           ( (\<forall>j\<le>ilen ( inth (x1a \<odot> xxs) 0). j \<le> ilen (lfuse (x1a \<odot> xxs))) \<and>
             (\<forall>i. 1 \<le> i \<and> i-1 \<le> ilen (xxs) \<longrightarrow> 
              (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) i). j \<le> ilen (lfuse (x1a \<odot> xxs)))) )"
       by auto
          (metis One_nat_def add.commute le_diff_conv le_zero_eq not_less_eq_eq old.nat.simps(4) 
           plus_1_eq_Suc)
 have 2: "(\<forall>j\<le>ilen ( inth (x1a \<odot> xxs) 0). j \<le> ilen (lfuse (x1a \<odot> xxs))) =
          (\<forall>j\<le>ilen (x1a). j \<le> ilen (fuse x1a (lfuse xxs)))"
        by simp   
 have 3: "(\<forall>j\<le>ilen (x1a). j \<le> ilen (fuse x1a (lfuse xxs))) =
          (\<forall>j\<le>ilen (x1a). j \<le> ilen (x1a) + ilen (lfuse xxs))  "
    by (simp add: fuse_ilen_a)
 have 4: "(\<forall>j\<le>ilen (x1a). j \<le> ilen (x1a) + ilen (lfuse xxs)) "
    by linarith 
 have 5: "(\<forall>i. 1 \<le> i \<and> i-1 \<le> ilen (xxs) \<longrightarrow> 
           (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) i). j \<le> ilen (lfuse (x1a \<odot> xxs))))
          = 
          (\<forall>i. 0 \<le> i \<and> i \<le> ilen (xxs) \<longrightarrow> 
            (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) (Suc i)). j \<le> ilen (lfuse (x1a \<odot> xxs))))"
       by (metis add_diff_cancel_left' ilen_gr_zero prefix_ilen_good le_add1 
          ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc)
 have 6: "(\<forall>i. 0 \<le> i \<and> i \<le> ilen (xxs) \<longrightarrow> 
           (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) (Suc i)). j \<le> ilen (lfuse (x1a \<odot> xxs)))) =
          (\<forall>i. 0 \<le> i \<and> i \<le> ilen (xxs) \<longrightarrow>
           (\<forall>j\<le>ilen (inth ( xxs) (i)). j \<le> ilen x1a + ilen (lfuse ( xxs))))" 
         by (simp add: fuse_ilen_a)     
 have 7: "\<forall>i. 0 \<le> i \<and> i \<le> ilen xxs \<longrightarrow> (\<forall>j\<le>ilen ( inth xxs i). j \<le> ilen (lfuse xxs)) "
     using ICons.hyps ICons.prems lastfirst.simps(2) by blast
 have 8: "(\<forall>i. 0 \<le> i \<and> i \<le> ilen (xxs) \<longrightarrow> 
           (\<forall>j\<le>ilen (inth ( xxs) (i)). j \<le> ilen x1a + ilen (lfuse ( xxs)))) " 
          by (simp add: "7" trans_le_add2)
 show ?thesis using "1" "3" "5" "6" "8" by auto
 qed
qed


lemma lfuse_split:
 assumes "lastfirst xxs \<and> (\<forall> j \<le> ilen (xxs). ilen(inth (xxs) j) > 0)"
 shows   "(\<forall>i\<le> ilen(xxs).
               (\<forall>ia<ilen ((inth (xxs) i)).
                 f (sub (inth (inth (xxs) i) ia)
                        (inth (inth (xxs) i) (Suc ia))
                        \<sigma>))) =
                  (\<forall>j<ilen (lfuse (xxs)).
                 f (sub (inth (lfuse (xxs)) j) 
                        (inth (lfuse (xxs)) (Suc j)) 
                        \<sigma>))"
using assms
proof (induct xxs)
case (INil x)
then show ?case by auto
next
case (ICons x1a xxs)
then show ?case 
 proof -
  have 1: "(\<forall>i\<le>ilen (x1a \<odot> xxs).
             \<forall>ia<ilen (inth (x1a \<odot> xxs) i).
              f (sub (inth (inth (x1a \<odot> xxs) i) ia) (inth (inth (x1a \<odot> xxs) i) (Suc ia)) \<sigma>)) =
           (\<forall>i\<le>ilen (xxs)+1.
              \<forall>ia<ilen (inth (x1a \<odot> xxs) i).
               f (sub (inth (inth (x1a \<odot> xxs) i) ia) (inth (inth (x1a \<odot> xxs) i) (Suc ia)) \<sigma>))" 
     by simp
  have 2: "... =
           ( (\<forall>ia<ilen (inth (x1a \<odot> xxs) 0).
              f (sub (inth (inth (x1a \<odot> xxs) 0) ia) (inth (inth (x1a \<odot> xxs) 0) (Suc ia)) \<sigma>))
              \<and>
           (\<forall>i. 1 \<le> i \<and> i\<le>ilen (xxs)+1\<longrightarrow>
            (\<forall>ia<ilen (inth (x1a \<odot> xxs) i).
             f (sub (inth (inth (x1a \<odot> xxs) i) ia) (inth ( inth (x1a \<odot> xxs) i) (Suc ia)) \<sigma>))))"
       by (metis One_nat_def Suc_leI add_nonneg_nonneg gr_zeroI ilen_gr_zero zero_le_one)
  have 3: "(\<forall>ia<ilen (inth (x1a \<odot> xxs) 0).
              f (sub (inth (inth (x1a \<odot> xxs) 0) ia) (inth (inth (x1a \<odot> xxs) 0) (Suc ia)) \<sigma>)) =
           (\<forall>ia<ilen x1a.
              f (sub (inth (x1a) ia) (inth (x1a) (Suc ia)) \<sigma>))" 
     by simp
  have 4: "(\<forall>i. 1 \<le> i \<and> i\<le>ilen (xxs)+1\<longrightarrow>
             (\<forall>ia<ilen (inth (x1a \<odot> xxs) i).
              f (sub (inth (inth (x1a \<odot> xxs) i) ia) (inth (inth (x1a \<odot> xxs) i) (Suc ia)) \<sigma>))) =
           (\<forall>i. 0 \<le> i-1 \<and> i-1\<le>ilen (xxs)\<longrightarrow>
             (\<forall>ia<ilen (inth (x1a \<odot> xxs) ((i-1)+1)).
              f (sub (inth (inth (x1a \<odot> xxs) ((i-1)+1)) ia) 
                     (inth (inth (x1a \<odot> xxs) ((i-1)+1)) (Suc ia)) \<sigma>)))" 
       by (auto simp add: Nitpick.case_nat_unfold) 
  have 5: " ... =
            (\<forall>i. 0 \<le> i \<and> i\<le>ilen (xxs)\<longrightarrow>
             (\<forall>ia<ilen (inth (x1a \<odot> xxs) ((i)+1)).
              f (sub (inth (inth (x1a \<odot> xxs) ((i)+1)) ia) 
                     (inth (inth (x1a \<odot> xxs) ((i)+1)) (Suc ia)) \<sigma>)))" 
          using "4" by auto
  have 6: " ... =
            (\<forall>i. 0 \<le> i \<and> i\<le>ilen (xxs)\<longrightarrow>
             (\<forall>ia<ilen (inth (xxs) ((i))).
              f (sub (inth (inth (xxs) ((i))) ia) (inth (inth (xxs) ((i))) (Suc ia)) \<sigma>)))
          "
                by simp
  have 7: "lastfirst xxs"
    using ICons.prems lastfirst.simps(2) by blast 
  have 8: "ilast x1a = ifirst(ifirst xxs)"
    using ICons.prems lastfirst.simps(2) by blast
  have 9: "(\<forall> j \<le> ilen (x1a\<odot>xxs). ilen(inth (x1a\<odot>xxs) j) > 0)" 
    using ICons.prems by blast
  have 10: "ilen(inth (x1a\<odot>xxs) 0) > 0"
    using ICons.prems by blast 
  have 11: "(\<forall>j. 1 \<le> j \<and> j \<le> ilen (xxs)+1 \<longrightarrow> ilen(inth (x1a\<odot>xxs) j) > 0)" 
     using ICons.prems by auto
  have 12: "(\<forall>j. 0 \<le> j-1 \<and> j-1 \<le> ilen (xxs) \<longrightarrow> ilen(inth (x1a\<odot>xxs) ((j-1)+1)) > 0)" 
       using ICons.prems by auto
  have 13: "(\<forall>j. j \<le> ilen (xxs) \<longrightarrow> ilen(inth (x1a\<odot>xxs) ((j)+1)) > 0)"     
     using ICons.prems by auto
  have 14: "(\<forall>j.  j \<le> ilen (xxs) \<longrightarrow> ilen(inth (xxs) ((j))) > 0)" 
     using "13" by simp
  have 15: "(\<forall>i.  i\<le>ilen (xxs)\<longrightarrow>
             (\<forall>ia<ilen (inth (xxs) ((i))).
              f (sub (inth (inth (xxs) ((i))) ia) (inth (inth (xxs) ((i))) (Suc ia)) \<sigma>))) =
            (\<forall>j<ilen (lfuse (xxs)).
                 f (sub (inth (lfuse (xxs)) j) 
                        (inth (lfuse (xxs)) (Suc j)) 
                        \<sigma>))" 
        by (simp add: "14" "7" ICons.hyps)
  have 16: "(\<forall>j<ilen (lfuse (x1a \<odot> xxs)). 
              f (sub (inth (lfuse (x1a \<odot> xxs)) j) (inth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>)) =
            (\<forall>j<ilen (fuse x1a (lfuse (xxs))). 
              f (sub (inth (lfuse (x1a \<odot> xxs)) j) (inth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>))" 
             by simp
  have 17: " ... =
             (\<forall>j<ilen x1a + ilen ((lfuse (xxs))). 
              f (sub (inth (lfuse (x1a \<odot> xxs)) j) (inth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>))"
     by (simp add: fuse_ilen_a)
  have 18: " ... =
             ((\<forall>j<ilen x1a. 
              f (sub (inth (lfuse (x1a \<odot> xxs)) j) (inth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>)) \<and>
             (\<forall>j. ilen x1a \<le> j \<and> j<ilen x1a + ilen ((lfuse (xxs))) \<longrightarrow> 
              f (sub (inth (lfuse (x1a \<odot> xxs)) j) (inth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>))) " 
      using le_add1 less_le_trans not_less by blast
  have 19: "(\<forall>j<ilen x1a. 
              f (sub (inth (lfuse (x1a \<odot> xxs)) j) (inth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>)) =
            (\<forall>j<ilen x1a. 
              f (sub (inth (fuse x1a (lfuse (xxs))) j) (inth (fuse x1a (lfuse (xxs))) (Suc j)) \<sigma>))" 
          by simp
  have 20: " ... =
             (\<forall>j<ilen x1a. 
              f (sub (inth (x1a) j) (inth (x1a) (Suc j)) \<sigma>))" 
           proof -
           have "prefix (ilen x1a) (fuse x1a (lfuse xxs)) = x1a"
           by (metis (no_types) ICons.prems ifirst_lfuse_ifirst lastfirst.simps(2) prefix_fuse)
           then show ?thesis
           by (metis (no_types) Suc_leI inth_prefix le_simps(1))
           qed
  have 21: "(\<forall>j. ilen x1a \<le> j \<and> j<ilen x1a + ilen ((lfuse (xxs))) \<longrightarrow> 
              f (sub (inth (lfuse (x1a \<odot> xxs)) j) (inth (lfuse (x1a \<odot> xxs)) (Suc j)) \<sigma>)) =
            (\<forall>j. ilen x1a \<le> j \<and> j<ilen x1a + ilen ((lfuse (xxs))) \<longrightarrow> 
              f (sub (inth (fuse x1a (lfuse (xxs))) j) (inth (fuse x1a (lfuse (xxs))) (Suc j)) \<sigma>))"  
      by simp
  have 22: " ... =
             (\<forall>j. ilen x1a \<le> j \<and> j<ilen x1a + ilen ((lfuse (xxs))) \<longrightarrow> 
              f (sub (inth ((lfuse (xxs))) (j -ilen x1a)) 
                     (inth (fuse x1a (lfuse (xxs))) ((j)+1)) \<sigma>)) " (is "?L=?R")
           proof rule
            show "?L\<Longrightarrow>?R"
              by auto 
             (metis "8" ICons.prems fuse_ilen_a fuse_inth 
              ifirst_lfuse_ifirst less_imp_le_nat)
            show "?R \<Longrightarrow> ?L"
              by auto
             (metis "8" ICons.prems fuse_ilen_a fuse_inth ifirst_lfuse_ifirst less_imp_le_nat)
           qed
  have 23: " ... =
             (\<forall>j. ilen x1a \<le> j \<and> j<ilen x1a + ilen ((lfuse (xxs))) \<longrightarrow> 
              f (sub (inth ((lfuse (xxs))) (j -ilen x1a)) 
                     (inth ( (lfuse (xxs))) (((Suc j)-ilen x1a))) \<sigma>))" (is "?L=?R")
           proof 
            show "?L\<Longrightarrow>?R"
             by auto
                (metis "8" ICons.prems Suc_leI fuse_ilen_a fuse_inth ifirst_lfuse_ifirst le_SucI)
            show "?R \<Longrightarrow>?L"
             by auto
                (metis "8" ICons.prems Suc_leI fuse_ilen_a fuse_inth ifirst_lfuse_ifirst le_SucI)         
           qed
  have 24: " ... =
             (\<forall>j. 0 \<le> j  \<and> j < ilen ((lfuse (xxs))) \<longrightarrow> 
              f (sub (inth ((lfuse (xxs))) (j) ) (inth ( (lfuse (xxs))) (((Suc j) ))) \<sigma>))"
           by (rule interval_shift_index_to_zero_b)
  show ?thesis 
  using "15" "17" "18" "2" "20" "22" "23" "24" "4" "5" by auto
 qed
qed


end