(*  Title     : A Shallow Encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Interval Temporal  Algebra\<close>  

theory ITA
imports Fuse Semantics TimeReversal
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="
  cvc4 z3 e spass vampire
"]
*)
subsection \<open>Definition of Set of intervals and Operations on them\<close>

type_synonym 'a intervals = "'a interval set "

definition lan:: "('a:: world) formula \<Rightarrow> 'a intervals"
where "lan f = { \<sigma> . (\<sigma> \<Turnstile> f) }" 
 
definition fusion :: "'a intervals \<Rightarrow> 'a intervals \<Rightarrow> 'a intervals" (infixl "\<cdot>" 70)
where "X\<cdot>Y = {fuse \<sigma>1 \<sigma>2| \<sigma>1 \<sigma>2. \<sigma>1 \<in> X \<and> \<sigma>2 \<in> Y \<and> ilast \<sigma>1 = ifirst \<sigma>2}"

definition reverse :: "'a intervals \<Rightarrow> 'a intervals" ("(SRev _)" [85] 85)
where "(SRev X) = {irev \<sigma> |\<sigma>. \<sigma> \<in> X } "

definition sempty :: "'a intervals" ("SEmpty") 
where
  "SEmpty \<equiv> range INil"

definition smore :: "'a intervals" ("SMore")
where
  "SMore \<equiv>  - SEmpty"

definition sskip :: "'a intervals" ("SSkip")
where
  "SSkip \<equiv> -(SEmpty \<union> (SMore\<cdot>SMore))"

definition sfalse :: "'a intervals" ("SFalse")
where
  "SFalse \<equiv> {}" 

definition strue :: "'a intervals" ("STrue")
where
  "STrue \<equiv> -{}" 

definition sinit :: "'a intervals \<Rightarrow> 'a intervals" ("(SInit _)" [85] 85)
where
  "SInit X \<equiv> (X \<inter> SEmpty)\<cdot>STrue"

definition sfin :: "'a intervals \<Rightarrow> 'a intervals" ("(SFin _)" [85] 85)
where
  "SFin X \<equiv> STrue\<cdot>(X \<inter> SEmpty)"

definition ssometime :: "'a intervals \<Rightarrow> 'a intervals" ("(SSometime _)" [85] 85)
where
  "SSometime X \<equiv> STrue\<cdot>X"

definition salways :: "'a intervals \<Rightarrow> 'a intervals" ("(SAlways _)" [85] 85)
where
  "SAlways X \<equiv> -(SSometime (-X))"

definition sdi :: "'a intervals \<Rightarrow> 'a intervals" ("(SDi _)" [85] 85)
where
  "SDi X \<equiv> X\<cdot>STrue"

definition sbi :: "'a intervals \<Rightarrow> 'a intervals" ("(SBi _)" [85] 85)
where
  "SBi X \<equiv> -(SDi (-X))"

definition sda :: "'a intervals \<Rightarrow> 'a intervals" ("(SDa _)" [85] 85)
where
  "SDa X \<equiv> STrue\<cdot>X\<cdot>STrue"

definition sba :: "'a intervals \<Rightarrow> 'a intervals" ("(SBa _)" [85] 85)
where
  "SBa X \<equiv> -(SDa (-X))"

definition snext :: "'a intervals \<Rightarrow> 'a intervals" ("(SNext _)" [85] 85)
where 
  "SNext X \<equiv> SSkip\<cdot>X" 

definition swnext :: "'a intervals \<Rightarrow> 'a intervals" ("(SWnext _)" [85] 85)
where 
  "SWnext X \<equiv> (-(SSkip\<cdot>(-X)))" 
 
definition sprev :: "'a intervals \<Rightarrow> 'a intervals" ("(SPrev _)" [85] 85)
where 
  "SPrev X \<equiv> X\<cdot>SSkip" 

definition swprev :: "'a intervals \<Rightarrow> 'a intervals" ("(SWprev _)" [85] 85)
where 
  "SWprev X \<equiv> (-((-X)\<cdot>SSkip))"
 
primrec spower :: "'a intervals \<Rightarrow> nat \<Rightarrow> 'a intervals" ("(SPower _ _)" [88,88] 87)
where
   pwr_0  : "SPower X 0       = SEmpty"
 | pwr_Suc: "SPower X (Suc n) = ((X \<inter> SMore)\<cdot>(SPower X n))"

definition sstar :: "'a intervals \<Rightarrow> 'a intervals" ("(SStar _)" [85] 85)
where
  "SStar X \<equiv> (\<Union> n. SPower X n)"


subsection \<open>Simplification Lemmas\<close>

lemma snot_elim :
 "x \<in> -X \<longleftrightarrow> x \<notin> X "
by simp

lemma sor_elim :
 "x\<in> (X \<union> Y) \<longleftrightarrow> (x \<in> X \<or> x\<in>Y) "
by simp

lemma sand_elim :
  "x\<in> (X \<inter> Y) \<longleftrightarrow> (x \<in> X \<and> x\<in>Y) "
by simp

lemma sfalse_elim :
 " \<sigma> \<notin> SFalse"
by (simp add: sfalse_def)

lemma strue_elim :
 " \<sigma> \<in> STrue "
by (simp add: strue_def)

lemma sempty_elim :
    " \<sigma> \<in> SEmpty \<longleftrightarrow> ilen \<sigma> = 0"
by (simp add: image_iff INil_ilen sempty_def)

lemma smore_elim :
    " \<sigma> \<in> SMore \<longleftrightarrow> ilen \<sigma> > 0"
by (simp add: sempty_elim smore_def)

lemma fusion_iff:
    "\<sigma> \<in> X\<cdot>Y \<longleftrightarrow> (\<exists>\<sigma>1 \<sigma>2. \<sigma> = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in> X \<and> \<sigma>2 \<in> Y \<and> ilast \<sigma>1 = ifirst \<sigma>2)"
by (unfold fusion_def) auto

lemma fusion_iff_1:
    "\<sigma> \<in> X\<cdot>Y \<longleftrightarrow> (\<exists> i\<le> ilen \<sigma>. (prefix i \<sigma>)\<in>X \<and> (suffix i \<sigma>)\<in>Y )"
using fusion_iff[of "\<sigma>" "X" "Y"] 
by (metis ilast_ifirst fuse_ilen fuse_prefix_suffix prefix_fuse suffix_fuse le_add1) 

lemma smore_fusion_smore :
 " \<sigma> \<in> (SMore\<cdot>SMore) \<longleftrightarrow> ilen \<sigma> > 1"
using fusion_iff_1 
by (metis prefix_ilen_good suffix_ilen_good less_one 
          not_less not_less_iff_gr_or_eq smore_elim zero_less_diff)

lemma sskip_elim :
   " \<sigma> \<in> SSkip \<longleftrightarrow> ilen \<sigma> = 1"
using sskip_def smore_fusion_smore 
by (metis One_nat_def Suc_lessI Un_iff less_numeral_extra(4) sempty_elim smore_def smore_elim 
          snot_elim zero_neq_one)
  
lemma spower_elim_zero :
    " \<sigma> \<in> SPower X 0 \<longleftrightarrow> \<sigma> \<in> SEmpty  "
by simp

lemma spower_elim_suc :
    " \<sigma> \<in> SPower X (Suc n) \<longleftrightarrow> \<sigma> \<in> (X \<inter> SMore)\<cdot>(SPower X n)"
by simp
  
lemma spower_elim_suc_1 :
    " \<sigma> \<in> (X \<inter> SMore)\<cdot>(SPower X n) \<longleftrightarrow>
     (\<exists>\<sigma>1 \<sigma>2. \<sigma> = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in> X \<and> ilen \<sigma>1 >0 \<and> \<sigma>2 \<in> (SPower X n) \<and> 
           ilast \<sigma>1 = ifirst \<sigma>2)"
by (meson IntD1 IntD2 IntI smore_elim fusion_iff)        
 
lemma sstar_elim :
    " \<sigma> \<in> SStar X \<longleftrightarrow> (\<exists> n. \<sigma> \<in> SPower X n)"
by (simp add: sstar_def)
   
lemma sstar_elim_1 :
    " (\<exists> n . \<sigma> \<in> SPower X n) \<longleftrightarrow> 
     ( \<sigma> \<in> SPower X 0 \<or> (\<exists> n. \<sigma> \<in> SPower X (Suc n))) "
by (metis not0_implies_Suc)

lemma spower_suc :
    "(\<exists> n. \<sigma> \<in> SPower X (Suc n)) \<longleftrightarrow> 
     (\<exists> n. \<sigma> \<in> (X \<inter> SMore)\<cdot>(SPower X n)) "
by simp

lemma spower_suc_1 :
    "(\<exists> n. \<sigma> \<in> (X \<inter> SMore)\<cdot>(SPower X n)) \<longleftrightarrow>
    \<sigma> \<in> (X \<inter> SMore)\<cdot>(SStar X)"
by (metis fusion_iff sstar_elim)
    
lemma sstar_eqv :
    " \<sigma> \<in> SStar X \<longleftrightarrow> 
      ( \<sigma> \<in> SEmpty \<or> \<sigma> \<in> (X \<inter> SMore)\<cdot>(SStar X)) "
by (metis spower.simps(1) spower_elim_suc spower_suc_1 sstar_elim sstar_elim_1)
 
lemma spower_sskip_elim :
 "(\<sigma> \<in> SPower SSkip n) \<longleftrightarrow> ilen \<sigma> = n "
proof (induct n arbitrary: \<sigma>)
case 0
then show ?case by (auto simp add: sempty_elim)
next
case (Suc n)
then show ?case 
   proof auto
   show "(\<And>\<sigma>. (\<sigma> \<in> SPower SSkip n) = (ilen \<sigma> = n)) \<Longrightarrow> 
           \<sigma> \<in> SSkip \<inter> SMore \<cdot> SPower SSkip n \<Longrightarrow> ilen \<sigma> = Suc n"
     by (metis Suc.hyps fuse_ilen plus_1_eq_Suc spower_elim_suc_1 sskip_elim) 
   show "(\<And>\<sigma>. (\<sigma> \<in> SPower SSkip n) = (ilen \<sigma> = n)) \<Longrightarrow> 
           ilen \<sigma> = Suc n \<Longrightarrow> \<sigma> \<in> SSkip \<inter> SMore \<cdot> SPower SSkip n " 
   by (metis Suc.hyps Compl_Un Int_commute add_diff_cancel_left' fusion_iff_1 inf_sup_aci(4) 
       prefix_ilen_good suffix_ilen_good le_add1 plus_1_eq_Suc smore_def 
       sskip_def sskip_elim)
   qed
qed



lemma srev_elim:
 "\<sigma> \<in> (SRev X) \<longleftrightarrow> irev \<sigma> \<in> X"
using irev_swap by (auto simp add: reverse_def)

subsection \<open>Algebraic Laws\<close>

subsubsection \<open>Commutative Additive Monoid\<close>

lemma UnionCommute:
 " (X::'a intervals) \<union> Y = Y \<union> X"
by (simp add: Un_commute)

lemma UnionSFalse:
 " X \<union> SFalse = X"
by (simp add: sfalse_def)

lemma UnionAssoc:
 " (X::'a intervals) \<union> (Y \<union> Z) = (X \<union> Y) \<union> Z"
by (simp add: sup_assoc) 

subsubsection \<open> Boolean algebra\<close>

lemma Huntington:
  " (X::'a intervals) = -(-X \<union> -Y) \<union> -(-X \<union> Y) "
by auto

lemma Morgan:
 " (X::'a intervals) \<inter> Y = -(-X \<union> -Y)"
by auto

\<comment> \<open>identities\<close> 

lemma STrueTop:
 " STrue = X \<union> -X"
by (simp add: strue_def)

lemma SFalseBottom:
 " SFalse = X \<inter> -X"
by (simp add: sfalse_def)

subsubsection \<open>multiplicative monoid\<close>

lemma FusionSEmptyL : 
    "SEmpty \<cdot> X = X"
using fusion_iff_1 set_eqI[of "SEmpty\<cdot>X" "X"] 
by (metis ilen_gr_zero prefix_ilen_good suffix_zero sempty_elim)

lemma FusionSEmptyR : 
    "X \<cdot> SEmpty  = X"
using fusion_iff_1 set_eqI[of "X\<cdot>SEmpty" "X"] 
proof auto
 show "\<And>x. (\<And>\<sigma> X Y. (\<sigma> \<in> X \<cdot> Y) = (\<exists>i\<le>ilen \<sigma>. prefix i \<sigma> \<in> X \<and> suffix i \<sigma> \<in> Y)) \<Longrightarrow>
         ((\<And>x. (x \<in> X \<cdot> SEmpty) = (x \<in> X)) \<Longrightarrow> X \<cdot> SEmpty = X) \<Longrightarrow> x \<in> X \<cdot> SEmpty \<Longrightarrow> x \<in> X" 
 by (metis diff_diff_cancel diff_zero fusion_iff_1 prefix_ilen 
       suffix_ilen_good sempty_elim)
  show " \<And>x. (\<And>\<sigma> X Y. (\<sigma> \<in> X \<cdot> Y) = (\<exists>i\<le>ilen \<sigma>. prefix i \<sigma> \<in> X \<and> suffix i \<sigma> \<in> Y)) \<Longrightarrow>
         ((\<And>x. (x \<in> X \<cdot> SEmpty) = (x \<in> X)) \<Longrightarrow> X \<cdot> SEmpty = X) \<Longrightarrow> x \<in> X \<Longrightarrow> x \<in> X \<cdot> SEmpty " 
   using sempty_elim fusion_iff_1 by fastforce
qed

lemma FusionAssocA:
assumes "x \<in> X \<cdot> (Y \<cdot> Z)"
shows    "x \<in> (X \<cdot> Y) \<cdot> Z"
proof -
 have 1: "(\<exists>\<sigma>1 \<sigma>2. x = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in> X \<and> \<sigma>2 \<in> Y \<cdot> Z \<and> ilast \<sigma>1 = ifirst \<sigma>2)" 
  using assms fusion_iff[of "x" "X" "Y \<cdot> Z"] by auto
 obtain "\<sigma>1" "\<sigma>2" where 2: "x = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in> X \<and> \<sigma>2 \<in> Y \<cdot> Z \<and> ilast \<sigma>1 = ifirst \<sigma>2" 
   using 1 by auto
 have 3: "(\<exists>\<sigma>3 \<sigma>4. \<sigma>2 = fuse \<sigma>3 \<sigma>4 \<and> \<sigma>3 \<in> Y \<and> \<sigma>4 \<in> Z \<and> ilast \<sigma>3 = ifirst \<sigma>4)" 
  using 2 fusion_iff[of "\<sigma>2" "Y" "Z"]  by auto
 obtain "\<sigma>3" "\<sigma>4" where 4: " \<sigma>2 = fuse \<sigma>3 \<sigma>4 \<and> \<sigma>3 \<in> Y \<and> \<sigma>4 \<in> Z \<and> ilast \<sigma>3 = ifirst \<sigma>4" 
   using "3" by auto
 have 5: "x = fuse \<sigma>1 (fuse \<sigma>3 \<sigma>4)" 
   using "2" "4" by auto
 have 6: "x= fuse (fuse \<sigma>1 \<sigma>3) \<sigma>4"
   using "5"  "2" "4" FusionAssoc ifirst_fuse by fastforce
 show ?thesis 
   by (metis "2" "4" "6" fusion_iff ifirst_fuse ilast_fuse)
qed

lemma FusionAssocB:
assumes "x \<in> (X \<cdot> Y) \<cdot> Z"
shows    "x \<in> X \<cdot> (Y \<cdot> Z)"
proof - 
 have 1: "(\<exists>\<sigma>1 \<sigma>2. x = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in> X\<cdot>Y \<and> \<sigma>2 \<in> Z \<and> ilast \<sigma>1 = ifirst \<sigma>2)" 
   using assms fusion_iff[of "x" "X\<cdot>Y" "Z"] by auto
 obtain "\<sigma>1" "\<sigma>2" where 2: "x = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in> X\<cdot>Y \<and> \<sigma>2 \<in> Z \<and> ilast \<sigma>1 = ifirst \<sigma>2" 
  using 1 by auto
 have 3: "(\<exists>\<sigma>3 \<sigma>4. \<sigma>1 = fuse \<sigma>3 \<sigma>4 \<and> \<sigma>3 \<in> X \<and> \<sigma>4 \<in> Y \<and> ilast \<sigma>3 = ifirst \<sigma>4)" 
  using 2 fusion_iff[of "\<sigma>1" "X" "Y"] by auto
 obtain "\<sigma>3" "\<sigma>4" where 4: " \<sigma>1 = fuse \<sigma>3 \<sigma>4 \<and> \<sigma>3 \<in> X \<and> \<sigma>4 \<in> Y \<and> ilast \<sigma>3 = ifirst \<sigma>4" 
   using "3" by auto
 have 5: "x = fuse (fuse \<sigma>3 \<sigma>4) \<sigma>2" 
    using "2" "4" by auto
 have 6: "x = fuse \<sigma>3 (fuse \<sigma>4 \<sigma>2)"
   using "2" "4" FusionAssoc ilast_fuse by force
 show ?thesis 
 by (metis "2" "4" "6" fusion_iff ifirst_fuse ilast_fuse)
qed

lemma FusionAssoc :
 " X\<cdot>(Y\<cdot>Z) = (X\<cdot>Y)\<cdot>Z"
using set_eqI[of "X\<cdot>(Y\<cdot>Z)" "(X\<cdot>Y)\<cdot>Z"]
FusionAssocA FusionAssocB  by blast

\<comment> \<open>left and right distributivity\<close>

lemma FusionUnionDistL:
 " (X \<union> Y)\<cdot>Z = (X\<cdot>Z) \<union> (Y\<cdot>Z)"
using fusion_iff[of _ "X \<union> Y" "Z"] fusion_iff[of _ "X" "Z"]  fusion_iff[of _ "Y" "Z"]
    set_eqI[of "(X \<union> Y)\<cdot>Z " "(X\<cdot>Z) \<union> (Y\<cdot>Z)"] 
    by auto


lemma FusionUnionDistR:
 " X\<cdot>(Y \<union> Z) = (X\<cdot>Y) \<union> (X\<cdot>Z)"
using fusion_iff set_eqI[of "X\<cdot>(Y \<union> Z)" "(X\<cdot>Y) \<union> (X\<cdot>Z)"] 
fusion_iff[of _ "X" "Y \<union> Z"] fusion_iff[of _ "X" "Y"]  fusion_iff[of _ "X" "Z"]
by auto


\<comment> \<open>left and right annihilation\<close>
lemma SFalseFusion:
  "SFalse \<cdot> X = SFalse"
by (simp add: fusion_def sfalse_def)

lemma FusionSFalse:
 "X\<cdot>SFalse = SFalse"
by (simp add: fusion_def sfalse_def)

\<comment> \<open>idempotency\<close>
lemma UnionIdem:
 " (X::'a intervals) \<union> X = X"
by simp

subsubsection \<open>Subsumption order\<close>

lemma Subsumption:
 "((X :: 'a intervals) \<subseteq> Y) = (X \<union> Y = Y)"
by auto

subsubsection \<open>Helper lemmas\<close>

lemma FusionRuleR:
 assumes "X \<subseteq> Y"
 shows   " Z \<cdot> X \<subseteq> Z \<cdot> Y"
using assms FusionUnionDistR by (metis Subsumption)

lemma FusionRuleL:
 assumes "X \<subseteq> Y"
 shows   " X \<cdot> Z \<subseteq> Y \<cdot> Z"
using assms by (metis FusionUnionDistL subset_Un_eq)

lemma spower_commutes:
 "(X \<inter> SMore)\<cdot>(SPower X n) = (SPower X n)\<cdot>(X \<inter> SMore)"
proof (induct n)
case 0
then show ?case by (simp add: FusionSEmptyL FusionSEmptyR)
next
case (Suc n)
then show ?case by (simp add: FusionAssoc)
qed

lemma fusion_inductl:
 assumes "Y \<union> X\<cdot>Z \<subseteq> Z "
 shows   "(SPower X n)\<cdot>Y \<subseteq> Z"
using assms 
proof (induct n)
case 0
then show ?case by (simp add: FusionSEmptyL)
next
case (Suc n)
then show ?case 
proof -
have f1: "X \<cdot> (SPower X n \<cdot> Y) \<union> X \<cdot> Z = X \<cdot> Z"
by (metis FusionUnionDistR Suc.hyps assms subset_Un_eq) (* 8 ms *)
have "X \<cdot> SPower X n \<cdot> Y \<union> X \<inter> SMore \<cdot> SPower X n \<cdot> Y = X \<cdot> (SPower X n \<cdot> Y)"
by (metis (no_types) FusionAssoc FusionUnionDistL sup_inf_absorb) (* 20 ms *)
then have "SPower X (Suc n) \<cdot> Y \<union> Z = X \<cdot> Z \<union> (Y \<union> Z)"
using f1 assms by auto (* 5 ms *)
then show ?thesis
using assms by auto (* 2 ms *)
qed
qed

lemma fusion_inductr:
assumes "Y \<union> Z\<cdot>X \<subseteq> Z "
 shows   "Y\<cdot>(SPower X n) \<subseteq> Z"
using assms 
proof (induct n)
case 0
then show ?case by (simp add: FusionSEmptyR)
next
case (Suc n)
then show ?case 
proof -
have f1: "Y \<cdot> SPower X n \<union> Z = Z"
using Suc.hyps assms by blast (* 0.7 ms *)
have "Y \<cdot> (X \<inter> SMore) \<cdot> SPower X n = Y \<cdot> (SPower X n \<cdot> (X \<inter> SMore))"
by (metis (no_types) FusionAssoc spower_commutes) (* 10 ms *)
then have "Y \<cdot> (X \<inter> SMore) \<cdot> SPower X n \<subseteq> Z"
using f1 by (metis (no_types) FusionAssoc FusionUnionDistL FusionUnionDistR Un_subset_iff 
   assms sup_inf_absorb) (* 37 ms *)
then show ?thesis
by (simp add: FusionAssoc) (* 0.5 ms *)
qed
qed

lemma sstar_contlA: 
 assumes " x \<in> Y \<cdot> (SStar X)"
 shows   "x \<in>   (\<Union>n. Y \<cdot> (SPower X n))"
proof -
 have 1: "(\<exists>\<sigma>1 \<sigma>2. x = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in> Y \<and> \<sigma>2 \<in> (SStar X) \<and> ilast \<sigma>1 = ifirst \<sigma>2)" 
   using assms by (simp add: fusion_iff)
 obtain "\<sigma>1" "\<sigma>2" where 2: "x = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in> Y \<and> \<sigma>2 \<in> (SStar X) \<and> ilast \<sigma>1 = ifirst \<sigma>2"
  using "1" by auto
 have 3: "(\<exists> n. \<sigma>2 \<in> SPower X n) "  
  using "2" sstar_elim by blast
 obtain n where 4: "\<sigma>2 \<in> SPower X n" 
   using 3 by auto
 have 5: "(\<exists> n. x \<in> Y\<cdot>SPower X n)"
    using "2" "4" fusion_iff by blast
 from 5 show ?thesis by blast
qed

lemma sstar_contlB: 
 assumes " x \<in>   (\<Union>n. Y \<cdot> (SPower X n))"
 shows   "x \<in> Y \<cdot> (SStar X)"
proof -   
 have 1: "\<exists>n. x \<in> Y\<cdot> (SPower X n)"
   using assms by blast
 obtain n where 2: "x \<in> Y\<cdot> (SPower X n)"
  using 1 by auto
 have 3: "(\<exists>\<sigma>1 \<sigma>2. x = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in> Y \<and> \<sigma>2 \<in> (SPower X n) \<and> ilast \<sigma>1 = ifirst \<sigma>2)"  
  using "2" by (simp add: fusion_iff)
 obtain "\<sigma>1" "\<sigma>2" where 4: "x = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in> Y \<and> \<sigma>2 \<in> (SPower X n) \<and> ilast \<sigma>1 = ifirst \<sigma>2" 
   using 3 by auto
 have 5: "\<sigma>2\<in> (SStar X)"
   using "4" sstar_elim by auto 
 from 5 4 show ?thesis using fusion_iff by blast
qed

lemma sstar_contl: 
  " Y \<cdot> (SStar X) = (\<Union>n. Y \<cdot> (SPower X n))"
using set_eqI[of "Y \<cdot> (SStar X)" "(\<Union>n. Y \<cdot> (SPower X n))"] 
by (metis sstar_contlA sstar_contlB)


lemma sstar_contrA: 
assumes  " x\<in> (SStar X)\<cdot>Y"
 shows   " x\<in> (\<Union>n. (SPower X n)\<cdot>Y)"
proof -
 have 1: "(\<exists>\<sigma>1 \<sigma>2. x = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in>(SStar X)  \<and> \<sigma>2 \<in> Y \<and> ilast \<sigma>1 = ifirst \<sigma>2)" 
   using assms by (simp add: fusion_iff)
 obtain "\<sigma>1" "\<sigma>2" where 2: "x = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in>(SStar X)  \<and> \<sigma>2 \<in> Y \<and> ilast \<sigma>1 = ifirst \<sigma>2"
  using 1 by auto
 have 3: "\<exists>n. \<sigma>1 \<in>  (SPower X n)"  
   using 2 sstar_elim by blast
 obtain n where 4: "\<sigma>1 \<in>  (SPower X n)"
   using 3 by auto
 have 5: "(\<exists> n. x \<in> (SPower X n)\<cdot>Y)"
  by (metis "2" "4" fusion_iff_1 fuse_ilen prefix_fuse suffix_fuse le_add1)
  from 5 show ?thesis by blast
qed

lemma sstar_contrB: 
assumes  " x\<in> (\<Union>n. (SPower X n)\<cdot>Y)"
 shows   " x\<in> (SStar X)\<cdot>Y"  
proof -
 have 1: "\<exists>n. x \<in>  (SPower X n)\<cdot>Y"
   using assms by blast
 obtain n where 2: "x \<in> (SPower X n)\<cdot>Y"
  using 1 by auto
 have 3: "(\<exists>\<sigma>1 \<sigma>2. x = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in> (SPower X n) \<and> \<sigma>2 \<in>Y  \<and> ilast \<sigma>1 = ifirst \<sigma>2)"  
  using "2" by (simp add: fusion_iff)
 obtain "\<sigma>1" "\<sigma>2" where 4: "x = fuse \<sigma>1 \<sigma>2 \<and> \<sigma>1 \<in> (SPower X n) \<and> \<sigma>2 \<in> Y \<and> ilast \<sigma>1 = ifirst \<sigma>2" 
   using 3 by auto
 have 5: "\<sigma>1\<in> (SStar X)"
   using "4" sstar_elim by auto 
 from 5 4 show ?thesis using fusion_iff by blast
qed


lemma sstar_contr: 
  " (SStar X)\<cdot>Y = (\<Union>n. (SPower X n)\<cdot>Y)"
using set_eqI[of "(SStar X)\<cdot>Y" "(\<Union>n. (SPower X n)\<cdot>Y)"]
by (metis sstar_contrA sstar_contrB)


subsubsection \<open>Kleene Algebra\<close>

\<comment> \<open>left unfold\<close>
lemma UnfoldL:
 "SEmpty \<union> X\<cdot>(SStar X) = (SStar X)"
proof -
 have 1: "(SStar X) = SEmpty \<union> (X \<inter> SMore)\<cdot>(SStar X) " 
      by (meson Un_iff set_eqI sstar_eqv)
 have 2: " (X \<inter> SMore)\<cdot>(SStar X) \<subseteq> X\<cdot>(SStar X)" 
      by (simp add: FusionRuleL)
 have 3: " (SStar X) \<subseteq> SEmpty \<union> X\<cdot>(SStar X)" 
      using "1" "2" by blast
 have 4: " SEmpty \<subseteq> (SStar X)" 
      using "1" by auto
 have 5: " X \<subseteq> SEmpty \<union> (X \<inter> SMore)" 
      by (simp add: Un_Int_distrib smore_def)
 have 6: " X\<cdot>(SStar X) \<subseteq> (SStar X) \<union> (X \<inter> SMore)\<cdot>(SStar X)" 
      using "5" by (metis FusionRuleL FusionUnionDistL FusionSEmptyL)
 have 7: " (SStar X) \<subseteq> SEmpty \<union> (X \<inter> SMore)\<cdot>(SStar X)"
      using "1" by auto
 have 8: "X\<cdot>(SStar X) \<subseteq> SEmpty \<union> (X \<inter> SMore)\<cdot>(SStar X) "
      using "6" "7" by blast
 hence 9: " X\<cdot>(SStar X) \<subseteq> (SStar X)"
      using "1" by auto
 have 10: "SEmpty \<union> X\<cdot>(SStar X) \<subseteq> (SStar X) "
      using "9" "4" by simp
 from 3 10 show ?thesis by auto
qed


\<comment> \<open>Left induction\<close>
lemma SStarInductL:
 assumes "Y \<union> X\<cdot>Z \<subseteq> Z"
 shows   "(SStar X)\<cdot>Y \<subseteq> Z" 
by (metis UN_least assms fusion_inductl sstar_contr)

\<comment> \<open>Right induction\<close>
lemma SStarInductR:
 assumes "Y \<union> Z\<cdot>X \<subseteq> Z "
 shows   "Y\<cdot>(SStar X) \<subseteq> Z"
using sstar_contl assms fusion_inductr by blast



subsubsection \<open>ITL specific Laws\<close>

lemma PwrFusionInterL:
 "((((SPower SSkip n) \<inter> X)\<cdot>V) \<inter> (((SPower SSkip n) \<inter> Y)\<cdot>W)) = 
  (((SPower SSkip n) \<inter> X \<inter> Y)\<cdot>(V \<inter> W)) "
using set_eqI[of "((((SPower SSkip n) \<inter> X)\<cdot>V) \<inter> (((SPower SSkip n) \<inter> Y)\<cdot>W)) "
                 "(((SPower SSkip n) \<inter> X \<inter> Y)\<cdot>(V \<inter> W)) "]
by (simp add: fusion_iff_1 spower_sskip_elim)
   (metis min.absorb1)

lemma PwrFusionInterR:
 "((V\<cdot>((SPower SSkip n) \<inter> X)) \<inter> ((W\<cdot>((SPower SSkip n) \<inter> Y)))) = 
   ((V \<inter> W)\<cdot>((SPower SSkip n) \<inter> X \<inter> Y)) "
using set_eqI[of "((V\<cdot>((SPower SSkip n) \<inter> X)) \<inter> ((W\<cdot>((SPower SSkip n) \<inter> Y)))) "
                 "((V \<inter> W)\<cdot>((SPower SSkip n) \<inter> X \<inter> Y)) "]
by (simp add: fusion_iff_1 spower_sskip_elim)
   (metis diff_diff_cancel)

lemma SSkipFusionImpSMore:
 " SSkip\<cdot>STrue \<subseteq> SMore"
using subsetI[of "SSkip\<cdot>STrue" "SMore"]  
by (auto simp add: fusion_iff_1 sskip_elim smore_elim strue_elim)

      
lemma SStarSkip:
 " (SStar SSkip) = STrue"
using set_eqI[of "(SStar SSkip)" "STrue"]
by (simp add: strue_def spower_sskip_elim sstar_elim) 


(* 
 sledgehammer (del: sskip_elim fusion_iff fusion_iff_1 smore_fusion_smore spower_commutes 
               fusion_inductl fusion_inductr sstar_contl sstar_contr )
 
*)  

subsection \<open>Derived Laws\<close>


subsubsection \<open>Helper Lemmas\<close>

lemma B01:
 assumes "(X:: 'a intervals) \<subseteq> Y"
 shows   " -Y \<subseteq> -X"
using assms by auto

lemma B04:
 " ((X:: 'a intervals) = Y ) \<longleftrightarrow> (X \<subseteq> Y) \<and> (Y \<subseteq> X)"
by auto

lemma B09:
 assumes "-X \<union> Y = STrue"
 shows   "(X:: 'a intervals) \<subseteq> Y"
using assms using strue_def by auto

lemma B20:
 "(X:: 'a intervals) \<subseteq> Y \<union> Z \<longleftrightarrow> X \<inter> -Y \<subseteq> Z"
by auto

lemma B28:
 " ((X:: 'a intervals) \<inter> Y) \<union> (X \<inter> Z) = X \<inter> (Y \<union> Z) "
by auto
  
lemma CH01:
 "STrue\<cdot>STrue = STrue"
by (metis FusionSEmptyR FusionUnionDistR Int_commute SStarSkip STrueTop UnfoldL inf_sup_absorb )

lemma CH07:
 "(((SSkip \<inter> X)\<cdot>V) \<inter> ((SSkip \<inter> Y)\<cdot>W)) = ((SSkip \<inter> X \<inter> Y)\<cdot>(V \<inter> W))   "
using PwrFusionInterL[ of "1" "X" "V" "Y" "W"] 
by (simp add: FusionSEmptyR inf_commute smore_def sskip_def)

lemma CH08:
 "((V\<cdot>(SSkip \<inter> X)) \<inter> ((W\<cdot>(SSkip \<inter> Y)))) = ((V \<inter> W)\<cdot>(SSkip \<inter> X \<inter> Y)) "
using PwrFusionInterR[of "V" "1" "X" "W" "Y"] 
by (simp add: FusionSEmptyR inf_commute smore_def sskip_def)

lemma CH09:
 "(((X \<inter> SEmpty)\<cdot>V) \<inter> ((Y \<inter> SEmpty)\<cdot>W)) = (((X \<inter> Y) \<inter> SEmpty)\<cdot>(V \<inter> W))   "
using PwrFusionInterL[ of "0" "X" "V" "Y" "W"] 
by (metis (no_types, lifting) inf_assoc inf_commute pwr_0)

lemma CH10:
 "((V\<cdot>(X \<inter> SEmpty)) \<inter> ((W\<cdot>(Y \<inter> SEmpty)))) = ((V \<inter> W)\<cdot>((X \<inter> Y) \<inter> SEmpty)) "
using PwrFusionInterR[of "V" "0" "X" "W" "Y"]
by (metis (no_types, lifting) inf_assoc inf_commute pwr_0)

lemma ST13:
 " ((X \<inter> SEmpty)\<cdot>Z) \<inter> ((Y \<inter> SEmpty)\<cdot>Z) = ((X \<inter> Y) \<inter> SEmpty)\<cdot>Z"
by (simp add: CH09)

lemma ST15:
 "(SStar (X \<inter> SEmpty)) = SEmpty"
by (metis FusionSEmptyL inf.right_idem inf_le2 UnfoldL
          SStarInductR sup.orderE sup_inf_absorb)

lemma ST21:
 "((-X) \<inter> SEmpty) \<union> (X \<inter> SEmpty) = SEmpty "
by blast 

lemma ST24:
 " (SInit X) \<inter> (SInit Y) = (SInit (X \<inter> Y))"
by (simp add: ST13 sinit_def)

lemma ST25:
 " (SInit STrue) = STrue"
by (simp add: sinit_def strue_def FusionSEmptyL)

lemma ST26:
 " (SInit (-X)) \<union> (SInit X) = STrue"
by (metis Compl_disjoint2 ST21 ST25 Un_Int_distrib compl_bot_eq FusionUnionDistL 
          sinit_def strue_def sup_bot.right_neutral sup_top_right) 

lemma ST28:
 " (SDi (SInit X)) = (SInit X)"
by (metis compl_bot_eq FusionAssoc FusionUnionDistR FusionSEmptyR sdi_def 
          sinit_def strue_def sup_top_right UnionCommute)

lemma ST33:
 " (STrue \<inter> SEmpty)\<cdot>SEmpty = SEmpty"
by (simp add: strue_def FusionSEmptyL)

lemma ST36:
 " (SInit (-X)) \<subseteq> -(SInit X)"
by (metis Compl_disjoint ST24 compl_bot_eq disjoint_eq_subset_Compl double_complement  
          inf.coboundedI2  inf.orderE   sfalse_def SFalseFusion sinit_def strue_def)

lemma ST37:
 "-(SInit X) \<subseteq> (SInit (-X)) "
using B09 ST26 by auto

lemma ST38:
 " -(SInit X) = (SInit (-X))"
using ST37 ST36 by auto

lemma ST47:
 " X \<union> Y\<cdot>X = (SEmpty \<union> Y)\<cdot>X"
by (simp add: FusionUnionDistL FusionSEmptyL)

lemma SStar01:
 assumes "X\<cdot>(SStar Y) \<union> SEmpty \<subseteq> (SStar Y) "
 shows   "(SStar X) \<subseteq> (SStar Y)"
using assms 
by (metis Un_commute FusionSEmptyR SStarInductL) 

lemma SStar03:
 "(SStar X)\<cdot>(SStar X) \<subseteq> (SStar X) "
by (metis SStarInductL Un_absorb UnfoldL sup.absorb_iff1 sup.right_idem)
 
lemma SStar05:
 assumes "((SStar X)\<cdot>(SStar X)) \<union> SEmpty \<subseteq> (SStar X)"
 shows   " (SStar (SStar X)) \<subseteq> (SStar X) "
using assms 
by (simp add: SStar01) 

lemma SStar12:
 " (SEmpty \<union> (X\<cdot>(SStar X))) \<subseteq> (SStar X)"
using UnfoldL by blast

lemma SStar06:
 "((SStar X)\<cdot>(SStar X)) \<union> SEmpty \<subseteq> (SStar X) "
using SStar03 SStar12 by force

lemma SStar07:
 " (SStar X) \<subseteq> (SStar (SStar X))"
by (metis FusionUnionDistR FusionSEmptyR Subsumption Un_commute UnfoldL ST47 sup.right_idem)

lemma SStar08:
 " (SStar X) = (SStar (SStar X))"
by (meson B04 SStar05 SStar06 SStar07)

lemma SStar15:
 " SEmpty \<subseteq> (SStar SSkip)"
by (simp add: SStarSkip strue_def)

lemma SStar16:
 " SSkip \<subseteq> (SStar SSkip)"
by (simp add: SStarSkip strue_def)

lemma SStar22:
 " (SEmpty \<inter> X)\<cdot>(SStar (SEmpty \<inter> X)) = (SEmpty \<inter> X)"
by (metis ST15 FusionSEmptyR inf_commute)

lemma SStar23:
 " (SStar (SEmpty \<inter> X)) = SEmpty"
using SStar22 UnfoldL by auto

lemma SStar25:
 " (SStar STrue) = STrue"
by (metis SStar08 SStarSkip)

lemma SStar28:
 " (SStar X)\<cdot>X \<subseteq> X\<cdot>(SStar X)"
by (metis B04 FusionUnionDistR FusionSEmptyR UnfoldL SStarInductL)

lemma SStar29:
 " X\<cdot>(SStar X) \<subseteq> (SStar X)\<cdot>X"
by (metis B04 SStar28 SStarInductR UnfoldL FusionRuleL ST47 sup.mono)

lemma SStar17:
 " (SStar SSkip)\<cdot>SSkip \<subseteq> SSkip\<cdot>(SStar SSkip)"
by (simp add: SStar28)

lemma SStar18:
 " SSkip\<cdot>(SStar SSkip) \<subseteq> (SStar SSkip)\<cdot>SSkip"
by (simp add: SStar29)

lemma SStar19:
 " (SStar SSkip)\<cdot>SSkip = SSkip\<cdot>(SStar SSkip)"
using SStar17 SStar18 by auto

lemma SStar30:
 " X\<cdot>(SStar X) = (SStar X)\<cdot>X"
using SStar28 SStar29 by auto

lemma SStar34:
 assumes "SEmpty \<union> (X \<union> Y)\<cdot>((SStar X)\<cdot>(SStar (Y\<cdot>(SStar X))) ) \<subseteq> (SStar X)\<cdot>(SStar (Y\<cdot>(SStar X))) "
 shows  "(SStar (X \<union> Y)) \<subseteq> (SStar X)\<cdot>(SStar (Y\<cdot>(SStar X))) "
by (metis assms FusionSEmptyR SStarInductL)

lemma SStar35:
 "SEmpty \<union> (X \<union> Y)\<cdot>((SStar X)\<cdot>(SStar (Y\<cdot>(SStar X))) ) \<subseteq> (SStar X)\<cdot>(SStar (Y\<cdot>(SStar X)))" 
by (simp add: FusionAssoc FusionUnionDistL ST47 UnfoldL UnionAssoc UnionCommute)

lemma SStar36:
 " (SStar (X \<union> Y)) \<subseteq> (SStar X)\<cdot>(SStar (Y\<cdot>(SStar X)))"
using SStar34 SStar35 by blast

lemma SStar46:
 " (SStar X)\<cdot>(SStar (Y\<cdot>(SStar X))) \<subseteq> (SStar (X \<union> Y))"
proof -
 have "(SEmpty \<union> SStar (X \<union> Y) \<cdot> Y) \<cdot> SStar X \<subseteq> SStar (X \<union> Y)"
 by (metis (no_types) FusionUnionDistR SStar12 SStar30 SStarInductR sup.bounded_iff) 
 then show ?thesis by (simp add: SStarInductR ST47 FusionAssoc) 
qed

lemma SStar47:
 " (SStar Z) = (SStar (Z \<inter> SMore))"
proof -
 have 1: "(SStar Z) = (SStar ((SEmpty \<inter> Z) \<union> (SMore \<inter> Z))) " 
      by (metis Int_Un_distrib2 compl_bot_eq inf_top.left_neutral smore_def strue_def STrueTop)
 have 2: "(SStar ((SEmpty \<inter> Z) \<union> (SMore \<inter> Z))) =
          (SStar (SEmpty \<inter> Z))\<cdot>(SStar ((SMore \<inter> Z)\<cdot>(SStar (SEmpty \<inter> Z)))) " 
      by (simp add: SStar36 SStar46 subset_antisym)
 have 3: "(SStar (SEmpty \<inter> Z))\<cdot>(SStar ((SMore \<inter> Z)\<cdot>(SStar (SEmpty \<inter> Z)))) =
          (SStar (Z \<inter> SMore)) "  
      by (simp add: FusionSEmptyL FusionSEmptyR SStar23 inf_commute) 
 from 1 2 3 show ?thesis by auto
qed

lemma SStar48:
 " (SStar SMore) = STrue"
by (metis Compl_Un Compl_disjoint2 SStar25 SStar47 ST21 ST33 FusionSEmptyR 
          inf.right_idem smore_def strue_def)

lemma SStar50:
assumes "SSkip\<cdot>((-X) \<union> ((SStar SSkip)\<cdot>(X \<inter> (SSkip\<cdot>(-X))))) \<union> (-X)
         \<subseteq> (-X) \<union> (SStar SSkip)\<cdot>(X \<inter> (SSkip\<cdot>(-X))) "
shows " ((SStar SSkip)\<cdot>(-X)) \<subseteq> ((-X) \<union> ((SStar SSkip)\<cdot>(X \<inter> (SSkip\<cdot>(-X)))))"
using SStarInductL assms by blast

lemma SStar51:
 " SSkip\<cdot>((-X) \<union> ((SStar SSkip)\<cdot>(X \<inter> (SSkip\<cdot>(-X))))) \<union> (-X)
         \<subseteq> (-X) \<union> (SStar SSkip)\<cdot>(X \<inter> (SSkip\<cdot>(-X)))"
using FusionUnionDistR[of "SSkip"  ] UnfoldL[of "SSkip"] 
proof -
have f1: "- X \<union> (SEmpty \<union> SSkip \<cdot> SStar SSkip) \<cdot> (X \<inter> (SSkip \<cdot> - X)) \<subseteq> 
          - X \<union> SStar SSkip \<cdot> (X \<inter> (SSkip \<cdot> - X))"
by (simp add: \<open>SEmpty \<union> SSkip \<cdot> SStar SSkip = SStar SSkip\<close>) (* 0.6 ms *)
have f2: "SSkip \<cdot> - X \<subseteq> - X \<union> SStar SSkip \<cdot> (X \<inter> (SSkip \<cdot> - X))"
by (metis B20 Morgan ST47 UnionIdem \<open>SEmpty \<union> SSkip \<cdot> SStar SSkip = SStar SSkip\<close> 
    inf_commute inf_idem sup.cobounded1) (* 48 ms *)
have "SSkip \<cdot> (SStar SSkip \<cdot> (X \<inter> (SSkip \<cdot> - X))) \<subseteq> - X \<union> SStar SSkip \<cdot> (X \<inter> (SSkip \<cdot> - X))"
using f1 by (metis (no_types) FusionAssoc ST47 Un_subset_iff) (* 22 ms *)
then show ?thesis
using f2 by (simp add: \<open>\<And>Z Y. SSkip \<cdot> (Y \<union> Z) = SSkip \<cdot> Y \<union> SSkip \<cdot> Z\<close>) (* 1 ms *)
qed


lemma SStar52:
 " (SStar X) \<subseteq> SEmpty \<union> (X \<inter> SMore)\<cdot>(SStar X)"
by (metis B04 SStar47 UnfoldL)

lemma SStar53:
 " SEmpty \<union> (X \<inter> SMore)\<cdot>(SStar X) \<subseteq> (SStar X)"
by (metis SStar12 SStar47)

lemma BD45:
 "(SBi ((-X) \<union> X1)) \<inter> (X\<cdot>Y) \<subseteq> X1\<cdot>Y  "
proof -
 have 1: "(SBi ((-X) \<union> X1)) = -(-((-X) \<union> X1)\<cdot>(Y \<union> (-Y)))  " 
      by (metis sbi_def sdi_def STrueTop)
 have 2: "-(-((-X) \<union> X1)\<cdot>(Y \<union> (-Y))) \<inter> (X\<cdot>Y) \<subseteq>
           -(-((-X) \<union> X1)\<cdot>(Y)) \<inter> (X\<cdot>Y)" 
        using FusionUnionDistR by fastforce
 have 3: "-(-((-X) \<union> X1)\<cdot>(Y)) \<inter> (X\<cdot>Y) \<subseteq>  (((-X) \<union> X1) \<inter> X)\<cdot>Y  " 
   by (metis FusionRuleL FusionUnionDistL dual_order.refl inf_sup_aci(1) shunt1 sup_neg_inf)
 have 4: " (((-X) \<union> X1) \<inter> X)\<cdot>Y \<subseteq> X1\<cdot>Y" 
      by (metis B20 double_compl FusionRuleL inf.right_idem inf_le1)
 from 1 2 3 4 show ?thesis by blast
qed

lemma BD46:
 "(SAlways ((-Y) \<union> Y1)) \<inter> (X1\<cdot>Y) \<subseteq> (X1\<cdot>Y1)  "
proof -
 have 1: "(SAlways ((-Y) \<union> Y1)) = -((X1 \<union> (-X1))\<cdot>(-((-Y) \<union> Y1 ) ) ) " 
      by (metis salways_def ssometime_def STrueTop)
 have 2: "-((X1 \<union> (-X1))\<cdot>(-((-Y) \<union> Y1 ) ) ) \<inter> (X1\<cdot>Y) \<subseteq>
          -(X1\<cdot>(-((-Y) \<union> Y1 ) ) ) \<inter>  (X1\<cdot>Y) "
      using FusionUnionDistL by fastforce
 have 3: "-(X1\<cdot>(-((-Y) \<union> Y1 ) ) ) \<inter>  (X1\<cdot>Y) \<subseteq> X1\<cdot>(((-Y) \<union> Y1) \<inter> Y) " 
       by (metis (no_types, lifting) B20 B04 compl_inf FusionUnionDistR Huntington 
                 Morgan Subsumption sup_ge1 UnionCommute)
 have 4: "X1\<cdot>(((-Y) \<union> Y1) \<inter> Y) \<subseteq> (X1\<cdot>Y1) " 
      by (metis B20 double_compl FusionRuleR inf.right_idem inf_le1)
 from 1 2 3 4 show ?thesis by blast
qed


subsubsection \<open>ITL Axioms derived\<close>

lemma SBoxGen:
 assumes  "X=STrue"
 shows    "(SAlways X) = STrue"
using assms 
by (metis double_compl FusionSFalse salways_def sfalse_def ssometime_def strue_def)

lemma SBiGen:
 assumes  "X=STrue"
 shows    "(SBi X) = STrue"
using assms 
by (metis double_compl sbi_def sdi_def sfalse_def SFalseFusion strue_def)

lemma SMP:
 assumes "X \<subseteq> Y"
 assumes "X=STrue"
 shows   " Y=STrue"
using assms(1) assms(2) 
using strue_def by blast

lemma SChopAssoc:
 "X\<cdot>(Y\<cdot>Z) = (X\<cdot>Y)\<cdot>Z"
by (simp add: FusionAssoc)

lemma SOrChopImp:
 " (X \<union> Y)\<cdot>Z \<subseteq> (X\<cdot>Z) \<union> (Y\<cdot>Z)"
by (simp add: FusionUnionDistL)

lemma SChopOrImp:
 " X\<cdot>(Y \<union> Z) \<subseteq> (X\<cdot>Y) \<union> (X\<cdot>Z)"
by (simp add: FusionUnionDistR)

lemma SEmptyChop:
 " SEmpty\<cdot>X = X"
by (simp add: FusionSEmptyL)

lemma SChopEmpty:
 " X\<cdot>SEmpty = X"
by (simp add: FusionSEmptyR)

lemma SStateImpBi:
 "(SInit X) \<subseteq> (SBi (SInit X)) " 
by (simp add: ST28 ST38 sbi_def)

lemma SNextImpNotNextNot:
 "(SNext X) \<subseteq> -(SNext (-X))"
proof -
 have 1: "((SNext X) \<subseteq> -(SNext (-X))) = (((SNext X) \<inter> (SNext (-X))) \<subseteq> SFalse)" 
      by (simp add: disjoint_eq_subset_Compl sfalse_def)
 have 2: " ((SNext X) \<inter> (SNext (-X))) = SSkip\<cdot>(X \<inter> (-X)) " 
      by (metis CH07 SStar16 inf.orderE snext_def) 
 have 3: "(SSkip)\<cdot>(X \<inter> (-X)) = SSkip\<cdot>SFalse " 
      by (simp add: sfalse_def)
 have 4: " SSkip\<cdot>SFalse = SFalse" by (simp add: FusionSFalse)
 from 1 2 3 4 show ?thesis by auto
qed

lemma SBiBoxChopImpChop:
 " (SBi ((-X) \<union> X1)) \<inter> (SAlways ((-Y) \<union> Y1))  \<inter> (X\<cdot>Y) \<subseteq> (X1\<cdot>Y1)"
using BD45 BD46 by blast

lemma SBoxInduct:
 "(SAlways (-X \<union> (SWnext X))) \<inter> X \<subseteq> (SAlways X)"
proof -
 have 1: " ((SAlways (-X \<union> (SWnext X))) \<inter> X \<subseteq> (SAlways X)) =
           ((SSometime (-X)) \<subseteq> ((-X) \<union> (SSometime (X \<inter> (SNext (-X))) ))) " 
      by (simp add: salways_def snext_def swnext_def)
        blast
 have 2: "((SSometime (-X)) \<subseteq> ((-X) \<union> (SSometime (X \<inter> (SNext (-X))) ))) =
          ( ((SStar SSkip)\<cdot>(-X)) \<subseteq> ((-X) \<union> ((SStar SSkip)\<cdot>(X \<inter> (SSkip\<cdot>(-X)))))  ) " 
      by (simp add: SStarSkip snext_def ssometime_def)
 have 3: "( ((SStar SSkip)\<cdot>(-X)) \<subseteq> ((-X) \<union> ((SStar SSkip)\<cdot>(X \<inter> (SSkip\<cdot>(-X)))))  ) " 
      using SStar51 SStar50 by blast
 from 1 2 3 show ?thesis by auto
qed

lemma SChopstarEqv:
 " (SStar X) = SEmpty \<union> (X \<inter> SMore)\<cdot>(SStar X)"
using SStar52 SStar53 by blast

subsection \<open>Extra Laws\<close>

subsubsection \<open>Boolean Laws\<close>

lemma B02:
 assumes " -Y \<subseteq> -X"
 shows   "(X:: 'a intervals) \<subseteq> Y"
using assms by auto

lemma B03:
 "((X:: 'a intervals) = Y ) \<longleftrightarrow> (-X = -Y) "
by auto

lemma B05:
 assumes "(X:: 'a intervals) \<union> Y \<subseteq> Z "
 shows   "X \<subseteq> Z \<and> Y \<subseteq> Z "
using assms by auto

lemma B06:
 assumes "X \<subseteq> Z \<and> Y \<subseteq> Z "
 shows   "(X:: 'a intervals) \<union> Y \<subseteq> Z "
using assms by auto

lemma B07:
 "(X:: 'a intervals) \<union> Y \<subseteq> Z \<longleftrightarrow>
  X \<subseteq> Z \<and> Y \<subseteq> Z"
by auto

lemma B08:
 assumes "(X:: 'a intervals) \<subseteq> Y"
 shows   "-X \<union> Y = STrue"
using assms 
using strue_def by auto

lemma B10:
 "(X:: 'a intervals) \<subseteq> Y \<longleftrightarrow> -X \<union> Y = STrue"
using strue_def by auto

lemma B11:
 assumes "(X:: 'a intervals) \<subseteq> Y "
 shows   " X \<inter> -Y = SFalse"
using assms sfalse_def by auto

lemma B12:
 assumes " X \<inter> -Y = SFalse"
 shows   "(X:: 'a intervals) \<subseteq> Y "
using assms sfalse_def by auto

lemma B13:
 "(X:: 'a intervals) \<subseteq> Y \<longleftrightarrow> X \<inter> -Y = SFalse"
using sfalse_def by auto

lemma B14:
 assumes "(X:: 'a intervals) \<subseteq> Y"
 shows   "X \<inter> Y =X"
using assms by auto

lemma B15:
 assumes "(X:: 'a intervals) \<subseteq> Y \<inter> Z"
 shows   "X \<subseteq> Y \<and> X \<subseteq> Z "
using assms by auto

lemma B16:
 assumes "X \<subseteq> Y \<and> X \<subseteq> Z "
 shows   "(X:: 'a intervals) \<subseteq> Y \<inter> Z"
using assms by auto

lemma B17:
 "(X:: 'a intervals) \<subseteq> Y \<inter> Z \<longleftrightarrow> X \<subseteq> Y \<and> X \<subseteq> Z "
by auto

lemma B18:
 assumes "(X:: 'a intervals) \<subseteq> Y \<union> Z "
 shows   "X \<inter> -Y \<subseteq> Z"
using assms by auto

lemma B19:
 assumes "X \<inter> -Y \<subseteq> Z"
 shows   "(X:: 'a intervals) \<subseteq> Y \<union> Z "
using assms by auto

lemma B21:
 "(X:: 'a intervals) \<union> (Y \<inter> Z) \<subseteq> (X \<union> Y) \<inter> (X \<union> Z) \<longleftrightarrow>
  X \<union> (Y \<inter> Z) \<subseteq> (X \<union> Y) \<and> X \<union> (Y \<inter> Z) \<subseteq> (X \<union> Z)"
by auto

lemma B22:
  " (X:: 'a intervals) \<union> (Y \<inter> Z) \<subseteq> X \<union> Y"
by auto

lemma B23:
 " (X:: 'a intervals) \<union> (Y \<inter> Z) \<subseteq> X \<union> Z"
by auto

lemma B24:
 "((X:: 'a intervals) \<union> Y) \<inter> (X \<union> Z) \<subseteq> X \<union> (Y \<inter> Z) \<longleftrightarrow>
  ((X \<union> Y) \<inter> (X \<union> Z)) \<inter> -X \<subseteq> Y \<inter> Z "
by auto

lemma B25:
 "(((X:: 'a intervals) \<union> Y) \<inter> (X \<union> Z)) \<inter> -X \<subseteq> Y \<inter> Z \<longleftrightarrow>
  ((X \<union> Y) \<inter> (X \<union> Z)) \<inter> -X \<subseteq> Y \<and>
  ((X \<union> Y) \<inter> (X \<union> Z)) \<inter> -X \<subseteq> Z"
by auto

lemma B26:
 "(((X:: 'a intervals) \<union> Y) \<inter> (X \<union> Z)) \<inter> -X \<subseteq> Y "
by auto

lemma B27:
 "(((X:: 'a intervals) \<union> Y) \<inter> (X \<union> Z)) \<inter> -X \<subseteq> Z"
by auto

lemma B29:
 "(X:: 'a intervals) \<union> (Y \<inter> Z) = (X \<union> Y) \<inter> (X \<union> Z) "
by auto

subsubsection \<open>Chop\<close>

lemma CH02:
  " X\<cdot>Y \<inter> -(X\<cdot>Z) \<subseteq> X\<cdot>(Y \<inter> -Z)"
by (metis B20 FusionRuleR FusionUnionDistR inf.right_idem inf_le1)

lemma CH03:
 " X\<cdot>Y \<inter> -(Z\<cdot>Y) \<subseteq> (X \<inter> -Z)\<cdot>Y"
by (metis B20 FusionRuleL FusionUnionDistL inf.right_idem inf_le1)

lemma CH04:
 " X\<cdot>Y \<inter> -(X\<cdot>-Z) \<subseteq> X\<cdot>(Y \<inter> Z)"
using CH02 by fastforce

lemma CH05:
 " X\<cdot>Y \<inter> -(-Z\<cdot>Y) \<subseteq> (X \<inter> Z)\<cdot>Y"
using CH03 by fastforce

lemma CH06:
 assumes "X\<subseteq>X1"
         "Y\<subseteq>Y1"
 shows   "X\<cdot>Y\<subseteq>X1\<cdot>Y1"
using assms
by (metis FusionRuleL FusionRuleR order_trans)
  
lemma CH11:
 " ((X \<inter> (SPower SSkip n))\<cdot>STrue) \<inter> ((SPower SSkip n)\<cdot>Y) = (X \<inter> (SPower SSkip n))\<cdot>Y"
using PwrFusionInterL[of "n" "X" "STrue" "STrue" "Y"] 
by (simp add: inf_commute strue_def) 

lemma CH12:
 " (STrue\<cdot>(X \<inter> (SPower SSkip n))) \<inter> (Y\<cdot>(SPower SSkip n)) = (Y\<cdot>(X \<inter> (SPower SSkip n))) "
using PwrFusionInterR[of "STrue" "n" "X" "Y" "STrue"] 
by (metis STrueTop inf_commute inf_sup_absorb) 

lemma CH13:
 " (SPower SSkip n) \<cdot> (SPower SSkip m) = (SPower SSkip (n+m))"
proof
 (induct n arbitrary: m)
 case 0
 then show ?case by (simp add: FusionSEmptyL)
 next
 case (Suc n)
 then show ?case 
 by (metis FusionAssoc add_Suc pwr_Suc)
qed

subsubsection \<open>Next\<close>

lemma N01:
 " (SNext SEmpty) = SSkip"
by (simp add: FusionSEmptyR snext_def)

lemma N02:
 " (SNext SFalse) = SFalse "
by (simp add: FusionSFalse snext_def)

lemma N03:
 " (SNext X)\<cdot>Y = (SNext (X\<cdot>Y))"
by (simp add: snext_def FusionAssoc)

lemma N04:
 " (SNext (X \<union> Y)) = (SNext X) \<union> (SNext Y)"
by (simp add: FusionUnionDistR snext_def)

lemma N05:
 " (SNext (X \<inter> Y)) = (SNext X) \<inter> (SNext Y)"
by (metis CH07 SStar16 inf.orderE snext_def)

lemma N06:
 assumes "X \<subseteq> Y "
 shows   "(SNext X) \<subseteq> (SNext Y)"
using assms 
by (metis FusionUnionDistR Subsumption snext_def)

lemma N07:
 " (SNext ((-X) \<union> Y)) = (SNext (-X)) \<union> (SNext Y)"
by (simp add: N04)

lemma N08:
 " SMore \<subseteq> SSkip\<cdot>STrue"
by (simp add: smore_def)
  (metis B10 SStarSkip UnfoldL double_complement)

lemma N23:
 "(SWprev X) \<subseteq>  (SEmpty \<union>  (SPrev X)) "
proof -
have "X \<cdot> SSkip \<union> - X \<cdot> SSkip = SStar SSkip \<cdot> SSkip"
by (metis (no_types) Compl_empty_eq FusionUnionDistL SStarSkip strue_def sup_compl_top)
then have "- SWprev X \<union> (SEmpty \<union> SPrev X) = STrue"
by (metis (no_types) SStar19 SStarSkip UnfoldL UnionAssoc double_compl sprev_def sup_commute 
    swprev_def) 
then show ?thesis
by (meson B09) 
qed

lemma N24:
 " (SEmpty) \<subseteq> (SWprev X)" 
by (metis B10 B02 FusionRuleL SSkipFusionImpSMore SStar30 SStarSkip UnfoldL 
          compl_bot_eq double_compl smore_def strue_def subset_antisym swprev_def top_greatest)
 
lemma N25:
 " (SPrev X) \<subseteq> (SWprev X)"
proof -
 have 1: "((SPrev X) \<subseteq> (SWprev X) ) = (((SPrev X) \<inter> (SPrev (-X)))\<subseteq> SFalse )" 
      by (simp add: B10 sfalse_def sprev_def swprev_def)
 have 2: "((SPrev X) \<inter> (SPrev (-X))) = (X \<inter> (-X))\<cdot>SSkip " 
      by (metis CH08 SStar16 inf.orderE sprev_def)
 have 3: " (X \<inter> (-X))\<cdot>SSkip = SFalse\<cdot>SSkip" 
      by (simp add: sfalse_def)
 have 4: "SFalse\<cdot>SSkip = SFalse " 
      by (simp add: SFalseFusion)
 from 1 2 3 4 show ?thesis by auto
qed

lemma N26:
 "(SWprev X) =  (SEmpty \<union>  (SPrev X)) "
using N23 N24 N25 by blast

lemma N09:
 " SSkip \<union> SMore\<cdot>SSkip \<subseteq> SMore"
proof -
 have 1: " SSkip \<subseteq> SMore" by (simp add: smore_def sskip_def)
 have 2: " SMore\<cdot>SSkip \<subseteq> SMore" 
 by (metis N26 UnionCommute compl_le_swap1 smore_def sup_ge2 swprev_def)
 from 1 2 show ?thesis by auto
qed

lemma N10:
 assumes "SSkip \<union> SMore\<cdot>SSkip \<subseteq> SMore "
 shows   "SSkip\<cdot>(SStar SSkip) \<subseteq> SMore "
using assms 
using SStarInductR N09 by blast

lemma N11:
 " SSkip \<cdot> STrue \<subseteq> SMore"
by (metis SStarSkip N09 N10)

lemma N12:
 " (SNext X) = -(SWnext (-X))"
by (simp add: snext_def swnext_def)

lemma N13:
 " SMore\<cdot>STrue = SMore "
by (metis FusionAssoc N11 N08 SStar48 SStarSkip ST47 UnfoldL subset_antisym sup.right_idem)

lemma N14:
 " STrue\<cdot>SSkip \<subseteq> SMore"
by (metis N11 SStar19 SStarSkip)

lemma N15:
 " SMore \<subseteq> STrue\<cdot>SSkip"
by (metis N08 SStar19 SStarSkip)

lemma N19:
 "(SWnext X) \<subseteq> (SEmpty \<union> (SNext X)) "
 proof -
have "SSkip\<cdot>X \<union> SSkip\<cdot>(-X)  = SSkip\<cdot>SStar SSkip "
   using FusionUnionDistR[of "SSkip" "X" "-X"]  SStarSkip
   by (metis STrueTop)
then have "- SWnext X \<union> (SEmpty \<union> SNext X) = STrue"
   by (metis SStarSkip UnfoldL UnionCommute double_compl inf_sup_aci(7) snext_def swnext_def)
then show ?thesis by (simp add: B09)
qed

lemma N20:
 " (SEmpty ) \<subseteq> (SWnext X)"
proof -
 have 1: "((SEmpty ) \<subseteq> (SWnext X) ) = (-(SWnext X) \<subseteq> SMore)" 
      by (simp add: smore_def)
 have 2: "(-(SWnext X) \<subseteq> SMore) = ((SNext (-X)) \<subseteq> SMore) " 
      by (simp add: snext_def swnext_def)
 have 3: "(SNext (-X)) \<subseteq> SSkip\<cdot>STrue " 
      by (metis FusionUnionDistR STrueTop snext_def sup.orderI sup.right_idem)
 hence 4: "(SNext (-X)) \<subseteq> SMore " using SSkipFusionImpSMore by auto
 from 1 2 4 show ?thesis by auto
qed

lemma N21:
 " (SEmpty \<union> (SNext X)) \<subseteq> (SWnext X)"
using N20 
by (metis B06 SNextImpNotNextNot snext_def swnext_def)

lemma N22:
 "(SWnext X) = (SEmpty \<union> (SNext X)) "
using N21 N19 by blast

lemma N16:
 " (SNext X) = SMore \<inter> (SWnext X)"
using N12 N22 smore_def by blast

lemma N17:
 " (SWnext (X \<inter> Y)) = (SWnext X) \<inter> (SWnext Y)"
by (simp add: N05 N22 Un_Int_distrib)

lemma N18:
 " (SWnext (X \<union> Y)) = (SWnext X) \<union> (SWnext Y)"
by (simp add: swnext_def)
   (metis (no_types, lifting) CH07 SStar16 compl_inf inf.orderE)

lemma N27:
 " (SNext ((-X) \<union> Y)) \<subseteq> (-(SNext X) \<union> (SNext Y)) "
using N04 SNextImpNotNextNot by blast

lemma N28:
 "(SPrev ((-X) \<union> Y)) \<subseteq> ( -(SPrev X) \<union> (SPrev Y))"
unfolding sprev_def 
proof -
have "\<And>I. (SEmpty) \<union> I \<cdot> SSkip = - (- I \<cdot> SSkip)"
using N26 sprev_def swprev_def by blast (* 7 ms *)
then have "(Y \<union> - X) \<cdot> SSkip \<subseteq> Y \<cdot> SSkip \<union> - (X \<cdot> SSkip)"
using FusionUnionDistL by blast (* 2 ms *)
then show "(- X \<union> Y) \<cdot> SSkip \<subseteq> - (X \<cdot> SSkip) \<union> Y \<cdot> SSkip"
by (simp add: UnionCommute) (* 0.9 ms *)
qed
lemma N29:
 "(SPrev X) = -(SWprev (-X) )"
by (simp add: sprev_def swprev_def)

subsubsection \<open>SInit\<close>

lemma ST01:
 "(X \<inter> SEmpty)\<cdot>Y \<subseteq> Y "
by (metis Int_commute FusionUnionDistL FusionSEmptyL le_iff_sup sup_inf_absorb UnionCommute)

lemma ST02:
 " (X \<inter> SEmpty)\<cdot>Y \<subseteq> (X \<inter> SEmpty)\<cdot>STrue"
by (simp add: FusionRuleR strue_def)

lemma ST03:
 "(X \<inter> SEmpty)\<cdot>(X \<inter> SEmpty) \<subseteq> (X \<inter> SEmpty) "
using ST01 by auto

lemma ST04:
 "(X \<inter> SEmpty) \<subseteq> (X \<inter> SEmpty)\<cdot>(X \<inter> SEmpty) "
  by (metis (no_types, lifting) B04 CH10 FusionSEmptyL FusionSEmptyR Int_Un_eq(2) Int_Un_eq(4)
       inf_commute inf_sup_absorb)
 
lemma ST05:
 "(X \<inter> SEmpty) \<subseteq> -((-X) \<inter> SEmpty) "
by blast

lemma ST06:
 "((-X) \<inter> SEmpty) \<subseteq> -(X \<inter> SEmpty) "
by auto

lemma ST07:
 "(X \<inter> SEmpty) \<inter> (Y \<inter> SEmpty) \<subseteq> (X \<inter> SEmpty)\<cdot>STrue "
using ST02 FusionSEmptyR by blast

lemma ST08:
 "(X \<inter> SEmpty) \<inter> (Y \<inter> SEmpty) \<subseteq> (STrue \<inter> SEmpty)\<cdot>(Y \<inter> SEmpty) "
by (metis FusionSEmptyL FusionSEmptyR ST33 inf.cobounded2)

lemma ST09:
 "((X \<inter> SEmpty)\<cdot>STrue) \<inter> (STrue \<inter> SEmpty)\<cdot>(Y \<inter> SEmpty) \<subseteq> (X \<inter> SEmpty)\<cdot>(Y \<inter> SEmpty)"
by (metis compl_bot_eq eq_refl FusionAssoc FusionSEmptyR inf.commute inf_top.left_neutral 
          CH09 strue_def)

lemma ST10:
 "(X \<inter> SEmpty)\<cdot>(Y \<inter> SEmpty) \<subseteq> (X \<inter> SEmpty) "
by (metis FusionRuleR FusionSEmptyR inf_le2)

lemma ST11:
 "(X \<inter> SEmpty)\<cdot>(Y \<inter> SEmpty) \<subseteq> (Y \<inter> SEmpty) "
using ST01 by blast

lemma ST12:
 "(X \<inter> SEmpty) \<inter> (Y \<inter> SEmpty) = (X \<inter> SEmpty)\<cdot>SEmpty \<inter> (Y \<inter> SEmpty)\<cdot>SEmpty "
by (simp add: FusionSEmptyR)

lemma ST14:
 " ((X \<inter> Y) \<inter> SEmpty)\<cdot>SEmpty = ((X \<inter> Y) \<inter> SEmpty) "
by (simp add: FusionSEmptyR)

lemma ST16:
 " (X \<inter> SEmpty) \<inter> (Y \<inter> SEmpty) \<subseteq> SEmpty"
by (simp add: le_infI2)

lemma ST17:
 " (X \<inter> SEmpty)\<cdot>(Y \<inter> SEmpty) \<subseteq> SEmpty"
using ST10 by auto

lemma ST18:
 "-((X \<inter> SEmpty) \<union> (Y \<inter> SEmpty)) = -(X \<inter> SEmpty) \<inter> -(Y \<inter> SEmpty)"
by auto

lemma ST19:
 " (X \<inter> SEmpty)\<cdot>((-X) \<inter> SEmpty) \<subseteq> (X \<inter> SEmpty) "
using ST10 by blast

lemma ST20:
 " (X \<inter> SEmpty)\<cdot>((-X) \<inter> SEmpty) \<subseteq> ((-X) \<inter> SEmpty)"
using ST01 by auto

lemma ST22:
 " ((X \<inter> SEmpty)\<cdot>SSkip)\<cdot>(Y \<inter> SEmpty) \<subseteq> (X \<inter> SEmpty)\<cdot>SSkip"
using FusionRuleR FusionSEmptyR by blast

lemma ST23:
 " ((X \<inter> SEmpty)\<cdot>SSkip)\<cdot>(Y \<inter> SEmpty) \<subseteq> SSkip\<cdot>(Y \<inter> SEmpty)"
by (simp add: ST01 FusionRuleL)

lemma ST27:
 " (SInit X) \<inter> (Y\<cdot>Z) \<subseteq> ((SInit X) \<inter> Y)\<cdot>Z"
by (metis B04 compl_bot_eq FusionAssoc FusionSEmptyL inf_commute inf_top.left_neutral 
          CH09 sinit_def strue_def)

lemma ST29:
 " (SInit X)\<cdot>Y \<subseteq> (SInit X) "
using ST02 FusionAssoc sinit_def by fastforce

lemma ST30:
 " (SInit X) \<inter> (SDi Y) = (SDi ( (SInit X) \<inter> Y))" 
unfolding sdi_def sinit_def strue_def 
by (metis CH09 FusionAssoc FusionSEmptyL compl_bot_eq inf_top.left_neutral)


lemma ST31:
 " (X\<cdot>(STrue \<inter> SEmpty)) \<inter> (STrue\<cdot>(Y \<inter> SEmpty)) = X\<cdot>(Y \<inter> SEmpty) "
by (metis Int_commute compl_bot_eq inf_top.right_neutral CH10 strue_def)

lemma ST32:
 "(STrue \<inter> SEmpty)\<cdot>SEmpty \<inter> (SInit X) = (X \<inter> SEmpty) "
by (metis Compl_empty_eq Int_commute CH09 ST14 inf_top.right_neutral 
          sinit_def strue_def)

lemma ST34:
 "((X \<inter> SEmpty)\<cdot>Y) = (SInit X) \<inter> Y "
by (metis FusionSEmptyL Int_commute CH09 compl_bot_eq inf_top_right sinit_def strue_def)

lemma ST35:
 "((SInit X) \<inter> Y)\<cdot>Z \<subseteq> (SInit X) \<inter> (Y\<cdot>Z) "
by (metis B04 ST34 FusionAssoc)
 
lemma ST39:
 "SEmpty \<inter> (SInit X) \<subseteq> (X \<inter> SEmpty) "
using ST32 by blast

lemma ST40:
 " (X \<inter> SEmpty) \<subseteq> SEmpty \<inter> (SInit X)"
using ST32 by auto

lemma ST41:
 " SEmpty \<inter> (SInit X) = (X \<inter> SEmpty)"
using ST40 ST39 by auto

lemma ST42:
 " (X \<inter> SEmpty) \<subseteq> ((X \<union> Y) \<inter> SEmpty)"
by blast

lemma ST43:
 " (Y \<inter> SEmpty) \<subseteq> ((X \<union> Y) \<inter> SEmpty)"
by blast

lemma ST44:
 " (X \<inter> SEmpty) \<inter> ((-X) \<inter> SEmpty) = SFalse"
by (simp add: sfalse_def) 

lemma ST45:
 "((X \<union> Y) \<inter> SEmpty) \<subseteq> (X \<inter> SEmpty) \<union> (Y \<inter> SEmpty)"
by auto

lemma ST46:
 " (SInit X) \<union> (SInit Y) = (SInit (X \<union> Y))"
by (simp add: Int_Un_distrib2 FusionUnionDistL sinit_def)

lemma ST48:
 " -(STrue\<cdot>(X \<inter> SEmpty)) \<subseteq> STrue\<cdot>((-X) \<inter> SEmpty)"
by (metis B09 FusionSEmptyR FusionUnionDistR ST21 double_compl)

lemma ST49:
 "STrue\<cdot>((-X) \<inter> SEmpty) \<subseteq> -(STrue\<cdot>(X \<inter> SEmpty)) "
 by (metis CH10 Compl_disjoint2 FusionSFalse disjoint_eq_subset_Compl inf.idem inf_compl_bot_left2 
     sfalse_def)


lemma ST50:
 " -(STrue\<cdot>(X \<inter> SEmpty)) = STrue\<cdot>((-X) \<inter> SEmpty)"
using ST48 ST49 by blast 

subsubsection \<open>SStar\<close>

lemma SStar02:
 assumes " X \<subseteq> Y "
 shows   "X\<cdot>(SStar Y) \<union> SEmpty \<subseteq> (SStar Y)"
using assms 
by (metis FusionUnionDistL Int_lower1 SStar15 Un_commute Un_mono UnfoldL inf.orderE sup.orderE 
    sup.orderI)


lemma SStar04:
 " (SStar X) \<subseteq> (SStar X)\<cdot>(SStar X)"
by (metis Un_absorb UnfoldL UnionAssoc ST47 sup.absorb_iff2)

lemma SStar09:
 assumes " (X\<cdot>(SEmpty \<union> (X\<cdot>(SStar X)))) \<union> SEmpty \<subseteq> (SEmpty \<union> (X\<cdot>(SStar X))) "
 shows   " (SStar X) \<subseteq> SEmpty \<union> (X\<cdot>(SStar X))"
using assms 
by (simp add: UnfoldL) 

lemma SStar10:
 " (X\<cdot>(SEmpty \<union> (X\<cdot>(SStar X)))) \<subseteq> (SEmpty \<union> (X\<cdot>(SStar X)))"
by (metis UnfoldL sup_ge2)

lemma SStar11:
 " SEmpty \<subseteq> (SEmpty \<union> (X\<cdot>(SStar X)))"
by auto

lemma SStar13:
 " (SStar SSkip) = STrue"
by (simp add: SStarSkip)

lemma SStar14:
 " (SSometime X) = (SStar SSkip)\<cdot>X"
by (simp add: SStarSkip ssometime_def)

lemma SStar20:
 " (SStar SEmpty) = SEmpty"
by (metis FusionSEmptyR ST15 ST33)

lemma SStar21:
 " (SStar (SEmpty \<inter> X))\<cdot>(SEmpty \<inter> X) = (SEmpty \<inter> X)"
by (metis ST15 FusionSEmptyL inf_commute)

lemma SStar24:
 " (SStar SFalse) = SEmpty"
by (metis SStar20 SStar47 inf_compl_bot sfalse_def smore_def)

lemma SStar26:
 " X \<subseteq> (SStar X)"
by (metis FusionSEmptyR FusionUnionDistR  SStar08  UnCI UnfoldL  subsetI subset_iff)

lemma SStar27:
 " SEmpty \<subseteq> (SStar X)"
using UnfoldL by blast

lemma SStar31:
 assumes "X \<union> (X\<cdot>Y)\<cdot>(X\<cdot>(SStar (Y\<cdot>X))) \<subseteq> X\<cdot>(SStar (Y\<cdot>X))"
 shows   "(SStar (X\<cdot>Y))\<cdot>X \<subseteq> X\<cdot>(SStar (Y\<cdot>X)) "
using assms SStarInductL by blast

lemma SStar32:
 "X \<union> (X\<cdot>Y)\<cdot>(X\<cdot>(SStar (Y\<cdot>X))) \<subseteq> X\<cdot>(SStar (Y\<cdot>X)) "
by (metis B06 SStar10 SStar11 FusionAssoc FusionRuleR FusionSEmptyR UnfoldL)

lemma SStar33:
 "(SStar (X\<cdot>Y))\<cdot>X \<subseteq> X\<cdot>(SStar (Y\<cdot>X))"
using SStar31 SStar32 by blast

lemma SStar37:
 assumes "X\<cdot>Z \<subseteq> Z\<cdot>Y "
 shows   "(SStar X)\<cdot>Z \<subseteq> Z\<cdot>(SStar Y) "
proof -
have "Z \<cdot> SStar Y = Z \<cdot> SEmpty \<union> Z \<cdot> (Y \<cdot> SStar Y)"
by (metis FusionUnionDistR UnfoldL) 
then have "Z \<cdot> SStar Y \<union> (Z \<union> X \<cdot> (Z \<cdot> SStar Y)) = Z \<union> Z \<cdot> Y \<cdot> SStar Y \<union> X \<cdot> Z \<cdot> SStar Y"
using FusionAssoc FusionSEmptyR by blast 
then have "Z \<cdot> SStar Y \<union> (Z \<union> X \<cdot> (Z \<cdot> SStar Y)) = Z \<cdot> SStar Y"
by (metis (no_types) FusionAssoc FusionSEmptyR FusionUnionDistL FusionUnionDistR UnfoldL UnionAssoc
    assms sup.absorb_iff1) 
then show ?thesis
by (meson SStarInductL sup.absorb_iff1) 
qed

lemma SStar38:
 assumes " Z\<cdot>X \<subseteq> Y\<cdot>Z"
 shows   "Z\<cdot>(SStar X) \<subseteq> (SStar Y)\<cdot>Z "
using assms 
proof -
have f1: "Z \<union> SStar Y \<cdot> Y \<cdot> Z = SStar Y \<cdot> Z"
by (metis (no_types) SStar30 ST47 UnfoldL) 
have "SStar Y \<cdot> Y \<cdot> Z = SStar Y \<cdot> Z \<cdot> X \<union> SStar Y \<cdot> Y \<cdot> Z"
by (metis FusionAssoc FusionUnionDistR assms subset_Un_eq) 
then have "Z \<union> SStar Y \<cdot> Z \<cdot> X \<subseteq> SStar Y \<cdot> Z"
using f1 by blast 
then show ?thesis
by (simp add: SStarInductR) 
qed


lemma SStar39:
 " Y\<cdot>(SStar ((SStar X)\<cdot>Y)) \<subseteq> (SStar (Y\<cdot>(SStar X)))\<cdot>Y"
by (simp add: SStar38 FusionAssoc)

lemma SStar40:
 " (SStar (Y\<cdot>(SStar X)))\<cdot>Y \<subseteq> Y\<cdot>(SStar ((SStar X)\<cdot>Y))"
by (simp add: SStar33)

lemma SStar41:
 "Y\<cdot>(SStar ((SStar X)\<cdot>Y)) = (SStar (Y\<cdot>(SStar X)))\<cdot>Y "
using SStar39 SStar40 by blast

lemma SStar42:
 " Z\<cdot>(SStar (Y\<cdot>Z)) \<subseteq> (SStar (Z\<cdot>Y))\<cdot>Z"
by (simp add: SStar38 FusionAssoc)

lemma SStar43:
 " (SStar (Z\<cdot>Y))\<cdot>Z \<subseteq> Z\<cdot>(SStar (Y\<cdot>Z))"
by (simp add: SStar33)

lemma SStar44:
 " Z\<cdot>(SStar (Y\<cdot>Z)) = (SStar (Z\<cdot>Y))\<cdot>Z"
using SStar42 SStar43 by blast

lemma SStar49:
 " (SStar X) = SEmpty \<union> (SStar X)\<cdot>X"
using SStar30 UnfoldL  by blast

subsubsection \<open>Box and Diamond\<close>

lemma BD01:
 " (SSometime SEmpty) = STrue"
by (simp add: ssometime_def FusionSEmptyR)

lemma BD02:
 " X \<subseteq> (SSometime X)"
unfolding ssometime_def 
by (metis CH01 SStar25 ST47 UnfoldL subset_Un_eq)

lemma BD03: 
 " (SNext (SSometime X)) \<subseteq> (SSometime X)"
by (metis FusionUnionDistL N03 SStar16 SStar03 SStar04 SStarSkip snext_def ssometime_def 
          sup.absorb_iff2 sup.orderE)

lemma BD04:
 " (SSometime (SNext X)) \<subseteq> (SSometime X)"
by (metis CH01 FusionAssoc FusionUnionDistL FusionUnionDistR SStar16 SStarSkip 
          snext_def ssometime_def sup.absorb_iff2)

lemma BD05:
 " (SSometime X) \<union> (SSometime Y) = (SSometime (X \<union> Y))"
by (simp add: FusionUnionDistR ssometime_def)

lemma BD06:
 " (SSometime STrue) = STrue"
by (simp add: CH01 ssometime_def)

lemma BD07:
 " (SSometime (X \<inter> Y)) \<subseteq> (SSometime X) \<inter> (SSometime Y)"
by (simp add: FusionRuleR ssometime_def)

lemma BD08:
 " (SAlways STrue) = STrue"
by (simp add: SBoxGen)

lemma BD09:
 " -(SAlways X) = (SSometime (-X))"
by (simp add: salways_def)

lemma BD10:
 " (SAlways X) \<subseteq> (SSometime X)"
by (metis B02 BD02 BD09 subset_trans)


lemma BD11:
 " (SSometime (SSometime X)) = (SSometime X)"
by (simp add: CH01 ssometime_def FusionAssoc)

lemma BD12:
 " (SAlways X) \<subseteq> X"
by (simp add: B02 BD02 BD09)

lemma BD13:
 " (SDi STrue) = STrue"
by (simp add: CH01 sdi_def)

lemma BD14:
 " (SDi SEmpty) = STrue"
by (simp add: sdi_def FusionSEmptyL)

lemma BD15:
 " (SBi STrue) = STrue"
by (simp add: SBiGen) 

lemma BD16:
 " (SDi (X \<union> Y)) = (SDi X) \<union> (SDi Y)"
by (simp add: FusionUnionDistL sdi_def)

lemma BD17:
 assumes " X \<subseteq> Y"
 shows   "(SDi X) \<subseteq> (SDi Y)"
using assms 
by (metis FusionUnionDistL Subsumption sdi_def)

lemma BD18:
 " (SDi (SDi X)) = (SDi X)"
by (metis CH01 FusionAssoc sdi_def)

lemma BD19:
 " (SDa SEmpty) = STrue"
by (simp add: CH01 sda_def FusionSEmptyR)

lemma BD20:
 " (SDa STrue) = STrue"
by (simp add: CH01 sda_def)

lemma BD21:
 " (SBa STrue) = STrue"
by (metis BD15 BD08 BD09 sba_def sbi_def sda_def sdi_def ssometime_def)

lemma BD22:
 " (SDa (X \<union> Y)) = (SDa X) \<union> (SDa Y)"
by (simp add: FusionUnionDistL FusionUnionDistR sda_def) 

lemma BD23:
 assumes " X \<subseteq> Y" 
 shows   " (SDa X) \<subseteq> (SDa Y)"
using assms 
by (metis BD22 Subsumption)

lemma BD24:
 assumes " X \<subseteq> Y"
 shows   " (SDa (-Y)) \<subseteq> (SDa (-X))"
using assms 
by (simp add: BD23)

lemma BD25:
 " (SDi X) \<subseteq> (SDa X)"
by (metis BD02 FusionAssoc sda_def sdi_def ssometime_def)

lemma BD26:
 " (SSometime X) \<subseteq> (SDa X)"
by (metis BD01 BD02 FusionSEmptyR FusionUnionDistR SStar14 le_iff_sup sda_def)

lemma BD27:
 " (SBa X) \<subseteq> (SBi X)"
by (simp add: BD25 sba_def sbi_def)

lemma BD28:
 " (SBa X) \<subseteq> (SAlways X)"
by (simp add: B02 BD26 BD09 sba_def) 

lemma BD29:
 " (SAlways X) \<inter> (SAlways Y) = (SAlways (X \<inter> Y)) "
by (metis BD05 BD09 Morgan compl_inf salways_def)

lemma BD30:
 " (SAlways X) \<union> (SAlways Y) \<subseteq> (SAlways (X \<union> Y))"
using BD07 
by (metis B02 BD09 compl_sup)

lemma BD31:
 " (SDi (X \<inter> Y)) \<subseteq> (SDi X) \<inter> (SDi Y)"
by (simp add: BD17)

lemma BD32:
 " (SBi X) \<union> (SBi Y) \<subseteq> (SBi (X \<union> Y))"
using  BD31 
by (metis (mono_tags, lifting) B02 compl_sup double_compl sbi_def)

lemma BD33:
 " (SDa (X \<inter> Y)) \<subseteq> (SDa X) \<inter> (SDa Y)"
by (simp add: BD23)

lemma BD34:
 " (SBa X) \<union> (SBa Y) \<subseteq> (SBa (X \<union> Y))"
using BD33 
by (metis (mono_tags, lifting) B02 compl_sup double_compl sba_def)

lemma BD35:
 " (SAlways SEmpty) = SEmpty"
by (metis N13 SStar14 SStar30 SStar48 SStarSkip double_complement salways_def smore_def)

lemma BD36:
 " (SBi SEmpty) = SEmpty"
using N13 sbi_def sdi_def smore_def by fastforce

lemma BD37:
 " (SBa SEmpty) = SEmpty"
by (metis N13 SStar30 SStar48 double_complement sba_def sda_def smore_def)

lemma BD38:
 assumes " X \<subseteq> Y"
 shows   " (SAlways X) \<subseteq> (SAlways Y)"
using assms 
by (simp add: BD29 inf.absorb_iff2)

lemma BD39:
 assumes " X \<subseteq> Y"
 shows   " (SBi X) \<subseteq> (SBi Y)"
using assms 
by (simp add: BD17 sbi_def)

lemma BD40:
 assumes " X \<subseteq> Y"
 shows   " (SBa X) \<subseteq> (SBa Y)"
using assms 
by (simp add: BD24 sba_def)

lemma BD41:
 " (SBi (SBi X)) = (SBi X)"
by (simp add: BD18 sbi_def)

lemma BD42:
 " (SAlways (SAlways X)) = (SAlways X)"
by (simp add: BD11 salways_def)

lemma BD43:
 " (SDa (SDa X)) = (SDa X)"
by (metis CH01 FusionAssoc sda_def)

lemma BD44:
 " (SBa (SBa X)) = (SBa X)"
by (simp add: BD43 sba_def)

lemma BD47:
 " (SAlways ( (-X) \<union> Y)) \<subseteq> ( -(SAlways X) \<union> (SAlways Y)) "
by (metis B20 BD12 BD29 BD38 BD42 double_compl)

lemma BD48:
 " (SAlways X) \<subseteq> X \<inter> (SWnext (SAlways X))"
by (metis B02 B16 BD03 BD09 BD12 N12 salways_def)

lemma BD49:
 " (SBi ( (-X) \<union> Y)) \<subseteq> (-(SBi X) \<union> (SBi Y) )"
by (metis B20 BD45 Un_commute double_complement sbi_def sdi_def)

lemma BD50: 
 " (SPrev (SDi X)) \<subseteq> (SDi X)"
by (metis B04 FusionAssoc FusionUnionDistR N08 SSkipFusionImpSMore SStar19 SStarSkip 
    STrueTop sdi_def smore_def sprev_def sup_ge2)

lemma BD51: 
 " -(SBi X) = (SDi (-X))"
by (simp add: sbi_def)

lemma BD52:
 " X \<subseteq> (SDi X)"
by (metis FusionSEmptyR FusionUnionDistR ST33 Subsumption UnionCommute sdi_def sup_inf_absorb)

lemma BD53:
 " (SBi X) \<subseteq> X"
by (simp add: B02 BD51 BD52)

lemma BD54:
 " (SBi X) \<subseteq> X \<inter> (SWprev (SBi X))"
by (metis B02 B16 BD50 BD51 BD53 N29 sbi_def)

lemma BD55:
 " (SBi (SMore \<union> X)) = (SInit X)"
by (metis (no_types, lifting) ST38 compl_sup double_complement inf_commute sbi_def sdi_def
     sinit_def smore_def)


lemma BD56:
 " (SAlways (SMore \<union> X)) = STrue\<cdot>(X \<inter> SEmpty)"
by (simp add: SStar14 SStarSkip ST50 UnionCommute salways_def smore_def)
 

subsection \<open>Time Reversal\<close>

subsubsection \<open>Time Reversal Axioms\<close>

lemma SRevSEmpty:
 "(SRev SEmpty) = SEmpty"
using set_eqI[of "(SRev SEmpty) " "SEmpty"]
by (simp add: sempty_elim srev_elim)

lemma SRevSNot:
 "(SRev (- X)) = (- (SRev X))"
using set_eqI[of "(SRev (- X))" "(- (SRev X))"]
by (simp add: srev_elim)

lemma fuse_irev:
assumes "ilast xs = ifirst ys"
shows "irev (fuse xs ys) = fuse (irev ys) (irev xs)"
using assms fuse_prefix_suffix 
      prefix_fuse[of "xs" "ys"]  
      suffix_fuse[of "xs" "ys"] 
by (metis irev_ilen irev_prefix irev_suffix prefix_ilen_bound suffix_ilen)

lemma SRevFusionsem:
"x \<in>  (SRev (X\<cdot>Y)) = (x \<in> ((SRev Y)\<cdot>(SRev X)))"
by (auto simp add: srev_elim fusion_iff )
 (metis fuse_irev ilast_irev irev_irev_ident, metis fuse_irev ifirst_irev ilast_irev)

lemma SRevFusion:
 "(SRev (X\<cdot>Y)) = (SRev Y)\<cdot>(SRev X) "
using  set_eqI[of "(SRev (X\<cdot>Y)) " "(SRev Y)\<cdot>(SRev X) "] 
using SRevFusionsem by auto

     
lemma SRevUnion:
 "(SRev (X \<union> Y)) = (SRev X) \<union> (SRev Y)"
using set_eqI[of "(SRev (X \<union> Y))" "(SRev X) \<union> (SRev Y) "]
using srev_elim by auto

lemma SRevSPower:
 " (SRev (SPower X n)) = (SPower (SRev X) n)"
proof (induct n)
 case 0
 then show ?case by (simp add: SRevSEmpty)
 next
 case (Suc n)
 then show ?case 
  proof -
   have "SRev X \<inter> SMore = SRev (X \<inter> SMore)"
   by (metis (no_types) Morgan SRevSEmpty SRevSNot SRevUnion smore_def) 
   then show ?thesis
   by (simp add: SRevFusion Suc.hyps spower_commutes) 
  qed
qed

lemma SRevSStar:
 "(SRev (SStar X)) = (SStar (SRev X)) "
proof -
 have 1: "(SRev (SStar X)) = (SRev (\<Union> n. SPower X n)) " by (simp add: sstar_def)
 have 2: "(SRev (\<Union> n. SPower X n)) = (\<Union> n. SPower (SRev X) n) " 
      using set_eqI[of "(SRev (\<Union> n. SPower X n)) " "(\<Union> n. SPower (SRev X) n) "]
       SRevSPower[of "X"]  
       by (metis srev_elim sstar_def sstar_elim)
 have 3: "(\<Union> n. SPower (SRev X) n) = (SStar (SRev X)) " by (simp add: sstar_def)
 from 1 2 3 show ?thesis by auto
qed

lemma SRevSRev:
 " (SRev (SRev X)) = X"
using set_eqI[of "(SRev (SRev X))" "X"]
by (simp add: srev_elim)


subsubsection \<open>Time Reversal Laws\<close>

lemma TR01:
 "(SRev SMore) = SMore"
by (simp add: SRevSEmpty SRevSNot smore_def) 

lemma TR02:
 "(SRev SSkip) = SSkip"
by (metis SRevFusion SRevSEmpty SRevSNot SRevUnion TR01 sskip_def)

lemma TR03:
 "(SRev STrue) = STrue "
by (metis SRevSStar SStarSkip TR02)

lemma TR04:
 "(SRev SFalse) = SFalse"
by (metis Compl_eq_Compl_iff SRevSNot TR03 sfalse_def strue_def)

lemma TR05:
 " (SRev (SSometime X)) = (SDi (SRev X))"
by (simp add: SRevFusion TR03 sdi_def ssometime_def)

lemma TR06:
 " (SRev (SAlways X)) = (SBi (SRev X))"
by (simp add: SRevSNot TR05 salways_def sbi_def)

lemma TR07:
 " (SRev (SDi X)) = (SSometime (SRev X)) "
by (simp add: SRevFusion TR03 sdi_def ssometime_def)

lemma TR08:
 " (SRev (SBi X)) = (SAlways (SRev X))"
by (metis SRevSRev TR06)

lemma TR09:
 " (SRev (SNext X)) = (SPrev (SRev X))"
by (simp add: SRevFusion TR02 snext_def sprev_def)

lemma TR10:
 " (SRev (SWnext X)) = (SWprev (SRev X))"
by (simp add: SRevFusion SRevSNot TR02 swnext_def swprev_def)

lemma TR11:
 " (SRev (SPrev X)) = (SNext (SRev X))"
by (simp add: SRevFusion TR02 snext_def sprev_def)

lemma TR12:
 " (SRev (SWprev X)) = (SWnext (SRev X))"
by (metis SRevSRev TR10)

lemma TR13:
 " (SRev (SDa X)) = (SDa (SRev X))"
by (simp add: SRevFusion TR03 sda_def FusionAssoc)

lemma TR14:
 " (SRev (SBa X)) = (SBa (SRev X))"
by (simp add: SRevSNot TR13 sba_def)

lemma TR15:
 " (SRev (SPower SSkip n)) = (SPower SSkip n)"
by (simp add: SRevSPower TR02)

lemma TR16:
 assumes "X \<subseteq> Y"
 shows   "(SRev X) \<subseteq> (SRev Y)"
using assms by (metis SRevUnion le_iff_sup)

lemma TR17:
 assumes "X = Y"
 shows   "(SRev X) =(SRev Y)"
using assms TR16 by auto


subsection \<open>Link between Set of Intervals and ITL\<close>

lemma interval_lan [simp]:
 " \<sigma> \<in> (lan f) \<longleftrightarrow> (\<sigma> \<Turnstile> f)"
by (simp add: lan_def)

lemma valid_lan_eqv :
 "( (lan f) = (lan g) ) \<longleftrightarrow> ( \<turnstile> f = g ) "
using interval_lan lan_def Valid_def by fastforce

lemma valid_lan_imp :
 " ( (lan f) \<subseteq> (lan g) ) \<longleftrightarrow> (\<turnstile> f \<longrightarrow> g) "
using interval_lan lan_def Valid_def 
by (simp add: Valid_def lan_def Collect_mono_iff)

lemma valid_strue :
 " ( (lan f) = STrue) \<longleftrightarrow> (\<turnstile> f) "
using strue_def by fastforce

lemma strue_true:
 " \<sigma> \<in> STrue \<longleftrightarrow> (\<sigma> \<Turnstile> #True)"
by (simp add: strue_elim)

lemma strue_true_1:
 " STrue = (lan (LIFT #True))"
using lan_def strue_true by fastforce

lemma sfalse_false:
 " \<sigma> \<in> SFalse \<longleftrightarrow> (\<sigma> \<Turnstile> #False)"
by (simp add: sfalse_def)

lemma sfalse_false_1:
 " SFalse = (lan (LIFT(#False)))"
using sfalse_false using lan_def by fastforce

lemma not_negation:
 " \<sigma> \<in> (-(lan f)) \<longleftrightarrow> (\<sigma> \<Turnstile> \<not> f)"
by simp
 
lemma not_negation_1:
 "-(lan f ) = (lan (LIFT(\<not> f))) "
using interval_lan lan_def by fastforce

lemma inter_and:
 "(\<sigma> \<in> ( (lan f) \<inter> (lan g) )) \<longleftrightarrow> (\<sigma> \<Turnstile> f \<and> g) "
by (simp add: lan_def)

lemma inter_and_1:
 " ( (lan f) \<inter> (lan g) ) = (lan (LIFT(f \<and> g))) "
using inter_and lan_def by fastforce

lemma union_or:
 "(\<sigma> \<in> ( (lan f) \<union> (lan g) )) \<longleftrightarrow> (\<sigma> \<Turnstile> f \<or> g) "
by (simp add: lan_def)

lemma union_or_1:
 " ( (lan f) \<union> (lan g) ) = (lan (LIFT(f \<or> g))) "
using union_or lan_def by fastforce

lemma subset_impl:
 " (\<sigma> \<in> ( -(lan f) \<union> (lan g))) \<longleftrightarrow> (\<sigma> \<Turnstile> f \<longrightarrow> g)"
by simp

lemma subset_impl_1:
 " ( -(lan f) \<union> (lan g)) = (lan (LIFT(f \<longrightarrow> g)))"
using subset_impl lan_def by fastforce

lemma fusion_chop:
 "(\<sigma> \<in> ((lan f)\<cdot>(lan g))) \<longleftrightarrow> (\<sigma> \<Turnstile> f;g)   "
by (metis fusion_iff chop_fuse interval_lan)

lemma fusion_chop_1:
 " ((lan f)\<cdot>(lan g)) = (lan (LIFT(f;g)))   "
using fusion_chop lan_def by blast

lemma sempty_empty:
    "  \<sigma> \<in> SEmpty \<longleftrightarrow> (\<sigma> \<Turnstile> empty)  "
by (simp add: empty_defs sempty_elim)

lemma sempty_empty_1:
 " SEmpty = (lan (LIFT empty))"
using sempty_empty lan_def by fastforce

lemma smore_more:
 "\<sigma> \<in> SMore \<longleftrightarrow> (\<sigma> \<Turnstile> more)  "
by (simp add: more_defs smore_elim)

lemma smore_more_1:
 " SMore = (lan (LIFT more))  "
using smore_more lan_def by fastforce

lemma sskip_skip:
 " \<sigma> \<in> SSkip = (\<sigma> \<Turnstile> skip)"
by (simp add: skip_defs sskip_elim) 

lemma sskip_skip_1:
 " SSkip = (lan (LIFT skip))"
using sskip_skip lan_def by fastforce

lemma snext_next:
 " \<sigma> \<in> (SNext (lan f)) \<longleftrightarrow> (\<sigma> \<Turnstile> \<circle> f)"
by (metis snext_def fusion_chop next_d_def sskip_skip_1)

lemma snext_next_1:
 " (SNext (lan f)) = (lan (LIFT(\<circle> f)))"
using snext_next lan_def by fastforce

lemma swnext_wnext:
 " \<sigma> \<in> (SWnext (lan f)) \<longleftrightarrow> (\<sigma> \<Turnstile> wnext f)"
by (simp add: swnext_def fusion_chop_1 next_d_def not_negation_1 sskip_skip_1 wnext_d_def)

lemma swnext_wnext_1:
 " (SWnext (lan f)) = (lan (LIFT(wnext f)))"
using swnext_wnext lan_def by fastforce

lemma sprev_prev:
 " \<sigma> \<in> (SPrev (lan f)) \<longleftrightarrow> (\<sigma> \<Turnstile> prev f)"
by (metis fusion_chop prev_d_def sprev_def sskip_skip_1)

lemma sprev_prev_1:
 " (SPrev (lan f)) = (lan (LIFT(prev f)))"
using sprev_prev lan_def by fastforce

lemma swprev_wprev:
 " \<sigma> \<in> (SWprev (lan f)) \<longleftrightarrow> (\<sigma> \<Turnstile> wprev f)"
by (simp add: fusion_chop_1 not_negation_1 prev_d_def sskip_skip_1 swprev_def wprev_d_def)

lemma swprev_wprev_1:
 " (SWprev (lan f)) = (lan (LIFT(wprev f)))"
using swprev_wprev lan_def by fastforce

lemma sinit_init:
 " \<sigma> \<in> SInit (lan f) \<longleftrightarrow> (\<sigma> \<Turnstile> init f)"
by (simp add: Int_commute fusion_chop_1 init_d_def inter_and_1 sempty_empty_1 sinit_def strue_true_1)

lemma sinit_init_1:
 " SInit (lan f) = (lan (LIFT(init f)))"
using sinit_init lan_def by fastforce

lemma and_inter_more:
 " \<sigma> \<in> (((lan f) \<inter> SMore)) \<longleftrightarrow> (\<sigma> \<Turnstile> (f \<and> more) ) "
using smore_more inter_and by auto

lemma and_inter_more_1:
 " \<sigma> \<in> (((lan f) \<inter> SMore)) \<longleftrightarrow> (\<sigma> \<in> (lan (LIFT(f \<and> more)) )) "
using and_inter_more lan_def by (simp add: smore_more_1)

lemma and_inter_more_2:
 "  ((lan f) \<inter> SMore) =  (lan (LIFT(f \<and> more)) ) "
using and_inter_more_1 by blast

lemma and_chop:
 "\<sigma> \<in> (((lan f) \<inter> SMore)\<cdot> (lan g)) \<longleftrightarrow> (\<sigma> \<Turnstile> (f \<and> more);g)  "
by (metis fusion_chop inter_and_1 smore_more_1)

lemma and_chop_1:
 " (((lan f) \<inter> SMore)\<cdot> (lan g)) = (lan (LIFT((f \<and> more);g)))  "
using and_chop lan_def by blast

lemma spower_chop_power:
 "  (SPower (lan f) n) = (lan (LIFT(power (f \<and> more) n)))"
proof (induct n)
case 0
then show ?case by (simp add: sempty_empty_1)
next
case (Suc n)
then show ?case by (metis and_chop_1 pow_Suc pwr_Suc)
qed

lemma sstar_spower:
 " \<sigma> \<in> SStar (lan f) \<longleftrightarrow> (\<exists> n. \<sigma> \<in> SPower (lan f) n)"
by (simp add: sstar_def)

lemma sstar_chopstar:
 " \<sigma> \<in> (SStar (lan f)) \<longleftrightarrow> \<sigma> \<in> (lan (LIFT(f\<^sup>\<star>))) "
proof -
 have 1: "\<sigma> \<in> (SStar (lan f)) = (\<exists>n. \<sigma> \<in> SPower (lan f) n)  " 
  using  sstar_spower by blast
 have 2: "(\<exists>n. \<sigma> \<in> SPower (lan f) n) = 
          (\<exists>n. \<sigma> \<in> lan (LIFT(power (f \<and> more) n)))"
  using spower_chop_power by blast
 have 3: "(\<exists>n. \<sigma> \<in> lan (LIFT(power (f \<and> more) n))) =
          (\<exists>n.  (LIFT(power (f \<and> more) n)) \<sigma>  )" 
  using interval_lan  by simp
 have 4: "(\<exists>n.  (LIFT(power (f \<and> more) n)) \<sigma>  ) =
          (\<sigma> \<in> (lan (LIFT(f\<^sup>\<star>))))  " 
    by (simp add: chopstar_d_def powerstar_d_def)
 show ?thesis by (simp add: "1" "2" "4")
qed

lemma chopstar_sstar_1:
 " (SStar (lan f)) = (lan (LIFT(f\<^sup>\<star>)))"
using sstar_chopstar lan_def by blast

lemma chopstar_seqv:
 " \<sigma> \<in> (lan (LIFT(f\<^sup>\<star>))) \<longleftrightarrow> 
   \<sigma> \<in> (lan (LIFT(empty \<or>  (f \<and> more); f\<^sup>\<star>)))"
   using ChopstarEqv by fastforce

lemma chopstar_seqv_1:
 " (lan (LIFT(f\<^sup>\<star>))) = (lan (LIFT(empty \<or>  (f \<and> more); f\<^sup>\<star>)))"
using  chopstar_seqv lan_def by blast

lemma srev_irev:
 " \<sigma> \<in> (SRev (lan f)) \<longleftrightarrow> \<sigma> \<in> (lan (LIFT(f\<^sup>r)))"
by (simp add: reverse_d_def srev_elim)

lemma srev_irev_1:
 " (SRev (lan f)) = (lan (LIFT(f\<^sup>r)))"
using srev_irev lan_def by blast

end