(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)
section \<open>Finite Intervals\<close>

theory Interval
 imports
   Main 
begin

(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 z3 e spass vampire " ]
*)

(*
declare [[smt_timeout=120]]
*)


text \<open>
   An interval is a finite sequence of elements of a particular type. Intervals are similar to list 
   in Isabelle/HOL, the difference is that intervals have instead of nil a single
   element at the end. So an interval of length zero is a single element whereas for Isabelle's list 
   we have that an empty list is nil (no element present). 

   The usual operations on intervals are defined: length (\<open>ilen\<close>), \<open>prefix\<close>, 
   \<open>suffix\<close>, \<open>sub\<close>,  \<open>inth\<close>, \<open>ifirst\<close>, \<open>ilast\<close>, \<open>iapp\<close> 
   and \<open>irev\<close>. 

   We also introduce \<open>index_sequence\<close> which is a sequence of chop (fuse) points. This sequence
   is used in the old definition of chopstar which is an existential quantification over this 
   sequence. The type \<open>index_sequence\<close> is again of type interval but the elements are 
   natural numbers. Two functions \<open>shift\<close> and \<open>shiftm\<close> are introduced that are used to add (shift) 
   and subtract a natural number of each element in the sequence of chop (fuse) points. 
   The operation \<open>upt\<close> to produce a sequence of consecutive chop points between two natural numbers.
    
\<close>

subsection \<open>Definitions\<close>


datatype (iset: 'a) interval =
    INil  'a ("\<lceil>_\<rceil>")
  | ICons 'a "'a interval"  (infixr "\<odot>" 65)
for
  map: imap
  rel: interval_all2
  pred: interval_all

(*print_theorems*)

type_synonym  index = "nat interval"  

syntax
  \<comment> \<open>interval Enumeration\<close>
  "_interval" :: "args => 'a interval"    ("\<langle>(_)\<rangle>")

translations
  "\<langle>x, xs\<rangle>" == "x\<odot>\<langle>xs\<rangle>"
  "\<langle>x\<rangle>" == "\<lceil>x\<rceil>"
  
primrec (nonexhaustive) ilen :: "'a interval \<Rightarrow> nat" where
    " ilen \<langle>x\<rangle> = 0"
  | " ilen (x\<odot>xs) = 1+ (ilen xs)"
 
primrec (nonexhaustive) inth :: "'a interval => nat => 'a" where
    "inth \<langle>x\<rangle> n      = x"
  | "inth (x\<odot>xs) n = (case n of 0 \<Rightarrow> x | Suc k \<Rightarrow> inth xs k)"
   
primrec prefix:: "nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" where
    "prefix n \<langle>x\<rangle> = \<langle>x\<rangle>"   
  | "prefix n (x \<odot>xs) = (case n of 0 \<Rightarrow> \<langle>x\<rangle> | Suc m \<Rightarrow> x \<odot> (prefix m xs))"

primrec suffix:: "nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" where
   "suffix n \<langle>x\<rangle> = \<langle>x\<rangle>" 
 | "suffix n (x\<odot>xs) = (case n of 0 \<Rightarrow> (x\<odot>xs)  | Suc m \<Rightarrow> suffix m xs)"

definition sub:: "nat \<Rightarrow> nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" 
where
 "sub n k xs = prefix (k-n) (suffix n xs)"



abbreviation ifirst ::  "'a interval \<Rightarrow> 'a" where
 "ifirst xs \<equiv> (inth xs 0)" 

abbreviation ilast ::  "'a interval \<Rightarrow> 'a" where
 "ilast xs \<equiv> (inth xs (ilen xs))" 

primrec iapp :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> 'a interval" (infixr "\<ominus>" 65) where
iapp_INil:   "\<langle>x\<rangle> \<ominus> ys = x \<odot> ys" |
iapp_ICons: "(x\<odot>xs) \<ominus> ys = x \<odot> (xs \<ominus> ys)"

primrec irev :: "'a interval \<Rightarrow> 'a interval" where
  " irev \<langle>x\<rangle> = \<langle>x\<rangle>"
| " irev (x \<odot>xs) = (irev xs) \<ominus> \<langle>x\<rangle>"

definition index_sequence :: " nat \<Rightarrow> index \<Rightarrow> bool" where
  "index_sequence x idx \<equiv> (inth idx 0 = x) \<and> (\<forall> n. n<ilen idx \<longrightarrow> inth idx n < inth idx (Suc n))"  

definition shift :: "nat \<Rightarrow> nat \<Rightarrow> nat" where
  "shift k = (\<lambda> x. x+k)"

definition shiftm ::  "nat \<Rightarrow> nat \<Rightarrow> nat" where
  "shiftm k = (\<lambda> x. x-k)" 

primrec upt :: "nat \<Rightarrow> nat \<Rightarrow> nat interval" ("(1[_..\<le>/_'])")
 where
    upt_0  : "[i..\<le>0]       = \<langle>0\<rangle>"
  | upt_Suc: "[i..\<le>(Suc j)] = (if i \<le> j then [i..\<le>j] \<ominus> \<langle>(Suc j)\<rangle> else \<langle>(Suc j)\<rangle> )"


subsection \<open>Lemmas\<close>

text \<open>
 Basic lemmas are introduced for each of the above operations on intervals.  
\<close>

subsubsection \<open>Shifting indexes to zero\<close>

lemma interval_shift_index_to_zero_a:
shows "(\<forall> (i::nat). a \<le> i \<and> i < a+b \<longrightarrow> f ( g1(i) ) ( g2(Suc i))) =
       (\<forall> i. 0 \<le> i \<and> i < b \<longrightarrow> f (g1(i+a)) (g2((Suc i)+a)))" 
by simp (metis add.commute le_Suc_ex le_add2 nat_add_left_cancel_less)

lemma interval_shift_index_to_zero_b:
shows "(\<forall> (i::nat). a \<le> i \<and> i < a+b \<longrightarrow> f ( g(i-a) ) ( g2((Suc i)-a))) =
       (\<forall> i. 0 \<le> i \<and> i < b \<longrightarrow> f (g(i)) (g2(Suc i)))" (is "?L=?R")
proof -
 have 1: "?L \<Longrightarrow> ?R" 
   by (metis  add_Suc add_diff_cancel_left' add_diff_cancel_right' le_add2 less_diff_conv)
 have 2: "?R \<Longrightarrow> ?L"
   by (metis Nat.add_diff_assoc add_diff_cancel_left' le0 le_add_diff_inverse2 less_diff_conv 
       plus_1_eq_Suc)
 show ?thesis using "1" "2" by blast
qed


subsubsection \<open>Interval Length\<close>

lemma ilen_gr_zero [simp]:
    "ilen xs \<ge> 0"
by auto

lemma ilen_ICons [simp]: 
    " (ilen (x\<odot>xs)) = (ilen xs) +1"
by simp
  
lemma ilen_ICons_1 :
    " ilen l > 0 = (\<exists> x ls. l = x\<odot>ls) "
by (induct l) simp_all

lemma ilen_imap [simp]:
    " ilen (imap f xs) = ilen xs"
by (induct xs) simp_all

lemma ilen_iapp [simp]:
 " ilen (xs \<ominus> ys) = (ilen xs) + (ilen ys) +1"
by (induct xs arbitrary: ys) simp_all

lemma irev_ilen [simp]:
 " ilen (irev xs) = ilen xs"
by (induct xs) simp_all



subsubsection \<open>inth\<close>

lemma inth_zero [simp]:
    " inth (x\<odot>xs) 0  = x"
by simp

lemma inth_Suc [simp]:
    "inth (x\<odot>xs) (Suc n) = inth xs n"
by auto

lemma inth_last:
    " inth (x\<odot>xs) (ilen (x\<odot>xs)) = inth xs (ilen xs) "
by simp


lemma inth_last_stutter:
 " inth xs (ilen xs + i) = inth xs (ilen xs)"
proof (induction xs arbitrary:i)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case by simp  
qed

lemma inth_ICons_a:
 assumes "0<i"
 shows   "inth(x\<odot>xs) i     = inth xs (i-1)"
using assms by (metis Suc_diff_1 inth_Suc)

lemma inth_ICons_b:
 shows   "inth(x\<odot>xs) (i+1) = inth xs i"
by simp

lemma inth_ICons:
 assumes "0<i"
 shows   "inth(x\<odot>xs) i     = inth xs (i-1) \<and>
          inth(x\<odot>xs) (i+1) = inth xs i"
by (meson assms inth_ICons_a inth_ICons_b)

lemma inth_zero_ifirst:
 " ifirst xs = inth xs 0   "
by simp 

lemma inth_ilen_ilast:
 " ilast xs = inth xs (ilen xs)  "
by simp

lemma INil_ilen :
    " (xs = \<langle>x\<rangle>) \<longleftrightarrow> ilen xs = 0 \<and> inth xs 0 = x" 
by (cases xs) auto

lemma interval_eq_inth_eq :
    " (xs =ys) = (ilen xs =ilen ys \<and> (\<forall> i\<le> ilen xs. inth xs i = inth ys i))"
proof 
 (induct xs arbitrary: ys)
 case (INil x)
 then show ?case by (metis INil_ilen le_numeral_extra(3))
 next
 case (ICons x1a xs)
 then show ?case 
  proof (cases ys)
  case (INil x1)
  then show ?thesis by simp
  next
  case (ICons x21 x22)
  then show ?thesis 
  using ICons.hyps by fastforce
  qed
 qed

lemma inth_imap :
    "inth (imap f xs) i = f (inth xs i)"
proof 
 (induct xs arbitrary: i)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by simp
   next
   case (Suc nat)
   then show ?thesis using ICons.hyps by simp
   qed
 qed




subsubsection \<open>prefix, suffix and sub\<close>



lemma prefix_INil [simp]: 
    " prefix m \<langle>x\<rangle> = \<langle>x\<rangle> " 
by simp

lemma prefix_suc [simp]:
    " prefix (Suc m) (x\<odot>xs) = x \<odot> (prefix m xs)" 
by auto

lemma prefix_zero [simp]:
    " prefix 0 (x\<odot>xs) = \<langle>x\<rangle>"
by auto

lemma prefix_zero_ifirst [simp]:
  " prefix 0 xs = \<langle>inth xs 0\<rangle>"
by (induct xs) simp_all

lemma ifirst_prefix [simp]:
 shows " ifirst (prefix i xs) = ifirst xs"
proof 
 (induct xs arbitrary: i)
 case (INil x)
 then show ?case by auto
 next
 case (ICons x1a xs)
 then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by auto
   next
   case (Suc nat)
   then show ?thesis using ICons.hyps by auto
   qed
 qed

  
lemma ilast_suffix [simp]:
 shows   "ilast (suffix i xs) = ilast xs" 
proof 
 (induct xs arbitrary: i)
 case (INil x)
 then show ?case by auto
 next
 case (ICons x1a xs)
 then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by auto
   next
   case (Suc nat)
   then show ?thesis using ICons.hyps by auto
   qed
 qed
 

lemma prefix_ilen [simp]:
    " (prefix (ilen xs) xs) = xs"
by (induct xs) simp_all

lemma prefix_ilen_gr_1 [simp]: 
    "(prefix ((ilen xs)+i) xs) = xs"
by (induct xs) simp_all

lemma ilen_prefix_ICons [simp]:
    " ilen( prefix (Suc i) (x\<odot>xs)) = 1 + ilen(prefix i xs)"
using ilen_ICons by auto

lemma prefix_ilen_code [code]: 
    " ilen (prefix i xs) = (if i\<le> ilen xs then i else ilen xs)"
proof 
 (induct xs arbitrary: i)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by auto
   next
   case (Suc nat)
   then show ?thesis using ICons.hyps by auto
   qed
 qed

lemma prefix_ilen_min [simp]: 
    " ilen (prefix i xs) = min i (ilen xs)"
by (simp add: prefix_ilen_code min_def)

lemma prefix_ilen_good [simp]:
   assumes " i\<le> ilen xs"
   shows   " (ilen ( prefix i xs)) = i" 
using assms by simp 

lemma prefix_ilen_bad : 
  assumes   " i > ilen xs"
  shows     "  ilen (prefix i xs) = ilen xs"
using assms by simp 
   
lemma prefix_ilen_bound :
 shows     " ilen (prefix i xs) \<le> ilen xs " 
by simp

lemma suffix_ilen_code [code]:
    " ilen (suffix i xs) = (if i\<le> ilen xs then (ilen xs)-i else 0) "
proof
 (induct xs arbitrary: i)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
  proof (cases i)
  case 0
  then show ?thesis by auto
  next
  case (Suc nat)
  then show ?thesis using ICons.hyps by auto
  qed
qed
  
lemma suffix_ilen [simp]:
 " ilen (suffix i xs ) = (ilen xs)-i"
by (simp add: suffix_ilen_code)
 
lemma suffix_ilen_good [simp]:
   assumes " i\<le> ilen xs"
   shows   " ilen (suffix i xs ) = (ilen xs)-i"
using assms by simp
  
lemma suffix_ilen_bad:
  assumes  " i> ilen xs"
  shows    "  ilen (suffix i xs ) = 0"
using assms by simp
 
lemma suffix_ilen_bound:
 "ilen(suffix i xs) \<le> ilen xs"
by simp

lemma inth_prefix [simp]:
 assumes "  k \<le> i"
 shows   "  inth (prefix i xs) k = inth xs k"
using assms
 proof 
  (induct i arbitrary: xs k)
  case 0
  then show ?case 
    proof (cases xs)
    case (INil x1)
    then show ?thesis by auto
    next
    case (ICons x21 x22)
    then show ?thesis using "0.prems" by auto
    qed
  next
  case (Suc i)
  then show ?case 
    proof (cases xs)
    case (INil x1)
    then show ?thesis by auto
    next
    case (ICons x21 x22)
    then show ?thesis 
       proof (cases k)
       case 0
       then show ?thesis by (simp add: ICons)
       next
       case (Suc nat)
       then show ?thesis 
       using Suc.hyps Suc.prems ICons by auto
       qed
   qed
qed
 
lemma inth_suffix [simp]:
 assumes " k \<le> ilen xs -i"
 shows   " inth (suffix i xs) k = inth xs (i+k)"
using assms
proof (induct xs arbitrary: i k)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by auto
   next
   case (Suc nat)
   then show ?thesis 
     proof auto
       assume a0: " i = Suc nat" 
       show " inth (suffix nat xs) k =  inth xs (nat + k) "
       using a0 ICons.hyps ICons.prems by auto
     qed
   qed
qed

lemma suffix_prefix_help_1:
 assumes " ia+i \<le> ilen xs"
         "  k \<le> ia"
 shows " inth (prefix ia (suffix i xs)) k = inth (suffix i (prefix (ia +i) xs)) k"
proof -
 have 1: " inth (prefix ia (suffix i xs)) k = inth (suffix i xs) k" 
 using inth_prefix assms by metis 
 have 2: " inth (suffix i xs) k = inth xs (i+k) " 
 using inth_suffix assms by (simp add: add_le_imp_le_diff)
 have 3: " inth xs (i+k) = inth (prefix (ia+i) xs) (i+k)" 
 using inth_prefix assms by simp
 have 4: " inth (prefix (ia+i) xs) (i+k) = inth (suffix i (prefix (ia +i) xs)) k" 
 using inth_suffix assms by simp
 from 1 2 3 4 show ?thesis by auto
qed

lemma suffix_prefix_help_2:
 assumes "ia+i \<le> ilen xs"
 shows   " (\<forall> k \<le> ia . inth (prefix ia (suffix i xs)) k = inth (suffix i (prefix (ia +i) xs)) k)"
using suffix_prefix_help_1 using assms by fastforce

lemma suffix_prefix_help_3:
 assumes "ia+i \<le> ilen xs "
 shows   "ilen (prefix ia (suffix i xs)) = ilen (suffix i (prefix (ia +i) xs)) "
using assms prefix_ilen_good suffix_ilen_good by auto

lemma suffix_prefix_swap:
 assumes "ia+i \<le> ilen xs"
 shows   " prefix ia (suffix i xs) = suffix i (prefix (ia +i) xs)"
using assms using interval_eq_inth_eq by fastforce


lemma prefix_prefix_zero [simp]:
    " prefix 0 ( prefix 0 xs ) = prefix 0 xs" 
by (induct xs) simp_all
  
lemma pref_pref [simp]:
    " (prefix i (prefix i xs)) = prefix i xs"
by (metis prefix_ilen prefix_ilen_gr_1 prefix_ilen_good less_imp_add_positive not_less)

lemma pref_pref_3 [simp]:
    " (prefix i (prefix (i+k) xs)) = prefix i xs"
proof 
  (induct xs arbitrary: i k)
  case (INil x)
  then show ?case by simp
  next
  case (ICons x1a xs)
  then show ?case 
   proof (cases i)
   case 0
   then show ?thesis 
    by (auto simp add: Nitpick.case_nat_unfold)
   next
   case (Suc nat)
   then show ?thesis 
    by (simp add: ICons.hyps)
   qed
qed
 
lemma pref_help:
 assumes "i\<le>ilen (prefix (ilen xs - Suc 0) xs)"
 shows   " (prefix i (prefix (ilen xs - Suc 0) xs)) = (prefix i xs)  "
using assms 
by (metis diff_le_self pref_pref_3 prefix_ilen_good 
     ordered_cancel_comm_monoid_diff_class.add_diff_inverse)

lemma pref_pref_help:
 assumes "ilen xs >0"
         "  ia<ilen (xs)"
 shows   "(prefix ia (prefix (ilen xs - Suc 0) xs)) = (prefix ia  xs) "
using assms 
by (metis Suc_leI Suc_le_mono Suc_pred diff_le_self pref_help prefix_ilen_good)

lemma interval_pref_pref_help_1:
 assumes "i>0"
         "i\<le> ilen xs"
 shows   "(prefix (ilen (prefix i xs) - Suc 0) (prefix i xs)) = 
          (prefix (ilen (prefix i xs) - Suc 0) xs) "
using assms pref_pref_3 by (metis diff_le_self prefix_ilen_good le_iff_add)

lemma suffix_suc [simp]:
    " suffix (Suc m) (x \<odot> xs) = suffix m xs"
by auto
  
lemma suffix_zero [simp]:
    " suffix 0 xs = xs "
by (induct xs) simp_all
    
lemma interval_hd_tail:
 assumes "ilen xs >0"
 shows   "xs = (ifirst xs) \<odot> (suffix 1 xs)"
by (metis One_nat_def assms ilen_ICons_1 inth_zero  suffix_suc suffix_zero)

lemma suffix_ilen_last [simp]:
    " suffix (ilen xs) xs = \<langle>(inth xs (ilen xs))\<rangle> "
by (induct xs) simp_all

lemma suffix_ilast:
   " suffix (ilen xs) xs = \<langle>ilast xs\<rangle>"
by simp

lemma suffix_suffix [simp]:
    " suffix i (suffix j xs) = suffix (i+j) xs "
proof 
  (induct xs arbitrary: i j)
  case (INil x)
  then show ?case by simp
  next
  case (ICons x1a xs)
  then show ?case 
   proof (cases i)
   case 0
   then show ?thesis 
     by auto
   next
   case (Suc nat)
   then show ?thesis 
     by (simp add: Nitpick.case_nat_unfold add.commute ICons)
   qed
qed

lemma prefix_suffix_ilen_code [code]:
    "ilen (prefix ia (suffix i xs)) =
     (if i \<le> ilen xs then
      (if ia\<le> ilen xs -i  then ia else (ilen xs) -i )
      else 0)"
using suffix_ilen_code by auto

lemma prefix_suffix_ilen [simp]:
    "ilen (prefix ia (suffix i xs)) = min ia (ilen xs -i)"
by auto

lemma prefix_suffix_ilen_good [simp]:
  assumes  "ia\<le> ilen xs -i"
           "i \<le> ilen xs"
  shows    " ilen (prefix ia (suffix i xs)) = ia "
using assms by auto

lemma prefix_suffix_ilen_bad_0:
  assumes  "i> ilen xs"
  shows    "ilen (prefix ia (suffix i xs)) = 0"
using assms by simp 

lemma prefix_suffix_ilen_bad_1 :
  assumes  "i \<le> ilen xs"
           " ia > ilen xs -i"
  shows    "ilen (prefix ia (suffix i xs)) = (ilen xs) -i  "
using assms by simp

lemma suffix_suffix_3:
 assumes "i>0"
         "ia<i"
         "i\<le> ilen xs"
 shows   "(suffix (i-ia) (suffix ((ilen xs)-i) xs)) = (suffix (((ilen xs)-ia)) xs)"
using assms by simp

lemma sub_zero_prefix :
    " sub 0 k xs = prefix k xs"
by (simp add:  sub_def) 

lemma sub_suffix : 
  assumes  "i < j"
           "j\<le> (ilen xs)-k"
  shows    "(sub (i+k) (j+k) xs) = (sub i j (suffix k xs))"
using assms by (simp add:  sub_def) 
  
lemma sub_prefix_suffix_0:
  assumes "0 \<le> i"
          "ia+i \<le> ilen xs"
  shows   "(sub i (i+ia) xs) = (prefix (ia) (suffix i xs))" 
using assms by (simp add:  sub_def) 
  
lemma sub_prefix_suffix:
 assumes "0 \<le> i"
         "i\<le>j"
         "j \<le> ilen xs"
 shows   "(sub i j xs) = (prefix (j-i) (suffix i xs))"
using assms by (simp add:  sub_def) 

lemma ilast_prefix:
 assumes "k \<le> ilen xs"
 shows   "ilast(prefix k xs) = (inth xs k)"
using assms prefix_ilen_good by fastforce

lemma ifirst_suffix:
 assumes "k \<le> ilen xs"
 shows   "ifirst(suffix k xs) = (inth xs k)"
by (simp add: assms)

lemma suffix_gr:
 assumes "i>ilen xs"
 shows   "suffix i xs = \<langle>ilast(xs)\<rangle>"
 by (metis add.commute assms suffix_ilast suffix_suffix 
     less_imp_add_positive suffix.simps(1))

lemma ilast_ifirst:
    " (ilast (prefix i xs)) = (ifirst (suffix i xs))"
proof -
  have 1: "(ilast (prefix i xs)) = (inth (prefix i xs) (ilen (prefix i xs)))" 
   by simp  
  have 2: "i \<le> ilen xs \<longrightarrow> ilen (prefix i xs) = i  "
   using prefix_ilen_good by blast
  have 3: "i>ilen xs \<longrightarrow> ilen (prefix i xs) = ilen xs"
    using prefix_ilen_bad by blast
  have 4: "i \<le> ilen xs \<longrightarrow> 
           (inth (prefix i xs) (ilen (prefix i xs))) = (inth xs i)"
    using ilast_prefix by auto
  have 5: "i>ilen xs \<longrightarrow> 
           (inth (prefix i xs) (ilen (prefix i xs))) = (inth xs (ilen xs))"
    using "3" by auto
  have 6: "(ifirst (suffix i xs)) = (inth (suffix i xs) 0)"
      by simp 
  have 7: "i \<le> ilen xs \<longrightarrow> 
           (inth (suffix i xs) 0) = (inth xs i)"
    by simp 
  have 8: "i>ilen xs \<longrightarrow> 
           (inth (suffix i xs) 0) = (inth xs (ilen xs)) "
     by (simp add: suffix_gr)
  show ?thesis using "4" "5" "8" by (metis "7" leI)
qed


lemma ilen_sub [simp]:
 assumes "k\<le> n"
         "n \<le> ilen xs"
 shows   "ilen(sub k n xs ) = (n-k)"
using assms 
by (metis  sub_def prefix_ilen_good suffix_ilen suffix_prefix_swap le_add_diff_inverse2)

lemma inth_sub [simp]:
 assumes "k\<le> n"
         "n \<le> ilen xs"
         "j\<le>n-k"
 shows    "inth(sub k n xs) j = (inth xs (k+j))"
proof -
 have 1: "inth(sub k n xs) j = inth (prefix (n-k) (suffix k xs)) j "
   by (simp add:  sub_def) 
 have 2: "n-k \<le> ilen (suffix k xs) "
    using  sub_def assms by auto
 have 3: "j\<le> (n-k)"
    using assms by auto
 have 4: "inth (prefix (n-k) (suffix k xs)) j =
          inth (suffix k xs) j"
    using "2" assms inth_prefix by blast  
 have 5: "inth (suffix k xs) j = inth xs (k+j)"
       using assms by auto 
 show ?thesis 
 by (simp add: "1" "4" "5")
qed       

lemma ilast_sub:
 assumes "k \<le> n"
         "n \<le> ilen xs"
 shows   "ilast (sub k n xs) = (inth xs n)"
by (simp add: assms)

lemma ifirst_sub:
 assumes "k \<le> n"
         "n \<le> ilen xs"
 shows   "ifirst (sub k n xs) = (inth xs k)"
by (simp add: assms)


lemma sub_sub:
 assumes "n1 \<le> n2"
         "n0 \<le> n4"
         "n2 \<le> n4-n0"
         "n4 \<le> n3"
         "n3 \<le> ilen xs"
 shows   "  (sub n1 n2 (sub n0 n3 xs)) = (sub n1 n2 (sub n0 n4 xs))"
proof -
 have 1: " ilen(sub n0 n3 xs) = n3 -n0"
     by (meson assms ilen_sub le_trans)
 have 2: " ilen (sub n1 n2 (sub n0 n3 xs)) = n2 -n1 "
    using ilen_sub assms by auto
 have 3: " ilen(sub n0 n4 xs) = n4 -n0"
     using assms ilen_sub le_trans by blast
 have 4: " ilen (sub n1 n2 (sub n0 n4 xs)) = n2 -n1"  
    by (simp add: "3" assms)
 have 5: " \<And>i. i\<le> (n3 -n0) \<longrightarrow> (inth (sub n0 n3 xs) i ) = (inth xs (n0+i))"
  using assms by auto
 have 6: " \<And>i . i\<le> (n2-n1) \<longrightarrow> (inth (sub n1 n2 (sub n0 n3 xs)) i) = (inth (sub n0 n3 xs) (n1+i))"
    using inth_sub assms
    by (metis "1" Nat.le_diff_conv2 le_trans)
 have 7: "\<And>i . i\<le> (n2-n1) \<longrightarrow> (inth (sub n1 n2 (sub n0 n3 xs)) i) = (inth xs (n0+(n1+i))) "
    using "5" "6" assms by auto
 have 8: " n0\<le> n4 \<and> n4 \<le> ilen xs"
   using assms le_trans by blast
 have 9: " \<And>i. i\<le> (n4 -n0) \<longrightarrow> (inth (sub n0 n4 xs) i ) = (inth xs (n0+i))"
   using "8" inth_sub by blast
 have 10: "\<And>i. i\<le> (n2-n1) \<longrightarrow> (inth (sub n1 n2 (sub n0 n4 xs)) i) = (inth (sub n0 n4 xs) (n1+i))"
   by (simp add: "3" assms)
 have 11: "\<And>i. i\<le> (n2-n1) \<longrightarrow> (inth (sub n1 n2 (sub n0 n4 xs)) i) = (inth xs (n0 +(n1+i))) "
    by (metis "3" Nat.le_diff_conv2 add.commute assms inth_sub le_trans)
 have 12: "\<And>i. i\<le> (n2-n1) \<longrightarrow> 
           (inth (sub n1 n2 (sub n0 n3 xs)) i) = (inth (sub n1 n2 (sub n0 n4 xs)) i) "
   by (simp add: "11" "7")
 from 12 2 4 show ?thesis by (simp add: interval_eq_inth_eq)
qed

lemma sub_sub_1:
 assumes "n1 \<le> n2"
         "n0 \<le> n3"
         "n2 \<le> n3-n0"
         "n3 \<le> ilen xs"
 shows   "(sub n1 n2 (sub n0 n3 xs)) = (sub (n0+n1) (n0+n2) xs)"
proof -
 have 1: "ilen(sub n1 n2 (sub n0 n3 xs)) = ilen(sub (n1+n0) (n2+n0) xs)"
   using assms by auto
 have 2: "\<And>i. i\<le> (n2-n1) \<longrightarrow> (inth (sub n1 n2 (sub n0 n3 xs)) i) = (inth xs (n0+(n1+i))) "
   using assms by auto 
 have 3: "\<And>i. i\<le> (n2-n1) \<longrightarrow> (inth (sub (n0+n1) (n0+n2) xs) i) = (inth xs (n0+(n1+i))) "
   using assms by (metis ab_semigroup_add_class.add_ac(1) diff_add_inverse inth_sub le_Suc_ex le_add1)
 have 4: "\<And>i. i\<le> (n2-n1) \<longrightarrow> (inth (sub n1 n2 (sub n0 n3 xs)) i) = (inth (sub (n0+n1) (n0+n2) xs) i)"
   by (simp add: "2" "3")
 show ?thesis 
 by (metis "1" "4" add.commute assms interval_eq_inth_eq ilen_sub)
qed

lemma suf_first_upto:
assumes "(\<exists>i<k. f (suffix i xs))"
        "k \<le> ilen xs +1"
shows   "(\<exists>i<k. f (suffix i xs) \<and>
           (\<forall>j<i . \<not> (f (suffix j xs))))"
using assms
proof (induct xs arbitrary:k)
case (INil x)
then show ?case by auto
next
case (ICons x1a xs)
then show ?case 
  proof -
   have 0: "k=0 \<longrightarrow> ( \<exists>i<k. f (suffix i (x1a \<odot> xs)) \<and> (\<forall>j<i. \<not> f (suffix j (x1a \<odot> xs)))) "
     using ICons.prems(1) by blast
   have 1: "\<And>n. k = (Suc n) \<longrightarrow> (\<exists>i< (Suc n). f (suffix i (x1a \<odot> xs)) \<and> 
            (\<forall>j<i. \<not> f (suffix j (x1a \<odot> xs)))) =
            ( f (x1a \<odot> xs) \<or>
            (\<exists>i. 1\<le> i \<and> i<(Suc n) \<and> f (suffix i (x1a \<odot> xs)) \<and> 
                 (\<forall>j<i. \<not> f (suffix j (x1a \<odot> xs)))))
           " by auto 
             (metis Nitpick.case_nat_unfold less_antisym not_le_imp_less not_less0)
  have 2: "\<And>n. k = (Suc n) \<longrightarrow> (\<exists>i. 1\<le> i \<and> i<(Suc n) \<and> f (suffix i (x1a \<odot> xs)) \<and> 
                 (\<forall>j<i. \<not> f (suffix j (x1a \<odot> xs)))) =
            (\<exists>i.  i<n  \<and> f (suffix (Suc i) (x1a \<odot> xs)) \<and> 
                 (\<forall>j<(Suc i). \<not> f (suffix j (x1a \<odot> xs))))
           "         
        by auto
           (metis Suc_le_D nat.case(2) not_less not_less_eq_eq)
  have 3: "\<And>n. k= (Suc n) \<longrightarrow> (\<exists>i.  i<n  \<and> f (suffix (Suc i) (x1a \<odot> xs)) \<and> 
                 (\<forall>j<(Suc i). \<not> f (suffix j (x1a \<odot> xs)))) =
           (\<exists>i.  i<n  \<and> f (suffix i xs) \<and> 
                 \<not>(f (suffix 0 (x1a\<odot>xs))) \<and>
                 (\<forall>j. 1 \<le> j \<and> j<(Suc i) \<longrightarrow> \<not> f (suffix j (x1a \<odot> xs))))
          " 
         by (metis suffix_suc le_add1 less_Suc_eq_0_disj plus_1_eq_Suc)
  have 4: "\<And>n. k= (Suc n) \<longrightarrow> (\<exists>i.  i<n  \<and> f (suffix i xs) \<and> 
                 \<not>(f (suffix 0 (x1a\<odot>xs))) \<and>
                 (\<forall>j. 1 \<le> j \<and> j<(Suc i) \<longrightarrow> \<not> f (suffix j (x1a \<odot> xs)))) =
           ( \<not>(f (x1a\<odot>xs)) \<and> 
              (\<exists>i.  i<n  \<and> f (suffix i xs) \<and> 
                 (\<forall>j.  j<i \<longrightarrow> \<not> f (suffix (Suc j) (x1a \<odot> xs)))) )
            " using ICons.hyps ICons.prems by (auto, auto simp add: less_Suc_eq_0_disj)
  have 5: "\<And>n. k =(Suc n) \<longrightarrow> ( \<not>(f (x1a\<odot>xs)) \<and> 
              (\<exists>i.  i<n  \<and> f (suffix i xs) \<and> 
                 (\<forall>j.  j<i \<longrightarrow> \<not> f (suffix (Suc j) (x1a \<odot> xs)))) ) =
             ( \<not>(f (x1a\<odot>xs)) \<and> 
              (\<exists>i<n.   f (suffix i xs) \<and> 
                 (\<forall>j<i . \<not> f (suffix j  xs))) ) "
      by auto  
  have 6: "\<And>n. k =(Suc n) \<longrightarrow>  (\<exists>i< (Suc n). f (suffix i (x1a \<odot> xs)) \<and> 
            (\<forall>j<i. \<not> f (suffix j (x1a \<odot> xs)))) = 
            ( f (x1a \<odot> xs) \<or> ( \<not>(f (x1a\<odot>xs)) \<and> 
              (\<exists>i<n.   f (suffix i xs) \<and> 
                 (\<forall>j<i . \<not> f (suffix j  xs))) )  )"  
       using "1" "2" "3" "4" by auto
  have 7: "\<And>n. k =(Suc n) \<longrightarrow> ( f (x1a \<odot> xs) \<or> ( \<not>(f (x1a\<odot>xs)) \<and> 
              (\<exists>i<n.   f (suffix i xs) \<and> 
                 (\<forall>j<i . \<not> f (suffix j  xs))) )  ) =
               (  f (x1a \<odot> xs) \<or> (\<exists>i<n.   f (suffix i xs) \<and> 
                 (\<forall>j<i . \<not> f (suffix j  xs))))
              " 
         by auto
  have 8: "\<And>n. k= (Suc n) \<longrightarrow> n \<le> ilen xs +1"
      using ICons.prems(2) by auto
  have 9: "\<And>n. k =(Suc n) \<longrightarrow> (  f (x1a \<odot> xs) \<or> (\<exists>i<n.   f (suffix i xs) \<and> 
                 (\<forall>j<i . \<not> f (suffix j  xs)))) =
             (  f (x1a \<odot> xs) \<or> (\<exists>i<n.   f (suffix i xs))) "
       using 7 8 ICons.hyps by fastforce
  have 10: "\<And>n. k= (Suc n) \<longrightarrow> (\<exists>i<k. f (suffix i (x1a \<odot> xs))) =
            (f (x1a \<odot> xs)  \<or> (\<exists>i<n. f (suffix i (xs))))
          " 
       using Nitpick.case_nat_unfold less_Suc_eq_0_disj by auto 
  have 11: "\<And>n. k =(Suc n) \<longrightarrow>  
             ( \<exists>i<k. f (suffix i (x1a \<odot> xs)) \<and> (\<forall>j<i. \<not> f (suffix j (x1a \<odot> xs))))  "  
          using 9 10 6 ICons.prems(1) by force 
   show ?thesis 
  using 0 11 less_imp_Suc_add by blast
 qed
qed

subsubsection \<open>iapp\<close>

lemma ilen_snoc_1:
 "ilen l > 0 = (\<exists> x ls. l = ls \<ominus> \<langle>x\<rangle>)"
proof (induct l)
case (INil x)
then show ?case by fastforce
next
case (ICons x1a l)
then show ?case 
   by (metis Suc_eq_plus1 iapp_ICons iapp_INil interval.exhaust ilen_iapp nat.simps(3) neq0_conv)
qed 


lemma prefix_iapp [simp]:
 "prefix (ilen xs -k) (xs \<ominus> ys) = prefix (ilen xs -k) xs"
proof 
  (induct xs arbitrary:k)
  case (INil x)
  then show ?case by simp
  next
  case (ICons x1a xs)
  then show ?case 
    proof (cases k)
    case 0
    then show ?thesis 
       by (metis ICons.hyps diff_zero iapp_ICons prefix_suc ilen.simps(2) plus_1_eq_Suc)
    next
    case (Suc nat)
    then show ?thesis 
       by (auto simp add: ICons.hyps Nitpick.case_nat_unfold)
    qed   
qed


lemma prefix_iapp2 [simp]:
 "prefix (ilen xs + k +1) (xs \<ominus> ys) = xs \<ominus> prefix k ys"
proof 
 (induct xs arbitrary: k)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
   proof (cases k)
   case 0
   then show ?thesis 
      by auto
         (metis ICons.hyps Suc_eq_plus1 add.right_neutral prefix_zero_ifirst)
   next
   case (Suc nat)
   then show ?thesis 
      by auto
         (metis ICons.hyps Suc_eq_plus1 add_Suc_right)
   qed
qed


lemma suffix_iapp [simp]:
 "suffix (ilen xs +m +1) (xs \<ominus> ys) = suffix (m) ys"
proof 
  (induct xs arbitrary:m)
  case (INil x)
  then show ?case by simp
  next
  case (ICons x1a xs)
  then show ?case 
   proof (cases m)
   case 0
   then show ?thesis 
    by  auto
        (metis ICons.hyps One_nat_def Suc_eq_plus1 suffix_zero plus_1_eq_Suc 
         semiring_normalization_rules(23))
   next
   case (Suc nat)
   then show ?thesis 
    by auto
       (metis ICons.hyps Suc_eq_plus1 suffix_suffix)
   qed
qed

lemma suffix_iapp2 [simp]: 
 "(suffix (ilen xs - k)  xs) \<ominus> ys = suffix (ilen xs - k) (xs \<ominus> ys)"
proof 
 (induct xs)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
      by (auto simp add: Nitpick.case_nat_unfold)
qed

lemma iapp_assoc [simp]:
 " (xs \<ominus> ys) \<ominus> zs = xs \<ominus> ( ys \<ominus> zs ) "
by (induct xs) simp_all

lemma iapp_inth:
 " inth (xs \<ominus> ys) k = (if k \<le> ilen xs 
                      then (inth xs k) 
                      else (inth ys (k - (ilen xs) -1)) )"
proof 
  (induct xs arbitrary: k)
  case (INil x)
  then show ?case 
    proof (cases k)
    case 0
    then show ?thesis by simp
    next
    case (Suc nat)
    then show ?thesis by simp
    qed
  next
  case (ICons x1a xs)
  then show ?case 
    proof (cases k)
    case 0
    then show ?thesis by simp
    next
    case (Suc nat)
    then show ?thesis by (simp add: ICons.hyps)
    qed
qed


lemma irev_iapp [simp]:
 " irev (xs \<ominus> ys) = (irev ys) \<ominus> (irev xs)"
by (induct xs) simp_all

lemma ilast_iapp [simp]:
 "ilast(xs \<ominus> \<langle>x\<rangle>) = x"
by (induct xs) simp_all

lemma ilast_iapp2 [simp]:
 "ilast (xs \<ominus> ys) = ilast ys"
by (induct xs arbitrary: ys) simp_all

lemma ifirst_iapp [simp]:
 "ifirst (\<langle>x\<rangle> \<ominus> xs) = x"
by (induct xs) simp_all

lemma ifirst_iapp2 [simp]:
 "ifirst(xs \<ominus> ys) = ifirst xs"
by (induct xs arbitrary: ys) simp_all

lemma iapp_not_INil [simp]:
 "xs \<ominus> ys \<noteq> \<langle>x\<rangle>"
by (induct xs arbitrary: ys) simp_all

lemma iapp_eq_iapp_conv [simp]:
 assumes "ilen xs = ilen ys \<or> ilen us = ilen vs"
 shows   "(xs \<ominus> us = ys \<ominus> vs) = (xs =ys \<and> us=vs)"
using assms 
proof 
  (induct xs arbitrary: ys)
  case (INil x)
  then show ?case 
    proof (cases ys)
    case (INil x1)
    then show ?thesis by simp
    next
    case (ICons x21 x22)
    then show ?thesis using INil.prems by auto
    qed
  next
  case (ICons x1a xs)
  then show ?case 
    proof (cases ys)
    case (INil x1)
    then show ?thesis using ICons.prems by auto
    next
    case (ICons x21 x22)
    then show ?thesis using ICons.hyps ICons.prems by auto
    qed
qed

lemma iapp_eq_iapp_conv2: 
 "(xs \<ominus> ys = zs \<ominus> ts) =
  (\<exists> us. xs = zs \<ominus> us \<and> us \<ominus> ys = ts \<or>
         xs =zs \<and> ys=ts \<or>
         xs \<ominus> us = zs \<and> ys = us \<ominus> ts)"
proof 
 (induct xs arbitrary: ys zs ts)
case (INil x)
then show ?case 
 proof (cases zs)
 case (INil x1)
 then show ?thesis by simp
 next
 case (ICons x21 x22)
 then show ?thesis by simp
 qed
next
case (ICons x1a xs)
then show ?case  
 proof (cases zs)
 case (INil x1)
 then show ?thesis by simp
 next
 case (ICons x21 x22)
 then show ?thesis by (auto simp add: ICons.hyps)
 qed
qed

lemma same_iapp_eq[iff, induct_simp]:
 "(xs \<ominus> ys = xs \<ominus> zs) = (ys = zs)"
using suffix_iapp by (metis suffix_zero)

lemma iapp_eq_conv[iff]:
 "(xs \<ominus> \<langle>x\<rangle> = ys \<ominus> \<langle>y\<rangle>) = (xs = ys \<and> x = y)"
by auto

lemma iapp_same_eq[iff, induct_simp]:
 "(ys \<ominus> xs = zs \<ominus> xs) = (ys = zs)"
by auto

lemma suffix1_iapp:
 "suffix 1 (xs \<ominus> ys) = (case xs of \<langle>x\<rangle> \<Rightarrow> ys | x \<odot> zs \<Rightarrow> zs \<ominus> ys)"
by (cases xs) simp_all


lemma ICons_eq_iapp_conv:
 "(x\<odot> xs = ys \<ominus> zs) =
  ((\<langle>x\<rangle> = ys \<and> xs =zs) \<or> (\<exists> ys'. x\<odot>ys' =  ys \<and>  xs = ys' \<ominus> zs))"
by (cases ys) simp_all

lemma iapp_eq_ICons_conv:
 "(ys \<ominus> zs = x \<odot> xs) =
  ((\<langle>x\<rangle> = ys \<and> zs =xs) \<or> (\<exists> ys'. ys = x\<odot>ys'  \<and>  ys' \<ominus> zs = xs))"
by (cases ys) auto

lemma ICons_eq_iappI:
assumes "x \<odot> xs1 = ys"
        "xs = xs1 \<ominus> zs"
shows   "x \<odot> xs = ys \<ominus> zs"
using assms by auto

lemma iapp_eq_iappI:
assumes "xs \<ominus> xs1 = zs"
        "ys = xs1 \<ominus> us"
shows   "xs \<ominus> ys = zs \<ominus> us"
using assms by auto


lemma iapp_gr_zero:
 " ilen (xs \<ominus> ys) >0"
by auto


lemma iapp_prefix_suffix:
assumes "i+1 \<le> ilen xs"
        "ilen xs >0 "
shows "xs = (prefix i xs) \<ominus> (suffix (i+1) xs)"
using assms 
proof (induct xs arbitrary:i)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
   proof (cases i)
   case 0
   then show ?thesis by auto
   next
   case (Suc nat)
   then show ?thesis using ICons.hyps ICons.prems(1) by auto
   qed
qed 


subsubsection \<open>Reverse\<close>


lemma irev_irev_ident [simp]: 
 " irev (irev xs) = xs"
by (induct xs) auto

lemma irev_swap :
 " ((irev xs) = ys) = (xs = irev ys)"
by auto

lemma irev_singleton_conv [simp]:
 "( irev xs = \<langle>x\<rangle>) = (xs = \<langle>x\<rangle>)"
by (metis irev_irev_ident irev.simps(1)) 

lemma single_irev_conv [simp]:
 " (\<langle>x\<rangle> = irev xs) = (\<langle>x\<rangle> = xs)"
by (metis irev_irev_ident irev.simps(1))

lemma irev_is_irev_conv [iff]:
 "(irev xs = irev ys) = (xs = ys)"
proof
 (induct xs arbitrary: ys)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
 using irev_swap by force
qed

lemma irev_induct [case_names INil snoc]:
assumes "\<And> y. P \<langle>y\<rangle>"
        "\<And>x xs. P xs \<Longrightarrow> P(xs \<ominus> \<langle>x\<rangle>)"
shows   "P xs"
using assms
using interval.induct[of "\<lambda> xs. P (irev xs)" "irev xs"]
by simp
  

lemma irev_exhaust [case_names INil snoc]:
assumes "\<And>x. xs = \<langle>x\<rangle> \<Longrightarrow> P"
        "\<And>ys y. xs = ys \<ominus> \<langle>y\<rangle> \<Longrightarrow> P"
shows   "  P"
using assms 
by (induct xs rule:irev_induct) auto

lemmas irev_cases = irev_exhaust

lemma interval_irev_eq_ICons_iff [iff]:
 "(irev xs = y\<odot>ys) = (xs = (irev ys) \<ominus> \<langle>y\<rangle>)"
by (metis irev_irev_ident irev.simps(2)) 


lemma irev_iapp_ICons:
 " irev (xs \<ominus> \<langle>x\<rangle>) = x\<odot>irev xs"
by (cases xs) simp_all

lemma ilast_irev:
 "ilast (irev xs) = ifirst xs"
proof (cases xs)
case (INil x1)
then show ?thesis 
 by simp
next
case (ICons x21 x22)
then show ?thesis 
 by  (metis ilast_iapp inth_zero irev.simps(2))
qed

lemma ifirst_irev:
 "ifirst (irev xs) = ilast xs"
proof 
  (induct xs)
  case (INil x)
  then show ?case by simp
  next
  case (ICons x1a xs)
  then show ?case 
    by (metis ifirst_iapp2 inth_last irev.simps(2))
qed


lemma irev_inth:
assumes "k\<le> ilen (irev xs)"
shows   "  (inth  (irev xs) k) = (inth  xs ((ilen xs) -k))"
using assms 
proof 
  (induct xs)
  case (INil x)
  then show ?case by simp
  next
  case (ICons x1a xs)
  then show ?case 
    proof (cases k)
    case 0
    then show ?thesis 
     by auto
        (metis ICons.hyps diff_zero ilen_gr_zero)
    next
    case (Suc nat)
    then show ?thesis 
     using ICons.hyps Suc_diff_le by (auto simp add: iapp_inth) fastforce
    qed
qed


lemma irev_prefix:
assumes "k\<le> ilen xs"
shows   "  irev( prefix k xs) = suffix ((ilen xs) - k) (irev xs)  "
proof 
 (induct xs arbitrary: k)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
   proof (cases k)
   case 0
   then show ?thesis 
    by auto 
       (metis Suc_eq_plus1 add.right_neutral ilast_iapp ilen_iapp 
        irev_ilen suffix_ilast ilen.simps(1))
   next
   case (Suc nat)
   then show ?thesis 
    by auto
       (metis ICons.hyps irev_ilen suffix_iapp2)
   qed
qed


lemma irev_suffix:
assumes "k\<le> ilen xs"
shows   " irev( suffix k xs) = prefix ((ilen xs) - k) (irev xs)  "
using assms
proof 
 (induct xs arbitrary: k)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case by (simp add: irev_prefix irev_swap)
qed

lemma irev_sub:
 assumes " 0 \<le> i "
         " i\<le>j "
         " j \<le> ilen xs"
 shows   " irev (sub i j xs) = sub ((ilen xs) -j) ((ilen xs) - i) (irev xs)"
using assms 
proof -
 have 1: "irev (sub i j xs) = irev (prefix (j-i) (suffix i xs))" 
   using assms sub_prefix_suffix by (simp add: sub_prefix_suffix)
 have 2: "irev (prefix (j-i) (suffix i xs)) = suffix ((ilen xs) - j) (irev (suffix i xs))"
   using assms irev_prefix[of "j-i" "suffix i xs"]  by auto
 have 3: "suffix ((ilen xs) - j) (irev (suffix i xs)) = 
           suffix ((ilen xs) -j) (prefix ((ilen xs) - i) (irev xs)) "
    using assms irev_suffix[of "i" "xs"] by auto
 have 4: "suffix ((ilen xs) -j) (prefix ((ilen xs) - i) (irev xs)) =
          sub ((ilen xs) -j) ((ilen xs) - i) (irev xs) "
    using assms by (simp add: diff_le_mono2 sub_prefix_suffix suffix_prefix_swap)
 from 1 2 3 4 show ?thesis by auto
qed


subsubsection \<open>Induction rule\<close>

lemma interval_ilen_induct:
assumes "(\<And>xs. \<forall>ys. ilen ys < ilen xs \<longrightarrow> P ys \<Longrightarrow> P xs)"
shows   "  P xs"
using assms by (fact measure_induct)

lemma interval_induct_12:
 assumes "\<And>x. P \<langle>x\<rangle>"
         "\<And>x y. P \<langle>x,y\<rangle>"
         "\<And>x y zs. P (y\<odot>zs) \<Longrightarrow> P (x\<odot>y\<odot>zs)"
 shows   "P xs" 
using assms 
proof (induction xs)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
by auto
   (metis add_cancel_right_right ilen_ICons_1 INil_ilen le_add1 le_eq_less_or_eq 
    less_add_same_cancel1)
qed

lemma interval_induct2 [consumes 1, case_names INil ICons]:
assumes "ilen xs = ilen ys"
        " (\<And> x y. P \<langle>x\<rangle> \<langle>y\<rangle>) "  
        "(\<And> x xs y ys. ilen xs = ilen ys \<Longrightarrow> P xs ys \<Longrightarrow> P (x\<odot>xs) (y\<odot>ys))"
 shows " P xs ys"
using assms 
proof (induction xs arbitrary: ys)
case (INil x)
then show ?case 
by (metis INil_ilen)
next
case (ICons x1a xs)
then show ?case 
by (metis ilen_ICons_1 ilen.simps(2) nat.simps(1) plus_1_eq_Suc)
qed



subsubsection \<open>Imap\<close>

lemma imap_ext:
 assumes "(\<forall>x. x \<in> iset xs \<longrightarrow> f x = g x)"
 shows   " imap f xs = imap g xs"
using assms
by (induct xs) simp_all


lemma imap_ident [simp]:
 "imap (\<lambda>x. x) = (\<lambda>xs .xs)"
 proof (rule ext) 
  show " \<And>xs. imap (\<lambda>x. x) xs = xs " 
   by (simp add: interval.map_ident)
qed


lemma imap_iapp [simp]:
 "imap f (xs \<ominus> ys) = imap f xs \<ominus> imap f ys"
by (induct xs) auto

lemma imap_imap [simp]:
 "imap f (imap g xs) = imap (f \<circ> g) xs"
by (simp add: interval.map_comp)

lemma imap_comp_imap [simp]:
 "((imap f) \<circ> (imap g)) = imap(f \<circ> g)"
by (rule ext) simp

lemma irev_imap:
 "irev (imap f xs) = imap f(irev xs)"
by (induct xs) auto

lemma imap_eq_conv [simp]:
 "(imap f xs = imap g xs) = (\<forall> x\<in> iset xs. (f x) = (g x))"
by (induct xs) auto

lemma imap_cong [fundef_cong]:
 assumes "xs = ys"
         "(\<forall>x. x \<in> iset ys \<longrightarrow> f x = g x)"
 shows   " imap f xs = imap g ys"
using assms by simp

lemma imap_injective:
 assumes "imap f xs = imap f ys"
         " inj f"
 shows  " xs = ys"
using assms by (meson injD interval.inj_map)

lemma inj_imap_eq_imap [simp]:
 assumes "inj f"
 shows   "  (imap f xs = imap f ys)  = (xs = ys)"
using assms by (blast dest: imap_injective)

lemma inj_imapI:
 assumes "inj f"
 shows   " inj (imap f)"
using assms interval.inj_map by blast

lemma inj_imapD:
 assumes "inj (imap f)"
 shows   "  inj f"
using assms by (metis inj_def interval.map(1) interval.simps(1))

lemma inj_imap[iff]:
 "inj (imap f) = inj f"
by (blast dest: inj_imapD intro: inj_imapI)

lemma imap_idI: 
 assumes "(\<forall>x. x \<in> iset xs \<longrightarrow> f x = x)"
 shows   "  imap f xs = xs"
using assms by (induct xs) auto

lemma imap_is_INil_conv[iff]:
 "(imap f xs = \<langle>x\<rangle>) = (\<exists> y. xs = \<langle>y\<rangle> \<and> f y = x)"
proof (cases xs)
case (INil x1)
then show ?thesis by simp
next
case (ICons x21 x22)
then show ?thesis by simp
qed

lemma INil_is_imap_conv [iff]:
 "(\<langle>x\<rangle> = imap f xs) = (\<exists> y. \<langle>y\<rangle> = xs \<and> f y = x)"
proof (cases xs)
case (INil x1)
then show ?thesis by auto
next
case (ICons x21 x22)
then show ?thesis by auto
qed

lemma imap_eq_ICons_conv:
 "(imap f xs = y\<odot>ys) = (\<exists> z zs. xs = z\<odot>zs \<and> f z =y \<and> imap f zs = ys)"
by (cases xs) auto

lemma ICons_eq_imap_conv:
 "(y\<odot>ys= imap f xs ) = (\<exists> z zs. z\<odot>zs=xs \<and> f z =y \<and> ys = imap f zs)"
by (cases xs) auto

lemma ex_imap_conv:
  "(\<exists>xs. ys = imap f xs) = (\<forall>y \<in> iset ys. \<exists>x. y = f x)"
by (induct ys) (auto simp add: ICons_eq_imap_conv)


functor imap: imap
by (simp_all add: id_def)

declare imap.id [simp]


lemma ifirst_imap:
 "ifirst (imap f xs) = f (ifirst xs)"
by (cases xs) simp_all


lemma ilast_imap:
 "ilast (imap f xs) = f (ilast xs)"
proof (cases xs rule: irev_cases)
case (INil x)
then show ?thesis by simp
next
case (snoc ys y)
then show ?thesis by (simp add: iapp_inth)
qed  


lemma imap_tail:
shows   " imap f (suffix 1 xs) = (suffix 1 (imap f xs))"
by (cases xs) simp_all

lemma imap_eq_imp_ilen_eq:
 assumes "imap f xs = imap g ys"
 shows "ilen xs = ilen ys"
using assms
proof (induct ys arbitrary: xs)
case (INil x)
then show ?case by auto
next
case (ICons x1a ys)
then show ?case by (metis ilen_imap)
qed

lemma iset_imap [simp]: 
  "iset (imap f xs) = f`(iset xs)"
by (induct xs) auto

lemma imap_inj_on:
 assumes imap: "imap f xs = imap f ys" and
         inj: "inj_on f (iset xs \<union> iset ys)"
  shows "xs =ys"
using imap_eq_imp_ilen_eq[OF imap] assms
proof (induction rule: interval_induct2)
case (INil x y)
then show ?case by auto
next
case (ICons x xs y ys)
then show ?case 
  by (simp add: inj_on_Un_image_eq_iff inj_on_eq_iff)
qed

lemma inj_on_imap_eq_imap:
assumes "inj_on f (iset xs \<union>  iset ys)"
shows   "(imap f xs = imap f ys) = (xs = ys)"
using assms by (blast dest:imap_inj_on)

lemma inj_on_imapI:
 assumes "inj_on f (\<Union> (iset ` A))"
 shows   " inj_on (imap f) A"
using assms by (blast intro:inj_onI dest:inj_onD imap_inj_on)

lemma imap_prefix:
 assumes "k \<le> ilen xs"
 shows   " imap f (prefix k xs) = prefix k (imap f xs)"
using assms
proof (induction xs arbitrary: k)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
  proof (cases k)
  case 0
  then show ?thesis by auto
  next
  case (Suc nat)
  then show ?thesis using ICons.IH ICons.prems by auto
  qed
qed


lemma imap_suffix:
 assumes "k \<le> ilen xs"
 shows   " imap f (suffix k xs) = suffix k (imap f xs)"
using assms
proof (induction xs arbitrary: k)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
 proof (cases k)
 case 0
 then show ?thesis by auto
 next
 case (Suc nat)
 then show ?thesis using ICons.IH ICons.prems by auto
 qed
qed


subsubsection \<open>index sequence\<close>

lemma idx_less:
    assumes  " index_sequence x idx"
             " n+k<ilen idx" 
    shows "  inth idx n < inth idx (Suc(n+k)) "
using index_sequence_def assms by (induct k) auto

lemma idx_less_eq:
 assumes "index_sequence x l"
         "k\<le>j"
         "j\<le>ilen l"
 shows    "inth l k \<le> inth l j"
using assms 
proof (cases "k=j")
 show "index_sequence x l \<Longrightarrow> k \<le> j \<Longrightarrow> j \<le> ilen l \<Longrightarrow> k = j \<Longrightarrow> inth l k \<le> inth l j" 
   by blast
 show "index_sequence x l \<Longrightarrow> k \<le> j \<Longrightarrow> j \<le> ilen l \<Longrightarrow> k \<noteq> j \<Longrightarrow> inth l k \<le> inth l j " 
   by (metis Suc_le_lessD idx_less le_SucE le_eq_less_or_eq le_zero_eq 
       ordered_cancel_comm_monoid_diff_class.add_diff_inverse zero_induct)
qed


lemma idx_mono:
 assumes "index_sequence x l"
 shows    "mono (\<lambda> x. inth l x)"
proof - 
 have 1: "\<forall>x y. x \<le> y \<longrightarrow> inth l x \<le> inth l y"
   proof 
     fix x 
     show "\<forall>y\<ge>x. inth l x \<le> inth l y " 
     proof
      fix y
      show "x \<le> y \<longrightarrow> inth l x \<le> inth l y" 
       proof -
        have 2: "x\<le> y \<and> y \<le> ilen l \<Longrightarrow> inth l x \<le> inth l y" 
         using assms idx_less_eq by blast
        have 3: "x\<le> y  \<and> x>ilen l \<Longrightarrow> inth l x \<le> inth l y" 
           using assms inth_last_stutter 
           by (metis "2" le_cases less_imp_add_positive ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
        have 4: "x\<le> y  \<and> x\<le>ilen l \<and> y > ilen l \<Longrightarrow> inth l x \<le> inth l y"   
           by (metis assms idx_less_eq inth_last_stutter less_imp_add_positive order_refl)
        show ?thesis 
         using "2" "3" "4" not_less by blast
       qed
      qed
     qed
    show ?thesis 
    by (simp add: "1" monoI)
qed


lemma idx_less_last :
 assumes " index_sequence x idx"
         " i<ilen idx "
         " i+(ilen idx - (i+1))< ilen idx" 
 shows   " inth idx i < inth idx (Suc(i+(ilen idx -(i+1))))"
using assms idx_less by blast

lemma idx_less_last_1:
 assumes  " index_sequence x idx"
          " i<ilen idx" 
 shows    " inth idx i < inth idx (ilen idx) "
using assms idx_less_last by auto

lemma idx_greater_first:
 assumes  " index_sequence x idx"
          " 0 < i"
          " i\<le> ilen idx"
 shows    " x < inth idx i"
using assms 
proof -
 have 1: " inth idx 0 = x "
    using assms by (simp add: index_sequence_def)
 have 2: "\<forall> i. 0 < i \<and> i\<le> ilen idx \<longrightarrow> (inth idx 0) < (inth idx i) " 
     proof 
       fix i
       show "0 < i \<and> i \<le> ilen idx \<longrightarrow> inth idx 0 < inth idx i " 
        by (meson Suc_leI assms(1) dual_order.strict_trans1 index_sequence_def idx_less_eq)
     qed
 show ?thesis 
 using "1" "2" assms(2) assms(3) by blast 
qed

lemma idx_ICons:
    " index_sequence y (x\<odot>ls) = 
       (x=y \<and> x<inth ls 0 \<and> index_sequence (inth ls 0) ls)"
using less_Suc_eq_0_disj by (simp add: index_sequence_def) auto

lemma idx_shift_mono:
    "mono (shift k)" 
by (simp add:  shift_def mono_def)     

lemma idx_expand:
 assumes  "index_sequence 0 l"
          "(inth l (ilen l)) = (ilen xs)"
          " i< (ilen l)"
 shows    "  (inth l i) \<le>  (inth l (i+1)) \<and> (inth l (i+1)) \<le> (ilen xs)"
using assms
by (metis Suc_eq_plus1 Suc_lessI add.right_neutral idx_less idx_less_last_1 
    less_imp_le_nat order_refl)

lemma idx_shift_idx [simp]:
    " ( index_sequence (x+k) (imap (shift k)  idx)) = (index_sequence x idx) " 
by (simp add:  shift_def index_sequence_def inth_imap)    

lemma idx_shiftm :
  assumes "index_sequence k lsk"
 shows    "index_sequence 0 (imap (shiftm k) lsk) \<and> k \<le> (inth lsk 0)"
using assms
proof (auto simp add:  index_sequence_def shiftm_def inth_imap )
  show "\<And>n. \<forall>n<ilen lsk. inth lsk n < inth lsk (Suc n) \<Longrightarrow>
         k = inth lsk 0 \<Longrightarrow>
         n < ilen lsk \<Longrightarrow> inth lsk n - inth lsk 0 < inth lsk (Suc n) - inth lsk 0" 
  by (metis assms diff_less_mono idx_greater_first le_eq_less_or_eq neq0_conv)
qed  

lemma idx_shiftm_a :
  assumes "index_sequence 0 (imap (shiftm k) lsk)"
          "k \<le> (inth lsk 0)"
 shows    "index_sequence k lsk"
using assms 
by (auto simp add: index_sequence_def shiftm_def inth_imap) 

lemma idx_shiftm_b :
 "index_sequence k lsk = (index_sequence 0 (imap (shiftm k) lsk) \<and> k \<le> (inth lsk 0))"
using idx_shiftm idx_shiftm_a by blast

lemma idx_shiftm_c :
 "index_sequence (inth lsk 0) lsk = index_sequence 0 (imap (shiftm (inth lsk 0)) lsk)"
using idx_shiftm_b by blast

lemma lsk_ls :
    "(index_sequence k (lsk) \<and> lsk = imap (shift k) ls \<and> index_sequence 0 (ls) ) =
      (index_sequence k (lsk) \<and> ls = imap (shiftm k) lsk \<and> index_sequence 0 (ls) ) "
proof (simp add: interval_eq_inth_eq index_sequence_def shift_def shiftm_def inth_imap)
  show "(inth lsk 0 = k \<and>
     (\<forall>n<ilen lsk. inth lsk n < inth lsk (Suc n)) \<and>
     ilen lsk = ilen ls \<and>
     (\<forall>i\<le>ilen lsk. inth lsk i = inth ls i + k) \<and>
     inth ls 0 = 0 \<and> (\<forall>n<ilen ls. inth ls n < inth ls (Suc n))) =
    (inth lsk 0 = k \<and>
     (\<forall>n<ilen lsk. inth lsk n < inth lsk (Suc n)) \<and>
     ilen ls = ilen lsk \<and>
     (\<forall>i\<le>ilen ls. inth ls i = inth lsk i - k) \<and>
     inth ls 0 = 0 \<and> (\<forall>n<ilen ls. inth ls n < inth ls (Suc n)))" (is "?L=?R")
    proof rule
     show "?L \<Longrightarrow> ?R"
    by (metis (no_types, lifting) add_diff_cancel_right')
     show "?R \<Longrightarrow> ?L "
     by (metis Suc_le_lessD add.commute diff_is_0_eq nat_le_linear not0_implies_Suc 
         ordered_cancel_comm_monoid_diff_class.add_diff_inverse zero_order(3)) 
   qed
qed

lemma idx_link_shiftm:
    "(index_sequence k (lsk) \<and> ls = imap (shiftm k) lsk ) =
     (index_sequence k (lsk) \<and> ls = imap (shiftm k) lsk \<and> 
      index_sequence 0 (ls) \<and> (ilen ls) =(ilen lsk)) "
using idx_shiftm using ilen_imap by blast
    
lemma idx_link:
    "(lsk = imap (shift k) ls \<and> index_sequence 0 (ls) ) =
     (lsk = imap (shift k) ls \<and> index_sequence k (lsk) \<and> index_sequence 0 (ls)\<and> 
      (ilen ls) =(ilen lsk))"
by (metis add.left_neutral idx_shift_idx ilen_imap)

lemma idx_bound_0 :
 assumes "index_sequence 0 ls"
         "inth ls (ilen ls) = ilen (suffix k xs)"
         " i\<le>ilen ls"
 shows   " inth ls i \<le>  ilen (suffix k xs) "
using assms
by (metis eq_iff idx_less_last_1 le_neq_implies_less less_imp_le_nat) 
 
lemma idx_bound_1:
  "(index_sequence 0 (ls) \<and> (inth (ls) (ilen (ls))) = (ilen (suffix k xs))) =
     (index_sequence 0 (ls) \<and> (inth (ls) (ilen (ls))) = (ilen (suffix k xs)) \<and>
     (\<forall>i. (i\<le>ilen ls) \<longrightarrow> ((inth ls (i)) \<le> (ilen (suffix k xs)))) )" 
using idx_bound_0 by blast


lemma idx_less_equal:
assumes "index_sequence 0 l"
        "(inth l (ilen l)) = ilen xs"
        "i \<le> ilen l"
        "n \<le> ilen l"
shows   "\<forall> j . j\<le>n \<longrightarrow> inth l j \<le> inth l n"
using assms 
using idx_less_eq by blast


lemma idx_less_than:
assumes "index_sequence 0 l"
        "(inth l (ilen l)) = ilen xs"
        "i \<le> ilen l"
        "n \<le> ilen l"
shows   "\<forall> j. j>n \<and> j \<le> ilen l\<longrightarrow> inth l j > inth l n"
by (meson Suc_leI assms idx_less_equal index_sequence_def less_le_trans)


lemma idx_sub:
 assumes "k\<le>n"
         "n \<le> ilen l"
         "index_sequence 0 l"
 shows   "index_sequence (inth l k) (sub k n l)"
proof -
 have 1: "index_sequence (inth l k) (sub k n l) = 
          ((inth (sub k n l) 0) = (inth l k) \<and> 
           (\<forall>i. i<ilen(sub k n l) \<longrightarrow> (inth (sub k n l) i) < (inth (sub k n l) (Suc i))))"
    using index_sequence_def by auto 
 have 2: "(inth (sub k n l) 0) = (inth l k) "
       using assms ifirst_sub by auto
 have 3: " ilen(sub k n l) = (n -k)"
   by (simp add: assms)
 have 4: "(\<forall>i. i<ilen(sub k n l) \<longrightarrow> (inth (sub k n l) i) < (inth (sub k n l) (Suc i))) "
  proof 
   fix i
   show "i < ilen (sub k n l) \<longrightarrow> inth (sub k n l) i < inth (sub k n l) (Suc i) "
   proof -
    have 41: "i < (n-k) \<longrightarrow> inth (sub k n l) i = inth l (k+i) "
    by (simp add: assms)
    have 42: "i < (n-k) \<longrightarrow> inth (sub k n l) (Suc i) = inth l (k+(Suc i)) "
      by (simp add: assms)
    have 43: "i < (n-k) \<longrightarrow> inth l (k+i) < inth l (k+(Suc i))"
      using assms 
      using index_sequence_def by auto
    show ?thesis using "3" "41" "42" "43" by auto
   qed
  qed
 show ?thesis by (simp add: "1" "2" "4")
qed

lemma idx_split:
 assumes " n\<le> ilen l"
 shows   "index_sequence 0 l = 
          (index_sequence 0 (prefix n l) \<and> index_sequence (inth l n) (suffix n l))"
proof -
 have 1: "index_sequence 0 l \<longrightarrow> 
          index_sequence 0 (prefix n l) \<and> index_sequence (inth l n) (suffix n l) "
   using idx_sub using assms index_sequence_def by auto
 have 2: "index_sequence 0 (prefix n l) =
          ( (inth (prefix n l) 0) = 0 \<and>
            (\<forall>i. i< ilen(prefix n l) \<longrightarrow> (inth (prefix n l) i) < (inth (prefix n l) (Suc i))))"
   using index_sequence_def by blast
 have 3: "( (inth (prefix n l) 0) = 0 \<and>
            (\<forall>i. i< ilen(prefix n l) \<longrightarrow> (inth (prefix n l) i) < (inth (prefix n l) (Suc i)))) =
            ( (inth l 0) = 0 \<and> 
              (\<forall>i. i<n \<longrightarrow> (inth l i) < (inth l (Suc i)))) "
    using assms by auto      
 have 4: "index_sequence (inth l n) (suffix n l) =
          ( (inth (suffix n l) 0) = (inth l n) \<and>
            (\<forall> i. i<ilen(suffix n l) \<longrightarrow> (inth (suffix n l) i) < (inth (suffix n l) (Suc i)))) "
    using index_sequence_def by blast
 have 5: "( (inth (suffix n l) 0) = (inth l n) \<and>
            (\<forall> i. i<ilen(suffix n l) \<longrightarrow> (inth (suffix n l) i) < (inth (suffix n l) (Suc i)))) =
            ( (inth l n) = (inth l n) \<and>
            (\<forall> i. i<ilen l - n \<longrightarrow> (inth l (i+n) ) < (inth l ((Suc i)+n)))) "
            by (simp add: add.commute)
 have 6: "(\<forall> i. i<ilen l - n \<longrightarrow> (inth l (i+n) ) < (inth l ((Suc i)+n))) =
          (\<forall> i. n \<le> (i+n) \<and> (i+n)<ilen l  \<longrightarrow> (inth l (i+n) ) < (inth l ((Suc (i+n))))) "
       by auto
 have 61: "(\<forall> i. n \<le> (i+n) \<and> (i+n)<ilen l  \<longrightarrow> (inth l (i+n) ) < (inth l ((Suc (i+n))))) =
           (\<forall> j. n \<le> j \<and> j<ilen l  \<longrightarrow> (inth l j ) < (inth l ((Suc j))))"
           by (metis diff_add)  
 have 7: " (index_sequence 0 (prefix n l) \<and> index_sequence (inth l n) (suffix n l)) \<longrightarrow>
          (inth l 0) = 0 \<and> (\<forall>i. i<n \<longrightarrow> (inth l i) < (inth l (Suc i))) " 
       using "2" "3" by blast
 have 8: "(index_sequence 0 (prefix n l) \<and> index_sequence (inth l n) (suffix n l)) \<longrightarrow>
           (\<forall> i. n \<le> i \<and> i<ilen l  \<longrightarrow> (inth l (i) ) < (inth l ((Suc i)))) "
           using "4" "5" "6" "61" by blast
 have 9: "(index_sequence 0 (prefix n l) \<and> index_sequence (inth l n) (suffix n l)) \<longrightarrow>
          (inth l 0) = 0 \<and> (\<forall>i. i<ilen l \<longrightarrow> (inth l i) < (inth l (Suc i))) "
  using "7" "8" not_le by blast
 have 10: "(index_sequence 0 (prefix n l) \<and> index_sequence (inth l n) (suffix n l)) \<longrightarrow>
           index_sequence 0 l "
    using "9" index_sequence_def by blast
 from 10 1 show ?thesis by blast
qed

lemma idx_suffixa:
 assumes "n\<le> ilen l"
         "index_sequence (inth l n) (suffix n l)"
 shows   "index_sequence 0 ((imap (shiftm (inth l n))  (suffix n l)))"
using assms idx_shiftm by blast

lemma idx_greater:
 assumes "index_sequence k l"
 shows   "(\<forall>i. i\<le> ilen l \<longrightarrow> k \<le> (inth l i))"
by (metis assms eq_iff index_sequence_def idx_greater_first less_imp_le neq0_conv)

lemma idx_suffixb:
 assumes "n\<le> ilen l"
         "index_sequence 0 ((imap (shiftm (inth l n))  (suffix n l)))"
 shows   "index_sequence (inth l n) (suffix n l)"
by (metis assms idx_shiftm_c ilast_ifirst ilast_prefix)

lemma idx_suffix:
 assumes "n\<le> ilen l"
 shows   "index_sequence (inth l n) (suffix n l) =
          index_sequence 0 ((imap (shiftm (inth l n))  (suffix n l)))"
using assms idx_shiftm idx_suffixb by blast
 
lemma idx_ifirst:
 assumes "index_sequence 0 (x1a\<odot> l)"
  shows  " x1a < ifirst(l)"
by (metis assms idx_ICons) 

lemma idx_expand1:
 " (index_sequence x1a (x1a \<odot> l)) = ( x1a < ifirst l \<and> index_sequence (ifirst l) l )"
using inth_zero_ifirst less_Suc_eq_0_disj by (auto simp add: index_sequence_def)


lemma idx_ilen_leq_ilast_ifirst:
 assumes "index_sequence (ifirst l) l"
 shows   "ilen (l) \<le> (ilast l - ifirst l)"
using assms
proof
 (induct l)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a l)
 then show ?case 
   proof -
    have 1: "ilen (x1a \<odot> l) = ilen l +1"
     by simp
    have 2: "index_sequence (ifirst l) l" 
     using ICons.prems idx_expand1 by auto 
    have 3: "ilast (x1a \<odot> l) = ilast l" 
     by simp
    have 4: "ifirst (x1a \<odot> l) = x1a"
     by simp
    have 5: "x1a < ifirst l"
     using ICons.prems idx_expand1 by auto
    have 6: "ilen l \<le> ilast l - ifirst l" 
     using "2" ICons.hyps by blast
    have 7: "ilen l +1 \<le> (ilast l - ifirst l)+1"
       using "6" add_le_cancel_right by blast
    have 8: " (ilast l - ifirst l) +1 \<le> ilast l - x1a "
    using "5" "6" le_less_linear by fastforce
    show ?thesis using "7" "8" by auto
   qed      
 qed

lemma idx_ilen_leq:
assumes "index_sequence (ifirst l) l"
        "ilast(l) \<le> ilen xs"
shows "ilen (l) \<le> ilen (sub (ifirst l) (ilast l) xs) "
proof -
 have 1: "ifirst l \<le> ilast l"
   using assms by (metis eq_iff gr0I index_sequence_def idx_less_last_1 less_imp_le_nat) 
 have 2: "ilen (sub (ifirst l) (ilast l) xs) = (ilast l - ifirst l)"
     using "1" assms ilen_sub by blast
 have 3: "ilen (l) \<le> (ilast l - ifirst l)" 
     using assms idx_ilen_leq_ilast_ifirst by blast
 show ?thesis using "2" "3" by linarith
qed


lemma idx_shiftm_sub_inth:
assumes "index_sequence 0 l"
        " (inth l (ilen l)) = ilen xs"
        "  k \<le> n "
        " n \<le> ilen l"
 shows  "\<forall> j. j \<le> n-k \<longrightarrow> 
           inth (imap (shiftm (inth l k)) (sub k n l)) j = inth l (k+j) - (inth l k)"
by (simp add: Nat.le_diff_conv2 assms inth_imap shiftm_def)


lemma idx_shiftm_suffix_inth:
assumes "index_sequence 0 l"
        "  (inth l (ilen l)) = ilen xs"
        "  n \<le> ilen l"
 shows  "\<forall> j. j \<le> ilen l -n \<longrightarrow> 
           inth (imap (shiftm (inth l n)) (suffix n l)) j = inth l (n+j) - (inth l n)"
using assms by (metis  inth_imap inth_suffix shiftm_def) 
 


subsubsection \<open>upt\<close>

lemma upt_rec[code]:  
 "[i..\<le>j] = (if i<j then i \<odot> [Suc i..\<le>j] else \<langle>j\<rangle>)"
by (induct j) auto

lemma upt_conv_INil [simp]:
assumes "j< i"
shows   "[i..\<le>j] = \<langle>j\<rangle>"
using assms
by (metis  upt.simps(1)  upt.simps(2) Suc_leD less_Suc_eq_0_disj not_le)

lemma upt_same:
 "[i..\<le>i] = \<langle>i\<rangle>"
by (metis  upt.simps(1)  upt.simps(2) less_Suc_eq less_Suc_eq_0_disj not_le)

lemma upt_eq_INil_conv[simp]:
 "([i..\<le>j] = \<langle>j\<rangle>) = (j\<le>i)"
by (simp add:  upt_rec)

lemma upt_eq_ICons_conv:
 "([i..\<le>j] = x\<odot>xs) = (i<j \<and> i=x \<and> [Suc i..\<le>j] = xs)"
using  upt_rec by (induct j arbitrary: x xs) auto

lemma upt_suc_iapp:
assumes "i \<le> j"
shows  " [i..\<le>(Suc j)] = [i..\<le>j]\<ominus>\<langle>(Suc j)\<rangle>"
using assms by simp

lemma upt_conv_ICons:
assumes "i<j"
shows   "  [i..\<le>j] = i \<odot> [(Suc i)..\<le>j]"
using assms by (simp add: upt_rec)

lemma upt_conv_ICons_ICons:
 "( m \<odot> n \<odot> ns = [m..\<le>q]) = (n \<odot> ns = [(Suc m)..\<le>q])"
proof (cases "m \<le> q")
case True
then show ?thesis by (simp add:  upt_rec)
next
case False
then show ?thesis by auto
qed

lemma upt_add_eq_iapp:
assumes "i \<le> j"
        " k>0"
shows   "[i..\<le>j+k] = [i..\<le>j]\<ominus>[Suc j..\<le>j+k]"
using assms 
proof
 (induct k)
 case 0
 then show ?case by blast
 next
 case (Suc k)
 then show ?case using Suc_less_eq le_simps(2) by auto
qed

lemma upt_ilen:
 "ilen [i..\<le>j] = j - i"
by (induct j) (auto simp add: Suc_diff_le)

lemma upt_inth_help:
 " inth [i..\<le>i + k] k = i + k"
proof 
   (induct k arbitrary: i)
   case 0
   then show ?case by (simp add: upt_same)
   next
   case (Suc k)
   then show ?case 
   by (metis  upt_rec add_Suc_shift inth_Suc less_add_same_cancel1 zero_less_Suc)
qed

lemma upt_inth:
assumes "i + k \<le> j"
shows   "(inth [i..\<le>j] k) = i + k"
using assms 
proof 
  (induct j arbitrary: k i)
  case 0
  then show ?case by simp
  next
  case (Suc j)
  then show ?case 
   by (metis  upt.upt_Suc Nat.le_diff_conv2 add.commute add_leD1 
          iapp_inth le_SucE upt_ilen upt_inth_help)
qed

lemma upt_ifirst:
assumes "i \<le> j"
shows   "ifirst [i..\<le>j] = i"
using assms by (simp add:  upt_rec)

lemma upt_ilast:
 " ilast [i..\<le>j] = j"
by (metis add_diff_inverse_nat INil_ilen order_refl upt_conv_INil upt_ilen upt_inth)

lemma prefix_upt:
assumes "i+m \<le> n"
shows   "prefix m [i..\<le>n] = [i..\<le>i+m]"
using assms 
proof
 (induct m arbitrary: i)
 case 0
 then show ?case by (simp add: upt_inth upt_same)
 next
 case (Suc m)
 then show ?case using  upt_rec by auto
qed

lemma suffix_upt:
 " suffix m [i..\<le>j] = [i+m..\<le>j]"
proof
 (induct m arbitrary: i j)
 case 0
 then show ?case by simp
 next
 case (Suc j)
 then show ?case using  upt_rec 
 by (metis add_Suc_shift suffix_suc not_less_eq not_less_iff_gr_or_eq suffix.simps(1))
qed

lemma imap_suc_upt:
 " imap Suc [m..\<le>n] = [Suc m..\<le>Suc n]"
proof
 (induct n arbitrary: m)
 case 0
 then show ?case by simp
 next
 case (Suc n)
 then show ?case by simp
qed

lemma imap_add_upt:
 "imap (\<lambda>i. i + n) [0..\<le>m] = [n..\<le>m+n]"
proof
 (induct m)
 case 0
 then show ?case by (simp add: upt_same)
 next
 case (Suc m)
 then show ?case by simp
qed



subsubsection \<open>Set\<close>

lemma iset_iapp [simp]:
 "iset (xs \<ominus> ys) = (iset xs \<union> iset ys)"
by (induct xs) auto

lemma interval_finite_iset [iff]: 
  "finite (iset (xs:: 'a interval) )"
by (induct xs) auto

lemma interval_hd_in_iset [simp]: 
  "  x \<in> iset (x\<odot>xs)"
by simp

lemma iset_subset_ICons: 
   "iset xs \<subseteq> iset (x \<odot> xs)"
by auto

lemma iset_IConsD: 
 assumes "y \<in> iset (x \<odot> xs)"
 shows   " y=x \<or> y \<in> iset xs"
using assms by auto

lemma exists_ICons:
 "(\<exists> ys \<in> iset ( (x\<odot>xs)). P ys) \<longleftrightarrow> 
  (P x \<and> (\<exists> ys \<in> iset xs. P ys)) \<or> (\<not> P x \<and> (\<exists> ys \<in> iset xs. P ys)) \<or>
  (P x \<and> (\<forall> ys \<in> iset xs. \<not> P ys))"
by auto

lemma interval_iset_nonempty:
 " iset xs \<noteq> {} "
by (induct xs) auto

lemma iset_irev [simp]: 
  "iset (irev xs) = iset xs"
by (induct xs) auto



lemma split_interval:
assumes "x \<in> iset xs " 
shows   " \<exists> ys zs. xs = ys \<ominus> (x \<odot> zs) \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> xs= ys \<ominus> \<langle>x\<rangle>"
using assms 
proof (induct xs)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
by (metis iapp_INil interval.inject(2) interval.set_cases iapp_assoc)
qed

lemma interval_in_iset_conv_decomp:
 "x \<in> iset xs =
  (\<exists> ys zs. xs = ys \<ominus> (x \<odot> zs) \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> xs= ys \<ominus> \<langle>x\<rangle>)"
by (auto elim: split_interval)

lemma split_interval_first:
assumes "x \<in> iset xs"
shows   "  \<exists> ys zs. xs = ys \<ominus> (x \<odot> zs) \<and> x \<notin> iset ys \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           xs = ys \<ominus> \<langle>x\<rangle> \<and> x\<notin> iset ys"
using assms 
proof (induct xs)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
  proof (cases "x = x1a")
  case True
  then show ?thesis by blast
  next
  case False
  then show ?thesis 
    using ICons.prems ICons.hyps
    proof auto
     show "\<And>ys zs.
           x \<noteq> x1a \<Longrightarrow>
           x \<notin> iset ys \<Longrightarrow>
           xs = ys \<ominus> x \<odot> zs \<Longrightarrow>
            \<exists>ysa. (\<exists>zsa. x1a \<odot> ys \<ominus> x \<odot> zs = ysa \<ominus> x \<odot> zsa) \<and> x \<notin> iset ysa \<or>
               x1a \<odot> ys \<ominus> x \<odot> zs = ysa \<ominus> \<langle>x\<rangle> \<and> x \<notin> iset ysa" 
        by (meson ICons_eq_iappI iset_IConsD)
     show "x \<noteq> x1a \<Longrightarrow> 
           xs = \<langle>x\<rangle> \<Longrightarrow> 
           \<exists>ys. (\<exists>zs. \<langle>x1a, x\<rangle> = ys \<ominus> x \<odot> zs) \<and> x \<notin> iset ys \<or> \<langle>x1a, x\<rangle> = ys \<ominus> \<langle>x\<rangle> \<and> x \<notin> iset ys"
        by (metis False Set.set_insert iapp_INil interval.simps(15) singleton_insert_inj_eq') 
     show "\<And>zs. x \<noteq> x1a \<Longrightarrow> 
             xs = x \<odot> zs \<Longrightarrow> 
             \<exists>ys. (\<exists>zsa. x1a \<odot> x \<odot> zs = ys \<ominus> x \<odot> zsa) \<and> x \<notin> iset ys \<or> 
                  x1a \<odot> x \<odot> zs = ys \<ominus> \<langle>x\<rangle> \<and> x \<notin> iset ys"
        by (metis iapp_INil interval.simps(15) singleton_iff) 
     show "\<And>ys. x \<noteq> x1a \<Longrightarrow>
             x \<notin> iset ys \<Longrightarrow>
             xs = ys \<ominus> \<langle>x\<rangle> \<Longrightarrow> 
             \<exists>ysa. (\<exists>zs. x1a \<odot> ys \<ominus> \<langle>x\<rangle> = ysa \<ominus> x \<odot> zs) \<and> x \<notin> iset ysa \<or> 
               x1a \<odot> ys \<ominus> \<langle>x\<rangle> = ysa \<ominus> \<langle>x\<rangle> \<and> x \<notin> iset ysa " 
       by (meson ICons_eq_iappI iset_IConsD)
   qed
  qed
qed

lemma in_iset_conv_decomp_first:
 "x \<in> iset xs = 
  (\<exists> ys zs. xs = ys \<ominus> (x \<odot> zs) \<and> x \<notin> iset ys \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           xs = ys \<ominus> \<langle>x\<rangle> \<and> x\<notin> iset ys)"
by (auto dest!: split_interval_first)

lemma split_interval_last:
 assumes "x \<in> iset xs"
  shows " \<exists> ys zs. xs = ys \<ominus> (x \<odot> zs) \<and> x \<notin> iset zs \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<and> x \<notin> iset zs\<or> 
           xs = ys \<ominus> \<langle>x\<rangle> "
using assms
proof (induct xs rule: irev_induct)
case (INil y)
then show ?case by simp
next
case (snoc x1a xs)
  then show ?case proof (cases "x = x1a")
  case True
  then show ?thesis by blast
  next
  case False
  then show ?thesis using snoc by fastforce
qed 
qed

lemma interval_in_iset_conv_decomp_last:
 "x \<in> iset xs = 
  (\<exists> ys zs. xs = ys \<ominus> (x \<odot> zs) \<and> x \<notin> iset zs \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<and> x \<notin> iset zs\<or> 
           xs = ys \<ominus> \<langle>x\<rangle>)" 
by (auto dest!: split_interval_last)

lemma interval_list_prop:
assumes "\<exists> x \<in> iset xs. P x"
shows   "(\<exists> ys x zs. (xs = ys \<ominus> (x \<odot> zs) \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           xs = ys \<ominus> \<langle>x\<rangle>) \<and> P x)"
using assms
proof (induct xs)
case (INil x)
then show ?case by auto
next
  case (ICons x1a xs)
  then show ?case 
  by (meson split_interval)
qed

lemma split_interval_propE:
 assumes "\<exists> x\<in> iset xs. P x"
 obtains ys x zs where "xs = ys \<ominus> (x \<odot> zs) \<or> xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           xs = ys \<ominus> \<langle>x\<rangle>" and "P x" 
using interval_list_prop [OF assms] by blast

lemma split_interval_first_prop:
assumes "\<exists> x \<in> iset xs. P x" 
 shows  "(\<exists> ys x zs. ((xs = ys \<ominus> (x \<odot> zs) \<and> (\<forall>y \<in> iset ys. \<not> P y)) \<or>
           xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           (xs = ys \<ominus> \<langle>x\<rangle> \<and> (\<forall>y \<in> iset ys. \<not> P y))) \<and> P x 
  )"
using assms
proof (induct xs)
case (INil x)
then show ?case by auto
next
  case (ICons x1a xs)
  then show ?case proof (cases "P x1a")
  case True
  then show ?thesis by blast
  next
  case False
  then show ?thesis 
    using ICons 
    proof auto
     show "\<And>ys xa zs.
       \<not> P x1a \<Longrightarrow>
       P xa \<Longrightarrow>
       \<forall>y\<in>iset ys. \<not> P y \<Longrightarrow>
       xs = ys \<ominus> xa \<odot> zs \<Longrightarrow>
       \<exists>ysa x.
          (\<exists>zsa. x1a \<odot> ys \<ominus> xa \<odot> zs = ysa \<ominus> x \<odot> zsa \<and> (\<forall>y\<in>iset ysa. \<not> P y) \<or>
          x1a = x \<and> ys \<ominus> xa \<odot> zs = zsa \<or> x1a \<odot> ys \<ominus> xa \<odot> zs = ysa \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>iset ysa. \<not> P y)) \<and>
          P x" 
        by (metis insertE iapp_ICons interval.simps(16))
     show "\<And>x ys xa zs.
       \<not> P x1a \<Longrightarrow>
       P x \<Longrightarrow>
       P xa \<Longrightarrow>
       \<forall>y\<in>iset ys. \<not> P y \<Longrightarrow>
       xs = ys \<ominus> xa \<odot> zs \<Longrightarrow>
       x \<in> iset zs \<Longrightarrow>
       \<exists>ysa x.
          (\<exists>zsa. x1a \<odot> ys \<ominus> xa \<odot> zs = ysa \<ominus> x \<odot> zsa \<and> (\<forall>y\<in>iset ysa. \<not> P y) \<or>
          x1a = x \<and> ys \<ominus> xa \<odot> zs = zsa \<or> x1a \<odot> ys \<ominus> xa \<odot> zs = ysa \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>iset ysa. \<not> P y)) \<and>
          P x"  
        by (metis ICons_eq_iappI iset_IConsD)
     show "\<And>xa. \<not> P x1a \<Longrightarrow>
          P xa \<Longrightarrow>
          xs = \<langle>xa\<rangle> \<Longrightarrow>
          \<exists>ys x.
             (\<exists>zs. \<langle>x1a, xa\<rangle> = ys \<ominus> x \<odot> zs \<and> (\<forall>y\<in>iset ys. \<not> P y) \<or> 
               x1a = x \<and> \<langle>xa\<rangle> = zs \<or> \<langle>x1a, xa\<rangle> = ys \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>iset ys. \<not> P y)) \<and> P x" 
        by (metis iapp_INil interval.simps(15) singleton_iff)
     show "\<And>xa zs.
       \<not> P x1a \<Longrightarrow>
       P xa \<Longrightarrow>
       xs = xa \<odot> zs \<Longrightarrow>
       \<exists>ys x.
          (\<exists>zsa. x1a \<odot> xa \<odot> zs = ys \<ominus> x \<odot> zsa \<and> (\<forall>y\<in>iset ys. \<not> P y) \<or>
                 x1a = x \<and> xa \<odot> zs = zsa \<or> x1a \<odot> xa \<odot> zs = ys \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>iset ys. \<not> P y)) \<and>
          P x"
        by (metis iapp_INil interval.simps(15) singleton_iff) 
     show " \<And>x xa zs.
       \<not> P x1a \<Longrightarrow>
       P x \<Longrightarrow>
       P xa \<Longrightarrow>
       xs = xa \<odot> zs \<Longrightarrow>
       x \<in> iset zs \<Longrightarrow>
       \<exists>ys x.
          (\<exists>zsa. x1a \<odot> xa \<odot> zs = ys \<ominus> x \<odot> zsa \<and> (\<forall>y\<in>iset ys. \<not> P y) \<or>
                 x1a = x \<and> xa \<odot> zs = zsa \<or> x1a \<odot> xa \<odot> zs = ys \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>iset ys. \<not> P y)) \<and>
          P x" 
       by (metis iapp_INil interval.simps(15) singleton_iff)
     show "\<And>ys xa.
       \<not> P x1a \<Longrightarrow>
       P xa \<Longrightarrow>
       \<forall>y\<in>iset ys. \<not> P y \<Longrightarrow>
       xs = ys \<ominus> \<langle>xa\<rangle> \<Longrightarrow>
       \<exists>ysa x.
          (\<exists>zs. x1a \<odot> ys \<ominus> \<langle>xa\<rangle> = ysa \<ominus> x \<odot> zs \<and> (\<forall>y\<in>iset ysa. \<not> P y) \<or>
                x1a = x \<and> ys \<ominus> \<langle>xa\<rangle> = zs \<or> x1a \<odot> ys \<ominus> \<langle>xa\<rangle> = ysa \<ominus> \<langle>x\<rangle> \<and> (\<forall>y\<in>iset ysa. \<not> P y)) \<and>
          P x " 
    by (metis insertE iapp.simps(2) interval.set(2))
   qed   
  qed 
qed

lemma split_interval_first_propE:
 assumes "\<exists> x \<in> iset xs. P x"
 obtains ys x zs where "((xs = ys \<ominus> (x \<odot> zs) \<and> (\<forall>y \<in> iset ys. \<not> P y)) \<or>
           xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           (xs = ys \<ominus> \<langle>x\<rangle> \<and> (\<forall>y \<in> iset ys. \<not> P y)))" and "P x"
 using split_interval_first_prop [OF assms] by blast

lemma split_first_prop_iff:
 "(\<exists> x \<in> iset xs. P x) \<longleftrightarrow>
  (\<exists> ys x zs. ((xs = ys \<ominus> (x \<odot> zs) \<and> (\<forall>y \<in> iset ys. \<not> P y)) \<or>
           xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<or> 
           (xs = ys \<ominus> \<langle>x\<rangle> \<and> (\<forall>y \<in> iset ys. \<not> P y))) \<and> P x 
  )"
by (rule, erule split_interval_first_prop) auto

lemma split_interval_last_prop:
assumes "\<exists> x \<in> iset xs. P x "
shows   "(\<exists> ys x zs. ((xs = ys \<ominus> (x \<odot> zs) \<and> (\<forall>y \<in> iset zs. \<not> P y)) \<or>
           xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<and> (\<forall>y \<in> iset zs. \<not> P y) \<or> 
           (xs = ys \<ominus> \<langle>x\<rangle>)) \<and> P x 
  )"
using assms
proof (induct xs rule: irev_induct)
case (INil y)
then show ?case by auto
next
case (snoc x1a xs)
  then show ?case proof (cases "P x1a")
  case True
  then show ?thesis by blast
  next
  case False
  then show ?thesis 
   using snoc by fastforce
  qed
qed

lemma split_interval_last_propE:
 assumes "\<exists> x \<in> iset xs. P x" 
 obtains ys x zs where "((xs = ys \<ominus> (x \<odot> zs) \<and> (\<forall>y \<in> iset zs. \<not> P y)) \<or>
           xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<and> (\<forall>y \<in> iset zs. \<not> P y) \<or> 
           (xs = ys \<ominus> \<langle>x\<rangle>))" and "P x"
 using split_interval_last_prop [OF assms] by blast

lemma split_interval_last_prop_iff:
 "(\<exists> x \<in> iset xs. P x) \<longleftrightarrow>
  (\<exists> ys x zs. ((xs = ys \<ominus> (x \<odot> zs) \<and> (\<forall>y \<in> iset zs. \<not> P y)) \<or>
           xs = \<langle>x\<rangle> \<or> xs = x \<odot> zs \<and> (\<forall>y \<in> iset zs. \<not> P y) \<or> 
           (xs = ys \<ominus> \<langle>x\<rangle>)) \<and> P x 
  )"
by (rule, erule split_interval_last_prop, auto)

lemma inth_and_iset:
 "x \<in> iset xs = (\<exists> i\<le>ilen xs. (inth xs i) =x)"
proof (induct xs)
case (INil x)
then show ?case by auto
next
case (ICons x1a xs)
then show ?case 
by (metis Suc_le_mono insert_iff interval.simps(16) ilen_ICons ilen_gr_zero 
    inth_Suc inth_ICons inth_zero ilen.simps(2) le_diff_conv neq0_conv plus_1_eq_Suc)
qed

lemma card_ilen:
 "card (iset xs) \<le> ilen xs+1"
proof (induct xs)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case by (simp add: card_insert_le_m1)
qed

lemma iset_inth:
 "iset xs = { (inth xs k) | k. k\<le> ilen xs}"
proof (induct xs)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
  proof auto
   show "iset xs = {inth xs k |k. k \<le> ilen xs} \<Longrightarrow>
    \<exists>k. x1a = (case k of 0 \<Rightarrow> x1a | Suc x \<Rightarrow> inth xs x) \<and> k \<le> Suc (ilen xs)"
   by force
   show "\<And>k. iset xs = {inth xs k |k. k \<le> ilen xs} \<Longrightarrow>
         k \<le> ilen xs \<Longrightarrow>
         \<exists>ka. inth xs k = (case ka of 0 \<Rightarrow> x1a | Suc x \<Rightarrow> inth xs x) \<and> ka \<le> Suc (ilen xs)"
   by force
   show "\<And>k. iset xs = {inth xs k |k. k \<le> ilen xs} \<Longrightarrow>
         (case k of 0 \<Rightarrow> x1a | Suc x \<Rightarrow> inth xs x) \<noteq> x1a \<Longrightarrow>
         k \<le> Suc (ilen xs) \<Longrightarrow>
         \<exists>ka. (case k of 0 \<Rightarrow> x1a | Suc x \<Rightarrow> inth xs x) = inth xs ka \<and> ka \<le> ilen xs "
   by (metis (mono_tags, lifting) Nitpick.case_nat_unfold Suc_eq_plus1 le_diff_conv)
  qed
qed

lemma prefix_iset:
assumes "k\<le> ilen xs"
 shows  "  iset (prefix k xs) = {(inth xs i) |i. i\<le>k}"
proof -
 have 1: " iset (prefix k xs) = { (inth (prefix k xs) i) | i. i\<le> ilen (prefix k xs)}"
   by (simp add: iset_inth)
 have 2: " k\<le> ilen xs"
   using assms by auto
 have 3: "{ (inth (prefix k xs) i) | i. i\<le> ilen (prefix k xs)} =
          { (inth xs i) | i. i\<le> k}" 
   using "2" by (metis inth_prefix prefix_ilen_good)
 show ?thesis using "1" "3" by auto
qed

 lemma suffix_iset:
assumes "k\<le> ilen xs"
 shows  "  iset (suffix k xs) = {(inth xs (k+i)) |i. i\<le>ilen xs - k}"
proof -
 have 1: "iset (suffix k xs) = { (inth (suffix k xs) i) | i. i\<le> ilen (suffix k xs)}" 
   by (simp add: iset_inth)
 have 2: "k \<le> ilen xs"
   using assms by auto 
 have 3: "{ (inth (suffix k xs) i) | i. i\<le> ilen (suffix k xs)} =
          { (inth xs (k+i)) | i. i\<le> ilen xs - k}"
          by force
 show ?thesis  using "1" "3" by auto
qed

lemma suffix_iset_a:
assumes "k\<le> ilen xs"
 shows  "  iset (suffix k xs) = {(inth xs i) |i. k\<le> i \<and> i\<le>ilen xs }"
proof -
 have 1: "iset (suffix k xs) = { (inth xs (k+i)) | i. i\<le> ilen xs - k}" 
   using assms suffix_iset by blast
 have 2: "\<forall> x \<in> { (inth xs (k+i)) | i. i\<le> ilen xs - k} . 
            x \<in> {(inth xs i) |i. k\<le> i \<and> i\<le>ilen xs }"
    using assms by auto
 have 3: "\<forall> x \<in> {(inth xs i) |i. k\<le> i \<and> i\<le>ilen xs } . 
            x \<in> { (inth xs (k+i)) | i. i\<le> ilen xs - k} "
     using nat_le_iff_add by force
 have 4: "{ (inth xs (k+i)) | i. i\<le> ilen xs - k} =
          {(inth xs i) |i. k\<le> i \<and> i\<le>ilen xs }" 
  using "2" "3" by blast  
 show ?thesis by (simp add: "1" "4")
qed

lemma sub_interval_iset:
 assumes "k \<le> n"
         "n \<le> ilen xs"
 shows   " iset (sub k n xs) = { (inth xs (k+i)) | i. i \<le> n-k }"
proof -
 have 1: " iset (sub k n xs) = { (inth (sub k n xs) i) | i. i \<le> ilen (sub k n xs) }"
    by (simp add: iset_inth)
 have 2: " k\<le> n "
   using assms by auto
 have 3: " n \<le> ilen xs"
   using assms by auto
 have 4: " { (inth (sub k n xs) i) | i. i \<le> ilen (sub k n xs) } =
           { (inth xs (k+i) ) | i. i \<le> n-k } "
   using "2" "3" by force
 show ?thesis by (simp add: "1" "4")
qed


lemma sub_iset_a:
 assumes "k \<le> n"
         "n \<le> ilen xs"
 shows   " iset (sub k n xs) = { (inth xs i) | i. k \<le> i \<and> i \<le> n }"
proof -
 have 1: "iset (sub k n xs) = { (inth xs (k+i)) | i. i \<le> n-k }" 
   using assms sub_interval_iset by blast
 have 2: "\<forall> x \<in> { (inth xs (k+i)) | i. i \<le> n-k } .
            x \<in> { (inth xs i) | i. k \<le> i \<and> i \<le> n }"
   using assms by auto 
 have 3: "\<forall> x \<in> { (inth xs i) | i. k \<le> i \<and> i \<le> n }.
            x \<in> { (inth xs (k+i)) | i. i \<le> n-k }" 
    using assms using le_Suc_ex by auto fastforce
 have 4: "{ (inth xs (k+i)) | i. i \<le> n-k } = { (inth xs i) | i. k \<le> i \<and> i \<le> n }" 
     using "2" "3" by blast
  show ?thesis by (simp add: "1" "4")
qed

lemma inth_iset:
 assumes "k\<le> ilen xs"
 shows   " (inth xs k) \<in> iset xs"
using assms 
by (meson inth_and_iset)



end
