(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>The First Occurrence Operator in finite ITL\<close>

theory First
 imports
   Theorems TimeReversal
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="z3 cvc4 e spass vampire " ]
*)

(*declare [[smt_timeout=120]]*)

(*declare [[show_types]]*)


text \<open>
Runtime verification (RV) has gained significant interest in recent years. 
The behaviour of a program can be verified in real time by analysing its evolving trace. 
This approach has two significant benefits over static verification techniques such as model 
checking. Firstly, it is only necessary to verify actual execution paths rather than all 
possible paths. Secondly, it is possible to react at runtime should the program diverge from 
its specified behaviour. RV does not replace traditional verification techniques but it 
does provide an extra layer of security.

Linear Temporal Logic (LTL) is a popular formalism for writing specifications from which RV 
monitors can be derived automatically. By contrast, Interval Temporal Logic (ITL) has not 
been as widely represented in this field despite being more expressive and 
compositional. The principal issue is efficiency. ITL uses non-deterministic operators to 
IConstruct sequential and iterative specifications (chop and chop-star, respectively) and 
these introduce combinatorial complexity. Approaches to mitigate this include using a 
deterministic subset of ITL or adapting the semantics to include a deterministic chop operator.
 This work proposes an alternative approach, wholly within existing ITL, and based upon a new,
 derived operator called “first occurrence”.

A theory of first occurrence is developed and used to derive an algebra of RV monitors.

\<close>

subsection \<open>Definitions\<close>

subsubsection \<open>Definitions Strict Initial and Final\<close>


definition bs_d :: "('a::world) formula \<Rightarrow> 'a formula"
where 
  "bs_d f \<equiv> LIFT(empty \<or> ((bi f) ; skip))" 

definition bt_d :: "('a::world) formula \<Rightarrow> 'a formula" 
where 
  "bt_d f \<equiv> LIFT(empty \<or> (skip;(\<box> f)))"

syntax 
 "_bs_d" :: "lift \<Rightarrow> lift" ("(bs _)" [88] 87)
 "_bt_d" :: "lift \<Rightarrow> lift" ("(bt _)" [88] 87)

syntax (ASCII)
 "_bs_d" :: "lift \<Rightarrow> lift" ("(bs _)" [88] 87)
 "_bt_d" :: "lift \<Rightarrow> lift" ("(bt _)" [88] 87)

translations
 "_bs_d" \<rightleftharpoons> "CONST bs_d"
 "_bt_d" \<rightleftharpoons> "CONST bt_d"


definition ds_d :: "('a::world) formula \<Rightarrow> 'a formula"
where 
  "ds_d f \<equiv> LIFT (\<not> (bs (\<not> f)))"  

definition dt_d :: "('a::world) formula \<Rightarrow> 'a formula"
where
  "dt_d f \<equiv> LIFT (\<not> (bt (\<not> f)))"

syntax
 "_ds_d" :: "lift \<Rightarrow> lift" ("(ds _)" [88] 87)
 "_dt_d" :: "lift \<Rightarrow> lift" ("(dt _)" [88] 87)

syntax (ASCII)
 "_ds_d" :: "lift \<Rightarrow> lift" ("(ds _)" [88] 87)
 "_dt_d" :: "lift \<Rightarrow> lift" ("(dt _)" [88] 87)

translations
 "_ds_d" \<rightleftharpoons> "CONST ds_d"
 "_dt_d" \<rightleftharpoons> "CONST dt_d"

subsubsection \<open>Definition First and Last Operators\<close>

definition first_d :: "('a::world) formula \<Rightarrow> 'a formula"
where   
  "first_d f \<equiv>  LIFT (f \<and>  (bs (\<not> f)))"

definition last_d :: "('a::world) formula \<Rightarrow> 'a formula"
where   
  "last_d f \<equiv>  LIFT (f \<and>  (bt (\<not> f)))"

syntax
 "_first_d" :: "lift \<Rightarrow> lift" ("(\<rhd> _)" [88] 87)
 "_last_d" :: "lift \<Rightarrow> lift" ("(\<lhd> _)" [88] 87)

syntax (ASCII)
 "_first_d" :: "lift \<Rightarrow> lift" ("(first _)" [88] 87)
 "_last_d" :: "lift \<Rightarrow> lift" ("(last _)" [88] 87)

translations
 "_first_d" \<rightleftharpoons> "CONST first_d"
 "_last_d"  \<rightleftharpoons> "CONST last_d"



subsection \<open>First and Time Reversal\<close>

lemma BsEqvRule:
 assumes "\<turnstile> f = g"
 shows   "\<turnstile> bs f = bs g"
proof -
 have 1: "\<turnstile> f = g"  using assms by auto
 hence 2: "\<turnstile> bi(f) = bi(g)" by (simp add: BiEqvBi)
 hence 3: "\<turnstile> bi(f);skip  = bi(g);skip" by (simp add: LeftChopEqvChop)
 hence 4: "\<turnstile> (empty \<or> bi(f);skip)  = (empty \<or> bi(g);skip)" by auto
 hence 5: "\<turnstile> bs(f) = bs(g)" by (simp add: bs_d_def)
 from 1 2 3 4 5 show ?thesis by auto
qed

lemma BtEqvRule:
 assumes "\<turnstile> f = g"
 shows   "\<turnstile> bt f = bt g"
proof -
 have 1: "\<turnstile> f = g"  using assms by auto
 hence 2: "\<turnstile> \<box>(f) = \<box>(g)" by (simp add: BoxEqvBox)
 hence 3: "\<turnstile> skip;\<box>(f)  = skip;\<box>(g)" using RightChopEqvChop by blast
 hence 4: "\<turnstile> (empty \<or> skip;\<box>(f))  = (empty \<or> skip;\<box>(g))" by auto
 hence 5: "\<turnstile> bt(f) = bt(g)" by (simp add: bt_d_def)
 from 1 2 3 4 5 show ?thesis by auto
qed

lemma FstEqvRule:
 assumes "\<turnstile> f = g"
 shows   "\<turnstile> \<rhd>f = \<rhd>g "
proof -
 have 1: "\<turnstile> f = g" using assms by auto
 hence 2: "\<turnstile> (\<not> f) = (\<not> g)" by auto
 hence 3: "\<turnstile> bs(\<not> f) = bs(\<not> g)" by (simp add: BsEqvRule)
 hence 4: "\<turnstile> (f \<and> bs(\<not> f)) = (g \<and> bs(\<not> g)) " using "1" by fastforce
 from 4 show ?thesis by (simp add:first_d_def)
qed

lemma LstEqvRule:
 assumes "\<turnstile> f = g"
 shows   "\<turnstile> \<lhd>f = \<lhd>g "
proof -
 have 1: "\<turnstile> f = g" using assms by auto
 hence 2: "\<turnstile> (\<not> f) = (\<not> g)" by auto
 hence 3: "\<turnstile> bt(\<not> f) = bt(\<not> g)" by (simp add: BtEqvRule)
 hence 4: "\<turnstile> (f \<and> bt(\<not> f)) = (g \<and> bt(\<not> g)) " using "1" by fastforce
 from 4 show ?thesis by (simp add:last_d_def)
qed

lemma RBsEqvBt:
 "\<turnstile> (bs f)\<^sup>r = (bt (f\<^sup>r))"
proof -
 have 1: "\<turnstile> (bs f)\<^sup>r = (empty \<or> ((bi f) ; skip))\<^sup>r" 
      by (simp add: bs_d_def)
 have 2: "\<turnstile> (empty \<or> ((bi f) ; skip))\<^sup>r = (empty\<^sup>r \<or> ((bi f) ; skip)\<^sup>r)" 
      using ROr by blast
 have 3: "\<turnstile> (empty\<^sup>r \<or> ((bi f) ; skip)\<^sup>r) = (empty \<or> (skip\<^sup>r;(bi f)\<^sup>r))" 
      using REmptyEqvEmpty RevChop by fastforce
 have 4: "\<turnstile> (empty \<or> (skip\<^sup>r;(bi f)\<^sup>r)) = (empty \<or> (skip;\<box> (f\<^sup>r)))" 
      by (metis "3" RBiEqvBox RevSkip int_eq)
 have 5: "\<turnstile> (empty \<or> (skip;\<box> (f\<^sup>r))) = (bt (f\<^sup>r))" 
      by (simp add: bt_d_def)
 from 1 2 3 4 5 show ?thesis by fastforce
qed

lemma RRBsEqvBt:
 "\<turnstile> (bs (f\<^sup>r))\<^sup>r = (bt (f))"
proof -
 have 1: "\<turnstile> (bs (f\<^sup>r))\<^sup>r = bt ((f\<^sup>r)\<^sup>r)" using RBsEqvBt by blast
 have 2: "\<turnstile> bt ((f\<^sup>r)\<^sup>r) = bt f" using EqvReverseReverse using BtEqvRule by blast
 from 1 2 show ?thesis by fastforce
qed

lemma RBtEqvBs:
 "\<turnstile> (bt f)\<^sup>r = (bs (f\<^sup>r))"
proof -
 have 1: "\<turnstile> (bt f)\<^sup>r = (empty \<or> (skip;\<box> f))\<^sup>r " 
      by (simp add: bt_d_def)
 have 2: "\<turnstile> (empty \<or> (skip;\<box> f))\<^sup>r = (empty\<^sup>r \<or> (skip;\<box> f)\<^sup>r)"
      using ROr by blast
 have 3: "\<turnstile> (empty\<^sup>r \<or> (skip;\<box> f)\<^sup>r) = (empty \<or> (\<box> f)\<^sup>r;skip\<^sup>r)" 
      using REmptyEqvEmpty RevChop by fastforce
 have 4: "\<turnstile> (empty \<or> (\<box> f)\<^sup>r;skip\<^sup>r) = (empty \<or> (bi (f\<^sup>r));skip)" 
      by (metis "3" RBoxEqvBi RevSkip int_eq)
 show ?thesis
 by (metis "3" "4" all_rev_eq(3) bs_d_def bt_d_def inteq_reflection) 
qed

lemma RRBtEqvBs:
 "\<turnstile> (bt (f\<^sup>r))\<^sup>r = (bs (f))"
proof -
 have 1: "\<turnstile> (bt (f\<^sup>r))\<^sup>r = bs ((f\<^sup>r)\<^sup>r)" using RBtEqvBs by blast
 have 2: "\<turnstile> bs ((f\<^sup>r)\<^sup>r) = bs f" using EqvReverseReverse using BsEqvRule by blast
 from 1 2 show ?thesis by fastforce
qed

lemma RFirstEqvLast:
 "\<turnstile> (\<rhd> f)\<^sup>r = (\<lhd> (f\<^sup>r))"
proof -
 have 1: "\<turnstile> (\<rhd> f)\<^sup>r = (f \<and> bs(\<not> f))\<^sup>r" by (simp add: first_d_def)
 have 2: "\<turnstile> (f \<and> bs(\<not> f))\<^sup>r = (f\<^sup>r \<and> (bs (\<not> f))\<^sup>r)" using RAnd by blast
 have 3: "\<turnstile> (f\<^sup>r \<and> (bs (\<not> f))\<^sup>r) = (f\<^sup>r \<and> bt ((\<not> f)\<^sup>r))" using RBsEqvBt by fastforce
 have 4: "\<turnstile> (f\<^sup>r \<and> bt ((\<not> f)\<^sup>r)) = (f\<^sup>r \<and> bt (\<not>(f\<^sup>r)))" using RNot int_eq by fastforce  
 have 5: "\<turnstile> (f\<^sup>r \<and> bt (\<not>(f\<^sup>r))) = (\<lhd> (f\<^sup>r))" by (simp add: last_d_def)
 from 1 2 3 4 5 show ?thesis by fastforce
qed

lemma RRFirstEqvLast:
 "\<turnstile> (\<rhd> (f\<^sup>r))\<^sup>r = (\<lhd> (f))"
proof -
 have 1: "\<turnstile> (\<rhd> (f\<^sup>r))\<^sup>r = \<lhd> ((f\<^sup>r)\<^sup>r)" using RFirstEqvLast by blast
 have 2: "\<turnstile> \<lhd> ((f\<^sup>r)\<^sup>r) = \<lhd> f" using EqvReverseReverse using LstEqvRule by blast
 from 1 2 show ?thesis by fastforce
qed

lemma RLastEqvFirst:
 "\<turnstile> (\<lhd> f)\<^sup>r = (\<rhd> (f\<^sup>r))"
proof -
 have 1: "\<turnstile> (\<lhd> f)\<^sup>r = (f \<and> bt(\<not> f))\<^sup>r" by (simp add: last_d_def)
 have 2: "\<turnstile> (f \<and> bt(\<not> f))\<^sup>r = (f\<^sup>r \<and> (bt (\<not>f))\<^sup>r)" using RAnd by blast
 have 3: "\<turnstile> (f\<^sup>r \<and> (bt (\<not>f))\<^sup>r) = (f\<^sup>r \<and> bs ( (\<not>f)\<^sup>r))" using RBtEqvBs by fastforce
 have 4: "\<turnstile> (f\<^sup>r \<and> bs ( (\<not>f)\<^sup>r)) = (f\<^sup>r \<and> bs(\<not>(f\<^sup>r)))"  using RNot int_eq by fastforce  
 have 5: "\<turnstile> (f\<^sup>r \<and> bs(\<not>(f\<^sup>r))) = (\<rhd> (f\<^sup>r))" by (simp add: first_d_def)
 from 1 2 3 4 5 show ?thesis by fastforce
qed

lemma RRLastEqvFirst:
 "\<turnstile> (\<lhd> (f\<^sup>r))\<^sup>r = (\<rhd> (f))"
proof -
 have 1: "\<turnstile> (\<lhd> (f\<^sup>r))\<^sup>r = \<rhd> ((f\<^sup>r)\<^sup>r)" using RLastEqvFirst by blast
 have 2: "\<turnstile> \<rhd> ((f\<^sup>r)\<^sup>r) = \<rhd> f" using EqvReverseReverse using FstEqvRule by blast
 from 1 2 show ?thesis by fastforce
qed


subsection \<open>Semantic Theorems\<close>

subsubsection \<open>Semantics First and Last Operators\<close>

lemma FstAndBisem:
 "(ilen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> ( \<sigma> \<Turnstile>  bi (\<not>f);skip)) =
  (ilen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> (\<forall>ia<ilen (\<sigma>). (prefix ia  \<sigma> \<Turnstile> \<not>f)) ) "
proof -
 have " (ilen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> ( \<sigma> \<Turnstile>  bi (\<not>f);skip)) =
        (0 < ilen \<sigma> \<and> (\<sigma> \<Turnstile> f) \<and>
           (\<exists>i. (i \<le> ilen \<sigma> \<longrightarrow> (\<forall>ia\<le>i. \<not> (prefix ia (prefix i \<sigma>) \<Turnstile> f)) \<and>
           ilen \<sigma> - i = Suc 0) \<and> i \<le> ilen \<sigma>)
          )"
     using le_trans by  (auto simp add: chop_defs bi_defs skip_defs, blast)        
 also have  " ... = 
          (0 < ilen \<sigma> \<and> (\<sigma> \<Turnstile> f) \<and>
           (\<exists>i. (i \<le> ilen \<sigma> \<longrightarrow> (\<forall>ia\<le>i. \<not> (prefix ia (prefix i \<sigma>) \<Turnstile> f)) \<and>
           i = ilen \<sigma> - Suc 0) \<and> i \<le> ilen \<sigma>)
          )" 
      by auto
 also have "... =
          (0 < ilen \<sigma> \<and> (\<sigma> \<Turnstile> f) \<and>
           (\<forall>ia\<le>(ilen \<sigma> - Suc 0). \<not> (prefix ia (prefix (ilen \<sigma> - Suc 0) \<sigma>) \<Turnstile> f) )
          )" 
      using diff_le_self by blast
 also have "... =
          (ilen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and>
            (\<forall>ia<ilen ( \<sigma>). \<not> (prefix ia (prefix (ilen \<sigma> - Suc 0) \<sigma>) \<Turnstile> f))
          ) " by (metis Suc_pred less_Suc_eq_le)
 also have "... =
          (ilen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and>
           (\<forall>ia<ilen ( \<sigma>). (prefix ia (prefix (ilen \<sigma> - Suc 0) \<sigma>) \<Turnstile> \<not>f))
          ) " 
      by auto
 also have "... = 
          (ilen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> (\<forall>ia<ilen (\<sigma>). (prefix ia  \<sigma> \<Turnstile> \<not>f)) ) " 
          by (simp add: pref_pref_help) 
 finally show " (ilen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> ( \<sigma> \<Turnstile>  bi (\<not>f);skip)) =
               (ilen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> (\<forall>ia<ilen (\<sigma>). (prefix ia  \<sigma> \<Turnstile> \<not>f)) ) " .
qed
     
lemma Fstsem_0:
 "(\<sigma> \<Turnstile> \<rhd>f) = 
  ( 
   ( \<sigma> \<Turnstile> f \<and> empty) \<or> (ilen \<sigma> >0  \<and> (\<sigma> \<Turnstile> f) \<and> ( \<sigma> \<Turnstile>  bi (\<not>f);skip)) 
  )  "
using empty_defs by  (auto simp add: first_d_def bs_d_def)

lemma Emptysem:
 "(\<sigma> \<Turnstile> f \<and> empty) = ((\<sigma> \<Turnstile> f) \<and> ilen \<sigma> = 0) "
using empty_defs by auto

lemma Fstsem:
 "(\<sigma> \<Turnstile> \<rhd>f) =
  (  
   ( (\<sigma> \<Turnstile> f) \<and> ilen \<sigma> = 0) \<or>
   ( ilen \<sigma>>0 \<and> (\<sigma> \<Turnstile> f) \<and> (\<forall>ia<ilen (\<sigma>). (prefix ia  \<sigma> \<Turnstile> \<not>f)))
  ) "
using Fstsem_0 Emptysem FstAndBisem by metis

lemma Lstsem:
 "(\<sigma> \<Turnstile> \<lhd> f) =
  ( ( (\<sigma> \<Turnstile> f) \<and> ilen \<sigma> = 0) \<or>
    ( ilen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> (\<forall>ia < ilen \<sigma>. (suffix ((ilen \<sigma>) -ia)  \<sigma> \<Turnstile> \<not>f)) )
  ) 
 "
proof -
 have "(\<sigma> \<Turnstile> \<lhd> f) = (\<sigma> \<Turnstile> (\<rhd> (f\<^sup>r))\<^sup>r)"  
      using  RRFirstEqvLast by fastforce
 also have " ... = (irev \<sigma> \<Turnstile> \<rhd> (f\<^sup>r))" 
      by (metis reverse_d_def)
 also have " ... =
  (  
   ( (irev \<sigma> \<Turnstile> f\<^sup>r) \<and> ilen (irev \<sigma>) = 0) \<or>
   ( ilen (irev \<sigma>)>0 \<and> (irev \<sigma> \<Turnstile> f\<^sup>r) \<and> 
     (\<forall>ia<ilen (irev \<sigma>). (prefix ia  (irev \<sigma>) \<Turnstile> \<not>(f\<^sup>r))))
  )" 
      using Fstsem by blast
 also have " ... =
  (  
   ( ( \<sigma> \<Turnstile> f) \<and> ilen (\<sigma>) = 0) \<or>
   ( ilen (\<sigma>)>0 \<and> ( \<sigma> \<Turnstile> f) \<and> 
     (\<forall>ia<ilen (irev \<sigma>). (prefix ia  (irev \<sigma>) \<Turnstile> (\<not>(f))\<^sup>r)))
  )" 
  by (simp add: reverse_d_def)
 also have " ... =
  (  
   ( ( \<sigma> \<Turnstile> f) \<and> ilen (\<sigma>) = 0) \<or>
   ( ilen (\<sigma>)>0 \<and> ( \<sigma> \<Turnstile> f) \<and> 
     (\<forall>ia<ilen (irev \<sigma>). (irev (prefix ia  (irev \<sigma>)) \<Turnstile> (\<not>(f)))))
  )" 
  by (simp add: reverse_d_def)
 also have " ... =
  (  
   ( ( \<sigma> \<Turnstile> f) \<and> ilen (\<sigma>) = 0) \<or>
   ( ilen (\<sigma>)>0 \<and> ( \<sigma> \<Turnstile> f) \<and> 
     (\<forall>ia<ilen (\<sigma>). ( (suffix ((ilen \<sigma>) - ia)  (\<sigma>)) \<Turnstile> (\<not>(f)))))
  )" 
      by (simp add: irev_prefix)
 finally show " 
  (\<sigma> \<Turnstile> \<lhd> f) =
  ( ( (\<sigma> \<Turnstile> f) \<and> ilen \<sigma> = 0) \<or>
    ( ilen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> 
      (\<forall>ia < ilen \<sigma>. (suffix ((ilen \<sigma>) -ia)  \<sigma> \<Turnstile> \<not>f)) )
  )  " .
qed

subsubsection \<open>Various Semantic Lemmas\<close>

lemma DiLensem:
 "(\<sigma> \<Turnstile> di (f \<and> len(i))) =
   ( (prefix i \<sigma> \<Turnstile> f) \<and> i\<le>ilen \<sigma>) "
using prefix_ilen_good by (auto simp add: itl_defs)

lemma PrefixFstsem:
 "( (prefix i \<sigma> \<Turnstile> \<rhd>f) \<and> i\<le>ilen \<sigma>) =
  ( i\<le>ilen \<sigma> \<and>
   ( 
    ( (prefix i \<sigma> \<Turnstile> f) \<and> i = 0) \<or>
    ( i>0 \<and> (prefix i \<sigma> \<Turnstile> f) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>f)))
   ) 
 ) "
proof -
 have 1: "( ((prefix i \<sigma>) \<Turnstile> \<rhd>f)) =
  (  
   ( ((prefix i \<sigma>) \<Turnstile> f) \<and> ilen (prefix i \<sigma>) = 0) \<or>
   ( ilen (prefix i \<sigma>)>0 \<and> ((prefix i \<sigma>) \<Turnstile> f) \<and> 
     (\<forall>ia<ilen (prefix i \<sigma>). (prefix ia  (prefix i \<sigma>) \<Turnstile> \<not>f)) ) 
  ) " 
       using Fstsem by blast
 hence 2: "( ((prefix i \<sigma>) \<Turnstile> \<rhd>f) \<and> i\<le>ilen \<sigma>) =
  (  i\<le>ilen \<sigma> \<and> (
   ( ((prefix i \<sigma>) \<Turnstile> f) \<and> ilen (prefix i \<sigma>) = 0) \<or>
   ( ilen (prefix i \<sigma>)>0 \<and> ((prefix i \<sigma>) \<Turnstile> f) \<and> 
     (\<forall>ia<ilen (prefix i \<sigma>). (prefix ia  (prefix i \<sigma>) \<Turnstile> \<not>f)) )
    )
  ) " 
       by auto
 hence 3: "( ((prefix i \<sigma>) \<Turnstile> \<rhd>f) \<and> i\<le>ilen \<sigma>) =
  (  i\<le>ilen \<sigma> \<and> (
   ( ((prefix i \<sigma>) \<Turnstile> f) \<and> i = 0) \<or>
   ( i>0 \<and> ((prefix i \<sigma>) \<Turnstile> f) \<and> (\<forall>ia<i. (prefix ia  (prefix i \<sigma>) \<Turnstile> \<not>f)))
    )
  ) " 
       by auto
 hence 4: "( ((prefix i \<sigma>) \<Turnstile> \<rhd>f) \<and> i\<le>ilen \<sigma>) =
  (  i\<le>ilen \<sigma> \<and> (
   ( ((prefix i \<sigma>) \<Turnstile> f) \<and> i = 0) \<or>
   ( i>0 \<and> ((prefix i \<sigma>) \<Turnstile> f) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>f)))
    )
  ) " 
      using pref_pref_3 using less_imp_add_positive by fastforce
 from 4 show ?thesis by auto
qed

lemma PrefixFstAndsem:
 "( (prefix i \<sigma> \<Turnstile> \<rhd>f \<and> g) \<and> i\<le>ilen \<sigma>) =
  ( i\<le>ilen \<sigma> \<and>
   ( 
    ( (prefix i \<sigma> \<Turnstile> f \<and> g ) \<and> i = 0) \<or>
    ( i>0 \<and> (prefix i \<sigma> \<Turnstile> f \<and> g) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>f)))
   ) 
 ) "
using PrefixFstsem[of "f" "i" "\<sigma>"]    by (metis unl_lift2)

lemma DiLenFstsem:
 "(\<sigma> \<Turnstile> di (\<rhd>f \<and> len(i))) =
   ( i\<le>ilen \<sigma> \<and>
   ( 
    ( (prefix i \<sigma> \<Turnstile> f) \<and> i = 0) \<or>
    ( i>0 \<and> (prefix i \<sigma> \<Turnstile> f) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>f)))
   ) 
 ) "
by (simp add: DiLensem PrefixFstsem)

lemma DiLenFstAndsem:
 "(\<sigma> \<Turnstile> di ((\<rhd>f \<and> g) \<and> len(i))) =
   ( i\<le>ilen \<sigma> \<and>
   ( 
    ( (prefix i \<sigma> \<Turnstile> f \<and> g) \<and> i = 0) \<or>
    ( i>0 \<and> (prefix i \<sigma> \<Turnstile> f \<and> g) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>f)))
   ) 
 ) "
using DiLensem PrefixFstAndsem by metis

lemma FstLenSamesem:
 "( ( i\<le>ilen \<sigma> \<and>
   ( 
    ( (prefix i \<sigma> \<Turnstile> f) \<and> i = 0) \<or>
    ( i>0 \<and> (prefix i \<sigma> \<Turnstile> f) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>f)))
   ) 
 ) \<and>
  ( j\<le>ilen \<sigma> \<and>
   ( 
    ( (prefix j \<sigma> \<Turnstile> f) \<and> j = 0) \<or>
    ( j>0 \<and> (prefix j \<sigma> \<Turnstile> f) \<and> (\<forall>ia<j. (prefix ia \<sigma> \<Turnstile> \<not>f)))
   ) 
 )
) \<longrightarrow> (i=j)
"
by (metis not_less_iff_gr_or_eq unl_lift)


subsection \<open>Theorems\<close>

subsubsection \<open>Fixed length intervals\<close>

lemma LenZeroEqvEmpty:
 "\<turnstile> len(0) = empty"
by (simp add: len_d_def)


lemma LenOneEqvSkip:
 "\<turnstile> len(1) = skip "
by (simp add: len_d_def ChopEmpty) 

lemma LenNPlusOneA:
 "\<turnstile>  len(n+1) = skip;len(n)"
by (simp add: len_d_def)

lemma LenEqvLenChopLen:
 "\<turnstile> len(i+j) = len(i);len(j)"
proof 
 (induct i)
 case 0
 then show ?case 
 by (metis EmptyChop LenZeroEqvEmpty add.left_neutral inteq_reflection)
 next
 case (Suc i)
 then show ?case
 by (metis ChopAssoc add_Suc int_eq len_d_def pow_Suc) 
qed 

lemma LenNPlusOneB:
 "\<turnstile> len(n+1) = len(n);skip"
proof -
 have  1: "\<turnstile> len(n+1) = len(n);len(1)" by (rule LenEqvLenChopLen)
 have  2: "\<turnstile> len(1) = skip" by (rule LenOneEqvSkip)
 hence 3: "\<turnstile> len(n);len(1) = len(n);skip" using RightChopEqvChop by blast
 from 1 3 show ?thesis by fastforce
qed

lemma LenCommute:
 "\<turnstile> (skip;(len n)) = (len n);skip"
proof
  (induct n)
  case 0
  then show ?case 
  by (metis LenEqvLenChopLen LenNPlusOneA LenOneEqvSkip inteq_reflection)
  next
  case (Suc n)
  then show ?case 
  by (metis LenEqvLenChopLen LenNPlusOneA LenOneEqvSkip inteq_reflection)
qed

lemma SkipTrueEqvTrueSkip: 
 "\<turnstile> skip;#True = #True;skip "
using TrueChopSkipEqvSkipChopTrue by fastforce

lemma PowerCommute:
 "\<turnstile> (f;(power f n)) = ((power f n);f)"
proof
  (induct n)
  case 0
  then show ?case using EmptyChop ChopEmpty pow_0 by (metis int_eq)
  next
  case (Suc n)
  then show ?case using ChopAssoc pow_Suc by (metis inteq_reflection)
qed

lemma PowerRev:
 "\<turnstile> (power skip n)\<^sup>r = (power skip n)" 
proof
 (induct n)
 case 0
 then show ?case using REmptyEqvEmpty by auto
 next
 case (Suc n)
 then show ?case using PowerCommute RevChop pow_Suc by (metis RevSkip int_eq)
qed

lemma RLenEqvLen:
 "\<turnstile> (len k)\<^sup>r = (len k)"
proof
  (induct k)
  case 0
  then show ?case
  by (metis PowerRev len_d_def) 
  next
  case (Suc k)
  then show ?case 
  by (metis PowerRev len_d_def)
qed

lemma PowerSkipEqvLen:
 "\<turnstile> (power skip n) = (len n)"
by (simp add: len_d_def)

lemma ExistsLen:
 "\<turnstile> \<exists> k. len(k)"
by (simp add: len_defs Valid_def)

lemma AndExistsLen:
 "\<turnstile> f = (f \<and> (\<exists>k. len(k)))"
using ExistsLen by fastforce

lemma AndExistsLenChop:
 "\<turnstile> (f;g) =  (\<exists>k. (f \<and> len(k));g)  "
by (simp add: Valid_def len_defs chop_defs)

lemma AndExistsLenChopR:
 "\<turnstile> (f;g) = (\<exists>k. f;(g \<and> len(k))) "
by (simp add: Valid_def len_defs chop_defs)

lemma LFixedAndDistr:
 "\<turnstile> ((f0 \<and> len(k));g0 \<and> (f1 \<and> len(k));g1) = ((f0 \<and> f1) \<and> len(k));(g0 \<and> g1) "
by (auto simp add: Valid_def len_defs chop_defs)
  

lemma RFixedAndDistr:
 "\<turnstile> (f0;(g0 \<and> len(k)) \<and> f1;(g1 \<and> len(k))) = (f0 \<and> f1);((g0 \<and> g1) \<and> len(k))"
by (simp add: Valid_def len_defs chop_defs) (metis diff_diff_cancel)

lemma LFixedAndDistrA:
 "\<turnstile> ((f0 \<and> len(k));g0 \<and> (f1 \<and> len(k));g0) = ((f0 \<and> f1) \<and> len(k));g0 "
proof -
 have 1: "\<turnstile> ((f0 \<and> len(k));g0 \<and> (f1 \<and> len(k));g0) = ((f0 \<and> f1) \<and> len(k));(g0 \<and> g0) " 
      by (rule LFixedAndDistr)
 have 2: "\<turnstile> ((f0 \<and> f1) \<and> len(k));(g0 \<and> g0) = ((f0 \<and> f1) \<and> len(k));g0" 
      by auto
 from 1 2 show ?thesis by fastforce
qed

lemma LFixedAndDistrB:
 "\<turnstile> ((f0 \<and> len(k));g0 \<and> (f0 \<and> len(k));g1) = (f0 \<and> len(k));(g0 \<and> g1)"
proof -
 have 1: "\<turnstile> ((f0 \<and> len(k));g0 \<and> (f0 \<and> len(k));g1) = ((f0 \<and> f0) \<and> len(k));(g0 \<and> g1) " 
      by (rule LFixedAndDistr)
 have 2: "\<turnstile> ((f0 \<and> f0) \<and> len(k));(g0 \<and> g1) = (f0 \<and> len(k));(g0 \<and> g1)" 
     by auto
 from 1 2 show ?thesis by fastforce
qed

lemma LFixedAndDistrB1:
 "\<turnstile> (len(k);f \<and> len(k);g) = len(k);(f \<and> g) "
proof -
 have 1: "\<turnstile> len(k);f = (#True \<and> len(k));f " 
      by auto
 have 2: "\<turnstile> len(k);g = (#True \<and> len(k));g " 
      by auto
 have 3: "\<turnstile> (len(k);f \<and> len(k);g) = ((#True \<and> len(k));f \<and> (#True \<and> len(k));g)" 
      using "1" "2" by auto
 have 4: "\<turnstile> ((#True \<and> len(k));f \<and> (#True \<and> len(k));g) = (#True \<and> len(k));(f \<and> g)" 
      using LFixedAndDistrB by blast
 have 5: "\<turnstile> (#True \<and> len(k));(f \<and> g) = (len(k));(f \<and> g)" 
      by auto
 from 1 2 3 4 5 show ?thesis by auto
qed

lemma RFixedAndDistrA:
 "\<turnstile> (f0;(g0 \<and> len(k)) \<and> f0;(g1 \<and> len(k))) = f0;((g0 \<and> g1) \<and> len(k)) "
proof -
 have 1: "\<turnstile> (f0;(g0 \<and> len(k)) \<and> f0;(g1 \<and> len(k))) = (f0 \<and> f0);((g0 \<and> g1) \<and> len(k)) " 
      by (rule RFixedAndDistr)
 have 2: "\<turnstile> (f0 \<and> f0);((g0 \<and> g1) \<and> len(k)) = f0;((g0 \<and> g1) \<and> len(k)) " 
      by auto
from 1 2 show ?thesis by fastforce
qed

lemma RFixedAndDistrB:
 "\<turnstile> (f0;(g0 \<and> len(k)) \<and> f1;(g0 \<and> len(k))) = (f0 \<and> f1);(g0 \<and> len(k)) "
proof -
 have 1: "\<turnstile> (f0;(g0 \<and> len(k)) \<and> f1;(g0 \<and> len(k))) = (f0 \<and> f1);((g0 \<and> g0) \<and> len(k)) " 
      by (rule RFixedAndDistr)
 have 2: "\<turnstile> (f0 \<and> f1);((g0 \<and> g0) \<and> len(k)) = (f0 \<and> f1);(g0 \<and> len(k)) " 
      by auto
from 1 2 show ?thesis by fastforce
qed 

lemma ChopSkipAndChopSkip:
 "\<turnstile> (f0;skip \<and> f1;skip) = (f0 \<and> f1);skip"
proof -
 have 1: "\<turnstile> (f0;(#True \<and> len(1)) \<and> f1;(#True \<and> len(1))) = (f0 \<and> f1);(#True \<and> len(1)) " 
      by (rule RFixedAndDistrB)
 have 2: "\<turnstile> (#True \<and> len(1)) = skip " 
      using LenOneEqvSkip by fastforce
 hence 3: "\<turnstile> f0;(#True \<and> len(1)) = f0;skip " 
      using RightChopEqvChop by blast
 have 4: "\<turnstile> f1;(#True \<and> len(1)) = f1;skip " 
      using "2" RightChopEqvChop by blast
 have 5: "\<turnstile> (f0;(#True \<and> len(1)) \<and> f1;(#True \<and> len(1))) = (f0;skip \<and> f1;skip) " 
      using "3" "4" by fastforce
 have 6: "\<turnstile> (f0 \<and> f1);(#True \<and> len(1)) = (f0 \<and> f1);skip" 
      using "2" RightChopEqvChop by blast
 from 1 5 6 show ?thesis by fastforce
qed

lemma BiAndChopSkipEqv:
 "\<turnstile> (bi (f \<and> g));skip = ((bi f);skip \<and> (bi g);skip)"
proof -
 have 1: "\<turnstile> bi (f \<and> g) = ((bi f) \<and> (bi g)) " 
      by (auto simp add: bi_defs Valid_def)
 hence 2: "\<turnstile> (bi (f \<and> g));skip = (bi f \<and> bi g);skip" 
      by (rule LeftChopEqvChop)
 have 3: "\<turnstile> (bi f \<and> bi g);skip = ((bi f);skip \<and> (bi g);skip)" 
      using ChopSkipAndChopSkip  by fastforce
 from 2 3 show ?thesis by fastforce
qed

lemma DiAndChopSkipEqv:
 "\<turnstile> (di (f \<and> g));skip \<longrightarrow> (di f);skip \<and> (di g);skip"
proof -
 have 1: "\<turnstile> di (f \<and> g) \<longrightarrow> (di f) \<and> (di g) " 
      by (simp add: DiAndImpAnd)
 hence 2: "\<turnstile> (di (f \<and> g));skip \<longrightarrow> (di f \<and> di g);skip" 
      by (rule LeftChopImpChop)
 have 3: "\<turnstile> (di f \<and> di g);skip = ((di f);skip \<and> (di g);skip)" 
      using ChopSkipAndChopSkip by fastforce
 from 2 3 show ?thesis by fastforce
qed

lemma ChopEmptyAndEmpty:
 "\<turnstile> (f;g \<and> empty) = (f \<and> g \<and> empty)"
by (simp add: Valid_def itl_defs)
   (metis prefix_ilen suffix_zero le_zero_eq)

lemma ChopSkipImpMore:
 "\<turnstile> f;skip \<longrightarrow> more"
using ChopImpDiamond MoreEqvSkipChopTrue SkipTrueEqvTrueSkip TrueChopEqvDiamond by fastforce

lemma MoreEqvMoreChopTrue:
 "\<turnstile> more = more;#True"
proof -
 have 1: "\<turnstile> more = skip;#True" 
      using MoreEqvSkipChopTrue by blast
 have 2: "\<turnstile> #True = #True;#True" 
      by (auto simp add: Valid_def chop_defs)
 hence 3: "\<turnstile> skip;#True = skip;(#True;#True)" 
      using RightChopEqvChop by blast
 have 4: "\<turnstile> skip;(#True;#True) = (skip;#True);#True" 
      using ChopAssoc by blast
 have 5: "\<turnstile> (skip;#True);#True = more;#True" 
      using MoreEqvSkipChopTrue by (simp add: more_d_def next_d_def)
 from 1 3 4 5 show ?thesis by fastforce
qed


lemma NotNotChopSkip:
 "\<turnstile> (\<not>((\<not> f) ;skip)) = (empty \<or> (f;skip)) "
by (metis WprevEqvEmptyOrPrev prev_d_def wprev_d_def)

lemma NotChopFixed:
 "\<turnstile> (\<not>(f;(g \<and> len(k)))) = (\<not>(\<diamond>(g \<and> len(k))) \<or> ((\<not>f);(g \<and> len(k))))"
by (auto simp add: itl_defs Valid_def)
   (metis diff_diff_cancel)

lemma NotFixedChop:
 "\<turnstile> (\<not>((g \<and> len(k));f)) = (\<not>(di(g \<and> len(k))) \<or> ((g \<and> len(k));(\<not>f)))"
by (auto simp add: itl_defs Valid_def) 

lemma NotChopNotSkip:
 "\<turnstile> (\<not>(f;skip)) = (empty \<or> ((\<not> f);skip))"
proof -
 have 1: "\<turnstile> (\<not>((\<not>(\<not>f));skip)) = (empty \<or> ((\<not> f);skip))" using NotNotChopSkip by blast
 have 2: "\<turnstile> (\<not>((\<not>(\<not>f));skip)) = (\<not>(f;skip))" by auto
 from 1 2 show ?thesis by auto
qed

subsubsection \<open>Additional ITL theorems\<close>

lemma BiOrBiImpBiOr:
 "\<turnstile> bi f \<or> bi g \<longrightarrow> bi(f \<or> g) "
proof -
 have 1: "\<turnstile> f \<longrightarrow> f \<or> g" by auto
 hence 2: "\<turnstile> bi f \<longrightarrow> bi(f \<or> g)" by (rule BiImpBiRule)
 have 3: "\<turnstile> g \<longrightarrow> f \<or> g" by auto
 hence 4: "\<turnstile> bi g \<longrightarrow> bi(f \<or> g)" by (rule BiImpBiRule)
 from 2 4 show ?thesis by fastforce
qed

lemma MoreAndBiImpBiChopSkip:
 "\<turnstile> more \<and> bi f \<longrightarrow> (bi f);skip "
proof -
 have 1: "\<turnstile>  (bi f);skip = ((\<not>(di (\<not> f)));skip) " by (simp add: bi_d_def)
 have 2: "\<turnstile> (\<not>( (\<not>(di (\<not> f)));skip)) = (empty \<or> (di (\<not> f));skip)" by (rule NotNotChopSkip)
 have 3: "\<turnstile> empty \<longrightarrow> empty \<or> di (\<not> f) " by auto
 have 4: "\<turnstile> (di (\<not> f));skip \<longrightarrow> di (\<not> f)" using ChopImpDi DiEqvDiDi by fastforce
 hence 5: "\<turnstile> (di (\<not> f));skip \<longrightarrow> empty \<or> di (\<not> f)" by (rule Prop05)
 have 6: "\<turnstile> \<not>( (\<not>(di (\<not> f)));skip) \<longrightarrow> empty \<or> di (\<not> f)" using "2" "3" "5" by fastforce
 hence 7: "\<turnstile> \<not>(empty \<or> di (\<not> f)) \<longrightarrow> \<not>(\<not>( (\<not>(di (\<not> f)));skip)) "  by fastforce
 have 8: "\<turnstile> (\<not>( \<not>( (\<not>(di (\<not> f)));skip))) = ((\<not>(di (\<not> f)));skip)" by auto
 have 9: "\<turnstile> (\<not>(empty \<or> di (\<not> f))) = (more \<and> \<not>( di (\<not> f))) "
 using NotEmptyEqvMore by auto 
 have 10: "\<turnstile> (more \<and> \<not>( di (\<not> f))) = (more \<and> bi f)" by (simp add: bi_d_def)
 from 1 6 7 8 9 10 show ?thesis by (metis int_eq)
qed

lemma DiChopImpDiB:
 "\<turnstile> di(f;g) \<longrightarrow> di f "
proof -
 have 1: "\<turnstile> f ; (g;#True) \<longrightarrow> di f" by (rule ChopImpDi)
 have 2: "\<turnstile> f ; (g;#True) = (f;g);#True " by (rule ChopAssoc)
 from 1 2 show ?thesis by (metis di_d_def int_eq)
qed

lemma BiBiOrImpBi:
 "\<turnstile> bi ( bi f \<or> bi g) \<longrightarrow> bi f \<or> bi g"
using BiElim by auto

lemma BiImpBiBiOr:
 "\<turnstile> bi f  \<longrightarrow> bi ( bi f \<or> bi g)"
proof -
 have 1: "\<turnstile> bi f \<longrightarrow> bi f \<or> bi g" by auto
 hence 2: "\<turnstile> bi (bi f) \<longrightarrow> bi(bi f \<or> bi g)  " using BiImpBiRule by blast
 have 3: "\<turnstile> bi (bi f) = bi f" using BiEqvBiBi by fastforce
 from 2 3 show ?thesis by fastforce
qed

lemma BiImpBiBiOrB:
 "\<turnstile> bi g  \<longrightarrow> bi ( bi f \<or> bi g)"
proof -
 have 1: "\<turnstile> bi g \<longrightarrow> bi f \<or> bi g" by auto
 hence 2: "\<turnstile> bi (bi g) \<longrightarrow> bi(bi f \<or> bi g)  " using BiImpBiRule by blast
 have 3: "\<turnstile> bi (bi g) = bi g" using BiEqvBiBi by fastforce
 from 2 3 show ?thesis by fastforce
qed

lemma BiBiOrEqvBi:
 "\<turnstile> bi ( bi f \<or> bi g) = bi f \<or> bi g "
proof -
 have 1: "\<turnstile> bi ( bi f \<or> bi g) \<longrightarrow> bi f \<or> bi g " by (rule BiBiOrImpBi)
 have 2: "\<turnstile> bi f  \<longrightarrow> bi ( bi f \<or> bi g) " by (rule BiImpBiBiOr)
 have 3: "\<turnstile> bi g  \<longrightarrow> bi ( bi f \<or> bi g) " by (rule BiImpBiBiOrB)
 have 4: "\<turnstile> bi f  \<or> bi g \<longrightarrow> bi ( bi f \<or> bi g) " using "2" "3" by fastforce
 from 1 4 show ?thesis by fastforce
qed

lemma DiEqvOrDiChopSkipA:
 "\<turnstile> di f = (f \<or> di(f;skip))"
proof -
 have 1: "\<turnstile> di f = f ;#True " by (simp add: di_d_def)
 hence 2: "\<turnstile> di f = f; ( empty \<or> more)" by (simp add: empty_d_def)
 hence 3: "\<turnstile>f; ( empty \<or> more) = (f;empty \<or> f;more) " using ChopOrEqv by blast
 have 4: "\<turnstile> f;empty = f" by (rule ChopEmpty)
 have 5: "\<turnstile> more = skip;#True" using MoreEqvSkipChopTrue by blast
 hence 6: "\<turnstile> f;more = f;(skip;#True)" using RightChopEqvChop by blast
 have 7: "\<turnstile> f;(skip;#True) = (f;skip);#True" by (rule ChopAssoc)
 from 2 3 4 6 7 show ?thesis by (metis di_d_def int_eq)
qed 

lemma DiEqvOrDiChopSkipB:
 "\<turnstile> di f = (f \<or> (di f);skip)"
proof -
 have 1: "\<turnstile> (di f) = (f \<or> di(f;skip))" by (rule DiEqvOrDiChopSkipA)
 have 2: "\<turnstile>  di(f;skip) = (f;skip);#True " by (simp add: di_d_def)
 have 3: "\<turnstile> (f;skip);#True = f;(skip;#True)" by (rule ChopAssocB)
 have 4: "\<turnstile> di(f;skip) = f;(skip;#True) " using "2" "3" by fastforce
 have 5: "\<turnstile> skip;#True = #True;skip" by (rule SkipTrueEqvTrueSkip)
 hence 6: "\<turnstile> f;(skip;#True) = f;(#True;skip)" using RightChopEqvChop by blast
 have 7: "\<turnstile> di(f;skip) = f;(#True;skip)" using "4" "6" by fastforce
 have 8: "\<turnstile> f;(#True;skip) = (f;#True);skip" by (rule ChopAssoc)
 have 9: "\<turnstile> (f;#True);skip = (di f);skip" by (simp add: di_d_def)
 have 10 : "\<turnstile> di(f;skip) =(di f);skip " using "7" "8" "9" by fastforce
 hence 11: "\<turnstile> (f \<or> di(f;skip)) = (f \<or> (di f);skip) " by auto
 from 1 11 show ?thesis by fastforce
qed
  
lemma BiEqvAndEmptyOrBiChopSkip:
 "\<turnstile> bi f = (f \<and> (empty \<or> (bi f);skip)) "
proof -
 have 1: "\<turnstile> di (\<not> f) = (\<not> f \<or> (di (\<not> f);skip))" by (rule DiEqvOrDiChopSkipB)
 have 2: "\<turnstile> di (\<not> f) = (\<not>( bi f ))" by (rule DiNotEqvNotBi)
 have 3: "\<turnstile> (\<not>( bi f )) =  (\<not> f \<or> (di (\<not> f);skip)) " using "1" "2" by fastforce
 hence 4: "\<turnstile> bi f = (\<not>(\<not> f \<or> (di (\<not> f);skip))) " by auto
 have 5: "\<turnstile> (\<not>(\<not> f \<or> (di (\<not> f);skip))) = (f \<and> \<not>(di (\<not> f);skip)) " by auto
 have 6: "\<turnstile> di (\<not> f);skip = ((\<not>(bi f));skip) " by (simp add: "2" LeftChopEqvChop)
 hence 7: "\<turnstile> (\<not>(di (\<not> f);skip)) = (\<not>((\<not>(bi f));skip)) " by auto
 have 8: "\<turnstile> (\<not>((\<not>(bi f));skip)) = (empty \<or> (bi f);skip) " using NotNotChopSkip by blast
 hence 9: "\<turnstile> (f \<and> \<not>(di (\<not> f);skip)) = (f \<and> (empty \<or> (bi f);skip)) " using "7" "8" by fastforce
 from 4 5 9 show ?thesis by fastforce
qed

lemma DiDiAndEqvDi:
 "\<turnstile> di ( di f \<and> di g) = (di f \<and> di g) "
proof -
 have 1: "\<turnstile> bi ( bi (\<not> f) \<or> bi (\<not> g)) = (bi (\<not>f) \<or> bi (\<not> g)) " 
      by (meson BiBiOrImpBi BiImpBiBiOr BiImpBiBiOrB Prop02 int_iffI)
 have 2: "\<turnstile> bi (\<not>f) = (\<not> (di f)) " 
      by (simp add: bi_d_def)
 have 3: "\<turnstile> bi (\<not>g) = (\<not> (di g)) " 
      by (simp add: bi_d_def)
 have 4: "\<turnstile> (bi (\<not>f) \<or> bi (\<not> g)) = (\<not> (di f) \<or> \<not> (di g)) " 
      using "2" "3" by fastforce
 have 5: "\<turnstile> (\<not> (di f) \<or> \<not> (di g)) = (\<not>(di f \<and> di g)) " 
      by auto
 have 6: "\<turnstile> bi ( bi (\<not> f) \<or> bi (\<not> g)) = (\<not>(di f \<and> di g)) " 
      using "1" "5" "4" by fastforce
 hence 7: "\<turnstile> (\<not>(bi ( bi (\<not> f) \<or> bi (\<not> g)))) = (di f \<and> di g)  " 
      by auto
 have 8 : "\<turnstile> (\<not>(bi ( bi (\<not> f) \<or> bi (\<not> g)))) = di ( \<not>(bi (\<not> f) \<or> bi (\<not> g)))  " 
      using DiNotEqvNotBi  by fastforce
 have 9 : "\<turnstile> (\<not>(bi (\<not> f) \<or> bi (\<not> g))) = (di f \<and> di g) " 
      using "1" "7" by fastforce
 hence 10: "\<turnstile> di ( \<not>(bi (\<not> f) \<or> bi (\<not> g))) = di ( di f \<and> di g) " 
      using DiEqvDi by blast
 from 7 8 10 show ?thesis by fastforce 
qed

lemma BiInduct:
 "\<turnstile> bi(f \<longrightarrow> wprev f) \<and> f \<longrightarrow> bi f "
proof -
 have 1: "\<turnstile> \<box>((f\<^sup>r) \<longrightarrow> wnext(f\<^sup>r)) \<and> f\<^sup>r \<longrightarrow> \<box>(f\<^sup>r)" using BoxInduct by blast
 hence 2: "\<turnstile> (\<box>((f\<^sup>r) \<longrightarrow> wnext(f\<^sup>r)) \<and> f\<^sup>r \<longrightarrow> \<box>(f\<^sup>r))\<^sup>r" using ReverseEqv by blast
 have 3: "\<turnstile> ((f\<^sup>r)\<^sup>r ) = f" by (simp add: EqvReverseReverse)
 have 4: "\<turnstile> (\<box> (f\<^sup>r))\<^sup>r = bi (f)  " using RRBoxEqvBi by blast
 have 5: "\<turnstile> ((f\<^sup>r) \<longrightarrow> wnext(f\<^sup>r))\<^sup>r = ( (f\<^sup>r)\<^sup>r \<longrightarrow> (wnext(f\<^sup>r))\<^sup>r )" by (simp add: rev_fun2)
 have 6: "\<turnstile> (wnext(f\<^sup>r))\<^sup>r = wprev(f)" using RRWNextEqvWPrev by blast 
 have 7: "\<turnstile> ( (f\<^sup>r)\<^sup>r \<longrightarrow> (wnext(f\<^sup>r))\<^sup>r ) = ( f \<longrightarrow> wprev(f) )" using "6" "3" by fastforce
 have 8: "\<turnstile> bi( (f\<^sup>r)\<^sup>r \<longrightarrow> (wnext(f\<^sup>r))\<^sup>r ) = bi( f \<longrightarrow> wprev(f) )" using "7" "3" BiEqvBi by blast
 have 9: "\<turnstile> (\<box>((f\<^sup>r) \<longrightarrow> wnext(f\<^sup>r)))\<^sup>r = bi ( ((f\<^sup>r) \<longrightarrow> wnext(f\<^sup>r))\<^sup>r ) " using RBoxEqvBi by blast
 have 10: "\<turnstile> (\<box>((f\<^sup>r) \<longrightarrow> wnext(f\<^sup>r)))\<^sup>r = bi( f \<longrightarrow> wprev(f) )" using "8" "9" "5" int_eq by fastforce
 have 11: "\<turnstile> (\<box>((f\<^sup>r) \<longrightarrow> wnext(f\<^sup>r)) \<and> f\<^sup>r \<longrightarrow> \<box>(f\<^sup>r))\<^sup>r = 
             (((\<box>((f\<^sup>r) \<longrightarrow> wnext(f\<^sup>r)))\<^sup>r \<and> (f\<^sup>r)\<^sup>r \<longrightarrow> (\<box>(f\<^sup>r))\<^sup>r))"  by (metis int_eq rev_fun2)
 have 12: "\<turnstile> ((\<box>((f\<^sup>r) \<longrightarrow> wnext(f\<^sup>r)))\<^sup>r \<and> (f\<^sup>r)\<^sup>r \<longrightarrow> (\<box>(f\<^sup>r))\<^sup>r) =
            (bi( f \<longrightarrow> wprev(f) ) \<and> f \<longrightarrow> bi f ) " using "8" "3" "4" "10" by fastforce
 from 2 11 12 show ?thesis using MP by fastforce
qed

lemma PrevLoop:
 assumes "\<turnstile> f \<longrightarrow> prev f"
 shows   "\<turnstile> \<not> f"
proof -
 have  1: "\<turnstile> f \<longrightarrow> prev f " using assms by auto
 hence 2: "\<turnstile> f \<longrightarrow> ( more \<and> wprev f) " 
   by (metis ChopSkipImpMore Prop05 Prop12 WprevEqvEmptyOrPrev inteq_reflection 
       lift_imp_trans prev_d_def)
 hence 3: "\<turnstile> f \<longrightarrow> wprev f " by auto
 hence 4: "\<turnstile> bi(f \<longrightarrow> wprev f)" by (rule BiGen)
 have  5: "\<turnstile> bi(f \<longrightarrow> wprev f) \<and> f \<longrightarrow> bi f " by (rule BiInduct)
 hence 6: "\<turnstile> bi(f \<longrightarrow> wprev f) \<longrightarrow> (f \<longrightarrow> bi f)" by fastforce
 have  7: "\<turnstile> (f \<longrightarrow> bi f)" using "4" "6" MP by blast
 have  8: "\<turnstile> bi f \<longrightarrow> f" by (rule BiElim)
 have  9: "\<turnstile> f = bi f" using "7" "8" by fastforce
 have 10: "\<turnstile> f \<longrightarrow> more" using "2" by auto
 hence 11: "\<turnstile> bi f \<longrightarrow>  bi more" using BiImpBiRule by blast
 have  12: "\<turnstile> \<not>(bi more)" using DiEmpty bi_d_def empty_d_def by (simp add: bi_d_def empty_d_def)
 from 7 9 11 12 show ?thesis using MP by fastforce 
qed

lemma PrevImpNotPrevNot:
  "\<turnstile> prev f \<longrightarrow> \<not>(prev (\<not> f)) "
by (metis (no_types, lifting) NextImpNotNextNot RPrevEqvNext ReverseEqv inteq_reflection 
    rev_fun1 rev_fun2)

lemma BiEqvAndWprevBi:
 "\<turnstile> bi f = (f \<and> wprev(bi f))"
by (metis BiEqvAndEmptyOrBiChopSkip WprevEqvEmptyOrPrev inteq_reflection prev_d_def) 

 
lemma DiIntroLoop:
 assumes "\<turnstile> (f \<and> \<not> g) \<longrightarrow> prev f"
 shows "\<turnstile> f \<longrightarrow> di g"
using assms DiamondIntro[of "LIFT(f\<^sup>r)" "LIFT(g\<^sup>r)"]  
by (metis (no_types, lifting) RDiEqvDiamond RPrevEqvNext ReverseEqv inteq_reflection 
    rev_fun2 rev_fun1)


lemma DiEqvOrChopMore:
 "\<turnstile> di f = (f \<or> f;more)"
proof -
 have 1: "\<turnstile> di f = f;#True" by (simp add: di_d_def)
 hence 2: "\<turnstile> di f = f; (empty \<or> more) " by (simp add: empty_d_def)
 have  3: "\<turnstile> f; (empty \<or> more) = (f;empty \<or> f;more) " by (simp add: ChopOrEqv)
 have  4: "\<turnstile> f;empty = f " by (rule ChopEmpty)
 from 2 3 4 show ?thesis by fastforce
qed


lemma DiAndDiEqvDiAndDiOrDiAndDi:
 "\<turnstile> (di f \<and> di g) = (di(f \<and> di g) \<or> di(g \<and> di f)) "
proof - 
 have 1: "\<turnstile> di f  = (f \<or> f;more) " 
      using DiEqvOrChopMore by blast
 have 2: "\<turnstile> di g  = (g \<or> g;more) " 
      using DiEqvOrChopMore by blast
 have 3: "\<turnstile> (di f \<and> di g) = ((f \<or> f;more) \<and> (g \<or> g;more))  "  
      using "1" "2" by fastforce
 have 4: "\<turnstile> ((f \<or> f;more) \<and> (g \<or> g;more)) =
            ((f \<and> g) \<or> ( f \<and> g;more) \<or> (g \<and> f;more) \<or> (f;more \<and> g;more )) " 
      by auto
 have 5: "\<turnstile> more = #True;skip" 
      using MoreEqvSkipChopTrue SkipTrueEqvTrueSkip by fastforce
 hence 6: "\<turnstile> f;more = f;(#True;skip)" 
      using RightChopEqvChop by blast
 have  7: "\<turnstile> f;(#True;skip) = (f;#True);skip " 
      by (rule ChopAssoc)
 have  8: "\<turnstile> f;more = prev (di f) " 
      using "6" "7" by (metis di_d_def int_eq prev_d_def)
 have  9: "\<turnstile> g;more = g;(#True;skip) " 
      using "5"  RightChopEqvChop by blast
 have 10: "\<turnstile> g;(#True;skip) = (g;#True);skip " 
      by (rule ChopAssoc)
 have 11: "\<turnstile> g;more = prev (di g) " 
      using "9" "10" by (metis di_d_def int_eq prev_d_def)
 have 12: "\<turnstile> (f;more \<and> g;more) = (prev (di f) \<and> prev (di g)) " 
      using "8" "11" by fastforce
 hence 13: "\<turnstile> (f;more \<and> g;more) = prev (di f \<and> di g) " 
      by (metis ChopSkipAndChopSkip int_eq prev_d_def)
 have 14: "\<turnstile> (di f \<and> di g) = 
            ((f \<and> g) \<or> ( f \<and> g;more) \<or> (g \<and> f;more)) \<or> (f;more \<and> g;more) " 
      using "3" "4" by auto
 have 15: "\<turnstile> (di f \<and> di g) = 
            ((f \<and> g) \<or> ( f \<and> g;more) \<or> (g \<and> f;more)) \<or> prev (di f \<and> di g) "
            by (metis "10" "14" "6" "7" "9" ChopSkipAndChopSkip di_d_def int_eq prev_d_def) 
 hence 16: "\<turnstile> (di f \<and> di g) \<longrightarrow> 
            ((f \<and> g) \<or> ( f \<and> g;more) \<or> (g \<and> f;more)) \<or> prev (di f \<and> di g) " 
      by fastforce
 hence 17: "\<turnstile> (di f \<and> di g) \<and> \<not>((f \<and> g) \<or> ( f \<and> g;more) \<or> (g \<and> f;more)) \<longrightarrow>  
               prev (di f \<and> di g)  " 
      by fastforce
 hence 18: "\<turnstile> (di f \<and> di g) \<longrightarrow> di((f \<and> g) \<or> ( f \<and> g;more) \<or> (g \<and> f;more))" 
      using DiIntroLoop by blast
 have  19: "\<turnstile> di((f \<and> g) \<or> ( f \<and> g;more) \<or> (g \<and> f;more)) = 
              (di(f \<and> g) \<or> di( f \<and> g;more) \<or> di(g \<and> f;more))" 
      by (meson DiOrEqv Prop06)
 have  20: "\<turnstile> f   \<longrightarrow> di f  " 
      using DiIntro by blast 
 hence 21: "\<turnstile> f \<and> g \<longrightarrow> g \<and> di f" 
      by auto
 hence 22: "\<turnstile> di(f \<and> g) \<longrightarrow> di(g \<and> di f) " 
      using DiImpDi by blast
 hence 23: "\<turnstile> di(f \<and> g) \<longrightarrow> di(g \<and> di f) \<or> di(f \<and> di g) " 
      by auto
 have  24: "\<turnstile> g;more \<longrightarrow> di g" 
      by (simp add: ChopImpDi)
 hence 25: "\<turnstile> f \<and> g;more \<longrightarrow> f \<and> di g" 
      by auto
 hence 26: "\<turnstile> di(f \<and> g;more) \<longrightarrow> di(f \<and> di g)" 
      using DiImpDi by blast
 hence 27: "\<turnstile> di(f \<and> g;more) \<longrightarrow>    di(f \<and> di g) \<or> di(g \<and> di f) " 
      by auto
 have  28: "\<turnstile> f;more \<longrightarrow> di f" 
      by (simp add: ChopImpDi)
 hence 29: "\<turnstile> g \<and> f;more \<longrightarrow> g \<and> di f" 
      by auto
 hence 30: "\<turnstile> di(g \<and> f;more) \<longrightarrow> di(g \<and> di f)" 
      using DiImpDi by blast
 hence 31: "\<turnstile> di(g \<and> f;more) \<longrightarrow>  di(f \<and> di g) \<or>  di(g \<and> di f)  " 
      by auto
 have  32: "\<turnstile>  di(f \<and> g) \<or> di( f \<and> g;more) \<or> di(g \<and> f;more) \<longrightarrow> 
               di(f \<and> di g) \<or>  di(g \<and> di f)  " 
      using "23" "27" "31" by fastforce
 have  33: "\<turnstile> di((f \<and> g) \<or> ( f \<and> g;more) \<or> (g \<and> f;more)) \<longrightarrow> 
              di(f \<and> di g) \<or>  di(g \<and> di f)  " 
      using "19" "32" by fastforce
 have  34: "\<turnstile> (di f \<and> di g) \<longrightarrow> di(f \<and> di g) \<or>  di(g \<and> di f)" 
      using "18" "33" by fastforce
 have  35: "\<turnstile>  f  \<longrightarrow> di f "  
      using DiIntro by blast 
 hence 36: "\<turnstile> f \<and> di g \<longrightarrow> di f \<and> di g" 
      by auto
 hence 37: "\<turnstile> di (f \<and> di g) \<longrightarrow> di (di f \<and> di g) "  
      using DiImpDi by blast
 have  38: "\<turnstile> di (di f \<and> di g) = (di f \<and> di g)" 
      using DiDiAndEqvDi by blast
 have  39: "\<turnstile> di (f \<and> di g) \<longrightarrow> di f \<and> di g " 
      using "37" "38" by fastforce
 have  40: "\<turnstile>  g  \<longrightarrow> di g "  
      using DiIntro by blast 
 hence 41: "\<turnstile>   g \<and> di f \<longrightarrow> di f \<and> di g" 
      by auto
 hence 42: "\<turnstile> di (g \<and> di f) \<longrightarrow> di (di f \<and> di g) "  
      using DiImpDi by blast
 have  43: "\<turnstile> di (di f \<and> di g) = (di f \<and> di g)" 
      using DiDiAndEqvDi by fastforce
 have  44: "\<turnstile> di (g \<and> di f) \<longrightarrow> di f \<and> di g " 
      using "42" "43" by fastforce
 have  45: "\<turnstile> di (f \<and> di g) \<or> di (g \<and> di f) \<longrightarrow> di f \<and> di g" 
      using "39" "44"  by fastforce
 from 34 45 show ?thesis by fastforce
qed 

lemma BoxStateEqvBiFinState:
 "\<turnstile> \<box> (init w) = bi (fin (init w))"
proof -
  have 1: "\<turnstile> \<diamond> (\<not> (init w)) = #True ; (\<not>(init w))" 
       by (simp add: sometimes_d_def)
  have 2: "\<turnstile> \<diamond> (init(\<not> w)) = #True ; init (\<not> w)" 
       by (simp add: sometimes_d_def)
  have 3: "\<turnstile> di (#True \<and> fin (init (\<not> w))) = #True ; init (\<not> w)" 
       using DiAndFinEqvChopState by blast
  have 4: "\<turnstile> \<diamond> (init(\<not> w)) = di (#True \<and> fin (init (\<not> w))) " 
       using 1 2 3 by fastforce
  have 5: "\<turnstile> (\<not> (\<diamond> (init(\<not> w)))) = (\<not> (di (#True \<and> fin (init (\<not> w)))))" 
       using 4 by fastforce
  have 6: "\<turnstile> \<box> (init w) = (\<not> (di (#True \<and> fin (init (\<not> w)))))" 
       using 5 always_d_def Initprop(2) by (metis int_eq)
  have 7: "\<turnstile> \<box> (init w) = bi (\<not> (fin (init (\<not> w))))" 
       using 6 by (simp add: bi_d_def)
  have 8: "\<turnstile> init (\<not> w) = (\<not> (init w))" 
       using Initprop(2) by fastforce
  have 9: "\<turnstile> fin (init (\<not> w)) = fin (\<not> (init w))" 
       using 8 FinEqvFin by blast
  have 10: "\<turnstile> fin (init (\<not> w)) = (\<not> (fin (init w)))" 
       using 8 FinNotStateEqvNotFinState FinEqvFin by blast
  have 11: "\<turnstile> (\<not> (fin (init (\<not> w)))) = (fin (init w))" 
       using 10  by fastforce
  have 12: "\<turnstile> bi (\<not> (fin (init (\<not> w)))) = bi (fin (init w))" 
       using 11  by (simp add: BiEqvBi)
  have 13: "\<turnstile> \<box> (init w) = bi (fin (init w))" 
       using 7 12 by fastforce
  from 13 show ?thesis by simp
qed

lemma DiamondStateEqvDiFinState:
 "\<turnstile> \<diamond> (init w) = di (fin (init w))"
proof -
  have 1: "\<turnstile> \<box> (init (\<not> w)) = bi (fin (init (\<not> w)))" 
       using BoxStateEqvBiFinState by blast
  have 2: "\<turnstile> (\<not> (\<box> (init (\<not> w)))) = (\<not> (bi (fin (init (\<not> w)))))" 
       using 1 by auto
  have 3: "\<turnstile> \<diamond> (\<not> (init (\<not> w))) = di (\<not> (fin (init (\<not> w))))" 
       using 2 by (simp add: always_d_def bi_d_def)
  have 4: "\<turnstile> \<diamond> (init  w) = di (\<not> (fin (init (\<not> w))))" 
       by (metis "3" DiEqvNotBiNot DiState Initprop(2) StateEqvBi int_eq)
  have 5: "\<turnstile> \<diamond> (init  w) = di (fin (init w))" using 4 FinNotStateEqvNotFinState 
       by (metis DiEqvNotBiNot DiNotEqvNotBi inteq_reflection)
  from 1 2 3 4 5 show ?thesis by simp
qed

lemma OrDiEqvDi:
 "\<turnstile> (f \<or> di f) = di f"
proof -
  have 1: "\<turnstile> f \<longrightarrow> di f" using DiIntro by blast
 from 1 show ?thesis by auto
qed

lemma AndDiEqv:
 "\<turnstile> (f \<and> di f) = f"
proof -
  have 1: "\<turnstile> f \<longrightarrow> di f" using DiIntro by blast
 from 1 show ?thesis by auto
qed

lemma BiEmptyEqvEmpty:
 "\<turnstile> bi empty = empty"
proof -
 have 1: "\<turnstile> bi empty = (\<not> (di (\<not> empty)))" by (simp add: bi_d_def)
 have 2: "\<turnstile> (\<not> (di (\<not> empty))) = (\<not> ((\<not> empty);#True))" by (simp add: di_d_def)
 have 3: "\<turnstile> (\<not> ((\<not> empty);#True)) = (\<not> (more;#True))" by (simp add: empty_d_def)
 have 4: "\<turnstile> more;#True =  more" using MoreEqvMoreChopTrue by auto
 hence 5: "\<turnstile> (\<not>(more;#True)) =  (\<not> more)" by fastforce
 from 1 2 3 5 show ?thesis using NotEmptyEqvMore by fastforce
qed

lemma EmptyChopSkipInduct:
 assumes "\<turnstile> empty \<longrightarrow> f"
         "\<turnstile> prev f \<longrightarrow> f"
 shows   "\<turnstile> f"
proof -
 have 1: "\<turnstile> empty  \<longrightarrow> f" using assms(1) by auto
 have 2: "\<turnstile> prev f \<longrightarrow> f " using assms(2) by blast
 have 3: "\<turnstile> (empty \<or> prev f) \<longrightarrow> f" using "1" "2"  by fastforce
 have 4: "\<turnstile> wprev f = (empty \<or> prev f)" by (simp add: WprevEqvEmptyOrPrev)
 hence 5: "\<turnstile> wprev f \<longrightarrow> f" using 3 by fastforce 
 hence 6: "\<turnstile> \<not>f \<longrightarrow> \<not> (wprev f) " by fastforce
 hence 7: "\<turnstile> \<not>f \<longrightarrow> prev (\<not> f) " by (simp add: wprev_d_def)
 hence 8: "\<turnstile> \<not> \<not> f " by (rule PrevLoop)
 from 8 show ?thesis by auto
qed

lemma MoreImpImpChopSkipEqv:
 "\<turnstile> more \<longrightarrow> ( (f\<longrightarrow>g);skip = ((f;skip)\<longrightarrow>(g;skip)) )  "
proof -
 have 01: "\<turnstile> (f\<longrightarrow>g) = (\<not> f \<or> g)" by auto
 hence 02: "\<turnstile> (f\<longrightarrow>g);skip = (\<not> f \<or> g);skip " by (simp add: LeftChopEqvChop)
 hence 1: "\<turnstile> (more \<and> (f\<longrightarrow>g);skip) = (more \<and>  (\<not> f \<or> g);skip)" by fastforce
 have 2: "\<turnstile> (\<not> f \<or> g);skip = ((\<not> f);skip \<or> g;skip)" 
      using OrChopEqv by auto
 hence 3: "\<turnstile> (more \<and>  (\<not> f \<or> g);skip) = (more \<and> ((\<not> f);skip \<or> g;skip))" 
      by auto
 have 4: "\<turnstile> (\<not>((\<not> f);skip)) = (empty \<or> (f;skip)) " 
      using NotNotChopSkip by blast
 hence 5: "\<turnstile> ((\<not> f);skip) = (\<not>(empty \<or> (f;skip)))" 
      by fastforce
 have 6: "\<turnstile> \<not>(empty \<or> (f;skip)) = (more \<and> \<not>(f;skip))" 
      using "5" NotChopSkipEqvMoreAndNotChopSkip by fastforce
 have 7: "\<turnstile> ((\<not> f);skip \<or> g;skip) = ( (more \<and> \<not>(f;skip)) \<or> g;skip)" 
      using "5" "6" by fastforce
 hence 8: "\<turnstile> (more \<and>(\<not> f;skip \<or> g;skip)) = (more \<and> ( (more \<and> \<not>(f;skip)) \<or> g;skip)) " 
      by auto
 have 9: "\<turnstile> (more \<and> ( (more \<and> \<not>(f;skip)) \<or> g;skip)) = (more \<and> (\<not>(f;skip) \<or> g;skip))" 
      by auto
 have 10: "\<turnstile> (more \<and> (\<not>(f;skip) \<or> g;skip)) = (more \<and> ((f;skip)\<longrightarrow>(g;skip)))  " 
      by auto
 have 11: "\<turnstile> (more \<and> (f\<longrightarrow>g);skip) = (more \<and> ((f;skip)\<longrightarrow>(g;skip)))" 
      using "1" "2" "3" "8" "9" "10"  "7" by fastforce
 from 11 show ?thesis using MP by fastforce
qed

lemma MoreImpImpPrevEqv:
 "\<turnstile> more \<longrightarrow> ( prev(f\<longrightarrow>g) = (prev f\<longrightarrow> prev g) )"
by (simp add: MoreImpImpChopSkipEqv prev_d_def)

lemma BiBoxNotEqvNotTrueChopChopTrue:
 "\<turnstile> bi(\<box> (\<not> f)) = (\<not>((#True;f);#True )) "
by (simp add: bi_d_def always_d_def di_d_def sometimes_d_def)

(* changed proof *)
lemma DiAndEmptyEqvAndEmpty:
 "\<turnstile> (di f \<and> empty) = (f  \<and> empty)"
proof -
 have 1 : "\<turnstile> di f = (f \<or> di f;skip) " 
      using DiEqvOrDiChopSkipB by blast
 hence 2: "\<turnstile>  (di f \<and> empty) = ((f \<or> di f;skip) \<and> empty)" 
      by fastforce
 have 3 :  "\<turnstile> ((f \<or> di f;skip) \<and> empty) = ((f \<and> empty) \<or> (di f;skip \<and> empty)) " 
      by auto
 have 4: "\<turnstile>  \<not>(di f;skip \<and> empty)"
   by (metis AndDiEqv DiSkipEqvMore NotChopAndMoreAndEmpty inteq_reflection)
 hence 5 : "\<turnstile> ((f \<and> empty) \<or> (di f;skip \<and> empty)) = (f \<and> empty)" 
      by auto
 from 2 3 5 show ?thesis by fastforce
qed

subsubsection \<open>Strict initial intervals\<close>

lemma DsMoreDi:
 "\<turnstile> ds f = (more \<and> (di f);skip)"
proof -
 have 1: "\<turnstile> ds f = (\<not>(bs (\<not> f))) " 
      by (simp add: ds_d_def)
 have 2: "\<turnstile> (\<not>(bs (\<not> f))) = (\<not>(empty \<or> (bi (\<not> f));skip)) " 
      by (simp add: bs_d_def)
 have 3: "\<turnstile> (\<not>(empty \<or> (bi (\<not> f));skip)) = (\<not>empty \<and> \<not>((bi (\<not>f));skip)) " 
      by auto
 have 4: "\<turnstile> (\<not>empty \<and> \<not>((bi (\<not>f));skip)) = (more \<and> \<not>((bi (\<not>f));skip)) " 
     using NotEmptyEqvMore by auto 
 have 5: "\<turnstile> (more \<and> \<not>((bi (\<not>f));skip)) = (more \<and> \<not>(\<not>(di f);skip)) " 
      by (metis DiEqvNotBiNot DiIntro DiSkipEqvMore NotChopSkipEqvMoreAndNotChopSkip 
                Prop10 RightChopImpMoreRule int_simps(4) inteq_reflection lift_and_com)
 have 6: "\<turnstile> (more \<and> \<not>((\<not>(di f));skip)) = (more \<and> (empty \<or> (di f);skip))" 
      using NotNotChopSkip by fastforce
 have 7: "\<turnstile> (more \<and> (empty \<or> (di f);skip)) = (more \<and>   (di f);skip)" 
      using NotEmptyEqvMore by auto
 from 1 2 3 4 5 6 7 show ?thesis by fastforce
qed

lemma DsDi:
 "\<turnstile> ds f = (di f);skip"
proof -
 have 1: "\<turnstile> ds f = (more \<and> (di f);skip)" by (rule DsMoreDi)
 have 2: "\<turnstile> (di f);skip \<longrightarrow> more " by (metis DiIntro DiSkipEqvMore RightChopImpMoreRule int_eq)
 hence 3: "\<turnstile> (more \<and> (di f);skip) = (di f);skip" by auto
 from 1 2 show ?thesis by fastforce
qed

lemma BsEqvNotDsNot:
 "\<turnstile> bs f = (\<not>(ds (\<not> f)))"
proof -
 have 1: "\<turnstile> ds (\<not> f)  = (more \<and> (di (\<not> f));skip)" 
      by (rule DsMoreDi)
 hence 2: "\<turnstile> (\<not>(ds (\<not> f)))  = (\<not>(more \<and> (di (\<not> f));skip))" 
      by auto
 have 3: "\<turnstile> (\<not>(more \<and> (di (\<not> f));skip)) = (empty \<or> \<not>((di (\<not> f));skip))" 
      using NotEmptyEqvMore by auto
 have 4: "\<turnstile> (empty \<or> \<not>((di (\<not> f));skip)) = (empty \<or> \<not>((\<not>(bi f));skip))" 
      using DiNotEqvNotBi by (metis "3" inteq_reflection) 
 have 5: "\<turnstile>  (\<not>((\<not>(bi f));skip)) = (empty \<or> (bi f);skip) " 
      by (rule NotNotChopSkip)
 hence 6: "\<turnstile> (empty \<or> \<not>((\<not>(bi f));skip)) = (empty \<or> (bi f);skip) " 
      by auto
 from 2 3 4 6 show ?thesis by (metis bs_d_def inteq_reflection)
qed

lemma NotBsEqvDsNot:
 "\<turnstile> (\<not>(bs f)) = ds (\<not> f)"
proof -
 have 1: "\<turnstile> bs f = (\<not>(ds (\<not> f)))" by (rule BsEqvNotDsNot)
 hence 2: "\<turnstile> (\<not>( bs f)) = (\<not>\<not>(ds (\<not> f))) " by auto
 from 2 show ?thesis by auto
qed

lemma NotDsEqvBsNot:
 "\<turnstile> (\<not> (ds f)) = bs (\<not> f)"
proof -
 have 1: "\<turnstile> (\<not>(ds  f)) = (\<not>\<not>(bs (\<not>f)))" by (simp add: ds_d_def)
 from 1 show ?thesis by auto
qed

lemma NotDsAndEmpty:
 "\<turnstile> \<not>(ds f \<and> empty)"
proof -
 have 1: "\<turnstile> ds f = (more \<and> (di f);skip)" by (rule DsMoreDi)
 have 2: "\<turnstile> more \<and> (di f);skip \<and> empty \<longrightarrow> #False" using NotEmptyEqvMore by auto
 from 1 2 show ?thesis by fastforce
qed

lemma BsMoreEqvEmpty:
 "\<turnstile> bs more = empty"
proof -
 have 1: "\<turnstile> bs more = (empty \<or> (bi more);skip)" by (simp add: bs_d_def)
 have 2: "\<turnstile> bi more \<longrightarrow> #False" using DiEmpty NotEmptyEqvMore by (simp add: bi_d_def empty_d_def)
 hence 3: "\<turnstile> (bi more);skip \<longrightarrow> #False;skip " using LeftChopImpChop by blast
 have 31: "\<turnstile> #False;skip \<longrightarrow> #False " by (simp add: Valid_def skip_defs chop_defs)
 have 32: "\<turnstile> (bi more);skip \<longrightarrow> #False " using "3" "31" by fastforce 
 hence 4: "\<turnstile> (empty \<or> ((bi more);skip)) = empty"  by fastforce
 from 1 4 show ?thesis by fastforce
qed

lemma  BsAndEqv:
 "\<turnstile> (bs f \<and> bs g) = bs(f \<and> g)"
proof -
 have 1: "\<turnstile> bs f = (empty \<or> (bi f) ;skip)" 
      by (simp add: bs_d_def)
 have 2: "\<turnstile> bs g = (empty \<or> (bi g) ;skip)" 
      by (simp add: bs_d_def)
 have 3: "\<turnstile> (bs f \<and> bs g) = ((empty \<or> (bi f) ;skip) \<and> (empty \<or> (bi g) ;skip)) " 
      using "1" "2" by fastforce
 have 4: "\<turnstile> ((empty \<or> (bi f) ;skip) \<and> (empty \<or> (bi g) ;skip)) = 
             (empty \<or> ((bi f) ;skip \<and> (bi g) ;skip)) " 
      by auto
 have 5: "\<turnstile> (((bi f) ;skip \<and> (bi g) ;skip)) = bi(f \<and> g);skip " 
      using BiAndChopSkipEqv  by fastforce
 hence 6: "\<turnstile> (empty \<or> ((bi f) ;skip \<and> (bi g) ;skip)) = (empty \<or> bi(f \<and> g);skip) " 
      by auto
 from 3 4 6 show ?thesis by (metis bs_d_def inteq_reflection)
qed

lemma DsEqvRule:
 assumes "\<turnstile> f = g"
 shows "\<turnstile> ds f = ds g"
using assms using int_eq by force

lemma DsOrEqv:
 "\<turnstile> (ds f \<or> ds g) = ds (f \<or> g)"
proof -
 have 1: "\<turnstile> ds f = (\<not>(bs (\<not> f)))" by (simp add: ds_d_def)
 have 2: "\<turnstile> ds g = (\<not>(bs (\<not> g)))" by (simp add: ds_d_def)
 have 3: "\<turnstile> (ds f \<or> ds g) = (\<not>(bs (\<not> f)) \<or> \<not>(bs (\<not> g)))" using "1" "2" by fastforce
 have 4: "\<turnstile> (\<not>(bs (\<not> f)) \<or> \<not>(bs (\<not> g))) = (\<not>(bs (\<not> f) \<and> bs (\<not> g)))" by auto
 have 5: "\<turnstile> (bs (\<not> f) \<and> bs (\<not> g)) = bs( \<not> f \<and> \<not>g)" by (rule BsAndEqv)
 hence 6: "\<turnstile> (\<not>(bs (\<not> f) \<and> bs (\<not> g))) = (\<not>(bs( \<not> f \<and> \<not>g))) " by auto
 have 7: "\<turnstile> (\<not>(bs( \<not> f \<and> \<not>g))) = ds (\<not>(\<not> f \<and> \<not>g)) " by (rule NotBsEqvDsNot)
 have 8: "\<turnstile> (\<not>(\<not> f \<and> \<not>g)) = (f \<or> g)" by auto
 hence 9: "\<turnstile> ds(\<not>(\<not> f \<and> \<not>g)) = ds (f \<or> g) " by (rule DsEqvRule)
 from 3 4 6 7 9 show ?thesis by fastforce
qed

lemma BsOrImp:
 "\<turnstile> bs f \<or> bs g \<longrightarrow> bs(f \<or> g)"
proof -
 have 1: "\<turnstile> bi f \<or> bi g \<longrightarrow> bi(f \<or> g)" 
      by (rule BiOrBiImpBiOr)
 hence 2: "\<turnstile> (bi f \<or> bi g);skip \<longrightarrow> (bi(f \<or> g));skip" 
      by (rule LeftChopImpChop)
 have 3: "\<turnstile> (bi f);skip \<or> (bi g);skip \<longrightarrow> (bi(f \<or> g));skip" 
      using "1" OrChopEqv "2" by fastforce
 hence 4: "\<turnstile> empty \<or> (bi f);skip \<or> (bi g);skip \<longrightarrow>  empty \<or> (bi(f \<or> g));skip" 
      by auto
 hence 5: "\<turnstile> (empty \<or> (bi f);skip) \<or> (empty \<or> (bi g);skip) \<longrightarrow>  empty \<or> (bi(f \<or> g));skip " 
      by auto
 from 5 show ?thesis by (simp add: bs_d_def)
qed

lemma DsAndImp:
 "\<turnstile> ds (f \<and> g) \<longrightarrow> ds f \<and> ds g "
proof -
 have 1: "\<turnstile> bs (\<not>f) \<or> bs (\<not>g) \<longrightarrow> bs(\<not>f \<or> \<not>g)" by (rule BsOrImp)
 have 2: "\<turnstile> (\<not>f \<or> \<not>g) = (\<not>(f \<and> g))" by auto
 hence 3: "\<turnstile> bs(\<not>f \<or> \<not>g) = bs (\<not>(f \<and> g))" by (rule BsEqvRule)
 have 4: "\<turnstile> bs (\<not>f) \<or> bs (\<not>g) \<longrightarrow> bs (\<not>(f \<and> g))" using "1" "3" by fastforce
 have 5: "\<turnstile> bs (\<not>f) = (\<not>(ds f))" using NotDsEqvBsNot by fastforce
 have 6: "\<turnstile> bs (\<not>g) = (\<not>(ds g))" using NotDsEqvBsNot by fastforce
 have 7: "\<turnstile> bs (\<not>(f \<and> g)) =  (\<not>(ds (f \<and> g)))" using NotDsEqvBsNot by fastforce
 have 8: "\<turnstile>  \<not>(ds f) \<or>  \<not>(ds g) \<longrightarrow>  \<not>(ds (f \<and> g))" using "4" "5" "6" "7" by fastforce
 hence 9: "\<turnstile> \<not>(ds f \<and> ds g)  \<longrightarrow>  \<not>(ds (f \<and> g))" by auto
 from 9 show ?thesis by auto
qed

lemma DsAndImpElimL:
 "\<turnstile> ds (f \<and> g) \<longrightarrow> ds f "
using DsAndImp by fastforce

lemma DsAndImpElimR:
 "\<turnstile> ds (f \<and> g) \<longrightarrow> ds g "
using DsAndImp by fastforce

lemma BiImpBs:
 "\<turnstile> bi f \<longrightarrow> bs f  "
proof -
 have 1: "\<turnstile> empty \<longrightarrow> empty \<or> (bi f);skip" by auto
 hence 2: "\<turnstile> empty \<and> bi f \<longrightarrow> empty \<or> (bi f);skip" by auto
 have 2: "\<turnstile> more \<and> bi f \<longrightarrow> (bi f);skip" by (rule MoreAndBiImpBiChopSkip)
 hence 3: "\<turnstile> more \<and> bi f \<longrightarrow> empty \<or> (bi f);skip " by auto
 have 4: "\<turnstile> bi f = ((bi f \<and> empty) \<or> (bi f \<and> more))" by  (auto simp add: empty_d_def) 
 have 5: "\<turnstile> (empty \<or> (bi f);skip) = bs f " by (simp add: bs_d_def)
 from 2 3 4 5 show ?thesis by fastforce
qed

lemma BsImpBsBs:
 "\<turnstile> bs f \<longrightarrow> bs ( bs f)"
proof -
 have 1: "\<turnstile> bi f \<longrightarrow> bs f" by (rule BiImpBs)
 hence 2: "\<turnstile> bi (bi f) \<longrightarrow> bi(bs f)" by (rule BiImpBiRule)
 hence 3: "\<turnstile> (bi f) \<longrightarrow> bi(bs f)" using BiEqvBiBi by fastforce
 hence 4: "\<turnstile> (bi f);skip \<longrightarrow> (bi(bs f));skip" by (rule LeftChopImpChop)
 hence 5: "\<turnstile> empty \<or> (bi f);skip \<longrightarrow> empty \<or> (bi(bs f));skip" by auto
 from 5 show ?thesis by (simp add: bs_d_def)
qed

lemma DsImpDi:
 "\<turnstile> ds f \<longrightarrow> di f"
proof -
 have 1: "\<turnstile> bi (\<not> f) \<longrightarrow> bs (\<not> f)" by (rule BiImpBs)
 hence 2: "\<turnstile> \<not>(bs (\<not> f)) \<longrightarrow> \<not>(bi (\<not> f)) " by fastforce
 from 2 show ?thesis using NotBsEqvDsNot DiEqvNotBiNot by fastforce
qed

lemma BsImpBsRule:
 assumes "\<turnstile> f \<longrightarrow> g"
 shows "\<turnstile> bs f  \<longrightarrow> bs g"
proof -
 have 1: "\<turnstile> f \<longrightarrow> g" using assms by auto
 hence 2: "\<turnstile> bi f \<longrightarrow> bi g " by (rule BiImpBiRule)
 hence 3: "\<turnstile> (bi f);skip \<longrightarrow> (bi g);skip " by (rule LeftChopImpChop)
 hence 4: "\<turnstile> empty \<or> (bi f);skip \<longrightarrow> empty \<or> (bi g);skip" by auto
 from 4 show ?thesis by (simp add: bs_d_def)
qed
 
lemma DsChopImpDsB:
 "\<turnstile> ds (f;g) \<longrightarrow> ds f"
proof -
 have 1: "\<turnstile> di(f;g) \<longrightarrow> di f" by (rule DiChopImpDiB)
 hence 2: "\<turnstile> (di(f;g));skip \<longrightarrow> (di f);skip " by (rule LeftChopImpChop)
 from 2 show ?thesis using DsDi by fastforce
qed

lemma NotBsImpBsNotChop:
 "\<turnstile> bs (\<not> f) \<longrightarrow> bs ( \<not>(f;g))"
proof -
 have 1: "\<turnstile> ds (f;g) \<longrightarrow> ds f" by (rule DsChopImpDsB)
 hence 2: "\<turnstile> \<not>(ds f) \<longrightarrow> \<not>(ds (f;g))" by fastforce
 from 2 show ?thesis using NotDsEqvBsNot by fastforce
qed

(* proof has changed notice the crucial lemma BiBiOrEqvBi that is needed!*)
lemma BsOrBsEqvBsBiOrBi:
 "\<turnstile> (bs f \<or> bs g) = bs(bi f \<or> bi g) "
proof -
 have 1: "\<turnstile> (bs f \<or> bs g) = ((empty \<or> (bi f);skip) \<or> (empty \<or> (bi g);skip))" 
      by (simp add: bs_d_def)
 have 2: "\<turnstile> ((empty \<or> (bi f);skip) \<or> (empty \<or> (bi g);skip)) = (empty \<or> (bi f);skip \<or> (bi g);skip)" 
      by auto
 have 3: "\<turnstile> ((bi f);skip \<or> (bi g);skip) = (bi f \<or> bi g);skip" 
      using OrChopEqv by fastforce
 hence 4: "\<turnstile> (empty \<or> (bi f);skip \<or> (bi g);skip) = (empty \<or> (bi f \<or> bi g);skip)" 
      by auto
 have 5: "\<turnstile> (bi f \<or> bi g) = bi ( bi f \<or> bi g)  " 
      by (meson BiBiOrImpBi BiImpBiBiOr BiImpBiBiOrB Prop02 int_iffI)
 hence 6: "\<turnstile> (bi f \<or> bi g);skip = bi ( bi f \<or> bi g);skip  " 
      by (simp add: LeftChopEqvChop)
 hence 7: "\<turnstile> (empty \<or> bi ( bi f \<or> bi g);skip) = (empty \<or> (bi f \<or> bi g);skip)" 
      by auto
 have 8: "\<turnstile> (empty \<or> (bi f \<or> bi g);skip) = bs(bi f \<or> bi g)" using bs_d_def 
      by (metis "4" "5" inteq_reflection)
 from 1 2 4 8 show ?thesis by (metis inteq_reflection) 
qed

lemma DiOrDsEqvDi:
 "\<turnstile> di f \<or> ds f = di f"
proof -
 have 1: "\<turnstile> di f \<longrightarrow> di f \<or> ds f" by auto
 have 2: "\<turnstile> di f \<longrightarrow> di f" by auto
 have 3: "\<turnstile> ds f \<longrightarrow> di f" by (rule DsImpDi)
 have 4: "\<turnstile> di f \<or> ds f \<longrightarrow> di f" using "2" "3" by auto
 from 1 4 show ?thesis by auto
qed

lemma DiAndDsEqvDs:
 "\<turnstile> (di f \<and> ds f) = ds f "
proof -
 have 1: "\<turnstile> di f \<and> ds f \<longrightarrow> ds f" by auto
 have 2: "\<turnstile> ds f \<longrightarrow> ds f " by auto
 have 3: "\<turnstile> ds f \<longrightarrow> di f" by (rule DsImpDi)
 have 4: "\<turnstile> ds f \<longrightarrow> di f \<and> ds f" using "2" "3" by auto
 from 1 4 show ?thesis by auto
qed

lemma OrDsEqvDi:
 "\<turnstile> (f \<or> ds f) = di f"
proof -
 have 1: "\<turnstile> ds f =  (di f);skip" by (rule DsDi)
 hence 2: "\<turnstile> (f \<or> ds f) = (f \<or> (di f);skip)" by auto
 from 2 show ?thesis using DiEqvOrDiChopSkipB[of "f"]  by fastforce
qed

lemma AndBsEqvBi:
 "\<turnstile> (f \<and> bs f) = bi f "
proof -
 have 1: "\<turnstile> (f \<and> bs f) = (f \<and> (empty \<or> (bi f);skip)) " by (simp add: bs_d_def)
 from 1 show ?thesis
 by (meson BiEqvAndEmptyOrBiChopSkip Prop04 int_simps(4)) 
qed

lemma BsEqvBsBi:
 "\<turnstile> bs f = bs (bi f) "
proof -
 have 1: "\<turnstile> bs f = (empty \<or> (bi f);skip)" by (simp add: bs_d_def)
 have 2: "\<turnstile> bi f = bi ( bi f) " by (rule BiEqvBiBi)
 hence 3: "\<turnstile> (bi f);skip = bi (bi f);skip " using LeftChopEqvChop by blast
 hence 4: "\<turnstile>  (empty \<or> (bi f);skip) =  (empty \<or> bi (bi f);skip) " by auto
 from 1 4 show ?thesis by (simp add: bs_d_def)
qed

lemma StateImpBs:
 "\<turnstile> init w \<longrightarrow> bs (init w)"
proof -
 have 1: "\<turnstile> init w = bi (init w) "  by (rule StateEqvBi)
 have 2: "\<turnstile> bi (init w) \<longrightarrow> bs (init w) " by (rule BiImpBs)
 from 1 2 show ?thesis using StateImpBi by fastforce
qed

lemma DsAndDsEqvDsAndDiOrDsAndDi:
 "\<turnstile> (ds f \<and> ds g) =( ds (f \<and> di g) \<or> ds(g \<and> di f))"
proof -
 have 1: "\<turnstile> (di f \<and> di g) = (di(f \<and> di g) \<or> di(g \<and> di f))" 
      by (rule DiAndDiEqvDiAndDiOrDiAndDi)
 hence 2: "\<turnstile> (di f \<and> di g);skip = (di(f \<and> di g) \<or> di(g \<and> di f));skip" 
      by (rule LeftChopEqvChop)
 have 3: "\<turnstile> (di f \<and> di g);skip = ((di f);skip \<and> (di g);skip) " 
      using ChopSkipAndChopSkip by fastforce
 have 4: "\<turnstile> ((di f);skip \<and> (di g);skip) = (di(f \<and> di g) \<or> di(g \<and> di f));skip " 
      using "2" "3" by fastforce
 have 5: "\<turnstile> (di(f \<and> di g) \<or> di(g \<and> di f));skip = (di(f \<and> di g);skip \<or> di(g \<and> di f);skip)" 
      using OrChopEqv by blast
 have 6: "\<turnstile> ds f = (di f);skip  " 
      using DsDi by blast
 have 7: "\<turnstile> ds g = (di g);skip " 
      using DsDi by blast
 have 8: "\<turnstile>  ((di f);skip \<and> (di g);skip) = (ds f \<and>  ds g)" 
      using "6" "7" by fastforce
 have 9: "\<turnstile> ds(f \<and> di g) =  di(f \<and> di g);skip" 
      using DsDi by blast
 have 10: "\<turnstile> ds(g \<and> di f) = di(g \<and> di f);skip  " 
      using DsDi by blast
 have 11: "\<turnstile> (di(f \<and> di g);skip \<or> di(g \<and> di f);skip) = (ds(f \<and> di g) \<or> ds(g \<and> di f))" 
      using "9" "10" by fastforce
 from 4 5 8 11 show ?thesis by (meson Prop04) 
qed

lemma BsEqvBiMoreImpChop:
 "\<turnstile> bs f = bi(more \<longrightarrow> f;skip) "
proof -
 have 1: "\<turnstile> bs f = (empty \<or> (bi f;skip))" 
      by (simp add: bs_d_def)
 have 2: "\<turnstile> (empty \<or> (bi f;skip)) = ((\<not>(\<not>(bi f));skip))  " 
      using NotNotChopSkip by fastforce
 have 3: "\<turnstile> \<not>((\<not>(bi f));skip) = (\<not>(di (\<not> f);skip))" 
      by (simp add: bi_d_def)
 have 4: "\<turnstile> (\<not>(di (\<not> f);skip)) = (\<not>(((\<not> f) ;#True);skip)) " 
      by (simp add:di_d_def)
 have 5: "\<turnstile> (\<not>(((\<not> f) ;#True);skip)) = (\<not>((\<not> f) ;(#True;skip)))" 
      using ChopAssocB by fastforce
 have 6: "\<turnstile> (\<not>((\<not> f) ;(#True;skip))) = (\<not>((\<not> f) ;(skip;#True)))"
      by (metis "5" SkipTrueEqvTrueSkip int_eq) 
 have 7: "\<turnstile> (\<not>((\<not> f) ;(skip;#True))) = (\<not>(((\<not> f) ;skip);#True)) " 
      using ChopAssoc by fastforce
 have 8: "\<turnstile> (\<not>(((\<not> f) ;skip);#True)) = (\<not>(di ((\<not> f);skip)))  "  
      by (simp add: di_d_def)
 have 9: "\<turnstile> (\<not>(di ((\<not> f);skip))) = bi (\<not>((\<not> f) ;skip))  " 
      using NotDiEqvBiNot by blast
 have 10: "\<turnstile> bi (\<not>((\<not> f) ;skip)) = bi( empty \<or> (f;skip))" 
      using NotNotChopSkip using BiEqvBi by blast
 have 11: "\<turnstile> bi( empty \<or> (f;skip)) = bi( \<not> more \<or> (f;skip))" 
      by (simp add: empty_d_def)
 have 12: "\<turnstile> ( \<not> more \<or> (f;skip)) = (more \<longrightarrow> f;skip) " by auto
 have 13: "\<turnstile> bi( \<not> more \<or> (f;skip)) = bi(more \<longrightarrow> f;skip) " using "12" using BiEqvBi by blast
 have 14: "\<turnstile> bs f = (\<not> (((\<not> f);skip);#True)) " using "1" "2" "3" "4" "5" "6" "7" by fastforce
 have 15: "\<turnstile> (\<not> (((\<not> f);skip);#True)) = bi(more \<longrightarrow> f;skip) " using "8" "9" "10" "11" "13" by fastforce
 from 14 15 show ?thesis by fastforce
qed

lemma BoxMoreStateEqvBsFinState:
 "\<turnstile> \<box>(more \<longrightarrow> \<not> (init w)) = bs(\<not>(fin(init w))) "
proof -
 have 1: "\<turnstile> \<box>(more \<longrightarrow> \<not> (init w)) = (\<not>(\<diamond>(\<not>(more \<longrightarrow> \<not> (init w))))) " 
      by (simp add: always_d_def)
 have 01: "\<turnstile> (\<not>(more \<longrightarrow> \<not> (init w))) = (init w \<and>  more)" by auto  
 hence 2: "\<turnstile> \<not>(\<diamond>(\<not>(more \<longrightarrow> \<not> (init w)))) = (\<not>(#True;(init w \<and>  more)))" 
      by (metis int_eq int_iffD1 int_simps(14) int_simps(6) sometimes_d_def)
 have 3: "\<turnstile> more = #True; skip" 
      using MoreEqvSkipChopTrue SkipTrueEqvTrueSkip by fastforce
 have 4: "\<turnstile> (init w \<and>  more) = (init w \<and> (#True; skip)) " 
      using "3" by auto
 have 5: "\<turnstile> (init w \<and> (#True; skip))   = ((init w \<and> empty);(#True;skip))" 
      using StateAndEmptyChop by fastforce
 have 6: "\<turnstile> (init w \<and>  more) = ((init w \<and> empty);(#True;skip)) " 
      using "4" "5" by fastforce
 have 7: "\<turnstile>  (#True;(init w \<and>  more)) = (#True;((init w \<and> empty);(#True;skip)) )" 
      using "6" RightChopEqvChop by blast
 have 8: "\<turnstile> (#True;((init w \<and> empty);(#True;skip)) ) =
            (((#True;(init w \<and> empty));(#True;skip)) ) " 
      using ChopAssoc by blast
 have 9: "\<turnstile> (((#True;(init w \<and> empty));(#True;skip)) ) = 
            ((((#True;(init w \<and> empty));#True);skip) )" 
      using ChopAssoc by blast
 have 10: "\<turnstile> (#True;(init w \<and>  more)) = 
             ((((#True;(init w \<and> empty));#True);skip) )" 
      using "7" "8" "9" by fastforce
 hence 11: "\<turnstile> (\<not>(#True;(init w \<and>  more))) = 
              (\<not>((((#True;(init w \<and> empty));#True);skip) ))" 
      by auto
 have 12: "\<turnstile> \<not>((((#True;(init w \<and> empty));#True);skip) ) =
             empty \<or> (\<not>((#True;(init w \<and> empty));#True);skip)" 
      using NotChopNotSkip by fastforce
 have 13: "\<turnstile> (\<not>((#True;(init w \<and> empty));#True)) = bi(\<box> (\<not>(init w \<and> empty)))" 
      using BiBoxNotEqvNotTrueChopChopTrue  by fastforce
 hence 14: "\<turnstile> (\<not>((#True;(init w \<and> empty));#True));skip = 
              (bi(\<box> (\<not>(init w \<and> empty))));skip" 
      using RightChopEqvChop by (simp add: LeftChopEqvChop)
 hence 15: "\<turnstile> empty \<or> (\<not>((#True;(init w \<and> empty));#True));skip = 
              empty \<or>(bi(\<box> (\<not>(init w \<and> empty))));skip" 
      by auto
 have 16: "\<turnstile> (\<not>((((#True;(init w \<and> empty));#True);skip) )) =
              (empty \<or> (bi(\<box> (\<not>(init w \<and> empty)));skip)) " 
      using "12" "15" using "14" NotChopNotSkip int_eq by fastforce
 have 171: "\<turnstile> (\<not>(init w \<and> empty)) = (\<not>(init w) \<or> \<not>empty)" 
      by auto
 hence 172: "\<turnstile> \<box>(\<not>(init w \<and> empty)) = \<box>(\<not>(init w) \<or> \<not>empty) " 
      by (simp add: BoxEqvBox)
 hence 173: "\<turnstile> bi(\<box>(\<not>(init w \<and> empty))) = bi(\<box>(\<not>(init w) \<or> \<not>empty)) " 
      by (simp add: BiEqvBi)
 hence 174: "\<turnstile> bi(\<box>(\<not>(init w \<and> empty)));skip = bi(\<box>(\<not>(init w) \<or> \<not>empty));skip "
      using LeftChopEqvChop by blast
 hence 17: "\<turnstile> (empty \<or> (bi(\<box> (\<not>(init w \<and> empty)));skip)) =
             (empty \<or> (bi(\<box> (\<not>(init w) \<or> \<not>empty));skip))" 
      by auto
 have 181: "\<turnstile>  (\<not>(init w) \<or> \<not>empty) =  (\<not>empty \<or> \<not>(init w)  )" 
      by auto
 hence 18: "\<turnstile> \<box> (\<not>(init w) \<or> \<not>empty) = \<box> (\<not>empty \<or> \<not>(init w)  )" 
      by (simp add: BoxEqvBox)
 have 191: "\<turnstile>  (\<not>empty \<or> \<not>(init w)  ) = (empty \<longrightarrow> \<not>(init w) )" 
      by auto
 hence 19: "\<turnstile> \<box> (\<not>empty \<or> \<not>(init w)  ) = \<box>(empty \<longrightarrow> \<not>(init w) )" 
      by (simp add: BoxEqvBox)
 have 20: "\<turnstile> \<box>(empty \<longrightarrow> \<not>(init w) ) = fin (\<not>(init w) )" 
      by (simp add: fin_d_def)
 have 21: "\<turnstile> fin (\<not>(init w) ) = (\<not>(fin (init w))) " 
      using FinEqvFin FinNotStateEqvNotFinState Initprop(2) by fastforce
 have 22: "\<turnstile> bi(\<box> (\<not>(init w) \<or> \<not>empty)) = bi (\<not>(fin (init w)))" 
      using "18" "19" "20" "21" BiEqvBi by (metis int_eq)
 hence 23: "\<turnstile> (bi(\<box> (\<not>(init w) \<or> \<not>empty)));skip = (bi (\<not>(fin (init w))));skip " 
      using RightChopEqvChop by (simp add: LeftChopEqvChop)
 hence 24: "\<turnstile> (empty \<or> (bi(\<box> (\<not>(init w) \<or> \<not>empty)));skip) = 
              (empty \<or> (bi (\<not>(fin (init w))));skip)" 
      by auto
 hence 25: "\<turnstile> (empty \<or> (bi (\<not>(fin (init w))));skip) = bs(\<not>(fin (init w))) " 
      by (simp add:bs_d_def)
  show ?thesis
  by (metis "01" "1" "11" "16" "172" "18" "19" "20" "21" "25" inteq_reflection sometimes_d_def) 
qed

lemma BsFalseEqvEmpty:
 "\<turnstile> bs #False = empty"
proof -
 have 1: "\<turnstile> bs #False = (empty \<or> bi #False;skip) " 
      by (simp add: bs_d_def)
 have 2: "\<turnstile> \<not>(bi #False;skip)"  
      by (metis BiEqvAndWprevBi MoreEqvSkipChopTrue NotChopSkipEqvMoreAndNotChopSkip 
          SkipTrueEqvTrueSkip int_eq int_iffD1 int_simps(14) int_simps(19) int_simps(2) 
          int_simps(21))
 from 1 2 show ?thesis by fastforce
qed

subsubsection \<open>First occurrence\<close>

lemma FstWithAndImp:
 "\<turnstile> \<rhd>f \<and> g \<longrightarrow> \<rhd> ( f \<and> g)"
proof -
 have 1: "\<turnstile> (\<rhd>f \<and> g) = ((f \<and> (bs (\<not>f))) \<and> g) " 
      by (simp add: first_d_def) 
 have 2: "\<turnstile> ((f \<and> (bs (\<not>f))) \<and> g) = (f \<and> \<not>(ds f) \<and> g)" 
      using NotDsEqvBsNot by fastforce
 have 3: "\<turnstile> \<not>(ds f) \<longrightarrow> \<not>(ds(f \<and> g))" 
      using DsAndImpElimL by fastforce
 hence 4: "\<turnstile> f \<and> \<not>(ds f) \<and> g \<longrightarrow> f \<and> g \<and> \<not>(ds(f \<and> g)) " 
      by auto
 have  5: "\<turnstile> (f \<and> g \<and> \<not>(ds(f \<and> g))) = ((f \<and> g) \<and> (bs (\<not>(f \<and> g))))  " 
      using NotDsEqvBsNot by fastforce
 have  6: "\<turnstile> ((f \<and> g) \<and> (bs (\<not>(f \<and> g)))) = \<rhd>(f \<and> g) " 
      by (simp add: first_d_def)
 from 1 2 4 5 6 show ?thesis by fastforce
qed
 
lemma FstWithOrEqv:
 "\<turnstile> \<rhd>(f \<or> g) = ((\<rhd>f \<and> bs (\<not>g)) \<or> (\<rhd>g \<and> bs (\<not>f)))" 
proof -
 have 1: "\<turnstile> \<rhd>(f \<or> g) = ((f \<or> g) \<and> bs (\<not>(f \<or> g)))" 
      by (simp add: first_d_def)
 have 2: "\<turnstile>  (\<not>(f \<or> g)) =  (\<not>f \<and> \<not>g)" 
      by auto
 hence 3: "\<turnstile> bs (\<not>(f \<or> g)) =  bs (\<not>f \<and> \<not>g)" 
      using BsEqvRule by blast
 have  4: "\<turnstile>  bs (\<not>f \<and> \<not>g) = (bs (\<not> f) \<and> bs (\<not> g))" 
      using BsAndEqv  by fastforce
 have  5: "\<turnstile> ((f \<or> g) \<and> bs (\<not>(f \<or> g))) = ((f \<or> g) \<and> bs (\<not> f) \<and> bs (\<not> g))" 
      using "3" "4" by fastforce
 have  6: "\<turnstile> ((f \<or> g) \<and> bs (\<not> f) \<and> bs (\<not> g)) = 
             (((f \<and> bs (\<not> f)) \<and> bs (\<not> g)) \<or> (g \<and> bs (\<not> f) \<and> bs (\<not> g))) " 
      by auto
 have  7: "\<turnstile> ((f \<and> bs (\<not> f)) \<and> bs (\<not> g)) = (\<rhd>f \<and>  bs (\<not> g))  " 
      by (simp add: first_d_def)
 have  8: "\<turnstile> (g \<and> bs (\<not> f) \<and> bs (\<not> g)) = ((g \<and> bs (\<not> g)) \<and> bs (\<not> f))" 
      by auto
 have  9: "\<turnstile> ((g \<and> bs (\<not> g)) \<and> bs (\<not> f)) = (\<rhd>g \<and>  bs (\<not> f))" 
      by (simp add: first_d_def)
 have 10: "\<turnstile> ((f \<and> bs (\<not> f)) \<and> bs (\<not> g)) \<or> (g \<and> bs (\<not> f) \<and> bs (\<not> g)) = 
            (\<rhd>f \<and>  bs (\<not> g)) \<or> (\<rhd>g \<and>  bs (\<not> f)) " 
      using "7" "8" "9" by fastforce
 from 1 5 6 10 show ?thesis by (metis "7" "8" "9" int_eq)
qed

lemma FstFstAndEqvFstAnd:
 "\<turnstile> \<rhd>(\<rhd>f \<and> g) = (\<rhd>f \<and> g)"
proof -
  have 1: "\<turnstile> (\<rhd>f \<and> g) = ((f \<and> (bs (\<not>f))) \<and> g)" by (simp add: first_d_def)
  hence 2: "\<turnstile> \<rhd>f \<and> g \<longrightarrow> (bs (\<not>f))" by auto
  hence 3: "\<turnstile> \<rhd>f \<and> g \<longrightarrow> \<rhd>f \<and> g \<and> (bs (\<not>f)) " by auto
  have 4: "\<turnstile> \<not> f \<longrightarrow> \<not> f \<or> \<not>(bs (\<not>f)) \<or> \<not> g" by auto
  hence 5: "\<turnstile> bs (\<not> f) \<longrightarrow> bs(\<not> f \<or> \<not>(bs (\<not>f)) \<or> \<not> g)" using BsImpBsRule by blast
  have 6: "\<turnstile> (\<not> f \<or> \<not>(bs (\<not>f)) \<or> \<not> g) = (\<not>(f \<and> bs (\<not> f) \<and> g))" by auto
  hence 7: "\<turnstile> bs(\<not> f \<or> \<not>(bs (\<not>f)) \<or> \<not> g) = bs(\<not>(f \<and> bs (\<not> f) \<and> g))" using BsEqvRule by blast
  have 8: "\<turnstile> ((f \<and> bs (\<not> f)) \<and> g) = (\<rhd>f \<and> g)  "  by (simp add: first_d_def)
  hence 9: "\<turnstile> (\<not>(f \<and> bs (\<not> f) \<and> g)) = (\<not>(\<rhd>f \<and> g))" by auto
  hence 10: "\<turnstile> bs (\<not>(f \<and> bs (\<not> f) \<and> g)) = bs (\<not>(\<rhd>f \<and> g))" using BsEqvRule by blast
  have 11: "\<turnstile> \<rhd>f \<and> g \<longrightarrow> (\<rhd>f \<and> g) \<and>   bs (\<not>(\<rhd>f \<and> g))" using "3" "5" "7" "10" by fastforce
  hence 12: "\<turnstile> \<rhd>f \<and> g \<longrightarrow> \<rhd>(\<rhd>f \<and> g)" by (simp add: first_d_def)
  have 13: "\<turnstile> \<rhd>(\<rhd>f \<and> g) = ((\<rhd>f \<and> g) \<and>   bs (\<not>(\<rhd>f \<and> g))) " by (simp add: first_d_def)
  hence 14: "\<turnstile> \<rhd>(\<rhd>f \<and> g) \<longrightarrow> \<rhd>f \<and> g" by auto
 from 12 14 show ?thesis by fastforce
qed

lemma FstTrue:
 "\<turnstile> \<rhd> #True = empty"
proof -
 have 1: "\<turnstile> \<rhd> #True = (#True \<and> bs (\<not> #True))" 
      by (simp add: first_d_def)
 have 2: "\<turnstile> bs (\<not> #True) = (empty \<or> (bi (\<not> #True));skip) " 
      by (simp add: bs_d_def)
 have 3: "\<turnstile> \<not>(bi (\<not> #True)) " 
      using BiElim by fastforce
 have 4: "\<turnstile> \<not>((bi (\<not> #True));skip)" 
      by (metis AndChopA BiEqvAndEmptyOrBiChopSkip MoreEqvSkipChopTrue 
          NotChopSkipEqvMoreAndNotChopSkip SkipTrueEqvTrueSkip int_eq int_simps(14) int_simps(21))
 have 5: "\<turnstile> bs (\<not> #True) = empty " 
      using "2" "4" by fastforce
 from 1 5 show ?thesis by fastforce
qed

lemma FstFalse:
  "\<turnstile> \<not>(\<rhd> #False) "
proof -
 have 1: "\<turnstile> \<rhd> #False = (#False \<and> bs #True)" by (simp add: first_d_def)
 from 1 show ?thesis by auto
qed

lemma FstChopFalseEqvFalse:
 "\<turnstile> \<not>(\<rhd>f ; #False)"
by (simp add:Valid_def chop_defs)

lemma FstEmpty:
 "\<turnstile> \<rhd> empty = empty"
proof -
 have 1: "\<turnstile> \<rhd> empty = (empty \<and> bs (\<not> empty))" by (simp add: first_d_def)
 have 2: "\<turnstile> bs (\<not> empty) = (empty \<or> bi (\<not> empty);skip)" by (simp add: bs_d_def)
 from 1 2 show ?thesis by fastforce
qed

lemma FstAndEmptyEqvAndEmpty:
 "\<turnstile> (\<rhd>f \<and> empty) = (f \<and> empty)"
proof -
 have 1: "\<turnstile> (\<rhd>f \<and> empty) = ((f \<and> bs (\<not> f)) \<and> empty)" by (simp add: first_d_def)
 have 2: "\<turnstile> bs (\<not> f) = (empty \<or> bi (\<not>f);skip)" by (simp add: bs_d_def)
 from 1 2 show ?thesis by fastforce
qed

lemma FstEmptyOrEqvEmpty:
 "\<turnstile> \<rhd>(empty \<or> f) = empty"
proof -
 have 1: "\<turnstile> \<rhd>(empty \<or> f) = ((\<rhd>empty \<and> bs (\<not> f)) \<or> (\<rhd>f \<and> bs (\<not> empty)))" using FstWithOrEqv by blast
 have 2: "\<turnstile>  (\<not> empty) = more"  by (simp add: empty_d_def)
 hence 3: "\<turnstile> bs (\<not> empty) = bs more" using BsEqvRule by blast
 have 4: "\<turnstile> bs more = empty" using BsMoreEqvEmpty by blast
 have 5: "\<turnstile> (\<rhd>f \<and> bs (\<not> empty)) = (\<rhd>f \<and> empty)" using "3" "4" by fastforce
 have  6: "\<turnstile> \<rhd>empty = empty" using FstEmpty by blast
 hence 7: "\<turnstile> (\<rhd>empty \<and> bs (\<not> f)) = (empty \<and> bs (\<not> f))" by auto
 have 8: "\<turnstile> (empty \<and>  bs (\<not> f)) = (empty \<and>(empty \<or> bi (\<not> f);skip))" by (simp add:bs_d_def)
 have 9: "\<turnstile> (empty \<and>(empty \<or> bi (\<not> f);skip)) = empty" by auto
 have 10: "\<turnstile> (empty \<and>  bs (\<not> f)) = empty " using "8" "9" by auto
 have 11: "\<turnstile> ((\<rhd>empty \<and> bs (\<not> f)) \<or> (\<rhd>f \<and> bs (\<not> empty))) =
          (empty \<or> (\<rhd>f \<and> empty))" using "7" "10" "5" by fastforce
 have 12: "\<turnstile> (empty \<or> (\<rhd>f \<and> empty)) = empty" by auto
 from 1 11 12 show ?thesis by fastforce
qed

lemma FstChopEmptyEqvFstChopFstEmpty:
 "\<turnstile> (\<rhd>f;g \<and> empty) = (\<rhd>f;\<rhd>g \<and> empty)"
proof -
 have 1: "\<turnstile> (\<rhd>f;g \<and> empty) = (\<rhd>f \<and> g \<and> empty)" using ChopEmptyAndEmpty by blast
 have 2: "\<turnstile> (\<rhd>g \<and> empty) = (g \<and> empty)  " using FstAndEmptyEqvAndEmpty by blast
 hence 3: "\<turnstile> (\<rhd>f \<and> g \<and> empty) = (\<rhd>f \<and> \<rhd>g \<and> empty)" by auto
 have 4: "\<turnstile> (\<rhd>f;\<rhd>g \<and> empty) = (\<rhd>f \<and> \<rhd>g \<and> empty)" using ChopEmptyAndEmpty by blast
 from 1 3 4 show ?thesis by fastforce
qed

lemma FstMoreEqvSkip:
 "\<turnstile> \<rhd> more = skip "
proof -
 have 1: "\<turnstile> \<rhd> more = (more \<and> bs (\<not> more)) " by (simp add: first_d_def)
 have 2: "\<turnstile> (more \<and> bs (\<not> more)) = (more \<and> (empty \<or> bi (\<not> more);skip)) " by (simp add:bs_d_def) 
 have 3: "\<turnstile> (more \<and> (empty \<or> bi (\<not> more);skip)) = (more \<and> bi (\<not> more);skip)" using empty_d_def 
 using MoreAndEmptyOrEqvMoreAnd by fastforce
 have 4: "\<turnstile> (more \<and> ((bi (\<not> more));skip)) = ((bi (\<not> more));skip)" using ChopSkipImpMore by fastforce
 have 5: "\<turnstile> ((bi (\<not> more));skip) = bi empty;skip" by (simp add: empty_d_def)
 have 6: "\<turnstile> bi empty = empty" using BiEmptyEqvEmpty by auto
 hence 7: "\<turnstile> bi empty;skip = empty;skip" using LeftChopEqvChop by blast
 have 8: "\<turnstile> empty;skip = skip " using EmptyChop by blast
 from 1 2 3 4 5 7 8 show ?thesis by (metis int_eq)
qed

lemma FstEqvBsNotAndDi:
 "\<turnstile> \<rhd>f = (bs (\<not> f) \<and> di f)"
proof -
 have 1: "\<turnstile> bs (\<not> f)  = (\<not>(ds f)) " by (simp add: ds_d_def)
 hence 2: "\<turnstile> (bs (\<not> f)  \<and> di f) = (\<not>(ds f) \<and> di f)" by auto
 have 3: "\<turnstile>  di f = (ds f \<or> f)" using OrDsEqvDi by fastforce
 hence 4 : "\<turnstile> (\<not>(ds f)  \<and> di f) = (\<not>(ds f)  \<and> (ds f \<or> f))" by auto
 have 5 : "\<turnstile> (\<not>(ds f)  \<and> (ds f \<or> f)) = (\<not>(ds f)  \<and> f)" by auto
 have 6 : "\<turnstile> (\<not>(ds f)  \<and> f) = (f \<and>  bs (\<not> f))" using 1 by auto
 from 2 4 5 6 show ?thesis by (metis first_d_def int_eq)
qed

lemma FstOrDiEqvDi:
 "\<turnstile> (\<rhd>f \<or> di f) = di f"
proof -
 have 1: "\<turnstile> (\<rhd>f \<or> di f) = ((f \<and> bs (\<not>f)) \<or> di f)" by (simp add: first_d_def)
 have 2: "\<turnstile> ((f \<and> bs (\<not>f)) \<or> di f) = ((f \<or> di f) \<and> (bs (\<not> f) \<or> di f))" by auto
 have 3: "\<turnstile> (f \<or> di f) = di f" 
 by (metis "2" DiIntro RRDiamondEqvDi int_eq Prop02 Prop03 Prop11 Prop12)
 hence 4: "\<turnstile> ((f \<or> di f) \<and> (bs (\<not> f) \<or> di f)) = (di f \<and> (bs (\<not> f) \<or> di f))" by auto
 have 5: "\<turnstile> (di f \<and> (bs (\<not> f) \<or> di f)) = di f" by auto
 from 1 2 4 5 show ?thesis by fastforce
qed

lemma FstAndDiEqvFst:
  "\<turnstile> (\<rhd>f \<and> di f) = \<rhd>f"
proof -
 have 1: "\<turnstile> (\<rhd>f \<and> di f) = ((f \<and> bs (\<not>f)) \<and> di f)" by (simp add: first_d_def)
 have 2: "\<turnstile> (f \<and> di f) = f" by (meson DiIntro Prop10 Prop11)
 hence 3: "\<turnstile> (f \<and> bs (\<not>f) \<and> di f) = (f \<and> bs (\<not>f))" by auto
 from 1 3 show ?thesis by (metis first_d_def int_iffD2 int_iffI Prop12)
qed

lemma DiEqvDiFst:
 "\<turnstile> di f = di (\<rhd> f)"
proof -
 have 1: "\<turnstile>  di (\<rhd> f) = di (f \<and> bs (\<not>f)) " 
      by (simp add: first_d_def)
 have 2: "\<turnstile> di (f \<and> bs (\<not>f)) \<longrightarrow> di f \<and> di (bs (\<not> f))" 
      using DiAndImpAnd by auto
 hence 3: "\<turnstile> di (f \<and> bs (\<not>f)) \<longrightarrow> di f" 
      by auto
 have 4: "\<turnstile> di (\<rhd> f) \<longrightarrow> di f" using "1" "3" 
      by fastforce
 have 5: "\<turnstile> (di f \<and> empty) = (f \<and> empty)" 
      using DiAndEmptyEqvAndEmpty by blast
 have 6: "\<turnstile> (\<rhd> f \<and> empty) = (f \<and> empty) " 
      using FstAndEmptyEqvAndEmpty by auto
 have 7: "\<turnstile> di f \<and> empty \<longrightarrow> \<rhd>f" 
      using "5" "6" by fastforce
 have 8: "\<turnstile> \<rhd>f \<longrightarrow> di (\<rhd> f)" 
      using DiIntro by auto
 have 9: "\<turnstile>  di f \<and> empty \<longrightarrow> di (\<rhd> f)" 
      using "7" "8" using lift_imp_trans by blast
 hence 10: "\<turnstile> empty \<longrightarrow> (di f \<longrightarrow> di (\<rhd> f)) " 
      by auto
 have 11: "\<turnstile> prev ( di f \<longrightarrow> di (\<rhd> f))\<longrightarrow> more " 
      by (simp add: ChopSkipImpMore prev_d_def)
 have 12: "\<turnstile> more \<longrightarrow>( prev ( di f \<longrightarrow> di (\<rhd> f)) = (prev(di f) \<longrightarrow> prev(di (\<rhd>f)))) " 
      using MoreImpImpPrevEqv by auto
 have 13: "\<turnstile> (more \<and> prev ( di f \<longrightarrow> di (\<rhd> f))) = (more \<and> (prev(di f) \<longrightarrow> prev(di (\<rhd>f))))" 
      using "12" by fastforce
 have 14: "\<turnstile> prev ( di f \<longrightarrow> di (\<rhd> f)) = (more \<and> (prev(di f) \<longrightarrow> prev(di (\<rhd>f))))" 
      using "11" "13" by fastforce
 have 15: "\<turnstile> di f = (f \<or> ds f)" 
      using OrDsEqvDi by fastforce
 have 16: "\<turnstile> di f = (di f \<and> (bs (\<not> f) \<or> \<not>(bs (\<not> f))))  " 
      by auto
 have 17: "\<turnstile> (di f \<and> (bs (\<not> f) \<or> \<not>(bs (\<not> f)))) = ((di f \<and> bs (\<not> f)) \<or> (di f \<and> \<not>(bs (\<not> f)))) " 
      by auto
 have 18: "\<turnstile> (di f \<and> bs (\<not> f)) = ((f \<or> ds f) \<and> bs (\<not> f))" 
      using "15" by auto
 have 19: "\<turnstile> ((f \<or> ds f) \<and> bs (\<not> f)) = ((f \<and> bs (\<not> f)) \<or> (ds f \<and> bs (\<not> f))) " 
      by auto
 have 20: "\<turnstile> \<not>(ds f \<and> bs (\<not> f))" 
      by (simp add: ds_d_def)
 have 21: "\<turnstile> ((f \<and> bs (\<not> f)) \<or> (ds f \<and> bs (\<not> f))) = (f \<and> bs (\<not> f))" 
      using "20" by auto
 have 22: "\<turnstile> (di f \<and> bs (\<not> f)) = (f \<and> bs (\<not> f))" 
      using "18" "19" "21" by fastforce 
 have 23: "\<turnstile> (f \<and> bs (\<not> f)) = \<rhd>f" 
      by (simp add: first_d_def)
 have 24: "\<turnstile> (\<rhd>f) \<longrightarrow> di (\<rhd>f)" 
      using DiIntro by auto
 have 25: "\<turnstile> (f \<and> bs (\<not> f)) \<longrightarrow> di (\<rhd>f)" 
      using "23" "24" by fastforce
 have 26: "\<turnstile> (di f \<and> bs (\<not> f)) \<longrightarrow> di (\<rhd>f)" 
      using "25" "22" by fastforce
 hence 27: "\<turnstile> (di f \<and> bs (\<not> f) \<and> (prev (di f \<longrightarrow> di (\<rhd> f)))) \<longrightarrow> di (\<rhd>f) " 
      by auto
 have 28: "\<turnstile> (di f \<and> \<not>(bs (\<not> f))) = (di f \<and> ds f) " 
      by (simp add: ds_d_def)
 hence 29: "\<turnstile> (di f \<and> \<not>(bs (\<not> f)) \<and> (prev (di f \<longrightarrow> di (\<rhd> f))))  = 
              (di f \<and> ds f \<and>  (prev (di f \<longrightarrow> di (\<rhd> f))))" 
      by auto
 have 30: "\<turnstile>  ds f = prev(di f)" 
      using DsDi by (metis prev_d_def)
 hence 31: "\<turnstile> (di f \<and> ds f \<and>  (prev (di f \<longrightarrow> di (\<rhd> f)))) =  
              (di f \<and> prev(di f) \<and>  (prev (di f \<longrightarrow> di (\<rhd> f))))" 
      by auto
 have 32: "\<turnstile> prev ( di f \<longrightarrow> di (\<rhd> f)) \<longrightarrow>  (prev(di f) \<longrightarrow> prev(di (\<rhd>f)))" 
      using "14" by auto
 hence 33: "\<turnstile> di f \<and> prev(di f) \<and> prev ( di f \<longrightarrow> di (\<rhd> f)) \<longrightarrow> 
              di f \<and> prev(di f) \<and> (prev(di f) \<longrightarrow> prev(di (\<rhd>f)))" 
       by auto
 have 34: "\<turnstile> di f \<and> prev(di f) \<and> (prev(di f) \<longrightarrow> prev(di (\<rhd>f))) \<longrightarrow> prev(di (\<rhd>f)) " 
      by auto
 have 35: "\<turnstile> prev(di (\<rhd>f))= (di (\<rhd>f));skip" 
      by (simp add: prev_d_def)
 have 36: "\<turnstile> (di (\<rhd>f));skip \<longrightarrow> di(di (\<rhd>f))" 
      using ChopImpDi by auto
 have 37: "\<turnstile> di(di (\<rhd>f)) = di (\<rhd>f)" 
      using DiEqvDiDi  by fastforce
 have 38: "\<turnstile> di f \<and> prev(di f) \<and> (prev(di f) \<longrightarrow> prev(di (\<rhd>f))) \<longrightarrow> di (\<rhd>f)" 
      using "37" "36" "35"  "34" by fastforce
 have 39: "\<turnstile> di f \<and> \<not>(bs (\<not> f)) \<and> (prev (di f \<longrightarrow> di (\<rhd> f))) \<longrightarrow> di (\<rhd>f)" 
      using "29" "31" "33" "38" by fastforce
 hence 40: "\<turnstile> \<not>(bs (\<not> f)) \<and> (prev (di f \<longrightarrow> di (\<rhd> f))) \<longrightarrow> (di f \<longrightarrow> di (\<rhd>f))" 
      by fastforce
 have 41: "\<turnstile> bs (\<not> f) \<and> (prev (di f \<longrightarrow> di (\<rhd> f))) \<longrightarrow>(di f  \<longrightarrow> di (\<rhd>f))" 
      using "27" by fastforce
 have 42: "\<turnstile> (\<not>(bs (\<not> f)) \<or> bs (\<not> f)) \<and> (prev (di f \<longrightarrow> di (\<rhd> f))) \<longrightarrow>(di f  \<longrightarrow> di (\<rhd>f)) " 
      using "40" "41" by fastforce
 have 43: "\<turnstile> (\<not>(bs (\<not> f)) \<or> bs (\<not> f))" 
      by auto
 have 44: "\<turnstile> (prev (di f \<longrightarrow> di (\<rhd> f))) \<longrightarrow>(di f  \<longrightarrow> di (\<rhd>f))" 
      using "42" "43" by fastforce
 have 45: "\<turnstile> di f  \<longrightarrow> di (\<rhd>f) " 
      using "10" "44" EmptyChopSkipInduct by blast
 from 4 45 show ?thesis by fastforce
qed

lemma FstDiEqvFst:
 "\<turnstile> \<rhd>(di f) = \<rhd>f"
proof -
 have 1: "\<turnstile> \<rhd>(di f) = (di f \<and> bs (\<not> (di f)))" by (simp add: first_d_def)
 have 2: "\<turnstile>  (\<not> (di f)) = bi (\<not> f)  " by (simp add: NotDiEqvBiNot)
 hence 3: "\<turnstile> bs (\<not> (di f)) = bs (bi (\<not> f))" using BsEqvRule by blast
 have 4: "\<turnstile> bs (bi (\<not> f)) = bs ( \<not> f)" using BsEqvBsBi by fastforce
 hence 5: "\<turnstile> (di f \<and> bs (\<not> (di f))) = (di f \<and> bs ( \<not> f))" using "3" by fastforce
 have 6 : "\<turnstile> di f = (f \<or> ds f)" using OrDsEqvDi by fastforce
 hence 7: "\<turnstile> (di f \<and> bs ( \<not> f)) = ((f \<or> ds f) \<and> bs ( \<not> f)) " by auto
 have 8: "\<turnstile> ((f \<or> ds f) \<and> bs ( \<not> f)) = ((f \<and> bs ( \<not> f)) \<or> (ds f \<and> bs ( \<not> f)))" by auto
 have 9: "\<turnstile> \<not>(ds f \<and> bs ( \<not> f)) " by (simp add: ds_d_def)
 have 10: "\<turnstile> (f \<and> bs ( \<not> f)) = \<rhd>f" by (simp add: first_d_def)
 have 11: "\<turnstile> ((f \<and> bs ( \<not> f)) \<or> (ds f \<and> bs ( \<not> f))) = \<rhd>f" using "9" "10" by fastforce
 from 1 5 7 8 11 show ?thesis by (metis int_eq)
qed

lemma DiAndFstOrEqvFstOrDiAnd:
 "\<turnstile> (di f \<and> (\<rhd>f \<or> g)) = (\<rhd>f \<or> (di f \<and> g))"
proof -
 have 1: "\<turnstile> (di f \<and> (\<rhd>f \<or> g)) = (\<rhd>f \<and> di f) \<or> (di f \<and> g)" by auto
 have 2: "\<turnstile> (\<rhd>f \<and> di f) = \<rhd>f"  using FstAndDiEqvFst by blast
 from 1 2 show ?thesis by auto
qed

lemma DiOrFstAndEqvDi:
 "\<turnstile> di f \<or> (\<rhd>f \<and> g) = di f"
proof -
 have 1: "\<turnstile> (di f \<or> (\<rhd>f \<and> g)) = ((\<rhd>f \<or> di f) \<and> (di f \<or> g)) " by auto
 have 2: "\<turnstile> (\<rhd>f \<or> di f) = di f" using FstOrDiEqvDi by blast
 from 1 2 show ?thesis by auto
qed

lemma FstDiAndDiEqv:
 "\<turnstile> \<rhd>(di f \<and> di g) = ((\<rhd>f \<and> di g) \<or> (\<rhd>g \<and> di f))"
proof -
 have 1: "\<turnstile> \<rhd>(di f \<and> di g) = ((di f \<and> di g) \<and> bs (\<not>(di f \<and> di g)))" by (simp add: first_d_def)
 have 2: "\<turnstile> (\<not>(di f \<and> di g)) = (bi (\<not>f) \<or> bi (\<not>g)) " by (auto simp add: bi_d_def) 
 hence 3: "\<turnstile> bs (\<not>(di f \<and> di g)) = bs(bi (\<not>f) \<or> bi (\<not>g))" using BsEqvRule by blast
 hence 4: "\<turnstile>  ((di f \<and> di g) \<and> bs (\<not>(di f \<and> di g))) =
              (di f \<and> di g \<and> bs(bi (\<not>f) \<or> bi (\<not>g))) " by auto
 have 5: "\<turnstile> (bs (\<not>f) \<or> bs (\<not>g)) = bs(bi (\<not>f) \<or> bi (\<not>g))" using BsOrBsEqvBsBiOrBi by blast
 hence 6: "\<turnstile> (di f \<and> di g \<and> bs(bi (\<not>f) \<or> bi (\<not>g))) =
             (di f \<and> di g \<and>(bs (\<not>f) \<or> bs (\<not>g)))" by auto
 have 7: "\<turnstile> (di f \<and> di g \<and>(bs (\<not>f) \<or> bs (\<not>g))) = 
            ((  bs (\<not>f) \<and> di f \<and> di g) \<or> (di f \<and> bs (\<not>g) \<and> di g  ))" by auto
 have 8: "\<turnstile> \<rhd>f =  (bs (\<not>f)  \<and> di f)" using FstEqvBsNotAndDi by blast
 hence 9: "\<turnstile> (bs (\<not>f) \<and> di f \<and> di g) = (\<rhd>f \<and> di g)" by auto
 have 10: "\<turnstile> \<rhd>g =  (bs (\<not>g)  \<and> di g)" using FstEqvBsNotAndDi by blast
 hence 11: "\<turnstile> (di f \<and> bs (\<not>g) \<and> di g) = (di f \<and> \<rhd>g)" by auto
  show ?thesis
  by (metis "11" "4" "6" "7" "9" first_d_def inteq_reflection lift_and_com)
qed

lemma BiNotFstEqvBiNot: 
 "\<turnstile> bi (\<not> (\<rhd>f)) = bi (\<not> f)"
proof -
 have 1: "\<turnstile> di f = di (\<rhd>f)" using DiEqvDiFst by blast
 hence 2: "\<turnstile> (\<not>(di f)) = (\<not>( di (\<rhd>f)))" by auto
 from 1 2 show ?thesis using NotDiEqvBiNot by fastforce
qed

lemma BsNotFstEqvBsNot:
 "\<turnstile> bs (\<not> (\<rhd>f)) = bs (\<not> f)"
proof -
 have 1: "\<turnstile> bs (\<not> (\<rhd>f)) = (empty \<or> bi (\<not> (\<rhd>f));skip) " by (simp add: bs_d_def)
 have 2: "\<turnstile> bi (\<not> (\<rhd>f)) = bi (\<not> f) " using BiNotFstEqvBiNot by blast
 hence 3: "\<turnstile> bi (\<not> (\<rhd>f));skip = bi (\<not> f);skip" using LeftChopEqvChop by blast
 hence 4: "\<turnstile> (empty \<or> bi (\<not> (\<rhd>f));skip) = (empty \<or> bi (\<not> f);skip)" by auto
 from 1 4 show ?thesis by (simp add: bs_d_def)
qed

lemma FstState:
 "\<turnstile> \<rhd> (init w) = (empty \<and> init w)"
proof -
 have 1: "\<turnstile> \<rhd> (init w) = (init w \<and> bs (\<not>(init w)))" by (simp add: first_d_def)
 hence 2: "\<turnstile> \<rhd> (init w) \<longrightarrow> init w" by auto
 have 3: "\<turnstile> init w \<longrightarrow> bs (init w)" using StateImpBs by auto
 have 4: "\<turnstile> \<rhd> (init w) \<longrightarrow> bs (init w)" using "2" "3" by fastforce
 have 5: "\<turnstile> \<rhd> (init w) \<longrightarrow> bs (\<not>(init w))" using "1" by auto
 have 6: "\<turnstile> \<rhd> (init w) \<longrightarrow> bs (init w) \<and> bs (\<not>(init w)) " using "4" "5" by fastforce
 have 7: "\<turnstile> (bs (init w) \<and> bs (\<not>(init w))) = (bs((init w) \<and> \<not>(init w))) " using BsAndEqv by blast
 have 8: "\<turnstile> ((init w) \<and> \<not>(init w)) =  #False"  by auto
 hence 9: "\<turnstile> (bs((init w) \<and> \<not>(init w))) =  bs #False" using BsEqvRule by blast
 have 10: "\<turnstile> bs #False = empty" using BsFalseEqvEmpty by auto
 have 11: "\<turnstile> \<rhd> (init w) \<longrightarrow> empty" using "10" "9" "7" "6" by fastforce
 have 12: "\<turnstile> \<rhd> (init w) \<longrightarrow> empty \<and> init w" using "11" "2" by fastforce
 have 13: "\<turnstile> empty \<and> init w \<longrightarrow> empty" by auto
 hence 14: "\<turnstile> empty \<and> init w \<longrightarrow> empty \<or> bi (\<not>(init w));skip" by auto
 hence 15: "\<turnstile> empty \<and> init w \<longrightarrow> bs (\<not>(init w))" by (simp add: bs_d_def)
 have 16: "\<turnstile> empty \<and> init w \<longrightarrow> init w  " by auto
 have 17: "\<turnstile> empty \<and> init w \<longrightarrow> init w \<and> bs (\<not>(init w))" using "16" "15" by auto
 hence 18: "\<turnstile> empty \<and> init w \<longrightarrow> \<rhd>(init w)" by (simp add: first_d_def)
 from 12 18 show ?thesis by fastforce
qed

lemma FstStateAndBsNotEmpty:
 "\<turnstile> (\<rhd> (init w) \<and> bs (\<not> empty)) = \<rhd> (init w)"
proof -
 have 1: "\<turnstile> (\<rhd> (init w) \<and> bs (\<not> empty)) = (\<rhd> (init w) \<and> bs more) " 
      using BsEqvRule NotEmptyEqvMore by (simp add: empty_d_def)
 have 2: "\<turnstile> (\<rhd> (init w) \<and> bs more) = (\<rhd> (init w) \<and> empty)" 
      using BsMoreEqvEmpty by fastforce
 have 3: "\<turnstile> \<rhd> (init w)  =  (empty \<and> (init w))  " 
      using FstState by blast
 hence 4: "\<turnstile> (\<rhd> (init w)  \<and> empty) =  (empty \<and> (init w)  \<and> empty) " 
      by auto
 have 5: "\<turnstile> (empty \<and> (init w)  \<and> empty) = (empty \<and> (init w))  " 
      by auto
 have 6: "\<turnstile> (empty \<and> (init w)) = \<rhd>(init w)" 
      using FstState by fastforce
 from 1 2 4 5 6 show ?thesis by fastforce
qed

lemma FstStateImpFstStateOr:
 "\<turnstile> \<rhd>(init w) \<longrightarrow> \<rhd>(init w \<or> f)"
proof -
 have 1: "\<turnstile> \<rhd>(init w) = (empty \<and> init w)" 
      using FstState by blast
 have 2: "\<turnstile>  (empty \<and> init w) =  (empty \<and> (empty \<or> bi (\<not> f);skip) \<and> init w)" 
      by auto
 have 3: "\<turnstile>  (empty \<and> (empty \<or> bi (\<not> f);skip) \<and> init w) = 
             (empty \<and> bs (\<not> f) \<and> init w)" 
      by (simp add: bs_d_def)
 have 4: "\<turnstile>  (empty \<and> bs (\<not> f) \<and> init w) = (empty \<and> init w \<and> bs (\<not> f)) " 
      by auto
 have 5: "\<turnstile> (empty \<and> init w)  = \<rhd> (init w) " 
      using FstState  by fastforce
 hence 6: "\<turnstile> (empty \<and> init w \<and> bs (\<not> f)) = (\<rhd> (init w) \<and> bs (\<not> f))" 
      by auto
 have 7: "\<turnstile> \<rhd> (init w) \<and> bs (\<not> f) \<longrightarrow> (\<rhd> (init w) \<and> bs (\<not> f)) \<or> (\<rhd>f \<and> bs (\<not>(init w)))" 
      by auto
 have 8: "\<turnstile> \<rhd>(init w \<or> f) = ((\<rhd> (init w) \<and> bs (\<not> f)) \<or> (\<rhd>f \<and> bs (\<not>(init w))))" 
      using FstWithOrEqv by blast
  show ?thesis using "1" "3" "8" by fastforce 
qed


lemma FstLenSame:
  "(\<forall> \<sigma>. (\<sigma> \<Turnstile>  di (\<rhd>f \<and> len(i)) \<and> di (\<rhd>f \<and> len(j))) \<longrightarrow> (i=j))"
by (simp add: DiLenFstsem FstLenSamesem)

lemma FstLenSame_1:
  "\<turnstile>  di (\<rhd>f \<and> len(i)) \<and> di (\<rhd>f \<and> len(j)) \<longrightarrow> (#i=#j)"
using FstLenSame Valid_def by fastforce

lemma FstAndLenSame:
  "(\<forall> \<sigma>. (\<sigma> \<Turnstile>  di ((\<rhd>f \<and> g1) \<and> len(i)) \<and> di ((\<rhd>f \<and> g2) \<and> len(j))) \<longrightarrow> (i=j))"
using linorder_neqE_nat by (simp  add: DiLenFstAndsem) blast 

lemma FstAndLenSame_1:
  "\<turnstile>  di ((\<rhd>f \<and> g1) \<and> len(i)) \<and> di ((\<rhd>f \<and> g2) \<and> len(j)) \<longrightarrow> (#i=#j)"
using FstAndLenSame Valid_def by fastforce

lemma FstLenSameChop:
  "(\<forall> \<sigma>. (\<sigma> \<Turnstile>  ((\<rhd>f \<and> g1) \<and> len(i));h1 \<and>  ((\<rhd>f \<and> g2) \<and> len(j));h2) \<longrightarrow> (i=j))"
proof
  fix \<sigma>
  show "(\<sigma> \<Turnstile>  ((\<rhd>f  \<and> g1) \<and> len(i));h1 \<and>  ((\<rhd>f \<and> g2) \<and> len(j));h2) \<longrightarrow> (i=j) "
  proof 
  assume 0: "(\<sigma> \<Turnstile>  ((\<rhd>f \<and> g1) \<and> len(i));h1 \<and>  ((\<rhd>f \<and> g2) \<and> len(j));h2)"
   have 1: "(\<sigma> \<Turnstile>  ((\<rhd>f \<and> g1) \<and> len(i));h1) " using "0" by auto
   have 2: "(\<sigma> \<Turnstile>  ((\<rhd>f \<and> g1) \<and> len(i));h1) \<longrightarrow>
            (\<sigma> \<Turnstile>  ((\<rhd>f \<and> g1) \<and> len(i));#True) " by (metis ChopImpDi Valid_def di_d_def unl_lift2)
   have 3: "(\<sigma> \<Turnstile>  di((\<rhd>f \<and> g1) \<and> len(i))) " using "1" "2" by (simp add: di_d_def)
   have 4: "(\<sigma> \<Turnstile>  ((\<rhd>f \<and> g2) \<and> len(j));h2) " using "0" by auto
   have 5: "(\<sigma> \<Turnstile>  ((\<rhd>f \<and> g2) \<and> len(j));h2) \<longrightarrow>
            (\<sigma> \<Turnstile>  ((\<rhd>f \<and> g2) \<and> len(j));#True) " by (metis ChopImpDi Valid_def di_d_def unl_lift2)
   have 6: "(\<sigma> \<Turnstile>  di((\<rhd>f \<and> g2) \<and> len(j))) " using "4" "5" by (simp add: di_d_def)
   have 7: "(\<sigma> \<Turnstile>  di((\<rhd>f \<and> g1) \<and> len(i)) \<and> di((\<rhd>f \<and> g2) \<and> len(j))) " using "3" "6" by auto
   thus "(i=j)" using FstAndLenSame by blast
   qed
qed

lemma FstLenSameChop_1:
  "\<turnstile>  ((\<rhd>f \<and> g1) \<and> len(i));h1 \<and>  ((\<rhd>f \<and> g2) \<and> len(j));h2 \<longrightarrow> (#i=#j)"
using FstLenSameChop Valid_def by fastforce

lemma DiImpExistsOneDiLenAndFst:
 "(\<forall>\<sigma>. (\<sigma>\<Turnstile> di f) \<longrightarrow> (\<exists>! k. (\<sigma> \<Turnstile> di(  \<rhd>f \<and> len(k)))))"
proof
 fix \<sigma>
 show "(\<sigma>\<Turnstile> di f) \<longrightarrow> (\<exists>! k. (\<sigma> \<Turnstile> di(  \<rhd>f \<and> len(k)))) "
 proof 
  assume 0: "(\<sigma>\<Turnstile> di f) "
  have 1: "(\<sigma> \<Turnstile> di (\<rhd> f)) " 
       using "0" DiEqvDiFst Valid_def by force
  have 2: "(\<sigma> \<Turnstile> \<rhd> f) = ( (\<sigma>\<Turnstile>\<rhd> f) \<and> (\<exists>k. (\<sigma> \<Turnstile> len(k))))  " 
       using AndExistsLen[of "TEMP \<rhd> f"] by (simp add: Valid_def) 
  have 3: "((\<sigma>\<Turnstile>\<rhd> f) \<and> (\<exists>k. (\<sigma> \<Turnstile> len(k)))) =
           (\<exists>k.  (\<sigma>\<Turnstile>\<rhd> f) \<and>  (\<sigma> \<Turnstile> len(k))) " 
       by auto
  have 4: "(\<sigma> \<Turnstile> di(\<rhd> f)) = (\<exists>k. (\<sigma> \<Turnstile> di(\<rhd>f \<and> len(k))))" 
       using "2" "3" by (metis "1" DiLensem di_defs)
  have 5: "(\<exists>k. (\<sigma> \<Turnstile> di(\<rhd>f \<and> len(k)))) " 
       using "1" using "4" by auto
  then obtain i where 6: "(\<sigma> \<Turnstile> di(\<rhd>f \<and> len(i))) " by blast
  from 5 obtain j where 7: "(\<sigma> \<Turnstile> di(\<rhd>f \<and> len(j))) " by blast
  have 8: "(\<sigma> \<Turnstile> di(\<rhd>f \<and> len(i))) \<and> (\<sigma> \<Turnstile> di(\<rhd>f \<and> len(j))) " 
       using "6" "7" by auto
  hence 9: "(\<sigma> \<Turnstile> di(\<rhd>f \<and> len(i)) \<and> di(\<rhd>f \<and> len(j)))  " 
       by simp
  hence 10: "i=j " 
       using FstLenSame by blast
  have 11: "\<And>j. (\<sigma> \<Turnstile> di(\<rhd>f \<and> len(j))) \<longrightarrow> (j=i) " 
       using "9" "10" using FstLenSame by auto
  thus "(\<exists>! k. (\<sigma> \<Turnstile> di(  \<rhd>f \<and> len(k) )))" 
       using "11" "5" by blast  
  qed
qed

lemma DiImpExistsOneDiLenAndFst_1:
 "\<turnstile> di f \<longrightarrow> (\<exists>! k. ( di( \<rhd>f \<and> len(k))))"
using Valid_def DiImpExistsOneDiLenAndFst by fastforce

lemma LFstAndDist_help:
 "(\<sigma> \<Turnstile>((\<rhd>f \<and> g1) \<and> len(k));h1  \<and> ((\<rhd>f \<and> g2) \<and> len(k));h2) =
  (\<sigma> \<Turnstile>(((\<rhd>f \<and> g1) \<and> (\<rhd>f \<and> g2)) \<and> len(k));(h1 \<and> h2) )"
using LFixedAndDistr by fastforce

lemma LFstAndDist_help_1:
 "(\<exists> k. (\<sigma> \<Turnstile>((\<rhd>f \<and> g1) \<and> len(k));h1  \<and> ((\<rhd>f \<and> g2) \<and> len(k));h2)) =
            (\<exists> k. (\<sigma> \<Turnstile>(((\<rhd>f \<and> g1) \<and> (\<rhd>f \<and> g2)) \<and> len(k));(h1 \<and> h2) )) "
proof 
  assume 0: "\<exists>k. \<sigma> \<Turnstile>  ((\<rhd> f \<and> g1) \<and> len k) ; h1  \<and>  ((\<rhd> f \<and> g2) \<and> len k) ; h2"
  obtain k where 1: " \<sigma> \<Turnstile>  ((\<rhd> f \<and> g1) \<and> len k) ; h1  \<and>  ((\<rhd> f \<and> g2) \<and> len k) ; h2" 
  using "0" by auto
  hence 2: "(\<sigma> \<Turnstile>(((\<rhd>f \<and> g1) \<and> (\<rhd>f \<and> g2)) \<and> len(k));(h1 \<and> h2)) " 
  using  LFstAndDist_help by blast
  show "(\<exists> k. (\<sigma> \<Turnstile>(((\<rhd>f \<and> g1) \<and> (\<rhd>f \<and> g2)) \<and> len(k));(h1 \<and> h2) )) "  
  using "2" by auto
  next
  assume 3: "(\<exists> k. (\<sigma> \<Turnstile>(((\<rhd>f \<and> g1) \<and> (\<rhd>f \<and> g2)) \<and> len(k));(h1 \<and> h2) )) "
  obtain k where 4: "(\<sigma> \<Turnstile>(((\<rhd>f \<and> g1) \<and> (\<rhd>f \<and> g2)) \<and> len(k));(h1 \<and> h2) ) " 
  using "3" by auto
  hence 5: "(\<sigma> \<Turnstile>((\<rhd>f \<and> g1) \<and> len(k));h1  \<and> ((\<rhd>f \<and> g2) \<and> len(k));h2) " 
  using  LFstAndDist_help by blast
  show "(\<exists> k. (\<sigma> \<Turnstile>((\<rhd>f \<and> g1) \<and> len(k));h1  \<and> ((\<rhd>f \<and> g2) \<and> len(k));h2)) " 
  using "5" by auto
qed

lemma LFstAndDistrsem:
 "(\<forall> \<sigma>. (\<sigma> \<Turnstile>((\<rhd>f \<and> g1);h1 \<and> (\<rhd>f \<and> g2);h2) = (\<rhd>f \<and> g1 \<and> g2);(h1 \<and> h2)))"
proof 
 fix \<sigma> 
 show "(\<sigma> \<Turnstile>((\<rhd>f \<and> g1);h1 \<and> (\<rhd>f \<and> g2);h2) = (\<rhd>f \<and> g1 \<and> g2);(h1 \<and> h2))" 
 proof -
  have 1: "(\<sigma> \<Turnstile>(\<rhd>f \<and> g1);h1) = (\<exists> i. (\<sigma> \<Turnstile>((\<rhd>f \<and> g1) \<and> len(i));h1) ) " 
  using AndExistsLenChop[of "TEMP (\<rhd> f \<and> g1)"] by fastforce
  have 2: "(\<sigma> \<Turnstile>(\<rhd>f \<and> g2);h2) = (\<exists> j. (\<sigma> \<Turnstile>((\<rhd>f \<and> g2) \<and> len(j));h2) ) " 
  using AndExistsLenChop[of "TEMP (\<rhd>f \<and> g2)"] by fastforce
  have 3: "(\<sigma> \<Turnstile>(\<rhd>f \<and> g1);h1 \<and> (\<rhd>f \<and> g2);h2) =
          ( (\<exists> i j. (\<sigma> \<Turnstile>((\<rhd>f \<and> g1) \<and> len(i));h1  \<and>
                             ((\<rhd>f \<and> g2) \<and> len(j));h2) )
          ) " 
  using "1" "2" by auto
  have 4: "( (\<exists> i j. (\<sigma> \<Turnstile>((\<rhd>f \<and> g1) \<and> len(i));h1  \<and>
                             ((\<rhd>f \<and> g2) \<and> len(j));h2) )
          ) =
          ( (\<exists> k. (\<sigma> \<Turnstile>((\<rhd>f \<and> g1) \<and> len(k));h1  \<and>
                             ((\<rhd>f \<and> g2) \<and> len(k));h2) )
          )  " 
  using FstLenSameChop by blast
  have 5: "(\<exists> k. (\<sigma> \<Turnstile>((\<rhd>f \<and> g1) \<and> len(k));h1  \<and> ((\<rhd>f \<and> g2) \<and> len(k));h2)) =
            (\<exists> k. (\<sigma> \<Turnstile>(((\<rhd>f \<and> g1) \<and> (\<rhd>f \<and> g2)) \<and> len(k));(h1 \<and> h2) ))  " 
  using LFstAndDist_help_1 by blast 
  have 6 : " (\<exists> k. (\<sigma> \<Turnstile>(((\<rhd>f \<and> g1) \<and> (\<rhd>f \<and> g2)) \<and> len(k));(h1 \<and> h2) )) =
             (\<sigma> \<Turnstile>((\<rhd>f \<and> g1) \<and> (\<rhd>f \<and> g2));(h1 \<and> h2)) "   
  using AndExistsLenChop[of "TEMP ((\<rhd> f \<and> g1) \<and> \<rhd> f \<and> g2)"] by fastforce
  have 7 : "(\<sigma> \<Turnstile>((\<rhd>f \<and> g1) \<and> (\<rhd>f \<and> g2));(h1 \<and> h2)) =
            (\<sigma> \<Turnstile>(\<rhd>f \<and> g1 \<and> g2);(h1 \<and> h2)) " 
  by (auto simp add: chop_defs)
  from 3 4 5 6 7 show ?thesis by auto
  qed
qed

lemma LFstAndDistr:
 "\<turnstile> ((\<rhd>f \<and> g1);h1 \<and> (\<rhd>f \<and> g2);h2) = (\<rhd>f \<and> g1 \<and> g2);(h1 \<and> h2) "
using  LFstAndDistrsem by fastforce

lemma LFstAndDistrA:
 "\<turnstile> ((\<rhd>f \<and> g1);h \<and> (\<rhd>f \<and> g2);h) = (\<rhd>f \<and> g1 \<and> g2);h "
proof -
 have 1: "\<turnstile> ((\<rhd>f \<and> g1);h \<and> (\<rhd>f \<and> g2);h) = (\<rhd>f \<and> g1 \<and> g2);(h \<and> h) " using LFstAndDistr by blast
 have 2: "\<turnstile> (\<rhd>f \<and> g1 \<and> g2);(h \<and> h) = (\<rhd>f \<and> g1 \<and> g2);h " by auto
 from 1 2 show ?thesis by auto
qed

lemma LFstAndDistrB:
 "\<turnstile> ((\<rhd>f \<and> g);h1 \<and> (\<rhd>f \<and> g);h2) = (\<rhd>f \<and> g);(h1 \<and> h2) " 
proof -
 have 1: "\<turnstile> ((\<rhd>f \<and> g);h1 \<and> (\<rhd>f \<and> g);h2) = (\<rhd>f \<and> g \<and> g);(h1 \<and> h2) " using LFstAndDistr by blast
 have 2: "\<turnstile> (\<rhd>f \<and> g \<and> g);(h1 \<and> h2) = (\<rhd>f \<and> g);(h1 \<and> h2) " by auto
 from 1 2 show ?thesis by auto
qed

lemma LFstAndDistrC:
  "\<turnstile> ((\<rhd>f );h1 \<and> (\<rhd>f );h2) = (\<rhd>f);(h1 \<and> h2)"
proof -
 have 1: "\<turnstile> ((\<rhd>f \<and> #True);h1 \<and> (\<rhd>f \<and> #True);h2) = (\<rhd>f \<and> #True \<and> #True);(h1 \<and> h2) " 
      using LFstAndDistr by blast
 have 2: "\<turnstile> (\<rhd>f \<and> #True);h1 = (\<rhd>f );h1 " 
      by auto
 have 3: "\<turnstile> (\<rhd>f \<and> #True);h2 = (\<rhd>f );h2 " 
      by auto
 have 4: "\<turnstile> (\<rhd>f \<and> #True \<and> #True);(h1 \<and> h2) = (\<rhd>f);(h1 \<and> h2)  " 
      by auto
 from 1 2 3 4 show ?thesis by auto
qed

lemma LFstAndDistrD:
 "\<turnstile> (di(\<rhd>f \<and> g1) \<and> di(\<rhd>f \<and> g2)) = di(\<rhd>f \<and> g1 \<and> g2) "
proof -
 have 1: "\<turnstile> ((\<rhd>f \<and> g1);#True \<and> (\<rhd>f \<and> g2);#True) = (\<rhd>f \<and> g1 \<and> g2);(#True \<and> #True) " 
      using LFstAndDistr by blast
 have 2: "\<turnstile> (\<rhd>f \<and> g1);#True = di(\<rhd>f \<and> g1) " 
      by (simp add: di_d_def)
 have 3: "\<turnstile> (\<rhd>f \<and> g2);#True = di(\<rhd>f \<and> g2) " 
      by (simp add: di_d_def)
 have 4: "\<turnstile> (\<rhd>f \<and> g1 \<and> g2);(#True \<and> #True) = di(\<rhd>f \<and> g1 \<and> g2) "  
      by (simp add: di_d_def)
 from 1 2 3 4 show ?thesis by fastforce
qed

lemma LstAndDistr:
 "\<turnstile> (h1;(\<lhd> f \<and> g1) \<and> h2;(\<lhd> f \<and> g2)) = (h1 \<and> h2);(\<lhd> f \<and> g1 \<and> g2) "
proof -
 have 1: "\<turnstile> ((\<rhd>(f\<^sup>r) \<and> g1\<^sup>r);(h1\<^sup>r) \<and> (\<rhd>(f\<^sup>r) \<and> (g2\<^sup>r));(h2\<^sup>r)) = 
            (\<rhd>(f\<^sup>r) \<and> (g1\<^sup>r) \<and> (g2\<^sup>r));((h1\<^sup>r) \<and> (h2\<^sup>r)) " 
      using LFstAndDistr by blast
 hence 2: "\<turnstile> ((\<rhd>(f\<^sup>r) \<and> g1\<^sup>r);(h1\<^sup>r) \<and> (\<rhd>(f\<^sup>r) \<and> (g2\<^sup>r));(h2\<^sup>r))\<^sup>r = 
            ((\<rhd>(f\<^sup>r) \<and> (g1\<^sup>r) \<and> (g2\<^sup>r));((h1\<^sup>r) \<and> (h2\<^sup>r)))\<^sup>r" 
      using "1" REqvRule by blast
 have 3: "\<turnstile>  (((\<rhd>(f\<^sup>r) \<and> g1\<^sup>r);(h1\<^sup>r))\<^sup>r \<and> ((\<rhd>(f\<^sup>r) \<and> (g2\<^sup>r));(h2\<^sup>r))\<^sup>r) =
             ((\<rhd>(f\<^sup>r) \<and> g1\<^sup>r);(h1\<^sup>r) \<and> (\<rhd>(f\<^sup>r) \<and> (g2\<^sup>r));(h2\<^sup>r))\<^sup>r 
         " 
      using RAnd by fastforce
 have 4: "\<turnstile> ((h1\<^sup>r)\<^sup>r;(\<rhd>(f\<^sup>r) \<and> g1\<^sup>r)\<^sup>r \<and> (h2\<^sup>r)\<^sup>r;(\<rhd>(f\<^sup>r) \<and> (g2\<^sup>r))\<^sup>r) =
            (((\<rhd>(f\<^sup>r) \<and> g1\<^sup>r);(h1\<^sup>r))\<^sup>r \<and> ((\<rhd>(f\<^sup>r) \<and> (g2\<^sup>r));(h2\<^sup>r))\<^sup>r) 
         " 
      using RevChop by fastforce
 have 5: "\<turnstile> (h1\<^sup>r)\<^sup>r = h1" 
      using EqvReverseReverse by blast
 have 6: "\<turnstile> (h2\<^sup>r)\<^sup>r = h2" 
      using EqvReverseReverse by blast
 have 7: "\<turnstile> (g1\<^sup>r)\<^sup>r = g1" 
      using EqvReverseReverse by blast
 have 8: "\<turnstile> (g2\<^sup>r)\<^sup>r = g2" 
      using EqvReverseReverse by blast
 have 9: "\<turnstile> (f\<^sup>r)\<^sup>r  = f"
      using EqvReverseReverse by blast
 have 10: "\<turnstile> (\<rhd>(f\<^sup>r) \<and> g1\<^sup>r)\<^sup>r = ((\<rhd>(f\<^sup>r))\<^sup>r  \<and> (g1\<^sup>r)\<^sup>r)" 
      using RAnd by blast
 have 11: "\<turnstile> (\<rhd>(f\<^sup>r) \<and> g2\<^sup>r)\<^sup>r = ((\<rhd>(f\<^sup>r))\<^sup>r  \<and> (g2\<^sup>r)\<^sup>r)" 
      using RAnd by blast
 have 12: "\<turnstile> (\<rhd>(f\<^sup>r))\<^sup>r  = \<lhd> (f)  " 
      using RRFirstEqvLast by blast 
 have 13: "\<turnstile> ((\<rhd>(f\<^sup>r))\<^sup>r  \<and> (g1\<^sup>r)\<^sup>r) = (\<lhd> f \<and> g1)" 
      using  "12" "7" by fastforce
 have 14: "\<turnstile> ((\<rhd>(f\<^sup>r))\<^sup>r  \<and> (g2\<^sup>r)\<^sup>r) = (\<lhd> f \<and> g2)" 
      using  "12" "8" by fastforce
 have 15: "\<turnstile> (h1;(\<lhd> f \<and> g1) \<and> h2;(\<lhd> f \<and> g2)) = 
             ((h1\<^sup>r)\<^sup>r;(\<rhd>(f\<^sup>r) \<and> g1\<^sup>r)\<^sup>r \<and> (h2\<^sup>r)\<^sup>r;(\<rhd>(f\<^sup>r) \<and> (g2\<^sup>r))\<^sup>r) 
              " 
      using "14" "13" "10" "11" "5" "6" by (metis "4" int_eq)
 have 16: "\<turnstile> (((\<rhd>(f\<^sup>r) \<and> (g1\<^sup>r) \<and> (g2\<^sup>r));((h1\<^sup>r) \<and> (h2\<^sup>r)))\<^sup>r) = 
             ((h1\<^sup>r) \<and> (h2\<^sup>r))\<^sup>r;((\<rhd> (f\<^sup>r)) \<and> (g1\<^sup>r) \<and> (g2\<^sup>r))\<^sup>r" 
      by (simp add: RevChop)
 have 17: "\<turnstile> ((\<rhd> (f\<^sup>r)) \<and> (g1\<^sup>r) \<and> (g2)\<^sup>r)\<^sup>r = ((\<rhd> (f\<^sup>r))\<^sup>r \<and> (g1\<^sup>r)\<^sup>r \<and> (g2\<^sup>r)\<^sup>r)" 
      by (metis inteq_reflection rev_fun2)
 have 18: "\<turnstile> ((\<rhd> (f\<^sup>r))\<^sup>r \<and> (g1\<^sup>r)\<^sup>r \<and> (g2\<^sup>r)\<^sup>r) = (\<lhd> f \<and> g1 \<and> g2)" 
      using "12" "7" "8" by fastforce
 have 19: "\<turnstile> ((h1\<^sup>r) \<and> (h2\<^sup>r))\<^sup>r = (h1 \<and> h2) " 
      using RRAnd by auto
 have 20: "\<turnstile> ((h1\<^sup>r) \<and> (h2\<^sup>r))\<^sup>r;((\<rhd> (f\<^sup>r)) \<and> (g1\<^sup>r) \<and> (g2\<^sup>r))\<^sup>r =
             (h1 \<and> h2);(\<lhd> f \<and> g1 \<and> g2)" 
      using "19" "17" "18" using ChopEqvChop by (metis int_eq)
 from 15 4 3 2 16 20 show ?thesis using int_eq by metis
qed

lemma LstAndDistrA:
  "\<turnstile> (h;(\<lhd> f \<and> g1) \<and> h;(\<lhd> f \<and> g2)) = h;(\<lhd> f \<and> g1 \<and> g2) "
proof -
 have 1: "\<turnstile> (h;(\<lhd> f \<and> g1) \<and> h;(\<lhd> f \<and> g2)) = (h \<and> h);(\<lhd> f \<and> g1 \<and> g2) " 
      using LstAndDistr by blast
 have 2: "\<turnstile> (h \<and> h);(\<lhd> f \<and> g1 \<and> g2) = h;(\<lhd> f \<and> g1 \<and> g2) " 
      by auto
 from 1 2 show ?thesis by auto
qed

lemma LstAndDistrB:
 "\<turnstile>  (h1;(\<lhd> f \<and> g) \<and> h2;(\<lhd> f \<and> g)) = (h1 \<and> h2);(\<lhd> f \<and> g) "
proof -
 have 1: "\<turnstile> (h1;(\<lhd> f \<and> g) \<and> h2;(\<lhd> f \<and> g)) = (h1 \<and> h2);(\<lhd> f \<and> g \<and> g) " 
      using LstAndDistr by blast
 have 2: "\<turnstile> (h1 \<and> h2);(\<lhd> f \<and> g \<and> g) = (h1 \<and> h2);(\<lhd> f \<and> g) " 
      by auto
 from 1 2 show ?thesis by auto
qed

lemma LstAndDistrC:
 "\<turnstile>  (h1;(\<lhd> f ) \<and> h2;(\<lhd> f )) = (h1 \<and> h2);(\<lhd> f ) "
proof -
 have 1: "\<turnstile> (h1;(\<lhd> f \<and> #True) \<and> h2;(\<lhd> f \<and> #True)) = (h1 \<and> h2);(\<lhd> f \<and> #True \<and> #True  )  " 
      using LstAndDistr by blast
 have 2: "\<turnstile> (h1 \<and> h2);(\<lhd> f \<and> #True \<and> #True  ) = (h1 \<and> h2);(\<lhd> f ) " 
      by auto
 have 3: "\<turnstile> h1;(\<lhd> f \<and> #True) = h1;(\<lhd> f ) " 
      by auto
 have 4: "\<turnstile> h2;(\<lhd> f \<and> #True) = h2;(\<lhd> f ) " 
      by auto
 from 1 2 3 4 show ?thesis by auto
qed

lemma LstAndDistrD:
 "\<turnstile>  (\<diamond>(\<lhd> f \<and> g1) \<and> \<diamond>(\<lhd> f \<and> g2)) = \<diamond>(\<lhd> f \<and> g1 \<and> g2) "
proof -
 have 1: "\<turnstile> (#True;(\<lhd> f \<and> g1) \<and> #True;(\<lhd> f \<and> g2)) = (#True \<and> #True );(\<lhd> f \<and> g1 \<and> g2) " 
      using LstAndDistr by blast
 have 2: "\<turnstile> (#True \<and> #True );(\<lhd> f \<and> g1 \<and> g2) = \<diamond>(\<lhd> f \<and> g1 \<and> g2) " 
      by (simp add: sometimes_d_def)
 have 3: "\<turnstile> #True;(\<lhd> f \<and> g1) = \<diamond>(\<lhd> f \<and> g1) " 
      by (simp add: sometimes_d_def)
 have 4: "\<turnstile> #True;(\<lhd> f \<and> g2) = \<diamond>(\<lhd> f \<and> g2) " 
      by (simp add: sometimes_d_def)
 from 1 2 3 4 show ?thesis by fastforce
qed

lemma NotFstChop:
 "\<turnstile> (\<not>(\<rhd>f ;g)) = (\<not>(di (\<rhd>f)) \<or> (\<rhd>f;(\<not>g)))"
proof -
 have  1: "\<turnstile> g \<longrightarrow> #True " by auto
 hence 2: "\<turnstile> \<rhd>f;g \<longrightarrow> \<rhd>f;#True " using RightChopImpChop by blast
 hence 3: "\<turnstile> \<rhd>f;g \<longrightarrow> di(\<rhd>f)" by (simp add:di_d_def)
 hence 4: "\<turnstile> \<not>(di(\<rhd>f)) \<longrightarrow> \<not>(\<rhd>f;g) " by auto
 have  5: "\<turnstile> (\<rhd>f; (\<not>g) \<longrightarrow> \<not>(\<rhd>f;g)) = ((\<rhd>f; (\<not>g)) \<and> (\<rhd>f;g) \<longrightarrow> #False ) " by auto
 have  6: "\<turnstile> ((\<rhd>f; (\<not>g)) \<and> (\<rhd>f;g)) = \<rhd>f;(\<not>g \<and> g) " using LFstAndDistrC by blast
 have  7: "\<turnstile> \<not>(\<rhd>f;(\<not>g \<and> g)) " by (simp add: FstChopFalseEqvFalse)
 have  8: "\<turnstile> \<rhd>f; (\<not>g) \<longrightarrow> \<not>(\<rhd>f;g) " using "5" "6" "7" by fastforce
 have  9: "\<turnstile> \<not>(di(\<rhd>f)) \<or> (\<rhd>f; (\<not>g)) \<longrightarrow> \<not>(\<rhd>f;g) " using "4" "8" by fastforce
 have 10: "\<turnstile> di(\<rhd>f) \<or> \<not>(di(\<rhd>f))" by auto
 hence 11: "\<turnstile> (\<rhd>f;#True) \<or> \<not>(di(\<rhd>f)) " by (simp add: di_d_def)
 hence 12: "\<turnstile> (\<rhd>f;(g \<or> \<not>g)) \<or> \<not>(di(\<rhd>f))" by  auto
 have 13: "\<turnstile> (\<rhd>f;(g \<or> \<not>g)) = ((\<rhd>f;g) \<or> (\<rhd>f;(\<not>g)))  "  using ChopOrEqv by fastforce
 have 14: "\<turnstile> ((\<rhd>f;g) \<or> (\<rhd>f;(\<not>g))) \<or> \<not>(di(\<rhd>f))  " using "12" "13" by fastforce
 hence 15: "\<turnstile> \<not>(\<rhd>f;g) \<longrightarrow> \<not>(di(\<rhd>f)) \<or> (\<rhd>f; (\<not>g))"  by auto
 from 9 15 show ?thesis by fastforce
qed

lemma BsNotFstChop:
 "\<turnstile> bs(\<not>(\<rhd>f;g)) = (empty \<or> \<not>(di(\<rhd>f)) \<or> (\<rhd>f;bs(\<not>g)))"
proof -
 have 1: "\<turnstile> bs(\<not>(\<rhd>f;g))= (empty \<or> bi(\<not>(\<rhd>f;g));skip) " 
      by (simp add:bs_d_def)
 have 2: "\<turnstile> (empty \<or> bi(\<not>(\<rhd>f;g));skip) = (empty \<or> (\<not>(di(\<rhd>f;g)));skip)" 
      by (metis "1" NotDiEqvBiNot int_eq)
 have 3: "\<turnstile> (empty \<or> (\<not>(di(\<rhd>f;g)));skip) = (empty \<or> (\<not>((\<rhd>f;g);#True));skip)" 
      by (simp add: di_d_def)
 have 4: "\<turnstile>  (\<not>((\<rhd>f;g);#True));skip =  (\<not>(\<rhd>f;(g;#True)));skip" 
      by (metis ChopAssocB LeftChopEqvChop int_simps(15) inteq_reflection)
 hence 5: "\<turnstile> (empty \<or> (\<not>((\<rhd>f;g);#True));skip) =  (empty \<or> (\<not>(\<rhd>f;(g;#True)));skip)" 
      by auto
 have 6: "\<turnstile> (empty \<or> (\<not>(\<rhd>f;(g;#True)));skip) = (empty \<or> (\<not>(\<rhd>f;di(g)));skip)" 
      by (simp add: di_d_def)
 have 7: "\<turnstile> (empty \<or> (\<not>(\<rhd>f;di(g)));skip) = (empty \<or> \<not>(\<not>((\<not>(\<rhd>f;di(g)));skip)))  " 
      by auto
 have 8: "\<turnstile> \<not>(\<not>((\<not>(\<rhd>f;di(g)));skip)) =  (\<not>(empty \<or> (\<rhd>f;di(g));skip )) " 
      using NotNotChopSkip  by fastforce
 hence 9: "\<turnstile> (empty \<or> \<not>(\<not>((\<not>(\<rhd>f;di(g)));skip))) =  (empty \<or> \<not>(empty \<or> (\<rhd>f;di(g));skip ))" 
      by auto
 have 10: "\<turnstile> (empty \<or> \<not>(empty \<or> (\<rhd>f;di(g));skip )) = (empty \<or> (more \<and> \<not>((\<rhd>f;di(g));skip )))" 
      by (meson "6" "7" "9" NotChopSkipEqvMoreAndNotChopSkip Prop04 Prop06)
 have 11: "\<turnstile> (empty \<or> (more \<and> \<not>((\<rhd>f;di(g));skip ))) = (empty \<or> \<not>((\<rhd>f;di(g));skip ))" 
      by (auto simp add: empty_d_def)
 have 12: "\<turnstile>  (empty \<or> \<not>((\<rhd>f;di(g));skip )) =  (empty \<or> \<not>(\<rhd>f;(di(g);skip) ))" 
      using ChopAssocB "11" by fastforce
 have 13: "\<turnstile>  (\<not>(\<rhd>f;(di(g);skip) )) =  (\<not>(\<rhd>f;(ds(g)) )) " 
      using DsDi using RightChopEqvChop by fastforce
 hence 14: "\<turnstile> (empty \<or> \<not>(\<rhd>f;(di(g);skip) )) =  (empty \<or> \<not>(\<rhd>f;(ds(g)) )) " 
      by auto
 have 15: "\<turnstile> (empty \<or> \<not>(\<rhd>f;(ds(g)) )) = (empty \<or> \<not>(di (\<rhd>f)) \<or> (\<rhd>f;(\<not>(ds g))))" 
      using NotFstChop by fastforce
 have 16: "\<turnstile>  (\<rhd>f;(\<not>(ds g)))=   (\<rhd>f;(bs (\<not>g)))" 
      using NotDsEqvBsNot RightChopEqvChop by blast
 hence 17: "\<turnstile> ((empty \<or> \<not>(di (\<rhd>f))) \<or> (\<rhd>f;(\<not>(ds g)))) = ((empty \<or> \<not>(di (\<rhd>f))) \<or>  (\<rhd>f;(bs (\<not>g))))" 
      by auto
  show ?thesis
  by (metis "11" "12" "15" "16" "2" "4" DsDi 
      NotChopSkipEqvMoreAndNotChopSkip bs_d_def di_d_def inteq_reflection) 
qed

lemma FstFstChopEqvFstChopFst:
 "\<turnstile> \<rhd>(\<rhd>f;g) = \<rhd>f;\<rhd>g "
proof -
 have 1: "\<turnstile> \<rhd>(\<rhd>f;g) = ((\<rhd>f;g) \<and> bs (\<not>(\<rhd>f;g))) " 
      by (simp add: first_d_def)
 have 2: "\<turnstile> bs (\<not>(\<rhd>f;g)) = (empty \<or> \<not>(di(\<rhd>f)) \<or> (\<rhd>f;bs(\<not>g))) " 
      using BsNotFstChop by auto
 hence 3: "\<turnstile> ((\<rhd>f;g) \<and> bs (\<not>(\<rhd>f;g))) = ((\<rhd>f;g) \<and> (empty \<or> \<not>(di(\<rhd>f)) \<or> (\<rhd>f;bs(\<not>g))))" 
      by auto
 have 4: "\<turnstile> ((\<rhd>f;g) \<and> (empty \<or> \<not>(di(\<rhd>f)) \<or> (\<rhd>f;bs(\<not>g)))) =
            (((\<rhd>f;g) \<and> empty) \<or> ((\<rhd>f;g) \<and> \<not>(di(\<rhd>f))) \<or> ((\<rhd>f;g) \<and> (\<rhd>f;bs(\<not>g)))) " 
      by auto
 have 5: "\<turnstile> \<not>((\<rhd>f;g) \<and> \<not>(di(\<rhd>f)))" 
      using ChopImpDi by fastforce
 hence 6: "\<turnstile> (((\<rhd>f;g) \<and> empty) \<or> ((\<rhd>f;g) \<and> \<not>(di(\<rhd>f))) \<or> ((\<rhd>f;g) \<and> (\<rhd>f;bs(\<not>g)))) =
             (((\<rhd>f;g) \<and> empty) \<or> ((\<rhd>f;g) \<and> (\<rhd>f;bs(\<not>g))))" 
      by auto
 have 7: "\<turnstile>  ((\<rhd>f;g) \<and> (\<rhd>f;(bs(\<not>g)))) = ((\<rhd>f;(g \<and> (bs(\<not>g)))))" 
      using LFstAndDistrC by blast
 hence 8: "\<turnstile> (((\<rhd>f;g) \<and> empty) \<or> ((\<rhd>f;g) \<and> (\<rhd>f;(bs(\<not>g))))) = 
             (((\<rhd>f;g) \<and> empty) \<or> ((\<rhd>f;(g \<and> (bs(\<not>g)))))) " 
      by auto
 have 9: "\<turnstile> (((\<rhd>f;g) \<and> empty) \<or> ((\<rhd>f;(g \<and> (bs(\<not>g)))))) = (((\<rhd>f;g) \<and> empty) \<or> \<rhd>f;\<rhd>g)" 
      by (simp add: first_d_def)
 have 10: "\<turnstile> ((\<rhd>f;g) \<and> empty) = ((\<rhd>f;\<rhd>g) \<and> empty)" 
      using FstChopEmptyEqvFstChopFstEmpty by blast
 hence 11: "\<turnstile> (((\<rhd>f;g) \<and> empty) \<or> \<rhd>f;\<rhd>g) = (((\<rhd>f;\<rhd>g) \<and> empty) \<or> \<rhd>f;\<rhd>g) " 
      by auto
 have 12: "\<turnstile> (((\<rhd>f;\<rhd>g) \<and> empty) \<or> \<rhd>f;\<rhd>g) = \<rhd>f;\<rhd>g" 
      by auto
 from 1 3 4 6 8 9 11 12 show ?thesis by (metis inteq_reflection)
qed

lemma FstFixFst:
 "\<turnstile>\<rhd>(\<rhd>f) = \<rhd>f"
proof -
 have 1: "\<turnstile> \<rhd>f = (\<rhd>f);empty " using ChopEmpty by (metis int_eq) 
 hence 2: "\<turnstile>\<rhd>( \<rhd>f ) = \<rhd>((\<rhd>f);empty) " using FstEqvRule by blast
 have 3: "\<turnstile> \<rhd>((\<rhd>f);empty) = \<rhd>f;\<rhd>empty" using FstFstChopEqvFstChopFst by auto
 have 4: "\<turnstile> \<rhd>f;\<rhd>empty = \<rhd>f;empty" using FstEmpty using RightChopEqvChop by blast
 have 5: "\<turnstile> \<rhd>f;empty = \<rhd>f" using ChopEmpty by blast
 from 2 3 4 5 show ?thesis by fastforce
qed

lemma FstCSEqvEmpty:
 "\<turnstile> \<rhd>(f\<^sup>\<star>) = empty"
proof -
 have 1: "\<turnstile> \<rhd>(f\<^sup>\<star>) = \<rhd>(empty \<or> ((f \<and> more);f\<^sup>\<star>))" using ChopstarEqv FstEqvRule by blast
 from 1 show ?thesis using FstEmptyOrEqvEmpty by fastforce
qed

lemma FstIterFixFst:
 "\<turnstile> power (\<rhd> f) n = \<rhd>(power (\<rhd> f) n)"
proof 
 (induct n)
 case 0
 then show ?case 
 proof -
  have 1: "\<turnstile> power (\<rhd> f) 0 = empty" by auto
  have 2: "\<turnstile> empty = \<rhd> empty" using FstEmpty by auto
  have 3: "\<turnstile> \<rhd> empty = \<rhd>(power (\<rhd>f) 0)" by auto
  from 1 2 3 show ?thesis by auto
 qed
 next
  case (Suc n)
  then show ?case 
  proof -
   have 4: "\<turnstile> (power (\<rhd> f) (Suc n)) = (\<rhd> f) ;(power (\<rhd>f) n) " 
   by (simp)
   have 5: "\<turnstile> (\<rhd> f) ;(power (\<rhd>f) n) = (\<rhd> f) ; \<rhd> (power (\<rhd> f) n) " 
   using RightChopEqvChop Suc.hyps by blast
   have 6: "\<turnstile> (\<rhd> f) ; \<rhd> (power (\<rhd> f) n) = \<rhd>(\<rhd>f;(power (\<rhd> f) n))" 
   using FstFstChopEqvFstChopFst by fastforce
   have 7: "\<turnstile> \<rhd>(\<rhd>f;(power (\<rhd> f) n)) = \<rhd>(power (\<rhd> f) (Suc n))" 
   by simp
   from 4 5 6 7 show ?thesis by fastforce
  qed
qed 

lemma DsImpNotFst:
 "\<turnstile> ds f \<longrightarrow> (\<not>(\<rhd>f))"
proof -
 have 1: "\<turnstile> (ds f \<and> \<rhd>f) = (ds f \<and> (f \<and> bs (\<not> f)))" by (simp add: first_d_def)
 have 2: "\<turnstile> (ds f \<and> (f \<and> bs (\<not> f))) = (ds f \<and> f \<and> \<not>(ds f)) " using NotDsEqvBsNot by fastforce
 from 1 2 show ?thesis by fastforce
qed

lemma FstLenAndEqvLenAnd:
 "\<turnstile> \<rhd>(len(k) \<and> f) = (len(k) \<and> f)"
proof -
 have 1: "\<turnstile> len(k) \<and> f \<and> ds(len(k)\<and> f) \<longrightarrow> ds (len(k))  " 
      using DsAndImpElimL by fastforce
 hence 2: "\<turnstile> len(k) \<and> f \<and> ds(len(k)\<and> f) \<longrightarrow> (di(len(k)));skip " 
      using DsDi by fastforce
 hence 3: "\<turnstile> len(k) \<and> f \<and> ds(len(k)\<and> f) \<longrightarrow> ((len(k);#True));skip" 
      by (simp add: di_d_def)
 hence 4: "\<turnstile> len(k) \<and> f \<and> ds(len(k)\<and> f) \<longrightarrow> (len(k);(#True;skip))" 
      using ChopAssocB by fastforce
 hence 5: "\<turnstile> len(k) \<and> f \<and> ds(len(k)\<and> f) \<longrightarrow> (len(k);(skip;#True))"
      by (metis SkipTrueEqvTrueSkip inteq_reflection) 
 hence 6: "\<turnstile> len(k) \<and> f \<and> ds(len(k)\<and> f) \<longrightarrow> (len(k);(skip;#True)) \<and> len(k) " 
      by auto
 hence 7: "\<turnstile> len(k) \<and> f \<and> ds(len(k)\<and> f) \<longrightarrow> (len(k);(skip;#True)) \<and> len(k);empty" 
      using ChopEmpty by (metis int_eq)
 hence 8: "\<turnstile> len(k) \<and> f \<and> ds(len(k)\<and> f) \<longrightarrow> (len(k);((skip;#True) \<and> empty)) " 
      using LFixedAndDistrB1 by fastforce
 have 9: "\<turnstile> \<not>(len(k);((skip;#True) \<and> empty))" 
      by (simp add: empty_d_def more_d_def next_d_def chop_defs Valid_def) 
 have 10: "\<turnstile> len(k) \<and> f \<longrightarrow> \<not>(ds(len(k)\<and> f)) " 
      using "8" "9" by fastforce
 hence 11: "\<turnstile> len(k) \<and> f \<longrightarrow> bs (\<not>(len(k)\<and> f))" 
      using NotDsEqvBsNot by fastforce
 hence 12: "\<turnstile> len(k) \<and> f \<longrightarrow> (len(k) \<and> f) \<and> bs (\<not>(len(k)\<and> f))  " 
      by auto
 hence 13: "\<turnstile> len(k) \<and> f \<longrightarrow> \<rhd>(len(k) \<and> f)" 
      by (simp add: first_d_def)
 have 14: "\<turnstile> \<rhd>(len(k) \<and> f) \<longrightarrow> len(k) \<and> f" 
      by (auto simp add: first_d_def)
 from 13 14 show ?thesis by fastforce
qed

lemma FstAndElimL:
 "\<turnstile> \<rhd>f \<longrightarrow> f"
by (auto simp add: first_d_def)

lemma FstImpNotDiChopSkip:
 "\<turnstile> \<rhd>f \<longrightarrow> \<not>(di f;skip)"
proof -
 have 1: "\<turnstile> \<rhd>f \<longrightarrow> bs (\<not> f)" by (auto simp add: first_d_def)
 hence 2: "\<turnstile> \<rhd>f \<longrightarrow> \<not>(ds f)" using NotDsEqvBsNot by fastforce
 have 3: "\<turnstile> ds f =  di f ; skip" using DsDi by blast
 hence 4: "\<turnstile> (\<not>(ds f)) = (\<not>(di f;skip)) " by auto
 from 2 4 show ?thesis by fastforce
qed

lemma FstImpNotDiChopSkipB:
 "\<turnstile> \<rhd>f \<longrightarrow> \<not>(di (f;skip))"
proof -
 have 1: "\<turnstile> \<rhd>f \<longrightarrow> bs (\<not> f)" 
      by (auto simp add: first_d_def)
 hence 2: "\<turnstile> \<rhd>f \<longrightarrow> \<not>(ds f)" 
      using NotDsEqvBsNot by fastforce
 have 3: "\<turnstile> ds f =  di f ; skip" 
      using DsDi by blast
 have 4: "\<turnstile> di f ; skip = (f;#True);skip" 
      by (simp add: di_d_def)
 have 5: "\<turnstile>(f;#True);skip = f;(#True;skip) " 
      using ChopAssocB by blast
 have 6: "\<turnstile> f;(#True;skip) = f;(skip;#True)" 
      using SkipTrueEqvTrueSkip using TrueChopSkipEqvSkipChopTrue RightChopEqvChop by blast
 have 7: "\<turnstile> f;(skip;#True) = (f;skip);#True" 
      using ChopAssoc by blast
 have 8: "\<turnstile> (f;skip);#True = di(f;skip)" 
      by (simp add: di_d_def)
 have 9: "\<turnstile>  (\<not>(ds f)) = (\<not>(di(f;skip))) " 
      using "3" "4" "5" "6" "7" "8" by fastforce
 from 2 9 show ?thesis by fastforce
qed

lemma FstImpDiEqv:
 "\<turnstile> \<rhd>f \<longrightarrow> (di f = f)"
proof -
 have 1: "\<turnstile> \<rhd>f  \<longrightarrow> \<not>(di f;skip)" using FstImpNotDiChopSkip by blast
 have 2: "\<turnstile> di f \<longrightarrow> f \<or> (di f;skip) " using DiEqvOrDiChopSkipB by fastforce
 have 3: "\<turnstile> \<rhd>f  \<and> di f \<longrightarrow> (f \<or> (di f;skip)) \<and> \<not>(di f;skip) " using "1" "2" by fastforce
 have 4: "\<turnstile> ((f \<or> (di f;skip)) \<and> \<not>(di f;skip)) = (f \<and> \<not>(di f;skip))" by auto 
 have 5: "\<turnstile>  \<rhd>f  \<and> di f \<longrightarrow> f \<and> \<not>(di f;skip) " using "3" "4" by fastforce
 hence 6: "\<turnstile>  \<rhd>f  \<and> di f \<longrightarrow> f " by fastforce
 hence 7: "\<turnstile> \<rhd>f \<longrightarrow> (di f \<longrightarrow> f)" using FstAndElimL by fastforce
 have 8: "\<turnstile> f  \<longrightarrow> di f " using DiIntro by auto
 hence 9: "\<turnstile> \<rhd>f \<longrightarrow> (f  \<longrightarrow> (di f))" by auto
 from 7 9 show ?thesis by fastforce
qed

lemma FstAndDiFstAndEqvFstAnd:
 "\<turnstile> (\<rhd>f \<and>  di(\<rhd>f \<and> g)) = (\<rhd>f \<and> g) "
proof -
 have 1: "\<turnstile> \<rhd>f \<and>  di(\<rhd>f \<and> g) \<longrightarrow> \<rhd> f " 
      by auto
 have 2: "\<turnstile> \<rhd>f \<and>  di(\<rhd>f \<and> g) \<longrightarrow> di(\<rhd>f \<and> g)" 
      by auto
 have 3: "\<turnstile> di(\<rhd>f \<and> g) = ((\<rhd>f \<and> g) \<or> di((\<rhd>f \<and> g);skip))" 
      using DiEqvOrDiChopSkipA by blast
 have 4: "\<turnstile> di((\<rhd>f \<and> g);skip) = ((\<rhd>f \<and> g);skip);#True " 
      by (simp add: di_d_def)
 have 5: "\<turnstile> \<rhd>f \<and>  di(\<rhd>f \<and> g) \<longrightarrow> (\<rhd>f \<and> g) \<or> ((\<rhd>f \<and> g);skip);#True " 
      using "2" "3" "4" by fastforce
 have 6: "\<turnstile> \<rhd>f \<and> g \<longrightarrow> f" 
      using FstAndElimL by fastforce
 hence 7: "\<turnstile> ((\<rhd>f \<and> g);skip);#True  \<longrightarrow> (f;skip);#True" 
      by (simp add: LeftChopImpChop)
 hence 8: "\<turnstile> ((\<rhd>f \<and> g);skip);#True  \<longrightarrow> di(f;skip)" 
      by (simp add: di_d_def)
 have  9: "\<turnstile> \<rhd>f \<longrightarrow> \<not>(di(f;skip))" 
      using FstImpNotDiChopSkipB by  blast
 have 10: "\<turnstile> \<rhd>f \<and>  di(\<rhd>f \<and> g) \<longrightarrow> ((\<rhd>f \<and> g) \<or> di(f;skip)) " 
      using "5" "8"  by fastforce
 have 11: "\<turnstile> \<rhd>f \<and>  di(\<rhd>f \<and> g) \<longrightarrow> \<not>(di(f;skip)) \<and> ((\<rhd>f \<and> g) \<or> di(f;skip))" 
      using "9" "10" "1" by fastforce
 have 12: "\<turnstile> (\<not>(di(f;skip)) \<and> ((\<rhd>f \<and> g) \<or> di(f;skip))) = (\<not>(di(f;skip)) \<and> ((\<rhd>f \<and> g))) " 
      by auto
 have 13: "\<turnstile> \<rhd>f \<and>  di(\<rhd>f \<and> g) \<longrightarrow> (\<rhd>f \<and> g)"  
      using "11" "12" by auto
 have 14: "\<turnstile> (\<rhd>f \<and> g) \<longrightarrow> \<rhd>f" 
      by auto
 hence 15: "\<turnstile> (\<rhd>f \<and> g) \<longrightarrow> di(\<rhd>f \<and> g ) " 
      using DiIntro by auto
 have 16: "\<turnstile> (\<rhd>f \<and> g) \<longrightarrow> \<rhd>f \<and> di(\<rhd>f \<and> g)"  
      using "14" "15" by auto
 from 13 16 show ?thesis by fastforce
qed

lemma FstAndDiImpBsNotAndDi:
 "\<turnstile> (\<rhd>f \<and> di g) \<longrightarrow> (bs (\<not>(di f \<and> g))) "
proof -
 have 1: "\<turnstile> (\<rhd>f \<and> di g) \<and> \<not>(bs (\<not>(di f \<and> g))) \<longrightarrow> ds(di f \<and> g)  " 
      by (auto simp add: ds_d_def)
 hence 2: "\<turnstile> (\<rhd>f \<and> di g) \<and> \<not>(bs (\<not>(di f \<and> g))) \<longrightarrow> ds(di f)  " 
      using DsAndImp by fastforce
 hence 3: "\<turnstile> (\<rhd>f \<and> di g) \<and> \<not>(bs (\<not>(di f \<and> g))) \<longrightarrow> di(di f);skip" 
      using DsDi by fastforce
 hence 4: "\<turnstile> (\<rhd>f \<and> di g) \<and> \<not>(bs (\<not>(di f \<and> g))) \<longrightarrow> di f;skip" 
      using DiEqvDiDi by (metis int_eq)
 hence 5: "\<turnstile> (\<rhd>f \<and> di g) \<and> \<not>(bs (\<not>(di f \<and> g))) \<longrightarrow> ds f" 
      using DsDi by fastforce
 hence 6: "\<turnstile>  (\<rhd>f \<and> di g) \<and> \<not>(bs (\<not>(di f \<and> g))) \<longrightarrow>\<not> (\<rhd> f) " 
      using DsImpNotFst by fastforce
 from 6 show ?thesis by auto
qed
 
lemma FstFstOrEqvFstOrL:
 "\<turnstile> \<rhd>(\<rhd>f \<or> g) = \<rhd>(f \<or> g)"
proof -
 have 1: "\<turnstile> \<rhd>(f \<or> g) = ((f \<or> g) \<and> bs (\<not>(f \<or> g)))" 
      by (simp add: first_d_def)
 have 2: "\<turnstile>  (\<not>(f \<or> g)) =  (\<not>f \<and> \<not> g)" 
      by auto
 hence 3: "\<turnstile>  bs(\<not>(f \<or> g)) =  bs (\<not>f \<and> \<not> g)" 
      using BsEqvRule by blast
 have 4: "\<turnstile> bs (\<not>f \<and> \<not> g) = (bs (\<not> f) \<and> bs (\<not> g))" 
      using BsAndEqv by fastforce
 hence 5: "\<turnstile> ((f \<or> g) \<and> bs(\<not>(f \<or> g))) =  ((f \<or> g) \<and> bs (\<not> f) \<and> bs (\<not> g)) " 
      using "4" "3" by fastforce
 have 6: "\<turnstile> ((f \<or> g) \<and> bs (\<not> f) \<and> bs (\<not> g)) =
            (((f \<and> bs (\<not> f)) \<or> (g \<and> bs (\<not> f))) \<and> bs (\<not> g))" 
      by auto
 have 7: "\<turnstile> (((f \<and> bs (\<not> f)) \<or> (g \<and> bs (\<not> f))) \<and> bs (\<not> g)) =
            ((\<rhd>f \<or> (g \<and> bs (\<not> f))) \<and> bs (\<not> g)) " 
      by (simp add: first_d_def)
 have 8: "\<turnstile> ((\<rhd>f \<or> (g \<and> bs (\<not> f))) \<and> bs (\<not> g)) =
            (((\<rhd>f \<or> g) \<and> (\<rhd>f \<or> bs (\<not>f))) \<and> bs (\<not> g)) " 
      by auto
 have 9: "\<turnstile> (((\<rhd>f \<or> g) \<and> (\<rhd>f \<or> bs (\<not>f))) \<and> bs (\<not> g)) =
            (((\<rhd>f \<or> g) \<and> ((f \<and> bs (\<not> f)) \<or> bs (\<not>f))) \<and> bs (\<not> g))" 
      by (simp add: first_d_def)
 have 10: "\<turnstile> (((\<rhd>f \<or> g) \<and> ((f \<and> bs (\<not> f)) \<or> bs (\<not>f))) \<and> bs (\<not> g)) =
             ((\<rhd>f \<or> g) \<and> bs (\<not> f) \<and> bs (\<not> g))" 
      by auto
 have 11: "\<turnstile> ((\<rhd>f \<or> g) \<and> bs (\<not> f) \<and> bs (\<not> g)) = 
              ((\<rhd>f \<or> g) \<and> bs(\<not>(\<rhd>f))\<and> bs (\<not> g))  " 
      using BsNotFstEqvBsNot by fastforce
 have 12: "\<turnstile> ((\<rhd>f \<or> g) \<and> bs(\<not>(\<rhd>f))\<and> bs (\<not> g)) =
             ((\<rhd>f \<or> g) \<and> bs (\<not>(\<rhd>f) \<and> \<not> g))" 
      using BsAndEqv by fastforce
 have 13: "\<turnstile>  (\<not>(\<rhd>f) \<and> \<not> g) = (\<not>(\<rhd>f \<or> g))" 
      by auto
 hence 14: "\<turnstile> bs (\<not>(\<rhd>f) \<and> \<not> g) = bs (\<not>(\<rhd>f \<or> g))" 
      using BsEqvRule by blast
 hence 15: "\<turnstile> ((\<rhd>f \<or> g) \<and> bs (\<not>(\<rhd>f) \<and> \<not> g)) = ((\<rhd>f \<or> g) \<and> bs (\<not>(\<rhd>f \<or> g)))" 
      by auto
 have 16: " \<turnstile> ((\<rhd>f \<or> g) \<and> bs (\<not>(\<rhd>f \<or> g))) = \<rhd>(\<rhd>f \<or> g)" 
      by (simp add: first_d_def)
 from 16 15 12 11 10 9 8 7 6 5 1 show ?thesis by (metis int_eq)
qed

lemma FstFstOrEqvFstOrR:
 "\<turnstile> \<rhd>(f \<or> \<rhd>g)= \<rhd>(f \<or> g)"
proof -
 have 1: "\<turnstile> (f \<or> \<rhd>g)= (\<rhd>g \<or> f)" by auto
 hence 2: "\<turnstile> \<rhd> (f \<or> \<rhd>g)= \<rhd>(\<rhd>g \<or> f) " using FstEqvRule by blast
 have 3: "\<turnstile> \<rhd>(\<rhd>g \<or> f) = \<rhd>(g \<or> f)" using FstFstOrEqvFstOrL by blast
 have 4: "\<turnstile> (g \<or> f) = (f \<or> g)" by auto
 hence 5: "\<turnstile> \<rhd>(g \<or> f) = \<rhd>(f \<or> g)"  using FstEqvRule by blast
 from 2 3 5 show ?thesis by fastforce
qed

lemma FstFstOrEqvFstOr:
 "\<turnstile> \<rhd>(\<rhd>f \<or> \<rhd>g) = \<rhd>(f \<or> g)"
proof -
 have 1: "\<turnstile>  \<rhd>(\<rhd>f \<or> \<rhd>g) = \<rhd>(f \<or> \<rhd>g)" using FstFstOrEqvFstOrL by blast
 have 2: "\<turnstile> \<rhd>(f \<or> \<rhd>g) = \<rhd>(f \<or> g)" using FstFstOrEqvFstOrR by blast
 from 1 2 show ?thesis by fastforce
qed

lemma FstLenEqvLen:
 "\<turnstile> \<rhd>( len(k)) = len(k) "
proof -
 have 1: "\<turnstile> \<rhd>( len(k) \<and> #True) = (len(k) \<and> #True)" using FstLenAndEqvLenAnd by blast
 have 2: "\<turnstile> (len(k) \<and> #True) = len(k)" by auto
 hence 3: "\<turnstile> \<rhd>( len(k) \<and> #True) = \<rhd>(len(k)) " using FstEqvRule by blast
 from 1 2 3 show ?thesis by auto
qed

lemma FstSkip:
 "\<turnstile> \<rhd> skip = skip"
proof -
 have 1: "\<turnstile> skip = len(1)  " using LenOneEqvSkip by fastforce
 hence 2: "\<turnstile> \<rhd>skip = \<rhd> (len(1))" using FstEqvRule by blast
 have 3: "\<turnstile> \<rhd>( len(1)) = len(1)" using FstLenEqvLen by blast
 from 1 2 3 show ?thesis using LenOneEqvSkip by fastforce
qed

lemma HaltStateEqvFstFinState:
 "\<turnstile> halt (init w) = \<rhd> (fin (init w))"
proof -
 have 1: "\<turnstile> halt (init w) = \<box>(empty = (init w))" by (simp add: halt_d_def)
 have 21: "\<turnstile> (empty = (init w)) = (((empty \<longrightarrow> (init w)) \<and> ((init w) \<longrightarrow> empty)))"  
      by auto
 hence 2: "\<turnstile> \<box>(empty = (init w)) = (\<box>((empty \<longrightarrow> (init w)) \<and> ((init w) \<longrightarrow> empty)))"
      by (simp add: BoxEqvBox)
 have 3: "\<turnstile> (\<box>((empty \<longrightarrow> (init w)) \<and> ((init w) \<longrightarrow> empty))) = 
            (\<box>((empty \<longrightarrow> (init w))) \<and> \<box>((init w) \<longrightarrow> empty))" 
      by (metis "21" BoxAndBoxEqvBoxRule int_eq)
 have 4: "\<turnstile> ((init w) \<longrightarrow> empty) = (more \<longrightarrow> \<not>(init w))" 
      by (auto simp add: empty_d_def)
 hence 5: "\<turnstile> \<box> ((init w) \<longrightarrow> empty) = \<box> (more \<longrightarrow> \<not>(init w))" using BoxEqvBox by blast
 have 6: "\<turnstile> \<box> (more \<longrightarrow> \<not>(init w)) = bs(\<not>(fin(init w)))" using BoxMoreStateEqvBsFinState by blast
 have 7: "\<turnstile> \<box>((empty \<longrightarrow> (init w))) = fin(init w)" by (simp add: fin_d_def)
 have 8: "\<turnstile>  (\<box>((empty \<longrightarrow> (init w))) \<and> \<box>((init w) \<longrightarrow> empty)) =
             (fin(init w) \<and> bs(\<not>(fin(init w))))" using "5" "6" "7" by fastforce
 from 1 2 3 8 show ?thesis by (metis first_d_def inteq_reflection)
qed

lemma FstLenEqvLenFst:
 "\<turnstile> \<rhd>(len k ; f) = len k ; \<rhd> f"
proof -
  have 1: "\<turnstile> len k ; f = \<rhd>(len k) ; f " using FstLenEqvLen LeftChopEqvChop by fastforce
  have 2: "\<turnstile> \<rhd>(len k ; f) = \<rhd> (\<rhd>(len k) ; f) " using 1 FstEqvRule by blast
  have 3: "\<turnstile> \<rhd> (\<rhd>(len k) ; f) = \<rhd>(len k) ; \<rhd>f" using FstFstChopEqvFstChopFst by blast
  have 4: "\<turnstile> \<rhd>(len k) ; \<rhd>f = len k ; \<rhd>f" using FstLenEqvLen LeftChopEqvChop by fastforce
  from 2 3 4 show ?thesis by fastforce
qed

lemma FstNextEqvNextFst:
 "\<turnstile> \<rhd>(\<circle> f) = \<circle>( \<rhd> f)"
proof -
  have 1: "\<turnstile> \<rhd>(\<circle> f) = \<rhd>(skip ; f) " using FstEqvRule by (simp add: next_d_def)
  have 2: "\<turnstile> skip ; f = \<rhd>skip ; f " using FstSkip using LeftChopEqvChop by fastforce
  have 3: "\<turnstile> \<rhd>(skip ; f) = \<rhd> (\<rhd>skip ; f) " using 2 FstEqvRule LeftChopEqvChop by blast
  have 4: "\<turnstile> \<rhd> (\<rhd>skip ; f) = \<rhd>skip ; \<rhd>f " using 3 FstFstChopEqvFstChopFst by blast
  have 5: "\<turnstile> \<rhd>skip ; \<rhd>f = skip ; \<rhd>f " using 4 FstSkip LeftChopEqvChop by blast
  have 6: "\<turnstile> skip ; \<rhd>f  =  \<circle>( \<rhd> f)" by (simp add: next_d_def)
  from 1 2 3 4 5 6 show ?thesis by fastforce
qed

lemma FstDiamondStateEqvHalt:
 "\<turnstile> \<rhd> (\<diamond> (init w)) = halt (init w)"
proof -
  have 1: "\<turnstile> \<diamond> (init w) = \<diamond> ((init w) \<and> #True)" by simp
  have 2: "\<turnstile> fin (init w) ; #True = \<diamond> ((init w) \<and> #True)" using 1 FinChopEqvDiamond by blast
  have 3: "\<turnstile> fin (init w) ; #True = di (fin (init w))" by (simp add: di_d_def)
  have 4: "\<turnstile> (\<diamond> (init w)) = (di (fin (init w)))" using 1 2 3 by fastforce
  have 5: "\<turnstile> \<rhd> (\<diamond> (init w)) = \<rhd> (di (fin (init w)))" using 4 FstEqvRule by blast
  hence 6: "\<turnstile> \<rhd> (\<diamond> (init w)) = \<rhd> (fin (init w))" using FstDiEqvFst by fastforce
  hence 7: "\<turnstile> \<rhd> (\<diamond> (init w)) = halt (init w)" using HaltStateEqvFstFinState by fastforce
  from 7 show ?thesis by simp
qed

lemma FstBoxStateEqvStateAndEmpty:
 "\<turnstile> \<rhd> (\<box> (init w)) = ((init w) \<and> empty)"
proof -
  have 1: "\<turnstile> ((init w) \<and> (\<box> (init w))\<^sup>\<star>) = \<box> (init w)" 
  using BoxCSEqvBox by blast
  have 2: "\<turnstile> \<box> (init w) = ((init w) \<and> (\<box> (init w))\<^sup>\<star>)" 
  using 1 by auto
  hence 3: "\<turnstile> \<box> (init w) = ((init w) \<and> (\<box> (init w))\<^sup>\<star>)" 
  by blast
  have 4: "\<turnstile> ((init w) \<and> empty) ; (\<box> (init w))\<^sup>\<star> = ((init w) \<and> (\<box> (init w))\<^sup>\<star>)" 
  using StateAndEmptyChop by blast
  have 5: "\<turnstile> ((init w) \<and> (\<box> (init w))\<^sup>\<star>)  =  ((init w) \<and> empty) ; (\<box> (init w))\<^sup>\<star>" 
  using 4 by fastforce
  have 6: "\<turnstile> \<box> (init w) = ((init w) \<and> empty) ; (\<box> (init w))\<^sup>\<star> " 
  using 3 5 by fastforce
  have 7: "\<turnstile> ((init w) \<and> empty) ; (\<box> (init w))\<^sup>\<star> = \<rhd> (init w) ; (\<box> (init w))\<^sup>\<star> " 
  using FstState by (metis AndChopCommute int_eq)
  have 8: "\<turnstile> \<box> (init w) = \<rhd> (init w) ; (\<box> (init w))\<^sup>\<star> " 
  using 6 7 by fastforce
  have 9: "\<turnstile> \<rhd> (\<box> (init w)) = \<rhd> (\<rhd> (init w) ; (\<box> (init w))\<^sup>\<star>) " 
  using 8 FstEqvRule by blast
  have 10: "\<turnstile> \<rhd> (\<rhd> (init w) ; (\<box> (init w))\<^sup>\<star>) = \<rhd> (init w) ;  \<rhd> ((\<box> (init w))\<^sup>\<star>) " 
  using FstFstChopEqvFstChopFst by blast
  have 11: "\<turnstile> \<rhd> (init w) ;  \<rhd> ((\<box> (init w))\<^sup>\<star>) = \<rhd> (init w) ; empty" 
  using RightChopEqvChop FstCSEqvEmpty by blast
  have 12: "\<turnstile> \<rhd> (init w) ; empty = \<rhd> (init w)" 
  using RightChopEqvChop ChopEmpty by blast
  have 13: "\<turnstile> \<rhd> (init w) = ((init w) \<and> empty)" 
  using FstState by fastforce
  from 9 10 11 12 13 show ?thesis by fastforce
qed

lemma FstAndFstStarEqvFst:
 "\<turnstile> (\<rhd>f \<and> (\<rhd>f)\<^sup>\<star>) = \<rhd>f "
proof -
  have 1: "\<turnstile> (\<rhd>f)\<^sup>\<star> = (empty \<or> (\<rhd>f);(\<rhd>f)\<^sup>\<star>) " 
       using CSEqvOrChopCS by fastforce
  have 2: "\<turnstile> ((\<rhd>f)\<^sup>\<star> \<and> \<rhd>f) = ((empty \<or> (\<rhd>f);(\<rhd>f)\<^sup>\<star>) \<and> \<rhd>f)" 
       using 1 by fastforce
  have 3: "\<turnstile> ((empty \<or> (\<rhd>f);(\<rhd>f)\<^sup>\<star>) \<and> \<rhd>f) = ((empty \<and> \<rhd>f) \<or> ((\<rhd>f);(\<rhd>f)\<^sup>\<star> \<and> \<rhd>f))" 
       by auto
  have 4: "\<turnstile> ((\<rhd>f)\<^sup>\<star> \<and> \<rhd>f) = ((empty \<and> \<rhd>f) \<or> ((\<rhd>f);(\<rhd>f)\<^sup>\<star> \<and> \<rhd>f))" 
       using 2 3 by fastforce
  have 5: "\<turnstile> ((\<rhd>f);(\<rhd>f)\<^sup>\<star> \<and> \<rhd>f) = ((\<rhd>f);(\<rhd>f)\<^sup>\<star> \<and> \<rhd>f;empty)" 
       using ChopEmpty by (metis inteq_reflection)
  have 6: "\<turnstile> ((\<rhd>f);(\<rhd>f)\<^sup>\<star> \<and> \<rhd>f;empty) = (\<rhd>f);((\<rhd>f)\<^sup>\<star> \<and> empty)"  
       using LFstAndDistrC by blast
  have 7: "\<turnstile> ((\<rhd>f)\<^sup>\<star> \<and> empty) = empty" 
       using EmptyImpCS by fastforce
  have 8: "\<turnstile> (\<rhd>f);((\<rhd>f)\<^sup>\<star> \<and> empty) = \<rhd>f" 
       using 7 ChopEmpty by (metis inteq_reflection)
  have 9: "\<turnstile> ((\<rhd>f);(\<rhd>f)\<^sup>\<star> \<and> \<rhd>f) = \<rhd>f" 
       using 5 6 8 by fastforce
  have 10: "\<turnstile> ((\<rhd>f)\<^sup>\<star> \<and> \<rhd>f) = ((empty \<and> \<rhd>f) \<or> \<rhd>f)" 
       using 4 9 by fastforce
  have 11: "\<turnstile> ((empty \<and> \<rhd>f) \<or> \<rhd>f) = \<rhd>f" 
       by auto
  have 12:  "\<turnstile> ((\<rhd>f)\<^sup>\<star> \<and> \<rhd>f) = \<rhd>f" 
       using 10 11 by fastforce
  from 12 show ?thesis by auto
qed

lemma HaltStateEqvFstHaltState:
   "\<turnstile> halt(init(w)) = \<rhd>(halt(init(w)))"
proof -
  have 1:  "\<turnstile> halt (init w) = \<rhd> (fin (init w))"
       by (simp add: HaltStateEqvFstFinState)
  have 2:  "\<turnstile> \<rhd> (fin (init w)) = \<rhd> (\<rhd> (fin (init w)))"
    using FstEqvRule FstFixFst by fastforce
  have 3:  "\<turnstile> \<rhd> (\<rhd> (fin (init w))) = \<rhd>(halt(init(w)))"
      using FstEqvRule HaltStateEqvFstFinState by fastforce
    from 1 2 3 show ?thesis by fastforce
qed


lemma DiHaltAndDiHaltAndEqvDiHaltAndAnd:
   "\<turnstile> (di(halt (init w) \<and> f) \<and> di(halt (init w) \<and> g)) = di(halt (init w) \<and> f \<and> g)"
proof -
  have 1: "\<turnstile> (di(halt (init w) \<and> f) \<and> di(halt (init w) \<and> g)) = 
             (di(\<rhd>(fin (init w)) \<and> f) \<and> di (\<rhd>(fin (init w)) \<and> g))" 
       using HaltStateEqvFstFinState by (metis LFstAndDistrD inteq_reflection)
  have 2: "\<turnstile> (di(\<rhd>(fin (init w)) \<and> f) \<and> di(\<rhd>(fin (init w)) \<and> g)) = 
             di(\<rhd>(fin (init w)) \<and> f \<and> g)" 
       using LFstAndDistrD by fastforce
  have 3: "\<turnstile> di(\<rhd>(fin (init w)) \<and> f \<and> g)  = di(halt (init w) \<and> f \<and> g)" 
       using HaltStateEqvFstFinState  by (metis DiEqvDi int_eq lift_and_com)  
  from 1 2 3 show ?thesis using int_eq by metis
qed

(* This doesn't work and is not a theorem.
      see    counter_ex_lhs and counter_ex_rhs                                                    
lemma FstAndFstChopEqvFstChopAndFstChop:
   "\<turnstile> ((\<rhd>f \<and> \<rhd>g) ; h) = ((\<rhd>f ; h) \<and> (\<rhd>g ; h))"
*)

lemma counter_ex_lhs:
 "\<turnstile> ((\<rhd>(len(5)) \<and> \<rhd>(len(2))) ; (len(5) \<or> len(2))) = #False "
proof -
 have 1: "\<turnstile> ((\<rhd>(len(5)) \<and> \<rhd>(len(2))) ; (len(5) \<or> len(2))) =
            (len(5) \<and> len(2)); (len(5) \<or> len(2))" 
       by (metis FstLenAndEqvLenAnd FstLenEqvLen LeftChopEqvChop inteq_reflection)
 have 2: "\<turnstile> (len(5) \<and> len(2)) = #False"
      by (simp add: Valid_def len_defs) 
 have 3: "\<turnstile> ((len(5) \<and> len(2)); (len(5) \<or> len(2))) = (#False;(len(5) \<or> len(2))) "  
      by (simp add: "2" LeftChopEqvChop)  
 have 4: "\<turnstile> (#False;(len(5) \<or> len(2))) = #False" 
      by (simp add: Valid_def chop_defs)
 from 1 3 4 show ?thesis by fastforce
qed

lemma counter_ex_rhs:
 "\<turnstile> ((\<rhd> (len(5)) ; (len(5) \<or> len(2))) \<and> (\<rhd> (len(2)) ; (len(5) \<or> len(2)))) = len(7)"  
proof -
 have 1: "\<turnstile> (\<rhd> (len(5)) ; (len(5) \<or> len(2))) = 
            len(5);(len(5) \<or> len(2)) " 
      using FstLenEqvLen LeftChopEqvChop by blast
 have 2: "\<turnstile> (\<rhd> (len(2)) ; (len(5) \<or> len(2))) =
            len(2) ;(len(5) \<or> len(2))" 
      using FstLenEqvLen LeftChopEqvChop by blast
 have 3: "\<turnstile> len(5);(len(5) \<or> len(2)) =
            ((len(5);len(5)) \<or> (len(5);len(2)))" 
      by (simp add: ChopOrEqv)
 have 4: "\<turnstile> ((len(5);len(5)) \<or> (len(5);len(2))) =
            (len(10) \<or> len(7))" 
      using LenEqvLenChopLen inteq_reflection by fastforce
 have 5: "\<turnstile> len(2) ;(len(5) \<or> len(2)) =
            ((len(2);len(5)) \<or> (len(2);len(2)))" 
      by (simp add: ChopOrEqv)
 have 6: "\<turnstile> ((len(2);len(5)) \<or> (len(2);len(2))) =
            (len(7) \<or> len(4))" 
      using LenEqvLenChopLen inteq_reflection by fastforce
 have 7: "\<turnstile> ((len(10) \<or> len(7)) \<and> (len(7) \<or> len(4))) = 
            ((len(7) \<or> len(10)) \<and> (len(7) \<or> len(4)))" 
      by fastforce
 have 8: "\<turnstile> ((len(7) \<or> len(10)) \<and> (len(7) \<or> len(4))) = 
             (len(7) \<or> (len(10) \<and> len(4))) "  
      by fastforce
 have 9: "\<turnstile> (len(10) \<and> len(4)) = #False" 
      by (simp add: Valid_def len_defs) 
 have 10 : "\<turnstile> (len(7) \<or> (len(10) \<and> len(4))) = len(7)" 
      using "9" by auto
 have 11: "\<turnstile> ((\<rhd> (len(5)) ; (len(5) \<or> len(2))) \<and> (\<rhd> (len(2)) ; (len(5) \<or> len(2)))) =
             (len(5);(len(5) \<or> len(2)) \<and> len(2) ;(len(5) \<or> len(2)))" 
      using "1" "2" by fastforce
 have 12: "\<turnstile> (len(5);(len(5) \<or> len(2)) \<and> len(2) ;(len(5) \<or> len(2))) = len(7) "
      by (metis "10" "4" "6" "7" "8" ChopOrEqv inteq_reflection)
 from 11 12 show ?thesis by fastforce
qed



end
