(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)
section \<open>Filter on Intervals\<close>

theory IFilter
 imports
   Interval
begin

(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 z3 e spass vampire " ]

*)


text \<open>
The ifilter operator on intervals is defined. The definition of ifilter is slightly more complicated than 
the one for lists as an interval has at least one state and one needs to ensure that the ifilter
operator always returns an interval. The lemmas involving the ifilter on intervals are similar to those for the
filter operator on lists only a bit more complicated.  
\<close>

subsection \<open>Definitions\<close>

definition opfx :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> bool"
 where "opfx xs ys = (\<exists>zs. ys = xs \<ominus> zs \<or> ys = xs)"

definition sopfx :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> bool"
where "sopfx xs ys \<longleftrightarrow> opfx xs ys \<and> xs \<noteq> ys"

interpretation opfx_order: order opfx sopfx
proof  standard
 show "\<And>x y. sopfx x y = (opfx x y \<and> \<not> opfx y x)"
 by (metis add_leD1 add_le_same_cancel1 ilen_iapp le_add1 not_one_le_zero opfx_def 
     sopfx_def)
 show "\<And>x. opfx x x"
   by (auto simp add: opfx_def sopfx_def)
 show "\<And>x y z. opfx x y \<Longrightarrow> opfx y z \<Longrightarrow> opfx x z"  
  by (auto simp add: opfx_def sopfx_def)
 show "\<And>x y. opfx x y \<Longrightarrow> opfx y x \<Longrightarrow> x = y "
 by (metis add_leD1 add_le_same_cancel1 ilen_iapp le_add1 not_one_le_zero opfx_def) 
qed

definition osfx :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> bool"
where "osfx xs ys = (\<exists>zs. ys = zs \<ominus> xs \<or> ys =xs)"


definition sosfx :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> bool"
where "sosfx xs ys \<longleftrightarrow> osfx xs ys \<and> xs \<noteq> ys "

interpretation osfx_order: order osfx sosfx
proof standard
 show "\<And>x y. sosfx x y = (osfx x y \<and> \<not> osfx y x)"
 by (metis irev_iapp irev_irev_ident opfx_def opfx_order.dual_order.antisym 
     osfx_def sosfx_def) 
 show "\<And>x. osfx x x" 
  by (auto simp add: osfx_def sosfx_def)
 show "\<And>x y z. osfx x y \<Longrightarrow> osfx y z \<Longrightarrow> osfx x z"
   by (auto simp add: osfx_def sosfx_def)
      (metis iapp_assoc)
 show "\<And>x y. osfx x y \<Longrightarrow> osfx y x \<Longrightarrow> x = y "
 using \<open>\<And>y x. sosfx x y = (osfx x y \<and> \<not> osfx y x)\<close> sosfx_def by blast
qed

primrec idistinct :: "'a interval \<Rightarrow> bool"
 where "idistinct \<langle>x\<rangle> \<longleftrightarrow> True"
      | "idistinct (x\<odot>xs) \<longleftrightarrow> x \<notin> iset xs \<and> idistinct xs"

primrec iremdups :: "'a interval \<Rightarrow> 'a interval"
 where "iremdups \<langle>x\<rangle> = \<langle>x\<rangle>"
      | "iremdups (x\<odot>xs) = (if x \<in> iset xs then iremdups xs else x \<odot> iremdups xs)"

primrec ifilter:: "('a \<Rightarrow> bool) \<Rightarrow> 'a interval \<Rightarrow> 'a interval" 
 where  "ifilter P \<langle>x\<rangle> = \<langle>x\<rangle>"
      | "ifilter P (x\<odot>xs) = (if (\<exists>y\<in>iset xs. P y) then
                              (if P x then x \<odot> ifilter P xs else ifilter P xs)
                             else \<langle>x\<rangle>)"

primrec nfilter:: "('a \<Rightarrow> bool) \<Rightarrow> 'a interval \<Rightarrow> nat \<Rightarrow> nat interval" 
 where  "nfilter P \<langle>x\<rangle> n     = \<langle>n\<rangle>"
      | "nfilter P (x\<odot>xs) n = (if (\<exists>y\<in>iset xs. P y) then
                                (if P x then n \<odot> (nfilter P xs (Suc n)) 
                                        else nfilter P xs (Suc n))
                               else \<langle>n\<rangle>)"


primrec prefixes :: "'a interval \<Rightarrow> 'a interval interval"
 where
    "prefixes \<langle>x\<rangle> = \<langle>\<langle>x\<rangle>\<rangle>"
  | "prefixes (x\<odot>xs) = \<langle>x\<rangle> \<odot> (imap ((\<odot>) x) (prefixes xs))"


primrec suffixes :: "'a interval \<Rightarrow> 'a interval interval" 
where
  "suffixes \<langle>x\<rangle> = \<langle>\<langle>x\<rangle>\<rangle>"
| "suffixes (x\<odot>xs) = (x\<odot>xs) \<odot> (suffixes xs)"


subsection \<open>Lemmas\<close>

subsubsection \<open>opfx and sopfx\<close>

lemma opfxI [intro?]:
 assumes "ys = xs \<ominus> zs \<or> ys = xs"
 shows   "  opfx xs ys"
using assms unfolding opfx_def by blast

lemma opfxE [elim?]:
 assumes "opfx xs ys"
 obtains zs where "ys = xs \<ominus> zs \<or> ys = xs"
 using assms unfolding opfx_def by blast

lemma sopfxI' [intro?]:
 " ys = xs \<ominus> (zs) \<Longrightarrow> sopfx xs ys"
unfolding sopfx_def opfx_def 
by (metis add_le_same_cancel1 ilen_iapp le_add1 not_one_le_zero)

lemma sopfxE' [elim?]:
 assumes "sopfx xs ys"
 obtains zs where "ys = xs \<ominus> zs"
 using assms  unfolding sopfx_def opfx_def by blast

lemma opfx_INil [simp]:
 "opfx \<langle>ifirst xs\<rangle> xs"
unfolding opfx_def
by (metis iapp_INil ifirst_iapp2 interval.exhaust inth.simps(1))  


lemma opfx_snoc [simp]:
 "opfx xs (ys \<ominus> \<langle>y\<rangle>) \<longleftrightarrow> xs = ys\<ominus>\<langle>y\<rangle> \<or> opfx xs ys"
unfolding opfx_def 
by (metis iapp_eq_iapp_conv2 iapp_not_INil) 

lemma ICons_pfx_ICons [simp]:
 " opfx (x\<odot>xs) (y\<odot>ys) = (x =y \<and> opfx xs ys)"
by (auto simp add: opfx_def)

lemma opfx_code [code]:
 "opfx \<langle>ifirst xs\<rangle> xs \<longleftrightarrow> True"
 "opfx (x\<odot>xs) \<langle>y\<rangle> \<longleftrightarrow> False"
 "opfx (x\<odot>xs) (y\<odot>ys) \<longleftrightarrow> (x=y \<and> opfx xs ys)"
proof simp_all
 show "\<not> opfx (x \<odot> xs) \<langle>y\<rangle> " 
 by (simp add: opfx_def)
qed

lemma same_opfx_opfx [simp]:
 "opfx (xs \<ominus> ys) (xs\<ominus>zs) = opfx ys zs"
by (induct xs) simp_all

lemma same_opfx_INil [simp]:
 " opfx (xs \<ominus> ys) (xs \<ominus> \<langle>y\<rangle>) =  (ys = \<langle>y\<rangle>)"
by (meson same_iapp_eq opfx_order.less_le_not_le opfx_snoc sopfxI')

lemma opfx_opfx [simp]:
 assumes "opfx xs ys"
 shows   "  opfx xs (ys\<ominus>zs)"
using assms unfolding opfx_def by fastforce

lemma iapp_opfxD:
 assumes "opfx (xs \<ominus> ys) zs"
 shows   "   opfx xs zs"
using assms by (auto simp add: opfx_def)

lemma opfx_ICons: 
 "opfx xs (y\<odot>ys) = (xs = \<langle>y\<rangle> \<or> (\<exists> zs.  xs = y\<odot>zs \<and> opfx zs ys))"
by (case_tac xs) (auto simp add: opfx_def)

lemma opfx_iapp:
 "opfx xs (ys\<ominus>zs) = (opfx xs ys \<or> (\<exists> us. xs = ys \<ominus> us \<and> opfx us zs))"
proof (induct zs rule: irev_induct)
case (INil y)
then show ?case by (meson opfx_snoc same_opfx_opfx)
next
case (snoc x xs)
then show ?case by (metis iapp_assoc opfx_snoc)
qed

lemma iapp_one_opfx:
assumes "opfx xs ys"
        "ilen xs < ilen ys"
  shows "opfx (xs\<ominus> \<langle>inth ys ((ilen xs)+1)\<rangle>) ys"
using assms
proof (unfold opfx_def)
 assume 1: "\<exists>zs. ys = xs \<ominus> zs \<or> ys = xs" 
   then obtain sk where 2: " ys = xs \<ominus> sk \<or> ys = xs"
   by fastforce
 assume 3: "ilen xs < ilen ys"
 have  4: "ilen sk \<ge> 0"
   by auto
 have 5: "\<exists> v. ys = xs \<ominus> ((ifirst sk) \<odot> v) \<or> ys = xs \<ominus> \<langle>ifirst sk\<rangle> "
 by (metis "2" "3" iapp_INil less_irrefl opfx_def opfx_INil)
 have 6: "ys \<noteq> xs"
     using "3" by blast
 have 7: " (ifirst sk) = inth (xs \<ominus> sk) ((ilen xs) + 1) " 
    by (simp add: iapp_inth)
 have 8: "  ys = xs \<ominus> sk"
   using "2" "6" by auto
 have 9: "inth (xs \<ominus> sk) ((ilen xs) + 1) = inth (ys) ((ilen xs) + 1)" 
   using "8" by blast  
 thus "\<exists>zs. ys = (xs \<ominus> \<langle>inth ys ((ilen xs) + 1)\<rangle>) \<ominus> zs \<or> ys = xs \<ominus> \<langle>inth ys ((ilen xs) + 1)\<rangle>"
     using "5" "7" "9"  
   by simp
qed   

lemma opfx_ilen_le:
 assumes "opfx xs ys"
 shows   "ilen xs \<le> ilen ys"
 using assms by (auto simp add: opfx_def)

lemma opfx_same_cases:
 assumes "opfx xs1 ys"
         "opfx xs2 ys"
 shows   "opfx xs1 xs2 \<or> opfx xs2 xs1"
using assms unfolding opfx_def using iapp_eq_iapp_conv2 
by metis

lemma opfx_ilen_opfx:
 assumes "opfx ps xs"
         "opfx qs xs"
         "ilen ps \<le> ilen qs"
 shows   "opfx ps qs"
using assms 
by (auto simp: opfx_def) 
   (metis assms(2) dual_order.antisym iapp_eq_iapp_conv opfx_iapp opfx_ilen_le)

lemma iset_mono_opfx:
 assumes "opfx xs ys"
 shows   "iset xs \<subseteq> iset ys"
using assms by (auto simp: opfx_def)

lemma prefix_is_opfx:
 "opfx (prefix n xs) xs"
 by (metis Suc_eq_plus1 add_diff_inverse_nat iapp_prefix_suffix 
     prefix_ilen_gr_1 le_simps(2) less_Suc_eq_0_disj not_less opfx_def)

lemma imap_mono_opfx:
 assumes "opfx xs ys"
 shows   "opfx (imap f xs) (imap f ys)"
using assms by (auto simp: opfx_def)

lemma opfx_ilen_less:
 assumes "sopfx xs ys"
 shows   "ilen xs < ilen ys"
using assms by (auto simp: sopfx_def opfx_def)

lemma opfx_snocD:
 assumes "opfx (xs\<ominus>\<langle>x\<rangle>) ys"
 shows   "sopfx xs ys"
using assms 
using opfx_order.less_le_trans sopfxI' by blast

lemma sopfx_simps [simp, code]:
 "sopfx xs \<langle>y\<rangle> \<longleftrightarrow> False"
 "sopfx \<langle>x\<rangle> (x\<odot>xs) \<longleftrightarrow> True"
 "sopfx (x\<odot>xs) (y\<odot>ys) \<longleftrightarrow> x=y \<and> sopfx xs ys"
proof (simp_all add: sopfx_def)
 show "opfx xs \<langle>y\<rangle> \<longrightarrow> xs = \<langle>y\<rangle>" 
   by (metis iapp_not_INil opfxE)
 show "opfx \<langle>x\<rangle> (x \<odot> xs)" 
   by (simp add: opfx_ICons)
 show "(x = y \<and> opfx xs ys \<and> (x = y \<longrightarrow> xs \<noteq> ys)) = (x = y \<and> opfx xs ys \<and> xs \<noteq> ys)" 
   by auto
qed

lemma prefix_sopfx:
 assumes "sopfx xs ys"
 shows   "sopfx (prefix n xs) ys"
using assms 
proof (induct n arbitrary: xs ys)
case 0
then show ?case 
  proof (cases ys)
  case (INil x1)
  then show ?thesis 
   using "0.prems" by auto
  next
  case (ICons x21 x22)
  then show ?thesis 
   using "0.prems" opfx_order.le_less_trans prefix_is_opfx by blast
  qed
next
case (Suc n)
then show ?case 
 using opfx_order.order.strict_trans1 prefix_is_opfx by blast
qed

lemma osfxI [intro?]:
 assumes "ys = zs \<ominus> xs \<or> ys =xs"
 shows   "  osfx xs ys"
using assms unfolding osfx_def by blast

lemma osfxE [elim?]:
 assumes "osfx xs ys"
 obtains zs where "ys= zs \<ominus> xs \<or> ys = xs"   
  using assms unfolding osfx_def by blast

lemma osfx_tl [simp]:
 assumes "ilen xs >0"
 shows  "osfx (suffix 1 xs) xs"
using assms 
proof (induct xs)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case by (auto simp: osfx_def) (metis iapp.simps(1))
qed


lemma osfx_suffix [simp]:
 shows  "osfx (suffix i xs) xs"
proof -
have 1: "i>ilen xs \<Longrightarrow> osfx (suffix i xs) xs"
   by (metis ilast_iapp ilen_snoc_1 neq0_conv osfxI suffix_gr suffix_ilen_last suffix_zero)
have 2: "i\<le>ilen xs \<and> i = 0 \<Longrightarrow> osfx (suffix i xs) xs" 
    by simp
have 3: "i\<le>ilen xs \<and> 0<i \<Longrightarrow> osfx (suffix i xs) xs" 
   unfolding osfx_def using iapp_prefix_suffix[of "i-1" xs]
    by (metis One_nat_def Suc_eq_plus1 Suc_pred less_le_trans) 
show ?thesis 
using "1" "2" "3" leI by blast 
qed

lemma osfx_prefix [simp]:
assumes  "osfx xs ys"
 shows   "ys = (prefix (ilen ys - ilen xs -1) ys) \<ominus> xs   \<or> ys =xs" 
using assms 
by (auto simp add: osfx_def)
   (metis diff_zero prefix_iapp prefix_ilen)


lemma osfx_snoc [simp]:
 "osfx xs (ys\<ominus> \<langle>y\<rangle>)  \<longleftrightarrow> 
  xs = \<langle>y\<rangle> \<or> (\<exists>zs. xs = zs \<ominus> \<langle>y\<rangle> \<and> osfx zs ys)  "
proof (cases xs)
case (INil x1)
then show ?thesis by (metis ilast_iapp ilast_iapp2 osfxI osfx_prefix)
next
case (ICons x21 x22)
then show ?thesis 
  proof auto
   show "xs = x21 \<odot> x22 \<Longrightarrow> osfx (x21 \<odot> x22) (ys \<ominus> \<langle>y\<rangle>) \<Longrightarrow> \<exists>zs. x21 \<odot> x22 = zs \<ominus> \<langle>y\<rangle> \<and> osfx zs ys" 
    using iapp_eq_iapp_conv2 iapp_not_INil unfolding osfx_def 
       by (metis interval.distinct(1) )
   show "\<And>zs. xs = zs \<ominus> \<langle>y\<rangle> \<Longrightarrow> x21 \<odot> x22 = zs \<ominus> \<langle>y\<rangle> \<Longrightarrow> osfx zs ys \<Longrightarrow> osfx (zs \<ominus> \<langle>y\<rangle>) (ys \<ominus> \<langle>y\<rangle>)  "
    unfolding osfx_def by (metis iapp_assoc )
    qed
qed

lemma snoc_osfx_snoc [simp]: 
"osfx (xs \<ominus> \<langle>x\<rangle>) (ys \<ominus> \<langle>y\<rangle>) = (x = y \<and> osfx xs ys)"
by (simp add: osfx_def)
   (metis iapp_assoc iapp_eq_conv)

lemma same_osfx_osfx [simp]: 
"osfx (ys \<ominus> xs) (zs \<ominus> xs) = osfx ys zs"
unfolding osfx_def 
by (metis iapp_assoc iapp_same_eq)

lemma same_suffix_INil [simp]: 
"osfx (ys \<ominus> xs) (x\<odot>xs) = (ys = \<langle>x\<rangle>)"
unfolding osfx_def 
by (metis iapp_INil iapp_assoc iapp_not_INil iapp_same_eq)

lemma osfx_ICons:
 "osfx xs (y\<odot>ys) \<longleftrightarrow> xs = y\<odot>ys \<or> osfx xs ys"
unfolding osfx_def 
by (auto simp: ICons_eq_iapp_conv)

lemma osfx_iapp:
 "osfx xs (ys \<ominus> zs) \<longleftrightarrow> 
   osfx xs zs  \<or> (\<exists>xs'. xs = xs' \<ominus> zs \<and> osfx xs' ys)"
by (auto simp: osfx_def iapp_eq_iapp_conv2)

lemma osfx_ilen:
assumes "osfx xs ys"
 shows   "  ilen xs \<le> ilen ys"
using assms by (auto simp add: osfx_def)

lemma osfx_same_cases:
assumes "osfx xs\<^sub>1 ys"
        " osfx xs\<^sub>2 ys"
shows   "  osfx xs\<^sub>1 xs\<^sub>2 \<or> osfx xs\<^sub>2 xs\<^sub>1"
using assms unfolding osfx_def by (metis iapp_eq_iapp_conv2)

lemma osfx_ilen_osfx:
assumes "osfx ps xs"
        "osfx qs xs"
        "ilen ps \<le> ilen qs"
shows   "osfx ps qs"
using assms 
by (auto simp: osfx_def iapp_eq_iapp_conv2)

lemma osfx_ilen_less:
assumes "sosfx xs ys"
shows   "ilen xs < ilen ys"
using assms by (auto simp: sosfx_def osfx_def)

lemma osfx_IConsD':
assumes "osfx (x\<odot>xs) ys"
shows   "sosfx xs ys"
using assms
by (metis One_nat_def add_diff_cancel_right' diff_is_0_eq ilen.simps(2) not_one_le_zero osfx_ICons 
    osfx_ilen osfx_order.dual_order.refl osfx_order.dual_order.trans plus_1_eq_Suc sosfx_def) 


lemma suffix_sosfx:
assumes "sosfx xs ys"
shows   "sosfx (suffix n xs) ys"
using assms 
proof (induct n arbitrary: xs ys)
case 0
then show ?case by simp
next
case (Suc n)
then show ?case 
    proof (cases xs)
    case (INil x1)
    then show ?thesis using Suc.prems by auto
    next
    case (ICons x21 x22)
    then show ?thesis 
    using Suc.hyps Suc.prems osfx_IConsD' osfx_order.less_imp_le by fastforce
    qed
qed

lemma sosfx_tl [simp]:
 assumes "ilen xs >0"
 shows   "sosfx (suffix 1 xs) xs"
using assms 
proof (induct xs)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case using osfx_IConsD' by force
qed

lemma INil_osfx [simp]:
 " osfx \<langle>ilast xs\<rangle> xs "
by (metis osfx_suffix suffix_ilen_last)


lemma osfx_INil [simp]:
 " (osfx xs \<langle>ilast xs\<rangle>) = (xs = \<langle>ilast xs\<rangle>)"
by (simp add: osfx_order.antisym_conv)

lemma osfx_IConsI:
assumes "osfx xs ys"
shows   "osfx xs (x\<odot> ys)" 
using assms 
by (simp add: osfx_ICons)


lemma osfx_IConsD:
assumes "osfx (x\<odot>xs) ys"
shows   "osfx xs ys"
using assms 
by (meson osfx_IConsD' osfx_order.less_imp_le)

lemma osfx_iappI:
assumes "osfx xs ys"
shows   "osfx xs (zs \<ominus> ys)"
using assms by (metis iapp_assoc osfx_def)

lemma osfx_iappD:
assumes "osfx (zs \<ominus> xs) ys"
shows   "osfx xs ys"
using assms osfxI osfx_order.dual_order.trans by blast

lemma sosfx_iset_subset:
assumes "sosfx xs ys"
shows   "iset xs \<subseteq> iset ys"
using assms by (auto simp: sosfx_def osfx_def)

lemma iset_mono_osfx:
assumes "osfx xs ys"
shows   "iset xs \<subseteq> iset ys"
using assms by (auto simp: osfx_def)

lemma osfx_IConsD2:
assumes "osfx (x\<odot>xs) (y\<odot>ys)"
shows   "osfx xs ys"
using assms 
proof -
 assume "osfx (x\<odot>xs) (y\<odot>ys)"
 then obtain zs where "y\<odot>ys = zs \<ominus> (x\<odot>xs) \<or> y\<odot>ys = x\<odot>xs" 
   using osfxE by blast
 then show ?thesis  
 by (metis assms interval.inject(2) osfx_IConsD osfx_ICons)
qed

lemma osfx_to_opfx [code]:
 "osfx xs ys \<longleftrightarrow> opfx (irev xs) (irev ys)"
unfolding opfx_def 
by (metis irev_iapp irev_irev_ident osfx_def)

lemma sosfx_to_sopfx [code]:
 "sosfx xs ys \<longleftrightarrow> sopfx (irev xs) (irev ys)"
by (auto simp: osfx_to_opfx sosfx_def sopfx_def)


lemma imap_mono_osfx:
assumes "osfx xs ys"
shows   "osfx (imap f xs) (imap f ys)"
using assms by (auto elim!: osfxE intro: osfxI)

lemma prefix_subset:
assumes "k\<le>ilen xs"
 shows  " iset (prefix k xs) \<le> iset xs"
using assms by (simp add: prefix_is_opfx iset_mono_opfx)

lemma suffix_subset:
assumes "k\<le>ilen xs"
shows   " iset (suffix k xs) \<le> iset xs"
using assms by (simp add: iset_mono_osfx)

subsubsection \<open>idistinct and iremdups\<close>

lemma idistinct_iapp [simp]:
 "idistinct (xs\<ominus>ys) = (idistinct xs \<and> idistinct ys \<and> iset xs \<inter> iset ys = {})"
by (induct xs)  auto

lemma idistinct_osfx:
assumes "idistinct ys"
        "osfx xs ys"
shows   "idistinct xs"
using assms 
proof (clarsimp elim!: osfxE)
 show "\<And>zs. idistinct ys \<Longrightarrow> ys = zs \<ominus> xs \<or> ys = xs \<Longrightarrow> idistinct xs" 
 using idistinct_iapp by blast
qed

lemma idistinct_tl:
assumes "idistinct xs"
        "ilen xs > 0"
shows   "idistinct (suffix 1 xs)"
using assms 
by (cases xs) auto

lemma idistinct_irev [simp]:
 "idistinct (irev xs) = idistinct xs"
by (induct xs)  auto

lemma iset_iremdups [simp]:
 "iset (iremdups xs) = iset xs"
by (induct xs) auto

lemma idistinct_iremdups [iff]:
 "idistinct (iremdups xs)"
by (induct xs) auto

lemma idistinct_iremdups_id: 
assumes  "idistinct xs"
shows    "iremdups xs = xs"
using assms 
by (induct xs) auto
 
lemma iremdups_id_iff_idistinct [simp]: 
 "iremdups xs = xs \<longleftrightarrow> idistinct xs"
by (metis idistinct_iremdups idistinct_iremdups_id)

lemma idistinct_imap:
 "idistinct (imap f xs) = (idistinct xs \<and> inj_on f (iset xs))"
by (induct xs) auto

lemma idistinct_prefix [simp]:
assumes "idistinct xs"
shows   "idistinct (prefix i xs)"
using assms 
proof (induct xs arbitrary: i)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
  proof (cases i)
  case 0
  then show ?thesis by auto
  next
  case (Suc nat)
  then show ?thesis 
  using ICons.hyps ICons.prems prefix_is_opfx iset_mono_opfx by fastforce
  qed
qed

lemma idistinct_suffix [simp]:
assumes "idistinct xs"
shows   "idistinct (suffix i xs)"
using assms 
proof 
 (induct xs arbitrary: i)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
   proof (cases i)
   case 0
   then show ?thesis using ICons.prems by auto
   next
   case (Suc nat)
   then show ?thesis using iapp_prefix_suffix  ICons.hyps ICons.prems by auto
   qed
qed


lemma idistinct_conv_inth:
 "idistinct xs = 
  (\<forall> i \<le> ilen xs. (\<forall> j \<le> ilen xs. i \<noteq> j \<longrightarrow> inth xs i \<noteq> inth xs j))"
proof 
 (induction xs)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
   by (auto simp add: ICons inth_ICons inth_and_iset split: nat.split_asm) 
      blast+
qed

lemma idistinct_Ex1:
assumes "idistinct xs"
        "x \<in> iset xs"
shows   "(\<exists>!i. i\<le> ilen xs \<and> (inth xs i) = x)"
using assms
by (metis idistinct_conv_inth inth_and_iset)


lemma inj_on_inth:
assumes "idistinct xs"
shows   "(\<forall> i \<in> I. i \<le> ilen xs \<Longrightarrow> inj_on (inth xs) I)"
using assms 
by (meson idistinct_conv_inth inj_onI)

lemma bij_betw_inth:
 assumes "idistinct xs"
         "A = {..< ilen xs+1}"
         "B = iset xs"
 shows   "bij_betw ((inth) xs) A B"
 using assms unfolding bij_betw_def 
proof (auto intro!: inj_on_inth simp: )
   show "\<And>xa. idistinct xs \<Longrightarrow>
          A = {..<Suc (ilen xs)} \<Longrightarrow>
          B = iset xs \<Longrightarrow> xa < Suc (ilen xs) \<Longrightarrow> inth xs xa \<in> iset xs" 
     by (meson inth_and_iset less_Suc_eq_le)
   show "\<And>x. idistinct xs \<Longrightarrow>
         A = {..<Suc (ilen xs)} \<Longrightarrow>
         B = iset xs \<Longrightarrow> x \<in> iset xs \<Longrightarrow> x \<in> inth xs ` {..<Suc (ilen xs)} " 
     using inth_and_iset by (metis image_iff lessThan_iff less_Suc_eq_le)
qed

lemma card_idistinct:
 assumes "card (iset xs) = ilen xs +1"
 shows   "idistinct xs"
using assms 
proof 
 (induct xs)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
  proof (cases "x1a \<in> iset xs")
  case True
  then show ?thesis 
  by (metis ICons.prems iremdups.simps(2) iset_iremdups add_Suc card_ilen ilen.simps(2) 
      not_less_eq_eq plus_1_eq_Suc)
  next
  case False
  then show ?thesis using ICons.hyps ICons.prems by auto
  qed
qed


lemma finite_interval:
 assumes "finite A"
 shows   "(A\<noteq> {} \<longrightarrow> (\<exists>xs. iset xs = A)) "
using assms 
proof (induct rule:finite_induct)
case empty
then show ?case by simp
next
case (insert x F)
then show ?case by (metis insert_is_Un interval.set(1) iset_iapp)
qed

lemma finite_idistinct_interval:
assumes "finite A"
        "A\<noteq> {}"
shows   " (\<exists> xs. iset xs = A \<and> idistinct xs)"
using assms by (metis idistinct_iremdups finite_interval iset_iremdups)

lemma iremdups_eq_INil_iff [simp]: 
 "(iremdups xs = \<langle>x\<rangle>) = (\<forall> i \<le> ilen xs. (inth xs i) = x )"
proof 
 (induct xs)
 case (INil x)
 then show ?case by auto
 next
 case (ICons x1a xs)
 then show ?case 
 proof -
  have 1: "(\<forall>i\<le>ilen (x1a \<odot> xs). inth (x1a \<odot> xs) i = x) =
           ( x1a = x \<and> (\<forall>i\<le>ilen (xs). inth (xs) i = x)) "
      by auto
         (metis Nitpick.case_nat_unfold Suc_eq_plus1 le_diff_conv)
  have 2: "(iremdups (x1a \<odot> xs) = \<langle>x\<rangle>) = 
           ((if x1a \<in> iset xs then iremdups xs else x1a \<odot> iremdups xs) = \<langle>x\<rangle>)" 
       by simp
  have 3: "((if x1a \<in> iset xs then iremdups xs else x1a \<odot> iremdups xs) = \<langle>x\<rangle>) = 
           (iremdups xs  = \<langle>x\<rangle> \<and> x1a \<in> iset xs)"
    by auto 
  have 4: "(iremdups xs  = \<langle>x\<rangle> \<and> x1a \<in> iset xs) =
           ((\<forall>i\<le>ilen (xs). inth (xs) i = x) \<and> x1a \<in> iset xs)  "
    using ICons.hyps by blast
  have 5: "(\<forall>i\<le>ilen (xs). inth (xs) i = x) \<and> x1a \<in> iset xs \<longrightarrow> x1a = x " 
           by (metis inth_and_iset)
  show ?thesis 
  by (metis "1" "2" "3" "5" ICons.hyps iset_iremdups interval.set_intros(1))
  qed 
qed

lemma iremdups_eq_INil_right_iff [simp]: 
 "(\<langle>x\<rangle> = iremdups xs) = (\<forall> i \<le> ilen xs. (inth xs i) = x )"
by (metis iremdups_eq_INil_iff)

lemma ilen_remdups_leq [iff]: 
 "ilen(iremdups xs) \<le> ilen xs"
by (induct xs) auto

lemma ilen_iremdups_eq[iff]:
  "(ilen (iremdups xs) = ilen xs) = (iremdups xs = xs)"
proof 
 (induct xs)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
 by (metis ilen_remdups_leq iremdups.simps(2) add_left_cancel ilen.simps(2) not_less_eq_eq 
     plus_1_eq_Suc)
qed

lemma idistinct_card:
assumes "idistinct xs"
shows   "card(iset xs) = Suc(ilen xs)"
using assms
by (induct xs) simp_all

subsubsection \<open>prefixes and suffixes\<close>

lemma in_iset_prefixes [simp]:
 "xs \<in> iset (prefixes ys) \<longleftrightarrow> opfx xs ys"
proof 
 (induct xs arbitrary: ys)
 case (INil x)
 then show ?case 
   proof (cases ys)
   case (INil x1)
   then show ?thesis using opfxE by force
   next
   case (ICons x21 x22)
   then show ?thesis unfolding opfx_def by auto
   qed
 next
 case (ICons x1a xs)
 then show ?case 
    proof (cases ys)
    case (INil x1)
    then show ?thesis by (simp add: opfx_code(2))
    next
    case (ICons x21 x22)
    then show ?thesis using ICons.hyps by auto
    qed
qed


lemma ilen_prefixes [simp]:
 "ilen (prefixes xs) = ilen xs"
by (induction xs)  auto

lemma idistinct_prefixes [intro]:
 "idistinct (prefixes xs)"
proof (induction xs)
case (INil x)
then show ?case by (auto simp: idistinct_imap)
next
case (ICons x1a xs)
then show ?case by (auto simp: idistinct_imap)
                   (meson inj_onI interval.inject(2))
qed 


lemma prefixes_snoc [simp]:
 "prefixes (xs\<ominus>\<langle>x\<rangle>) = (prefixes xs) \<ominus> \<langle>xs\<ominus>\<langle>x\<rangle>\<rangle>"
by (induction xs) auto

lemma ifirst_prefixes [simp]:
 "ifirst (prefixes xs) = \<langle>ifirst xs\<rangle>"
by (cases xs)  auto

lemma ilast_prefixes [simp]:
 "ilast (prefixes xs) = xs"
by (induction xs) 
   (simp_all add: ilast_imap inth_imap)

lemma prefixes_iapp:
 "prefixes (xs\<ominus>ys) = 
  prefixes xs \<ominus> imap (\<lambda>ys'. xs \<ominus> ys') (prefixes ys)"
proof 
 (induction xs arbitrary: ys)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
   proof -
    have 1: "prefixes ((x1a \<odot> xs) \<ominus> ys) = \<langle>x1a\<rangle> \<odot> (imap ((\<odot>) x1a) (prefixes (xs \<ominus> ys))) "
       by simp
    have 2: "prefixes (xs \<ominus> ys) = prefixes xs \<ominus> imap ((\<ominus>) xs) (prefixes ys)"
       by (simp add: ICons.IH) 
    have 3: "\<langle>x1a\<rangle> \<odot> (imap ((\<odot>) x1a) (prefixes (xs \<ominus> ys))) =
             \<langle>x1a\<rangle> \<odot> (imap ((\<odot>) x1a) (prefixes xs \<ominus> imap ((\<ominus>) xs) (prefixes ys)))
             " 
        using "2" by auto
    show ?thesis by (simp add: "3")
   qed
qed

lemma prefixes_eq_snoc:
 "prefixes ys = xs \<ominus> \<langle>x\<rangle> \<longleftrightarrow>
  (\<exists> z zs. ys = zs\<ominus>\<langle>z\<rangle> \<and> xs = prefixes zs) \<and> x = ys"
proof (cases ys rule: irev_cases)
case (INil x)
then show ?thesis using iapp_not_INil by (metis prefixes.simps(1))
next
case (snoc ys y)
then show ?thesis by auto
qed 

lemma iset_prefixes_eq:
 "iset (prefixes xs) = {ys. opfx ys xs}"
by auto

lemma card_iset_prefixes [simp]:
 "card (iset (prefixes xs)) = Suc (ilen xs)"
by (simp add: idistinct_card idistinct_prefixes)

lemma iset_prefixes_append:
 "iset (prefixes (xs\<ominus>ys)) = iset (prefixes xs) \<union> {xs \<ominus> ys' | ys'. ys' \<in> iset (prefixes ys)}"
by (subst prefixes_iapp) auto 
 

lemma in_iset_suffixes [simp]:
 "xs \<in> iset(suffixes ys) \<longleftrightarrow> osfx xs ys"
proof (induct ys)
case (INil x)
then show ?case using osfx_prefix by fastforce
next
case (ICons x1a ys)
then show ?case by (simp add: osfx_ICons)
qed

lemma interval_sfx_INil:
 "iset(suffixes \<langle>x\<rangle>) = {\<langle>x\<rangle>}"
by simp

lemma interval_sfx_ICons:
 "iset(suffixes (x\<odot>xs)) = {x\<odot>xs} \<union> iset(suffixes xs)"
by auto

lemma iset_suffixes_sfx:
 "iset (suffixes xs) = {suffix i xs| i. i\<le>ilen xs}"
proof 
 (induction xs)
 case (INil x)
 then show ?case by auto
 next
 case (ICons x1a xs)
 then show ?case 
   proof -
    have 1: "{suffix i (x1a\<odot>xs )| i. i\<le>ilen (x1a\<odot>xs)} =
             {suffix i (x1a \<odot> xs)| i. i=0} \<union> {suffix i (x1a \<odot> xs)| i. 0<i \<and> i \<le>ilen (x1a \<odot> xs)}"
      using neq0_conv by (auto, fastforce, force)
    have 2: "{suffix i (x1a \<odot> xs)| i. 0<i \<and> i \<le>ilen (x1a \<odot> xs)} =
              {suffix i (xs)| i. i \<le>ilen (xs)}" 
      by (auto,
          metis Nitpick.case_nat_unfold Suc_eq_plus1 le_diff_conv neq0_conv,
          metis Nitpick.case_nat_unfold Suc_le_mono add_diff_cancel_left' gr_implies_not_zero 
          less_Suc0 not_less_eq plus_1_eq_Suc)     
    show ?thesis using "1" "2" ICons.IH by auto  
  qed
qed

lemma iset_suffixes_osfx:
 "iset(suffixes xs) = {ys. osfx ys xs}"
by auto


lemma idistinct_suffixes [intro]:
 " idistinct(suffixes xs)"
proof (induct xs)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
  by (metis in_iset_suffixes iremdups.simps(2) iremdups_id_iff_idistinct osfx_IConsI 
      osfx_order.dual_order.eq_iff suffixes.simps(2))
qed

lemma ilen_suffixes [simp]:
 "ilen (suffixes xs) = ilen xs"
by (induct xs) simp_all

lemma suffixes_snoc [simp]:
"suffixes (xs \<ominus> \<langle>x\<rangle>) = (imap (\<lambda> ys. ys \<ominus> \<langle>x\<rangle>) (suffixes xs)) \<ominus> \<langle>\<langle>x\<rangle>\<rangle>" 
by (induct xs) simp_all

lemma ifirst_suffixes [simp]:
 "ifirst (suffixes xs) = xs"
by (induct xs) simp_all

lemma ilast_suffixes [simp]:
"ilast (suffixes xs) = \<langle>ilast xs\<rangle>"
by (induct xs) simp_all


lemma suffixes_iapp:
 "suffixes (xs \<ominus> ys) = imap (\<lambda> xs'. xs' \<ominus> ys) (suffixes xs) \<ominus> (suffixes ys)"
proof 
 (induction xs arbitrary: ys)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a ys)
 then show ?case 
   proof (cases ys)
   case (INil x1)
   then show ?thesis by simp
   next
   case (ICons x21 x22)
   then show ?thesis by (auto simp add: ICons.IH)
   qed
qed

lemma card_iset_suffixes [simp]:
" card (iset (suffixes xs)) = Suc (ilen xs)"
by (simp add: idistinct_card idistinct_suffixes)

lemma iset_suffixes_iapp:
"iset (suffixes (xs\<ominus>ys)) = {xs' \<ominus> ys | xs'. xs' \<in> iset (suffixes xs)} \<union> iset (suffixes ys)"
proof (subst suffixes_iapp)
 show "iset (imap (\<lambda>xs'. xs' \<ominus> ys) (suffixes xs) \<ominus> suffixes ys) =
      {xs' \<ominus> ys |xs'. xs' \<in> iset (suffixes xs)} \<union> iset (suffixes ys) " 
 proof (cases xs)
 case (INil x1)
 then show ?thesis by simp
 next
 case (ICons x21 x22)
 then show ?thesis by force
 qed
qed

lemma imap_first_suffixes [simp]:
 "imap (\<lambda> xs. inth xs 0) (suffixes xs) = xs"
by (induct xs) auto

lemma suffixes_conv_prefixes:
 " (suffixes xs) = irev (imap irev (prefixes (irev xs)))"
by (induction xs) auto

lemma prefixes_conv_suffixes:
 " (prefixes xs) = irev (imap irev (suffixes (irev xs)))"
by (induction xs) (auto simp add: irev_imap)

lemma prefixes_irev:
 " prefixes (irev xs) = irev (imap irev (suffixes xs))"
by (induction xs) auto

lemma suffixes_irev:
 " suffixes (irev xs) = irev (imap irev (prefixes xs))" 
by (induction xs) (auto simp add: irev_imap)

lemma inth_suffixes:
 assumes " i\<le> ilen(suffixes xs)"
 shows   " inth (suffixes xs) i = (suffix i xs)"
using assms 
proof (induct xs arbitrary:i)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case by (simp add: Nitpick.case_nat_unfold)
qed


lemma suffix_suffixes:
 assumes "i \<le> ilen (suffixes xs)"
 shows   "suffix i (suffixes xs) = suffixes (suffix i xs)"
using assms 
proof (induct xs arbitrary:i)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case by (simp add: Nitpick.case_nat_unfold) 
qed


subsubsection \<open>ifilter and nfilter\<close>

lemma sfxfilter_ilen_a:
 assumes "\<exists> ys \<in> iset (suffixes \<langle>x\<rangle>). P ys"
 shows   "ilen (ifilter P (suffixes \<langle>x\<rangle>)) = 0"
using assms by simp

lemma sfxfilter_ilen_b:
 assumes "(\<exists> ys \<in> iset (suffixes (x\<odot>xs)). P ys)"
          "(\<exists> ys \<in> iset (suffixes xs). P ys)"
          "P (x\<odot>xs)"
 shows   " ilen (ifilter P (suffixes (x\<odot>xs))) = ilen(ifilter P (suffixes xs))+1"
using assms by simp 

lemma sfxfilter_ilen_c:
 assumes "(\<exists> ys \<in> iset (suffixes (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> iset (suffixes xs). P ys)"
         "\<not> P (x\<odot>xs)"
 shows   " ilen (ifilter P (suffixes (x\<odot>xs))) = ilen(ifilter P (suffixes xs))"
using assms by simp

lemma sfxfilter_ilen_d:
 assumes "(\<exists> ys \<in> iset (suffixes (x\<odot>xs)). P ys)"
         "\<not>(\<exists> ys \<in> iset (suffixes xs). P ys)"
 shows   "ilen (ifilter P (suffixes (x\<odot>xs))) = 0"
using assms by simp

lemma ifilter_ilen_a:
 assumes "\<exists> ys \<in> iset \<langle>x\<rangle>. P ys"
 shows   "ilen (ifilter P \<langle>x\<rangle>) = 0"
using assms by simp 

lemma nfilter_ilen_a:
 assumes "\<exists> ys \<in> iset \<langle>x\<rangle>. P ys"
 shows   "ilen (nfilter P \<langle>x\<rangle> n) = 0"
using assms by simp 

lemma ifilter_ilen_b:
 assumes "(\<exists> ys \<in> iset (x\<odot>xs). P ys)"
         "(\<exists> ys \<in> iset (xs). P ys)"
         "P (x)"
 shows   " ilen (ifilter P (x\<odot>xs)) = ilen(ifilter P (xs))+1"
using assms by simp

lemma nfilter_ilen_b:
 assumes "(\<exists> ys \<in> iset (x\<odot>xs). P ys)"
         "(\<exists> ys \<in> iset (xs). P ys)"
         "P (x)"
 shows   " ilen (nfilter P (x\<odot>xs) n) = ilen(nfilter P (xs) (Suc n))+1"
using assms by simp

lemma ifilter_ilen_c:
 assumes "(\<exists> ys \<in> iset (x\<odot>xs). P ys)"
         "(\<exists> ys \<in> iset (xs). P ys)"
         "\<not> P (x)"
 shows   " ilen (ifilter P (x\<odot>xs)) = ilen(ifilter P (xs))"
using assms by simp

lemma nfilter_ilen_c:
 assumes "(\<exists> ys \<in> iset (x\<odot>xs). P ys)"
         "(\<exists> ys \<in> iset (xs). P ys)"
         "\<not> P (x)"
 shows   " ilen (nfilter P (x\<odot>xs) n) = ilen(nfilter P (xs) (Suc n))"
using assms by simp

lemma ifilter_ilen_d:
 assumes "(\<exists> ys \<in> iset ( (x\<odot>xs)). P ys)"
         "\<not>(\<exists> ys \<in> iset ( xs). P ys)"
 shows   " ilen (ifilter P ( (x\<odot>xs))) = 0"
using assms by simp

lemma nfilter_ilen_d:
 assumes "(\<exists> ys \<in> iset ( (x\<odot>xs)). P ys)"
         "\<not>(\<exists> ys \<in> iset ( xs). P ys)"
 shows   " ilen (nfilter P (x\<odot>xs) n) = 0"
using assms by simp

lemma sfxfilter_inth_a:
 assumes "(\<exists> ys \<in> iset (suffixes \<langle>x\<rangle>). P ys)"
         "j\<le> ilen(ifilter P (suffixes \<langle>x\<rangle>))"
 shows   "inth (ifilter P (suffixes \<langle>x\<rangle>)) j = \<langle>x\<rangle>"
using assms by simp

lemma sfxfilter_inth_b1:
 assumes "(\<exists> ys \<in> iset (suffixes (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> iset (suffixes xs). P ys)"
         "P (x\<odot>xs)"
 shows   " inth (ifilter P (suffixes (x\<odot>xs))) 0 = x\<odot>xs"
using assms by simp

lemma sfxfilter_inth_b2:
 assumes "(\<exists> ys \<in> iset (suffixes (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> iset (suffixes xs). P ys)"
         "P (x\<odot>xs)"
         "(Suc j) \<le> ilen (ifilter P (suffixes (x\<odot>xs)))"
 shows   " inth (ifilter P (suffixes (x\<odot>xs))) (Suc j) = inth (ifilter P (suffixes (xs))) j"
using assms by auto

lemma sfxfilter_inth_c:
 assumes "(\<exists> ys \<in> iset (suffixes (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> iset (suffixes xs). P ys)"
          "\<not> P (x\<odot>xs)"
          "j \<le> ilen (ifilter P (suffixes (x\<odot>xs)))"
 shows   " inth (ifilter P (suffixes (x\<odot>xs))) j =  inth (ifilter P (suffixes (xs))) j"
using assms by auto

lemma sfxfilter_inth_d:
 assumes "(\<exists> ys \<in> iset (suffixes (x\<odot>xs)). P ys)"
         "\<not>(\<exists> ys \<in> iset (suffixes xs). P ys)"
         "j\<le> ilen (ifilter P (suffixes (x\<odot>xs)))"
 shows   " inth (ifilter P (suffixes (x\<odot>xs))) j  = x\<odot>xs"
using assms by auto

lemma nfilter_inth_a:
 assumes "(\<exists> ys \<in> iset ( \<langle>x\<rangle>). P ys)"
         "j\<le> ilen(nfilter P ( \<langle>x\<rangle>) n)"
 shows   "inth (nfilter P ( \<langle>x\<rangle>) n) j = n"
using assms by auto

lemma ifilter_inth_a:
 assumes "(\<exists> ys \<in> iset ( \<langle>x\<rangle>). P ys)"
         "j\<le> ilen(ifilter P ( \<langle>x\<rangle>))"
 shows   "inth (ifilter P ( \<langle>x\<rangle>)) j = x"
using assms by simp

lemma nfilter_inth_b1:
 assumes "(\<exists> ys \<in> iset ( (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> iset ( xs). P ys)"
         "P (x)"
 shows   " inth (nfilter P ( (x\<odot>xs)) n) 0 = n"
using assms by simp

lemma ifilter_inth_b1:
 assumes "(\<exists> ys \<in> iset ( (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> iset ( xs). P ys)"
         "P (x)"
 shows   " inth (ifilter P ( (x\<odot>xs))) 0 = x"
using assms by simp

lemma nfilter_inth_b2:
 assumes "(\<exists> ys \<in> iset ( (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> iset ( xs). P ys)"
         "P (x)"
         "(Suc j) \<le> ilen (nfilter P ( (x\<odot>xs)) n)  "
 shows   " inth (nfilter P ( (x\<odot>xs)) n) (Suc j) = inth (nfilter P ( (xs)) (Suc n)) j"
using assms by auto

lemma ifilter_inth_b2:
 assumes "(\<exists> ys \<in> iset ( (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> iset ( xs). P ys)"
         "P (x)"
         "(Suc j) \<le> ilen (ifilter P ( (x\<odot>xs)))  "
 shows   " inth (ifilter P ( (x\<odot>xs))) (Suc j) = inth (ifilter P ( (xs))) j"
using assms by auto

lemma nfilter_inth_c:
 assumes "(\<exists> ys \<in> iset ( (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> iset ( xs). P ys)"
         "\<not> P (x)"
         "j \<le> ilen (nfilter P ( (x\<odot>xs)) n)"
 shows   " inth (nfilter P ( (x\<odot>xs)) n) j =  inth (nfilter P ( (xs)) (Suc n)) j"
using assms by auto

lemma ifilter_inth_c:
 assumes "(\<exists> ys \<in> iset ( (x\<odot>xs)). P ys)"
         "(\<exists> ys \<in> iset ( xs). P ys)"
         "\<not> P (x)"
         "j \<le> ilen (ifilter P ( (x\<odot>xs)))"
 shows   " inth (ifilter P ( (x\<odot>xs))) j =  inth (ifilter P ( (xs))) j"
using assms by auto

lemma nfilter_inth_d:
 assumes "(\<exists> ys \<in> iset ( (x\<odot>xs)). P ys)"
         "\<not>(\<exists> ys \<in> iset ( xs). P ys)"
         "j\<le> ilen (nfilter P ( (x\<odot>xs)) n)"
 shows   " inth (nfilter P ( (x\<odot>xs)) n) j  = n"
using assms by auto

lemma ifilter_inth_d:
 assumes "(\<exists> ys \<in> iset ( (x\<odot>xs)). P ys)"
         "\<not>(\<exists> ys \<in> iset ( xs). P ys)"
         "j\<le> ilen (ifilter P ( (x\<odot>xs)))"
 shows   " inth (ifilter P ( (x\<odot>xs))) j  = x"
using assms by auto

lemma sfxfilter_inth_ICons:
 "inth (ifilter P (suffixes (x\<odot>xs))) j =
  (if (\<exists> ys \<in> iset (suffixes xs). P ys) then 
      (if P (x\<odot>xs) then 
        (if j=0 then (x\<odot>xs) else inth (ifilter P (suffixes xs)) (j-1))
       else inth (ifilter P (suffixes xs)) j)
   else (x\<odot>xs))"
by (simp add: Nitpick.case_nat_unfold)

lemma sfxfilter_inth_ICons_a:
 "inth (ifilter P (suffixes (x\<odot>xs))) j =
  (if (\<exists> ys \<in> iset (suffixes xs). P ys) then 
      (if P (x\<odot>xs) then 
        (case j of 0 \<Rightarrow> x\<odot>xs | Suc m \<Rightarrow> inth (ifilter P (suffixes xs)) m)
       else inth (ifilter P (suffixes xs)) j)
   else (x\<odot>xs))"
by (simp add: Nitpick.case_nat_unfold)

lemma nfilter_inth_ICons:
 "inth (nfilter P ( (x\<odot>xs)) n) j =
  (if (\<exists> ys \<in> iset ( xs). P ys) then 
      (if P (x) then 
        (if j=0 then (n) else inth (nfilter P ( xs) (Suc n)) (j-1))
       else inth (nfilter P  (xs) (Suc n)) j)
   else (n))"
by (simp add: Nitpick.case_nat_unfold)

lemma nfilter_inth_ICons_a:
 "inth (nfilter P ( (x\<odot>xs)) n) j =
  (if (\<exists> ys \<in> iset ( xs). P ys) then 
      (if P (x) then 
        (case j of 0 \<Rightarrow> n | Suc m \<Rightarrow> inth (nfilter P ( xs) (Suc n)) m)
       else inth (nfilter P  (xs) (Suc n)) j)
   else (n))"
by (simp add: Nitpick.case_nat_unfold)

lemma ifilter_inth_ICons:
 "inth (ifilter P ( (x\<odot>xs))) j =
  (if (\<exists> ys \<in> iset ( xs). P ys) then 
      (if P (x) then 
        (if j=0 then (x) else inth (ifilter P ( xs)) (j-1))
       else inth (ifilter P  (xs)) j)
   else (x))"
by (simp add: Nitpick.case_nat_unfold)

lemma ifilter_inth_ICons_a:
 "inth (ifilter P ( (x\<odot>xs))) j =
  (if (\<exists> ys \<in> iset ( xs). P ys) then 
      (if P (x) then 
        (case j of 0 \<Rightarrow> x | (Suc m) \<Rightarrow> inth (ifilter P ( xs)) m)
       else inth (ifilter P  (xs)) j)
   else (x))"
by (simp add: Nitpick.case_nat_unfold)

lemma sfxfilter_inth:
 assumes "(\<exists> ys \<in> iset (suffixes xs). P ys)"
         "i \<le> ilen (ifilter P (suffixes xs))"
 shows   " P (inth (ifilter P (suffixes xs)) i)"
using assms
proof 
 (induction xs arbitrary: i)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
   proof (cases "\<exists>a\<in>iset (suffixes xs). P a")
   case True
   then show ?thesis using ICons by (auto simp add: nat.split_sels(2)) 
   next
   case False
   then show ?thesis using ICons by simp
   qed
qed


lemma nfilter_ilen:
 assumes "(\<exists> x \<in> iset xs. P x)"
 shows   "ilen(nfilter P xs n) = ilen (ifilter P xs)"
using assms
proof 
 (induction xs arbitrary: n)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case by simp
qed


lemma nfilter_upper_bound:
 assumes "(\<exists> x \<in> iset xs. P x)"
         "i\<le>ilen (nfilter P xs n)"
 shows   "(inth (nfilter P xs n) i) \<le> n+ilen xs"
using assms
proof 
 (induct xs arbitrary: i n)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
 proof (cases "\<exists> x \<in> iset xs. P x")
 case True
 then show ?thesis 
    proof (cases i)
    case 0
    then show ?thesis using ICons True 
    by  simp_all 
        (metis ICons.prems(1) ICons.prems(2) add_Suc nfilter_ilen_c)
    next
    case (Suc nat)
    then show ?thesis using ICons True by  simp_all fastforce
    qed
 next
 case False
 then show ?thesis using Cons by simp
 qed 
qed

lemma nfilter_lower_bound:
 assumes "(\<exists> x \<in> iset xs. P x)"
         "i\<le>ilen (nfilter P xs n)"
 shows   " n \<le> (inth (nfilter P xs n) i) "
using assms
proof 
 (induction xs arbitrary: i n)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case
  proof   (cases "\<exists> x \<in> iset xs. P x")
  case True
  then show ?thesis 
    proof (cases i)
    case 0
    then show ?thesis using ICons True Suc_leD by (simp_all, blast)
    next
    case (Suc nat)
    then show ?thesis using ICons True by simp_all fastforce
    qed
  next
  case False
  then show ?thesis using Cons False 
  by auto
  qed
qed

lemma nfilter_ifilter:
 assumes "(\<exists> x \<in> iset xs. P x)"
         "i\<le>ilen (nfilter P xs n)"
 shows   " (inth xs ((inth (nfilter P xs n) i)-n)  ) = (inth (ifilter P xs) i)"
using assms 
proof 
 (induct xs arbitrary: i n)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
   proof -
    have 1: "(\<exists> x \<in> iset xs. P x) \<and> P x1a \<and> i=0 \<longrightarrow> 
              inth (x1a \<odot> xs) (inth (nfilter P (x1a \<odot> xs) n) i - n) =
              inth (ifilter P (x1a \<odot> xs)) i"
     by auto
    have 2: "\<And>j. (\<exists> x \<in> iset xs. P x) \<and> P x1a \<and> j<ilen (nfilter P (x1a \<odot>xs) n)\<longrightarrow> 
              inth (x1a \<odot> xs) (inth (nfilter P (x1a \<odot> xs) n) (Suc j) - n) =
              inth (ifilter P (x1a \<odot> xs)) (Suc j)"
     by auto
        (metis ICons.hyps  inth.simps(2) Suc_diff_Suc inth_Suc less_Suc_eq_le 
         less_eq_Suc_le nfilter_lower_bound)  
    have 3: "(\<exists> x \<in> iset xs. P x) \<and> \<not> P x1a \<and> i=0 \<longrightarrow> 
              inth (x1a \<odot> xs) (inth (nfilter P (x1a \<odot> xs) n) i - n) =
              inth (ifilter P (x1a \<odot> xs)) i" 
      using nfilter_lower_bound[of "xs" "P" "0" "Suc n"]  ICons
     by auto
        (metis  inth.simps(2) Suc_diff_Suc inth_Suc le0 less_eq_Suc_le)
    have 4: "\<And>j. (\<exists> x \<in> iset xs. P x) \<and> \<not>P x1a \<and> j<ilen (nfilter P (x1a \<odot>xs) n)\<longrightarrow> 
              inth (x1a \<odot> xs) (inth (nfilter P (x1a \<odot> xs) n) (Suc j) - n) =
              inth (ifilter P (x1a \<odot> xs)) (Suc j)" 
      by auto
        (metis ICons.hyps  inth.simps(2) Suc_diff_Suc inth_Suc less_eq_Suc_le 
         nfilter_lower_bound)
    have 5: "\<not>(\<exists> x \<in> iset xs. P x) \<longrightarrow> inth (x1a \<odot> xs) (inth (nfilter P (x1a \<odot> xs) n) i - n) =
              inth (ifilter P (x1a \<odot> xs)) i" 
      by simp
    from 1 2 3 4 5 show ?thesis 
    by (metis ICons.prems(2) add_less_cancel_left le_0_eq le_SucE le_imp_less_Suc plus_1_eq_Suc 
        zero_induct)
  qed     
qed

lemma iset_ifilter [simp]:
assumes "(\<exists> x \<in> iset xs. P x)"
shows "iset (ifilter P xs) = {x. x \<in> iset xs \<and> P x}"
using assms 
proof 
 (induct xs)
 case (INil x)
 then show ?case by (simp add: Collect_conv_if)
 next
 case (ICons x1a xs)
 then show ?case by auto
qed

lemma iset_nfilter [simp]:
assumes "(\<exists> x \<in> iset xs. P x)"
shows "iset (nfilter P xs n) = {n+k|k .  k\<le> ilen xs \<and> P (inth xs k)}"
using assms
proof 
 (induction xs arbitrary: n)
 case (INil x)
 then show ?case by auto
 next
 case (ICons x1a xs)
 then show ?case 
   proof -
    have 1: "(\<exists> x \<in> iset xs. P x) \<and> P x1a \<longrightarrow>
             iset (nfilter P (x1a \<odot> xs) n) = 
             {n + k |k. k \<le> ilen (x1a \<odot> xs) \<and> P (inth (x1a \<odot> xs) k)}" 
     using ICons by auto
         (metis (mono_tags, lifting) Nitpick.case_nat_unfold One_nat_def Suc_le_mono Suc_pred 
          gr_implies_not0)+
    have 2: "(\<exists> x \<in> iset xs. P x) \<and> \<not> P x1a \<longrightarrow>
             iset (nfilter P (x1a \<odot> xs) n) = 
             {n + k |k. k \<le> ilen (x1a \<odot> xs) \<and> P (inth (x1a \<odot> xs) k)} " 
     using ICons by auto
        (metis (mono_tags, lifting) Nitpick.case_nat_unfold One_nat_def Suc_eq_plus1 Suc_pred 
         le_0_eq le_diff_conv not_le_imp_less)
     have 3: "\<not>(\<exists> x \<in> iset xs. P x)  \<longrightarrow>
             iset (nfilter P (x1a \<odot> xs) n) = 
             {n + k |k. k \<le> ilen (x1a \<odot> xs) \<and> P (inth (x1a \<odot> xs) k)}" 
      using ICons by auto
          (metis Nitpick.case_nat_unfold Suc_eq_plus1 le_diff_conv inth_iset)
     show ?thesis 
     using "1" "2" "3" by linarith
    qed     
qed

lemma iset_minus_ifilter_out:
assumes "(\<exists> z \<in> iset xs. (\<lambda> x. \<not>(x =y)) z)"
 shows  "iset xs - {y} = iset(ifilter (\<lambda> x. \<not>(x =y)) xs)"
using assms
by (induct xs) auto

lemma ifilter_ifilter [simp]:
assumes "\<exists> x \<in> iset (ifilter Q xs). P x"
        "\<exists> x \<in> iset xs. Q x"
        "\<exists> x \<in> iset xs. P x \<and> Q x" 
shows "ifilter P (ifilter Q xs) = ifilter (\<lambda>x. P x \<and> Q x) xs"
using assms
by (induct xs) auto

lemma ilen_nfilter_le [simp]: 
 "ilen (nfilter P xs n) \<le> ilen xs"
by (induct xs arbitrary: n) (auto simp add: le_SucI)

lemma ilen_ifilter_le [simp]: 
 "ilen (ifilter P xs) \<le> ilen xs"
by (induct xs) (auto simp add: le_SucI)

lemma sfxfilter_bound:
 assumes "(\<exists> ys \<in> iset (suffixes xs). P ys)"
 shows  " ilen (ifilter P (suffixes xs)) \<le> ilen xs"
using assms by (metis ilen_ifilter_le ilen_suffixes)

lemma ifilter_bound:
 assumes "(\<exists> ys \<in> iset ( xs). P ys)"
 shows  " ilen (ifilter P (xs)) \<le> ilen xs"
using assms by auto

lemma sfxfilter_inth_bound:
 assumes "(\<exists> ys \<in> iset (suffixes xs). P ys) "
         " j \<le> ilen (ifilter P (suffixes xs))"
 shows  " ilen ((inth (ifilter P (suffixes xs)) j)) \<le> ilen xs"
using assms iset_ifilter[of "suffixes xs" "P"]  
by (metis (mono_tags, lifting) in_iset_suffixes mem_Collect_eq inth_iset osfx_ilen)

lemma sfxfilter_inth_suffix:
 assumes "(\<exists> ys \<in> iset (suffixes xs). P ys)"
         "j\<le> ilen (ifilter P (suffixes xs))"
 shows " inth (ifilter P (suffixes xs)) j = 
         suffix (ilen xs - ilen(inth(ifilter P (suffixes xs)) j)) xs"
using assms 
proof 
 (induction xs arbitrary: j)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
   proof -
    have 1: "(\<exists> ys \<in> iset (suffixes xs). P ys) \<and> P (x1a\<odot>xs) \<and> j=0 \<longrightarrow> 
             inth (ifilter P (suffixes (x1a \<odot> xs))) j =
           suffix (ilen (x1a \<odot> xs) - ilen (inth (ifilter P (suffixes (x1a \<odot> xs))) j)) (x1a \<odot> xs)" 
      using Cons by auto
    have 2: "(\<exists> ys \<in> iset (suffixes xs). P ys) \<and> \<not>P (x1a\<odot>xs) \<and> j=0 \<longrightarrow> 
             inth (ifilter P (suffixes (x1a \<odot> xs))) j =
           suffix (ilen (x1a \<odot> xs) - ilen (inth (ifilter P (suffixes (x1a \<odot> xs))) j)) (x1a \<odot> xs)" 
      using ICons sfxfilter_inth_bound[of "xs" "P" "0"]  using Suc_diff_le by auto 
    have 3: "\<And>x. (\<exists> ys \<in> iset (suffixes xs). P ys) \<and> P (x1a\<odot>xs) \<and> 
                  x < ilen (ifilter P (suffixes (x1a\<odot>xs))) \<longrightarrow> 
             inth (ifilter P (suffixes (x1a \<odot> xs))) (Suc x) =
       suffix (ilen (x1a \<odot> xs)-ilen (inth (ifilter P (suffixes (x1a \<odot> xs))) (Suc x))) (x1a \<odot> xs)"  
     using ICons sfxfilter_inth_bound[of "xs" "P"]  Suc_diff_le in_iset_suffixes less_Suc_eq_le by auto
    have 4: "\<And>x. (\<exists> ys \<in> iset (suffixes xs). P ys) \<and> \<not>P (x1a\<odot>xs) \<and> 
                  x < ilen (ifilter P (suffixes (x1a\<odot>xs))) \<longrightarrow> 
             inth (ifilter P (suffixes (x1a \<odot> xs))) (Suc x) =
       suffix (ilen (x1a \<odot> xs)-ilen (inth (ifilter P (suffixes (x1a \<odot> xs))) (Suc x))) (x1a \<odot> xs)"  
      using ICons sfxfilter_inth_bound[of "xs" "P"] using Suc_diff_le by auto
    have 5: "\<not>(\<exists> ys \<in> iset (suffixes xs). P ys) \<longrightarrow> 
             inth (ifilter P (suffixes (x1a \<odot> xs))) j =
           suffix (ilen (x1a \<odot> xs) - ilen (inth (ifilter P (suffixes (x1a \<odot> xs))) j)) (x1a \<odot> xs)" 
      using Cons by auto  
    show ?thesis 
    by (metis "1" "2" "3" "4" "5" ICons.prems(2) Suc_less_SucD le_SucE le_imp_less_Suc le_zero_eq 
        zero_induct) 
   qed
qed

lemma initfilter_sfxfilter_exists:
 " (\<exists> ys \<in> iset (suffixes xs). P ( (inth ys 0 ))) = (\<exists> x \<in> iset xs. P x)"
by (metis inth_and_iset inth_imap ilen_suffixes imap_first_suffixes)

lemma initfilter_sfxfilter:
 assumes "\<exists> ys \<in> iset (suffixes xs). P ( (inth ys 0 ))"
 shows   "ifilter P xs =  imap (\<lambda>s. (inth s 0)) (ifilter (\<lambda>ys. P(inth ys 0)) (suffixes xs)) "
using assms 
proof 
 (induction xs)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
   proof -
    have 1: "(\<exists> x \<in> iset xs. P x) \<and> P x1a \<longrightarrow> 
             ifilter P (x1a \<odot> xs) =
              imap (\<lambda>s.  inth s 0) (ifilter (\<lambda>ys. P (inth ys 0)) (suffixes (x1a \<odot> xs)))" 
     using ICons initfilter_sfxfilter_exists[of "xs" "P"]  by auto
    have 2: "(\<exists> x \<in> iset xs. P x) \<and> \<not>P x1a \<longrightarrow> 
             ifilter P (x1a \<odot> xs) =
              imap (\<lambda>s.  inth s 0) (ifilter (\<lambda>ys. P (inth ys 0)) (suffixes (x1a \<odot> xs)))"  
     using ICons initfilter_sfxfilter_exists[of "xs" "P"] by auto
    have 3: "\<not>(\<exists> x \<in> iset xs. P x)  \<longrightarrow> 
             ifilter P (x1a \<odot> xs) =
              imap (\<lambda>s.  inth s 0) (ifilter (\<lambda>ys. P (inth ys 0)) (suffixes (x1a \<odot> xs)))" 
     using ICons initfilter_sfxfilter_exists[of "xs" "P"] by auto
   show ?thesis using "1" "2" "3" by blast
  qed
qed

lemma ifilter_inth:
 assumes "(\<exists> x \<in> iset xs. P x)"
         "i\<le> ilen (ifilter P xs)"
 shows   "(\<exists> k \<le> ilen xs. inth(ifilter P xs) i = inth xs k)"
using assms using iset_ifilter[of "xs" "P"] 
by (metis (no_types, lifting) inth_and_iset mem_Collect_eq)


lemma interval_sfx_inth_zero:
 " iset xs = {(inth ys 0)| ys. ys \<in> iset(suffixes xs) }"
proof 
 (induct xs)
 case (INil x)
 then show ?case 
 by auto
 next
 case (ICons x1a xs)
 then show ?case 
   by auto (metis   inth_zero)
qed

lemma interval_sfx_1:
assumes "ys \<in> iset(suffixes xs)"
shows   "(inth ys 0) \<in> iset xs"
using assms interval_sfx_inth_zero by fastforce

lemma sum_ilen_ifilter_compl_help:
 assumes "\<exists> x \<in> iset xs. P x"
         "\<exists> x \<in> iset xs. \<not> P x"
 shows   "ilen xs > 0"
using assms 
proof 
 (induct xs)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
  using ilen_ICons_1 by blast
qed

lemma ifilter_id_conv: 
 assumes "\<exists> x \<in> iset xs. P x"
 shows "(ifilter P xs = xs) = (\<forall>x\<in>iset xs. P x)"
using assms
proof 
 (induct xs)
 case (INil x)
 then show ?case by auto
 next
 case (ICons x1a xs)
 then show ?case 
  by (auto simp add: interval_iset_nonempty)
     (metis ilen_ifilter_le Suc_n_not_le_n ilen.simps(2) plus_1_eq_Suc)
qed

lemma sum_ilen_ifilter_compl:
assumes "\<exists> x \<in> iset xs. P x"
        "\<exists> x \<in> iset xs. \<not> P x"
 shows "ilen(ifilter P xs) + ilen(ifilter (\<lambda>x. \<not>P x) xs) +1 = ilen xs "
using assms
proof 
 (induct xs)
 case (INil x)
 then show ?case by auto
 next
 case (ICons x1a xs)
 then show ?case 
   proof -
    have 1: "(\<exists> x \<in> iset xs. P x) \<and> (\<exists> x \<in> iset xs. \<not> P x) \<and> P x1a \<longrightarrow> 
             ilen (ifilter P (x1a \<odot> xs)) + ilen (ifilter (\<lambda>x. \<not> P x) (x1a \<odot> xs)) + 1 = 
             ilen (x1a \<odot> xs)"
      using ICons by auto
    have 2: "(\<exists> x \<in> iset xs. P x) \<and> (\<exists> x \<in> iset xs. \<not> P x) \<and> \<not>P x1a \<longrightarrow> 
             ilen (ifilter P (x1a \<odot> xs)) + ilen (ifilter (\<lambda>x. \<not> P x) (x1a \<odot> xs)) + 1 = 
             ilen (x1a \<odot> xs)" 
      using ICons by auto
    have 3: "\<not>(\<exists> x \<in> iset xs. P x)  \<longrightarrow> 
             ilen (ifilter P (x1a \<odot> xs)) + ilen (ifilter (\<lambda>x. \<not> P x) (x1a \<odot> xs)) + 1 = 
             ilen (x1a \<odot> xs)" 
      using ICons by auto
                  (metis ifilter_id_conv)
    have 4: "(\<exists> x \<in> iset xs. P x) \<and> \<not>(\<exists> x \<in> iset xs. \<not> P x) \<and> P x1a \<longrightarrow> 
             ilen (ifilter P (x1a \<odot> xs)) + ilen (ifilter (\<lambda>x. \<not> P x) (x1a \<odot> xs)) + 1 = 
             ilen (x1a \<odot> xs)" 
     using ICons by auto
    have 5: "(\<exists> x \<in> iset xs. P x) \<and> \<not>(\<exists> x \<in> iset xs. \<not> P x) \<and> \<not>P x1a \<longrightarrow> 
             ilen (ifilter P (x1a \<odot> xs)) + ilen (ifilter (\<lambda>x. \<not> P x) (x1a \<odot> xs)) + 1 = 
             ilen (x1a \<odot> xs)" 
      using ICons by auto
                  (metis ifilter_id_conv)
    show ?thesis 
    using "1" "2" "3" "4" "5" by blast
  qed   
qed
 
lemma ifilter_ilen_imp:
assumes "\<exists> x \<in> iset xs. P x \<and> Q x"
shows   "ilen (ifilter (\<lambda>x. P x \<and> Q x) xs) \<le> ilen (ifilter P xs)"
using assms 
proof (induct xs)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case by force
qed 

lemma subset_ifilter:
assumes "(\<exists> x \<in> iset xs. P x)"
shows   " iset (ifilter P xs) \<le> iset (ifilter (\<lambda> x. P x \<or> Q x) xs)"
proof -
 have 1: "(\<exists> x \<in> iset xs. P x \<or> Q x)"
   using assms by blast 
 have 2: " iset (ifilter (\<lambda> x. P x \<or> Q x) xs) = {x \<in> iset xs. P x \<or> Q x}" 
   using assms iset_ifilter[of "xs" "(\<lambda> x. P x \<or> Q x)"] by blast
 have 3: "iset (ifilter P xs) = {x \<in> iset xs. P x}"
   using assms iset_ifilter by auto
 have 4: "{x \<in> iset xs. P x} \<le> {x \<in> iset xs. P x \<or> Q x}"
    by auto
 show ?thesis by (simp add: "2" "3" "4")  
qed

lemma iset_ifilter_not:
 assumes " \<exists> x \<in> iset xs. P x" 
         " \<exists> x \<in> iset xs. \<not> (P x)"
 shows   "iset (ifilter (\<lambda>x. \<not> (P x)) xs) = iset xs - iset (ifilter P xs)"
using assms
 proof (induct xs)
 case (INil x)
 then show ?case by auto
 next
 case (ICons x1a xs)
 then show ?case
    proof - 
     have 1: "(\<exists> x \<in> iset xs. P x) \<and> (\<exists> x \<in> iset xs. \<not> (P x)) \<and> P x1a \<longrightarrow>
              iset (ifilter (\<lambda>x. \<not> P x) (x1a \<odot> xs)) = iset (x1a \<odot> xs) - iset (ifilter P (x1a \<odot> xs))" 
      using ICons by auto
     have 2: "(\<exists> x \<in> iset xs. P x) \<and> (\<exists> x \<in> iset xs. \<not> (P x)) \<and> \<not>P x1a \<longrightarrow>
              iset (ifilter (\<lambda>x. \<not> P x) (x1a \<odot> xs)) = iset (x1a \<odot> xs) - iset (ifilter P (x1a \<odot> xs))" 
      using ICons by auto
     have 3: "(\<exists> x \<in> iset xs. P x) \<and> \<not>(\<exists> x \<in> iset xs. \<not> (P x)) \<and> P x1a \<longrightarrow>
              iset (ifilter (\<lambda>x. \<not> P x) (x1a \<odot> xs)) = iset (x1a \<odot> xs) - iset (ifilter P (x1a \<odot> xs))" 
      using ICons by auto
     have 4: "(\<exists> x \<in> iset xs. P x) \<and> \<not>(\<exists> x \<in> iset xs. \<not> (P x)) \<and> \<not>P x1a \<longrightarrow>
              iset (ifilter (\<lambda>x. \<not> P x) (x1a \<odot> xs)) = iset (x1a \<odot> xs) - iset (ifilter P (x1a \<odot> xs))" 
      using ICons by auto
     have 5: "\<not>(\<exists> x \<in> iset xs. P x) \<longrightarrow>
              iset (ifilter (\<lambda>x. \<not> P x) (x1a \<odot> xs)) = iset (x1a \<odot> xs) - iset (ifilter P (x1a \<odot> xs)) " 
      using ICons by auto
     show ?thesis using "1" "2" "3" "4" "5" by linarith
    qed
qed

lemma iset_ifilter_cap:
 assumes "\<exists> x \<in> iset xs. f x"
         "\<exists> x \<in> iset xs. \<not> f x"
 shows   " iset (ifilter f xs) \<inter> iset (ifilter (\<lambda> x. \<not> f x) xs) = {}"
using assms by auto

lemma ifilter_inth_or:
 assumes "\<exists> x \<in> iset xs. P x"
 shows   "\<exists> x \<in> iset (ifilter (\<lambda>x. P x \<or> Q x) xs). P x"
proof -
 have 1: "\<exists> i \<le> ilen(ifilter (\<lambda>x. P x \<or> Q x) xs). P (inth (ifilter (\<lambda>x. P x) xs) i)"
   using assms by (metis (mono_tags, lifting) iset_ifilter ilen_gr_zero 
   mem_Collect_eq inth_iset)  
 obtain i where 2: "i \<le> ilen(ifilter (\<lambda>x. P x \<or> Q x) xs) \<and> P (inth (ifilter (\<lambda>x. P x) xs) i) \<and>
                    (inth (ifilter (\<lambda>x. P x) xs) i) \<in> iset (ifilter P xs)" 
    using "1" 
    by (metis (mono_tags, lifting) iset_ifilter assms ilen_gr_zero 
        mem_Collect_eq inth_iset)
 have 3: "iset (ifilter P xs) \<le> iset (ifilter (\<lambda>x. P x \<or> Q x) xs)"
    using assms subset_ifilter[of "xs" "P"] by auto
 have 4: "(inth (ifilter (\<lambda>x. P x) xs) i) \<in> iset (ifilter P xs)"  
   using "2" by auto
 have 5: "(inth (ifilter (\<lambda>x. P x) xs) i) \<in> iset (ifilter (\<lambda>x. P x \<or> Q x) xs)" 
   using "3" "4" by blast
  from "2" "5" show ?thesis by blast
qed

lemma ifilter_iapp1:
assumes "\<forall>x\<in>iset xs. \<not> P x"
        "\<exists> y \<in> iset ys. P y"
shows   "ifilter P (xs \<ominus> ys) = ifilter P ys"
using assms
by (induct xs) auto

lemma ifilter_iapp2:
assumes "\<forall>x\<in>iset xs. \<not> P x"
        "\<exists>y\<in>iset ys. P y"
shows   "ifilter P (ys \<ominus> xs) = ifilter P ys"
using assms
by (induct ys) auto

lemma ifilter_iapp [simp]:
assumes "(\<exists> x \<in> iset (xs \<ominus> ys). P x)"
        "(\<exists> x \<in> iset xs. P x)"
        "(\<exists> x \<in> iset ys. P x)"
shows "ifilter P (xs\<ominus>ys) = (ifilter P xs) \<ominus> (ifilter P ys)"
using assms 
proof 
 (induct xs arbitrary: ys)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
  proof 
   (cases "(\<exists> x \<in> iset xs. P x)")
   case True
   then show ?thesis using ICons by auto
   next
   case False
   then show ?thesis using ICons using ifilter_iapp1[of "xs" "P"] 
   using iset_IConsD iset_iapp by auto  
 qed
qed

lemma ifilter_irev:
 assumes "(\<exists> x \<in> iset xs. P x)"
 shows   " irev (ifilter P xs) = ifilter P (irev xs)"
using assms 
proof 
 (induct xs)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
  then show ?case 
    proof (cases "(\<exists> x \<in> iset xs. P x)")
    case True
    then show ?thesis 
        proof (cases "P x1a")
        case True
        then show ?thesis using ICons True 
            by (simp add: ifilter_iapp1)
        next
        case False
        then show ?thesis using ICons False \<open>(\<exists> x \<in> iset xs. P x)\<close> 
        by (simp add: ifilter_iapp2)      
        qed
    next
    case False
    then show ?thesis using False ICons 
    by (simp add: ifilter_iapp1)
    qed
qed



lemma ifilter_True:
assumes "\<forall>x \<in> iset xs. P x"
shows   "ifilter P xs = xs"
using assms 
by (meson ifilter_id_conv ilen_ifilter_le inth_iset)

lemma nfilter_imap:
 assumes "\<exists> x \<in> iset (imap f xs). P x"
shows   " nfilter P (imap f xs) n =  (nfilter (P\<circ>f) xs n)"
using assms 
by (induct xs arbitrary: n) auto


lemma ifilter_imap:
assumes "\<exists> x \<in> iset (imap f xs). P x"
shows   " ifilter P (imap f xs) = imap f (ifilter (P\<circ>f) xs)"
using assms
by (induct xs) auto

lemma  ilen_nfilter_imap[simp]:
assumes "\<exists> x \<in> iset (imap f xs). P x"
shows  "ilen (nfilter P (imap f xs) n) = ilen(nfilter (P \<circ> f) xs n)" 
using assms by (simp add:nfilter_imap)         

lemma ilen_ifilter_imap[simp]:
assumes "\<exists> x \<in> iset (imap f xs). P x"
shows  "ilen (ifilter P (imap f xs)) = ilen(ifilter (P \<circ> f) xs)"
using assms by (simp add:ifilter_imap)

lemma nfilter_is_subset [simp]:
 assumes "\<exists> x \<in> iset xs. P x"
 shows   "iset (nfilter P xs n) \<le> {n+k|k. k\<le>ilen xs}"
using assms by auto

lemma ifilter_is_subset [simp]:
 assumes "\<exists> x \<in> iset xs. P x"
 shows   "iset (ifilter P xs) \<le> iset xs"
using assms by auto

lemma ilen_nfilter_less:
assumes "\<exists>  x \<in> iset xs. P x"
        "x \<in> iset xs"
        "\<not> P x "
 shows "ilen(nfilter P xs n) < ilen xs"
using assms 
proof 
 (induct xs arbitrary: n)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case using le_imp_less_Suc ilen_nfilter_le by (simp_all, blast)
qed

lemma ilen_ifilter_less:
assumes "\<exists>  x \<in> iset xs. P x"
        " x \<in> iset xs" 
        " \<not> P x "
 shows " ilen(ifilter P xs) < ilen xs"
using assms 
proof 
 (induct xs)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
       using ilen_ifilter_le le_imp_less_Suc by (simp_all, blast)
qed


lemma nfilter_iset:
 assumes "\<exists>  x \<in> iset xs. P x"
 shows   "iset (nfilter P xs n) = {n+i|i. i\<le>ilen xs \<and> P(inth xs i)}"
using assms by auto

lemma nfilter_cong[fundef_cong]:
 assumes "xs = ys"
         " (\<And>x. x \<in> iset ys \<Longrightarrow> P x = Q x) "
 shows   " nfilter P xs n = nfilter Q ys n"
using assms by (induction xs arbitrary: ys n) auto


lemma ifilter_cong[fundef_cong]:
 assumes "xs = ys"
         " (\<And>x. x \<in> iset ys \<Longrightarrow> P x = Q x) "
 shows   " ifilter P xs = ifilter Q ys"
using assms by (induct ys arbitrary: xs) auto

lemma iremdups_ifilter: 
 assumes "\<exists> x \<in> iset xs . P x"
shows "iremdups(ifilter P xs) = ifilter P (iremdups xs)"
using assms
by (induct xs) auto

lemma idistinct_imap_ifilter:
assumes "\<exists> x \<in> iset xs . P x"
        " idistinct (imap f xs)"
shows   " idistinct (imap f (ifilter P xs))"
using assms by (induct xs) auto

lemma idistinct_nfilter [simp]:
 "idistinct (nfilter P xs n)"
by (induction xs arbitrary: n) auto

lemma idistinct_ifilter [simp]: 
assumes "idistinct xs"
shows   "idistinct (ifilter P xs)"
using assms by (induct xs) auto

lemma idistinct_ilen_ifilter: 
assumes "\<exists> x \<in> iset xs . P x"
        " idistinct xs "
shows " ilen (ifilter P xs) +1 = card ({x. P x} Int iset xs)"
using assms by (induct xs) auto


lemma idx_nfilter_mono:
assumes "\<exists> x \<in> iset xs . P x"
        "na<ilen (nfilter P xs n)" 
shows   "inth (nfilter P xs n) na < inth (nfilter P xs n) (Suc na) " 
using assms 
proof (induct xs arbitrary: n na)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
   proof -
    have 1: "(\<exists> x \<in> iset xs . P x) \<and> P x1a \<and> na=0\<longrightarrow>
             inth (nfilter P (x1a \<odot> xs) n) na < inth (nfilter P (x1a \<odot> xs) n) (Suc na)" 
     using ICons by auto
         (meson Suc_le_lessD ilen_gr_zero nfilter_lower_bound)+
    have 5: "\<not>(\<exists> x \<in> iset xs . P x) \<longrightarrow>
             inth (nfilter P (x1a \<odot> xs) n) na < inth (nfilter P (x1a \<odot> xs) n) (Suc na)" 
     using ICons gr_implies_not_zero by (auto, fastforce) 
    show ?thesis
    using "1" "5" ICons.hyps ICons.prems(2) less_Suc_eq_0_disj by auto   
  qed
qed


lemma idx_nfilter:
assumes "\<exists> x \<in> iset xs . P x"
shows   "index_sequence (ifirst((nfilter P xs n))) (nfilter P xs n)"
using assms by (simp add: index_sequence_def idx_nfilter_mono)

lemma idx_nfilter_expand:
assumes "\<exists> x \<in> iset xs . P x"
shows   "\<forall>na<ilen (nfilter P xs n). inth (nfilter P xs n) na < inth (nfilter P xs n) (Suc na)"
using assms idx_nfilter by (simp add: index_sequence_def)


lemma idx_nfilter_gr_eq:
 assumes "\<exists> x \<in> iset xs . P x"
         "k \<le> j"
         "j \<le> ilen(nfilter P xs n)" 
  shows  "inth (nfilter P xs n) k \<le>  inth (nfilter P xs n) j"
using assms by (meson idx_nfilter idx_less_eq)

lemma idx_nfilter_gr:
 assumes "\<exists> x \<in> iset xs . P x"
  shows  "(\<forall> j .  k< j \<and> j \<le> ilen(nfilter P xs n)\<longrightarrow>
              inth (nfilter P xs n) k <  inth (nfilter P xs n) j)"
using assms 
by (meson Suc_leI dual_order.strict_trans1 idx_nfilter_expand idx_nfilter_gr_eq)

lemma idx_nfilter_less_eq:
 assumes "\<exists> x \<in> iset xs . P x"
         " k\<le> ilen(nfilter P xs n)" 
  shows  " (\<forall> j \<le> k.  inth (nfilter P xs n) j \<le>  inth (nfilter P xs n) k)"
using assms by (simp add: idx_nfilter_gr_eq)

lemma idx_nfilter_less:
 assumes "\<exists> x \<in> iset xs . P x"
         "k\<le> ilen(nfilter P xs n)" 
  shows  "(\<forall> j < k.  inth (nfilter P xs n) j <  inth (nfilter P xs n) k)"
using assms 
by (simp add: idx_nfilter_gr)

lemma nfilter_prefix_iset_0:
assumes "k\<le> ilen xs"
        "\<exists> x \<in> iset (prefix k xs) . P x"
 shows "ilen (nfilter P (prefix k xs) n) \<le> ilen (nfilter P xs n)" 
using assms 
proof (induct xs arbitrary: n k)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
   proof (cases k)
   case 0
   then show ?thesis using ICons by simp
   next
   case (Suc nat)
   then show ?thesis using ICons  prefix_subset by simp blast
   qed
qed


lemma nfilter_subset:
 assumes "\<exists> x \<in> iset (prefix k xs) . P x"
         "k\<le> ilen xs" 
 shows " iset (nfilter P (prefix k xs) n) \<le> iset (nfilter P xs n)"
using assms 
proof (induct xs arbitrary:k n)
case (INil x)
then show ?case by auto
next
case (ICons x1a xs)
then show ?case 
  proof (cases k)
  case 0
  then show ?thesis using ICons by simp
  next
  case (Suc nat)
  then show ?thesis using ICons by (auto simp add: inth_iset)  
  qed
qed

lemma nfilter_prefix_iset:
 assumes "\<exists> x \<in> iset (prefix k xs) . P x"
         "k\<le> ilen xs " 
 shows   "iset (nfilter P (prefix k xs) n) = {n+i|i. i\<le> k \<and> P(inth xs i)}"
using assms
by auto

lemma nfilter_suffix_iset:
 assumes "\<exists> x \<in> iset (suffix k xs) . P x"
         " k\<le> ilen xs " 
 shows " iset (nfilter P (suffix k xs) n) = {n+i|i. i \<le> ilen xs-k \<and> P(inth xs (k+i))}"
using assms by auto

lemma nfilter_suffix_iset_a:
 assumes "\<exists> x \<in> iset (suffix k xs) . P x"
         " k\<le> ilen xs " 
 shows   " iset (nfilter P (suffix k xs) n) = {n+(j-k)|j. k\<le>j \<and>  j \<le> ilen xs \<and> P(inth xs j)}"
using assms  Nat.le_diff_conv2 by auto fastforce

lemma nfilter_suffix_iset_b:
 assumes "\<exists> x \<in> iset (suffix k xs) . P x"
         " k\<le> ilen xs " 
 shows " iset (nfilter P (suffix k xs) n) = {n+i|i. i \<le> ilen xs-k \<and> P(inth xs (i+k))}"
using assms  nfilter_suffix_iset by (auto simp add: add.commute)
  
lemma nfilter_card:
assumes "\<exists> x \<in> iset xs . P x"
 shows " ilen (nfilter P xs n) +1 = card (iset (nfilter P xs n)) " 
using assms 
by (induct xs arbitrary: n) auto

lemma nfilter_prefix_iset_1:
assumes "k\<le> ilen xs"
        "\<exists> x \<in> iset (prefix k xs) . P x"
 shows   "iset (prefix (ilen(nfilter P (prefix k xs) n)) (nfilter P xs n)) =
            {(inth (nfilter P xs n) i) | i. i\<le> ilen(nfilter P (prefix k xs) n)}"
using  prefix_iset[of "(ilen(nfilter P (prefix k xs) n))" "(nfilter P xs n)"]
by (simp add: assms(1) assms(2) nfilter_prefix_iset_0)

lemma nfilter_prefix_subset:
assumes "\<exists> x \<in> iset xs . P x"
        " k \<le> ilen (nfilter P xs n) " 
shows   " iset (prefix k (nfilter P xs n)) \<le> iset (nfilter P xs n)"
using assms 
using prefix_subset by blast


lemma ilen_ifilter_conv_card:
assumes "\<exists>  x \<in> iset xs. P x"
 shows  "ilen (ifilter P xs) +  1 = card {i. i\<le> ilen xs \<and> P (inth xs i)}"
proof -
 have 1: " ilen (ifilter P xs) = ilen(nfilter P xs 0)" 
  by (simp add: assms nfilter_ilen)
 have 2: "  ilen (nfilter P xs 0) +1 = card (iset (nfilter P xs 0))"  
   by (meson assms nfilter_card) 
 have 3: " iset (nfilter P xs 0) = {i|i. i\<le> ilen xs \<and> P(inth xs i)}"
   using assms by auto
 show ?thesis 
 using "1" "2" "3" by auto
qed

lemma ifilter_nfilter_prefix_ilen_0:
   assumes "P (inth xs   ( (inth (nfilter P xs n) k) -n))"
           "k \<le> ilen (ifilter P xs)" 
   shows   "(\<exists> x \<in> iset xs. P x)"
using assms 
by (induction xs arbitrary: k) auto


lemma nfilter_ilen_n_zero:
  assumes "(\<exists> x \<in> iset xs. P x)"
  shows   "ilen (nfilter P xs n) = ilen (nfilter P xs 0)"
using assms 
proof 
 (induction xs arbitrary: n)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
   proof (cases n)
   case 0
   then show ?thesis using ICons by blast
   next
   case (Suc nat)
   then show ?thesis using ICons by (simp add: nfilter_ilen)
   qed
qed

lemma nfilter_inth_n_zero_a:
  assumes "(\<exists> x \<in> iset xs. P x)"
          " k \<le> ilen (nfilter P xs n)"
  shows   " n \<le> (inth (nfilter P xs n) k)"
using assms by (simp add: nfilter_lower_bound)

lemma nfilter_inth_n_zero:
  assumes "(\<exists> x \<in> iset xs. P x)"
          " k \<le> ilen (nfilter P xs n)"
  shows   "(inth (nfilter P xs n) k) - n  = inth (nfilter P xs 0) k"
using assms 
proof 
 (induction xs arbitrary: n k)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xs)
 then show ?case 
    proof (cases "(\<exists> x \<in> iset xs. P x)")
    case True
    then show ?thesis 
       proof  (cases "P x1a")
       case True
       then show ?thesis 
          proof  (cases k)
          case 0
          then show ?thesis using ICons True by simp
          next
          case (Suc nat)
          then show ?thesis using ICons True \<open>(\<exists> x \<in> iset xs. P x)\<close>
               proof auto
                 fix x
                 assume a0: "k = Suc nat" 
                 assume a1: "(\<And>k n. k \<le> ilen (nfilter P xs n) \<Longrightarrow> 
                              inth (nfilter P xs n) k - n = inth (nfilter P xs 0) k)"
                 assume a2: "nat \<le> ilen (nfilter P xs (Suc n))" 
                 assume a3: "P x1a" 
                 assume a4: "x \<in> iset xs" 
                 assume a5: "P x" 
                 show "inth (nfilter P xs (Suc n)) nat - n = inth (nfilter P xs (Suc 0)) nat" 
                 proof -
                 have 1: "inth (nfilter P xs (Suc n)) nat - (Suc n) = inth (nfilter P xs 0) nat"
                    using a1 a2 a3 a4 a5 by blast
                 have 2: "inth (nfilter P xs (Suc 0)) nat - (Suc 0) = inth (nfilter P xs 0) nat"
                    by (metis a2 a1 a4 a5 nfilter_ilen_n_zero) 
                 from 1 2 show ?thesis
                 by (metis One_nat_def Suc_diff_le \<open>\<exists>x\<in>iset xs. P x\<close> a2 diff_Suc_Suc 
                     nfilter_ilen_n_zero nfilter_lower_bound 
                     ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc) 
               qed
             qed
            qed
       next
       case False
       then show ?thesis using ICons True 
          proof  auto
            fix x
            assume b0: "\<not> P x1a" 
            assume b1: "(\<And>k n. k \<le> ilen (nfilter P xs n) \<Longrightarrow> 
                          inth (nfilter P xs n) k - n = inth (nfilter P xs 0) k)" 
            assume b2: "k \<le> ilen (nfilter P xs (Suc n))" 
            assume b3: "x \<in> iset xs"
            assume b4: "P x"  
            show " inth (nfilter P xs (Suc n)) k - n = inth (nfilter P xs (Suc 0)) k "
            proof -
            have 3: "inth (nfilter P xs (Suc n)) k - (Suc n) = inth (nfilter P xs 0) k"
              using b1 b2 by blast 
            have 4: "inth (nfilter P xs (Suc 0)) k - (Suc 0) = inth (nfilter P xs 0) k"
              by (metis True b1 b2 nfilter_ilen_n_zero) 
            show ?thesis 
            by (metis "3" "4" One_nat_def Suc_diff_le True b2 diff_Suc_Suc nfilter_ilen_n_zero 
                nfilter_lower_bound ordered_cancel_comm_monoid_diff_class.add_diff_inverse 
                plus_1_eq_Suc)
           qed
         qed
       qed
    next
    case False
    then show ?thesis using ICons False by auto
    qed
qed  

lemma nfilter_n_zero:
  assumes "(\<exists> x \<in> iset xs. P x)"
  shows   " (nfilter P xs n) = imap (\<lambda>i. i+n) (nfilter P xs 0)"
using assms 
proof -
 have 1: "ilen (nfilter P xs n) = ilen (imap (\<lambda>i. i+n) (nfilter P xs 0))" 
  using assms nfilter_ilen_n_zero by fastforce
 have 2: "\<And>k. k\<le> ilen (nfilter P xs n) \<longrightarrow>
               inth (nfilter P xs n) k = inth (imap (\<lambda>i. i+n) (nfilter P xs 0)) k"
  using assms nfilter_inth_n_zero[of "xs" "P" _ "n"] 
  by (metis inth_imap le_add_diff_inverse2 nfilter_lower_bound) 
 show ?thesis by (simp add: "1" "2" interval_eq_inth_eq)
qed

lemma nfilter_n_zero_a:
  assumes "(\<exists> x \<in> iset xs. P x)"
  shows   "  (nfilter P xs 0) = imap (\<lambda>i. i-n) (nfilter P xs n)"
proof -
 have 1: "ilen(nfilter P xs 0) = ilen( imap (\<lambda>i. i-n) (nfilter P xs n))"
     by (metis assms ilen_imap nfilter_ilen_n_zero) 
 have 2: "\<And>k. k\<le> ilen (nfilter P xs 0) \<longrightarrow>
              inth (nfilter P xs 0) k = inth (imap (\<lambda>i. i-n) (nfilter P xs n)) k"
    using assms 
    by (simp add: "1" inth_imap nfilter_inth_n_zero) 
 show ?thesis 
 using "1" "2" interval_eq_inth_eq by blast
qed


lemma nfilter_count:
  assumes "(\<exists> x \<in> iset xs. P x)"
  shows   " card {( inth (nfilter P xs n) k)| k. k \<le> ilen (nfilter P xs n)} = 
            ilen (nfilter P xs n) +1"
using assms nfilter_card[of "xs" "P" "n"]   iset_nfilter[of "xs" "P" "n"]   
 inth_and_iset by (simp add: iset_inth)

lemma nfilter_holds:
 assumes "(\<exists> x \<in> iset xs. P x)"
  shows  " (\<forall> x \<in> iset (nfilter P xs n). P (inth xs (x-n)))"
using assms by auto


lemma nfilter_holds_not:
 assumes "(\<exists> x \<in> iset xs. P x)"
  shows  " (\<forall> x \<in> ({i+n| i. i\<le> ilen xs} - (iset (nfilter P xs n))). \<not> P (inth xs (x-n)))"
using assms by auto

lemma nfilter_holds_a:
 assumes "(\<exists> x \<in> iset xs. P x)"
  shows  " (\<forall> i\<le>ilen xs.  (i+n) \<in> iset (nfilter P xs n) \<longrightarrow> P (inth xs i))"
using assms by auto

lemma nfilter_holds_not_a:
 assumes "(\<exists> x \<in> iset xs. P x)"
  shows  " (\<forall> i\<le> ilen xs.  P (inth xs i) \<longrightarrow> (i+n) \<in> iset (nfilter P xs n))"
using assms by auto

lemma nfilter_holds_b:
 assumes "(\<exists> x \<in> iset xs. P x)"
  shows  " (\<forall> i\<le>ilen xs.  (i+n) \<in> iset (nfilter P xs n) = P (inth xs i))"
using assms by auto

lemma nfilter_holds_c:
 assumes "(\<exists> x \<in> iset xs. P x)"
         " i\<le>ilen xs "
  shows  "   (i+n) \<in> iset (nfilter P xs n) = P (inth xs i)"
by (simp add: assms(1) assms(2))

lemma nfilter_holds_d:
 assumes "(\<exists> x \<in> iset xs. P x)"
         " n \<le> i "
         " i \<le>ilen xs +n "
  shows  "   i \<in> iset (nfilter P xs n) = P (inth xs (i-n))"
using assms 
by (metis diff_add le_diff_conv nfilter_holds_b)

lemma nfilter_holds_not_b:
 assumes "(\<exists> x \<in> iset xs. P x)"
         " n \<le> i "
         " i \<le>ilen xs +n "
  shows  "   i \<notin> iset (nfilter P xs n) = (\<not> P (inth xs (i-n)))"
using assms by auto

lemma nfilter_disjoint_iset_coset:
 assumes "(\<exists> x \<in> iset xs. P x)"
  shows  " ({i+n| i. i\<le> ilen xs} - (iset (nfilter P xs n))) \<inter> (iset (nfilter P xs n)) = {}"
using assms by auto

lemma nfilter_not_before:
 assumes "(\<exists> x \<in> iset xs. P x)"
         " i< (inth (nfilter P xs 0) 0)" 
 shows   " \<not> P (inth xs i)"
proof -
 have 0: "(inth (nfilter P xs 0) 0) \<le> ilen xs"
      by (metis add.left_neutral assms(1) ilen_gr_zero nfilter_upper_bound)
 have 1: " i \<notin> iset (nfilter P xs 0)"
   using assms 
   proof (induction xs arbitrary: i)
   case (INil x)
   then show ?case by simp
   next
   case (ICons x1a xs)
   then show ?case  
     by (metis idx_nfilter idx_greater inth_and_iset leD)
   qed
 have 2: "i \<notin> iset (nfilter P xs 0) \<and> i\<le> ilen xs \<longrightarrow> \<not> P (inth xs (i))" 
      by (metis add.right_neutral nfilter_holds_not_a inth_iset)
 have 3: "i \<le> ilen xs"
      using "0" assms(2) by linarith
 from 0 1 2 3 show ?thesis by auto
qed    

lemma nfilter_n_not_before:
 assumes "(\<exists> x \<in> iset (suffix n xs). P x)"
         "n \<le> ilen xs" 
         "n \<le>i"
         "i< (inth (nfilter P (suffix n xs) n) 0)"
 shows   "\<not> P (inth xs (i))"
proof -
 have 0: "(inth (nfilter P (suffix n xs) n) 0)  \<le> ilen xs "
     by (metis assms(1) assms(2) ilen_gr_zero suffix_ilen_good 
         le_add_diff_inverse nfilter_upper_bound)
 have 1: " i \<notin> iset (nfilter P (suffix n xs) n)"
   using assms 
   proof (induction xs arbitrary: i)
   case (INil x)
   then show ?case by simp
   next
   case (ICons x1a xs)
   then show ?case 
    by (metis idx_nfilter idx_greater inth_and_iset leD)
   qed
 have 2: "i \<notin> iset (nfilter P (suffix n xs) n) \<and> n \<le> i \<and> i\<le> ilen xs  \<longrightarrow> \<not> P (inth xs (i))" 
     using assms nfilter_holds_not_a[of "(suffix n xs)" "P" "n"] 
     by (metis add_le_imp_le_right inth_suffix suffix_ilen le_add_diff_inverse2 
         ordered_cancel_comm_monoid_diff_class.add_diff_inverse)  
 have 3: "n \<le> i \<and> i\<le> ilen xs" 
     using "0" assms(3) assms(4) by linarith
 from 0 1 2 3 show ?thesis by auto
qed   

lemma nfilter_not_after:
 assumes "(\<exists> x \<in> iset xs. P x)"
         "(inth (nfilter P xs 0) (ilen (nfilter P xs 0))) < i"
         "  i \<le> ilen xs " 
 shows   "  \<not> P (inth xs (i))"
proof -
 have 1: "i \<notin> iset (nfilter P xs 0)" 
   using assms 
     proof (induction xs arbitrary: i)
     case (INil x)
     then show ?case by auto
     next
     case (ICons x1a xs)
     then show ?case 
        by (metis idx_nfilter_gr_eq inth_and_iset leD le_refl)
   qed
 have 2: "i \<notin> iset (nfilter P xs 0) \<and> i \<le> ilen xs \<longrightarrow>\<not> P (inth xs (i)) " 
   by (metis add.right_neutral nfilter_holds_b inth_iset)
 have 3: "i \<le> ilen xs" 
    by (simp add: assms(3))
 from 1 2 3 show ?thesis by auto
qed

lemma nfilter_n_not_after:
 assumes "(\<exists> x \<in> iset (suffix n xs). P x)"
         "n \<le> ilen xs" 
         "(inth (nfilter P (suffix n xs) n) (ilen (nfilter P (suffix n xs) n))) < i"
         " i \<le> ilen xs "
 shows   " \<not> P (inth xs (i))"
proof -
 have 1: "i \<notin> iset (nfilter P (suffix n xs) n)" 
   using assms 
     proof (induction xs arbitrary: i)
     case (INil x)
     then show ?case by auto
     next
     case (ICons x1a xs)
     then show ?case 
         by (metis idx_nfilter_gr_eq inth_and_iset leD le_eq_less_or_eq)
     qed
 have 2: "i \<notin> iset (nfilter P (suffix n xs) n) \<and> n\<le> i \<and> i \<le> ilen xs \<longrightarrow>\<not> P (inth xs (i)) " 
       using assms nfilter_holds_not_a[of "suffix n xs" "P" "n"] 
       by (metis add_le_imp_le_left inth_suffix suffix_ilen le_add_diff_inverse2
        ordered_cancel_comm_monoid_diff_class.add_diff_inverse)   
 have 3: "n\<le> i \<and> i \<le> ilen xs"
     by (meson assms(1) assms(3) assms(4) dual_order.strict_implies_order dual_order.strict_trans2 
         nfilter_inth_n_zero_a order_refl)
 from 1 2 3 show ?thesis by auto
qed


lemma nfilter_not_between_help_a:
assumes "(\<And>k i. 
            \<exists>a\<in>iset xs. P a \<Longrightarrow>
            k < ilen (nfilter P xs 0) \<Longrightarrow>
            inth (nfilter P xs 0) k < i \<Longrightarrow>
            i < inth (nfilter P xs 0) (Suc k) \<Longrightarrow> 
            inth (nfilter P xs 0) (Suc k) \<le> ilen xs \<Longrightarrow> 
            i \<notin> iset (nfilter P xs 0))"
         "\<exists>a\<in>iset (x1a \<odot> xs). P a"
         "k < ilen (nfilter P (x1a \<odot> xs) 0)"
         "inth (nfilter P (x1a \<odot> xs) 0) k < i "
         "i < inth (nfilter P (x1a \<odot> xs) 0) (Suc k) "
         "inth (nfilter P (x1a \<odot> xs) 0) (Suc k) \<le> ilen (x1a \<odot> xs)"
         "\<exists>x\<in>iset xs. P x"
         "P x1a "
   shows     " i \<notin> iset (nfilter P (x1a \<odot> xs) 0)"
proof -
 have 1: "k=0 \<Longrightarrow> i \<notin> iset (nfilter P (x1a \<odot> xs) 0)"
     using assms by auto 
      (metis One_nat_def Suc_less_eq2 diff_Suc_1 ilen_gr_zero nfilter_not_before 
     nfilter_inth_n_zero)
 have 2: "\<And>n. k= (Suc n) \<Longrightarrow>  i \<notin> iset (nfilter P (x1a \<odot> xs) 0)"
   using assms 
   proof auto
    fix n
    fix x
    fix ka
    assume a0: "k= Suc n"
    assume a1: "(\<And>k i. k < ilen (nfilter P xs 0) \<Longrightarrow>
                  inth (nfilter P xs 0) k < i \<Longrightarrow>
                  i < inth (nfilter P xs 0) (Suc k) \<Longrightarrow>
                  inth (nfilter P xs 0) (Suc k) \<le> ilen xs \<Longrightarrow> 
                  \<not> P (inth xs i))" 
    assume a2: "n < ilen (nfilter P xs (Suc 0))"
    assume a3: " inth (nfilter P xs (Suc 0)) n < Suc ka"  
    assume a4: "Suc ka < inth (nfilter P xs (Suc 0)) (Suc n)" 
    assume a5: "inth (nfilter P xs (Suc 0)) (Suc n) \<le> Suc (ilen xs)"
    assume a6: "P x1a"
    assume a7: "x \<in> iset xs"
    assume a8: "P x"
    assume a9: "i = Suc ka"   
    assume a10: "P (inth xs ka)"
    show "False"
     proof -
      have 3: "n < ilen (nfilter P xs 0) \<Longrightarrow>
               inth (nfilter P xs 0) n < ka \<Longrightarrow>
               ka < inth (nfilter P xs 0) (Suc n) \<Longrightarrow> 
               inth (nfilter P xs 0) (Suc n) \<le> ilen xs \<Longrightarrow> 
               \<not> P (inth xs ka)"
         using a1[of "n" "ka"] by auto
      have 4: "n < ilen (nfilter P xs 0)" 
        by (metis a2 a7 a8 nfilter_ilen_n_zero)
      have 5: "inth (nfilter P xs 0) (Suc n) \<le> ilen xs" 
        by (metis "4" Suc_leI add.commute add.right_neutral assms(7) nfilter_upper_bound)
      have 6: "\<exists>x\<in>iset xs. P x" 
         using assms(7) by auto
      have 7: "Suc 0 \<le> ilen (nfilter P xs (Suc n)) " 
        by (metis One_nat_def Suc_le_mono Suc_mono a2 a7 a8 ilen_gr_zero le_SucE 
            nfilter_ilen_n_zero not_add_less1 plus_1_eq_Suc)
      have 8: "ka < inth (nfilter P xs 0) (Suc n)" 
         using nfilter_inth_n_zero[of "xs" "P" "Suc n" "Suc 0"]  a4 "6" a2 by linarith
      have 9: "inth (nfilter P xs 0) n < ka"
         using  a2 a3 a7 a8 nfilter_inth_n_zero[of "xs" "P" "n" "Suc 0" ]
         by (metis One_nat_def add.commute dual_order.strict_implies_order less_diff_conv2 
             nfilter_lower_bound plus_1_eq_Suc)  
      have 10: "  \<not> P (inth xs ka) "
          using 3 4 9 8 5 by auto
      from a10 10 show ?thesis by auto
   qed
  qed
 show ?thesis using 1 2 
 using gr0_implies_Suc by blast 
qed

lemma nfilter_not_between_help_b:
assumes "(\<And>k i. 
     \<exists>a\<in>iset xs. P a \<Longrightarrow>
     k < ilen (nfilter P xs 0) \<Longrightarrow>
     inth (nfilter P xs 0) k < i \<Longrightarrow>
     i < inth (nfilter P xs 0) (Suc k) \<Longrightarrow> 
     inth (nfilter P xs 0) (Suc k) \<le> ilen xs \<Longrightarrow> 
     i \<notin> iset (nfilter P xs 0))"
    "\<exists>a\<in>iset (x1a \<odot> xs). P a"
    "k < ilen (nfilter P (x1a \<odot> xs) 0)"
    "inth (nfilter P (x1a \<odot> xs) 0) k < i "
    "i < inth (nfilter P (x1a \<odot> xs) 0) (Suc k) "  
    "inth (nfilter P (x1a \<odot> xs) 0) (Suc k) \<le> ilen (x1a \<odot> xs)"
    "\<exists>x\<in>iset xs. P x"
    "\<not> P x1a "
 shows " i \<notin> iset (nfilter P (x1a \<odot> xs) 0)" 
proof -
 have 1: "k=0 \<Longrightarrow> i \<notin> iset (nfilter P (x1a \<odot> xs) 0)"
    using assms  
    proof auto
     fix x :: 'a and ka :: nat
     assume a1: "inth (nfilter P xs (Suc 0)) 0 < Suc ka"
     assume a2: "x \<in> iset xs"
     assume a3: "P x"
     assume a4: "\<And>k i. \<lbrakk>k < ilen (nfilter P xs 0); inth (nfilter P xs 0) k < i; 
                        i < inth (nfilter P xs 0) (Suc k); inth (nfilter P xs 0) (Suc k) \<le> ilen xs\<rbrakk> 
                        \<Longrightarrow> \<not> P (inth xs i)"
     assume a5: "P (inth xs ka)"
     assume a6: "0 < ilen (nfilter P xs (Suc 0))"
     assume a7: "Suc ka < inth (nfilter P xs (Suc 0)) (Suc 0)"
     assume a8: "inth (nfilter P xs (Suc 0)) (Suc 0) \<le> Suc (ilen xs)"
     have f9: "\<exists>a. a \<in> iset xs \<and> P a"
     using a3 a2 by blast 
     then have f10: "0 \<le> ilen (nfilter P xs (Suc 0)) \<longrightarrow> inth (nfilter P xs (Suc 0)) 0 \<noteq> 0"
     by (metis le_zero_eq nfilter_lower_bound not_less_eq_eq) 
     have f11: "ilen (nfilter P xs (Suc 0)) = ilen (nfilter P xs 0)"
     using f9 by (meson nfilter_ilen_n_zero) 
     obtain nn :: "nat \<Rightarrow> nat \<Rightarrow> nat" where
     f12: "inth (nfilter P xs (Suc 0)) (Suc 0) = Suc (nn (inth (nfilter P xs (Suc 0)) (Suc 0)) ka) \<and> 
           ka < nn (inth (nfilter P xs (Suc 0)) (Suc 0)) ka"
     using a7 by (meson Suc_less_eq2) 
     then have "inth (nfilter P xs 0) (Suc 0) = nn (inth (nfilter P xs (Suc 0)) (Suc 0)) ka"
     using f9 a6 by (metis One_nat_def diff_Suc_1 le_zero_eq neq0_conv 
                     nfilter_inth_n_zero not_less_eq_eq) 
     then have "\<not> 0 \<le> ilen (nfilter P xs 0)"
     using f12 f11 f10 f9 a8 a6 a5 a4 a1 
      by (metis One_nat_def Suc_le_mono diff_Suc_1 less_Suc_eq_0_disj nfilter_inth_n_zero) 
     then show False 
     by blast 
     qed
  have 2: "\<And>n. k= (Suc n) \<Longrightarrow>  i \<notin> iset (nfilter P (x1a \<odot> xs) 0)"    
    using assms
    proof auto
     fix n
     fix x
     fix ka
     assume a0: " k = Suc n"
     assume a1: "(\<And>k i. k < ilen (nfilter P xs 0) \<Longrightarrow>
                   inth (nfilter P xs 0) k < i \<Longrightarrow>
                   i < inth (nfilter P xs 0) (Suc k) \<Longrightarrow>
                   inth (nfilter P xs 0) (Suc k) \<le> ilen xs \<Longrightarrow> 
                   \<not> P (inth xs i)) "  
     assume a2: "Suc n < ilen (nfilter P xs (Suc 0))"
     assume a3: "inth (nfilter P xs (Suc 0)) (Suc n) < Suc ka"  
     assume a4: "Suc ka < inth (nfilter P xs (Suc 0)) (Suc (Suc n))"
     assume a5: "inth (nfilter P xs (Suc 0)) (Suc (Suc n)) \<le> Suc (ilen xs)"  
     assume a6: "\<not> P x1a"
     assume a7: "x \<in> iset xs"
     assume a8: "P x"
     assume a9: "i = Suc ka"   
     assume a10: "P (inth xs ka)"
     show "False"
     proof -
      have 3: "Suc n < ilen (nfilter P xs 0) \<Longrightarrow>
               inth (nfilter P xs 0) (Suc n) < ka \<Longrightarrow>
               ka < inth (nfilter P xs 0) (Suc (Suc n)) \<Longrightarrow> 
               inth (nfilter P xs 0) (Suc (Suc n)) \<le> ilen xs \<Longrightarrow> 
               \<not> P (inth xs ka)"
         using a1[of "Suc n" "ka"] by auto
      have 4: "(Suc n) < ilen (nfilter P xs 0)"  
          by (metis a2 a7 a8 nfilter_ilen_n_zero)
      have 5:  "inth (nfilter P xs 0) (Suc (Suc n)) \<le> ilen xs" 
        by (metis "4" Suc_leI a7 a8 add.left_neutral nfilter_upper_bound) 
      have 6: "\<exists>x\<in>iset xs. P x" 
        by (simp add: assms(7))
      have 7: "Suc 0 \<le> ilen (nfilter P xs (Suc n)) " 
        by (metis One_nat_def Suc_le_mono Suc_mono a2 a7 a8 ilen_gr_zero le_SucE 
            nfilter_ilen_n_zero not_add_less1 plus_1_eq_Suc)
      have 8: "ka < inth (nfilter P xs 0) (Suc (Suc n))" 
          using nfilter_inth_n_zero[of "xs" "P" "Suc (Suc n)" "Suc 0"]  a4 6 a2 by linarith
      have 9: "inth (nfilter P xs 0) (Suc n) < ka"
          using  a2 a3 a7 a8 nfilter_inth_n_zero[of "xs" "P" "Suc n" "Suc 0" ]
          by (metis One_nat_def add.commute dual_order.strict_implies_order less_diff_conv2
              nfilter_lower_bound plus_1_eq_Suc)
      have 10: "\<not> P (inth xs ka)"
         using 3 4 9 8 5 by auto
      from a10 10 show ?thesis by auto
     qed
    qed
  show ?thesis using 1 2 
  using gr0_implies_Suc by blast 
qed  


lemma nfilter_not_between_help:
 assumes "(\<exists> x \<in> iset xs. P x)"
         "k < ilen (nfilter P xs 0)"
         "(inth (nfilter P xs 0) k ) < i"  
         "i < (inth (nfilter P xs 0) (Suc k) )" 
         "(inth (nfilter P xs 0) (Suc k) ) \<le> ilen xs" 
 shows "i \<notin> iset (nfilter P xs 0)"  
using assms 
proof (induction xs arbitrary: i k)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
    proof (cases "(\<exists> x \<in> iset xs. P x)")
    case True
    then show ?thesis 
      proof  (cases "P x1a")
      case True
      then show ?thesis using ICons True using nfilter_not_between_help_a[of "xs" "P" "x1a" "k" "i" ]
          by fastforce
      next
      case False
      then show ?thesis using ICons False using nfilter_not_between_help_b[of "xs" "P" "x1a" "k" "i"]
         by simp
      qed 
    next
    case False
    then show ?thesis using ICons False by auto
    qed
qed

lemma nfilter_not_between:
 assumes "(\<exists> x \<in> iset xs. P x)"
         " (inth (nfilter P xs 0) k ) < i "
         " i < (inth (nfilter P xs 0) (Suc k) ) " 
         " k < ilen (nfilter P xs 0)  " 
  shows  " \<not> P (inth xs (i))"
proof -
 have 0: "(inth (nfilter P xs 0) (Suc k) ) \<le> ilen xs "
      by (metis Suc_leI add_cancel_right_left assms(1) assms(4) nfilter_upper_bound)
 have 1: " i \<le> ilen xs" 
      using "0" assms(3) by linarith
 have 2: " k < ilen (nfilter P xs 0) \<and> (inth (nfilter P xs 0) k ) < i \<and>
            i < (inth (nfilter P xs 0) (Suc k) )  \<and> (inth (nfilter P xs 0) (Suc k) ) \<le> ilen xs\<longrightarrow>
           i \<notin> iset (nfilter P xs 0)" 
   using assms(1) 
    proof (induction xs arbitrary: i k)
    case (INil x)
    then show ?case by simp
    next
    case (ICons x1a xs)
    then show ?case using assms nfilter_not_between_help 
    by metis    
    qed  
 have 3 : "i \<notin> iset (nfilter P xs 0) \<and>  i \<le> ilen xs \<longrightarrow>\<not> P (inth xs (i))" 
    by (metis add.right_neutral nfilter_holds_b inth_iset)
 from 0 1 2 3  show ?thesis using assms by blast
qed



lemma idx_imp_idistinct:
 assumes "index_sequence (inth xs 0) xs"
 shows   "idistinct xs"
using assms 
proof (induction xs)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
 by (auto simp add: idx_expand1)
    (metis idx_expand1 idx_greater_first inth_and_iset le_zero_eq not_less not_less_iff_gr_or_eq)
qed

lemma idx_iset_eq:
 assumes "index_sequence (inth xs 0) xs"
         "index_sequence (inth ys 0) ys"
         "iset xs = iset ys"
 shows   "xs = ys"
using assms 
proof 
 (induction xs arbitrary: ys)
 case (INil x)
 then show ?case
 by (metis idistinct.simps(2) iapp_INil iapp_not_INil idx_imp_idistinct 
     interval.exhaust interval_in_iset_conv_decomp_last) 
 next
 case (ICons x1a xs)
 then show ?case 
   proof (cases ys)
   case (INil x1)
   then show ?thesis 
   using ICons.prems(1) ICons.prems(3) idx_less_last_1 le_eq_less_or_eq inth_iset by fastforce
   next
   case (ICons x21 x22)
   then show ?thesis  
    proof (cases "x1a = x21")
    case True
    then show ?thesis
     proof -
      have 1: "ilen (x1a\<odot>xs) = ilen (x21\<odot>x22)" 
           by (metis ICons.prems(1) ICons.prems(2) ICons.prems(3) idistinct_card idx_imp_idistinct 
               ICons nat.inject)
      have 2: "index_sequence (inth xs 0) xs"
           using ICons.prems(1) idx_expand1 by auto
      have 3: "index_sequence (inth x22 0) x22"
           using ICons.prems(2) idx_expand1 ICons by auto
      have 4: "idistinct xs "
           using "2" idx_imp_idistinct by auto
      have 5: "idistinct x22"
           by (simp add: "3" idx_imp_idistinct)
      have 6: " iset (x1a\<odot>xs) = {x1a} \<union> iset xs"
           by auto
      have 7: "x1a \<notin> iset xs"
           by (meson ICons.prems(1) idistinct.simps(2) idx_imp_idistinct) 
      have 8: "iset (x21\<odot>x22) = {x21} \<union> iset x22 "
           by auto
      have 9: "x21 \<notin> iset x22"
           using ICons.prems(2) idx_imp_idistinct ICons by fastforce
      have 10: "iset xs =iset x22"
           using "7" "9" ICons.prems(3) True ICons by fastforce
      have 11: "xs = x22" 
           using "10" "2" "3" ICons.IH by blast
      have 12: "x1a\<odot> xs = x21\<odot>x22 "
           by (simp add: "11" True)
      show ?thesis by (simp add: "12" ICons)
     qed
    next
    case False
    then show ?thesis using ICons idx_imp_idistinct interval_hd_in_iset idx_expand1
    by (metis (full_types) ICons.prems(1) ICons.prems(2) ICons.prems(3) idistinct.simps(2) 
          inth_zero linorder_neqE_nat) 
   qed
  qed
qed

lemma ifilter_nfilter_prefix_idx_a:
  assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
  shows   "  index_sequence (inth (prefix k (nfilter P xs 0)) 0) (prefix k (nfilter P xs 0))"
using assms by (auto simp add: index_sequence_def)
               (metis diff_zero ifilter_nfilter_prefix_ilen_0 idx_nfilter_mono)

lemma ifilter_nfilter_prefix_idx_a_1:
  assumes "\<exists> x \<in> iset xs. P x"
           "k \<le> ilen (ifilter P xs)" 
  shows   "  index_sequence (inth (prefix k (nfilter P xs 0)) 0) (prefix k (nfilter P xs 0))"
using assms by (auto simp add: index_sequence_def)
               (meson idx_nfilter_mono)

lemma ifilter_nfilter_suffix_idx_a:
  assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
  shows   "  index_sequence (inth (suffix k (nfilter P xs 0)) 0) (suffix k (nfilter P xs 0))"
using assms by (simp add: index_sequence_def)
 (metis add.commute diff_zero ifilter_nfilter_prefix_ilen_0 idx_nfilter_mono less_diff_conv)

lemma ifilter_nfilter_suffix_idx_a_1:
  assumes "\<exists> x \<in> iset xs. P x"
           "k \<le> ilen (ifilter P xs)" 
  shows   "  index_sequence (inth (suffix k (nfilter P xs 0)) 0) (suffix k (nfilter P xs 0))"
using assms
by (simp add: index_sequence_def idx_nfilter_mono nfilter_ilen)

lemma ifilter_nfilter_prefix_idx_b:
  assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
  shows   "index_sequence (inth (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0) 0)
             (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0)"
using assms idx_nfilter[of "(prefix (inth (nfilter P xs 0) k) xs)" "P" "0"] 
by (metis add.right_neutral ilast_ifirst inth_suffix  le0 le_refl inth_iset)

lemma ifilter_nfilter_prefix_idx_b_1:
  assumes "\<exists> x \<in> iset xs. P x"
           "k \<le> ilen (ifilter P xs)" 
  shows   "index_sequence (inth (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0) 0)
             (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0)"
using assms 
by (metis ifilter_nfilter_prefix_idx_b nfilter_holds nfilter_ilen nfilter_inth_n_zero inth_iset)


lemma ifilter_nfilter_suffix_idx_b:
 assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
 shows   "index_sequence (inth (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs)
                              (inth (nfilter P xs 0) k)) 0)
             (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 
                        (inth (nfilter P xs 0) k))"
using assms 
by (auto simp add: index_sequence_def)
 (metis idx_nfilter_mono ifirst_suffix ilen_gr_zero  
   suffix_ilen_code ilen_nfilter_le less_le_trans not_less0 inth_iset)

lemma ifilter_nfilter_suffix_idx_b_1:
  assumes "\<exists> x \<in> iset xs. P x"
           "k \<le> ilen (ifilter P xs)" 
  shows   "index_sequence (inth (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs)
                                          (inth (nfilter P xs 0) k)) 0)
             (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) (inth (nfilter P xs 0) k))"
using assms 
by (metis ifilter_nfilter_suffix_idx_b nfilter_holds nfilter_ilen nfilter_inth_n_zero inth_iset)

lemma ifilter_nfilter_prefix_iset_eq:
  assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
  shows   " iset (prefix k (nfilter P xs 0)) =
            iset (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0)"
proof -
 have 1: "(inth (nfilter P xs 0) k)  \<le> ilen xs"
    by (metis assms(1) assms(2) diff_zero ifilter_nfilter_prefix_ilen_0 ilen_gr_zero 
    nfilter_ilen nfilter_upper_bound ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 2: "\<exists>x \<in> iset xs . P x "
    using "1" assms(1) inth_iset by blast 
 have 3: "{ i. i\<le> ilen(prefix ((inth (nfilter P xs 0) k)) xs) \<and> 
                    P (inth (prefix ((inth (nfilter P xs 0) k)) xs) i)} =

          { i. i\<le> (inth (nfilter P xs 0) k) \<and> P (inth xs i)}"
   using "1" by auto 
 have 4: "\<exists> x \<in> iset(prefix ((inth (nfilter P xs 0) k))    xs). P x"
  by (metis "1" assms(1) inth_prefix prefix_ilen_good inth_iset order_refl) 
 have 5: "{ i. i\<le> (inth (nfilter P xs 0) k) \<and> P (inth xs i)} = 
           { i. i\<le> (inth (nfilter P xs 0) k) \<and> i\<in>iset(nfilter P xs 0)}"
    using "4" "1" "2" nfilter_holds_b by auto
 have 6: "{ i. i\<le> (inth (nfilter P xs 0) k) \<and> i\<in>iset(nfilter P xs 0)} =
           { i. i\<le> (inth (nfilter P xs 0) k) \<and> 
                i\<in>{(inth (nfilter P xs 0) j)| j. j\<le> ilen(nfilter P xs 0)}   }"
      by (auto simp add: inth_and_iset)
 have 7: "{ i. i\<le> (inth (nfilter P xs 0) k) \<and>
               i\<in>{(inth (nfilter P xs 0) j)| j. j\<le> ilen(nfilter P xs 0)}   } =
           { (inth (nfilter P xs 0) j) | j. j\<le> k }"
    using assms "2" by (auto simp add: nfilter_ilen, 
        metis dual_order.antisym idx_nfilter idx_less_eq le_cases nfilter_ilen,
        metis idx_nfilter_less_eq nfilter_ilen)
 have 8: "k\<le> ilen (nfilter P xs 0)"
    by (simp add: "2" assms(2) nfilter_ilen)
 have 9: "{ (inth (nfilter P xs 0) j) | j. j\<le> k } = iset (prefix k (nfilter P xs 0))"
    using "8" "2" using prefix_iset by force
 have 10: "iset (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0) =
            { i. i\<le> ((inth (nfilter P xs 0) k)) \<and> 
                    P (inth (prefix ((inth (nfilter P xs 0) k)) xs) i)}"
   using "4" "1"  nfilter_prefix_iset by auto
 have 11: "ilen(prefix ((inth (nfilter P xs 0) k)) xs) = ((inth (nfilter P xs 0) k))"
   using "1" prefix_ilen_good by blast
 have 12: "{ i. i\<le> ((inth (nfilter P xs 0) k)) \<and> 
                    P (inth (prefix ((inth (nfilter P xs 0) k)) xs) i)} =
            { i. i\<le> ilen(prefix ((inth (nfilter P xs 0) k)) xs) \<and> 
                    P (inth (prefix ((inth (nfilter P xs 0) k)) xs) i)}"
    using "11" by auto
 show ?thesis 
 using "10" "12" "3" "5" "6" "7" "9" by auto
qed

lemma ifilter_nfilter_prefix_iset_eq_1:
  assumes "\<exists>x \<in> iset xs . P x"
           "k \<le> ilen (ifilter P xs)" 
  shows   " iset (prefix k (nfilter P xs 0)) =
            iset (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0)"
proof -
 have 1: "(inth (nfilter P xs 0) k)  \<le> ilen xs"
    by (metis assms(1) assms(2) diff_zero  ilen_gr_zero 
    nfilter_ilen nfilter_upper_bound ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 2: "\<exists>x \<in> iset xs . P x "
    using "1" assms(1) inth_iset by blast 
 have 3: "{ i. i\<le> ilen(prefix ((inth (nfilter P xs 0) k)) xs) \<and> 
                    P (inth (prefix ((inth (nfilter P xs 0) k)) xs) i)} =

          { i. i\<le> (inth (nfilter P xs 0) k) \<and> P (inth xs i)}"
   using "1" by auto 
 have 4: "\<exists> x \<in> iset(prefix ((inth (nfilter P xs 0) k))    xs). P x"
    using assms "1" "2" nfilter_holds[of "xs" "P" "0"] 
    by (metis ilast_prefix le_refl nfilter_ilen nfilter_inth_n_zero inth_iset) 
 have 5: "{ i. i\<le> (inth (nfilter P xs 0) k) \<and> P (inth xs i)} = 
           { i. i\<le> (inth (nfilter P xs 0) k) \<and> i\<in>iset(nfilter P xs 0)}"
    using "4" "1" "2" nfilter_holds_b by auto
 have 6: "{ i. i\<le> (inth (nfilter P xs 0) k) \<and> i\<in>iset(nfilter P xs 0)} =
           { i. i\<le> (inth (nfilter P xs 0) k) \<and> 
                i\<in>{(inth (nfilter P xs 0) j)| j. j\<le> ilen(nfilter P xs 0)}   }"
    by (auto simp add: inth_and_iset)
 have 7: "{ i. i\<le> (inth (nfilter P xs 0) k) \<and> 
               i\<in>{(inth (nfilter P xs 0) j)| j. j\<le> ilen(nfilter P xs 0)}   } =
           { (inth (nfilter P xs 0) j) | j. j\<le> k }"
   using assms "2" by (auto simp add: nfilter_ilen, 
        metis dual_order.antisym idx_nfilter idx_less_eq le_cases nfilter_ilen,
        metis idx_nfilter_less_eq nfilter_ilen)
 have 8: "k\<le> ilen (nfilter P xs 0)"
    by (simp add: "2" assms(2) nfilter_ilen)
 have 9: "{ (inth (nfilter P xs 0) j) | j. j\<le> k } = iset (prefix k (nfilter P xs 0))"
    using "8" "2" using prefix_iset by force
 have 10: "iset (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0) =
            { i. i\<le> ((inth (nfilter P xs 0) k)) \<and> 
                    P (inth (prefix ((inth (nfilter P xs 0) k)) xs) i)}"
   using "4" "1"  nfilter_prefix_iset by auto
 have 11: "ilen(prefix ((inth (nfilter P xs 0) k)) xs) = ((inth (nfilter P xs 0) k))"
   using "1" prefix_ilen_good by blast
 have 12: "{ i. i\<le> ((inth (nfilter P xs 0) k)) \<and> 
                    P (inth (prefix ((inth (nfilter P xs 0) k)) xs) i)} =
            { i. i\<le> ilen(prefix ((inth (nfilter P xs 0) k)) xs) \<and> 
                    P (inth (prefix ((inth (nfilter P xs 0) k)) xs) i)}"
    using "11" by auto
 show ?thesis 
 using "10" "12" "3" "5" "6" "7" "9" by auto
qed

lemma ifilter_nfilter_suffix_iset_eq:
  assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
  shows   " iset (nfilter P (suffix (inth (nfilter P xs 0) k)  xs) (inth (nfilter P xs 0) k)) =
            iset (suffix k (nfilter P xs 0)) 
            "
proof -
 have 1: "(inth (nfilter P xs 0) k)  \<le> ilen xs"
    by (metis assms(1) assms(2) diff_zero ifilter_nfilter_prefix_ilen_0 ilen_gr_zero 
    nfilter_ilen nfilter_upper_bound ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 2: "\<exists>x \<in> iset xs . P x "
    using "1" assms(1) inth_iset by blast 
 have 4: "\<exists> x \<in> iset(suffix ((inth (nfilter P xs 0) k))    xs). P x"
    using "1" assms(1) inth_and_iset by force
 have 10: "iset (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) (inth (nfilter P xs 0) k)) =
            { (inth (nfilter P xs 0) k)+i| i. i\<le> ilen xs - (inth (nfilter P xs 0) k) \<and> 
                    P (inth xs (i+(inth (nfilter P xs 0) k)) )}
          "
   using nfilter_suffix_iset_b[of "((inth (nfilter P xs 0) k))" "xs" "P" "(inth (nfilter P xs 0) k)"] 
    using "1" "4" by blast 
 have 5: " { (inth (nfilter P xs 0) k)+i|i.  i\<le> ilen xs - (inth (nfilter P xs 0) k) \<and> 
              P (inth xs (i+(inth (nfilter P xs 0) k))  )}
            = 
           { (inth (nfilter P xs 0) k)+i|i. i\<le> ilen xs - (inth (nfilter P xs 0) k) \<and> 
              i+(inth (nfilter P xs 0) k)\<in>iset(nfilter P xs 0)}"
    using "4" "1" "2" nfilter_holds_b by auto
 have 51: "{ (inth (nfilter P xs 0) k)+i |i. i\<le> ilen xs - (inth (nfilter P xs 0) k) \<and> 
               i+(inth (nfilter P xs 0) k)\<in>iset(nfilter P xs 0)} =
           {  (inth (nfilter P xs 0) k)+i| i. 
                (inth (nfilter P xs 0) k) \<le> i + (inth (nfilter P xs 0) k) \<and>
                i + (inth (nfilter P xs 0) k)\<le> ilen xs \<and> 
                i+(inth (nfilter P xs 0) k)\<in>iset(nfilter P xs 0)}
          " 
          using "1" by auto
 have 52: "{ (inth (nfilter P xs 0) k)+i| i. 
                (inth (nfilter P xs 0) k) \<le> i + (inth (nfilter P xs 0) k) \<and>
                i + (inth (nfilter P xs 0) k)\<le> ilen xs \<and> 
                i+(inth (nfilter P xs 0) k)\<in>iset(nfilter P xs 0)} =
           { j. (inth (nfilter P xs 0) k) \<le> j \<and> j\<le> ilen xs \<and> j\<in>iset(nfilter P xs 0)}
          " 
      by (metis (no_types, lifting) add_diff_cancel_right' le0 le_add_diff_inverse 
          le_add_same_cancel2)
 have 53: "{ j. (inth (nfilter P xs 0) k) \<le> j \<and> j\<le> ilen xs \<and> j\<in>iset(nfilter P xs 0)} =
           { j. (inth (nfilter P xs 0) k) \<le> j \<and>
                j\<le> ilen xs \<and> j\<in>{(inth (nfilter P xs 0) jj)| jj. jj\<le> ilen(nfilter P xs 0)}  }
          " 
     by (auto simp add: inth_and_iset)
 have 54: "{ j. (inth (nfilter P xs 0) k) \<le> j \<and>
                j\<le> ilen xs \<and> j\<in>{(inth (nfilter P xs 0) jj)| jj. jj\<le> ilen(nfilter P xs 0)}  } =
            { (inth (nfilter P xs 0) j) | j. k \<le>j \<and> j \<le> ilen (nfilter P xs 0) }
          " 
  using assms "2" by (auto,
      metis dual_order.antisym idx_nfilter_less_eq le_cases nfilter_ilen,
      metis idx_nfilter_gr_eq,
      metis add_cancel_right_left nfilter_upper_bound)
 have 8: "k\<le> ilen (nfilter P xs 0)"
    by (simp add: "2" assms(2) nfilter_ilen)
 have 9: "{ (inth (nfilter P xs 0) j) | j. k \<le>j \<and> j \<le> ilen (nfilter P xs 0) } = 
          iset (suffix k (nfilter P xs 0))"
    using "8" "2" using suffix_iset_a by blast
 show ?thesis 
 using "10"  "5" "51" "52" "53" "54" "9" by simp 
qed

lemma ifilter_nfilter_suffix_iset_eq_1:
  assumes "\<exists>x \<in> iset xs . P x"
           "k \<le> ilen (ifilter P xs)" 
  shows   " iset (nfilter P (suffix (inth (nfilter P xs 0) k)  xs) (inth (nfilter P xs 0) k)) =
            iset (suffix k (nfilter P xs 0)) 
            "
proof -
 have 1: "(inth (nfilter P xs 0) k)  \<le> ilen xs"
    by (metis assms(1) assms(2) diff_zero ilen_gr_zero 
    nfilter_ilen nfilter_upper_bound ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 2: "\<exists>x \<in> iset xs . P x "
    using  assms(1) by blast 
 have 4: "\<exists> x \<in> iset(suffix ((inth (nfilter P xs 0) k))    xs). P x"
   using nfilter_holds[of "xs" "P" "0"] 
   by (metis "1" "2" assms(2) ifirst_suffix le0 nfilter_ilen nfilter_inth_n_zero inth_iset) 
 have 10: "iset (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) (inth (nfilter P xs 0) k)) =
            { (inth (nfilter P xs 0) k)+i| i. i\<le> ilen xs - (inth (nfilter P xs 0) k) \<and> 
                    P (inth xs (i+(inth (nfilter P xs 0) k)) )}
          "
   using nfilter_suffix_iset_b[of "(inth (nfilter P xs 0) k)" "xs" "P" "(inth (nfilter P xs 0) k)"] 
     using "1" "4" by blast
 have 5: " { (inth (nfilter P xs 0) k)+i|i.  i\<le> ilen xs - (inth (nfilter P xs 0) k) \<and> 
              P (inth xs (i+(inth (nfilter P xs 0) k))  )}
            = 
           { (inth (nfilter P xs 0) k)+i|i. i\<le> ilen xs - (inth (nfilter P xs 0) k) \<and> 
              i+(inth (nfilter P xs 0) k)\<in>iset(nfilter P xs 0)}"
    using "4" "1" "2" nfilter_holds_b by auto
 have 51: "{ (inth (nfilter P xs 0) k)+i |i. i\<le> ilen xs - (inth (nfilter P xs 0) k) \<and> 
               i+(inth (nfilter P xs 0) k)\<in>iset(nfilter P xs 0)} =
           {  (inth (nfilter P xs 0) k)+i| i. 
                (inth (nfilter P xs 0) k) \<le> i + (inth (nfilter P xs 0) k) \<and>
                i + (inth (nfilter P xs 0) k)\<le> ilen xs \<and> 
                i+(inth (nfilter P xs 0) k)\<in>iset(nfilter P xs 0)}
          " 
          using "1" by auto
 have 52: "{ (inth (nfilter P xs 0) k)+i| i. 
                (inth (nfilter P xs 0) k) \<le> i + (inth (nfilter P xs 0) k) \<and>
                i + (inth (nfilter P xs 0) k)\<le> ilen xs \<and> 
                i+(inth (nfilter P xs 0) k)\<in>iset(nfilter P xs 0)} =
           { j. (inth (nfilter P xs 0) k) \<le> j \<and> j\<le> ilen xs \<and> j\<in>iset(nfilter P xs 0)}
          " 
      by (metis (no_types, lifting) add_diff_cancel_right' le0 le_add_diff_inverse 
          le_add_same_cancel2)
 have 53: "{ j. (inth (nfilter P xs 0) k) \<le> j \<and> j\<le> ilen xs \<and> j\<in>iset(nfilter P xs 0)} =
           { j. (inth (nfilter P xs 0) k) \<le> j \<and>
                j\<le> ilen xs \<and> j\<in>{(inth (nfilter P xs 0) jj)| jj. jj\<le> ilen(nfilter P xs 0)}  }
          " 
     by (auto simp add: inth_and_iset)
 have 54: "{ j. (inth (nfilter P xs 0) k) \<le> j \<and>
                j\<le> ilen xs \<and> j\<in>{(inth (nfilter P xs 0) jj)| jj. jj\<le> ilen(nfilter P xs 0)}  } =
            { (inth (nfilter P xs 0) j) | j. k \<le>j \<and> j \<le> ilen (nfilter P xs 0) }
          " 
    using assms "2" by (auto,
      metis  dual_order.antisym idx_nfilter_less_eq le_cases nfilter_ilen,
      simp add: "2" idx_nfilter_less_eq,
      metis add_cancel_right_left nfilter_upper_bound)
 have 8: "k\<le> ilen (nfilter P xs 0)"
    by (simp add: "2" assms(2) nfilter_ilen)
 have 9: "{ (inth (nfilter P xs 0) j) | j. k \<le>j \<and> j \<le> ilen (nfilter P xs 0) } = 
          iset (suffix k (nfilter P xs 0))"
    using "8" "2" using suffix_iset_a by blast
 show ?thesis 
 using "10"  "5" "51" "52" "53" "54" "9" by simp
qed

lemma nfilter_nfilter_prefix:
   assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
   shows   " (prefix k (nfilter P xs 0))   = 
             (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0)"
using assms  
by (meson ifilter_nfilter_prefix_idx_a ifilter_nfilter_prefix_idx_b 
    ifilter_nfilter_prefix_iset_eq idx_iset_eq)


lemma nfilter_nfilter_prefix_1:
   assumes "\<exists>x \<in> iset xs . P x "
           "k \<le> ilen (ifilter P xs)" 
   shows   " (prefix k (nfilter P xs 0))   = 
             (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0)"
using assms  
by (meson ifilter_nfilter_prefix_idx_a_1 ifilter_nfilter_prefix_idx_b_1 
    ifilter_nfilter_prefix_iset_eq_1 idx_iset_eq)



lemma nfilter_nfilter_suffix:
   assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
   shows   " (suffix k (nfilter P xs 0))   = 
             (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs)  (inth (nfilter P xs 0) k))"
using assms 
by (metis ifilter_nfilter_suffix_idx_a ifilter_nfilter_suffix_idx_b 
    ifilter_nfilter_suffix_iset_eq idx_iset_eq)

lemma nfilter_nfilter_suffix_1:
   assumes "\<exists>x \<in> iset xs . P x "
           "k \<le> ilen (ifilter P xs)" 
   shows   " (suffix k (nfilter P xs 0))   = 
             (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs)  (inth (nfilter P xs 0) k))"
proof -
 have 1: "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
   by (metis assms(1) assms(2) nfilter_holds nfilter_ilen nfilter_inth_n_zero inth_iset) 
 show ?thesis using assms "1" nfilter_nfilter_suffix by auto
qed

lemma nfilter_imap_ifilter:
 assumes "(\<exists> x \<in> iset xs. P x)"
 shows   " imap (\<lambda>n. (inth xs n)) (nfilter P xs 0) = ifilter P xs"
proof -
 have 1: " ilen (imap (\<lambda>n. (inth xs n)) (nfilter P xs 0)) = ilen (ifilter P xs)"
   by (simp add: assms nfilter_ilen)
 have 2: "\<And>i. i\<le> ilen (imap (\<lambda>n. (inth xs n)) (nfilter P xs 0)) \<longrightarrow>
           (inth (imap (\<lambda>n. (inth xs n)) (nfilter P xs 0)) i) =
           (inth (ifilter P xs) i)"
    by (metis assms diff_zero ilen_imap inth_imap nfilter_ifilter)
  from 1 2 show ?thesis 
  using interval_eq_inth_eq by blast
qed

lemma ifilter_nfilter_prefix:
   assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
   shows   " (prefix k (ifilter P xs ))   = 
             (ifilter P (prefix ((inth (nfilter P xs 0) k) ) xs) )"
proof -
 have 1: "\<exists>x \<in> iset xs . P x "
   by (metis assms(1) assms(2) diff_zero ifilter_nfilter_prefix_ilen_0)
 have 2: "(ifilter P xs ) = imap (\<lambda>s. inth xs s) (nfilter P xs 0)"
    by (simp add: "1" nfilter_imap_ifilter)
 have 3: "(prefix k (ifilter P xs )) =
          (prefix k (imap (\<lambda>s. inth xs s) (nfilter P xs 0)))"
  by (simp add: "2")
 have 4: "(prefix k (imap (\<lambda>s. inth xs s) (nfilter P xs 0))) =
          imap (\<lambda>s. inth xs s) (prefix k (nfilter P xs 0))"
    by (simp add: "1" assms(2) imap_prefix nfilter_ilen)
 have 5: "\<exists> x \<in> iset(prefix ((inth (nfilter P xs 0) k))    xs). P x"
   by (metis "1" add.left_neutral assms(1) assms(2) ilast_prefix inth_and_iset 
       nfilter_ilen nfilter_upper_bound order_refl)
 have 6: "(ifilter P (prefix ((inth (nfilter P xs 0) k) ) xs) ) =
          imap (\<lambda>s. (inth (prefix ((inth (nfilter P xs 0) k) ) xs) s)) 
           (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0)"
      by (simp add: "5" nfilter_imap_ifilter)
 have 7: "imap (\<lambda>s. inth xs s) (prefix k (nfilter P xs 0)) =
          imap (\<lambda>s. inth xs s) (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0)" 
   by (simp add: assms(1) assms(2) nfilter_nfilter_prefix)
 have 8: "imap (\<lambda>s. inth xs s) (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0) =
          imap (\<lambda>s. (inth (prefix ((inth (nfilter P xs 0) k) ) xs) s)) 
           (nfilter P (prefix ((inth (nfilter P xs 0) k) ) xs) 0) 
         " 
     using "1" "5" by simp
 show ?thesis 
 by (simp add: "3" "4" "6" "7" "8")     
qed

lemma ifilter_nfilter_prefix_1:
   assumes "\<exists>x \<in> iset xs . P x"
           "k \<le> ilen (ifilter P xs)" 
   shows   " (prefix k (ifilter P xs ))   = 
             (ifilter P (prefix ((inth (nfilter P xs 0) k) ) xs) )"
proof -
 have 1: "P (inth xs   ( (inth (nfilter P xs 0) k)) )" 
    by (metis assms(1) assms(2) nfilter_holds nfilter_ilen nfilter_inth_n_zero inth_iset)
 show ?thesis using assms "1" ifilter_nfilter_prefix by auto
qed

lemma ifilter_nfilter_prefix_ilen:
   assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
   shows   " ilen(prefix k (ifilter P xs))   = 
             ilen(ifilter P (prefix ((inth (nfilter P xs 0) k) ) xs))"
using assms 
by (simp add: ifilter_nfilter_prefix)

lemma ifilter_nfilter_prefix_ilen_1:
   assumes "\<exists>x \<in> iset xs . P x"
           "k \<le> ilen (ifilter P xs)" 
   shows   " ilen(prefix k (ifilter P xs))   = 
             ilen(ifilter P (prefix ((inth (nfilter P xs 0) k) ) xs))"
using assms 
by (simp add: ifilter_nfilter_prefix_1)

lemma ifilter_nfilter_prefix_inth:
   assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
           " j\<le>  ilen(prefix k (ifilter P xs))"
   shows   " inth (prefix k (ifilter P xs)) j  = 
             inth (ifilter P (prefix ((inth (nfilter P xs 0) k) ) xs)) j"
using assms 
by (simp add: ifilter_nfilter_prefix)

lemma ifilter_nfilter_prefix_inth_1:
   assumes "\<exists>x \<in> iset xs . P x"
           "k \<le> ilen (ifilter P xs)" 
           " j\<le>  ilen(prefix k (ifilter P xs)) "
   shows   " inth (prefix k (ifilter P xs)) j  = 
             inth (ifilter P (prefix ((inth (nfilter P xs 0) k) ) xs)) j"
using assms 
by (simp add: ifilter_nfilter_prefix_1)

lemma nfilter_imap_shift:
assumes "\<exists> x \<in> iset xs. P x"
shows "imap (\<lambda>s. inth xs (s+n)) (nfilter P xs 0) =
       imap (\<lambda>s. inth xs s) (nfilter P xs n)"
proof -
 have 1: " ilen (imap (\<lambda>s. inth xs (s+n)) (nfilter P xs 0)) =
           ilen (imap (\<lambda>s. inth xs s) (nfilter P xs n))"
   by (metis assms(1) ilen_imap nfilter_ilen_n_zero)
 have 2: "\<And>i. inth (imap (\<lambda>s. inth xs (s+n)) (nfilter P xs 0)) i =
              (inth xs ( (inth (nfilter P xs 0) i) +n))"
     by (simp add: inth_imap)
 have 3: "\<And>i. inth ( imap (\<lambda>s. inth xs s) (nfilter P xs n) ) i =
              (inth xs ( (inth (nfilter P xs n) i)))"
    by (simp add: inth_imap)
 have 4: "\<And>i. (inth (nfilter P xs 0) i) +n = (inth (nfilter P xs n) i)"
    by (metis assms(1) inth_imap nfilter_n_zero)
 show ?thesis 
 by (metis "1" "2" "3" "4" interval_eq_inth_eq)
qed

lemma nfilter_imap_shift_suffix:
 assumes " \<exists> x \<in> iset(suffix ((inth (nfilter P xs 0) k))    xs). P x"
 shows   "imap (\<lambda>s. inth xs (s+(inth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0) =
          imap (\<lambda>s. inth xs s) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) (inth (nfilter P xs 0) k))"
proof -
 have 1: "ilen (imap (\<lambda>s. inth xs (s+(inth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0)) =
          ilen (imap (\<lambda>s. inth xs s) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) (inth (nfilter P xs 0) k)))" 
    by (metis assms ilen_imap nfilter_ilen_n_zero)
 have 2: "\<And>i. inth (imap (\<lambda>s. inth xs (s+(inth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0)) i =
           inth xs 
           ((inth (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0) i) + (inth (nfilter P xs 0) k))
           " 
    using inth_imap by blast
 have 3: "\<And>i. 
          inth ( imap (\<lambda>s. inth xs s) 
                    (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs)
                               (inth (nfilter P xs 0) k))) i =
          inth xs (inth (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs)
                                 (inth (nfilter P xs 0) k)) i)
          " 
    using inth_imap by blast
 have 4: "\<And>i. (inth (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0) i) 
               + (inth (nfilter P xs 0) k) =
               (inth (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) (inth (nfilter P xs 0) k)) i) "
  by (metis assms inth_imap nfilter_n_zero)
 show ?thesis 
 by (metis "1" "2" "3" "4" interval_eq_inth_eq)
qed

lemma ifilter_nfilter_suffix:
   assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
   shows   " (suffix k (ifilter P xs ))   = 
             (ifilter P (suffix ((inth (nfilter P xs 0) k) ) xs) )"
proof -
 have 1: "\<exists>x \<in> iset xs . P x "
   by (metis assms(1) assms(2) diff_zero ifilter_nfilter_prefix_ilen_0)
 have 2: "(ifilter P xs ) = imap (\<lambda>s. inth xs s) (nfilter P xs 0)"
    by (simp add: "1" nfilter_imap_ifilter)
 have 3: "(suffix k (ifilter P xs )) =
          (suffix k (imap (\<lambda>s. inth xs s) (nfilter P xs 0)))"
  by (simp add: "2")
 have 4: "(suffix k (imap (\<lambda>s. inth xs s) (nfilter P xs 0))) =
          imap (\<lambda>s. inth xs s) (suffix k (nfilter P xs 0))"
  by (simp add: "1" assms(2) imap_suffix nfilter_ilen)
 have 5: "\<exists> x \<in> iset(suffix ((inth (nfilter P xs 0) k))    xs). P x"
   by (metis "1" add_cancel_right_left assms(1) assms(2) ifirst_suffix 
       ilen_gr_zero inth_and_iset nfilter_ilen nfilter_upper_bound)   
 have 6: "(ifilter P (suffix ((inth (nfilter P xs 0) k) ) xs) ) =
          imap (\<lambda>s. inth (suffix ((inth (nfilter P xs 0) k) ) xs) s) 
            (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0)
         "
     by (simp add: "5" nfilter_imap_ifilter)
 have 7: "imap (\<lambda>s. inth (suffix ((inth (nfilter P xs 0) k) ) xs) s) 
            (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0) =
          imap (\<lambda>s. inth xs (s+(inth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0)
         " 
 using "1" "5" by (simp add: add.commute)
 have 8: "imap (\<lambda>s. inth xs (s+(inth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0) =
          imap (\<lambda>s. inth xs s) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) (inth (nfilter P xs 0) k))
         " 
   using "5" nfilter_imap_shift_suffix by metis
 have 9: "imap (\<lambda>s. inth xs s) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) (inth (nfilter P xs 0) k)) =
          imap (\<lambda>s. inth xs s)
           (suffix k (nfilter P xs 0))
         " 
     by (simp add: assms(1) assms(2) nfilter_nfilter_suffix)
 show ?thesis 
 by (simp add: "3" "4" "6" "7" "8" "9")
qed


lemma ifilter_nfilter_suffix_1:
   assumes "\<exists> x \<in> iset xs. P x"
           "k \<le> ilen (ifilter P xs)" 
   shows   " (suffix k (ifilter P xs ))   = 
             (ifilter P (suffix ((inth (nfilter P xs 0) k) ) xs) )"
proof -
 have 1: "\<exists>x \<in> iset xs . P x "
   using assms by auto
 have 2: "(ifilter P xs ) = imap (\<lambda>s. inth xs s) (nfilter P xs 0)"
    by (simp add: "1" nfilter_imap_ifilter)
 have 3: "(suffix k (ifilter P xs )) =
          (suffix k (imap (\<lambda>s. inth xs s) (nfilter P xs 0)))"
  by (simp add: "2")
 have 4: "(suffix k (imap (\<lambda>s. inth xs s) (nfilter P xs 0))) =
          imap (\<lambda>s. inth xs s) (suffix k (nfilter P xs 0))"
  by (simp add: "1" assms(2) imap_suffix nfilter_ilen)
 have 5: "\<exists> x \<in> iset(suffix ((inth (nfilter P xs 0) k))    xs). P x"
  using assms inth_iset  "2"  inth_suffix nfilter_holds[of "xs" "P" "0"]
  by (metis add.right_neutral diff_zero ilen_imap le0)  
 have 6: "(ifilter P (suffix ((inth (nfilter P xs 0) k) ) xs) ) =
          imap (\<lambda>s. inth (suffix ((inth (nfilter P xs 0) k) ) xs) s) 
            (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0)
         "
     by (simp add: "5" nfilter_imap_ifilter)
 have 7: "imap (\<lambda>s. inth (suffix ((inth (nfilter P xs 0) k) ) xs) s) 
            (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0) =
          imap (\<lambda>s. inth xs (s+(inth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0)
         " 
   using "1" "5" 
   by (metis add.right_neutral inth_suffix suffix_suffix le0)
 have 8: "imap (\<lambda>s. inth xs (s+(inth (nfilter P xs 0) k))) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) 0) =
          imap (\<lambda>s. inth xs s) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) (inth (nfilter P xs 0) k))
         " 
   using "5" nfilter_imap_shift_suffix by metis
 have 9: "imap (\<lambda>s. inth xs s) 
           (nfilter P (suffix ((inth (nfilter P xs 0) k) ) xs) (inth (nfilter P xs 0) k)) =
          imap (\<lambda>s. inth xs s)
           (suffix k (nfilter P xs 0))
         " 
     by (simp add: assms(1) assms(2) nfilter_nfilter_suffix_1)
 show ?thesis 
 by (simp add: "3" "4" "6" "7" "8" "9")
qed

lemma ifilter_nfilter_suffix_ilen:
   assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
   shows   " ilen(suffix k (ifilter P xs))   = 
             ilen(ifilter P (suffix ((inth (nfilter P xs 0) k) ) xs))"
using assms 
by (simp add: ifilter_nfilter_suffix)

lemma ifilter_nfilter_suffix_ilen_1:
   assumes "\<exists> x \<in> iset xs. P x"
           "k \<le> ilen (ifilter P xs)" 
   shows   " ilen(suffix k (ifilter P xs))   = 
             ilen(ifilter P (suffix ((inth (nfilter P xs 0) k) ) xs))"
using assms 
by (simp add: ifilter_nfilter_suffix_1)

lemma ifilter_nfilter_suffix_inth:
   assumes "P (inth xs   ( (inth (nfilter P xs 0) k)) )"
           "k \<le> ilen (ifilter P xs)" 
           " j\<le>  ilen(suffix k (ifilter P xs)) " 
   shows   " inth (suffix k (ifilter P xs)) j  = 
             inth (ifilter P (suffix ((inth (nfilter P xs 0) k) ) xs)) j"
using assms 
by (simp add: ifilter_nfilter_suffix)

lemma ifilter_nfilter_suffix_inth_1:
   assumes "\<exists> x \<in> iset xs. P x"
           "k \<le> ilen (ifilter P xs)" 
           " j\<le>  ilen(suffix k (ifilter P xs)) " 
   shows   " inth (suffix k (ifilter P xs)) j  = 
             inth (ifilter P (suffix ((inth (nfilter P xs 0) k) ) xs)) j"
using assms 
by (simp add: ifilter_nfilter_suffix_1)

lemma nfilter_ilast:
 assumes " P (ilast (prefix k xs))"
         "k \<le> ilen xs" 
 shows   " (inth xs ((ilast (nfilter P (prefix k xs) n))-n)) = (ilast (prefix k xs))"
using assms
proof (induction xs arbitrary: n k)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
   proof (auto split:nat.split)
      fix x2
      fix x
      assume a0: "(\<And>k n. P (inth xs k) \<Longrightarrow>
               k \<le> ilen xs \<Longrightarrow>
               inth xs (inth (nfilter P (prefix k xs) n) (ilen (nfilter P (prefix k xs) n)) - n) =
               inth xs k)" 
      assume a1: "P (inth xs x2)" 
      assume a2: "x \<in> iset (prefix x2 xs)"
      assume a3: "P x"  
      assume a4: "k = Suc x2" 
      assume a5: "x2 \<le> ilen xs" 
      assume a6: "inth (nfilter P (prefix x2 xs) (Suc n)) (ilen (nfilter P (prefix x2 xs) (Suc n))) 
                  \<le> n" 
    show "x1a = inth xs x2" 
     using a0 a1 a2 a3 a6 by (metis le_refl nfilter_inth_n_zero_a not_less_eq_eq)
   next 
     fix x2
     fix x
     fix x2a
     assume b0: "(\<And>k n. P (inth xs k) \<Longrightarrow>
               k \<le> ilen xs \<Longrightarrow>
               inth xs (inth (nfilter P (prefix k xs) n) (ilen (nfilter P (prefix k xs) n)) - n) =
               inth xs k)" 
     assume b1: "P (inth xs x2)" 
     assume b2: "x \<in> iset (prefix x2 xs)"
     assume b3: "P x"
     assume b4: "k = Suc x2"
     assume b5: "x2 \<le> ilen xs"
     assume b6: "inth (nfilter P (prefix x2 xs) (Suc n))
                     (ilen (nfilter P (prefix x2 xs) (Suc n))) - n = Suc x2a"     
   show " inth xs x2a = inth xs x2" 
   using b0 b1 b2 b3 b4 b5 b6 
     by (metis Suc_eq_plus1 add_diff_cancel_left' diff_diff_left plus_1_eq_Suc)  
   next
    fix x2
    assume c0: "(\<And>k n. P (inth xs k) \<Longrightarrow>
             k \<le> ilen xs \<Longrightarrow>
             inth xs (inth (nfilter P (prefix k xs) n) (ilen (nfilter P (prefix k xs) n)) - n) =
                  inth xs k)" 
    assume c1: "P (inth xs x2)"
    assume c2: "\<forall>x\<in>iset (prefix x2 xs). \<not> P x" 
    assume c3: "k = Suc x2"  
    assume c4: "x2 \<le> ilen xs" 
   show "x1a = inth xs x2" 
   using c0 c1 c2 c3 c4 ilast_prefix inth_iset by fastforce
 qed
qed

lemma nfilter_ifirst:
 assumes " P (ifirst (suffix k xs)) "
         "k \<le> ilen xs" 
 shows   "ifirst (nfilter P (suffix k xs) n)= n "
using assms
proof (induction xs arbitrary: k n)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
     by (auto  split:nat.split)
qed

lemma ifilter_ilast:
 assumes " P (ilast (prefix n xs))"
         "n \<le> ilen xs"
 shows   " ilast (ifilter P (prefix n xs)) = (ilast (prefix n xs))"
using assms   
proof (induction n arbitrary: xs)
case 0
then show ?case by simp
next
case (Suc n)
then show ?case 
   proof (cases xs)
   case (INil x1)
   then show ?thesis 
   by simp
   next
   case (ICons x21 x22)
   then show ?thesis using Suc inth_iset by (auto, force)
   qed
qed

lemma ifilter_ifirst:
 assumes " P (ifirst (suffix n xs))"
 shows   " ifirst (ifilter P (suffix n xs)) = (ifirst (suffix n xs))"
using assms 
proof (induction xs arbitrary: n)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case by (auto split:nat.split)
qed

lemma ifilter_inth_aa:
 assumes "(\<exists> x \<in> iset xs. P x)"
         "n \<le> ilen (ifilter P xs)"
 shows   "P (inth (ifilter P xs) n)"
using assms iset_ifilter inth_iset[of "n" "(ifilter P xs)"]  by simp



lemma ifilter_ilen_zero_conv_a:
 assumes "(\<exists>x\<in>iset xs. P x)"
         "ilen (ifilter P xs) =0"
 shows   " (\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j \<noteq> k \<longrightarrow> \<not> P (inth xs j))) "
proof -
 have 1: "P (inth xs  (inth (nfilter P xs 0) 0)) "
   by (metis nfilter_imap_ifilter assms(1) assms(2) ifilter_inth_aa inth_imap 
       le_numeral_extra(3)) 
 have 2: "(inth (nfilter P xs 0) 0) \<le> ilen xs" 
   by (metis assms(1) diff_zero ilen_gr_zero nfilter_upper_bound 
       ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 3: "(\<forall> j \<le> ilen xs. j \<noteq> (inth (nfilter P xs 0) 0) \<longrightarrow> \<not> P (inth xs j)) " 
    using assms nfilter_not_after[of "xs" "P" ] nfilter_not_before[of "xs" "P"] 
    by (metis linorder_neqE_nat nfilter_ilen)
 show ?thesis
 using "1" "2" "3" by blast
qed

lemma ifilter_ilen_zero_conv_c:
 "(\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j \<noteq> k \<longrightarrow> \<not> P (inth xs j))) =
   (\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j<k \<or> k<j \<longrightarrow> \<not> P (inth xs j)))
 " 
using antisym_conv3 by  auto

lemma ifilter_ilen_zero_conv_d:
 " (\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j<k \<or> k<j \<longrightarrow> \<not> P (inth xs j))) =
   (\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
     (\<forall> j. j<k  \<longrightarrow> \<not> P (inth xs j)) \<and>
     (\<forall> j \<le> ilen xs. k<j  \<longrightarrow> \<not> P (inth xs j))
   )
 " 
by auto

lemma ifilter_ilen_zero_conv_b:
 assumes "(\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j \<noteq> k \<longrightarrow> \<not> P (inth xs j)))"
 shows   "(\<exists>x\<in>iset xs. P x) \<and>  ilen (ifilter P xs) =0 "
proof -
 have 1: "(\<exists>x\<in>iset xs. P x)" 
   using assms inth_iset by auto
 obtain k where 2: "k \<le> ilen xs \<and>  P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j \<noteq> k \<longrightarrow> \<not> P (inth xs j))" 
   using assms by auto
 have 3: "ilen (ifilter P xs) =0"
  using 1 2 
   proof (induct xs arbitrary: k)
   case (INil x)
   then show ?case by simp
   next
   case (ICons x1a xs)
   then show ?case 
        proof (cases k)
        case 0
        then show ?thesis using ICons 
          by (metis Suc_le_mono ifilter_ilen_d inth_Suc inth_and_iset ilen.simps(2) 
              le0  not_less_eq_eq order_refl plus_1_eq_Suc)
        next
        case (Suc nat)
        then show ?thesis using ICons.hyps ICons.prems 
        by (metis Suc_le_mono Zero_not_Suc add_diff_cancel_left' ifilter_ilen_c 
            ilen_gr_zero inth_Suc inth_zero iset_IConsD ilen.simps(2) plus_1_eq_Suc)
       qed  
   qed
 show ?thesis 
 using "1" "3" by blast
qed

lemma ifilter_ilen_zero_conv:
 " ((\<exists>x\<in>iset xs. P x) \<and>  ilen (ifilter P xs) =0) =
   (\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j \<noteq> k \<longrightarrow> \<not> P (inth xs j)))"
using ifilter_ilen_zero_conv_a[of "xs" "P"]  ifilter_ilen_zero_conv_b[of "xs" "P"]
by blast

lemma ifilter_ilen_zero_conv_1:
 " ((\<exists>x\<in>iset xs. P x) \<and>  ilen (ifilter P xs) =0) =
   (\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j<k \<or> k<j \<longrightarrow> \<not> P (inth xs j)))"
proof -
 have 1: "((\<exists>x\<in>iset xs. P x) \<and>  ilen (ifilter P xs) =0) = 
          (\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j \<noteq> k \<longrightarrow> \<not> P (inth xs j)))" 
    by (simp add: ifilter_ilen_zero_conv)
 have 2: "(\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j \<noteq> k \<longrightarrow> \<not> P (inth xs j))) =
          (\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j<k \<or> k<j \<longrightarrow> \<not> P (inth xs j)))" 
        by fastforce
 show ?thesis by (simp add: "1" "2")
qed

lemma ifilter_ilen_zero_conv_2:
 " ((\<exists>x\<in>iset xs. P x) \<and>  ilen (ifilter P xs) =0) =
   (\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
     (\<forall> j. j<k  \<longrightarrow> \<not> P (inth xs j)) \<and>
     (\<forall> j \<le> ilen xs. k<j  \<longrightarrow> \<not> P (inth xs j))
   )"
proof -
 have 1: "((\<exists>x\<in>iset xs. P x) \<and>  ilen (ifilter P xs) =0) =
   (\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j<k \<or> k<j \<longrightarrow> \<not> P (inth xs j)))"
   by (simp add: ifilter_ilen_zero_conv_1) 
 have 2: "(\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
             (\<forall> j \<le> ilen xs. j<k \<or> k<j \<longrightarrow> \<not> P (inth xs j))) =
   (\<exists> k \<le> ilen xs. P (inth xs k) \<and> 
     (\<forall> j. j<k  \<longrightarrow> \<not> P (inth xs j)) \<and>
     (\<forall> j \<le> ilen xs. k<j  \<longrightarrow> \<not> P (inth xs j))
   ) "
   using dual_order.strict_trans1 by auto 
 from 1 2 show ?thesis   by auto
qed 

lemma ifilter_suffixes_imap_help_0:
assumes "j \<le> inth(nfilter P xs 0) 0 " 
        "\<exists> x \<in> iset xs. P x" 
shows   "(ifilter P (suffix (inth(nfilter P xs 0) 0) xs)) = (ifilter P  (suffix j xs))"
using assms 
proof (induct xs arbitrary:j)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
   proof (cases "\<exists>a\<in>iset xs. P a")
   case True
   then show ?thesis 
     proof (cases "P x1a")
     case True
     then show ?thesis using ICons.prems
     by (metis le_zero_eq nfilter_inth_ICons)
     next
     case False
     then show ?thesis 
       proof (cases j)
       case 0
       then show ?thesis using ICons False \<open>\<exists>a\<in>iset xs. P a\<close> 
       by (metis ifilter_nfilter_suffix_1 suffix_zero le0)
       next
       case (Suc nat)
       then show ?thesis using ICons False \<open>\<exists>a\<in>iset xs. P a\<close>  by (auto split: nat.split)
              (metis One_nat_def add_diff_cancel_left' le0 nfilter_inth_n_zero plus_1_eq_Suc)
     qed
    qed
   next
   case False
   then show ?thesis 
   using ICons by auto
   qed
qed
    
lemma ifilter_suffixes_imap_help_0_a:
assumes "j \<le> inth(nfilter P (suffixes xs) 0) 0 " 
        "\<exists> x \<in> iset (suffixes xs). P x" 
shows   "(ifilter P (suffixes (suffix (inth(nfilter P (suffixes xs) 0) 0) xs))) = 
         (ifilter P (suffixes (suffix j xs)))"
proof -
 have 1: "(suffix (inth(nfilter P (suffixes xs) 0) 0) (suffixes xs)) =
          (suffixes (suffix (inth(nfilter P (suffixes xs) 0) 0) xs))" 
  by (metis assms(2) diff_zero ilen_gr_zero nfilter_upper_bound 
      ordered_cancel_comm_monoid_diff_class.add_diff_inverse suffix_suffixes)
 have 2: "inth(nfilter P (suffixes xs) 0) 0 \<le> ilen (suffixes xs)"
    by (metis add.left_neutral assms(2) ilen_gr_zero nfilter_upper_bound) 
 have 3: "(suffix j (suffixes xs)) = (suffixes (suffix j xs))" 
    using 2 suffix_suffixes assms le_trans by blast
 show ?thesis 
 using "1" "3" assms ifilter_suffixes_imap_help_0 by fastforce
qed

lemma ifilter_suffixes_imap_help_1:
assumes "j \<le> inth(nfilter P xs 0) 1 " 
        " 0 < ilen(ifilter P xs)" 
        "  inth(nfilter P xs 0) 0 < j" 
        "\<exists> x \<in> iset xs. P x" 
shows   "(ifilter P (suffix (inth(nfilter P xs 0) 1) xs)) = (ifilter P  (suffix j xs))"
using assms 
proof (induct xs arbitrary:j)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
  proof (cases "\<exists>a\<in>iset xs. P a")
  case True
  then show ?thesis 
    proof (cases "P x1a")
    case True
    then show ?thesis
       proof (cases j)
       case 0
       then show ?thesis 
       using ICons.prems by blast
       next
       case (Suc nat)
       then show ?thesis using ICons True \<open>\<exists>a\<in>iset xs. P a\<close> ifilter_suffixes_imap_help_0[of _ "P" "xs"] 
           nfilter_inth_n_zero[of "xs" "P"]  
          proof auto
          fix x :: 'a
          assume "j = Suc nat"
          assume a1: "Suc nat \<le> inth (nfilter P xs (Suc 0)) 0"
          assume a2: "\<And>k n. k \<le> ilen (nfilter P xs n) \<Longrightarrow> 
                      inth (nfilter P xs n) k - n = inth (nfilter P xs 0) k"
          assume a3: "\<And>j. j \<le> inth (nfilter P xs 0) 0 \<Longrightarrow> 
                       ifilter P (suffix (inth (nfilter P xs 0) 0) xs) = ifilter P (suffix j xs)"
          have "Suc (inth (nfilter P xs 0) 0) = inth (nfilter P xs 1) 0"
          using a2 a1 by (metis (no_types) One_nat_def Suc_le_D diff_Suc_Suc diff_zero le0) (* 22 ms *)
          then show "ifilter P (case inth (nfilter P xs (Suc 0)) 0 of 0 \<Rightarrow> x1a \<odot> xs | 
                               Suc n \<Rightarrow> suffix n xs) = ifilter P (suffix nat xs)"
          using a3 a2 a1 by (simp add: Nitpick.case_nat_unfold) (* 9 ms *)
          qed
      qed
    next
    case False
    then show ?thesis using False ICons \<open>\<exists>a\<in>iset xs. P a\<close>
       proof (auto split: nat.split)
        fix x2::nat
        fix x
        fix x2a
        assume a0: "\<not> P x1a"  
        assume a1: "(\<And>j. j \<le> inth (nfilter P xs 0) (Suc 0) \<Longrightarrow>
             inth (nfilter P xs 0) 0 < j \<Longrightarrow>
             ifilter P (suffix (inth (nfilter P xs 0) (Suc 0)) xs) = ifilter P (suffix j xs))"
        assume a2: "x2a \<le> x2"
        assume a3: "0 < ilen (ifilter P xs)"
        assume a4: "inth (nfilter P xs (Suc 0)) 0 < Suc x2a"
        assume a5: "x \<in> iset xs" 
        assume a6: "P x"   
        assume a7: "inth (nfilter P xs (Suc 0)) (Suc 0) = Suc x2"
        assume a8: "j = Suc x2a"  
        show "ifilter P (suffix x2 xs) = ifilter P (suffix x2a xs)"
        proof -
         have 1: "j>0"
          using a8 by auto 
         have 2: "x2a = j-1" 
          by (simp add: a8)
         have 3: "x2 = inth (nfilter P xs (Suc 0)) (Suc 0) - (Suc 0)"
          by (simp add: a7) 
         have 4: "(inth (nfilter P xs (Suc 0)) (Suc 0) - (Suc 0)) = (inth (nfilter P xs 0) (Suc 0))" 
          by (metis Suc_leI a3 a5 a6 nfilter_ilen nfilter_inth_n_zero)          
         have 5: "0 < ilen (ifilter P xs)"
          by (simp add: a3) 
         have 6: "ifilter P (suffix x2 xs) = ifilter P (suffix (inth (nfilter P xs 0) (Suc 0)) xs)"
          by (simp add: "3" "4") 
         have 7: "ifilter P (suffix x2a xs) = ifilter P (suffix (j-1) xs)"
          using 2 by blast
         have 8: "j-1 \<le> inth (nfilter P xs 0) 1"
           by (metis "2" "3" "4" One_nat_def a2)  
         have 9: "(inth (nfilter P xs (Suc 0)) 0)- (Suc 0) = (inth (nfilter P xs 0) 0)"
          using a5 a6 nfilter_inth_n_zero by fastforce 
         have 10: "(inth (nfilter P xs 0) 0)  < j - (Suc 0)"
           using nfilter_inth_ICons[of "P" "x1a" "xs" _ _] 
                 ifilter_inth_aa[of "xs" "P" "j-1"]
                 nfilter_inth_n_zero[of "xs" "P"] nfilter_holds[of "x1a\<odot>xs" "P"] 
           using 9 2 True a0 a3 a4 a5 a6 a8 
           by (metis "1" One_nat_def Suc_n_not_le_n diff_Suc_less less_Suc_eq less_imp_diff_less 
               less_imp_le_nat nfilter_ilen nfilter_lower_bound)
         have 11: "ifilter P (suffix (inth (nfilter P xs 0) 1) xs) = ifilter P (suffix (j-1) xs)"
          by (metis "10" "8" One_nat_def a1)    
         show ?thesis by (metis "11" "3" "4" "7" One_nat_def)
       qed
     qed
   qed
  next
  case False
  then show ?thesis 
  using ICons.prems(2) by auto
  qed
qed

lemma ifilter_suffixes_imap_help_j:
assumes "j \<le> inth(nfilter P xs 0) (Suc i) " 
        " i < ilen(ifilter P xs)" 
        "  inth(nfilter P xs 0) i < j" 
        "\<exists> x \<in> iset xs. P x" 
shows   "(ifilter P (suffix (inth(nfilter P xs 0) (Suc i)) xs)) = (ifilter P  (suffix j xs))"
using assms 
proof (induct xs arbitrary:j i)
case (INil x)
then show ?case by simp
next
case (ICons x1a xs)
then show ?case 
   proof (cases "(\<exists>a\<in>iset xs. P a)")
   case True
   then show ?thesis 
     proof (cases "P x1a")
     case True
     then show ?thesis using ICons True \<open>(\<exists>a\<in>iset xs. P a)\<close> 
       proof (auto split: nat.split)
        fix x
        fix x2::nat
        fix x2a
        assume a0: "(\<And>j i. j \<le> inth (nfilter P xs 0) (Suc i) \<Longrightarrow>
               i < ilen (ifilter P xs) \<Longrightarrow>
               inth (nfilter P xs 0) i < j \<Longrightarrow>
               ifilter P (suffix (inth (nfilter P xs 0) (Suc i)) xs) = ifilter P (suffix j xs))"
        assume a1: " x2a \<le> x2 "
        assume a2: "i < Suc (ilen (ifilter P xs))"
        assume a3: "(case i of 0 \<Rightarrow> 0 | Suc x \<Rightarrow> inth (nfilter P xs (Suc 0)) x) < Suc x2a"
        assume a4: "P x1a"
        assume a5: "x \<in> iset xs"   
        assume a6: "P x"  
        assume a7: "inth (nfilter P xs (Suc 0)) i = Suc x2 "
        assume a8: " j = Suc x2a"   
        show "ifilter P (suffix x2 xs) = ifilter P (suffix x2a xs) " 
        proof -
         have 1: "i=0 \<longrightarrow> ifilter P (suffix x2 xs) = ifilter P (suffix x2a xs)"
          using a0 a1 a2 a3 a4 a5 a6 a7 a8 
          by (metis One_nat_def add_diff_cancel_left' ilen_gr_zero nfilter_inth_n_zero 
              plus_1_eq_Suc ifilter_suffixes_imap_help_0)
         have 2: "i>0\<longrightarrow> (inth (nfilter P xs (Suc 0)) (i-1)) - (Suc 0) = inth (nfilter P xs 0) (i-1)"
          using a0 a1 a2 a3 a4 a5 a6 a7 a8  
          by (metis One_nat_def Suc_leI add_leD1 le_add_diff_inverse2 
              less_Suc_eq_le nfilter_ilen nfilter_inth_n_zero)
         have 3: "i>0 \<longrightarrow>  inth (nfilter P xs (Suc 0)) (i-1) <Suc  x2a"
          using a0 a1 a2 a3 a4 a5 a6 a7 a8  
          by (metis Nitpick.case_nat_unfold less_Suc0 not_less_eq)
         have 4: "i>0 \<longrightarrow> inth (nfilter P xs (Suc 0)) (i-1) >0"
          using a0 a1 a2 a3 a4 a5 a6 a7 a8  nfilter_lower_bound[of "xs" "P" "i-1" "Suc 0"] 
          by (metis One_nat_def Suc_le_lessD Suc_le_mono Suc_pred less_imp_le_nat nfilter_ilen)   
         have 5: "i>0 \<longrightarrow> inth (nfilter P xs 0) (i-1) < x2a"
          using 2 3 4  by linarith
         have 6: "i>0 \<longrightarrow> x2a \<le> inth (nfilter P xs 0) i"
          using a0 a1 a2 a3 a4 a5 a6 a7 a8  
          by (metis One_nat_def add_diff_cancel_left'  
              less_Suc_eq_le nfilter_ilen nfilter_inth_n_zero plus_1_eq_Suc)
         have 7: "i>0 \<longrightarrow> i - 1 < ilen (ifilter P xs)"
          using a2  by linarith
         have 8: "i>0 \<longrightarrow> ifilter P (suffix (inth (nfilter P xs 0) i) xs ) = ifilter P (suffix x2a xs)"
          using a0 5 6 7 by (metis Suc_pred')  
         have 9: "i>0 \<longrightarrow>  (inth (nfilter P xs 0) i) = x2"
          using 7 a0 a1 a2 a3 a4 a5 a6 a7 a8 
           by (metis  One_nat_def Suc_leI Suc_pred add_diff_cancel_left' 
               nfilter_ilen nfilter_inth_n_zero plus_1_eq_Suc)
         from 1 2 8 9 show ?thesis by auto
        qed
       qed
     next
     case False
     then show ?thesis using ICons False \<open>(\<exists>a\<in>iset xs. P a)\<close> 
      proof (auto split: nat.split) 
       fix x
       fix x2::nat
       fix x2a
       assume b0: "(\<And>j i. j \<le> inth (nfilter P xs 0) (Suc i) \<Longrightarrow>
               i < ilen (ifilter P xs) \<Longrightarrow>
               inth (nfilter P xs 0) i < j \<Longrightarrow>
               ifilter P (suffix (inth (nfilter P xs 0) (Suc i)) xs) = ifilter P (suffix j xs))"
       assume b1: "x2a \<le> x2"
       assume b2: "i < ilen (ifilter P xs)"
       assume b3: "inth (nfilter P xs (Suc 0)) i < Suc x2a"
       assume b4: "\<not> P x1a "
       assume b5: "x \<in> iset xs"
       assume b6: "P x"
       assume b7: "inth (nfilter P xs (Suc 0)) (Suc i) = Suc x2"
       assume b8: "j = Suc x2a"
       show "ifilter P (suffix x2 xs) = ifilter P (suffix x2a xs)"
       proof -
        have 1: "i=0 \<longrightarrow> (inth (nfilter P xs (Suc 0)) (Suc i)) - (Suc 0) = 
                          inth (nfilter P xs 0) (Suc i)" 
         using b0 b1 b2 b3 b4 b5 b6 b7 b8     
          by (metis Suc_leI  nfilter_ilen nfilter_inth_n_zero) 
        have 2: "i=0 \<longrightarrow> (inth (nfilter P xs 0) (Suc i)) = x2"
         using 1 b7  by linarith
        have 3: "i=0 \<longrightarrow>  inth (nfilter P xs 0) i < x2a"
         using b0 b1 b2 b3 b4 b5 b6 b7 b8  nfilter_lower_bound[of "xs" "P" "i" "Suc 0"]  
          nfilter_inth_n_zero[of "xs" "P" "i" "Suc 0"]  
          by (metis One_nat_def Suc_le_lessD le0 less_Suc_eq_le 
              ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc)
        have 4: "i=0 \<longrightarrow> i < ilen (ifilter P xs)"
         using b2 by blast
        have 5: "i=0 \<longrightarrow> x2a \<le> inth (nfilter P xs 0) (Suc i)"
         using 1 2 b1 by blast
        have 6: "i=0 \<longrightarrow> ifilter P (suffix x2 xs) = ifilter P (suffix x2a xs)"
         using 2 3 5 b0 b2 by blast  
        have 7: "i>0\<longrightarrow> (inth (nfilter P xs (Suc 0)) (i)) - (Suc 0) = inth (nfilter P xs 0) (i) "
         using b0 b1 b2 b3 b4 b5 b6 b7 b8  
         by (metis less_imp_le_nat nfilter_ilen nfilter_inth_n_zero) 
        have 8: "i>0 \<longrightarrow>  inth (nfilter P xs (Suc 0)) i <Suc  x2a"
         using b3 by auto
        have 9: "i>0 \<longrightarrow> inth (nfilter P xs (Suc 0)) (i) >0"
         using b0 b1 b2 b3 b4 b5 b6 b7 b8  
         by (metis gr_zeroI idx_nfilter_gr less_imp_le_nat nfilter_ilen not_less0)     
        have 10: "i>0 \<longrightarrow> inth (nfilter P xs 0) (i) < x2a"
         using 7 9 b3  by linarith
        have 11: "i>0 \<longrightarrow> x2a \<le> inth (nfilter P xs 0) (Suc i)"
         using b0 b1 b2 b3 b4 b5 b6 b7 b8 
         by (metis One_nat_def Suc_leI add_diff_cancel_left'  nfilter_ilen 
              nfilter_inth_n_zero plus_1_eq_Suc) 
        have 12: "i>0 \<longrightarrow> i  < ilen (ifilter P xs)"
          using b2 by simp
        have 13: "i>0 \<longrightarrow>  ifilter P (suffix (inth (nfilter P xs 0) (Suc i)) xs ) = 
                  ifilter P (suffix x2a xs)" 
         using 10 11 b0 b1 b2 b3 b4 b5 b6 b7 b8 by blast
        have 14: "i>0 \<longrightarrow> (inth (nfilter P xs 0) (Suc i)) = x2"
         using b0 b1 b2 b3 b4 b5 b6 b7 b8 
         by (metis One_nat_def Suc_leI add_diff_cancel_left' nfilter_ilen 
             nfilter_inth_n_zero plus_1_eq_Suc) 
        show ?thesis using "13" "14" "6" by blast 
       qed         
      qed
    qed
   next
   case False
   then show ?thesis using ICons False by auto
   qed
qed

lemma ifilter_suffixes_imap_help_j_aa:
assumes "j \<le> inth(nfilter P (suffixes xs) 0) (Suc i) " 
        " i < ilen(ifilter P (suffixes xs))" 
        "  inth(nfilter P (suffixes xs) 0) i < j" 
        "\<exists> x \<in> iset (suffixes xs). P x" 
shows   "(ifilter P (suffixes (suffix (inth(nfilter P (suffixes xs) 0) (Suc i)) xs))) = 
         (ifilter P (suffixes (suffix j xs)))"
proof -
 have 1: "inth(nfilter P (suffixes xs) 0) (Suc i) \<le> ilen (suffixes xs)"
   by (metis Suc_leI add_cancel_right_left assms(2) assms(4) nfilter_ilen nfilter_upper_bound)
 have 2: "(suffixes (suffix (inth(nfilter P (suffixes xs) 0) (Suc i)) xs)) =
          (suffix (inth(nfilter P (suffixes xs) 0) (Suc i)) (suffixes xs))"
   using "1" suffix_suffixes by fastforce
 have 3: "(suffixes (suffix j xs)) = (suffix j (suffixes xs))"
    using "1" assms le_trans suffix_suffixes[of "j" "xs"]  by fastforce 
 show ?thesis 
 by (simp add: "2" "3" assms(1) assms(2) assms(3) assms(4) ifilter_suffixes_imap_help_j)
qed

lemma ifilter_suffixes_imap:
assumes "(Suc i) \<le> ilen (ifilter f (suffixes \<sigma>)) " 
        "\<exists> x \<in> iset(suffixes \<sigma>). f x" 
shows   "(suffix (Suc i) (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))) =
           (imap (\<lambda>s. inth s 0)
                (ifilter f (suffixes (suffix (Suc (inth (nfilter f (suffixes \<sigma>) 0) i)) \<sigma>))))"
proof -
 have 1: "(suffix (Suc i) (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))) =
          (imap (\<lambda>s. inth s 0) (suffix (Suc i) (ifilter f (suffixes \<sigma>))))"
  by (simp add: assms(1) imap_suffix) 
 have 2: "(Suc (inth (nfilter f (suffixes \<sigma>) 0) i)) \<le> ilen(suffixes \<sigma>)"
      by (metis ilen_ifilter_le Suc_leD assms(1) assms(2) diff_is_0_eq diff_zero 
          ifilter_nfilter_suffix_1 suffix_ilen not_less_eq_eq)
 have 3: "(ifilter f (suffixes (suffix (Suc (inth (nfilter f (suffixes \<sigma>) 0) i)) \<sigma>))) =
          (ifilter f (suffix (Suc (inth (nfilter f (suffixes \<sigma>) 0) i)) (suffixes \<sigma>))) 
         " 
     using "2" suffix_suffixes by fastforce
 have 4: "(suffix (Suc i) (ifilter f (suffixes \<sigma>))) =
          (ifilter f (suffix (inth (nfilter f (suffixes \<sigma>) 0) (Suc i))  (suffixes \<sigma>))) " 
    by (simp add: assms(1) assms(2) ifilter_nfilter_suffix_1)
 have 5: "(Suc (inth (nfilter f (suffixes \<sigma>) 0) i)) \<le> inth(nfilter f (suffixes \<sigma>) 0) (Suc i)"
    by (simp add: Suc_leI Suc_le_lessD assms(1) assms(2) idx_nfilter_mono nfilter_ilen)
 have 6: " i < ilen(ifilter f (suffixes \<sigma>))"
    using Suc_le_eq assms(1) by blast 
 have 7: "inth (nfilter f (suffixes \<sigma>) 0) i  < (Suc (inth (nfilter f (suffixes \<sigma>) 0) i))"
    by simp
 have 8: "(ifilter f (suffix (inth (nfilter f (suffixes \<sigma>) 0) (Suc i))  (suffixes \<sigma>))) = 
          (ifilter f (suffix (Suc (inth (nfilter f (suffixes \<sigma>) 0) i))  (suffixes \<sigma>)))" 
  using 5 6 7 
  ifilter_suffixes_imap_help_j[of "(Suc (inth (nfilter f (suffixes \<sigma>) 0) i))" "f" "suffixes \<sigma>" "i" ]
     assms(2) by blast 
 show ?thesis 
 by (simp add: "1" "3" "4" "8")
qed          
  
lemma sfxfilter_suffix_ilen:
 assumes "k\<le>ilen \<sigma>"
         "\<exists>x \<in> iset (suffixes(suffix k \<sigma>)). f x"
 shows   "ilen (ifilter f (suffixes (suffix k \<sigma>))) \<le> ilen (ifilter f (suffixes \<sigma>))" 
using assms
proof (induct k arbitrary: \<sigma>)
case 0
then show ?case by  simp
next
case (Suc k)
then show ?case 
  proof (cases \<sigma>)
  case (INil x1)
  then show ?thesis 
  by simp
  next
  case (ICons x21 x22)
  then show ?thesis 
     proof auto
      fix y
      assume a0: " \<sigma> = x21 \<odot> x22"
      assume a1: "f (x21 \<odot> x22)" 
      assume a2: " osfx y x22"
      assume a3: "f y" 
      show "ilen (ifilter f (suffixes (suffix k x22))) \<le> Suc (ilen (ifilter f (suffixes x22)))"
        using Suc a0  
        by (metis ilen_ifilter_le Suc.hyps Suc.prems(2) suffix_ilen_code 
            suffix_suc ilen_suffixes le_0_eq le_SucI less_imp_le_nat zero_less_Suc)
     next
      assume b0: "\<sigma> = x21 \<odot> x22" 
      assume b1: "f (x21 \<odot> x22)" 
      assume b2: "\<forall>x\<in>iset (suffixes x22). \<not> f x" 
      show "ilen (ifilter f (suffixes (suffix k x22))) = 0"
       using b0 b1 b2 Suc.prems osfx_suffix
       by (simp add: osfx_order.order.trans) 
     next
      fix y
      assume c0: "\<sigma> = x21 \<odot> x22" 
      assume c1: "\<not> f (x21 \<odot> x22)" 
      assume c2: "osfx y x22" 
      assume c3: "f y" 
      show "ilen (ifilter f (suffixes (suffix k x22))) \<le> ilen (ifilter f (suffixes x22))" 
      using c0 c1 
        by (metis ilen_ifilter_le Suc.hyps Suc.prems(2) ilen_gr_zero 
            suffix_ilen_code suffix_suc ilen_suffixes le_0_eq)
     next
      assume d0: "\<sigma> = x21 \<odot> x22" 
      assume d1: "\<not> f (x21 \<odot> x22)"
      assume d2: "\<forall>x\<in>iset (suffixes x22). \<not> f x" 
      show  "ilen (ifilter f (suffixes (suffix k x22))) = 0 "
      using d0 d1 d2 Suc.prems osfx_suffix
      by (simp add: osfx_order.order.trans) 
     qed
  qed
qed

lemma sfxfilter_suffix_inth:
 assumes "k\<le>ilen \<sigma>"
         "\<exists>x \<in> iset (suffixes(suffix k \<sigma>)). f x"
         "j\<le> ilen (ifilter f (suffixes (suffix k \<sigma>)))"
 shows   "(inth (ifilter f (suffixes (suffix k \<sigma>))) j) =
             (inth (suffix (ilen(ifilter f (suffixes \<sigma>))-ilen((ifilter f (suffixes (suffix k \<sigma>)))))
                   (ifilter f (suffixes \<sigma>))) j)" 
using assms 
proof (induct k arbitrary: \<sigma>)
case 0
then show ?case by simp
next
case (Suc k)
then show ?case 
   proof (cases \<sigma>)
   case (INil x1)
   then show ?thesis 
   by simp
   next
   case (ICons x21 x22)
   then show ?thesis  
       proof auto
        fix y
        assume a0: "\<sigma> = x21 \<odot> x22" 
        assume a1: "f (x21 \<odot> x22)" 
        assume a2: "osfx y x22" 
        assume a3: "f y" 
        show " 
         inth (ifilter f (suffixes (suffix k x22))) j =
         inth
       (case Suc (ilen (ifilter f (suffixes x22))) - ilen (ifilter f (suffixes (suffix k x22))) of
          0 \<Rightarrow> (x21 \<odot> x22) \<odot> ifilter f (suffixes x22) | Suc m \<Rightarrow> suffix m (ifilter f (suffixes x22)))
          j" 
        using a0 Suc 
        by (simp add: Suc_diff_le sfxfilter_suffix_ilen)
       next
        assume b0: "\<sigma> = x21 \<odot> x22" 
        assume b1: "f (x21 \<odot> x22)" 
        assume b2: "\<forall>x\<in>iset (suffixes x22). \<not> f x " 
        show "inth (ifilter f (suffixes (suffix k x22))) j = x21 \<odot> x22" 
        using Suc b0 b1 b2 using osfx_suffix
        using osfx_order.order_trans by fastforce
       next
        fix y
        assume c0: "\<sigma> = x21 \<odot> x22" 
        assume c1: "\<not> f (x21 \<odot> x22)" 
        assume c2: "osfx y x22" 
        assume c4: " f y" 
        show "inth (ifilter f (suffixes (suffix k x22))) j =
         inth
          (suffix (ilen (ifilter f (suffixes x22)) - ilen (ifilter f (suffixes (suffix k x22))))
                  (ifilter f (suffixes x22)))
          j" 
         using Suc c0 by simp
       next
        assume d0: "\<sigma> = x21 \<odot> x22" 
        assume d1: "\<not> f (x21 \<odot> x22)" 
        assume d2: "\<forall>x\<in>iset (suffixes x22). \<not> f x" 
        show " inth (ifilter f (suffixes (suffix k x22))) j = x21 \<odot> x22"
         using Suc d0 d1 d2 osfx_suffix
         using osfx_order.order_trans by fastforce
      qed
   qed
qed

lemma sfxfilter_suffix_suffix:
 assumes "k\<le>ilen \<sigma>"
         "\<exists>x \<in> iset (suffixes(suffix k \<sigma>)). f x"
 shows   " (ifilter f (suffixes (suffix k \<sigma>))) = 
           (suffix (ilen(ifilter f (suffixes \<sigma>))-ilen((ifilter f (suffixes (suffix k \<sigma>)))))
                   (ifilter f (suffixes \<sigma>)))"
by (simp add: assms(1) assms(2) interval_eq_inth_eq sfxfilter_suffix_ilen sfxfilter_suffix_inth)

lemma sfxfilter_suffix_suffix_a:
assumes "jj\<le>ilen (ifilter f (suffixes xs))"
         "\<exists>x \<in> iset (suffixes xs). f x"
shows "(suffix jj (ifilter f (suffixes xs)))  =
        ifilter f (suffixes (suffix ((ilen xs) - ilen (inth (ifilter f (suffixes xs)) jj))   xs))"

proof -
 have 1: "(suffix jj (ifilter f (suffixes xs))) = 
           ifilter f (suffix (inth (nfilter f (suffixes xs) 0) jj) (suffixes xs))" 
      using ifilter_nfilter_suffix_1 
      by (simp add: ifilter_nfilter_suffix_1 assms(1) assms(2))
 have 2: "(inth (nfilter f (suffixes xs) 0) jj) \<le> ilen (suffixes xs)"
   by (metis add_cancel_right_left assms(1) assms(2) nfilter_ilen nfilter_upper_bound) 
 have 3: "(suffix (inth (nfilter f (suffixes xs) 0) jj) (suffixes xs)) =
           (suffixes (suffix (inth (nfilter f (suffixes xs) 0) jj) xs))"
   using suffix_suffixes[of "(inth (nfilter f (suffixes xs) 0) jj)" "xs" ] "2" by blast
 have 4: "(inth (ifilter f (suffixes xs)) jj) = inth (suffixes xs) (inth (nfilter f (suffixes xs) 0) jj) "
    by (metis nfilter_imap_ifilter assms(2) inth_imap)
 have 5: "ilen(suffixes xs) \<le>  ilen (xs)" 
    by simp
 have 6: " (inth (nfilter f (suffixes xs) 0) jj) \<le> ilen (xs)" 
  using 2 5 le_trans by blast
 have 7: "inth (suffixes xs) (inth (nfilter f (suffixes xs) 0) jj) =
          (suffix (inth (nfilter f (suffixes xs) 0) jj) xs)"
   by (simp add: "6" inth_suffixes)  
 have 8: "(inth (nfilter f (suffixes xs) 0) jj) =
          ((ilen xs) - ilen (inth (ifilter f (suffixes xs)) jj))" 
   by (simp add: "4" "6" "7")
 show ?thesis 
 using "1" "3" "8" by auto
qed
 
lemma sfx_suffix_upperbound:
 "(\<forall>jj<(ilen(ifilter f (suffixes \<sigma>)) - ilen (ifilter f (suffixes (suffix k \<sigma>)))). 
              ((ilen \<sigma>) - ilen (inth (ifilter f (suffixes \<sigma>)) jj)) < k
            )" 
proof 
 fix jj
 show "jj < ilen (ifilter f (suffixes \<sigma>)) - ilen (ifilter f (suffixes (suffix k \<sigma>))) \<longrightarrow>
          ilen \<sigma> - ilen (inth (ifilter f (suffixes \<sigma>)) jj) < k "  
 proof (induct k arbitrary: \<sigma> jj)
 case 0
 then show ?case by simp
 next
 case (Suc k)
 then show ?case 
   proof (cases \<sigma>)
   case (INil x1)
   then show ?thesis using Suc by simp
   next
   case (ICons x21 x22)
   then show ?thesis 
      proof (cases jj)
      case 0
      then show ?thesis using Suc ICons using Suc_diff_le less_Suc_eq_0_disj by auto 
      next
      case (Suc nat)
      then show ?thesis
        using Suc.hyps  ICons Suc_diff_le less_Suc_eq_0_disj diff_is_0_eq gr_implies_not_zero by auto 
      qed
   qed
 qed
qed



end
