(*  Title     : A Shallow Encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Projection operator\<close>   

theory Projection
imports Fuse TimeReversal
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="
  cvc4 z3 e spass vampire
"]
*)
text \<open>
This theory introduces the projection operator \cite{moszkowski95:_compos}. 
The projection operator is defined and we prove the soundness of the rules and axiom system. 
\<close>

subsection \<open>Definitions\<close>

primrec filt ::  "'a interval \<Rightarrow> index \<Rightarrow> 'a interval"
 where    "filt \<sigma> \<langle>l\<rangle>     = \<langle>inth \<sigma> l\<rangle>"
        | "filt \<sigma> (x\<odot>ls) = (inth \<sigma> x) \<odot> filt \<sigma> ls"
   
primrec lsum :: "'a interval interval \<Rightarrow> nat \<Rightarrow> index"
 where  "lsum \<langle>xs\<rangle> a = \<langle>a+(ilen xs)\<rangle>"
      | "lsum (xs \<odot> xxs) a = (a+(ilen xs)) \<odot> (lsum xxs (a+(ilen xs)))"

definition addzero :: "index \<Rightarrow> index"
 where "addzero ls = (if ilen ls = 0 then 
        (if ifirst ls = 0 then ls else 0 \<odot> ls) else 0 \<odot> ls)"

definition powerinterval :: "('a::world) formula \<Rightarrow> 'a interval \<Rightarrow> index \<Rightarrow> bool"
 where "powerinterval F \<sigma> l = (\<forall> i. i<ilen l \<longrightarrow> ((sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> F))"

definition cpl :: "('a::world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a interval \<Rightarrow> index"
 where "cpl f g \<sigma> = (\<some> l. index_sequence 0 l \<and> powerinterval f \<sigma> l \<and> 
                          (inth l (ilen l)) = ilen \<sigma> \<and> ((filt \<sigma> l) \<Turnstile> g))"

primrec lcpl :: " ('a ::world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a interval \<Rightarrow> index \<Rightarrow> nat interval interval"
 where  " lcpl  f g \<sigma> \<langle>x\<rangle> = \<langle>(imap (shift x) (cpl f g (sub x x \<sigma>)))\<rangle>"
      | " lcpl  f g \<sigma> (x \<odot> xs) =  
            (case xs of \<langle>y\<rangle> \<Rightarrow> \<langle>(imap (shift x) (cpl f g (sub x y \<sigma>)))\<rangle> 
             | y \<odot> ys \<Rightarrow> (imap (shift x) (cpl f g (sub x y \<sigma>))) \<odot> (lcpl f g \<sigma> xs)) " 


definition projection_d :: "('a:: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
where "projection_d F G \<equiv> \<lambda>s. (\<exists> l. index_sequence 0 l  \<and> (inth l (ilen l)) = (ilen s) \<and>
                                  powerinterval F s l \<and> ((filt s l) \<Turnstile> G)
                          )"

syntax
 "_projection_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<triangle> _)" [84,84] 83)

syntax (ASCII)
 "_projection_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ proj _)" [84,84] 83)

translations
 "_projection_d"    \<rightleftharpoons> "CONST projection_d"

definition uprojection_d :: "('a:: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
where "uprojection_d F G \<equiv> LIFT(\<not>(F \<triangle> (\<not> G)))"


syntax
 "_uprojection_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<nabla> _)" [84,84] 83)

syntax (ASCII)
 "_uprojection_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ uproj _)" [84,84] 83)

translations
 "_uprojection_d"    \<rightleftharpoons> "CONST uprojection_d"

definition dp_d :: "('a:: world) formula \<Rightarrow> 'a formula"
where "dp_d F \<equiv> LIFT(#True \<triangle> F)"

definition bp_d :: "('a:: world) formula \<Rightarrow> 'a formula"
where "bp_d F \<equiv> LIFT(#True \<nabla> F)"

syntax
 "_dp_d"     :: "lift \<Rightarrow> lift"        ("(dp _)" [88] 87)
 "_bp_d"     :: "lift \<Rightarrow> lift"        ("(bp _)" [88] 87)

syntax (ASCII)
 "_dp_d"     :: "lift \<Rightarrow> lift"        ("(dp _)" [88] 87)
 "_bp_d"     :: "lift \<Rightarrow> lift"        ("(bp _)" [88] 87)

translations
 "_dp_d"    \<rightleftharpoons> "CONST dp_d"
 "_bp_d"    \<rightleftharpoons> "CONST bp_d"


subsection \<open>Lemmas\<close>

subsubsection \<open>filt Lemmas\<close>

lemma filt_ilen:
 "ilen(filt s0 l) = ilen l"
by (induct l, simp, simp)

lemma filt_inth:
 assumes " i \<le> ilen (filt s0 l) "
 shows   " (inth (filt s0 l) i) = (inth s0 (inth l i)) "
using assms
proof (induct l arbitrary: i)
case (INil x)
then show ?case by simp
next
case (ICons x1a l)
then show ?case by simp (metis Nitpick.case_nat_unfold Suc_eq_plus1 le_diff_conv)
qed



lemma filt_expand: 
 " (s1 = (filt s0 l)) =
   ( ilen s1 = ilen l \<and>
     (\<forall> i \<le> ilen s1. (inth s1 i) = (inth s0 (inth l i))))"
by (metis filt_ilen filt_inth interval_eq_inth_eq )

lemma filt_fuse:
   " filt xs (fuse l1 l2) = (fuse (filt xs l1) (filt xs l2))"
by (induct l1 arbitrary: l2 xs) simp_all


lemma fuse_filt_ilen:
assumes "index_sequence 0 l"
        "(inth l (ilen l)) = ilen xs"
shows  "ilen (filt (fuse (prefix (inth l n) xs) (suffix (inth l n) xs)) l) =
        ilen (fuse (filt (prefix (inth l n) xs) (prefix n l)) 
               (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) )))"
proof -
 have 1: "ilen (filt (fuse (prefix (inth l n) xs) (suffix (inth l n) xs)) l) = ilen l "
    by (simp add: filt_ilen)
 have 2: "ilen (fuse (filt (prefix (inth l n) xs) (prefix n l)) 
                       (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) ))) =
          (ilen((filt (prefix (inth l n) xs) (prefix n l))) +
           ilen(filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) ))) "
    using fuse_ilen_a by blast
 have 3: "ilen((filt (prefix (inth l n) xs) (prefix n l))) = ilen(prefix n l) "
    using filt_ilen by blast
 have 4: "ilen(filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) )) =
          ilen(suffix n l)"
    by (simp add: filt_ilen)
 have 5: "ilen(prefix n l) + ilen(suffix n l) = ilen l"
   by (simp)
 from 1 2 3 4 5 show ?thesis by auto
qed
    

lemma fuse_filt_inth_a:
 assumes "index_sequence 0 l"
         "(inth l (ilen l)) = ilen xs"
         "i \<le> ilen l"
         "n \<le> ilen l"
 shows   " inth (filt (fuse (prefix (inth l n) xs) (suffix (inth l n) xs)) l) i =
           inth xs (inth l i)"
proof -
  have 1: "filt (fuse (prefix (inth l n) xs) (suffix (inth l n) xs)) l =
           filt xs l "
     using assms by (metis fuse_prefix_suffix idx_less_last_1 le_neq_implies_less
        less_imp_le_nat not_less)
  have 2: "inth (filt xs l) i = inth xs (inth l i)"
     using assms by (metis filt_inth filt_ilen  )
  from 1 2 show ?thesis by auto
qed


lemma fuse_filt_inth_b:
 assumes "index_sequence 0 l"
         "(inth l (ilen l)) = ilen xs"
         "i \<le> ilen l"
         "n \<le> ilen l"
 shows   "inth (fuse (filt (prefix (inth l n) xs) (prefix n l)) 
                  (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) ))) i =
          inth xs (inth l i)"
proof -
 have 1: "i \<le> ilen(fuse (filt (prefix (inth l n) xs) (prefix n l)) 
                  (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) ))) "
    using assms 
    by (metis filt_ilen fuse_ilen_a fuse_prefix_suffix_ilen ilen_imap) 
 have 2: " ilast(filt (prefix (inth l n) xs) (prefix n l)) = inth xs (inth l n) "
  using assms filt_inth[of _ "(prefix (inth l n) xs)" "(prefix n l)" ] 
  by (metis filt_ilen idx_less_equal ilast_prefix  prefix_ilen_good order_refl) 
 have 3: "ifirst(filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) )) =
          inth (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) )) 0"
    by simp
 have 4: "inth (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) )) 0 =
          inth (suffix (inth l n) xs) (inth (imap (shiftm (inth l n)) (suffix n l) ) 0)"
         using filt_inth by blast
 have 5: "inth (suffix (inth l n) xs) (inth (imap (shiftm (inth l n)) (suffix n l) ) 0) =
          inth  (suffix (inth l n) xs) (inth l (n+0) - (inth l n))"
    by (simp add: assms inth_imap shiftm_def)   
 have 6: "inth  (suffix (inth l n) xs) (inth l (n+0) - (inth l n)) = 
          inth  (suffix (inth l n) xs) 0 "
    by simp 
 have 7: "inth  (suffix (inth l n) xs) 0 = inth xs (inth l n) "
    using assms by (metis Nat.add_0_right inth_suffix le0  ) 
 have 8: "ilast(filt (prefix (inth l n) xs) (prefix n l)) =
          ifirst(filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) )) "
   using "2" "4" "5" "7" by auto
 have 10: "inth (fuse (filt (prefix (inth l n) xs) (prefix n l)) 
                  (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) ))) i =
          (if i \<le> ilen (filt (prefix (inth l n) xs) (prefix n l)) 
           then inth (filt (prefix (inth l n) xs) (prefix n l)) i
           else inth (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) )) 
                    (i - ilen (filt (prefix (inth l n) xs) (prefix n l)))) "
    using "1" "8" fuse_inth by auto
 have 11: "ilen (filt (prefix (inth l n) xs) (prefix n l)) = n "
   using assms by (metis filt_ilen prefix_ilen_good) 
 have 12: "i\<le>n \<longrightarrow> inth (filt (prefix (inth l n) xs) (prefix n l)) i = 
                   inth (prefix (inth l n) xs) (inth (prefix n l ) i)"
    by (simp add: "11" filt_inth)
 have 13: "i\<le>n \<longrightarrow> inth (prefix (inth l n) xs) (inth (prefix n l ) i) =
                   inth (prefix (inth l n) xs) (inth l i) "
    by (simp add: assms)
 have 15: " i\<le>n \<longrightarrow> inth (prefix (inth l n) xs) (inth l i) = inth xs (inth l i)"
         using assms(1) assms(2) assms(4) idx_less_equal inth_prefix by blast
 have 16: "inth (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) )) 
                    (i - ilen (filt (prefix (inth l n) xs) (prefix n l)))  =
           inth (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) )) 
                    (i - n)"
    by (simp add: "11") 
 have 17: "inth (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) )) (i - n) =
           inth (suffix (inth l n) xs) (inth (imap (shiftm (inth l n)) (suffix n l) ) (i -n)) "
    using assms filt_inth[of "i-n" "(suffix (inth l n) xs)" "(imap (shiftm (inth l n)) (suffix n l) )" ]
      by (simp add: filt_ilen) 
 have 18: "i> n \<longrightarrow> inth (imap (shiftm (inth l n)) (suffix n l) ) (i -n) =
           inth l (n+(i-n)) - (inth l n)"
    using assms 
    by (simp add: idx_shiftm_suffix_inth)
 have 19: "i>n \<longrightarrow> inth (suffix (inth l n) xs) (inth (imap (shiftm (inth l n)) (suffix n l) ) (i -n)) 
                   = inth (suffix (inth l n) xs) ((inth l i) - (inth l n))" 
    by (simp add: "18")    
 have 20 : "i>n \<longrightarrow> inth (suffix (inth l n) xs) ((inth l i) - (inth l n)) = 
                    inth xs ((inth l n) + ((inth l i) - (inth l n)))"
    using assms by (metis diff_le_mono idx_less_last_1 inth_suffix 
        le_neq_implies_less less_imp_le_nat nat_le_linear)
 have 22: "i>n \<longrightarrow> (inth l n) + ((inth l i) - (inth l n)) =  (inth l i)"
    using assms using idx_less_equal by fastforce
 have 23: "i>n \<longrightarrow> inth xs ((inth l n) + ((inth l i) - (inth l n))) = inth xs (inth l i) "
    by (simp add: "22")
 from 10 show ?thesis by (simp add: "11" "12" "13" "15" "17" "19" "20" "23")
qed

lemma fuse_filt_inth:
assumes "index_sequence 0 l"
        "(inth l (ilen l)) = ilen xs"
        "i \<le> ilen l"
        "n \<le> ilen l"
shows  "inth (filt (fuse (prefix (inth l n) xs) (suffix (inth l n) xs)) l) i =
        inth (fuse (filt (prefix (inth l n) xs) (prefix n l)) 
                  (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) ))) i"
using assms fuse_filt_inth_a[of "l" "xs" "i" "n"]   
            fuse_filt_inth_b[of "l" "xs" "i" "n"] by simp

lemma fuse_filt:
assumes "index_sequence 0 l"
        "(inth l (ilen l)) = ilen xs"
        "n \<le> ilen l"
shows  "filt (fuse (prefix (inth l n) xs) (suffix (inth l n) xs)) l =
        fuse (filt (prefix (inth l n) xs) (prefix n l)) 
             (filt (suffix (inth l n) xs) (imap (shiftm (inth l n)) (suffix n l) ))"
using assms fuse_filt_ilen fuse_filt_inth interval_eq_inth_eq by (metis filt_ilen)


lemma fuse_filt_a:
 assumes "index_sequence 0 l1"
         "index_sequence (ilast l1) l2"
         "ilast l2 = ilen xs"
 shows   " filt (fuse (prefix (ilast l1) xs) (suffix (ifirst l2) xs)) (fuse l1 l2) =
           fuse (filt (prefix (ilast l1) xs) l1) 
                (filt (suffix (ifirst l2) xs) (imap (shiftm (ifirst l2)) l2)) "
proof -
 have 1: "ilast l1 = ifirst l2" 
  using assms by (metis index_sequence_def)
 have 2: "index_sequence 0 (fuse l1 l2)"
   using assms by (metis index_sequence_def idx_fuse)
 have 3: "ilast(fuse l1 l2) = ilen xs"
   using assms by (metis "1" add.left_neutral fuse_inth_a fuse_ilen_a le_add2)  
 have 4: "ilen l1 \<le> ilen (fuse l1 l2) " 
       by (simp add: fuse_ilen_a)
 have 5: "filt (fuse (prefix (inth (fuse l1 l2) (ilen l1)) xs)
                     (suffix (inth (fuse l1 l2) (ilen l1)) xs))
               (fuse l1 l2)
          =
         fuse (filt (prefix (inth (fuse l1 l2) (ilen l1)) xs) (prefix (ilen l1) (fuse l1 l2))) 
              (filt (suffix (inth (fuse l1 l2) (ilen l1)) xs) 
                    (imap (shiftm (inth (fuse l1 l2) (ilen l1))) (suffix (ilen l1) (fuse l1 l2))))"
   using "2" "3" "4" fuse_filt by auto
 have 6: "filt (fuse (prefix (ilast l1) xs) (suffix (ifirst l2) xs)) (fuse l1 l2) =
          filt (fuse (prefix (inth (fuse l1 l2) (ilen l1)) xs) 
                     (suffix (inth (fuse l1 l2) (ilen l1)) xs)) 
               (fuse l1 l2)" 
    using "1" "4" ilast_prefix[of "ilast l1" "fuse l1 l2"]   
     prefix_fuse by (simp add: fuse_inth)
 have 7: "(filt (prefix (ilast l1) xs) l1) =
          (filt (prefix (inth (fuse l1 l2) (ilen l1)) xs) (prefix (ilen l1) (fuse l1 l2)))" 
    using "1" "4" ilast_prefix[of "(ilast l1)" "fuse l1 l2"]  
                  prefix_fuse 
                  by (simp add: prefix_fuse fuse_inth) 
 have 8: "(filt (suffix (ifirst l2) xs) (imap (shiftm (ifirst l2)) l2)) =
          (filt (suffix (inth (fuse l1 l2) (ilen l1)) xs) 
                (imap (shiftm (inth (fuse l1 l2) (ilen l1))) (suffix (ilen l1) (fuse l1 l2)) ))"
    using "1" "4" ilast_prefix prefix_fuse suffix_fuse by metis
 show ?thesis using "5" "6" "7" "8" by auto
qed  

lemma filt_prefix:
assumes "n \<le> ilen l"
shows   "  prefix n (filt \<sigma> l) = filt \<sigma> (prefix n l) "
proof -
 have 1: "ilen (prefix n (filt \<sigma> l)) = ilen ( filt \<sigma> (prefix n l))"
   by (simp add: assms filt_ilen)
 have 2: "\<forall> i\<le>ilen(prefix n (filt \<sigma> l)). 
           (inth (prefix n (filt \<sigma> l)) i) = (inth ( filt \<sigma> (prefix n l)) i)"
   by (simp add: assms filt_inth filt_ilen le_trans)
 show ?thesis using "1" "2" interval_eq_inth_eq by blast
qed


lemma filt_prefix_idx:
assumes "n \<le> ilen \<sigma>"
        "index_sequence 0 l"
        "inth l (ilen l) = n "
shows  " filt \<sigma> l = filt (prefix n \<sigma>) l"
proof -
 have 1: " ilen(filt \<sigma> l) = ilen (filt (prefix n \<sigma>) l) " 
   by (simp add: filt_ilen)
 have 2: "\<forall>i\<le>ilen(filt \<sigma> l). (inth (filt \<sigma> l) i) = (inth (filt (prefix n \<sigma>) l) i)"
   by (metis assms filt_inth filt_ilen idx_less_equal inth_prefix prefix_ilen_good order_refl) 
 show ?thesis 
 by (simp add: "1" "2" interval_eq_inth_eq)
qed


lemma filt_suffix:
assumes "n \<le> ilen l"
shows  "  suffix n (filt \<sigma> l) = filt \<sigma> (suffix n l) "
proof -
 have 1: "ilen (suffix n (filt \<sigma> l)) = ilen (filt \<sigma> (suffix n l))"
   by (simp add: assms filt_ilen)
 have 2: "\<forall>i\<le>ilen (suffix n (filt \<sigma> l)). 
   (inth (suffix n (filt \<sigma> l)) i) = (inth (filt \<sigma> (suffix n l)) i) "
   by (simp add: assms filt_inth filt_ilen 
       ordered_cancel_comm_monoid_diff_class.le_diff_conv2)
 show ?thesis using "1" "2" interval_eq_inth_eq by blast
qed


lemma filt_suffix_idx_ilen:
assumes "n \<le> ilen \<sigma>"
        "index_sequence 0 l"
        "inth l (ilen l) = ilen \<sigma> - n"
shows " ilen (filt \<sigma> (imap (shift n) l)) = ilen (filt (suffix n \<sigma>) l)"
using assms by (simp add: filt_ilen)

lemma filt_suffix_idx_inth:
assumes "n \<le> ilen \<sigma>"
        "index_sequence 0 l"
        "i\<le>ilen l"
        "inth l (ilen l) = ilen \<sigma> - n"
shows " inth (filt \<sigma> (imap (shift n) l)) i = inth (filt (suffix n \<sigma>) l) i"
proof -
 have 1: " inth (filt \<sigma> (imap (shift n) l)) i = inth \<sigma> (inth (imap (shift n) l) i)  "
 by (simp add: assms filt_inth filt_ilen) 
 have 2: "inth \<sigma> (inth (imap (shift n) l) i) = (inth \<sigma> ((inth l i) + n))"
 by (simp add:  shift_def inth_imap) 
 have 3: " inth (filt (suffix n \<sigma>) l) i = inth (suffix n \<sigma>) (inth l i)  "  
   by (simp add: assms filt_inth filt_ilen)
 have 4: "inth (suffix n \<sigma>) (inth l i) = (inth \<sigma> ((inth l i) + n)) " 
    by (metis add.commute assms eq_iff idx_less_equal inth_suffix 
        suffix_ilen_good)
 show ?thesis by (simp add: "1" "2" "3" "4")
qed


lemma filt_suffix_idx:
assumes "n \<le> ilen \<sigma>"
        "index_sequence 0 l"
        "inth l (ilen l) = ilen \<sigma> - n"
shows " filt \<sigma> (imap (shift n) l) = filt (suffix n \<sigma>) l"
using assms  filt_suffix_idx_ilen filt_suffix_idx_inth interval_eq_inth_eq
by (simp add: filt_suffix_idx_inth interval_eq_inth_eq filt_ilen)

lemma filt_filt:
 " (inth (filt (filt xxs l1) l2) k) = 
   (inth xxs  (inth l1 (inth l2 k)))"
by (metis filt_expand ilen_imap inth_imap)

lemma filt_filt_imap:
 "(filt (filt xxs l1) l2) = (filt xxs (imap (\<lambda> x. (inth l1 x)) l2))"
by (metis filt_expand ilen_imap inth_imap)

lemma filt_imap:
 " filt xs l = imap (\<lambda>x. inth xs x) l"
 by (metis filt_expand ilen_imap inth_imap)

lemma filt_imap_filt:
 "(filt (filt xxs l1) l2) = filt xxs (filt l1 l2)"
by (metis filt_filt_imap filt_imap)

lemma filt_sub:
 assumes "k \<le> n"
         "n \<le> ilen l"
 shows   " (sub k n (filt \<sigma> l)) = (filt \<sigma> (sub k n l))"
using assms 
by (simp add: sub_def filt_prefix filt_suffix)

lemma filt_lfuse_imap:
 "  (filt \<sigma> (lfuse (xxs)) )=
     (lfuse (imap (\<lambda> xs . (filt \<sigma> xs)) xxs))    "
proof (induct xxs)
case (INil x)
then show ?case by simp
next
case (ICons x1a xxs)
then show ?case by (simp add: filt_fuse)
qed



subsubsection \<open>powerinterval lemmas\<close>

lemma powerinterval_splita0:
 assumes " index_sequence 0 l"
         " n \<le> ilen l "
         " inth l (ilen l) = ilen \<sigma>"
         " i < ilen(prefix n l)"
 shows   "  (sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) (prefix (inth l n) \<sigma>)) =
            (sub (inth l i) (inth l (Suc i)) \<sigma>) 
              " 
proof -
 have 01: "(inth l n) \<le> ilen \<sigma>" 
   by (metis assms(1) assms(2) assms(3) idx_less_last_1 le_less)
 have 02: "(inth (prefix n l) i) \<le> (inth (prefix n l) (Suc i))"
   using assms(1) assms(3) assms(4) idx_expand by fastforce
 have 03: "(inth (prefix n l) (Suc i)) \<le> (inth l n)" 
    using assms(1) assms(2) assms(3) assms(4) idx_less_equal by fastforce
 have 1: "ilen (sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) (prefix (inth l n) \<sigma>)) =
          ilen(sub (inth l i) (inth l (Suc i)) (\<sigma>)) 
         " using ilen_sub 
         using "01" "02" "03" assms(4) by auto
 have 2: " (\<forall> j \<le> ilen (sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) (prefix (inth l n) \<sigma>)).
               (inth (sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) (prefix (inth l n) \<sigma>)) j) =
               (inth (sub (inth l i) (inth l (Suc i)) (\<sigma>)) j)) 
               "
   proof 
    fix j
    show "j \<le> ilen (sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) (prefix (inth l n) \<sigma>)) \<longrightarrow>
               (inth (sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) (prefix (inth l n) \<sigma>)) j) =
               (inth (sub (inth l i) (inth l (Suc i)) (\<sigma>)) j)
         "
    proof -
     have 21: "ilen (sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) (prefix (inth l n) \<sigma>)) = 
               (inth l (Suc i)) - (inth l i) "
        using "1" assms(1) assms(3) assms(4) idx_expand by fastforce
     have 22: "(inth (prefix n l) i) = (inth l i) "
      using assms by auto
     have 23: "(inth (prefix n l) (Suc i)) = (inth l (Suc i)) "
        using assms by auto
     have 24: "j  \<le> (inth l (Suc i)) - (inth l i) \<longrightarrow> 
               inth (sub (inth l i) (inth l (Suc i)) \<sigma>) j = (inth \<sigma> ((inth l i)+j)) "
        using assms idx_expand inth_sub less_le_trans by fastforce
     have 25: "j  \<le> (inth l (Suc i)) - (inth l i) \<longrightarrow> (inth l (Suc i)) \<le> ilen (prefix (inth l n) \<sigma>) "
         using assms(1) assms(2) assms(3) assms(4) idx_expand idx_less_equal 
         by fastforce
     have 26: "j \<le> (inth l (Suc i)) - (inth l i) \<longrightarrow> 
               (inth (sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) (prefix (inth l n) \<sigma>)) j)  =
               (inth (prefix (inth l n) \<sigma>) ((inth l i)+j)) "
               using "02" "22" "23" "25" by auto
     have 27: "(inth l (Suc i))  \<le>(inth l n)" 
        using assms(1) assms(2) assms(3) assms(4) idx_less_equal by fastforce
     have 28: "j \<le> (inth l (Suc i)) - (inth l i) \<longrightarrow> (inth l i)+j\<le>(inth l n)" 
         using "27" assms(1) assms(3) assms(4) idx_expand by fastforce
     have 29: "j \<le> (inth l (Suc i)) - (inth l i) \<longrightarrow> 
                (inth (prefix (inth l n) \<sigma>) ((inth l i)+j)) = (inth \<sigma> ((inth l i)+j))"
         using assms inth_prefix using "28" by blast
     show ?thesis 
     using "21" "22" "23" "24" "26" "29" by auto
   qed
  qed
  show ?thesis 
  using "1" "2" interval_eq_inth_eq by blast
qed

lemma powerinterval_splita:
 assumes " index_sequence 0 l"
         " n \<le> ilen l"
         " inth l (ilen l) = ilen \<sigma>"
         " powerinterval f \<sigma> l "
 shows   " powerinterval f (prefix (inth l n) \<sigma>) (prefix n l)"
proof -
 have 0: "(\<forall> i. i<ilen l \<longrightarrow> ( (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f)) "
   using assms by (simp add: powerinterval_def)
 have 1: " powerinterval f (prefix (inth l n) \<sigma>) (prefix n l) =
           (\<forall>i. i < ilen(prefix n l) \<longrightarrow>
                ((sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) (prefix (inth l n) \<sigma>)) \<Turnstile> f)) "
       using powerinterval_def by blast
 have 2: "(\<forall>i. i < ilen(prefix n l) \<longrightarrow>
                ((sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) (prefix (inth l n) \<sigma>)) \<Turnstile> f)) "
  proof 
    fix i
    show "i < ilen(prefix n l) \<longrightarrow>
                ((sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) (prefix (inth l n) \<sigma>)) \<Turnstile> f) " 
    proof -
     have 21: " i<n \<longrightarrow> ((sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f)  " 
       using "0" assms less_le_trans by blast
     have 22: "i<ilen(prefix n l) \<longrightarrow> 
               ((sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) \<sigma>) \<Turnstile> f) "
       by (simp add: "21" assms)
     have 23: "i<ilen(prefix n l) \<longrightarrow>  
               ((sub (inth (prefix n l) i) (inth (prefix n l) (Suc i)) (prefix (inth l n) \<sigma>)) \<Turnstile> f)"
      using "22" assms powerinterval_splita0 by fastforce     
     show ?thesis using "23" by blast
    qed
   qed
  show ?thesis using "1" "2" by blast
qed

lemma powerinterval_splitb0:
 assumes " index_sequence 0 l "
         " n \<le> ilen l"
         "  inth l (ilen l) = ilen \<sigma>"
         "  i < (ilen l -n) "
 shows   "(sub (inth (((imap (shiftm (inth l n))  (suffix n l)))) i) 
                   (inth (((imap (shiftm (inth l n))  (suffix n l)))) (Suc i)) 
                   (suffix (inth l n) \<sigma>)) =
          (sub (inth l (i+n)) (inth l ((Suc i)+n)) \<sigma>) "
proof -
 have 1: "ilen(((imap (shiftm (inth l n))  (suffix n l)))) = (ilen l) - n "
      by (simp add: assms) 
 have 2: " i<(ilen l -n) \<longrightarrow> 
               (inth (((imap (shiftm (inth l n))  (suffix n l)))) i) =
                (inth l (n + i)) - (inth l n) "
      using assms idx_shiftm_suffix_inth by force
 have 3: "i<(ilen l -n) \<longrightarrow> 
              (inth (((imap (shiftm (inth l n))  (suffix n l)))) (Suc i)) =
              (inth l (n + (Suc i))) - (inth l n) "
       using assms idx_shiftm_suffix_inth by fastforce
 have 4: "i<(ilen l -n) \<longrightarrow>
              (sub (inth (((imap (shiftm (inth l n))  (suffix n l)))) i) 
                   (inth (((imap (shiftm (inth l n))  (suffix n l)))) (Suc i)) 
                   (suffix (inth l n) \<sigma>)) =
                (sub ( (inth l (n + i)) - (inth l n) )
                     ( (inth l (n + (Suc i))) - (inth l n) )
                   (suffix (inth l n) \<sigma>)) " 
          by (simp add: "2" "3")
 have 5: " i<(ilen l -n) \<longrightarrow> (inth l (n + i))  < (inth l (n + (Suc i)))"
  using assms index_sequence_def by auto
 have 6: "i<(ilen l -n) \<longrightarrow> (inth l n) \<le> (inth l (n + i))" 
    using assms by (metis add.commute idx_less_equal le_add1 less_diff_conv less_imp_le_nat)
 have 7: "i<(ilen l -n) \<longrightarrow> (inth l n) \<le> (inth l (n + (Suc i)))" 
    using "5" "6" by linarith
 have 8: " i<(ilen l -n) \<longrightarrow> (inth l (n + i)) - (inth l n) < (inth l (n + (Suc i))) - (inth l n)" 
   using "5" "6" diff_less_mono by blast
 have 9: "i<(ilen l -n) \<longrightarrow> (inth l (n + (Suc i))) - (inth l n) \<le> ilen \<sigma> - (inth l n)"
   by (metis Suc_leI add.commute assms diff_le_mono idx_less_last_1 le_eq_less_or_eq 
             less_diff_conv ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 10: "i<(ilen l -n) \<longrightarrow> 
              (sub ( (inth l (n + i)) - (inth l n) )
                     ( (inth l (n + (Suc i))) - (inth l n) )
                   (suffix (inth l n) \<sigma>)) = 
               (sub (inth l (i+n)) (inth l ((Suc i)+n)) \<sigma>)  "
   using sub_suffix 
   by (metis "6" "7" "8" "9" add.commute le_add_diff_inverse2)
 show ?thesis 
 using "2" "3" "10" assms by auto   
qed

lemma powerinterval_splitb:
 assumes " index_sequence 0 l"
         "  n \<le> ilen l"
         "  inth l (ilen l) = ilen \<sigma> "
         " powerinterval f \<sigma> l "
 shows   " powerinterval f (suffix (inth l n) \<sigma>) ((imap (shiftm (inth l n))  (suffix n l)))"
proof -
 have 0: "(\<forall> i. i<ilen l \<longrightarrow> ( (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f)) "
   using assms by (simp add: powerinterval_def)
 have 1: "powerinterval f (suffix (inth l n) \<sigma>) ((imap (shiftm (inth l n))  (suffix n l))) =
         (\<forall> i. i<ilen(((imap (shiftm (inth l n))  (suffix n l)))) \<longrightarrow> 
            ( (sub (inth (((imap (shiftm (inth l n))  (suffix n l)))) i) 
                   (inth (((imap (shiftm (inth l n))  (suffix n l)))) (Suc i)) 
                   (suffix (inth l n) \<sigma>)) \<Turnstile> f))   "
     by (simp add: powerinterval_def)
 have 2: "(\<forall> i. i<ilen(((imap (shiftm (inth l n))  (suffix n l)))) \<longrightarrow> 
            ( (sub (inth (((imap (shiftm (inth l n))  (suffix n l)))) i) 
                   (inth (((imap (shiftm (inth l n))  (suffix n l)))) (Suc i)) 
                   (suffix (inth l n) \<sigma>)) \<Turnstile> f)) "
  proof 
   fix i
   show "i<ilen(((imap (shiftm (inth l n))  (suffix n l)))) \<longrightarrow> 
            ( (sub (inth (((imap (shiftm (inth l n))  (suffix n l)))) i) 
                   (inth (((imap (shiftm (inth l n))  (suffix n l)))) (Suc i)) 
                   (suffix (inth l n) \<sigma>)) \<Turnstile> f) "
   proof -
    have 21: " i< (ilen l) -n \<longrightarrow> 
               ((sub (inth l (i+n)) (inth l ((Suc i)+n)) \<sigma>) \<Turnstile> f)"
        by (simp add: "0")
    show ?thesis 
    using "21" assms powerinterval_splitb0 by fastforce
   qed
  qed
 show ?thesis using "1" "2" by blast
qed
       
lemma powerinterval_split:
 assumes " index_sequence 0 l"
         "  n \<le> ilen l"
         "  inth l (ilen l) = ilen \<sigma>"
 shows   " powerinterval f \<sigma> l =
           (powerinterval f (prefix (inth l n) \<sigma>) (prefix n l) \<and>
            powerinterval f (suffix (inth l n) \<sigma>) ((imap (shiftm (inth l n))  (suffix n l))))"
proof -
 have 1: "powerinterval f \<sigma> l \<Longrightarrow> 
          powerinterval f (prefix (inth l n) \<sigma>) (prefix n l) "
   by (simp add: assms powerinterval_splita)
 have 2: "powerinterval f \<sigma> l \<Longrightarrow> 
          powerinterval f (suffix (inth l n) \<sigma>) ((imap (shiftm (inth l n))  (suffix n l))) "
   by (simp add: assms powerinterval_splitb)
 have 3: "powerinterval f (prefix (inth l n) \<sigma>) (prefix n l) =
          (\<forall>i. i<n \<longrightarrow> ( (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f)) "
  by (metis assms prefix_ilen_good powerinterval_def powerinterval_splita0)
 have 4: "powerinterval f (suffix (inth l n) \<sigma>) ((imap (shiftm (inth l n))  (suffix n l))) =
         (\<forall> i. i<(ilen l) -n \<longrightarrow> ((sub (inth l (i+n)) (inth l ((Suc i)+n)) \<sigma>) \<Turnstile> f)) "
   by (simp add: assms powerinterval_def powerinterval_splitb0)
 have 5: "(\<forall> i. i<(ilen l) -n \<longrightarrow> ((sub (inth l (i+n)) (inth l ((Suc i)+n)) \<sigma>) \<Turnstile> f)) =
          (\<forall> i. n \<le> i \<and> i<(ilen l) \<longrightarrow> ((sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f)) "
     by (metis le_add2 le_add_diff_inverse2 less_diff_conv plus_nat.simps(2))
 have 5: "(powerinterval f (prefix (inth l n) \<sigma>) (prefix n l) \<and>
            powerinterval f (suffix (inth l n) \<sigma>) ((imap (shiftm (inth l n))  (suffix n l)))) \<Longrightarrow>
           powerinterval f \<sigma> l "
  using "3" "4" "5" assms powerinterval_def 
  by (metis not_less_eq not_less_less_Suc_eq order.order_iff_strict)
 show ?thesis 
 using "1" "2" "5" by blast  
qed

lemma powerinterval_fuse:
 assumes " index_sequence 0 l1"
         " index_sequence 0 l2"
         " ilast l1 = cp"
         "  cp \<le> ilen \<sigma> "
         " ilast l2  = ilen \<sigma> - cp"
 shows   " powerinterval f \<sigma> (fuse l1 (imap (shift cp) l2)) =
           (powerinterval f (prefix cp \<sigma>) l1 \<and> 
            powerinterval f (suffix cp \<sigma>) l2 )"
proof -
 have 1: "index_sequence 0 (fuse l1 (imap (shift cp) l2)) "
  using assms idx_fuse[of "l1" "(imap (shift cp) l2)"]   
  by (metis index_sequence_def idx_link) 
 have 2: "cp = (inth l1 (ilen l1))"
  using assms inth_ilen_ilast by blast
 have 3: "ilen l1 \<le> ilen (fuse l1 (imap (shift cp) l2))"
   by (simp add: fuse_ilen_a)
 have 4: "inth (fuse l1 (imap (shift cp) l2)) (ilen (fuse l1 (imap (shift cp) l2))) = ilen \<sigma>" 
    by (metis assms idx_fuse_ilen ilen_gr_zero)
 have 5: "cp = inth (fuse l1 (imap (shift cp) l2)) (ilen l1)"
    by (metis "3" assms fuse_inth idx_fuse_ifirst_ilast order_refl) 
 have 6: "(imap (shiftm cp) (imap (shift cp) l2)) = l2 "
  using assms by (metis idx_link lsk_ls)
 have 7: "ilast l1 = ifirst (imap (shift cp) l2)"
   by (metis assms idx_fuse_ifirst_ilast)
  show ?thesis 
  using "1" "3" "4" "5" "6" "7" prefix_fuse suffix_fuse powerinterval_split 
   by fastforce
qed


lemma powerinterval_idx:
 "(powerinterval (LIFT(f \<and> more)) \<sigma> l \<and> (inth l 0) = 0 \<and> (inth l (ilen l)) = ilen \<sigma>) =
  (index_sequence 0 l \<and> (inth l (ilen l)) = ilen \<sigma> \<and> powerinterval f \<sigma> l)"
proof (auto simp add: powerinterval_def index_sequence_def more_defs)
 show "\<And>n. \<forall>i<ilen l.
            f (sub ( inth l i) ( inth l (Suc i)) \<sigma>) \<and>
            0 < ilen (sub ( inth l i) ( inth l (Suc i)) \<sigma>) \<Longrightarrow>
          inth l 0 = 0 \<Longrightarrow>
          inth l (ilen l) = ilen \<sigma> \<Longrightarrow> n < ilen l \<Longrightarrow> 
          inth l n <  inth l (Suc n)"
   by (simp add:  sub_def)
 show "\<And>i.  inth l 0 = 0 \<Longrightarrow>
         \<forall>n<ilen l.  inth l n <  inth l (Suc n) \<Longrightarrow>
          inth l (ilen l) = ilen \<sigma> \<Longrightarrow>
         \<forall>i<ilen l. f (sub ( inth l i) ( inth l (Suc i)) \<sigma>) \<Longrightarrow>
         i < ilen l \<Longrightarrow> 0 < ilen (sub ( inth l i) ( inth l (Suc i)) \<sigma>) " 
   by (simp add:  sub_def) 
       (metis index_sequence_def idx_less_last_1)
qed

subsubsection \<open>cpl lemmas\<close>

lemma cpl_expand:
assumes "(\<exists> l. index_sequence 0 l \<and> powerinterval f \<sigma> l \<and> (inth l (ilen l)) = ilen \<sigma> \<and>
             ((filt \<sigma> l) \<Turnstile> g  )) "
shows  " index_sequence 0 (cpl f g \<sigma>) \<and> powerinterval f \<sigma> (cpl f g \<sigma>) \<and>
           (inth (cpl f g \<sigma>) (ilen (cpl f g \<sigma>))) = ilen \<sigma> \<and> ((filt \<sigma> (cpl f g \<sigma>)) \<Turnstile> g)"
proof -
 have 0: "cpl f g \<sigma> = (\<some> l. index_sequence 0 l \<and> powerinterval f \<sigma> l \<and> 
                         (inth l (ilen l)) = ilen \<sigma> \<and> ((filt \<sigma> l) \<Turnstile> g)) "
   using cpl_def by force
 have 1: "(\<exists> l. index_sequence 0 l \<and> powerinterval f \<sigma> l \<and> (inth l (ilen l)) = ilen \<sigma> \<and>
            ((filt \<sigma> l) \<Turnstile> g))"
      using assms by auto
 have 2: "index_sequence 0 (cpl f g \<sigma>) \<and> powerinterval f \<sigma> (cpl f g \<sigma>) \<and>
           (inth (cpl f g \<sigma>) (ilen (cpl f g \<sigma>))) = ilen \<sigma> \<and> ((filt \<sigma> (cpl f g \<sigma>)) \<Turnstile> g)"
  using "0" "1" 
   someI_ex[of "\<lambda>l. index_sequence 0 l \<and> powerinterval f \<sigma> l \<and> (inth l (ilen l)) = ilen \<sigma> \<and>
            ((filt \<sigma> l) \<Turnstile> g)"]  by simp
 show ?thesis 
 using "2" by blast
qed  

lemma cpl_projection:
 "(\<sigma> \<Turnstile> f \<triangle> g) =  
  ( index_sequence 0 (cpl f g \<sigma>) \<and> powerinterval f \<sigma> (cpl f g \<sigma>) \<and>
           (inth (cpl f g \<sigma>) (ilen (cpl f g \<sigma>))) = ilen \<sigma> \<and> g (filt \<sigma> (cpl f g \<sigma>)) )"
using cpl_expand by (simp add: projection_d_def, blast)

lemma cpl_empty:
 assumes "ilen \<sigma> = 0 \<and> (\<sigma> \<Turnstile> f \<triangle> g)"
 shows   "(cpl f g \<sigma>) = \<langle>0\<rangle>"
 using assms cpl_projection idx_less_last_1 INil_ilen by fastforce   

lemma cpl_empty_a:
 assumes "ilen \<sigma> = 0"
         "(cpl f g \<sigma>) = \<langle>0\<rangle>"
         "g(filt \<sigma> \<langle>0\<rangle>)"
 shows   "(\<sigma> \<Turnstile> f \<triangle> g)"
proof -
 have 1: "index_sequence 0 \<langle>0\<rangle> "
   by (simp add: index_sequence_def)
 have 2: "powerinterval f \<sigma> \<langle>0\<rangle> "
   by (simp add: powerinterval_def)
 have 3: "(inth \<langle>0\<rangle> (ilen \<langle>0\<rangle>)) = ilen \<sigma> "
    by (simp add: assms)
 have 4: " g(filt \<sigma> \<langle>0\<rangle>)"
    using assms by blast
 from 1 2 3 4 show ?thesis 
 by (simp add: assms cpl_projection)
qed  


lemma cpl_more:
 assumes "ilen \<sigma> >0"
         "(\<sigma> \<Turnstile> f \<triangle> g)"
 shows   "ilen(cpl f g \<sigma>)>0"
by (metis assms cpl_projection gr0I index_sequence_def)

lemma cpl_more_than_first:
 assumes "ilen \<sigma> >0"
         "(\<sigma> \<Turnstile> f \<triangle> g)"
 shows   "(inth (cpl f g \<sigma>) 0) = 0"
using assms cpl_projection index_sequence_def by auto 

lemma cpl_more_than_last:
 assumes "ilen \<sigma> >0"
         "(\<sigma> \<Turnstile> f \<triangle> g)"
 shows   "(inth (cpl f g \<sigma>) (ilen (cpl f g \<sigma>))) = ilen \<sigma>"
using assms cpl_projection by blast

lemma cpl_sub_more:
 assumes "n < k"
         "k \<le> ilen \<sigma>"
         "((sub n k \<sigma>) \<Turnstile> f \<triangle> g)"
 shows   " ilen(cpl f g (sub n k \<sigma>)) > 0"
using assms 
by (simp add: cpl_more)

lemma cpl_bounds:
 assumes "n < k"
         "k \<le> ilen \<sigma>"
         "((sub n k \<sigma>) \<Turnstile> f \<triangle> g)"
         "i < ilen (cpl f g (sub n k \<sigma>))" 
 shows   "0 \<le> (inth (cpl f g (sub n k \<sigma>)) i) \<and> (inth (cpl f g (sub n k \<sigma>)) i) \<le> k -n"
using assms 
by (metis cpl_projection idx_less_last_1 ilen_sub le0 less_imp_le_nat)

lemma cpl_imap_bounds:
 assumes "n < k"
         "k \<le> ilen \<sigma>"
         "((sub n k \<sigma>) \<Turnstile> f \<triangle> g)"
         "i < ilen (imap (shift n) (cpl f g (sub n k \<sigma>))) " 
 shows   "n \<le> (inth (imap (shift n) (cpl f g (sub n k \<sigma>))) i) \<and> 
               (inth (imap (shift n) (cpl f g (sub n k \<sigma>))) i) \<le> k "
using assms 
by (metis  shift_def Nat.le_diff_conv2 cpl_bounds ilen_imap 
    inth_imap le_add2 less_imp_le_nat)

lemma cpl_ifirst:
 assumes "(sub x1a (ifirst l) \<sigma>) \<Turnstile> f \<triangle> g"
 shows   "ifirst((imap (shift x1a) (cpl f g (sub x1a (ifirst l) \<sigma>))))  = x1a"
proof -
 have 1: "( index_sequence 0 (cpl f g (sub x1a (ifirst l) \<sigma>)) \<and> 
            powerinterval f (sub x1a (ifirst l) \<sigma>) (cpl f g (sub x1a (ifirst l) \<sigma>)) \<and>
           (inth (cpl f g (sub x1a (ifirst l) \<sigma>)) (ilen (cpl f g (sub x1a (ifirst l) \<sigma>)))) = 
           ilen (sub x1a (ifirst l) \<sigma>) \<and> 
           g (filt (sub x1a (ifirst l) \<sigma>) (cpl f g (sub x1a (ifirst l) \<sigma>))) ) " 
          using cpl_projection assms by auto
 have 2: "index_sequence 0 (cpl f g (sub x1a (ifirst l) \<sigma>)) "
   using "1" by auto
 have 3: " (inth (cpl f g (sub x1a (ifirst l) \<sigma>)) 0) =0"
    using 2 index_sequence_def by blast
 show ?thesis 
 by (metis "3"  shift_def add.left_neutral inth_imap)
qed

lemma cpl_ifirst_same:
 assumes "(sub x1a x1a \<sigma>) \<Turnstile> f \<triangle> g"
 shows   "ifirst((imap (shift x1a) (cpl f g (sub x1a x1a \<sigma>))))  = x1a"
proof -
 have 1: " ifirst (\<langle> x1a \<rangle>) = x1a"  
     by auto
 from 1 cpl_ifirst show ?thesis by (metis assms)
qed

lemma cpl_ilast:
 assumes " ((sub x (ifirst l) \<sigma>) \<Turnstile> f \<triangle> g) \<and> x < ifirst l \<and> ifirst l \<le> ilen \<sigma> "
 shows   "ilast((imap (shift x) (cpl f g (sub x (ifirst l) \<sigma>))))  = ifirst l"
 proof -
  have 01: "( index_sequence 0 (cpl f g (sub x (ifirst l) \<sigma>)) \<and> 
            powerinterval f (sub x (ifirst l) \<sigma>) (cpl f g (sub x (ifirst l) \<sigma>)) \<and>
           (inth (cpl f g (sub x (ifirst l) \<sigma>)) (ilen (cpl f g (sub x (ifirst l) \<sigma>)))) = 
           ilen (sub x (ifirst l) \<sigma>) \<and> 
           g (filt (sub x (ifirst l) \<sigma>) (cpl f g (sub x (ifirst l) \<sigma>))) ) "
     using cpl_projection assms by (simp add: cpl_projection)
  have 02: "(inth (cpl f g (sub x (ifirst l) \<sigma>)) (ilen (cpl f g (sub x (ifirst l) \<sigma>)))) = 
             ilen (sub x (ifirst l) \<sigma>) "
        using 01 by auto
  have 03: "ilast(cpl f g (sub x (ifirst l) \<sigma>)) = 
               (inth (cpl f g (sub x (ifirst l) \<sigma>)) (ilen (cpl f g (sub x (ifirst l) \<sigma>))))"
        by simp
  have 04: "x < ifirst l "
    using assms by blast
  have 05: "ilen (sub x (ifirst l) \<sigma>) = (ifirst l) -x "
   using assms ilen_sub less_imp_le_nat by blast
  have 06: "ilast((imap (shift x) (cpl f g (sub x (ifirst l) \<sigma>)))) =
             ((ifirst l) -x)+x "
    by (metis "02" "05"  shift_def ilen_imap inth_imap)
  show ?thesis using "04" "06" by auto
qed

lemma cpl_ilast_i:
 assumes "((sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         "(inth l i) < (inth l (Suc i))"
         "(inth l (Suc i)) \<le> ilen \<sigma> "
 shows   "ilast((imap (shift (inth l i)) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))))  = 
          (inth l (Suc i))"
 proof -
  have 01: "( index_sequence 0 (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) \<and> 
            powerinterval f (sub (inth l i) (inth l (Suc i)) \<sigma>) 
                            (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) \<and>
           (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) 
                (ilen (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)))) = 
               ilen (sub (inth l i) (inth l (Suc i)) \<sigma>) \<and> 
           g (filt (sub (inth l i) (inth l (Suc i)) \<sigma>) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))) )"
     using cpl_projection assms by (simp add: cpl_projection)
  have 02: "(inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) 
                 (ilen (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)))) = 
             ilen (sub (inth l i) (inth l (Suc i)) \<sigma>) "
        using 01 by auto
  have 03: "ilast(cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) = 
               (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) 
                    (ilen (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))))"
        by simp
  have 04: "(inth l i) < (inth l (Suc i)) "
        by (simp add: assms)
  have 05: "ilen (sub (inth l i) (inth l (Suc i)) \<sigma>) = (inth l (Suc i)) -(inth l i) "
        by (simp add:  assms less_imp_le_nat)
  have 06: "ilast((imap (shift (inth l i)) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)))) =
             ((inth l (Suc i)) -(inth l i))+(inth l i) "
    by (metis "02" "05"  shift_def ilen_imap inth_imap)
  show ?thesis using "04" "06" by auto
qed


subsubsection \<open>lcpl lemmas\<close>


lemma lcpl_inth:
 assumes "index_sequence (inth l 0) l"
         " i < ilen l"
 shows   " (inth (lcpl f g \<sigma> l) i) = 
           (imap (shift (inth l i)) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)))"
using assms
proof (induct l arbitrary: i)
case (INil x)
then show ?case by simp
next
case (ICons x1a ls)
then show ?case 
   proof (cases i)
   case 0
   then show ?thesis 
     proof (cases ls)
     case (INil x1)
     then show ?thesis by (simp add: "0")
     next
     case (ICons x21 x22)
     then show ?thesis by (simp add: "0")
     qed
   next
   case (Suc nat)
   then show ?thesis 
     proof (cases ls)
     case (INil x1)
     then show ?thesis 
     by (metis ICons.prems(2) One_nat_def Suc ilen.simps(1) ilen.simps(2) leD le_add1
          plus_1_eq_Suc)
     next
     case (ICons x21 x22)
     then show ?thesis 
     using ICons.hyps ICons.prems(1) ICons.prems(2) Suc idx_ICons by auto
     qed
   qed
qed


lemma lcpl_ilen:
 assumes "index_sequence (inth l 0) l"
         " ilen l > 0 " 
 shows   " ilen(lcpl f g \<sigma> l) = ilen l -1"
using assms 
 proof 
  (induct l)
  case (INil x)
  then show ?case by simp
  next
  case (ICons x1a l)
  then show ?case 
   proof (cases l)
   case (INil x1)
   then show ?thesis by simp
   next
   case (ICons x21 x22)
   then show ?thesis using ICons.hyps ICons.prems(1) idx_ICons by auto
   qed
qed

lemma lcpl_ilen_zero:
 assumes "index_sequence (inth l 0) l"
         "ilen l = 0" 
 shows   "ilen(lcpl f g \<sigma> l) = 0"
using assms
by (metis One_nat_def ilen.simps(1) ilen.simps(2) interval.exhaust lcpl.simps(1) leD le_add1 
    le_imp_less_Suc order_refl) 


lemma lcpl_last:
 assumes "index_sequence (inth l 0) l"
         "(inth l (ilen l)) = ilen \<sigma>"
         " ilen l >0 "
 shows   " ilast (lcpl f g \<sigma> l) = 
             (imap (shift (inth l (ilen l-1))) 
                 (cpl f g (sub (inth l (ilen l -1)) (inth l (ilen l)) \<sigma>)))"
proof -
 have 1: "ilast (lcpl f g \<sigma> l) = (inth (lcpl f g \<sigma> l) (ilen (lcpl f g \<sigma> l))) "
   by simp
 have 2: "(inth (lcpl f g \<sigma> l) (ilen (lcpl f g \<sigma> l))) =
          (imap (shift (inth l (ilen (lcpl f g \<sigma> l)))) 
               (cpl f g (sub (inth l (ilen (lcpl f g \<sigma> l))) 
                             (inth l (Suc (ilen (lcpl f g \<sigma> l)))) 
                             \<sigma>))) "
      using assms lcpl_inth 
      by (metis One_nat_def Suc_pred diff_le_self lcpl_ilen le_eq_less_or_eq n_not_Suc_n )
 have 3: "(imap (shift (inth l (ilen (lcpl f g \<sigma> l)))) 
               (cpl f g (sub (inth l (ilen (lcpl f g \<sigma> l))) 
                             (inth l (Suc (ilen (lcpl f g \<sigma> l)))) 
                             \<sigma>))) =
          (imap (shift (inth l (ilen l-1))) 
               (cpl f g (sub (inth l (ilen l-1)) (inth l (Suc (ilen l-1))) \<sigma>))) "
        using assms by (metis lcpl_ilen )
 show ?thesis by (simp add: "2" "3" assms(3))
qed

lemma lcpl_last_last:
 assumes "index_sequence (inth l 0) l"
         "(inth l (ilen l)) = ilen \<sigma>"
         "((sub (inth l (ilen l -1) ) (inth l (ilen l)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " ilen l > 0 "
 shows   " ilast (ilast (lcpl f g \<sigma> l)) = ilen \<sigma> "
using assms 
by (metis (no_types, lifting) One_nat_def Suc_pred' cpl_ilast_i diff_Suc_less 
    dual_order.order_iff_strict index_sequence_def lcpl_last)

lemma lcpl_zero_zero:
  assumes "index_sequence (inth l 0) l"
          " (inth l (ilen l)) = ilen \<sigma>"
          " ilen l = 0"
 shows   "(inth (lcpl f g \<sigma> l) 0) = 
          (imap (shift (inth l 0)) (cpl f g (sub (inth l 0) (inth l 0) \<sigma>)))"
using assms 
by simp
 (metis INil_ilen lcpl.simps(1))
 
lemma lcpl_ifirst:
 assumes "index_sequence (inth l 0) l"
         "(inth l (ilen l)) = ilen \<sigma> "
         "(\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)  "
         " ilen l > 0"
 shows   " ifirst(ifirst((lcpl f g \<sigma> l))) = ifirst l"
proof -
 have 01: "(ifirst((lcpl f g \<sigma> l))) = 
           (imap (shift (inth l 0)) (cpl f g (sub (inth l 0) (inth l (Suc 0)) \<sigma>)))" 
      using assms by (metis lcpl_inth  )
 have 02: "ilen l > 0 \<longrightarrow> 
           ifirst((imap (shift (inth l 0)) (cpl f g (sub (inth l 0) (inth l (Suc 0)) \<sigma>)))) =
            (inth l 0)" 
            using assms by (metis Suc_leI cpl_ifirst ilast_ifirst ilast_prefix)
 show ?thesis 
 using "01" "02" inth_zero_ifirst by (simp add: assms(4))
qed

 
lemma lcpl_lfuse_lastfirst:
 assumes "index_sequence (inth l 0) l"
         "(inth l (ilen l)) = ilen \<sigma>"
         "((ilen l = 0 \<and> ((sub (inth l 0) (inth l 0) \<sigma>) \<Turnstile> f \<triangle> g)) \<or>
           (ilen l > 0 \<and> (\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)))"
 shows   "lastfirst (lcpl f g \<sigma> l)"
using assms 
proof 
  (induct l)
  case (INil x)
  then show ?case by simp
  next
  case (ICons x1a ls)
  then show ?case 
   proof -
    have 1: "ilen ls = 0 \<longrightarrow> lastfirst (lcpl f g \<sigma> (x1a \<odot> ls))     "
          using ICons.prems by (metis One_nat_def add_diff_cancel_left' INil_ilen 
              ilen.simps(2) lastfirst.simps(1) lcpl_ilen plus_1_eq_Suc zero_less_one)
    have 2: "ilen ls > 0 \<longrightarrow> lastfirst ( lcpl f g \<sigma> (x1a \<odot> ls) ) = 
               lastfirst((imap (shift x1a) (cpl f g (sub x1a (ifirst ls) \<sigma>))) \<odot> (lcpl f g \<sigma> ls))"
    by (metis (no_types, lifting) interval.simps(6) ilen_ICons_1 inth_zero 
        lcpl.simps(2))
    have 3: "ilen ls > 0 \<longrightarrow> 
             lastfirst((imap (shift x1a) (cpl f g (sub x1a (ifirst ls) \<sigma>))) \<odot> (lcpl f g \<sigma> ls)) =
              ( ilast( (imap (shift x1a) (cpl f g (sub x1a (ifirst ls) \<sigma>))) ) = 
               ifirst(ifirst((lcpl f g \<sigma> ls))) \<and> lastfirst((lcpl f g \<sigma> ls)) )"
      using lastfirst.simps(2) by blast
   have 4: "x1a < (ifirst ls)"
      using ICons.prems by (metis  index_sequence_def inth_Suc inth_zero 
                ilen.simps(2) plus_1_eq_Suc zero_less_Suc) 
    have 5: " (ifirst ls) \<le> ilen \<sigma>"
        using ICons.prems by (metis Suc_lessI eq_iff idx_less_last_1 
            inth_Suc ilen.simps(2) less_imp_le_nat plus_1_eq_Suc zero_less_Suc)
    have 6: "ilen ls >0 \<longrightarrow> 
              ilast( (imap (shift x1a) (cpl f g (sub x1a (ifirst ls) \<sigma>))) ) = ifirst ls"
      using ICons.prems by (metis "4" "5" cpl_ilast inth_Suc inth_zero less_le_not_le)
    have 7: "index_sequence (inth ls 0) ls"
       using assms ICons.prems index_sequence_def by auto
    have 8: "inth ls (ilen ls) = ilen \<sigma> "
        using ICons.prems by auto
    have 9: "( (\<forall> i<ilen ( ls). (sub (inth ( ls) i) (inth ( ls) (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g))"
       using ICons.prems by auto   
    have 10: "ilen ls >0 \<longrightarrow> ifirst(ifirst((lcpl f g \<sigma> ls))) = ifirst ls  "
      using "7" "8" "9" 
      by (metis cpl_ifirst ilen_ICons_1 inth_Suc  lcpl_inth)
    have 11: "ilen ls >0 \<longrightarrow> 
                lastfirst((lcpl f g \<sigma> ls))"
      using "7" "8" "9" ICons.hyps by blast
   show ?thesis using "1" "10" "11" "2" "6" by auto
  qed
qed

lemma lcpl_lfuse_ilen:
 assumes "index_sequence (inth l 0) l"
         "(inth l (ilen l)) = ilen \<sigma> " 
         "((ilen l =0 \<and> ((sub (inth l 0 ) (inth l 0 ) \<sigma> \<Turnstile> f \<triangle> g))) \<or>
           (ilen l >0 \<and> (\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)))"
 shows   "(ilen l =0 \<longrightarrow> ilen(lfuse (lcpl f g \<sigma> l)) = 0) \<and>
          (ilen l >0 \<longrightarrow> 
            ilen(lfuse (lcpl f g \<sigma> l)) = (\<Sum>k=0..ilen l-1. ilen (inth (lcpl f g \<sigma> l) k)))"
proof -
 have 1: "lastfirst (lcpl f g \<sigma> l)" 
  using assms lcpl_lfuse_lastfirst by blast
 have 2: "ilen l = 0 \<longrightarrow> 
           lfuse (lcpl f g \<sigma> l) =  
           ((imap (shift (inth l 0)) (cpl f g (sub (inth l 0) (inth l 0) \<sigma>))))" 
        by (metis assms(1) assms(2) INil_ilen lcpl_ilen_zero lcpl_zero_zero lfuse_INil)
 have 3: "ilen l = 0 \<longrightarrow> 
           ilen ((imap (shift (inth l 0)) (cpl f g (sub (inth l 0) (inth l 0) \<sigma>)))) = 0"
    using assms cpl_empty ilen_sub by fastforce
 have 4: "ilen l = 0 \<longrightarrow> ilen(lfuse (lcpl f g \<sigma> l)) = 0"
    using "2" "3" by auto 
 have 5: "ilen l > 0 \<longrightarrow>
            ilen(lfuse (lcpl f g \<sigma> l)) = (\<Sum>k=0..ilen l-1. ilen (inth (lcpl f g \<sigma> l) k)) "
   using lfuse_ilen 
   by (metis assms lcpl_ilen lcpl_lfuse_lastfirst)
 from 4 5 show ?thesis by simp
qed

lemma lcpl_lfuse_idx:
 assumes "index_sequence 0 l"
         " (inth l (ilen l)) = ilen \<sigma>" 
         "(\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         "ilen l > 0 "
 shows   "index_sequence (ifirst (lfuse  (lcpl f g \<sigma> l)))  (lfuse ( (lcpl f g \<sigma> l)))"
proof -
 have 0: "ilen \<sigma> >0 \<longrightarrow> ilen l >0 "
  using assms gr_zeroI index_sequence_def by fastforce
 have 2: "ilen \<sigma> >0 \<longrightarrow> lastfirst (lcpl f g \<sigma> l)"
           using "0" assms index_sequence_def lcpl_lfuse_lastfirst by blast
 have 3: " (\<forall>i<ilen l. index_sequence 0 (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))) "
     using assms cpl_projection by auto
 have 4: " (\<forall>i<ilen l. 
              index_sequence (inth l i) 
                             (imap (shift (inth l i)) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))))"
    using "3" idx_link by blast
 have 5: "ilen \<sigma> >0 \<longrightarrow>
           (\<forall>i<ilen l. index_sequence (inth l i) (inth (lcpl f g \<sigma> l) i) ) "
    using assms by (metis "4" index_sequence_def lcpl_inth)
 have 6: " ilen \<sigma> >0 \<longrightarrow>
          (\<forall> i<ilen l. ifirst (inth (lcpl f g \<sigma> l) i) = (inth l i)) "
    by (metis "5" index_sequence_def)
 have 7: " ilen \<sigma> >0 \<longrightarrow>
            (\<forall>i<ilen l. 
              index_sequence (ifirst (inth (lcpl f g \<sigma> l) i)) (inth (lcpl f g \<sigma> l) i) ) "
     using "6" "5" by simp
 have 8: "ilen \<sigma> >0 \<longrightarrow>
           index_sequence (ifirst (lfuse  (lcpl f g \<sigma> l)))  (lfuse ( (lcpl f g \<sigma> l))) "
     using assms 
     by (metis (mono_tags, lifting) "2" "7" Suc_diff_1 index_sequence_def idx_lfuse 
         lcpl_ilen le_imp_less_Suc)
  from 8  show ?thesis 
  by (metis assms(1) assms(2) assms(4) index_sequence_def idx_less_last_1)
qed

lemma lcpl_ilen_inth_gr_zero:
 assumes "index_sequence 0 l"
         "(inth l (ilen l)) = ilen \<sigma>"
         "(\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " ilen \<sigma> >0"
 shows  " (\<forall> j \<le> ilen (lcpl f g \<sigma> l). ilen(inth (lcpl f g \<sigma> l) j) > 0) "
proof 
   fix j
   show "j \<le> ilen (lcpl f g \<sigma> l) \<longrightarrow> 0 < ilen ( inth (lcpl f g \<sigma> l) j) "
   proof -
   have 1: "ilen \<sigma>>0 \<longrightarrow> lastfirst (lcpl f g \<sigma> l)"
      by (metis assms index_sequence_def lcpl_lfuse_lastfirst neq0_conv)
   have 2: "ilen \<sigma>>0 \<longrightarrow> ilen l >0"
      using assms gr_zeroI index_sequence_def by fastforce
   have 3: "ilen \<sigma>>0 \<longrightarrow> j \<le> ilen (lcpl f g \<sigma> l) \<longrightarrow> 
           (inth (lcpl f g \<sigma> l) j) = 
             (imap (shift (inth l j)) (cpl f g (sub (inth l j) (inth l (Suc j)) \<sigma>)))" 
     using assms lcpl_inth[of "l" "j" "f" "g" "\<sigma>" ] 
     by (metis "2" One_nat_def Suc_pred index_sequence_def lcpl_ilen less_Suc_eq_le)
   have 4: "ilen \<sigma>>0 \<longrightarrow> j \<le> ilen (lcpl f g \<sigma> l) \<longrightarrow>
            ilen (imap (shift (inth l j)) (cpl f g (sub (inth l j) (inth l (Suc j)) \<sigma>))) > 0 "
     by (metis (no_types, lifting) "2" One_nat_def Suc_pred add.commute assms cpl_sub_more gr0I 
         index_sequence_def idx_expand ilen_imap lcpl_ilen 
         le_imp_less_Suc plus_1_eq_Suc)
   show ?thesis 
   using  "3" "4" by (simp add: assms(4))
   qed
qed  


lemma lcpl_ilast_inth:
 assumes "index_sequence 0 l"
         "(inth l (ilen l)) = ilen \<sigma>"
         "(\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " ilen \<sigma> > 0 " 
         " j \<le> ilen (lcpl f g \<sigma> l) "
 shows  "ilast(inth (lcpl f g \<sigma> l) j) = (inth l (Suc j))"
proof -
 have 0: "ilen \<sigma> >0 \<longrightarrow> ilen l >0"
    using assms gr_zeroI index_sequence_def by fastforce
 have 1: "ilen \<sigma>>0 \<longrightarrow> 
          j \<le> ilen (lcpl f g \<sigma> l)   \<longrightarrow>
           inth (lcpl f g \<sigma> l) j = 
            (imap (shift (inth l j)) (cpl f g (sub (inth l j) (inth l (Suc j)) \<sigma>)))
          " 
    using assms lcpl_inth[of "l" "j" "f" "g" "\<sigma>"] 
    proof -
    show ?thesis
    by (metis (no_types) "0" Suc_diff_1 
        \<open>\<lbrakk>index_sequence (inth l 0) l; j < ilen l\<rbrakk> \<Longrightarrow> 
          inth (lcpl f g \<sigma> l) j = 
          imap (shift ( inth l j)) (cpl f g (sub (inth l j) ( inth l (Suc j)) \<sigma>))\<close> 
        \<open>index_sequence 0 l\<close> index_sequence_def lcpl_ilen le_imp_less_Suc) 
    qed 
 have 2: "ilen \<sigma>>0 \<longrightarrow> 
          j \<le> ilen (lcpl f g \<sigma> l)   \<longrightarrow> (inth l (Suc j)) \<le> ilen \<sigma>"  
     using assms 
     by (metis "0" Suc_diff_1 Suc_eq_plus1 index_sequence_def idx_expand lcpl_ilen leD 
         not_less_eq)
 have 2: "ilen \<sigma>>0 \<longrightarrow> 
          j \<le> ilen (lcpl f g \<sigma> l)   \<longrightarrow>
          ilast (imap (shift (inth l j)) (cpl f g (sub (inth l j) (inth l (Suc j)) \<sigma>))) = 
           (inth l (Suc j))"
    using assms cpl_ilast_i[of "f" "g" "l" "j" "\<sigma>"]  
    by (metis "0" "2" One_nat_def Suc_pred index_sequence_def lcpl_ilen le_imp_less_Suc)
 show ?thesis using "1" "2" by (simp add: assms(4) assms(5)) 
qed  



lemma lcpl_lfuse_filt_power_help:
 assumes "index_sequence 0 l"
         "(inth l (ilen l)) = ilen \<sigma>" 
         "(\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " ilen \<sigma> > 0 " 
 shows   " (\<forall> i<ilen l. g (filt \<sigma> (inth (lcpl f g \<sigma> l) i)) )"
proof -
 have 1: "(\<forall> i<ilen l. g (filt (sub (inth l i) (inth l (Suc i)) \<sigma>) 
                                 (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) )) "
    using assms cpl_projection by blast 
 have 2: "ilen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<ilen l. 
            (filt \<sigma> (inth (lcpl f g \<sigma> l) i)) =
            (filt \<sigma> (imap (shift (inth l i)) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)))))
          "
    using assms by (metis index_sequence_def lcpl_inth) 
 have 3: "ilen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<ilen l. 
              ilen (filt \<sigma> (inth (lcpl f g \<sigma> l) i)) =
              ilen (filt (sub (inth l i) (inth l (Suc i)) \<sigma>) 
                                 (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) ) 
           ) 
          " 
        by (simp add: "2" filt_ilen)
 have 4: "ilen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<ilen l.
            (\<forall> j\<le> ilen (filt \<sigma> (inth (lcpl f g \<sigma> l) i)).
               (inth (filt \<sigma> (imap (shift (inth l i)) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)))) j) 
                =
                (inth \<sigma> ((inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) j) +(inth l i) ))
            )
          )"  
        using inth_imap shift_def filt_imap by metis
 have 5: "ilen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<ilen l.
            (\<forall> j\<le> ilen (filt \<sigma> (inth (lcpl f g \<sigma> l) i)).
               (inth (filt (sub (inth l i) (inth l (Suc i)) \<sigma>) 
                                 (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) ) j) =
               (inth (sub (inth l i) (inth l (Suc i)) \<sigma>) 
                    (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) j)) 
            ))"
       by (simp add: filt_imap inth_imap)  
 have 6: "ilen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<ilen l.
               (inth l (Suc i)) \<le> ilen \<sigma> 
           )"
      using assms idx_expand by auto 
 have 7: "ilen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<ilen l.
                (inth l i) \<le> (inth l (Suc i))
           ) "
       using assms index_sequence_def less_imp_le by blast 
 have 8: "ilen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<ilen l.
            (\<forall> j\<le> ilen (filt \<sigma> (inth (lcpl f g \<sigma> l) i)). 
                (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) j) \<le> (inth l (Suc i)) - (inth l i)
            ))"
            by (metis "3" "6" "7" assms(3) cpl_projection filt_ilen idx_less_last_1 ilen_sub 
                le_eq_less_or_eq)
 have 9: "ilen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<ilen l.
            (\<forall> j\<le> ilen (filt \<sigma> (inth (lcpl f g \<sigma> l) i)).
              (inth (sub (inth l i) (inth l (Suc i)) \<sigma>) 
                    (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) j)) =
              (inth \<sigma> ((inth  (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) j) + (inth l i)))
            ))
         "   using "6" 
         by (simp add:  "7" "8" add.commute)
  have 10: "ilen \<sigma> >0 \<longrightarrow> 
           (\<forall> i<ilen l.
            (\<forall> j\<le> ilen (filt \<sigma> (inth (lcpl f g \<sigma> l) i)).
               (filt (sub (inth l i) (inth l (Suc i)) \<sigma>) 
                                 (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) ) =
              (filt \<sigma> (inth (lcpl f g \<sigma> l) i)) 
             ))
               " 
            by (simp add: filt_expand) 
                (metis "2" "3" "4" "9" filt_ilen filt_imap inth_imap)
  show ?thesis using "1" "10" assms(4) by fastforce
qed



subsubsection \<open>lsum lemmas\<close>

lemma lsum_INil:
 "lsum \<langle>xs\<rangle> a = \<langle>a+ilen xs\<rangle>"
by simp

lemma lsum_addzero_INil:
 " addzero (lsum \<langle>xs\<rangle> 0) =  (if ilen xs = 0 then \<langle>0\<rangle> else  \<langle>0, ilen xs \<rangle>)"
by (simp add: addzero_def)

lemma lsum_addzero_ICons:
 " addzero (lsum (xs\<odot>xxs) 0) = 0 \<odot> (lsum (xs\<odot>xxs) 0)"
by (simp add: addzero_def)

lemma lsum_ifirst:
 "ifirst (lsum xxs a) = a+ilen(ifirst xxs)"
by (case_tac xxs) simp_all

lemma lsum_ilen:
 "ilen (lsum xxs a) = ilen xxs"
by (induct xxs arbitrary: a) simp_all
  
lemma lsum_addzero_ifirst:
 "ifirst (addzero (lsum xxs 0)) = 0"
by (simp add: addzero_def lsum_ifirst)

lemma lsum_addzero_ilen:
 "(ilen xxs = 0 \<and> ilen(ifirst xxs) = 0 \<longrightarrow> 
   ilen (addzero (lsum xxs 0)) =  0)
  \<and>
  (ilen xxs = 0 \<and> ilen(ifirst xxs) > 0 \<longrightarrow> 
   ilen (addzero (lsum xxs 0)) =  1)
  \<and>
  (ilen xxs > 0  \<longrightarrow>
   ilen (addzero (lsum xxs 0)) =  (ilen xxs) +1)
 "
by (simp add: addzero_def)
   (metis add.left_neutral lsum_ifirst lsum_ilen)

lemma lsum_inth_help:
assumes  "i>0"
         " i \<le> ilen xxs +1"
shows " (\<Sum>k = 0..(i-1). ilen ( inth (xxs) k)) =
            (\<Sum>k = 1..i. ilen ( inth (xxs) (k-1))) "
using assms 
proof 
  (induct i)
  case 0
  then show ?case by blast
  next
  case (Suc i)
  then show ?case 
  proof simp_all
  assume a1: "0 < i \<Longrightarrow> (\<Sum>k = 0..i - Suc 0. ilen ( inth xxs k)) = 
                        (\<Sum>k = Suc 0..i. ilen ( inth xxs (k - Suc 0)))"
  have f2: "\<forall>f. sum f {1::nat..0} = (0::nat)"
  by auto 
  have f3: "\<forall>n f. sum f {n::nat..n} = (f n::nat)"
  by simp 
  have f4: "\<forall>f n. (sum f {0::nat..n - 1} + (f n::nat) = sum f {0..n} \<or> 0 = n) \<or> \<not> 0 \<le> n"
  by (metis (no_types) One_nat_def Suc_pred le_eq_less_or_eq sum.atLeast0_atMost_Suc) (* 18 ms *)
  have f5: "\<forall>n. (0::nat) \<le> n"
  by blast 
  have "\<forall>n. (0::nat) + n = n"
  by linarith 
  then show "(\<Sum>n = 0..i. ilen ( inth xxs n)) = 
             (\<Sum>n = Suc 0..i. ilen ( inth xxs (n - Suc 0))) + ilen ( inth xxs i)"
  using f5 f4 f3 f2 a1 by (metis One_nat_def le_eq_less_or_eq) (* 103 ms *)
  qed
qed
  

lemma lsum_inth:
 assumes "i\<le> ilen xxs" 
 shows "inth (lsum xxs a) i = a+(\<Sum>k::nat= 0..i. ilen(inth xxs k))"
using assms
proof 
 (induct xxs arbitrary: a i)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xxs)
 then show ?case 
  proof -
   have 1: "inth (lsum (x1a \<odot> xxs) a) i = inth ((a+ilen x1a)\<odot> (lsum xxs (a+ilen x1a))) i"
    by simp
   have 2: "i \<le> ilen xxs +1"
    using ICons.prems by auto  
   have 3: "i =0 \<longrightarrow> 
            inth ((a+ilen x1a)\<odot> (lsum xxs (a+ilen x1a))) i = (a+ilen x1a)"
      by simp 
   have 4: "a+ilen x1a = a + (\<Sum>k = 0..0. ilen ( inth (x1a \<odot> xxs) k))" 
    by simp
   have 5: "i>0 \<and> i \<le> ilen xxs +1 \<longrightarrow> 
            inth ((a+ilen x1a)\<odot> (lsum xxs (a+ilen x1a))) i = 
             inth (lsum xxs (a+ilen x1a)) (i-1) "
        by (metis One_nat_def Suc_pred inth_Suc)
   have 6: "i>0 \<and> i \<le> ilen xxs +1 \<longrightarrow> 
            inth (lsum xxs (a+ilen x1a)) (i-1) = 
             (a+ilen x1a) + (\<Sum>k = 0..(i-1). ilen ( inth (xxs) k)) " 
      using ICons.hyps le_diff_conv by blast
   have 7: "i>0 \<and> i \<le> ilen xxs +1 \<longrightarrow> 
            (\<Sum>k = 0..(i-1). ilen ( inth (xxs) k)) =
            (\<Sum>k = 1..i. ilen ( inth (xxs) (k-1))) "
       using lsum_inth_help by blast
   have 8: "i>0 \<and> i \<le> ilen xxs +1 \<longrightarrow>  
             (\<Sum>k = 1..i. ilen ( inth (xxs) (k-1))) = 
             (\<Sum>k = 1..i. ilen ( inth (x1a \<odot> xxs) (k)))
             "
     by (metis (no_types, lifting) atLeastAtMost_iff inth_Suc
         ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc sum.cong) 
   have 9: "i>0 \<and> i \<le> ilen xxs +1 \<longrightarrow> 
            (a+ilen x1a) + (\<Sum>k = 1..i. ilen ( inth (x1a \<odot> xxs) (k))) =
             a + (\<Sum>k = 0..i. ilen ( inth (x1a \<odot> xxs) (k)))" 
          by (simp add: sum.atLeast_Suc_atMost)
   show ?thesis 
   by (metis "1" "2" "3" "4" "5" "6" "7" "8" "9" not_gr_zero)
  qed
qed


lemma lsum_addzero_inth:
 assumes "i\<le> ilen (addzero (lsum xxs 0))" 
 shows "(ilen xxs = 0 \<and> ilen(ifirst xxs) = 0 \<longrightarrow>
          inth (addzero (lsum xxs 0)) i = (inth (lsum xxs 0) i) )
        \<and>
        (ilen xxs = 0 \<and> ilen(ifirst xxs) > 0 \<longrightarrow>
          inth (addzero (lsum xxs 0)) i = (inth (0\<odot>(lsum xxs 0)) i) )
        \<and>
        (ilen xxs > 0 \<longrightarrow>
          inth (addzero (lsum xxs 0)) i = (inth (0\<odot>(lsum xxs 0)) i) )
       "
using assms
by (metis add.left_neutral addzero_def less_numeral_extra(3) lsum_ifirst lsum_ilen)


lemma lsum_ilast:
 " ilast (lsum xxs a) = a+(\<Sum>k::nat= 0..(ilen xxs). ilen(inth xxs k))"
by (metis le_refl lsum_ilen lsum_inth)


lemma lsum_addzero_ilast:
 "ilast (addzero (lsum xxs 0)) = ilast(lsum xxs 0)"
by (simp add: addzero_def)


lemma lsum_inth_leq_Suc:
 assumes "i < ilen xxs"
         " (\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)"
 shows " inth (lsum xxs a) i < inth (lsum xxs a) (Suc i)"
proof -
 have 1: "inth (lsum xxs a) i = a+(\<Sum>k::nat= 0..i. ilen(inth xxs k)) "
     using assms less_imp_le_nat lsum_inth by blast 
 have 2: "inth (lsum xxs a) (Suc i) = a+(\<Sum>k::nat= 0..(Suc i). ilen(inth xxs k))" 
   using assms by (simp add: lsum_inth Suc_leI)
 have 3: "(\<Sum>k::nat= 0..(Suc i). ilen(inth xxs k)) =
          (\<Sum>k::nat= 0..i. ilen(inth xxs k)) + ilen(inth xxs (Suc i))  " 
      using sum.atLeast0_atMost_Suc by blast
 have 4: " ilen(inth xxs (Suc i)) > 0"
    using assms by auto
 show ?thesis using "1" "2" "3" "4" by linarith 
qed 

lemma lsum_addzero_inth_leq_Suc:
 assumes "i < ilen(addzero (lsum xxs 0))"
         "(\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)"
 shows " inth (addzero (lsum xxs 0)) i < inth (addzero (lsum xxs 0)) (Suc i)"
proof (cases i)
case 0
then show ?thesis 
  by (metis add.left_neutral addzero_def assms(2) ilen_gr_zero inth_Suc 
      less_numeral_extra(3) lsum_addzero_ifirst lsum_ifirst)
next
case (Suc nat)
then show ?thesis
by (metis (no_types, lifting) Suc_less_eq2 add_diff_cancel_left' add_diff_cancel_right' addzero_def 
    assms(1) assms(2) inth_Suc lsum_addzero_ilen lsum_inth_leq_Suc neq0_conv not_add_less1 
    plus_1_eq_Suc) 
 
qed

lemma lsum_idx:
assumes "(\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)" 
shows "index_sequence (inth (lsum xxs a) 0) (lsum xxs a)"
by (simp add: assms index_sequence_def lsum_ilen lsum_inth_leq_Suc)

lemma lsum_addzero_idx:
assumes "(\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)" 
shows "index_sequence 0 (addzero (lsum xxs 0))"
by (metis idx_expand1 add.left_neutral addzero_def assms le0
           lsum_idx lsum_ifirst)


lemma filt_lfuse_lsum_a:
assumes "lastfirst (xs \<odot>xxs)"
        "(\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)"
        "ilen xs > 0"
shows "(filt (fuse xs (lfuse xxs)) (lsum xxs (ilen xs))) =
          (filt (lfuse xxs) ( (lsum xxs 0)))"
using assms 
proof 
  (induct xxs arbitrary: xs)
  case (INil x)
  then show ?case 
   proof -
    have 1: "filt (fuse xs (lfuse \<langle>x\<rangle>)) (lsum \<langle>x\<rangle> (ilen xs)) =
             filt (fuse xs x) (lsum \<langle>x\<rangle> (ilen xs))"
      by simp
    have 2: "filt (fuse xs x) (lsum \<langle>x\<rangle> (ilen xs)) =
             filt (fuse xs x) \<langle>ilen xs + ilen x\<rangle>"
      by simp
    have 3: "filt (fuse xs x) (\<langle>ilen xs + ilen x\<rangle>) =
             \<langle>(inth (fuse xs x) (ilen xs + ilen x))\<rangle> "
      by simp
    have 4: "\<langle>(inth (fuse xs x) (ilen xs + ilen x))\<rangle> =
             \<langle>(inth x (ilen x))\<rangle>"
    using INil.prems fuse_inth_a by auto
    have 5: "filt (lfuse \<langle>x\<rangle>) (lsum \<langle>x\<rangle> 0) =
             filt (x) (\<langle>ilen x\<rangle>) "
          by simp
    have 6: "filt (x) (\<langle>ilen x\<rangle>) = \<langle>(inth x (ilen x))\<rangle>"
            by auto  
    show ?thesis 
    by (simp add: "4")
   qed
  next
  case (ICons x1a xxs)
  then show ?case 
   proof -
    have 01: "filt (fuse xs (lfuse (x1a \<odot> xxs))) (lsum (x1a \<odot> xxs) (ilen xs)) =
              filt (fuse xs (fuse x1a (lfuse xxs))) (lsum (x1a \<odot> xxs) (ilen xs)) " 
      by simp
    have 02: "filt (fuse xs (fuse x1a (lfuse xxs))) 
                   (lsum (x1a \<odot> xxs) (ilen xs)) =
              filt (fuse xs (fuse x1a (lfuse xxs))) 
                   ((ilen xs + ilen x1a) \<odot>  lsum (xxs) (ilen xs +ilen x1a))" 
           by simp
    have 03: "filt (fuse xs (fuse x1a (lfuse xxs))) 
                   ((ilen xs + ilen x1a) \<odot>  lsum (xxs) (ilen xs +ilen x1a)) =
              (inth (fuse xs (fuse x1a (lfuse xxs))) (ilen xs + ilen x1a)) \<odot>
              filt (fuse xs (fuse x1a (lfuse xxs))) (lsum (xxs) (ilen xs +ilen x1a)) "
        by simp
    have 04: "(inth (fuse xs (fuse x1a (lfuse xxs))) (ilen xs + ilen x1a)) =
              (inth (fuse xs x1a) (ilen xs +ilen x1a)) "
        proof -
        have f1: " inth xs (ilen xs) =  inth x1a 0"
        using ICons.prems(1) by force 
        have f2: "lastfirst (x1a \<odot> xxs)"
        using ICons.prems(1) lastfirst.simps(2) by blast
        then have " inth x1a (ilen x1a) =  inth ( inth xxs 0) 0"
        by simp 
        then show ?thesis
        using f2 f1 
          by (metis ICons.prems(1) add.right_neutral fuse_ilen_a 
               fuse_inth_a ifirst_lfuse_ifirst  
             inth_zero le0 le_add1 lfuse_ICons) 
        qed
    have 05: "(inth (fuse xs x1a) (ilen xs +ilen x1a)) = (inth x1a (ilen x1a)) "
       using ICons.prems(1) fuse_inth_a by force
    have 06: " filt (fuse xs (fuse x1a (lfuse xxs))) (lsum (xxs) (ilen xs +ilen x1a)) =
               filt (fuse (fuse xs x1a) (lfuse xxs)) (lsum (xxs) (ilen(fuse xs x1a)))"
       using FusionAssoc[of xs x1a "(lfuse xxs)"]
       by (metis ICons.prems(1) fuse_ilen_a inth_zero lastfirst.simps(2) lastfirst_lfuse)   
    have 07: "lastfirst ( (fuse xs x1a)\<odot> xxs ) "
     by (metis "05" ICons.prems(1) fuse_ilen_a lastfirst.simps(2))
    have 08: " ilen (fuse xs x1a) > 0"
     by (simp add: ICons.prems fuse_ilen_a)
    have 09: "filt (fuse (fuse xs x1a) (lfuse xxs)) (lsum (xxs) (ilen(fuse xs x1a))) =
              filt (lfuse xxs) (lsum (xxs) 0)"
        by (metis "07" "08" ICons.prems(2) Suc_leI inth_Suc ilen.simps(2) less_Suc_eq_le 
           ICons(1) plus_1_eq_Suc)
    have 10: "filt (fuse xs (fuse x1a (lfuse xxs))) 
                   ((ilen xs + ilen x1a) \<odot>  lsum (xxs) (ilen xs +ilen x1a)) =
              (inth x1a (ilen x1a)) \<odot> filt (lfuse xxs) (lsum (xxs) 0)" 
     by (simp add: "04" "05" "06" "09")
    have 11: "filt (lfuse (x1a \<odot> xxs)) (lsum (x1a \<odot> xxs) 0) =
              filt (fuse x1a (lfuse xxs)) ((ilen x1a) \<odot> (lsum xxs (ilen x1a)))  "
     by simp
    have 12: "filt (fuse x1a (lfuse xxs)) ((ilen x1a) \<odot> (lsum xxs (ilen x1a))) =
              (inth (fuse x1a (lfuse xxs)) (ilen x1a)) \<odot> 
              (filt (fuse x1a (lfuse xxs)) (lsum xxs (ilen x1a)))"
        by simp
    have 13: "(inth (fuse x1a (lfuse xxs)) (ilen x1a)) = (inth x1a (ilen x1a))"
     by (metis ICons.prems(1) eq_iff fuse_ilen_a fuse_inth 
         ifirst_lfuse_ifirst lastfirst.simps(2) le_add1)
    have 14: "(filt (fuse x1a (lfuse xxs)) (lsum xxs (ilen x1a))) =
              filt (lfuse xxs) (lsum xxs 0)"
      using ICons.prems(1) ICons.prems(2) inth_zero ilen.simps(2) lastfirst.simps(2) 
        ICons(1) by fastforce
    have 15: "filt (fuse x1a (lfuse xxs)) ((ilen x1a) \<odot> (lsum xxs (ilen x1a))) = 
              (inth x1a (ilen x1a)) \<odot> (filt (lfuse xxs) (lsum xxs 0))"
     by (simp add: "13" "14") 
     show ?thesis using "10" "15" by auto
   qed
qed

lemma filt_lfuse_lsum:
assumes "lastfirst (xs \<odot>xxs)"
        " (\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)"
        " ilen xs > 0"
shows " (filt (lfuse (xs\<odot>xxs)) (addzero (lsum (xs\<odot>xxs) 0))) = 
        (ifirst xs)\<odot> (ilast xs)\<odot> (filt (lfuse xxs) (lsum xxs 0))"
proof -
 have 1: " lfuse (xs \<odot> xxs) = fuse xs (lfuse xxs) "
   by simp
 have 2: "(filt (lfuse (xs\<odot>xxs)) (addzero (lsum (xs\<odot>xxs) 0))) =
          (filt (fuse xs (lfuse xxs)) (addzero (lsum (xs\<odot>xxs) 0)))  "
  using "1" by simp
 have 3: "addzero (lsum (xs\<odot>xxs) 0) =   
         0\<odot>(ilen xs) \<odot> (lsum xxs (ilen xs)) "
   using lsum_addzero_ICons by auto
 have 4: "(filt (fuse xs (lfuse xxs)) (addzero (lsum (xs\<odot>xxs) 0))) =
          (filt (fuse xs (lfuse xxs)) (0\<odot>(ilen xs) \<odot> (lsum xxs (ilen xs)))) "
   using "3" by auto 
 have 5: " (filt (fuse xs (lfuse xxs)) (0\<odot>(ilen xs) \<odot> (lsum xxs (ilen xs)))) =
            (inth (fuse xs (lfuse xxs)) 0) \<odot> 
            (inth (fuse xs (lfuse xxs)) (ilen xs)) \<odot>
            (filt (fuse xs (lfuse xxs)) (lsum xxs (ilen xs)))  "
   by simp
 have 6: "(inth (fuse xs (lfuse xxs)) 0) = (inth xs 0)"
   using assms by (metis fuse_inth ifirst_lfuse_ifirst ilen_gr_zero 
       lastfirst.simps(2))
 have 7: "(inth (fuse xs (lfuse xxs)) (ilen xs)) = (inth xs (ilen xs)) "
   using assms by (metis fuse_ilen_a fuse_inth ifirst_lfuse_ifirst 
       lastfirst.simps(2) le_add1 order_refl)   
 have 8: "index_sequence 0 (addzero (lsum  xxs 0))"
   using assms lsum_addzero_idx by blast  
 have 9: "(filt (fuse xs (lfuse xxs)) (lsum xxs (ilen xs))) =
          (filt (lfuse xxs) (lsum xxs 0))"
     using assms filt_lfuse_lsum_a by blast
   show ?thesis 
   using "3" "6" "7" "9" inth_ilen_ilast inth_zero_ifirst by force
qed   
 
lemma filt_lfuse_lsum_1:
assumes "lastfirst (xs \<odot>xxs)" 
        " (\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)"
        " ilen xs > 0"
shows " (filt (lfuse (xs\<odot>xxs)) ((lsum (xs\<odot>xxs) 0))) = 
         (ilast xs)\<odot> (filt (lfuse xxs) (lsum xxs 0))"
using assms by (simp, 
                metis assms(1) eq_iff filt_lfuse_lsum_a fuse_ilen_a 
                fuse_inth ifirst_lfuse_ifirst lastfirst.simps(2) le_add1) 

lemma filt_lfuse_lsum_2:
assumes "lastfirst (xxs)"
        " (\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)"
        " j \<le> ilen xxs "
shows   " (inth (filt (lfuse (xxs)) ((lsum (xxs) 0))) j) = ilast(inth xxs j)"
using assms 
 proof (induct xxs arbitrary: j)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a xxs)
 then show ?case  
   proof (cases j)
   case 0
   then show ?thesis using ICons.hyps ICons.prems  filt_expand ifirst_lfuse_ifirst[of x1a xxs]
        by simp
           (metis "0" add.right_neutral fuse_inth_a le0)
   next
   case (Suc nat)
   then show ?thesis using ICons.hyps ICons.prems by simp_all
             (metis ICons.prems(1) Suc_le_mono filt_lfuse_lsum_a le_eq_less_or_eq old.nat.simps(4) 
          old.nat.simps(5) zero_less_Suc)
   qed
qed

lemma filt_lfuse_lsum_3:
assumes "lastfirst (xxs)"
        " (\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)"
        " j \<le> ilen (addzero (lsum xxs 0)) " 
shows   " ( j=0 \<longrightarrow> (inth (filt (lfuse (xxs)) (addzero(lsum (xxs) 0))) j) = ifirst(ifirst xxs))
          \<and>
          (j>0 \<longrightarrow> (inth (filt (lfuse (xxs)) (addzero(lsum (xxs) 0))) j) = ilast(inth xxs (j-1)))
        "
using assms 
proof -
 have 1: "j \<le> ilen (addzero (lsum xxs 0)) \<and> j=0 \<longrightarrow>
           (inth (filt (lfuse (xxs)) (addzero(lsum (xxs) 0))) j) = ifirst(ifirst xxs)  " 
     using assms by (metis filt_ilen filt_inth lastfirst_lfuse  
          lsum_addzero_ifirst)
 have 2: "j \<le> ilen (addzero (lsum xxs 0)) \<and> j>0 \<longrightarrow>
        (inth (filt (lfuse xxs) (addzero(lsum (xxs) 0))) j) = 
         (inth (lfuse xxs) (inth  (addzero(lsum (xxs) 0)) j))         
         " 
     by (simp add: filt_ilen filt_inth)
 have 3: "j \<le> ilen (addzero (lsum xxs 0)) \<and> j>0 \<longrightarrow> 
           inth  (addzero(lsum (xxs) 0)) j = inth (lsum xxs 0) (j-1)
         "    
    by (metis One_nat_def Suc_pred inth_Suc leD lsum_addzero_ilen lsum_addzero_inth neq0_conv)
 have 4: "j \<le> ilen (addzero (lsum xxs 0)) \<and> j>0 \<longrightarrow> 
          (inth (lfuse xxs) (inth (lsum xxs 0) (j-1))) = ilast(inth xxs (j -1)) " 
   by (metis assms diff_is_0_eq' filt_ilen filt_lfuse_lsum_2 filt_inth  
       le0 le_diff_conv lsum_addzero_ilen lsum_ilen neq0_conv)
 show ?thesis 
 using "1" "2" "3" "4" by (simp add: assms(3))
qed

   

lemma filt_lfuse_lsum_4:
 assumes "lastfirst (xxs)"
          "(\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)"
 shows   " (inth (addzero (lsum xxs 0)) (ilen (addzero (lsum xxs 0)))) \<le> ilen (lfuse xxs)"
using assms
by (metis add_cancel_right_left eq_iff  lfuse_ilen
    lsum_addzero_ilast lsum_ilen lsum_inth) 

lemma filt_lfuse_lsum_5:
 assumes "lastfirst (xxs)"
         " (\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)"
         " i\<le>ilen (addzero (lsum xxs 0)) "
 shows   " (inth (addzero (lsum xxs 0)) i) \<le> ilen (lfuse xxs)"
using assms filt_lfuse_lsum_4[of "xxs"] lsum_addzero_idx[of "xxs"]
proof -
have f1: " inth (addzero (lsum xxs 0)) (ilen (addzero (lsum xxs 0))) \<le> ilen (lfuse xxs)"
using \<open>\<forall>j\<le>ilen xxs. 0 < ilen ( inth xxs j)\<close> 
     \<open>\<lbrakk>lastfirst xxs; \<forall>j\<le>ilen xxs. 0 < ilen ( inth xxs j)\<rbrakk> \<Longrightarrow> 
       inth (addzero (lsum xxs 0)) (ilen (addzero (lsum xxs 0))) \<le> ilen (lfuse xxs)\<close> 
      \<open>lastfirst xxs\<close> by blast (* 0.4 ms *)
have "\<not> i < ilen (addzero (lsum xxs 0)) \<or> 
       inth (addzero (lsum xxs 0)) i < inth (addzero (lsum xxs 0)) (ilen (addzero (lsum xxs 0)))"
using \<open>\<forall>j\<le>ilen xxs. 0 < ilen ( inth xxs j) \<Longrightarrow> index_sequence 0 (addzero (lsum xxs 0))\<close> 
      \<open>\<forall>j\<le>ilen xxs. 0 < ilen ( inth xxs j)\<close> idx_less_last_1 by blast (* 1 ms *)
then show ?thesis
using f1 \<open>i \<le> ilen (addzero (lsum xxs 0))\<close> by force (* 6 ms *)
qed

lemma lfuse_ilen_b:
 assumes "lastfirst xxs"
  shows  "( \<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
                  (\<forall>j\<le>ilen (inth (xxs) (i)).
                     inth ((lsum xxs 0) ) (i-1) + j \<le> ilen (lfuse xxs)))"
proof -
 have 0: "(\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow>  
            (i-1) \<le> ilen xxs 
          ) "
     by linarith
 have 1: "(\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
                   inth ((lsum xxs 0) ) (i-1) = 
                    (\<Sum>k::nat= 0..(i-1). ilen(inth xxs k))
            ) "
         by (metis "0" add.left_neutral lsum_inth)
 have 2: "ilen (lfuse xxs) = (\<Sum>k::nat= 0..(ilen xxs). ilen(inth xxs k))" 
     using assms lfuse_ilen by blast
 have 3: "(\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
            (\<forall>j\<le>ilen (inth (xxs) (i)).
              (\<Sum>k::nat= 0..(i-1). ilen(inth xxs k)) + j \<le> 
               (\<Sum>k::nat= 0..(i). ilen(inth xxs k)) 
            ))
         " 
     by (metis (no_types, lifting) add_le_cancel_left le_add_diff_inverse plus_1_eq_Suc 
      sum.atLeast0_atMost_Suc)
 have 4: "(\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
            (\<Sum>k::nat= 0..(i). ilen(inth xxs k)) \<le> 
             (\<Sum>k::nat= 0..(ilen xxs). ilen(inth xxs k))
          )
        " 
        using sum.ub_add_nat[of _ _ "\<lambda> k. ilen(inth xxs k)" ] 
        by (metis Suc_eq_plus1 le_add1 le_add_diff_inverse le_simps(1) zero_less_Suc)    
 show ?thesis 
 using "1" "2" "3" "4" by fastforce
qed


lemma lsum_shift:
 assumes "lastfirst xxs"
         " (\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)"
         " i\<le>ilen xxs "  
 shows   "  inth (lsum xxs a) i = a+ inth (lsum xxs 0) i"
using assms by (simp add: lsum_inth)

lemma lsum_lfuse_inth_lsum_inth:
assumes "lastfirst xxs"
        " (\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)"
shows    " (\<forall> i\<le>ilen xxs. 
                 (\<forall> j\<le> ilen(inth xxs i). 
                  ( (inth (lfuse xxs) ((inth (addzero (lsum xxs 0)) i) +j) )) =
                  ( (inth (inth xxs i) j))  ))"
using assms
proof 
 (induct xxs)
 case (INil x)
 then show ?case 
 by (metis  inth.simps(1) add_cancel_right_left  ilen.simps(1) 
     le_zero_eq lfuse_INil lsum_addzero_ifirst)
 next
 case (ICons x1a xxs)
 then show ?case 
  proof -
   have 1: "(\<forall>i\<le>ilen (x1a \<odot> xxs).
             (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) i).
               inth (lfuse (x1a \<odot> xxs)) (inth (addzero (lsum (x1a \<odot> xxs) 0)) i + j) = 
               inth (inth (x1a \<odot> xxs) i) j)) 
           = 
           ( (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) 0).
                  inth (lfuse (x1a \<odot> xxs)) (inth (addzero (lsum (x1a \<odot> xxs) 0)) 0 + j) = 
                  inth (inth (x1a \<odot> xxs) 0) j) \<and> 
             (\<forall>i. 1 \<le> i \<and> i\<le>ilen (x1a \<odot> xxs) \<longrightarrow> 
                  (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) i).
                    inth (lfuse (x1a \<odot> xxs)) (inth (addzero (lsum (x1a \<odot> xxs) 0)) i + j) = 
                    inth (inth (x1a \<odot> xxs) i) j)) )          
         "
    by (metis One_nat_def Suc_leI ilen_gr_zero not_gr_zero)
   have 2: "(\<forall>j\<le>ilen (inth (x1a \<odot> xxs) 0).
                  inth (lfuse (x1a \<odot> xxs)) (inth (addzero (lsum (x1a \<odot> xxs) 0)) 0 + j) = 
                  inth (inth (x1a \<odot> xxs) 0) j) 
            =
            (\<forall>j\<le>ilen (x1a).
                  inth (lfuse (x1a \<odot> xxs)) ( j) = inth (x1a) j)
           "
       by (metis add_cancel_right_left inth_zero  
           lsum_addzero_ifirst)
   have 3: "ilast x1a = ifirst(ifirst xxs)"
      using ICons.prems lastfirst.simps(2) by blast
   have 4: "(\<forall>j\<le>ilen (x1a).
                  inth (lfuse (x1a \<odot> xxs)) ( j) = inth (x1a) j) " 
     by (metis ICons.prems(1) fuse_ilen_a fuse_inth 
         ifirst_lfuse_ifirst lastfirst.simps(2) le_add1 le_trans  lfuse_ICons)
   have 41: "(\<forall>j\<le>ilen (inth (x1a \<odot> xxs) 0).
                  inth (lfuse (x1a \<odot> xxs)) (inth (addzero (lsum (x1a \<odot> xxs) 0)) 0 + j) = 
                  inth (inth (x1a \<odot> xxs) 0) j)" 
     using "2" "4" by blast
   have 5: "(\<forall>i. 1 \<le> i \<and> i\<le>ilen (x1a \<odot> xxs) \<longrightarrow> 
             (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) i).
                    inth (lfuse (x1a \<odot> xxs)) (inth (addzero (lsum (x1a \<odot> xxs) 0)) i + j) = 
                    inth (inth (x1a \<odot> xxs) i) j)) =
            (\<forall>i. 0 \<le> i-1 \<and> i-1 \<le>ilen (xxs) \<longrightarrow> 
             (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) i).
                    inth (lfuse (x1a \<odot> xxs)) (inth (addzero (lsum (x1a \<odot> xxs) 0)) i + j) = 
                    inth (inth (x1a \<odot> xxs) i) j))" 
         using "1" "2" "4" le_diff_conv by auto
   have 6: "(\<forall>i. 0 \<le> i-1 \<and> i-1 \<le>ilen (xxs) \<longrightarrow> 
              (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) i).
                    inth (lfuse (x1a \<odot> xxs)) (inth (addzero (lsum (x1a \<odot> xxs) 0)) i + j) = 
                    inth (inth (x1a \<odot> xxs) i) j)) =
            (\<forall>i \<le>ilen (xxs).
              (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) (Suc i)).
                    inth (lfuse (x1a \<odot> xxs)) (inth (addzero (lsum (x1a \<odot> xxs) 0)) (Suc i) + j) = 
                    inth (inth (x1a \<odot> xxs) (Suc i)) j))" (is "?L = ?R")
            proof 
             show "?L \<Longrightarrow> ?R" 
              using "2" "4" by simp
             show "?R \<Longrightarrow> ?L" 
              using "2" "4" by (metis One_nat_def Suc_pred gr0I)
            qed
   have 7: "(\<forall>i \<le>ilen (xxs). 
             (\<forall>j\<le>ilen (inth (x1a \<odot> xxs) (Suc i)).
                    inth (lfuse (x1a \<odot> xxs)) (inth (addzero (lsum (x1a \<odot> xxs) 0)) (Suc i) + j) =
                    inth (inth (x1a \<odot> xxs) (Suc i)) j)) =
            (\<forall>i \<le>ilen (xxs). 
             (\<forall>j\<le>ilen (inth (xxs) (i)).
                    inth (lfuse (x1a \<odot> xxs)) (inth (addzero (lsum (x1a \<odot> xxs) 0)) (Suc i) + j) =
                    inth (inth (xxs) (i)) j))" 
            by simp
   have 8: "(\<forall>i \<le>ilen (xxs).
             (\<forall>j\<le>ilen (inth (xxs) (i)).
                    inth (lfuse (x1a \<odot> xxs)) (inth (addzero (lsum (x1a \<odot> xxs) 0)) (Suc i) + j) = 
                    inth (inth (xxs) (i)) j)) =
            (\<forall>i \<le>ilen (xxs).
             (\<forall>j\<le>ilen (inth (xxs) (i)).
                    inth (fuse x1a (lfuse xxs)) (inth ( (lsum (x1a \<odot> xxs) 0)) (i) + j) = 
                    inth (inth (xxs) (i)) j))"
       by (metis inth_Suc lfuse_ICons lsum_addzero_ICons)
   have 9: "(\<forall>i \<le>ilen (xxs).
             (\<forall>j\<le>ilen (inth (xxs) (i)).
                    inth (fuse x1a (lfuse xxs)) (inth ( (lsum (x1a \<odot> xxs) 0)) (i) + j) = 
                    inth (inth (xxs) (i)) j))
              =
            (\<forall>i \<le>ilen (xxs). 
             (\<forall>j\<le>ilen (inth (xxs) (i)).
                    inth (fuse x1a (lfuse xxs)) (inth ((ilen(x1a)\<odot>(lsum xxs) (ilen x1a)) ) i +j) =
                    inth (inth (xxs) (i)) j))
            " 
    by simp
  have 11: "(\<forall>i \<le>ilen (xxs). 
             (\<forall>j\<le>ilen (inth (xxs) (i)).
               inth (fuse x1a (lfuse xxs)) (inth ((ilen(x1a)\<odot> (lsum xxs) (ilen x1a)) ) (i) + j) = 
               inth (inth (xxs) (i)) j)) 
              =
            ((\<forall>i. i=0 \<longrightarrow> 
              (\<forall>j\<le>ilen (inth (xxs) (i)).
                inth (fuse x1a (lfuse xxs)) (inth ((ilen(x1a)\<odot> (lsum xxs) (ilen x1a)) ) (i) + j) = 
                inth (inth (xxs) (i)) j))
               \<and> 
             (\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
              (\<forall>j\<le>ilen (inth (xxs) (i)).
                inth (fuse x1a (lfuse xxs)) (inth ((ilen(x1a)\<odot> (lsum xxs) (ilen x1a)) ) (i) + j) = 
                inth (inth (xxs) (i)) j)) ) 
           "
           by (metis One_nat_def Suc_leI gr_zeroI ilen_gr_zero)
  have 12: "(\<forall>i. i=0 \<longrightarrow> 
             (\<forall>j\<le>ilen (inth (xxs) (i)).
               inth (fuse x1a (lfuse xxs)) (inth ((ilen(x1a)\<odot> (lsum xxs) (ilen x1a)) ) (i) + j) = 
               inth (inth (xxs) (i)) j)) =
            (\<forall>j\<le>ilen (inth (xxs) (0)).
               inth (fuse x1a (lfuse xxs)) (inth ((ilen(x1a)\<odot> (lsum xxs) (ilen x1a)) ) (0) + j) = 
               inth (inth (xxs) (0)) j)
          "
   by blast
  have 13: "(\<forall>j\<le>ilen (inth (xxs) (0)).
              inth (fuse x1a (lfuse xxs)) (inth ((ilen(x1a)\<odot> (lsum xxs) (ilen x1a)) ) (0) + j) = 
              inth (inth (xxs) (0)) j) =
            (\<forall>j\<le>ilen (inth (xxs) (0)).
              inth (fuse x1a (lfuse xxs)) (ilen(x1a) + j) = 
              inth (inth (xxs) (0)) j)
           " 
    by auto
  have 14: "(\<forall>j\<le>ilen (inth (xxs) (0)).
              inth (fuse x1a (lfuse xxs)) (ilen(x1a) + j) = 
              inth (inth (xxs) (0)) j) 
            = 
            (\<forall>j\<le>ilen (inth (xxs) (0)).
              inth ( (lfuse xxs)) (j) = 
              inth (inth (xxs) (0)) j)
           " 
    using ICons.prems fuse_inth_a ifirst_lfuse_ifirst lfuse_ilen_a 
    by (metis lastfirst.simps(2) le0)
  have 15: "(\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
             (\<forall>j\<le>ilen (inth (xxs) (i)).
               inth (fuse x1a (lfuse xxs)) (inth ((ilen(x1a)\<odot> (lsum xxs) (ilen x1a)) ) (i) + j) = 
               inth (inth (xxs) (i)) j)) =
            (\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
             (\<forall>j\<le>ilen (inth (xxs) (i)).
               inth (fuse x1a (lfuse xxs)) (inth ( (lsum xxs (ilen x1a)) ) (i-1) + j) = 
               inth (inth (xxs) (i)) j))
          " 
    by (metis inth_Suc ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc)
  have 16: "(\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow>
               inth ( (lsum xxs (ilen x1a)) ) (i-1) = 
               ilen x1a + inth (lsum xxs 0)  (i-1))
            " 
       by (simp add: le_diff_conv lsum_inth)
  have 17: "(\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
              (\<forall>j\<le>ilen (inth (xxs) (i)).
                inth (fuse x1a (lfuse xxs)) (inth ( (lsum xxs (ilen x1a)) ) (i-1) + j) = 
                inth (inth (xxs) (i)) j))
            =
            (\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
             (\<forall>j\<le>ilen (inth (xxs) (i)).
               inth (fuse x1a (lfuse xxs)) (ilen x1a + inth ((lsum xxs 0) ) (i-1) + j) =
               inth (inth (xxs) (i)) j))"
      using "16" by auto
  have 18: "(\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
             (\<forall>j\<le>ilen (inth (xxs) (i)).
               inth ((lsum xxs 0) ) (i-1) + j \<le> ilen (lfuse xxs)))"
     using ICons.prems lastfirst.simps(2) lfuse_ilen_b by blast
  have 19: "(\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
             (\<forall>j\<le>ilen (inth (xxs) (i)). 
                inth (fuse x1a (lfuse xxs)) (ilen x1a + inth ((lsum xxs 0) ) (i-1) +j) =
                inth ( (lfuse xxs)) ( inth ((lsum xxs 0) ) (i-1) + j))) 
              " 
    by (metis ICons.prems(1) ab_semigroup_add_class.add_ac(1) fuse_inth_a 
        ifirst_lfuse_ifirst lastfirst.simps(2) lfuse_ilen_b)
  have 20: "(\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
             (\<forall>j\<le>ilen (inth (xxs) (i)).
               inth (fuse x1a (lfuse xxs)) (ilen x1a + inth ((lsum xxs 0) ) (i-1) + j) = 
               inth (inth (xxs) (i)) j)) =
            (\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
             (\<forall>j\<le>ilen (inth (xxs) (i)).
               inth ( (lfuse xxs)) ( inth ((lsum xxs 0) ) (i-1) + j) = 
               inth (inth (xxs) (i)) j))
              "  
     using 19 by auto
  have 201: "((\<forall>i. i=0 \<longrightarrow> 
              (\<forall>j\<le>ilen (inth (xxs) (i)).
                inth (fuse x1a (lfuse xxs)) (inth ((ilen(x1a)\<odot> (lsum xxs) (ilen x1a)) ) (i) + j) = 
                inth (inth (xxs) (i)) j))
               \<and> 
             (\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
              (\<forall>j\<le>ilen (inth (xxs) (i)).
                inth (fuse x1a (lfuse xxs)) (inth ((ilen(x1a)\<odot> (lsum xxs) (ilen x1a)) ) (i) + j) = 
                inth (inth (xxs) (i)) j)) ) =
              ((\<forall>j\<le>ilen (inth (xxs) (0)).
                  inth ( (lfuse xxs)) (j) = 
                  inth (inth (xxs) (0)) j) \<and>
               (\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
                (\<forall>j\<le>ilen (inth (xxs) (i)).
                  inth ( (lfuse xxs)) ( inth ((lsum xxs 0) ) (i-1) + j) = 
                  inth (inth (xxs) (i)) j)) ) "
        using "14" "15" "17" "20" by auto
  have 21: "lastfirst xxs "
    using ICons.prems lastfirst.simps(2) by blast
  have 22: " (\<forall> j \<le> ilen xxs. ilen(inth xxs j) > 0)"
    using ICons.prems by auto
  have 23: "(\<forall>i\<le>ilen xxs.
             (\<forall>j\<le>ilen (inth xxs i).
               inth (lfuse xxs) (inth (addzero (lsum xxs 0)) i + j) = 
               inth (inth xxs i) j)) "       
     using "21" "22" ICons.hyps by blast
  have 24: "((\<forall>j\<le>ilen (inth xxs 0).
               inth (lfuse xxs) (inth (addzero (lsum xxs 0)) 0 + j) = 
               inth ( inth xxs 0) j) \<and>
             (\<forall>i. 1 \<le> i \<and> i \<le>ilen xxs \<longrightarrow>
              (\<forall>j\<le>ilen (inth xxs i).
                inth (lfuse xxs) (inth (addzero (lsum xxs 0)) i + j) = 
                inth (inth xxs i) j))  )"      
        using "23" by blast
   have 25: "(\<forall>j\<le>ilen (inth (xxs) (0)).
                  inth ( (lfuse xxs)) (j) = 
                  inth (inth (xxs) (0)) j)" 
     by (metis "24" add_cancel_right_left lsum_addzero_ifirst)
   have 26: "(\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
                  inth (addzero (lsum xxs 0)) i = inth ((lsum xxs 0) ) (i-1)
             )"
       by (metis addzero_def inth_Suc less_eq_Suc_le less_le_trans lsum_ilen 
           not_less_eq_eq ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc 
           zero_less_one) 
   have 27: "(\<forall>i. 1 \<le> i \<and> i \<le>ilen (xxs) \<longrightarrow> 
                (\<forall>j\<le>ilen (inth (xxs) (i)).
                  inth ( (lfuse xxs)) ( inth ((lsum xxs 0) ) (i-1) + j) = 
                  inth (inth (xxs) (i)) j)) =
             (\<forall>i. 1 \<le> i \<and> i \<le>ilen xxs \<longrightarrow>
              (\<forall>j\<le>ilen (inth xxs i).
                inth (lfuse xxs) (inth (addzero (lsum xxs 0)) i + j) = 
                inth (inth xxs i) j)) " 
     by (simp add: "26")
  show ?thesis 
  using "11" "201" "24" "25" "27" "6" "8" by auto
qed    

qed




lemma lcpl_lsum_less_th_equal:
 assumes "index_sequence 0 l"
         " (inth l (ilen l)) = ilen \<sigma>"
         " (\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " ilen \<sigma> > 0 "
         " i < ilen (addzero (lsum (lcpl f g \<sigma> l) 0)) " 
 shows   " (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i)) \<le>
           ilen( (lfuse (lcpl f g \<sigma> l)))"
using assms 
by (metis (no_types, lifting) Suc_leI filt_lfuse_lsum_5 index_sequence_def 
    lcpl_ilen_inth_gr_zero lcpl_lfuse_lastfirst neq0_conv)
 



lemma lcpl_lsum_ilen:
 assumes "index_sequence 0 l"
         " (inth l (ilen l)) = ilen \<sigma>"
         " (\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " ilen \<sigma> >0 " 
 shows   "ilen (addzero (lsum ((lcpl f g \<sigma> l)) 0)) = ilen l"
proof -
 have 1: "ilen \<sigma> >0 \<longrightarrow> ilen (lcpl f g \<sigma> l) = ilen l -1"
   using assms index_sequence_def lcpl_ilen by fastforce
 have 2: "ilen \<sigma>>0 \<longrightarrow> ilen l >0"
   using assms gr_zeroI index_sequence_def by fastforce
 have 3: "ilen \<sigma> > 0 \<longrightarrow> ilen (ifirst (lcpl f g \<sigma> l)) > 0"  
    by (metis assms ilen_gr_zero lcpl_ilen_inth_gr_zero)
 have 4: "ilen \<sigma> >0 \<longrightarrow> ilen (lcpl f g \<sigma> l) = 0 \<longrightarrow> 
           ilen (addzero (lsum ((lcpl f g \<sigma> l)) 0)) = ilen l " 
     using "1" "2" "3" lsum_addzero_ilen by fastforce
 have 5: "ilen \<sigma> >0 \<longrightarrow> ilen (lcpl f g \<sigma> l) > 0 \<longrightarrow> 
        ilen (addzero (lsum ((lcpl f g \<sigma> l)) 0)) = ilen l" 
    by (simp add: "1" lsum_addzero_ilen)
  show ?thesis using "4" "5" using assms(4) by blast
qed

lemma lcpl_lsum_inth:
 assumes "index_sequence 0 l"
         " (inth l (ilen l)) = ilen \<sigma> "
         "(\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " ilen \<sigma>>0 " 
         " j \<le> ilen (addzero (lsum (lcpl f g \<sigma> l) 0)) "
 shows  " ( j=0 \<longrightarrow> 
          (inth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) j) =
           ifirst(ifirst (lcpl f g \<sigma> l)) )
          \<and>
          (j>0 \<longrightarrow> (inth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) j) =
            ilast(inth (lcpl f g \<sigma> l) (j-1)))
         "
proof -
 have 0: "ilen \<sigma> >0 \<longrightarrow> ilen l >0" 
   using assms gr_zeroI index_sequence_def by fastforce
 have 1: "ilen \<sigma> >0 \<longrightarrow> lastfirst (lcpl f g \<sigma> l)" 
    using "0" assms index_sequence_def lcpl_lfuse_lastfirst by blast
 have 2: "ilen \<sigma>>0 \<longrightarrow> 
         (\<forall> j \<le> ilen (lcpl f g \<sigma> l). ilen(inth (lcpl f g \<sigma> l) j) > 0) "
   by (simp add: assms lcpl_ilen_inth_gr_zero)
 have 3: " ilen \<sigma>>0 \<longrightarrow> 
           ilen (addzero (lsum (lcpl f g \<sigma> l) 0)) =
           ilen (lcpl f g \<sigma> l) +1
          " 
          by (metis "2" One_nat_def Suc_eq_plus1 ilen_gr_zero  
         le_imp_less_or_eq lsum_addzero_ilen)
 have 4: " ilen \<sigma> > 0 \<longrightarrow>
           (inth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) 0) =
             ifirst(ifirst (lcpl f g \<sigma> l)) "
    by (simp add: "1" "2" filt_lfuse_lsum_3) 
 have 5: "ilen \<sigma> > 0 \<longrightarrow> 
          j \<le> ilen (addzero (lsum (lcpl f g \<sigma> l) 0)) \<and> j>0 \<longrightarrow>
          (inth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) (j))
           = ilast(inth (lcpl f g \<sigma> l) (j-1))"
   by (simp add: "1" "2" filt_lfuse_lsum_3)
 from 4 5 show ?thesis using assms(4) assms(5) by blast
qed   

lemma lcpl_lsum_inth_a:
 assumes "index_sequence 0 l"
         " (inth l (ilen l)) = ilen \<sigma> " 
         " (\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " ilen \<sigma>>0 "
         " j \<le> ilen l " 
 shows  " (inth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) j) =
          (inth l j)"
proof -
 have 1: "ilen \<sigma>>0 \<longrightarrow> ilen l>0 "
  using assms gr_zeroI index_sequence_def by fastforce 
 have 2: "ilen \<sigma>>0 \<longrightarrow> ilen (addzero (lsum (lcpl f g \<sigma> l) 0)) = ilen l"
    by (simp add: assms lcpl_lsum_ilen) 
 have 3: "ilen \<sigma>>0 \<longrightarrow> 
          j \<le> ilen l \<longrightarrow>
            ( j=0 \<longrightarrow> 
          (inth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) j) =
           ifirst(ifirst (lcpl f g \<sigma> l)) )
          \<and>
          (j>0 \<longrightarrow> (inth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) j) = 
           ilast(inth (lcpl f g \<sigma> l) (j-1)))
          " 
     by (metis "2" assms lcpl_lsum_inth)
 have 4: "ilen \<sigma>>0 \<longrightarrow> 
          j \<le> ilen l \<and> j=0 \<longrightarrow> ifirst(ifirst (lcpl f g \<sigma> l)) = (inth l j) " 
     using assms lcpl_ifirst[of "l" "\<sigma>" "f" "g"]  
     by (metis "1" index_sequence_def)
 have 5: "ilen \<sigma>>0 \<longrightarrow> 
          j \<le> ilen l \<and> j>0 \<longrightarrow> j-1 \<le> ilen (lcpl f g \<sigma> l)"
      using assms by (metis "1" diff_le_mono index_sequence_def lcpl_ilen)
 have 6: "ilen \<sigma>>0 \<longrightarrow> 
          j \<le> ilen l \<and> j>0 \<longrightarrow> ilast(inth (lcpl f g \<sigma> l) (j-1)) = (inth l j)"
    using lcpl_ilast_inth 
     by (metis "5" One_nat_def Suc_pred assms)
 show ?thesis 
 using "3" "4" "6" using assms(4) assms(5) by auto
qed

lemma lcpl_filt_lfuse_lsum:
 assumes "index_sequence 0 l"
         " (inth l (ilen l)) = ilen \<sigma>"
         " (\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " ilen \<sigma>>0 " 
 shows  " (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) = l"
using assms 
proof -
 have 0: "ilen \<sigma>>0 \<longrightarrow>   ilen l > 0" 
   using assms gr_zeroI index_sequence_def by fastforce 
 have 1: " ilen \<sigma>>0 \<longrightarrow>  
          ilen (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) = ilen l"
   by (simp add: assms filt_ilen lcpl_lsum_ilen)
 have 2: "ilen \<sigma>>0 \<longrightarrow>  
           (\<forall> j. j\<le>ilen l \<longrightarrow>
          (inth (filt (lfuse ((lcpl f g \<sigma> l))) (addzero(lsum ((lcpl f g \<sigma> l)) 0))) j) =
          (inth l j))  "
     by (simp add: assms lcpl_lsum_inth_a)
 from 1 2 show ?thesis by (simp add: assms(4) interval_eq_inth_eq) 
qed


lemma lcpl_lfuse_filt_power:
 assumes "index_sequence 0 l"
         " (inth l (ilen l)) = ilen \<sigma> "  
         " (\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         "  ilen \<sigma> >0 " 
 shows   "  powerinterval g (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) (addzero (lsum (lcpl f g \<sigma> l) 0))"
proof -
 have 0: "ilen \<sigma> > 0 \<longrightarrow> ilen l >0 "
 using assms gr_zeroI index_sequence_def by fastforce 
 have 01: "(\<forall> i<ilen l. 
            g (filt (sub (inth l i) (inth l (Suc i)) \<sigma>) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))))"
   using assms cpl_projection by auto
 have 02: "ilen \<sigma> > 0 \<longrightarrow> (\<forall> i<ilen l. g (filt \<sigma> (inth (lcpl f g \<sigma> l) i)) )"
   using "0" assms index_sequence_def lcpl_lfuse_filt_power_help by blast
 have 03 : "powerinterval g (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) (addzero (lsum (lcpl f g \<sigma> l) 0)) =
           ( \<forall>i<ilen (addzero (lsum (lcpl f g \<sigma> l) 0)).
             g (sub (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) 
                    (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                    (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))))" 
        by (simp add: powerinterval_def)
  have 04: "ilen \<sigma> >0 \<longrightarrow>  ilen (addzero (lsum (lcpl f g \<sigma> l) 0)) = ilen l"
    by (simp add: assms lcpl_lsum_ilen) 
  have 05: "ilen \<sigma> >0 \<longrightarrow>
           (\<forall> i<ilen l. 
              (filt \<sigma> (inth (lcpl f g \<sigma> l) i))  =
              (sub (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) 
                   (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                   (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))) )
      " 
   proof -
    have 06: "ilen \<sigma> >0 \<longrightarrow> 
              (\<forall> i<ilen l. ilen (filt \<sigma> (inth (lcpl f g \<sigma> l) i)) =
                    ilen (inth (lcpl f g \<sigma> l) i))"
      using filt_ilen by blast
    have 07: "ilen \<sigma> >0 \<longrightarrow> 
              (\<forall> i<ilen l. ilen (inth (lcpl f g \<sigma> l) i) =
                ilen (imap (shift (inth l i)) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)))) 
              " 
        using assms by (metis index_sequence_def lcpl_inth)
    have 08: "ilen \<sigma> >0 \<longrightarrow> 
              (\<forall> i<ilen l. 
                ilen (imap (shift (inth l i)) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)))  =
                ilen (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)))
              " 
         using ilen_imap by blast
    have 09: "ilen \<sigma> >0 \<longrightarrow> 
              (\<forall>i<ilen l. 
               ilen (sub (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i)
                           (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                           (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))) =
               (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i)) -  
               (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) )
             " 
             by (metis "04" assms filt_ilen ilen_sub lcpl_lsum_less_th_equal 
                  lcpl_ilen_inth_gr_zero le_eq_less_or_eq lsum_addzero_inth_leq_Suc)
    have 10: "ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l. (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i)) =
                 (inth (0\<odot>(lsum (lcpl f g \<sigma> l) 0)) (Suc i)))  
              " 
        using "04" addzero_def by auto
    have 11: "ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l. (inth (0\<odot>(lsum (lcpl f g \<sigma> l) 0)) (Suc i)) =
                 (\<Sum>k::nat= 0..(i). ilen(inth (lcpl f g \<sigma> l) k)) )
              "   using "04" 
        proof simp_all
          assume "0 < ilen \<sigma> \<longrightarrow> ilen (addzero (lsum (lcpl f g \<sigma> l) 0)) = ilen l"
          show   " 0 < ilen \<sigma> \<longrightarrow>
                  (\<forall>i<ilen l. inth (lsum (lcpl f g \<sigma> l) 0) i = 
                   (\<Sum>k = 0..i. ilen (inth (lcpl f g \<sigma> l) k))) "
           proof 
            assume " 0 < ilen \<sigma>"
            show " \<forall>i<ilen l. inth (lsum (lcpl f g \<sigma> l) 0) i = 
                   (\<Sum>k = 0..i. ilen (inth (lcpl f g \<sigma> l) k)) " 
            proof 
             fix i
             show "i < ilen l \<longrightarrow>
                   inth (lsum (lcpl f g \<sigma> l) 0) i = (\<Sum>k = 0..i. ilen (inth (lcpl f g \<sigma> l) k)) " 
             by (metis (no_types, lifting) "0" One_nat_def Suc_leI Suc_le_mono Suc_pred 
                 add.left_neutral assms(1) assms(4) index_sequence_def lcpl_ilen lsum_inth sum.cong)
            qed
           qed
        qed              
    have 12: "ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l. (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) =
                 (inth (0\<odot>(lsum (lcpl f g \<sigma> l) 0)) i) )
              "      
       using "04" addzero_def by auto
    have 13: "ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l. (inth (0\<odot>(lsum (lcpl f g \<sigma> l) 0)) i) =
               (case i of 0 \<Rightarrow> 0
                 | Suc j \<Rightarrow> (\<Sum>k::nat= 0..(j). ilen(inth (lcpl f g \<sigma> l) k)) ))    
              " using "11" by (simp_all add: Nitpick.case_nat_unfold)
    have 14: "ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l. 
              (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i)) -  
              (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) =
              (\<Sum>k::nat= 0..(i). ilen(inth (lcpl f g \<sigma> l) k)) -
              (case i of 0 \<Rightarrow> 0
                 | Suc j \<Rightarrow> (\<Sum>k::nat= 0..(j). ilen(inth (lcpl f g \<sigma> l) k)) ) )
             " 
      using "10" "11" "12" "13" by auto
    have 15: "ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l.  
                (\<Sum>k::nat= 0..(i). ilen(inth (lcpl f g \<sigma> l) k)) -
              (case i of 0 \<Rightarrow> 0
                 | Suc j \<Rightarrow> (\<Sum>k::nat= 0..(j). ilen(inth (lcpl f g \<sigma> l) k)) ) =
                (case i of 0 \<Rightarrow> ilen(inth (lcpl f g \<sigma> l) 0)
                  | Suc j \<Rightarrow> (\<Sum>k::nat= 0..(Suc j). ilen(inth (lcpl f g \<sigma> l) k)) -
                             (\<Sum>k::nat= 0..(j). ilen(inth (lcpl f g \<sigma> l) k)) ))
             " 
        by (simp add: Nitpick.case_nat_unfold)
     have 16: "ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l.  
                 (case i of 0 \<Rightarrow> ilen(inth (lcpl f g \<sigma> l) 0)
                  | Suc j \<Rightarrow> (\<Sum>k::nat= 0..(Suc j). ilen(inth (lcpl f g \<sigma> l) k)) -
                             (\<Sum>k::nat= 0..(j). ilen(inth (lcpl f g \<sigma> l) k)) ) = 
                 (case i of 0 \<Rightarrow> ilen(inth (lcpl f g \<sigma> l) 0)
                  | Suc j \<Rightarrow> ilen(inth (lcpl f g \<sigma> l) (Suc j))  ) )
              "
      by (metis (no_types, lifting) Nitpick.case_nat_unfold add_diff_cancel_left' 
           sum.atLeast0_atMost_Suc)
    have 17: "ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l. 
                  (case i of 0 \<Rightarrow> ilen(inth (lcpl f g \<sigma> l) 0)
                  | Suc j \<Rightarrow> ilen(inth (lcpl f g \<sigma> l) (Suc j))  ) =
                  ilen(inth (lcpl f g \<sigma> l) i))
             " 
      by (simp add: Nitpick.case_nat_unfold)
    have 18: "ilen \<sigma> >0 \<longrightarrow> 
              (\<forall> i<ilen l. ilen (filt \<sigma> (inth (lcpl f g \<sigma> l) i)) =
                ilen (sub (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) 
                            (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                            (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))) )
             "   
       by (simp add: "09" "14" "15" "16" "17" filt_ilen)
    have 19: "ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l. 
                (\<forall> j\<le> ilen(inth (lcpl f g \<sigma> l) i).
                 (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i)) \<le> 
                  ilen (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))))
             "
             by (simp add: "04" assms filt_ilen lcpl_lsum_less_th_equal)
    have 22: " ilen \<sigma> >0 \<longrightarrow>  lastfirst (lcpl f g \<sigma> l)"
      using "0" assms index_sequence_def lcpl_lfuse_lastfirst by blast
   have 23: " ilen \<sigma> >0 \<longrightarrow>  (\<forall> j \<le> ilen (lcpl f g \<sigma> l). ilen(inth (lcpl f g \<sigma> l) j) > 0)"
     by (simp add: assms lcpl_ilen_inth_gr_zero)
    have 190: "ilen \<sigma> >0 \<longrightarrow> 
                (\<forall> i<ilen l. 
                   (\<forall> j\<le> ilen(inth (lcpl f g \<sigma> l) i).
                      (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) \<le> 
                      (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                   ))
               " 
       by (simp add: "04" "23" less_imp_le_nat lsum_addzero_inth_leq_Suc)
    have 20: "ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l. 
                 (\<forall> j\<le> ilen(inth (lcpl f g \<sigma> l) i). 
                   (inth (sub (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) 
                             (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                             (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))) j) =
                   (inth (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) 
                        ((inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) +j) ) ) ) 
              " 
    by (simp add: "14" "15" "16" "17" "19" "190")
   have 21: "ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l. 
                 (\<forall> j\<le> ilen(inth (lcpl f g \<sigma> l) i).
                  (inth (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) 
                       ((inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) +j) ) =
                  (inth \<sigma> (inth (lfuse (lcpl f g \<sigma> l)) 
                              ((inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) +j) )) ) )
            " 
       by (simp add: filt_imap inth_imap)
   have 24: "ilen \<sigma> >0 \<longrightarrow>
             (\<forall> i\<le>ilen (lcpl f g \<sigma> l). 
                 (\<forall> j\<le> ilen(inth (lcpl f g \<sigma> l) i). 
                  ( (inth (lfuse (lcpl f g \<sigma> l)) ((inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) +j) )) =
                  ( (inth (inth (lcpl f g \<sigma> l) i) j))  ))
            " 
     by (simp add: "22" "23" lsum_lfuse_inth_lsum_inth)
   have 241: "ilen \<sigma> >0 \<longrightarrow> ilen (lcpl f g \<sigma> l) = ilen l -1"
     using "0" assms index_sequence_def lcpl_ilen by blast 
   have 25: " ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l. 
                 (\<forall> j\<le> ilen(inth (lcpl f g \<sigma> l) i). 
                  (inth \<sigma> (inth (lfuse (lcpl f g \<sigma> l)) 
                              ((inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i) +j) )) =
                  (inth (filt \<sigma> (inth (lcpl f g \<sigma> l) i)) j) ))
            " 
            by (simp add: "06" "24" "241" assms filt_inth)
    have 26: "ilen \<sigma> >0 \<longrightarrow> 
               (\<forall> i<ilen l. 
                 (\<forall> j\<le> ilen(inth (lcpl f g \<sigma> l) i). 
                  (inth (filt \<sigma> (inth (lcpl f g \<sigma> l) i)) j) =
                  (inth (sub (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) i)
                            (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (Suc i))
                            (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))) j) ) )
             "
     by (simp add: "20" "21" "25")
   from 18 26 show ?thesis 
   by (simp add: filt_ilen interval_eq_inth_eq)
   qed
 show ?thesis 
 using "02" "03" "04" "05" by (simp add: assms(4))
qed

lemma lcpl_lfuse_filt_ilen:
 assumes "index_sequence 0 l"
         " (inth l (ilen l)) = ilen \<sigma>" 
         " (\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " ilen \<sigma> >0 " 
 shows   " (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (ilen(addzero (lsum (lcpl f g \<sigma> l) 0)))) =
          ilen (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) "
proof -
 have 0: "ilen \<sigma> > 0 \<longrightarrow> ilen l > 0"
    using assms gr_zeroI index_sequence_def by fastforce 
 have 1: "ilen \<sigma> >0 \<longrightarrow>
          ilen (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) = ilen ( (lfuse (lcpl f g \<sigma> l)))  " 
   using filt_ilen by blast
 have 2: "ilen \<sigma> >0 \<longrightarrow> 
           ilen ( (lfuse (lcpl f g \<sigma> l))) = 
            (\<Sum>k::nat= 0..(ilen (lcpl f g \<sigma> l)). ilen(inth (lcpl f g \<sigma> l) k))"
   using "0" assms index_sequence_def lfuse_ilen lcpl_lfuse_lastfirst by blast 
 have 3: "ilen \<sigma> >0 \<longrightarrow> 
      ilast( addzero (lsum (lcpl f g \<sigma> l) 0)) = ilast( (lsum (lcpl f g \<sigma> l) 0))  "
    using lsum_addzero_ilast by blast
 have 4: "ilen \<sigma> >0 \<longrightarrow> 
           ilast( (lsum (lcpl f g \<sigma> l) 0)) = 
          (\<Sum>k::nat= 0..(ilen (lcpl f g \<sigma> l)). ilen(inth (lcpl f g \<sigma> l) k))"
   by (metis add_cancel_right_left lsum_ilast)
 show ?thesis using "1" "2" "3" "4" by (simp add: assms(4))
qed




subsection \<open>Soundness of Projection Axioms\<close>

subsubsection \<open>PJ1\<close>

lemma PJ1sem:
 "(\<sigma> \<Turnstile> f \<triangle> ( g \<or> h) \<longrightarrow> f \<triangle> g \<or> f \<triangle> h)"
by (simp add: projection_d_def) blast

lemma PJ1sema:
 "(\<sigma> \<Turnstile> f \<triangle> ( g \<or> h) = (f \<triangle> g \<or> f \<triangle> h))"
by (simp add: projection_d_def) blast

subsubsection \<open>PJ2\<close>

lemma PJ2sem:
 "(\<sigma> \<Turnstile> f \<triangle> empty = empty)"
proof auto
  show "(\<sigma> \<Turnstile> f \<triangle> empty) \<Longrightarrow> \<sigma> \<Turnstile> empty " 
  unfolding projection_d_def empty_defs index_sequence_def  
  by (metis filt.simps(2)  ilen_ICons_1 neq0_conv) 
  show "\<sigma> \<Turnstile> empty \<Longrightarrow> (\<sigma> \<Turnstile> f \<triangle> empty)" 
  unfolding projection_d_def empty_defs index_sequence_def powerinterval_def 
   by (metis  inth.simps(1) filt.simps(1) index_sequence_def ilen.simps(1) 
       not_less_zero )
qed

subsubsection \<open>PJ3\<close>

lemma PJ3help:
 "sub 0 (ilen \<sigma>) \<sigma> = \<sigma>"
by (simp add: sub_zero_prefix)

lemma PJ3help1:
assumes "f \<sigma> \<and> 0 < ilen \<sigma>"
shows   " (\<exists>l. index_sequence 0 l \<and>
        inth l (ilen l) = ilen \<sigma> \<and>
        (\<forall>i<ilen l. f (sub (inth l i) (inth l (Suc i)) \<sigma>)) \<and>
        (\<exists>s1. ilen s1 = ilen l \<and>
      (\<forall>i\<le>ilen s1. inth s1 i = inth \<sigma> (inth l i)) \<and> ilen s1 = Suc 0))"
proof -
 have 1: "index_sequence 0 \<langle>0,ilen \<sigma>\<rangle> " 
   by (simp add: assms index_sequence_def)
 have 2: "inth \<langle>0,ilen \<sigma>\<rangle> (ilen \<langle>0,ilen \<sigma>\<rangle>) = ilen \<sigma> " 
   by auto
 have 3: "(\<forall>i<ilen \<langle>0,ilen \<sigma>\<rangle>. f (sub (inth \<langle>0,ilen \<sigma>\<rangle> i) (inth \<langle>0,ilen \<sigma>\<rangle> (Suc i)) \<sigma>)) "     
   by (simp add: PJ3help assms)
 have 4: "ilen \<langle>inth \<sigma> (0), inth \<sigma> (ilen \<sigma>)\<rangle> = ilen \<langle>0,ilen \<sigma>\<rangle>" 
    by simp
 have 5: "(\<forall>i\<le>ilen  \<langle>inth \<sigma> (0), inth \<sigma> (ilen \<sigma>)\<rangle>. 
      inth  \<langle>inth \<sigma> (0), inth \<sigma> (ilen \<sigma>)\<rangle> i = inth \<sigma> (inth \<langle>0,ilen \<sigma>\<rangle> i)) "
   using antisym_conv2 by fastforce 
 have 6: "ilen \<langle>inth \<sigma> (0), inth \<sigma> (ilen \<sigma>)\<rangle> = Suc 0 " 
   by simp
 show ?thesis 
 using "1" "2" "3" "4" "5" "6" by blast
qed


lemma PJ3sem:
"(\<sigma> \<Turnstile> f \<triangle> skip = (f \<and> more))"
proof -
 have 1: "(\<sigma> \<Turnstile> f \<triangle> skip) \<Longrightarrow> (\<sigma> \<Turnstile> f \<and> more)"
   by (metis (mono_tags, lifting) One_nat_def PJ3help cpl_projection filt_expand index_sequence_def 
       more_defs powerinterval_def skip_defs unl_lift2 zero_less_one)
 have 2: "(\<sigma> \<Turnstile> f \<and> more) \<Longrightarrow> (\<sigma> \<Turnstile> f \<triangle> skip)"
   by (simp add:  projection_d_def skip_defs more_defs powerinterval_def,
       metis PJ3help1  filt_ilen)
 show ?thesis 
 using "1" "2" unl_lift2 by blast
qed


subsubsection \<open>PJ4\<close>

lemma PJ4semchaina:
assumes "(\<sigma> \<Turnstile> f \<triangle> (g;h))"
 shows  "(\<sigma> \<Turnstile> (f \<triangle> g) ; (f \<triangle> h))"
proof -
 have 1: "(\<sigma> \<Turnstile> f \<triangle> (g;h))" 
  using assms by auto
 have 2: "(\<exists>l. index_sequence 0 l \<and>
         inth l (ilen l) = ilen \<sigma> \<and>
         powerinterval f \<sigma> l \<and>
         (\<exists>n\<le>ilen (filt \<sigma> l). g (prefix n (filt \<sigma> l)) \<and> h (suffix n (filt \<sigma> l)))) "
  by (metis assms chop_defs projection_d_def)
  obtain l where 3: "index_sequence 0 l \<and>
         inth l (ilen l) = ilen \<sigma> \<and>
         powerinterval f \<sigma> l \<and>
         (\<exists>n\<le>ilen (filt \<sigma> l). g (prefix n (filt \<sigma> l)) \<and> h (suffix n (filt \<sigma> l)))" 
    using "2" by auto
 have 4: "index_sequence 0 l "  
   using "3" by auto
 have 5: "powerinterval f \<sigma> l" 
   using "3" by auto
 have 6: "inth l (ilen l) = ilen \<sigma>"
   using "3" by auto
 have 7: "(\<exists>n\<le>ilen (filt \<sigma> l). g (prefix n (filt \<sigma> l)) \<and> h (suffix n (filt \<sigma> l)))" 
   using "3" by auto
  obtain n where 8: "n\<le>ilen (filt \<sigma> l) \<and> g (prefix n (filt \<sigma> l)) \<and> h (suffix n (filt \<sigma> l)) " 
  using "7" by auto
 have 9: "n\<le>ilen (filt \<sigma> l) " 
  using "8" by auto
 have 10: " g (prefix n (filt \<sigma> l))" 
  using "8" by auto
 have 11: "h (suffix n (filt \<sigma> l))" 
  using "8" by auto
 have 12: "index_sequence 0 (prefix n l)" 
   by (metis "4" "8" filt_ilen idx_split)
 have 13: "index_sequence (inth l n) (suffix n l)"
   by (metis "4" "9" filt_ilen idx_split) 
 have 14: "index_sequence 0 ((imap (shiftm (inth l n))  (suffix n l)))" 
  using "13" idx_shiftm by blast
 have 15: "g  (filt \<sigma> (prefix n l))"
   by (metis "8" filt_ilen filt_prefix) 
 have 16: "h (filt \<sigma> (suffix n l))"
   by (metis "11" "9" filt_ilen filt_suffix) 
 have 17: "g  (filt (prefix (inth l n) \<sigma>) (prefix n l))"
   by (metis (no_types, lifting) "12" "15" "4" "6" "8" filt_ilen filt_prefix_idx 
       idx_less_equal ilast_prefix order_refl)
 have 18: "h (filt (suffix (inth l n) \<sigma>) ((imap (shiftm (inth l n))  (suffix n l))))" 
    proof -
     have 181: "ilen((filt \<sigma> (suffix n l))) = 
                ilen(filt (suffix (inth l n) \<sigma>) ((imap (shiftm (inth l n))  (suffix n l)))) " 
        by (simp add: filt_ilen)
     have 182: "(\<forall>j\<le> ilen((filt \<sigma> (suffix n l))).
                 (inth (filt \<sigma> (suffix n l)) j) =
                 (inth \<sigma> ((inth l (n+j)))) 
                )" 
          by (metis "9" filt_ilen filt_imap inth_imap inth_suffix 
              suffix_ilen_good)
     have 183: "(\<forall>j\<le> ilen((filt \<sigma> (suffix n l))).
                  (inth (filt (suffix (inth l n) \<sigma>) ((imap (shiftm (inth l n))  (suffix n l)))) j) =
                  (inth (suffix (inth l n) \<sigma>) (inth (imap (shiftm (inth l n))  (suffix n l)) j))
                )  "
             by (simp add: filt_imap inth_imap) 
     have 184: "(inth l n) \<le> ilen \<sigma>"
          by (metis "4" "6" "9" filt_ilen idx_less_equal order_refl)
     have 185: "(\<forall>j\<le> ilen((filt \<sigma> (suffix n l))). 
                 (inth (imap (shiftm (inth l n))  (suffix n l)) j) =
                  (inth (suffix n l) j) - (inth l n)
                ) "
            by (metis "4" "6" "9" filt_ilen inth_suffix idx_shiftm_suffix_inth 
                suffix_ilen_good) 
     have 186: "(\<forall>j\<le> ilen((filt \<sigma> (suffix n l))). 
                   (inth (suffix n l) j) - (inth l n) = (inth l (j+n)) - (inth l n)
                )"
           by (metis "9" add.commute filt_ilen inth_suffix suffix_ilen_good) 
     have 187: "(\<forall>j\<le> ilen((filt \<sigma> (suffix n l))).
                   (inth l (j+n)) - (inth l n) \<le> ilen \<sigma> - (inth l n)
                )"
          by (metis "4" "6" "9" add.commute diff_le_mono eq_imp_le filt_ilen 
             idx_less_equal suffix_ilen_good nat_add_left_cancel_le 
             ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
     have 188: "(\<forall>j\<le> ilen((filt \<sigma> (suffix n l))).
                 (inth (suffix (inth l n) \<sigma>) (inth (imap (shiftm (inth l n))  (suffix n l)) j)) =
                 (inth \<sigma> ((inth (imap (shiftm (inth l n))  (suffix n l))  j) + (inth l n) ))
                ) "
           using inth_suffix 
           by (simp add: "184" "185" "186" "187" add.commute) 
     have 189: "(\<forall>j\<le> ilen((filt \<sigma> (suffix n l))). 
                 (inth \<sigma> ((inth (imap (shiftm (inth l n))  (suffix n l))  j) + (inth l n) )) =
                 (inth \<sigma> ((inth (suffix n l) j)) )
                ) "
         by (metis "13" "185" filt_ilen idx_greater 
             ordered_cancel_comm_monoid_diff_class.le_imp_diff_is_add)
     have 190: "(\<forall>j\<le> ilen((filt \<sigma> (suffix n l))).
                 (inth (filt \<sigma> (suffix n l)) j) = 
                 (inth (filt (suffix (inth l n) \<sigma>) ((imap (shiftm (inth l n))  (suffix n l)))) j)
                )"
           using "183" "188" "189" filt_expand by fastforce
     show ?thesis 
     by (metis "16" "181" "190" interval_eq_inth_eq)
    qed
 have 19: "powerinterval f (prefix (inth l n) \<sigma>) (prefix n l)"
   by (metis "4" "5" "6" "8" filt_ilen powerinterval_splita)
 have 20: "powerinterval f (suffix (inth l n) \<sigma>) ((imap (shiftm (inth l n))  (suffix n l)))" 
   by (metis "4" "5" "6" "9" filt_ilen powerinterval_split)
 have 21: "(inth l n) \<le> ilen \<sigma> "
    by (metis "3" "9" filt_ilen idx_less_equal order_refl)
 have 22: "inth (prefix n l) (ilen (prefix n l)) = (inth l n)"
   by (metis "8" filt_ilen ilast_prefix) 
 have 23: " inth ((imap (shiftm (inth l n))  (suffix n l)))
                (ilen ((imap (shiftm (inth l n))  (suffix n l)))) = ilen \<sigma> -(inth l n)" 
   by (metis "4" "6" "9" eq_imp_le filt_ilen ilen_imap idx_shiftm_suffix_inth 
        suffix_ilen_good ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 24: "l = fuse (prefix n l) 
                    (imap (shift (inth l n) ) ((imap (shiftm (inth l n))  (suffix n l))))"
   using fuse_prefix_suffix 
   by (metis "13" "14" "8" filt_ilen lsk_ls)
 have 25: "(\<exists>l1. index_sequence 0 l1 \<and>
              inth l1 (ilen l1) = ilen (prefix (inth l n) \<sigma>) \<and>
             powerinterval f (prefix (inth l n) \<sigma>) l1 \<and> g (filt (prefix (inth l n) \<sigma>) l1))"
   by (metis "12" "17" "19" "21" "22" prefix_ilen_good) 
 have 26: "(\<exists>l2. index_sequence 0 l2 \<and> 
              inth l2 (ilen l2) = ilen (suffix (inth l n) \<sigma>) \<and>
             powerinterval f (suffix (inth l n) \<sigma>) l2 \<and> h (filt (suffix (inth l n) \<sigma>) l2))"
   using "18" "14" "20" "23" "21" by auto
 have 27: "(prefix (inth l n) \<sigma>) \<Turnstile> f\<triangle>g"
     by (metis "25" projection_d_def)
 have 28: " (suffix (inth l n) \<sigma>) \<Turnstile> f\<triangle>h"  
    by (metis "26" projection_d_def)
  show ?thesis 
  using "21" "27" "28" chop_defs by auto
qed

lemma PJ4semchainb:
assumes "(\<sigma> \<Turnstile> (f \<triangle> g) ; (f \<triangle> h))"
 shows  "(\<sigma> \<Turnstile> f \<triangle> (g;h))"
proof -
 have 1: "(\<exists>n\<le>ilen \<sigma>.
        (\<exists>l. index_sequence 0 l \<and>
              inth l (ilen l) = ilen (prefix n \<sigma>) \<and>
             powerinterval f (prefix n \<sigma>) l \<and> g (filt (prefix n \<sigma>) l)) \<and>
        (\<exists>l. index_sequence 0 l \<and>
              inth l (ilen l) = ilen (suffix n \<sigma>) \<and>
             powerinterval f (suffix n \<sigma>) l \<and> h (filt (suffix n \<sigma>) l))) " 
    using assms by (metis chop_defs cpl_projection)
 obtain cp where 2: "cp\<le>ilen \<sigma> \<and> 
        (\<exists>l. index_sequence 0 l \<and>
              inth l (ilen l) = ilen (prefix cp \<sigma>) \<and>
             powerinterval f (prefix cp \<sigma>) l \<and> g (filt (prefix cp \<sigma>) l)) \<and>
        (\<exists>l. index_sequence 0 l \<and>
              inth l (ilen l) = ilen (suffix cp \<sigma>) \<and>
             powerinterval f (suffix cp \<sigma>) l \<and> h (filt (suffix cp \<sigma>) l))"
   using "1" by auto
 have 3: "cp\<le>ilen \<sigma>"
  using "2" by auto
 have 4: "(\<exists>l. index_sequence 0 l \<and>
              inth l (ilen l) = ilen (prefix cp \<sigma>) \<and>
             powerinterval f (prefix cp \<sigma>) l \<and> g (filt (prefix cp \<sigma>) l))" 
  using "2" by auto
 obtain l1 where 5: "index_sequence 0 l1 \<and>
              inth l1 (ilen l1) = ilen (prefix cp \<sigma>) \<and>
             powerinterval f (prefix cp \<sigma>) l1 \<and> g (filt (prefix cp \<sigma>) l1)"
   using "4" by auto
 have 6: "index_sequence 0 l1"  
   using "5" by auto
 have 7: " inth l1 (ilen l1) = ilen (prefix cp \<sigma>)"  
   using "5" by auto
 have 8: "powerinterval f (prefix cp \<sigma>) l1" 
    using "5" by auto
 have 9: "g (filt (prefix cp \<sigma>) l1)"
   using "5" by auto
 have 10: "(\<exists>l. index_sequence 0 l \<and>
              inth l (ilen l) = ilen (suffix cp \<sigma>) \<and>
             powerinterval f (suffix cp \<sigma>) l \<and> h (filt (suffix cp \<sigma>) l))" 
   using "2" by auto
 obtain l2 where 11: "index_sequence 0 l2 \<and>
              inth l2 (ilen l2) = ilen (suffix cp \<sigma>) \<and>
             powerinterval f (suffix cp \<sigma>) l2 \<and> h (filt (suffix cp \<sigma>) l2)"  
   using "10" by auto
 have 12: "index_sequence 0 l2" 
   using "11" by auto
 have 13: " inth l2 (ilen l2) = ilen (suffix cp \<sigma>)"
     using "11" by auto
 have 14: "powerinterval f (suffix cp \<sigma>) l2" 
    using "11" by auto
 have 15: "h (filt (suffix cp \<sigma>) l2)" 
    using "11" by auto
 have 16: "index_sequence 0 (fuse l1 (imap (shift cp) l2))"
    by (metis "11" "12" "2" "5" "6" eq_imp_le idx_fuse_idx prefix_ilen_good 
        suffix_ilen_good)
 have 17: "ilast l1 = ifirst (imap (shift cp) l2)"
    by (metis "12" "13" "3" "6" "7" idx_fuse_ifirst_ilast prefix_ilen_good 
        suffix_ilen_good)
 have 18: "inth (fuse l1 (imap (shift cp) l2)) (ilen l1) = cp"
    by (metis "17" "3" "7" eq_imp_le fuse_ilen_a fuse_inth 
        prefix_ilen_good le_add1)
 have 19: "ilast (fuse l1 (imap (shift cp) l2))  = ilen \<sigma> " 
    proof -
      have 191: "ilast (fuse l1 (imap (shift cp) l2)) = ilast (imap (shift cp) l2)" 
         by (metis "17" eq_imp_le fuse_inth_a fuse_ilen_a)
      have 192: "ilast (imap (shift cp) l2) = ilast l2 + cp"
        by (metis  shift_def ilen_imap inth_imap) 
      have 193: "ilast l2 + cp = ilen \<sigma>" 
         by (metis "13" "3" Nat.le_imp_diff_is_add suffix_ilen_good)
      show ?thesis using "191" "192" "193" by auto
     qed
 have 20: "powerinterval f \<sigma> (fuse l1 (imap (shift cp) l2))" 
   using powerinterval_fuse[of "l1" "(l2)" "cp" "\<sigma>" "f"]  
   using "12" "13" "14" "3" "5" by auto
 have 21: "\<sigma> = fuse (prefix cp \<sigma>) (suffix cp \<sigma>)"
     by (simp add: "3" fuse_prefix_suffix) 
 have 22: "inth ((fuse l1 (imap (shift cp) l2))) (ilen l1) = cp"
     using "18" by blast
 have 23: " (prefix (ilen l1) (filt \<sigma> (fuse l1 (imap (shift cp) l2)) )) =
            (filt (prefix cp \<sigma>) l1)"
    by (metis "17" "2" "5" filt_prefix filt_prefix_idx fuse_ilen_a 
        prefix_fuse prefix_ilen_good le_add1)
 have 24: "g (prefix (ilen l1) (filt \<sigma> (fuse l1 (imap (shift cp) l2)) ))"
   by (simp add: "23" "9")    
 have 25: "(suffix (ilen l1) (filt \<sigma> (fuse l1 (imap (shift cp) l2)) )) =
           (filt (suffix cp \<sigma>) l2)" 
   by (metis "12" "13" "17" "2" "23" filt_ilen filt_suffix filt_suffix_idx 
       prefix_ilen_bound suffix_fuse suffix_ilen_good)
 have 26: "ilen l1 \<le> ilen (filt \<sigma> (fuse l1 (imap (shift cp) l2)))"
         by (metis "23" filt_ilen prefix_ilen_bound)
 have 27: "(\<exists>l. index_sequence 0 l \<and>
          inth l (ilen l) = ilen \<sigma> \<and>
         powerinterval f \<sigma> l \<and>
         (\<exists>n\<le>ilen (filt \<sigma> l). g (prefix n (filt \<sigma> l)) \<and> h (suffix n (filt \<sigma> l))))"
    by (metis "11" "16" "19" "20" "24" "25" "26")      
 show ?thesis
 by (metis "27" chop_fuse fuse_prefix_suffix ilast_ifirst projection_d_def)
qed 

lemma PJ4sem:
"(\<sigma> \<Turnstile> f \<triangle> (g;h) = (f \<triangle> g) ; (f \<triangle> h))"
using PJ4semchaina PJ4semchainb unl_lift2 by blast



subsubsection \<open>PJ5\<close>

lemma PJ5sem:
 "(\<sigma> \<Turnstile> f \<triangle> init(g) \<longrightarrow> init(g))"
by (simp add: projection_d_def init_defs)
   ( metis filt_inth filt_ilen index_sequence_def ilen_gr_zero )

subsubsection \<open>PJ6\<close>

lemma PJ6help1:
 assumes "index_sequence 0 l"
         " (inth l (ilen l)) = (ilen \<sigma>)"
 shows   "(\<forall> i. 0\<le> i \<and> i<ilen l \<longrightarrow> ilen (sub (inth l i) (inth l (Suc i)) \<sigma>)
          = (inth l (Suc i)) - (inth l i)) "
proof 
  fix i
  show "0 \<le> i \<and> i < ilen l \<longrightarrow>
         ilen (sub (inth l i) (inth l (Suc i)) \<sigma>) = inth l (Suc i) - inth l i"
  using assms 
  by (simp add: index_sequence_def sub_def) 
     (metis Suc_lessI assms(1) idx_less_last_1 le_diff_iff le_eq_less_or_eq min.orderE)  
qed


lemma PJ6help2:
assumes " index_sequence 0 l"
        " inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i<ilen l. inth l (Suc i) - inth l i = Suc 0) "
shows   " (\<forall>i\<le>ilen l. inth l i = i)"
proof 
 fix i
 show " i\<le>ilen l \<longrightarrow> inth l i = i"
 proof
  (induct i)
  case 0
  then show ?case using assms index_sequence_def by blast
  next
  case (Suc i)
  then show ?case 
  by (metis One_nat_def Suc_eq_plus1 Suc_leD Suc_le_lessD assms idx_expand  
      le_add_diff_inverse2 plus_1_eq_Suc)
  qed
qed


lemma PJ6help3:
assumes " index_sequence 0 l"
        " inth l (ilen l) = ilen \<sigma>"
        " (\<forall>i\<le>ilen l. inth l i = i) "
shows   " (\<forall>i<ilen l. inth l (Suc i) - inth l i = Suc 0)
          "
proof
 fix i
 show "i<ilen l \<longrightarrow> inth l (Suc i) - inth l i = Suc 0"
 by (simp add: assms)
qed


lemma PJ6help4:
 " (\<exists>l. index_sequence 0 l \<and> l = [0..\<le>ilen \<sigma>] \<and>
          inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i\<le>ilen l. inth l i = i) \<and> 
         (\<exists>s1. ilen s1 = ilen l \<and> (\<forall>i\<le>ilen s1. inth s1 i = inth \<sigma> (i)) \<and> s1 = \<sigma>))
 "
by (simp add: index_sequence_def upt_ilen upt_inth)



lemma PJ6help5:
 "(\<exists>l. index_sequence 0 l \<and>
          inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i<ilen l. ilen (sub (inth l i) (inth l (Suc i)) \<sigma>) = Suc 0) \<and>
         (\<exists>s1. ilen s1 = ilen l \<and> (\<forall>i\<le>ilen s1. inth s1 i = inth \<sigma> ( inth l i)) \<and> g s1))
   = g \<sigma> "
proof -
 have "(\<exists>l. index_sequence 0 l \<and>
          inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i<ilen l. ilen (sub (inth l i) (inth l (Suc i)) \<sigma>) = Suc 0) \<and>
         (\<exists>s1. ilen s1 = ilen l \<and> (\<forall>i\<le>ilen s1. inth s1 i = inth \<sigma> (inth l i)) \<and> g s1))
       = 
       (\<exists>l. index_sequence 0 l \<and>
          inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i<ilen l. inth l (Suc i) - inth l i = Suc 0) \<and>
         (\<exists>s1. ilen s1 = ilen l \<and> (\<forall>i\<le>ilen s1. inth s1 i = inth \<sigma> ( inth l i)) \<and> g s1))"
     using PJ6help1 by (metis zero_order(1))
  also have " ... = 
            (\<exists>l. index_sequence 0 l \<and>
          inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i\<le>ilen l. inth l i = i) \<and>
         (\<exists>s1. ilen s1 = ilen l \<and> (\<forall>i\<le>ilen s1. inth s1 i = inth \<sigma> (inth l i)) \<and> g s1))
            "
       using PJ6help2 PJ6help3 by blast
 also have " ... =
             (\<exists>l. index_sequence 0 l \<and>
          inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i\<le>ilen l. inth l i = i) \<and>
         (\<exists>s1. ilen s1 = ilen l \<and> (\<forall>i\<le>ilen s1. inth s1 i = inth \<sigma> (i)) \<and> g s1))
           "
       by metis
 also have " ... =
           (\<exists>l. index_sequence 0 l \<and>
          inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i\<le>ilen l. inth l i = i) \<and> 
         (\<exists>s1. ilen s1 = ilen l \<and> (\<forall>i\<le>ilen s1. inth s1 i = inth \<sigma> (i)) \<and> s1 = \<sigma> \<and> g \<sigma>))
           "
    by (metis interval_eq_inth_eq le_eq_less_or_eq) 
 also have " ... =
             g \<sigma>
           " 
   using PJ6help4 by blast
 finally show "(\<exists>l. index_sequence 0 l \<and>
          inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i<ilen l. ilen (sub (inth l i) (inth l (Suc i)) \<sigma>) = Suc 0) \<and>
         (\<exists>s1. ilen s1 = ilen l \<and> (\<forall>i\<le>ilen s1. inth s1 i = inth \<sigma> ( inth l i)) \<and> g s1))
   = g \<sigma>
              " .
qed

lemma PJ6sem:
"(\<sigma> \<Turnstile> skip \<triangle> g = g)"
proof -
 have 1: "(\<sigma> \<Turnstile> skip \<triangle> g = g) =
          ((\<exists>l. index_sequence 0 l \<and> inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i<ilen l.  (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> skip) \<and> g (filt \<sigma> l)) =
         g \<sigma>)" 
       by (simp add: projection_d_def powerinterval_def)
 have 2: "(\<exists>l. index_sequence 0 l \<and> inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i<ilen l.  (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> skip) \<and> g (filt \<sigma> l)) =
         (\<exists>l s1 . index_sequence 0 l \<and> inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i<ilen l.  (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> skip) \<and> 
          ilen s1 = ilen l \<and>
          (\<forall> i\<le> ilen s1. (inth s1 i) = (inth \<sigma> (inth l i))) \<and>
         g s1)" 
     using filt_expand by metis
  have 3: "(\<exists>l s1 . index_sequence 0 l \<and> inth l (ilen l) = ilen \<sigma> \<and>
         (\<forall>i<ilen l.  (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> skip) \<and> 
          ilen s1 = ilen l \<and>
          (\<forall> i\<le> ilen s1. (inth s1 i) = (inth \<sigma> (inth l i))) \<and>
         g s1) =
         (\<exists>l. index_sequence 0 l \<and> inth l (ilen l) = ilen \<sigma> \<and>
        (\<forall>i<ilen l. ilen (sub (inth l i) (inth l (Suc i)) \<sigma>) = Suc 0) \<and>
        (\<exists>s1. ilen s1 = ilen l \<and> (\<forall>i\<le>ilen s1. inth s1 i = inth \<sigma> (inth l i)) \<and> g s1)) "
   by (simp add: skip_defs)
  have 4: "(\<exists>l. index_sequence 0 l \<and> inth l (ilen l) = ilen \<sigma> \<and>
        (\<forall>i<ilen l. ilen (sub (inth l i) (inth l (Suc i)) \<sigma>) = Suc 0) \<and>
        (\<exists>s1. ilen s1 = ilen l \<and> (\<forall>i\<le>ilen s1. inth s1 i = inth \<sigma> (inth l i)) \<and> g s1)) =
        (g \<sigma>) "
    by (simp add: PJ6help5)
  from 1 2 3 4 show ?thesis 
  by simp
qed


subsubsection \<open>PJ7\<close>


lemma PJemptyImp:
assumes " ilen \<sigma> = 0"
 shows  "(\<sigma> \<Turnstile> (f \<triangle> g) = g)"  
using assms 
by (simp add: projection_d_def index_sequence_def powerinterval_def,
    auto, 
    metis filt.simps(1) INil_ilen lessI not_less_zero old.nat.exhaust,
    metis  inth.simps(1) filt.simps(1) suffix_ilast suffix_zero 
    ilen.simps(1) not_less_zero)

lemma PJ7empty:
 assumes " ilen \<sigma> = 0"
 shows   "(\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h) = (f \<triangle> g) \<triangle> h)"
proof -
 have 1: "(\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h) = (g \<triangle> h)) "
   using PJemptyImp assms by blast
 have 2: "(\<sigma> \<Turnstile> (g \<triangle> h) = h)"
   using PJemptyImp assms by blast
 have 3: "(\<sigma> \<Turnstile> (f \<triangle> g) \<triangle> h = h) "
   using PJemptyImp assms by blast
 from 1 2 3 show ?thesis by simp
qed

lemma PJ7helpchain1a_help_1:
 assumes "index_sequence 0 l"
         "(inth l (ilen l)) = ilen \<sigma> "  
         " (\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " ilen \<sigma> > 0 " 
 shows   " index_sequence 0  (lfuse ( (lcpl f g \<sigma> l)))"
proof -
 have 0:  "ilen \<sigma> > 0 \<longrightarrow> ilen l >0"
    using assms gr_zeroI index_sequence_def by fastforce
 have 01: "ifirst l = 0"
   using assms index_sequence_def by auto
 have 02: "lastfirst (lcpl f g \<sigma> l)"
    using assms lcpl_lfuse_lastfirst by (simp add: projection_d_def) 
     (metis "0" "01" assms(3) lcpl_lfuse_lastfirst)
 have 1: "ilen \<sigma> > 0 \<longrightarrow> ifirst (lfuse  (lcpl f g \<sigma> l)) = 0 "
   using assms 0  01  02 
     lcpl_ifirst[of "l" "\<sigma>" "f" "g" ]  
    by (metis (mono_tags, lifting)  lastfirst_lfuse )
 from 1 0 show ?thesis using assms lcpl_lfuse_idx[of "l" "\<sigma>" "f" "g"] by(simp add: projection_d_def)  
qed

lemma PJ7helpchain1a_help_2:
 assumes "index_sequence 0 l"
         " (inth l (ilen l)) = ilen \<sigma> "
         " (\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g)"
         " ilen \<sigma> > 0 " 
 shows   " powerinterval f \<sigma> (lfuse ( (lcpl f g \<sigma> l)))"
proof -
 have 1: "(\<forall> i<ilen l. 
             powerinterval f (sub (inth l i) (inth l (Suc i)) \<sigma>) 
                             (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)))"
   using assms cpl_projection by blast 
 have 2: "\<forall>i<ilen l.
           \<forall>ia<ilen (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)).
            f (sub (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) ia)
                   (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) (Suc ia))
                   (sub (inth l i) (inth l (Suc i)) \<sigma>))" 
   using "1" by (simp add: powerinterval_def)
 have 3: "\<forall>i<ilen l.
           \<forall>ia<ilen (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)).
              (sub (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) ia)
                   (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) (Suc ia))
                   (sub (inth l i) (inth l (Suc i)) \<sigma>)) =
            (sub (inth (imap (shift (inth l i)) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))) ia)
                  (inth (imap (shift (inth l i)) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))) (Suc ia))
                  \<sigma>)"
   proof -
    have 30: "\<forall>i<ilen l.
               \<forall>ia<ilen (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)).
                \<forall>j\<le>ilen (sub (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) ia)
                   (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) (Suc ia))
                   (sub (inth l i) (inth l (Suc i)) \<sigma>)).
                  (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) (Suc ia)) \<le> 
                  (inth l (Suc i)) - (inth l (i)) "
         using assms by (metis add.commute cpl_projection idx_expand ilen_sub
              plus_1_eq_Suc)
    have 31: "\<forall>i<ilen l.
               \<forall>ia<ilen (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)).
                ilen (sub (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) ia)
                            (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) (Suc ia))
                            (sub (inth l i) (inth l (Suc i)) \<sigma>)) =
                ilen (sub (inth (imap (shift (inth l i)) 
                                 (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))) ia)
                            (inth (imap (shift (inth l i)) 
                                 (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))) (Suc ia))
                  \<sigma>)"
         using assms unfolding shift_def inth_imap 
          by (metis (no_types, lifting) PJ6help1 add.commute cpl_projection idx_expand 
             sub_sub_1 le_add1 le_add_same_cancel1 plus_1_eq_Suc)
    have 32: "\<forall>i<ilen l.
               \<forall>ia<ilen (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)).
                \<forall>j\<le>ilen (sub (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) ia)
                   (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) (Suc ia))
                   (sub (inth l i) (inth l (Suc i)) \<sigma>)).
                  inth (sub (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) ia)
                           (inth (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)) (Suc ia))
                           (sub (inth l i) (inth l (Suc i)) \<sigma>)) j =
                  inth (sub (inth (imap (shift (inth l i)) 
                                (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))) ia)
                           (inth (imap (shift (inth l i)) 
                                (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))) (Suc ia))
                  \<sigma>) j"
              using assms 30 sub_sub_1[of _ _ _ _ "\<sigma>"] 
              by (auto simp add: shift_def inth_imap cpl_projection idx_expand )
                 (metis add.commute idx_expand plus_1_eq_Suc)           
    show ?thesis using "31" "32" interval_eq_inth_eq by (simp add: interval_eq_inth_eq)
   qed
  have 4: "\<forall>i<ilen l.
           \<forall>ia<ilen (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>)).
            f (sub (inth (imap (shift (inth l i)) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))) ia)
                  (inth (imap (shift (inth l i)) (cpl f g (sub (inth l i) (inth l (Suc i)) \<sigma>))) (Suc ia))
                  \<sigma>)" 
       using "2" "3" by auto
  have 5: "ilen(lcpl f g \<sigma> l) = ilen l -1"
      using assms index_sequence_def lcpl_ilen lcpl_ilen_zero by fastforce
  have 6: "ilen \<sigma> >0 \<longrightarrow> ilen l >0" 
      using assms gr_zeroI index_sequence_def by fastforce
  have 7: "\<forall>i< ilen l.
            \<forall>ia<ilen ((inth (lcpl f g \<sigma> l) i)).
             f (sub (inth (inth (lcpl f g \<sigma> l) i) ia)
                    (inth (inth (lcpl f g \<sigma> l) i) (Suc ia))
                    \<sigma>)" 
      using assms by (metis (no_types, lifting) "4" index_sequence_def ilen_imap lcpl_inth)
  have 8: "ilen \<sigma> >0 \<longrightarrow>
            (\<forall>i\<le> ilen(lcpl f g \<sigma> l).
              \<forall>ia<ilen ((inth (lcpl f g \<sigma> l) i)).
               f (sub (inth (inth (lcpl f g \<sigma> l) i) ia)
                      (inth (inth (lcpl f g \<sigma> l) i) (Suc ia))
                      \<sigma>))" 
     by (metis "5" "6" "7" One_nat_def Suc_pred le_imp_less_Suc)
  have 9: "ilen \<sigma> >0 \<longrightarrow> 
            powerinterval f \<sigma> (lfuse ( (lcpl f g \<sigma> l))) =
            (\<forall>i<ilen (lfuse (lcpl f g \<sigma> l)).
              f (sub (inth (lfuse (lcpl f g \<sigma> l)) i) (inth (lfuse (lcpl f g \<sigma> l)) (Suc i)) \<sigma>))"
   by (simp add: powerinterval_def)
  have 10: "ilen \<sigma> >0 \<longrightarrow> lastfirst (lcpl f g \<sigma> l)"
    using "6" assms index_sequence_def lcpl_lfuse_lastfirst by blast
  have 11: "ilen \<sigma> >0 \<longrightarrow>
             (\<forall> j \<le> ilen (lcpl f g \<sigma> l). ilen(inth (lcpl f g \<sigma> l) j) > 0)
             " 
      by (simp add: assms lcpl_ilen_inth_gr_zero)
  have 12: "ilen \<sigma> >0 \<longrightarrow>
             (\<forall>i\<le> ilen(lcpl f g \<sigma> l).
               (\<forall>ia<ilen ((inth (lcpl f g \<sigma> l) i)).
                 f (sub (inth (inth (lcpl f g \<sigma> l) i) ia)
                        (inth (inth (lcpl f g \<sigma> l) i) (Suc ia))
                        \<sigma>))) =
                  (\<forall>j<ilen (lfuse (lcpl f g \<sigma> l)).
                 f (sub (inth (lfuse (lcpl f g \<sigma> l)) j) 
                        (inth (lfuse (lcpl f g \<sigma> l)) (Suc j)) 
                        \<sigma>))
              " 
     using lfuse_split[of "(lcpl f g \<sigma> l)" "f" "\<sigma>"] "10" "11" by auto
   show ?thesis using "12" "8" "9" using assms(4) by blast
qed    

lemma PJ7helpchain1a_help_3:
 assumes "index_sequence 0 l"
         " (inth l (ilen l)) = ilen \<sigma>" 
         " (\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g) " 
         " h (filt \<sigma> l) "
         " ilen \<sigma> >0 " 
 shows   " ilast (lfuse (lcpl f g \<sigma> l)) = ilen \<sigma>"
proof -
 have 0: "ilen \<sigma> > 0 \<longrightarrow> ilen l >0"
    using assms gr_zeroI index_sequence_def by fastforce
 have 1: "ilen \<sigma> >0 \<longrightarrow>
            ilast (lfuse (lcpl f g \<sigma> l)) = ilast(ilast ( (lcpl f g \<sigma> l)))" 
      using assms 
      using "0" index_sequence_def lastfirst_lfuse_ilast lcpl_lfuse_lastfirst by blast
 have 2: "ilen \<sigma> >0 \<longrightarrow>
         ilast ( (lcpl f g \<sigma> l))
          = (inth (lcpl f g \<sigma> l) (ilen l -1)) " 
       using assms by (simp add: "0" index_sequence_def lcpl_ilen)
 have 3: "ilen \<sigma> >0 \<longrightarrow>
          (inth (lcpl f g \<sigma> l) (ilen l -1)) =
           (imap (shift (inth l (ilen l -1))) 
                (cpl f g (sub (inth l (ilen l -1)) (inth l (Suc (ilen l -1))) \<sigma>)))
         " 
       using assms by (simp add: "0" index_sequence_def lcpl_inth)
 have 4: "ilen \<sigma> >0 \<longrightarrow>
          ilast( ( (cpl f g (sub (inth l (ilen l -1)) (inth l (Suc (ilen l -1))) \<sigma>)))) =
           ilen ((sub (inth l (ilen l -1)) (inth l (Suc (ilen l -1))) \<sigma>))
          " using "0" assms 
          by (metis cpl_projection diff_less zero_less_one)         
 have 5: "ilen \<sigma> >0 \<longrightarrow>
         ilen ((sub (inth l (ilen l -1)) (inth l (Suc (ilen l -1))) \<sigma>)) =
            (inth l (Suc (ilen l -1))) - (inth l (ilen l -1))     
        "  
     using "0" PJ6help1 assms diff_less zero_less_one by blast 
 have 6: "ilen \<sigma> >0 \<longrightarrow>
          ilast ( (imap (shift (inth l (ilen l -1))) 
                         (cpl f g (sub (inth l (ilen l -1)) (inth l (Suc (ilen l -1))) \<sigma>)))) =
             (shift (inth l (ilen l -1))) 
              (ilast (cpl f g (sub (inth l (ilen l -1)) (inth l (Suc (ilen l -1))) \<sigma>)))
         "      
   by (metis ilen_imap inth_imap)
 have 7: "ilen \<sigma> >0 \<longrightarrow> 
          (shift (inth l (ilen l -1))) 
           (ilast (cpl f g (sub (inth l (ilen l -1)) (inth l (Suc (ilen l -1))) \<sigma>))) =
          (shift (inth l (ilen l -1))) ((inth l (Suc (ilen l -1))) - (inth l (ilen l -1)))
          " 
      using "4" "5" by auto
 have 8: "ilen \<sigma> >0 \<longrightarrow> 
          (shift (inth l (ilen l -1))) ((inth l (Suc (ilen l -1))) - (inth l (ilen l -1))) =
           ((inth l (Suc (ilen l -1))) - (inth l (ilen l -1))) + (inth l (ilen l -1))
         " 
    by (simp add:  shift_def)
 have 9: "ilen \<sigma> >0 \<longrightarrow>  
          ((inth l (Suc (ilen l -1))) - (inth l (ilen l -1))) + (inth l (ilen l -1)) 
          = (inth l (Suc (ilen l -1)))
         " 
    using assms
    by (metis "0" add.commute diff_add diff_less idx_expand plus_1_eq_Suc zero_less_one) 
 have 10: "ilen \<sigma> >0 \<longrightarrow> (inth l (Suc (ilen l -1))) = ilen \<sigma> " 
     by (simp add: "0" assms)
  show ?thesis 
  using "1" "10" "2" "3" "6" "7" "8" "9" assms(5) by presburger
qed

lemma PJ7helpchain1a_help_4:
 assumes "index_sequence 0 l"
         " (inth l (ilen l)) = ilen \<sigma>" 
         " (\<forall> i<ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g) " 
         "  h (filt \<sigma> l) "
         " ilen \<sigma> >0 "
 shows   " ((filt \<sigma> (lfuse (lcpl f g \<sigma> l))) \<Turnstile> g \<triangle> h) "
proof -
 have 0: "ilen \<sigma> > 0 \<longrightarrow> ilen l >0"
    using assms gr_zeroI index_sequence_def by fastforce
 have 1: "ilen \<sigma> >0 \<longrightarrow> index_sequence 0 (addzero (lsum (lcpl f g \<sigma> l) 0)) "
    by (simp add: assms lcpl_ilen_inth_gr_zero lsum_addzero_idx)
 have 2: "ilen \<sigma> > 0 \<longrightarrow> 
          (inth (addzero (lsum (lcpl f g \<sigma> l) 0)) (ilen(addzero (lsum (lcpl f g \<sigma> l) 0)))) =
          ilen (filt \<sigma> (lfuse (lcpl f g \<sigma> l)))"
  by (simp add: assms lcpl_lfuse_filt_ilen)
 have 3: "ilen \<sigma> >0 \<longrightarrow>  
          powerinterval g (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) (addzero (lsum (lcpl f g \<sigma> l) 0))
              " 
    by (simp add: assms lcpl_lfuse_filt_power)
 have 4: "ilen \<sigma> >0 \<longrightarrow>
           h (filt (filt \<sigma> (lfuse (lcpl f g \<sigma> l))) (addzero (lsum (lcpl f g \<sigma> l) 0))) "
   by (simp add: assms filt_imap_filt lcpl_filt_lfuse_lsum)
show ?thesis by (metis "1" "2" "3" "4" assms(5) projection_d_def)
qed


lemma PJ7helpchain1a:
assumes "ilen \<sigma> > 0"
        "(\<sigma> \<Turnstile> (f \<triangle> g) \<triangle> h)" 
shows " (\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h))"
proof -
 have 1: "ilen \<sigma> > 0"
  using assms by auto
 have 2: "(\<exists>l. index_sequence 0 l \<and> inth l (ilen l) = ilen \<sigma> \<and>
       powerinterval (LIFT(f \<triangle> g)) \<sigma> l \<and>
         h (filt \<sigma> l)) "  
 using assms using cpl_projection by blast
 obtain l where 3: "index_sequence 0 l \<and> inth l (ilen l) = ilen \<sigma> \<and>
       powerinterval (LIFT(f \<triangle> g)) \<sigma> l \<and>
         h (filt \<sigma> l)"
  using "2" by blast
 have 4: "index_sequence 0 l \<and> inth l (ilen l) = ilen \<sigma> \<and>
            (\<forall> i <ilen l. (sub (inth l i) (inth l (Suc i)) \<sigma>) \<Turnstile> f \<triangle> g) \<and>
         h (filt \<sigma> l)" 
 using "3"  by (simp add: powerinterval_def) 
 have 6: "ilen l>0"
  using "1" "4" gr_zeroI index_sequence_def by fastforce
 have 7: "index_sequence 0  (lfuse (lcpl f g \<sigma> l))" 
 by (metis (no_types, lifting) "1" "4" PJ7helpchain1a_help_1)
 have 8: "powerinterval f \<sigma> (lfuse (lcpl f g \<sigma> l))"   
  using "4" "6" PJ7helpchain1a_help_2 assms index_sequence_def by auto
 have 9: "(inth (lfuse (lcpl f g \<sigma> l)) (ilen (lfuse (lcpl f g \<sigma> l)))) = ilen \<sigma> " 
  by (metis "1" "4" PJ7helpchain1a_help_3)
 have 10: "((filt \<sigma> (lfuse (lcpl f g \<sigma> l))) \<Turnstile> g \<triangle> h) " 
   by (simp add: "1" "4" "6" PJ7helpchain1a_help_4) 
show ?thesis 
using "10" "7" "8" "9" by (metis projection_d_def)
qed
       

lemma PJ7helpchain1b:
assumes "ilen \<sigma> > 0"
        "(\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h))" 
shows " (\<sigma> \<Turnstile> (f \<triangle> g) \<triangle> h)"
proof -
 have 1: "ilen \<sigma> > 0" 
  using assms by auto
 have 2: "(\<exists>l. index_sequence 0 l \<and> inth l (ilen l) = ilen \<sigma> \<and>
         powerinterval f \<sigma> l \<and>
         ((filt \<sigma> l) \<Turnstile> g \<triangle> h) ) " 
  using assms by (simp add: projection_d_def)
 obtain l where 3: "index_sequence 0 l \<and> inth l (ilen l) = ilen \<sigma> \<and>
         powerinterval f \<sigma> l \<and>
         ((filt \<sigma> l) \<Turnstile> g \<triangle> h)" 
  using "2" by blast
 have 4: "(\<exists> la. index_sequence 0 la \<and> inth la (ilen la) = ilen(filt \<sigma> l) \<and>
           powerinterval g (filt \<sigma> l) la \<and>
           ((filt (filt \<sigma> l) la) \<Turnstile> h))" 
  using "3" using cpl_projection by blast
 obtain la where 5: "index_sequence 0 la \<and> inth la (ilen la) = ilen(filt \<sigma> l) \<and>
           powerinterval g (filt \<sigma> l) la \<and>
           ((filt (filt \<sigma> l) la) \<Turnstile> h) " 
  using "4" by blast
 have 6: "ilen l >0 "
   using "1" "3" gr0I index_sequence_def by force
 have 7: "ilen(filt \<sigma> l) = ilen l "
    by (simp add: filt_ilen)
 have 8: "(filt (filt \<sigma> l) la) = (filt \<sigma> (filt l la)) " 
   using filt_imap_filt by blast
 have 9: "(inth (filt l la) (ilen (filt l la))) = ilen \<sigma> "
  by (metis "3" "5" filt_expand order_refl)
 have 10: "ilen la >0"
   using "5" "6" "7" gr0I index_sequence_def by force
 have 11: "ilen (filt l la) > 0  "
    by (simp add: "10" filt_ilen)
 have 12: "index_sequence 0 (filt l la) "
    proof - 
     have 111: "inth (filt l la) 0 = 0"
       by (metis "3" "5" filt_inth index_sequence_def ilen_gr_zero)
     have 112: "ilen(filt l la) = ilen la"
        using filt_expand by blast
     have 113: "(\<forall> i< ilen la. (inth (filt l la) i) = (inth l (inth la i))) "
      by (simp add: filt_imap inth_imap)
     have 114: "(\<forall> i< ilen la. (inth (filt l la) (Suc i)) = (inth l (inth la (Suc i)))) "
       by (simp add: filt_imap inth_imap)
     have 115: "(\<forall> i< ilen la. (inth l (inth la i)) < (inth l (inth la (Suc i))) ) "
       by (metis "3" "5" "7" Suc_lessI idx_less_than idx_less_last_1 lessI 
           less_imp_le_nat)
     show ?thesis by (simp add: "111" "113" "114" "115" filt_ilen index_sequence_def)
    qed
  have 20: "powerinterval (LIFT(f\<triangle>g)) \<sigma> (filt l la)"
    proof -
     have 201: "powerinterval (LIFT(f\<triangle>g)) \<sigma> (filt l la) =
               (\<forall> i< ilen la. (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) \<Turnstile> f \<triangle> g)  "
       by (simp add: filt_imap inth_imap powerinterval_def)
     have 202: "(\<forall> i< ilen la. (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) \<Turnstile> f \<triangle> g) =
                (\<forall> i< ilen la. 
                  (\<exists> ll. index_sequence 0 ll 
                  \<and> ilast ll = ilen (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) \<and>
                    powerinterval f (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) ll \<and>
                    ((filt (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) ll) \<Turnstile> g)   
                     ))
                "
       by (simp add: projection_d_def)
      have 203: "(\<forall> i< ilen la.  ilen (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) =
                    (inth l (inth la (Suc i))) - (inth l (inth la i)))  "
           by (metis "3" "5" "7" Suc_leI Suc_lessI idx_less_equal idx_less_last_1 
               ilen_sub lessI less_imp_le_nat)
      have 2041: "(\<forall> i< ilen la.  
                      (inth la (Suc i)) \<le> ilen l 
                )
                "
       using "5" "7" Suc_lessI idx_less_last_1 by fastforce 
      have 204: "(\<forall> i< ilen la.  
                   ilen (imap (shiftm (inth l (inth la i))) (sub (inth la i) (inth la (Suc i)) l)) =
                     (inth la (Suc i)) - (inth la i))
                " 
                by (simp add: "5" PJ6help1 filt_ilen)
      have 205: "(\<forall> i< ilen la.  
                  (\<forall> j \<le> (inth la (Suc i)) - (inth la i). 
                     (inth (sub (inth la i) (inth la (Suc i)) l) j) =
                      (inth l ((inth la i) +j))
                  )
                )           
                " 
        using "2041" "5" index_sequence_def inth_sub order.strict_implies_order by blast
      have 2060: "(\<forall> i< ilen la.  
                    (inth la i) \<le> (inth la (Suc i)) )
                 " 
           using "5" idx_expand by fastforce
      have 206: "(\<forall> i< ilen la.  
                   (\<forall> j \<le> (inth la (Suc i)) - (inth la i).
                      (inth l (inth la i)) \<le> (inth l ((inth la i) +j))
                   )
                 )
                "  
             using "2041" "3" "2060" idx_less_equal by fastforce
     have 207: " (\<forall> i< ilen la.  
                   (\<forall> j \<le> (inth la (Suc i)) - (inth la i).
                    (inth (imap (shiftm (inth l (inth la i))) (sub (inth la i) (inth la (Suc i)) l)) j ) =
                      (inth l ((inth la i) +j)) - (inth l (inth la i))
                    ))
                " 
                by (simp add: "205" inth_imap shiftm_def)
     have 208: "(\<forall> i< ilen la.  
                   (inth (imap (shiftm (inth l (inth la i))) (sub (inth la i) (inth la (Suc i)) l)) 0) =0 
              ) "
              by (simp add: "207")
     have 209: "(\<forall> i< ilen la.  
                   ilast (imap (shiftm (inth l (inth la i))) (sub (inth la i) (inth la (Suc i)) l)) =
                   (inth l (inth la (Suc i))) - (inth l (inth la i))
                ) 
               "   
             using "204" "207" "5" "2060" by simp_all        
     have 210: "(\<forall> i< ilen la. 
              index_sequence 0 (imap (shiftm (inth l (inth la i))) (sub (inth la i) (inth la (Suc i)) l))
                ) 
               " 
               by (metis "3" "5" "7" add.commute idx_expand idx_shiftm 
                   idx_sub plus_1_eq_Suc)
     have 211: "(\<forall> i< ilen la. 
                 powerinterval f (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) 
                       (imap (shiftm (inth l (inth la i))) (sub (inth la i) (inth la (Suc i)) l))
                ) 
              " 
       proof -
        have 2111: "(\<forall> i< ilen la. 
                     powerinterval f (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) 
                                     (imap (shiftm (inth l (inth la i))) 
                                     (sub (inth la i) (inth la (Suc i)) l))
                    )  = 
                    (\<forall>i<ilen la.
                     (\<forall>ia<ilen (sub (inth la i) (inth la (Suc i)) l).
                       f (sub (inth (imap (shiftm (inth l (inth la i)))
                                        (sub (inth la i) (inth la (Suc i)) l)) ia)
                              (inth (imap (shiftm (inth l ( inth la i))) 
                                        (sub (inth la i) (inth la (Suc i)) l)) (Suc ia))
                              (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) ) 
                     ) 
                     ) " 
        by (simp add: powerinterval_def)
       have 2112: " ... = 
                    (\<forall>i<ilen la.
                     (\<forall>ia<  (inth la (Suc i)) - (inth la i).
                       f (sub (inth (imap (shiftm (inth l (inth la i))) 
                                        (sub (inth la i) (inth la (Suc i)) l)) ia)
                              (inth (imap (shiftm (inth l ( inth la i))) 
                                        (sub (inth la i) (inth la (Suc i)) l)) (Suc ia))
                              (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>))
                     )
                    )"
               using "204" by auto
       have 2113: "... = 
                    (\<forall>i<ilen la.
                     (\<forall>ia<  (inth la (Suc i)) - (inth la i).
                        f (sub  ((inth l ((inth la i) +ia)) - (inth l (inth la i)))   
                                ((inth l ((inth la i) +(Suc ia))) - (inth l (inth la i)))   
                                (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>))
                     )
                    )" 
               using "207" by auto
       have 2114: "
                    (\<forall>i<ilen la.
                      (let m = (inth la (Suc i)); n= (inth la i) in
                       (\<forall>ia<  m - n. 
                         (inth l (n +ia))  \<le> ((inth l (n+ (Suc ia))) ) 
                       )
                      )
                     )" 
                using "2041" "3" "2060" idx_less_equal by simp_all fastforce
         have 2115: "
                    (\<forall>i<ilen la.
                      (let m = (inth la (Suc i)); n= (inth la i) in
                       (\<forall>ia<  m - n. 
                         (inth l (n +ia)) - (inth l n) \<le> ((inth l ((Suc (n+ia)))) - (inth l n)) 
                       )
                      )
                     ) "
           by (metis "2114" add_Suc_right diff_le_mono)
        have 2116: "
                    (\<forall>i<ilen la.
                      (let m = (inth la (Suc i)); n= (inth la i) in
                       (\<forall>ia<  m - n. 
                        ((inth l ((Suc (n+ia)))) - (inth l n))  \<le> (inth l m) - (inth l n) 
                       )
                      )
                     ) "
            by (metis "3" "5" "7" idx_expand Suc_leI add.commute diff_le_mono 
            idx_less_equal less_diff_conv plus_1_eq_Suc)  
      have 2117: "(\<forall>i<ilen la.
                      (let m = (inth la (Suc i)); n= (inth la i) in 
                        (inth l m) \<le> ilen \<sigma> )) "
         by (metis "12" "9" Suc_lessI filt_inth filt_ilen idx_less_last_1  
                   less_or_eq_imp_le)
      have 2118: "(\<forall>i<ilen la.
                    (let m = (inth la (Suc i)); n= (inth la i) in
                     (\<forall>ia<  m - n.
                        (n+(Suc ia)) \<le> ilen l )
                     ) 
                  ) "
              using "5" "7" idx_expand less_diff_conv by fastforce
      have 21190: "(\<forall>i<ilen la.
                    (let m = (inth la (Suc i)); n= (inth la i) in
                     (\<forall>ia<  m - n. 
                       ((inth l (n +ia)) - (inth l n)) \<le> ((inth l (n+(Suc ia))) - (inth l n))
                     ) ))
                   " 
            by (meson "2114" diff_le_mono)
      have 21191: "(\<forall>i<ilen la.
                    (let m = (inth la (Suc i)); n= (inth la i) in
                     (\<forall>ia<  m - n. 
                      ((inth l (n+(Suc ia))) - (inth l n)) \<le>
                       ilen (sub (inth l n) (inth l m) \<sigma>)
                      )))
                  " 
          by (simp add: "203" "2116")
      have 2119: "(\<forall>i<ilen la.
                    (let m = (inth la (Suc i)); n= (inth la i) in
                     (\<forall>ia<  m - n. 
                        ilen (sub  ((inth l (n +ia)) - (inth l n))   
                                     ((inth l (n+(Suc ia))) - (inth l n))   
                                     (sub (inth l n) (inth l m) \<sigma>)) =
                        ((inth l (n+(Suc ia)))) - ((inth l (n +ia))) 
                     )
                    )
                   )
                   "
              by (metis (no_types, lifting) "206" "21190" "21191" 
                       diff_diff_add ilen_sub le_add_diff_inverse less_imp_le)  
      have 2120: "(\<forall>i<ilen la.
                    (let m = (inth la (Suc i)); n= (inth la i) in
                     (\<forall>ia<  m - n.
                      (\<forall>j \<le> ((inth l (n+(Suc ia)))) - ((inth l (n +ia))).
                         inth  (sub  ((inth l (n +ia)) - (inth l n))   
                                    ((inth l (n+(Suc ia))) - (inth l n))   
                                    (sub (inth l n) (inth l m) \<sigma>)) j =
                         inth (sub (inth l n) (inth l m) \<sigma>) (((inth l (n +ia)) - (inth l n))+j) 
               )   )))  " 
               by (metis "2119" "21190" "21191" ilen_sub inth_sub)
      have 212111: "(\<forall>i<ilen la.
                     (let m = (inth la (Suc i)); n= (inth la i) in
                      (\<forall>ia<  m - n.
                           ((inth l (n +ia))) \<le> ((inth l (n+(Suc ia))))
                      )
                     )
                    )
                 "
               using "2114" by blast
      have 212112: "(\<forall>i<ilen la.
                     (let m = (inth la (Suc i)); n= (inth la i) in
                      (\<forall>ia<  m - n.
                           ((inth l (n))) \<le> ((inth l (n+(ia))))
                      )
                     )
                    )" 
          by (simp add: "206")
      have 212113: "(\<forall>i<ilen la.
                     (let m = (inth la (Suc i)); n= (inth la i) in
                      (\<forall>ia<  m - n.
                       (\<forall> j \<le> ((inth l (n+(Suc ia)))) - ((inth l (n +ia))). 
                         ((inth l (n +ia)) )+j \<le> (inth l m) 
                       )
                      )
                     )
                    )" 
           using "212111"  "212112"  "206" "2060" "2116"  
           unfolding Let_def by fastforce
      have 21211: "(\<forall>i<ilen la.
                     (let m = (inth la (Suc i)); n= (inth la i) in
                      (\<forall>ia<  m - n.
                       (\<forall> j \<le> ((inth l (n+(Suc ia)))) - ((inth l (n +ia))). 
                         ((inth l (n +ia)) - (inth l n))+j \<le> (inth l m) - (inth l n)
                       )
                      )
                     )
                    )
                 " 
       by (metis "212112" "212113" Nat.add_diff_assoc2 diff_le_mono)
      have 2121: "(\<forall>i<ilen la.
                    (let m = (inth la (Suc i)); n= (inth la i) in
                     (\<forall>ia<  m - n.
                      (\<forall> j \<le> ((inth l (n+(Suc ia)))) - ((inth l (n +ia))). 
                        (inth (sub (inth l n) (inth l m) \<sigma>) (((inth l (n +ia)) - (inth l n))+j) ) =
                        (inth \<sigma> ((inth l n) + (((inth l (n +ia)) - (inth l n))+j))   ) 
                      ) 
                     ) 
                    ) 
                   ) 
                 " 
              by (metis "206" "2117" "21211" add_diff_inverse_nat inth_sub nat_diff_split 
                  not_less_zero order_refl)
      have 2122: "(\<forall>i<ilen la.
                    (let m = (inth la (Suc i)); n= (inth la i) in
                     (\<forall>ia<  m - n.
                      (\<forall> j \<le> ((inth l (n+(Suc ia)))) - ((inth l (n +ia))). 
                       (inth \<sigma> ((inth l n) + (((inth l (n +ia)) - (inth l n))+j))   ) =
                       (inth \<sigma> (  (((inth l (n +ia)))+j))   ) 
                      ) 
                     ) 
                    ) 
                   )   " 
            by (metis "206" add.assoc le_add_diff_inverse less_imp_le_nat)
      have 2123:  "(\<forall>i<ilen la.
                     (let m = (inth la (Suc i)); n= (inth la i) in
                      (\<forall>ia<  m - n.
                       (\<forall> j \<le> ((inth l (n+(Suc ia)))) - ((inth l (n +ia))).
                         inth (sub (inth l (n +ia)) ((inth l (n+(Suc ia)))) \<sigma>) j =
                         (inth \<sigma> (  (((inth l (n +ia)))+j)) )
                       )
                      )
                     )
                    )
                  " 
                  by (metis "2114" "2118" "3" eq_imp_le idx_less_equal inth_sub)
      have 2124: "(\<forall>i<ilen la.
                    (let m = (inth la (Suc i)); n= (inth la i) in
                     (\<forall>ia<  m - n. 
                       ilen (sub (inth l (n +ia)) ((inth l (n+(Suc ia)))) \<sigma>) =
                       ((inth l (n+(Suc ia)))) - ((inth l (n +ia)))
                     )
                    )
                   )
                 " 
              by (metis "2118" "3" Suc_eq_plus1 Suc_le_lessD add_Suc_right idx_expand 
                 ilen_sub)
    have 2125: "(\<forall>i<ilen la.
                  (let m = (inth la (Suc i)); n= (inth la i) in
                    (\<forall>ia<  m - n.
                      (\<forall> j \<le> ((inth l (n+(Suc ia)))) - ((inth l (n +ia))).
                         inth  (sub  ((inth l (n +ia)) - (inth l n))   
                                    ((inth l (n+(Suc ia))) - (inth l n))   
                                    (sub (inth l n) (inth l m) \<sigma>)) j =
                        (inth (sub (inth l (n +ia)) ((inth l (n+(Suc ia)))) \<sigma>) j) 
                      )
                     )
                    ) 
                   )
                  " 
     by (metis "2120" "2121" "2122" "2123")  
    have 2126: "(\<forall>i<ilen la.
                   (let m = (inth la (Suc i)); n= (inth la i) in
                     (\<forall>ia<  m - n. 
                       ilen (sub  ((inth l (n +ia)) - (inth l n))   
                                    ((inth l (n+(Suc ia))) - (inth l n))   
                                    (sub (inth l n) (inth l m) \<sigma>)) =
                       ilen (sub (inth l (n +ia)) ((inth l (n+(Suc ia)))) \<sigma>)
                     )
                    )
                  ) "
       by (metis "2119" "2124")
    have 2127: "(\<forall>i<ilen la.
                   (let m = (inth la (Suc i)); n= (inth la i) in
                     (\<forall>ia<  m - n. 
                      (sub  ((inth l (n +ia)) - (inth l n))   
                            ((inth l (n+(Suc ia))) - (inth l n))   
                            (sub (inth l n) (inth l m) \<sigma>)) =
                      (sub (inth l (n +ia)) ((inth l (n+(Suc ia)))) \<sigma>) 
                     )
                   )
                 )
              " 
        using interval_eq_inth_eq 2125 2126 
        by (metis "2119")
    have 2128: "(\<forall>i<ilen la.
                   (let m = (inth la (Suc i)); n= (inth la i) in
                     (\<forall>ia<  m - n. 
                      f (sub (inth l (n +ia)) ((inth l (n+(Suc ia)))) \<sigma>) 
                     )
                    )
                  )
                 " 
           by (metis "2041" "3" add.commute add_Suc_right less_diff_conv less_le_trans 
                powerinterval_def)
    show ?thesis using 2111 2112 2113 2127 2128 unfolding Let_def by (simp add: "2113")
   qed
   have 220: "(\<forall> i< ilen la. 
                 ((filt (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>)
                        (imap (shiftm (inth l (inth la i))) (sub (inth la i) (inth la (Suc i)) l)) )
                   \<Turnstile> g)    
                )     
               " 
         proof -
          have 2201: "(\<forall> i< ilen la. 
                        ilen (filt (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>)
                                     (imap (shiftm (inth l (inth la i)))
                                          (sub (inth la i) (inth la (Suc i)) l)) ) =
                      ilen (imap (shiftm (inth l (inth la i))) (sub (inth la i) (inth la (Suc i)) l)) 
                    )
                   " 
               using filt_ilen by blast
          have 2202: "(\<forall> i< ilen la.
                         ilen (imap (shiftm (inth l (inth la i))) 
                                     (sub (inth la i) (inth la (Suc i)) l)) =
                            (inth la (Suc i)) - (inth la i)
                      )
                     "
            using "204" by blast         
          have 2203: "(\<forall> i< ilen la.  
                       (\<forall> j\<le> (inth la (Suc i)) - (inth la i) .
                         inth ((filt (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>)
                                    (imap (shiftm (inth l (inth la i))) 
                                         (sub (inth la i) (inth la (Suc i)) l)) ) ) j =
                           inth (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) 
                               (inth (imap (shiftm (inth l (inth la i))) 
                                         (sub (inth la i) (inth la (Suc i)) l)) j) 
                        ))
                      " 
                by (simp add: filt_imap inth_imap)
          have 2204: "(\<forall> i< ilen la.  
                       (\<forall> j\<le> (inth la (Suc i)) - (inth la i) .
                          inth (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) 
                               (inth (imap (shiftm (inth l (inth la i))) 
                                         (sub (inth la i) (inth la (Suc i)) l)) j) =
                          inth (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) 
                              ((inth (sub (inth la i) (inth la (Suc i)) l) j) -(inth l (inth la i)))
                       )) 
                     " 
              by (simp add: inth_imap shiftm_def)   
          have 2205: "(\<forall> i< ilen la.  
                       (
                          (inth l (inth la i)) \<le> (inth l (inth la (Suc i))) \<and>
                          (inth l (inth la (Suc i))) \<le> ilen \<sigma> 
                       ))
                     "  
                     by (metis "12" "9" add.commute filt_ilen filt_imap idx_expand 
                          inth_imap  plus_1_eq_Suc)
          have 2206 : "(\<forall> i< ilen la.  
                       (\<forall> j\<le> (inth la (Suc i)) - (inth la i) . 
                         (inth (sub (inth la i) (inth la (Suc i)) l) j) =
                          (inth l ((inth la i)+j))
                       ))
                       "
            using "205" by blast 
          have 2207: "(\<forall> i< ilen la.  
                       (\<forall> j\<le> (inth la (Suc i)) - (inth la i) .
                        ((inth l (inth la i)) + ((inth (sub (inth la i) (inth la (Suc i)) l) j)
                                               -(inth l (inth la i)))) =
                           (inth l ((inth la i)+j))
                       ))
                      "  
              by (simp add: "206" "2206")      
          have 2208: "(\<forall> i< ilen la.  
                       (\<forall> j\<le> (inth la (Suc i)) - (inth la i) .
                         inth (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) 
                             ((inth (sub (inth la i) (inth la (Suc i)) l) j) -(inth l (inth la i))) =
                         inth \<sigma>  (inth l ((inth la i)+j))          
                       ))
                      " using "2041" "206" "2060" "2205" "3" idx_less_equal[of "l" "\<sigma>"]
                   using Nat.le_diff_conv2 by auto
          have 2209: "(\<forall> i< ilen la.  
                       (\<forall> j\<le> (inth la (Suc i)) - (inth la i) .
                          (inth (filt \<sigma> l) ((inth la i)+j)) = (inth \<sigma> (inth l ((inth la i)+j)))
                       ))
                      "
           by (simp add: filt_imap inth_imap)
         have 2210: "(\<forall> i< ilen la. 
                        ilen  (sub (inth la i) (inth la (Suc i)) (filt \<sigma> l)) =
                          (inth la (Suc i)) - (inth la i) 
                     )
                    " 
                    using "5" PJ6help1 by blast
         have 2211: "(\<forall> i< ilen la. 
                      (\<forall> j\<le> (inth la (Suc i)) - (inth la i) .
                       (inth (sub (inth la i) (inth la (Suc i)) (filt \<sigma> l)) j) =
                        (inth \<sigma> (inth l ((inth la i)+j)))
                      ))
                    " 
                    using "2209" "5" idx_expand inth_sub by fastforce
         have 2212: "(\<forall> i< ilen la. 
                      (\<forall> j\<le> (inth la (Suc i)) - (inth la i) .
                             inth ((filt (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>)
                                    (imap (shiftm (inth l (inth la i))) 
                                         (sub (inth la i) (inth la (Suc i)) l)) ) ) j =
                     (inth (sub (inth la i) (inth la (Suc i)) (filt \<sigma> l)) j)
                 ))
                "  
               by (simp add: "2203" "2204" "2208" "2211")
         have 2213: "(\<forall> i< ilen la. 
                 ilen ((filt (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>)
                               (imap (shiftm (inth l (inth la i))) 
                                    (sub (inth la i) (inth la (Suc i)) l)) ) ) =
                 ilen (sub (inth la i) (inth la (Suc i)) (filt \<sigma> l))
                )
               " 
               by (metis "2202" "2210" filt_ilen)    
         have 2214: "(\<forall> i< ilen la. 
                       (filt (sub (inth l (inth la i)) (inth l (inth la (Suc i))) \<sigma>) 
                             (imap (shiftm (inth l (inth la i))) (sub (inth la i) (inth la (Suc i)) l))) =
                       (sub (inth la i) (inth la (Suc i)) (filt \<sigma> l))
                     ) 
                    " 
         using interval_eq_inth_eq 2212 2213 using "2201" "2202" by fastforce  
         have 2215: "(\<forall> i< ilen la. 
                   (sub (inth la i) (inth la (Suc i)) (filt \<sigma> l)) \<Turnstile> g
                ) "
          using "5" powerinterval_def by blast  
      show ?thesis by (simp add: "2214" "2215") 
     qed      
    show ?thesis 
    using "201" "202" "203" "209" "210" "211" "220" by metis
   qed
  show ?thesis 
  by (metis "12" "20" "5" "8" "9" projection_d_def)          
qed      



lemma PJ7sem:
"(\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h) = (f \<triangle> g) \<triangle> h)  "
proof -
 have 1: " ilen \<sigma> >0 \<longrightarrow> (\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h) = (f \<triangle> g) \<triangle> h)"
  using PJ7helpchain1a PJ7helpchain1b unl_lift2 by blast
 have 2: " ilen \<sigma> = 0 \<longrightarrow> (\<sigma> \<Turnstile> f \<triangle> (g \<triangle> h) = (f \<triangle> g) \<triangle> h)"
   using PJ7empty by blast
 from 1 2 show ?thesis by auto
qed


subsubsection \<open>PJ8\<close>

lemma PJ8semhelp:
 assumes "index_sequence 0 l" 
         "  inth l (ilen l) = ilen \<sigma> "
         " (\<forall>n na. na + n \<le> ilen \<sigma> \<longrightarrow> f (sub n (n+ na) \<sigma>) \<longrightarrow> g (sub n (n + na) \<sigma>)) "
 shows "
         (\<forall>i<ilen l. f (sub ( inth l i) ( inth l (Suc i)) \<sigma>)
                       \<longrightarrow> g (sub ( inth l i) ( inth l (Suc i)) \<sigma>)
         )"
by (metis add.commute assms idx_expand 
    ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc)


lemma PJ8sem:
"(\<sigma> \<Turnstile> ba(f \<longrightarrow> g) \<longrightarrow> (f \<triangle> h)  \<longrightarrow> (g \<triangle> h))"
using PJ8semhelp by (simp add: projection_d_def ba_defs powerinterval_def) blast


subsubsection \<open>PJ9\<close>

lemma PJ9sem:
"(\<sigma> \<Turnstile> f \<nabla> ( g \<longrightarrow> h) \<longrightarrow> f \<triangle> g  \<longrightarrow> f \<triangle> h)"
by (simp add: uprojection_d_def projection_d_def, metis)



subsection \<open>Axioms\<close>

lemma BpGen:
 assumes "\<turnstile> f"
 shows   "\<turnstile> bp f"
using assms 
by (simp add: bp_d_def uprojection_d_def projection_d_def Valid_def)

lemma PJ1:
 "\<turnstile> f \<triangle> ( g \<or> h) \<longrightarrow> f \<triangle> g \<or> f \<triangle> h"
using PJ1sem Valid_def by blast

lemma PJ2:
 "\<turnstile> f \<triangle> empty = empty"
 using PJ2sem Valid_def by blast

lemma PJ3:
 "\<turnstile> f \<triangle> skip = (f \<and> more)" 
using PJ3sem Valid_def by blast

lemma PJ4:
 "\<turnstile> f \<triangle> (g;h) = (f \<triangle> g) ; ( f \<triangle> h)"
using PJ4sem Valid_def by blast

lemma PJ5:
 "\<turnstile> f \<triangle> init(g) \<longrightarrow> init(g)"
using PJ5sem Valid_def by blast

lemma PJ6:
 "\<turnstile> skip \<triangle> g = g"
using PJ6sem Valid_def by blast

lemma PJ7:
 "\<turnstile> f \<triangle> (g \<triangle> h) = (f \<triangle> g) \<triangle> h"
using PJ7sem Valid_def by blast

lemma PJ8:
 "\<turnstile> ba(f \<longrightarrow> g) \<longrightarrow> (f \<triangle> h)  \<longrightarrow> (g \<triangle> h)"
using PJ8sem Valid_def by blast

lemma PJ9:
 "\<turnstile> f \<nabla> (g \<longrightarrow> h) \<longrightarrow> f \<triangle> g  \<longrightarrow> f \<triangle> h"
using PJ9sem Valid_def by blast


subsection \<open>Time Reversal\<close>


lemma filt_iapp:
 "filt w (l \<ominus> \<langle>x\<rangle>) = (filt w l) \<ominus> \<langle>(inth w x)\<rangle>"
proof 
 (induct l)
 case (INil x)
 then show ?case by simp
 next
 case (ICons x1a l)
 then show ?case 
 by simp
 qed



lemma filt_irev:
assumes "\<forall> i \<le> ilen l. (inth l i) \<le> ilen w"
 shows "(filt (irev w) l) = (irev (filt w (imap (\<lambda>x. ilen w -x) (irev l)) )) " 
using assms
proof 
  (induct l)
  case (INil x)
  then show ?case 
    proof -
     have 01: "filt (irev w) \<langle>x\<rangle> = \<langle> inth (irev w) x \<rangle>"
       by simp
     have 02: "\<langle> inth (irev w) x \<rangle> = \<langle> inth w (ilen w -x) \<rangle>"  
       using INil.prems irev_inth by auto
     have 03: "irev (filt w (imap ((-) (ilen w)) \<langle>x\<rangle>)) = \<langle> inth w (ilen w -x) \<rangle>" 
       by simp
     from 01 02 03 show ?thesis by auto
    qed
  next
  case (ICons x1a l)
  then show ?case 
     proof -
     have 1: "filt (irev w) (x1a \<odot> l) =
               (inth (irev w) x1a) \<odot> (filt (irev w) l) " 
           by simp
     have 2: "(inth (irev w) x1a) = (inth w (ilen w - x1a) )"
         using ICons.prems irev_inth by fastforce
     have 3: "irev (filt w (x1a \<odot> l))  = irev ((inth w x1a) \<odot> (filt w l)) "
       by simp
     have 4: "irev (filt w (imap ((-) (ilen w)) (irev (x1a \<odot> l)))) =
              irev (filt w  (imap ((-) (ilen w)) ( (irev l) \<ominus> \<langle> x1a \<rangle> )))  
                "
       by simp
     have 5: "irev (filt w  (imap ((-) (ilen w)) ( (irev l) \<ominus> \<langle> x1a \<rangle> ))) =
              irev ( (filt w  (imap ((-) (ilen w)) ( (irev l)))) \<ominus> \<langle> inth w (ilen w -x1a) \<rangle>)"
        by (simp add: filt_iapp) 
     have 6: "irev ( (filt w  (imap ((-) (ilen w)) ( (irev l)))) \<ominus> \<langle> inth w (ilen w -x1a) \<rangle>) =
              (inth w (ilen w -x1a)) \<odot> irev( (filt w  (imap ((-) (ilen w)) ( (irev l)))))  "
        by auto 
     have 7: "\<forall> i \<le> ilen l. (inth l i) \<le> ilen w"
         using ICons(2) by auto
     have 8 : "(filt (irev w) l) = irev(filt w (imap ((-) (ilen w)) (irev l)))"
        using "7" ICons.hyps by blast
     show ?thesis 
     using "2" "5" "8" by auto
     qed
qed

lemma ProjectionRevsema:
 assumes "(\<sigma> \<Turnstile> (f \<triangle> g)\<^sup>r)"
 shows   "  (\<sigma> \<Turnstile> (f\<^sup>r) \<triangle> (g\<^sup>r))"
proof -
 have 1: " (\<sigma> \<Turnstile> (f \<triangle> g)\<^sup>r)" 
  using assms by auto
 have 2: "\<exists>l. index_sequence 0 l \<and>  inth l (ilen l) = ilen \<sigma> \<and> 
          powerinterval f (irev \<sigma>) l \<and> g (filt (irev \<sigma>) l)"
  using "1" by (simp add: projection_d_def reverse_d_def)
 obtain l where 3: "index_sequence 0 l \<and>  inth l (ilen l) = ilen \<sigma> \<and> 
          powerinterval f (irev \<sigma>) l \<and> g (filt (irev \<sigma>) l)" 
  using "2" by auto
 have 4: "index_sequence 0 l" 
  using "3" by auto
 have 5: " inth l (ilen l) = ilen \<sigma> "
  using "3" by auto
 have 6: "powerinterval f (irev \<sigma>) l" 
   using "3" by auto
 have 7: "g (filt (irev \<sigma>) l)" 
    using "3" by auto
 have 8: "\<forall> i \<le> ilen l. (inth l i) \<le> ilen \<sigma>"  
  using "4" "5" idx_less_last_1 le_eq_less_or_eq by fastforce
 have 9: "g ( irev(filt \<sigma> (imap (\<lambda>x. ilen \<sigma> -x) (irev l)) ))" 
  using "7" "8" filt_irev by fastforce
 have 10: "inth (imap ((-) (ilen \<sigma>)) (irev l)) 0 = 0"
      by (metis "5" diff_self_eq_0 ifirst_irev inth_imap )
 have 11: "(\<forall>n<ilen l.
             inth (imap ((-) (ilen \<sigma>)) (irev l)) n
             < inth (imap ((-) (ilen \<sigma>)) (irev l)) (Suc n))" 
      by (simp add: inth_imap irev_inth) 
          (metis (no_types, lifting) "4" "5" Suc_diff_Suc Suc_less_eq diff_less_Suc diff_less_mono2 
           index_sequence_def idx_less_last_1)
 have 12: "index_sequence 0 (imap (\<lambda>x. ilen \<sigma> -x) (irev l))" 
   by (simp add: "10" "11" index_sequence_def)
 have 13: "ilen (imap (\<lambda>x. ilen \<sigma> -x) (irev l)) = ilen l"
    by auto
 have 14: "inth (imap ((-) (ilen \<sigma>)) (irev l)) (ilen (imap ((-) (ilen \<sigma>)) (irev l))) = 
           ilen \<sigma>" 
   by (metis "4" diff_zero index_sequence_def ilast_irev ilen_imap 
        inth_imap)
 have 15: "\<forall>i<ilen l.
           (inth (imap ((-) (ilen \<sigma>)) (irev l)) i) \<le>
           (inth (imap ((-) (ilen \<sigma>)) (irev l)) (Suc i))
          " 
       by (simp add: "11" less_imp_le_nat)
 have 16: "\<forall>i<ilen l. 
            (inth (imap ((-) (ilen \<sigma>)) (irev l)) (Suc i)) \<le> ilen \<sigma>
          " 
     by (simp add: inth_imap)
 have 17: "\<forall>i<ilen l.
           irev
               (sub (inth (imap ((-) (ilen \<sigma>)) (irev l)) i)
                    (inth (imap ((-) (ilen \<sigma>)) (irev l)) (Suc i)) \<sigma>) =
              sub (ilen \<sigma> - (inth (imap ((-) (ilen \<sigma>)) (irev l)) (Suc i)))
                  (ilen \<sigma> - (inth (imap ((-) (ilen \<sigma>)) (irev l)) i))
                  (irev \<sigma>)  
          " 
      using irev_sub 
      using "15" "16" by blast
  have 18: "\<forall>i<ilen l.
             ilen \<sigma> - (inth (imap ((-) (ilen \<sigma>)) (irev l)) (Suc i)) =
             (inth (irev l) (Suc i)) 
           " 
     by (metis "10" "12" "13" "14" Suc_leI diff_diff_cancel idx_less_than 
         ilen_gr_zero inth_imap less_le zero_less_Suc zero_less_diff)
  have 19: "\<forall>i<ilen l.
             ilen \<sigma> - (inth (imap ((-) (ilen \<sigma>)) (irev l)) (i)) =
             (inth (irev l) (i)) " 
    by (metis "10" "18" "5" diff_zero ifirst_irev less_SucI less_Suc_eq_0_disj)
  have 20: "\<forall>i<ilen l.
              ( sub (ilen \<sigma> - (inth (imap ((-) (ilen \<sigma>)) (irev l)) (Suc i)))
                  (ilen \<sigma> - (inth (imap ((-) (ilen \<sigma>)) (irev l)) i))
                  (irev \<sigma>)) =
              ( sub (inth (irev l) (Suc i)) 
                  (inth (irev l) (i)) 
                  (irev \<sigma>))
           " 
    using "18" "19" by simp
  have 21: "\<forall>i<ilen l.
             ( sub (inth (irev l) (Suc i)) 
                  (inth (irev l) (i)) 
                  (irev \<sigma>)) =
            ( sub (inth l (ilen l -  (Suc i))) 
                  (inth l (ilen l - i))  
                  (irev \<sigma>))
             " 
      by (simp add: irev_inth)
  have 22: "\<forall>i<ilen l.
             f ( sub (inth l (ilen l -  (Suc i))) 
                  (inth l (ilen l - i))  
                  (irev \<sigma>))
            " 
       by (metis "6" Suc_diff_Suc Suc_less_eq diff_less_Suc powerinterval_def)
  have 24: "\<forall>i<ilen l.
            f (irev
               (sub (inth (imap ((-) (ilen \<sigma>)) (irev l)) i)
                    (inth (imap ((-) (ilen \<sigma>)) (irev l)) (Suc i)) \<sigma>))"
       by (simp add: "17" "20" "21" "22")
  have 25: "powerinterval (\<lambda>s. f (irev s)) \<sigma> (imap ((-) (ilen \<sigma>)) (irev l))" 
    by (simp add: "24" powerinterval_def)     
  have 26: "(\<exists>l. index_sequence 0 l \<and> 
                 inth l (ilen l) = ilen \<sigma> \<and> 
                 powerinterval (\<lambda>s. f (irev s)) \<sigma> l \<and> g (irev(filt \<sigma> l)))" 
  using "12" "14" "25" "9" by blast    
  from 26 show ?thesis
  by (simp add: projection_d_def reverse_d_def)
qed


lemma ProjectionRevsemb:
 assumes "  (\<sigma> \<Turnstile> (f\<^sup>r) \<triangle> (g\<^sup>r))"
 shows   "(\<sigma> \<Turnstile> (f \<triangle> g)\<^sup>r)"
proof -
 have 1: "(\<exists>l. index_sequence 0 l \<and> 
                 inth l (ilen l) = ilen \<sigma> \<and> 
                 powerinterval (\<lambda>s. f (irev s)) \<sigma> l \<and> g (irev(filt \<sigma> l)))"
  using assms by (simp add:  projection_d_def reverse_d_def)
 obtain l where 2: "index_sequence 0 l \<and> 
                 inth l (ilen l) = ilen \<sigma> \<and> 
                 powerinterval (\<lambda>s. f (irev s)) \<sigma> l \<and> g (irev(filt \<sigma> l))" 
 using "1" by auto
 have 3: "index_sequence 0 l" 
  using "2" by auto
 have 4: "inth l (ilen l) = ilen \<sigma>" 
  using "2" by auto
 have 5: "powerinterval (\<lambda>s. f (irev s)) \<sigma> l"
  using "2" by auto
 have 6: "g (irev(filt \<sigma> l))"
  using "2" by auto
 have 7: "ilen (imap ((-) (ilen \<sigma>)) (irev l)) = ilen l"
   by simp 
 have 8: "\<forall> i \<le> ilen l. (inth (imap ((-) (ilen \<sigma>)) (irev l)) i) \<le> ilen \<sigma>"
   by (simp add: inth_imap)
 have 9: "g (filt (irev \<sigma>) (imap ((-) (ilen \<sigma>)) (irev l)))" 
  by (metis "2" filt_irev idx_less_equal irev_ilen irev_irev_ident 
      le_refl)
 have 10: "inth (imap ((-) (ilen \<sigma>)) (irev l)) (ilen (imap ((-) (ilen \<sigma>)) (irev l))) = ilen \<sigma>" 
   by (metis "2" diff_zero index_sequence_def ilast_irev ilen_imap 
       inth_imap )
 have 11: "inth (imap ((-) (ilen \<sigma>)) (irev l)) 0 = 0"
      by (metis "4" diff_self_eq_0 ifirst_irev  inth_imap )
 have 12: "(\<forall>n<ilen l.
             inth (imap ((-) (ilen \<sigma>)) (irev l)) n
             < inth (imap ((-) (ilen \<sigma>)) (irev l)) (Suc n))" 
    by (simp add: inth_imap irev_inth)
      (metis (no_types, lifting) "2" Suc_diff_Suc Suc_less_eq diff_less_Suc diff_less_mono2 
       index_sequence_def idx_less_last_1)
 have 13: "index_sequence 0 (imap (\<lambda>x. ilen \<sigma> -x) (irev l))"
   by (simp add: "11" "12" index_sequence_def)
 have 14: "\<forall>i<ilen l. f (irev (sub (inth l i) (inth l (Suc i)) \<sigma>))"
  using "5" by (simp add:powerinterval_def)
 have 15: "\<forall>i<ilen l. (inth l (Suc i)) \<le> ilen \<sigma>"
    using "2" idx_expand by fastforce
 have 16: "\<forall>i<ilen l. (inth l i) \<le> (inth l (Suc i))" 
    using "2" idx_expand by fastforce
 have 17: "\<forall>i<ilen l. 
            f (sub ((ilen \<sigma>)-(inth l (Suc i))) ((ilen \<sigma>) - (inth l i)) (irev \<sigma>))  
          " 
    using "14" 
    by (simp add: irev_sub "15" "16")
 have 18: "\<forall>i<ilen l.
            (inth (imap ((-) (ilen \<sigma>)) (irev l)) i) =
            ilen \<sigma> - (inth (irev l) i) 
          " 
  using inth_imap by blast
 have 19: "\<forall>i<ilen l.
           (inth (imap ((-) (ilen \<sigma>)) (irev l)) (Suc i)) =
            ilen \<sigma> - (inth (irev l) (Suc i)) 
          " 
   using inth_imap by blast
 have 20: "\<forall>i<ilen l.
            (inth (irev l) i) = (inth l (ilen l - i))
          " 
     by (simp add: irev_inth)
 have 21: "\<forall>i<ilen l.
            (inth (irev l) (Suc i)) = (inth l (ilen l - (Suc i)))" 
   by (simp add: irev_inth)
 have 22: "\<forall>i<ilen l.
           f (sub (ilen \<sigma> - (inth l (ilen l - i))) 
                   (ilen \<sigma> - (inth l (ilen l - (Suc i))) )
                   (irev \<sigma>))
          " 
   by (metis "17" Suc_diff_Suc Suc_less_eq diff_less_Suc)
 have 23: "\<forall>i<ilen l.
            f (sub (ilen \<sigma> - (inth (irev l) i)) 
                   (ilen \<sigma> - (inth (irev l) (Suc i)) )
                   (irev \<sigma>))
          " 
   by (simp add: "20" "21" "22")
 have 24: "\<forall>i<ilen l.
            f (sub (inth (imap ((-) (ilen \<sigma>)) (irev l)) i)
                   (inth (imap ((-) (ilen \<sigma>)) (irev l)) (Suc i)) (irev \<sigma>))"
  by (simp add: "23" inth_imap)
 have 25: "powerinterval f (irev \<sigma>) (imap ((-) (ilen \<sigma>)) (irev l))" 
   by (simp add: "24" powerinterval_def)
 have 26: "\<exists>l. index_sequence 0 l \<and>  inth l (ilen l) = ilen \<sigma> \<and> 
          powerinterval f (irev \<sigma>) l \<and> g (filt (irev \<sigma>) l)" 
   using "10" "13" "25" "9" by blast
 from 26 show ?thesis by (simp add: projection_d_def reverse_d_def)
qed



lemma ProjectionRev:
 "\<turnstile> (f \<triangle> g)\<^sup>r = f\<^sup>r \<triangle> g\<^sup>r"
using ProjectionRevsema ProjectionRevsemb unl_lift2 by blast



subsection \<open>Theorems\<close>

subsubsection \<open>Projection\<close>

lemma PowerProjLen:
 "\<turnstile> f \<triangle> len n = power (f \<and> more) n"
proof 
  (induct n)
  case 0
  then show ?case by (metis PJ2 len_d_def pow_0)
  next
  case (Suc n)
  then show ?case 
  by (metis PJ3 PJ4 inteq_reflection len_d_def pow_Suc)
  qed  

lemma ProjLenExist:
 "\<turnstile> f \<triangle> (\<exists>n. len n) = (\<exists> n. f \<triangle> len n)"
  by (simp add: Valid_def projection_d_def, blast)

lemma PowerProjLenExist:
 "\<turnstile> (\<exists> n. f \<triangle> len n) = (\<exists> n. power (f \<and> more) n)"
using PowerProjLen by (simp add: Valid_def PowerProjLen, blast)


lemma RightProjImpProj:
 assumes "\<turnstile> g1 \<longrightarrow> g2"
 shows   "\<turnstile> f \<triangle> g1 \<longrightarrow> f \<triangle> g2"
using assms 
by (simp add: Valid_def projection_d_def, blast)

lemma LeftProjImpProj:
 assumes "\<turnstile> f1 \<longrightarrow> f2"
 shows   "\<turnstile> f1 \<triangle> g \<longrightarrow> f2 \<triangle> g"
using assms 
by (simp add: Valid_def projection_d_def powerinterval_def, blast)

lemma RightProjEqvProj:
 assumes "\<turnstile> g1 = g2"
 shows   "\<turnstile> f \<triangle> g1 = f \<triangle> g2"
using assms 
by (metis (mono_tags, lifting) Valid_def inteq_reflection unl_lift2)

lemma LeftProjEqvProj:
 assumes "\<turnstile> f1 = f2"
 shows   "\<turnstile> f1 \<triangle> g = f2 \<triangle> g"
using assms 
by (metis (mono_tags, lifting) Valid_def inteq_reflection unl_lift2)


lemma ProjTrueEqvChopstar:
 "\<turnstile> f \<triangle> #True = f\<^sup>\<star>"
proof -
 have 1: "\<turnstile>  #True = ( (\<exists>n. len n))"
   by (simp add: Valid_def len_defs)
 have 2: "\<turnstile> f \<triangle> #True = f \<triangle> (\<exists>n. len n) " 
   using "1" RightProjEqvProj by blast
 have 3: "\<turnstile> f \<triangle> (\<exists>n. len n) = (\<exists> n. power (f \<and> more) n)" 
  using PowerProjLenExist ProjLenExist by fastforce
 have 4: "\<turnstile> (\<exists> n. power (f \<and> more) n) = f\<^sup>\<star>" 
  by (simp add: chopstar_d_def powerstar_d_def)
 show ?thesis using "2" "3" "4" by fastforce
qed

lemma ProjChopstarEqvChopstarProj:
 "\<turnstile> f \<triangle> (g\<^sup>\<star>) = (f \<triangle> g)\<^sup>\<star>"
proof -
 have 1: "\<turnstile> f \<triangle> (g\<^sup>\<star>) = f \<triangle> (g \<triangle> #True)"
  by (metis ProjTrueEqvChopstar RightProjEqvProj inteq_reflection) 
 have 2: "\<turnstile> f \<triangle> (g \<triangle> #True) = (f \<triangle> g) \<triangle> #True"
  by (simp add: PJ7) 
 have 3: "\<turnstile> (f \<triangle> g) \<triangle> #True = (f \<triangle> g)\<^sup>\<star>" 
    by (simp add: ProjTrueEqvChopstar)
 show ?thesis using "1" "2" "3" by fastforce
qed  


lemma ProjAndImp:
 " \<turnstile> f \<triangle> (g1 \<and> g2) \<longrightarrow> f \<triangle> g1 \<and> f \<triangle> g2  "
by (meson Prop12 RightProjImpProj int_iffD1 lift_and_com)

lemma ProjOrDist:
 "\<turnstile> #True \<triangle> (f \<or>  g) = ( #True \<triangle> f  \<or>  #True \<triangle> g) "
using PJ1sema by blast

lemma StateImportProj:
 "\<turnstile> ((init w) \<and> f \<triangle> g) = f \<triangle> ((init w) \<and> g)"
by (auto simp add: Valid_def init_defs projection_d_def filt_inth index_sequence_def) 

lemma ProjStateAndNextEqvStateAndMoreChopProj:
" \<turnstile> f \<triangle> ( (init w) \<and> \<circle> g) = ((init w) \<and> (f \<and> more);(f\<triangle> g))"
proof -
 have 2: "\<turnstile> (f \<and> more);(f\<triangle> g) = f \<triangle> \<circle> g"  
   by (metis PJ3 PJ4 inteq_reflection next_d_def)   
 have 3: "\<turnstile> f \<triangle> ( (init w) ) \<longrightarrow> init w "
   by (simp add: PJ5)
 have 4: "\<turnstile> (init w \<and>   f \<triangle> \<circle> g) = f \<triangle> ( (init w) \<and> \<circle> g) "
  by (simp add: StateImportProj)
 have 5: "\<turnstile> f \<triangle> ( (init w) \<and> \<circle> g) \<longrightarrow> ((init w) \<and> (f \<and> more);(f\<triangle> g))"
    using "2" "3" ProjAndImp by fastforce
  from 5 4 show ?thesis using "2" by fastforce
qed

lemma ProjNext:
 "\<turnstile> f \<triangle> \<circle> g = (f \<and> more);(f\<triangle> g)"
by (metis PJ3 PJ4 inteq_reflection next_d_def)

lemma ProjWnext:
 "\<turnstile> f \<triangle> (wnext g) = (empty \<or> (f \<and> more);(f\<triangle> g))"
proof -
 have 1: "\<turnstile> f \<triangle> (wnext g) = f \<triangle> (empty \<or> \<circle> g)"
  by (simp add: RightProjEqvProj WnextEqvEmptyOrNext)
 have 2: "\<turnstile> f \<triangle> (empty \<or> \<circle> g) = (empty \<or> f \<triangle> (\<circle> g)) "
   using PJ1sema PJ2 by fastforce
 have 3: "\<turnstile> f \<triangle> (\<circle> g) = (f \<and> more);(f\<triangle> g)"
   by (metis PJ3 PJ4 inteq_reflection next_d_def)
 show ?thesis 
 using "1" "2" "3" by fastforce
qed


lemma ProjIntro:
 assumes "\<turnstile>   f \<and>  more  \<longrightarrow> (g \<and>  more ); f "
 shows   "\<turnstile> f \<longrightarrow> g \<triangle> #True "
using assms CSIntro ProjTrueEqvChopstar by force 
    

lemma RightBoxStateImportProj:
 "\<turnstile> \<box>(init w) \<and> f \<triangle> g \<longrightarrow> f \<triangle> (\<box>(init w) \<and> g)"
by (simp add: Valid_def always_defs init_defs projection_d_def)
 (metis filt_ilen filt_inth idx_bound_1 suffix_zero)


lemma LeftBoxStateImportProjhelp:
 "(\<forall>n\<le>ilen wa. w \<langle> inth wa n\<rangle>) \<and>
         (\<exists>l.  inth l 0 = 0 \<and>
              (\<forall>n<ilen l.  inth l n <  inth l (Suc n)) \<and>
               inth l (ilen l) = ilen wa \<and>
              (\<forall>i<ilen l. f (sub ( inth l i) ( inth l (Suc i)) wa)) \<and>
              g (filt wa l)) \<longrightarrow>
         (\<exists>l.  inth l 0 = 0 \<and>
              (\<forall>n<ilen l.  inth l n <  inth l (Suc n)) \<and>
               inth l (ilen l) = ilen wa \<and>
              (\<forall>i<ilen l.
                  f (sub ( inth l i) ( inth l (Suc i)) wa) \<and>
                  (\<forall>n\<le>ilen (sub ( inth l i) ( inth l (Suc i)) wa).
                      w \<langle> inth (sub ( inth l i) ( inth l (Suc i)) wa) n\<rangle>)) \<and>
              g (filt wa l))"
proof 
 assume 0: "(\<forall>n\<le>ilen wa. w \<langle> inth wa n\<rangle>) \<and>
    (\<exists>l.  inth l 0 = 0 \<and>
         (\<forall>n<ilen l.  inth l n <  inth l (Suc n)) \<and>
          inth l (ilen l) = ilen wa \<and>
         (\<forall>i<ilen l. f (sub ( inth l i) ( inth l (Suc i)) wa)) \<and>
         g (filt wa l))" 
 show   "\<exists>l.  inth l 0 = 0 \<and>
        (\<forall>n<ilen l.  inth l n <  inth l (Suc n)) \<and>
         inth l (ilen l) = ilen wa \<and>
        (\<forall>i<ilen l.
            f (sub ( inth l i) ( inth l (Suc i)) wa) \<and>
            (\<forall>n\<le>ilen (sub ( inth l i) ( inth l (Suc i)) wa).
                w \<langle> inth (sub ( inth l i) ( inth l (Suc i)) wa) n\<rangle>)) \<and>
        g (filt wa l)"
  proof -
   have 1: "(\<forall>n\<le>ilen wa. w \<langle> inth wa n\<rangle>) "
    using 0 by auto
   have 2: "(\<exists>l.  inth l 0 = 0 \<and>
         (\<forall>n<ilen l.  inth l n <  inth l (Suc n)) \<and>
          inth l (ilen l) = ilen wa \<and>
         (\<forall>i<ilen l. f (sub ( inth l i) ( inth l (Suc i)) wa)) \<and>
         g (filt wa l))"
     using 0 by auto
   obtain l where 3: " inth l 0 = 0 \<and>
         (\<forall>n<ilen l.  inth l n <  inth l (Suc n)) \<and>
          inth l (ilen l) = ilen wa \<and>
         (\<forall>i<ilen l. f (sub ( inth l i) ( inth l (Suc i)) wa)) \<and>
         g (filt wa l)"
      using "2" by auto
   have 4: " inth l 0 = 0" 
     using "3" by auto
   have 5: "(\<forall>n<ilen l.  inth l n <  inth l (Suc n))"  
     using "3" by auto
   have 6: " inth l (ilen l) = ilen wa" 
      using "3" by auto
   have 7: "(\<forall>i<ilen l. f (sub ( inth l i) ( inth l (Suc i)) wa))" 
      using "3" by auto
   have 8: "g (filt wa l)"
      using "3" by auto
   have 9: "(\<forall>i<ilen l.
            f (sub ( inth l i) ( inth l (Suc i)) wa) \<and>
            (\<forall>n\<le>(inth l (Suc i)) -(inth l i).
                w \<langle>inth wa ((inth l i)+n)\<rangle>))" 
   by (metis "1" "7" inth_last_stutter nat_le_iff_add nat_le_linear)
   have 10: "(\<forall>i<ilen l. 
               ilen (sub ( inth l i) ( inth l (Suc i)) wa) =
                (inth l (Suc i)) -(inth l i) ) " 
       by (simp add: "3" PJ6help1 index_sequence_def)
   have 11: "(\<forall>i<ilen l.
              (\<forall>n\<le>(inth l (Suc i)) -(inth l i).
                 inth (sub ( inth l i) ( inth l (Suc i)) wa) n =
                 inth wa ((inth l i)+n) )) "
       using "3" index_sequence_def idx_expand by fastforce
  have 12: "(\<forall>i<ilen l.
            f (sub ( inth l i) ( inth l (Suc i)) wa) \<and>
            (\<forall>n\<le>ilen (sub ( inth l i) ( inth l (Suc i)) wa).
                w \<langle> inth (sub ( inth l i) ( inth l (Suc i)) wa) n\<rangle>))"
    using "9" "10" "11" by simp 
  show ?thesis 
  using "12" "3" by blast
  qed
qed
 
lemma LeftBoxStateImportProj:
 "\<turnstile> \<box>(init w) \<and> f \<triangle> g \<longrightarrow> (f \<and> \<box> (init w)) \<triangle> g"
using LeftBoxStateImportProjhelp
 by (simp add: index_sequence_def Valid_def always_defs init_defs projection_d_def powerinterval_def)
 blast

subsubsection \<open>dp and bp\<close>

lemma NotDpEqvBpNot:
 "\<turnstile> (\<not>(dp f)) =  bp (\<not> f)"
by (simp add: bp_d_def dp_d_def uprojection_d_def)


lemma NotBpEqvDpNot:
 "\<turnstile> (\<not>(bp f)) = dp(\<not> f)"
by (simp add: bp_d_def dp_d_def uprojection_d_def)


lemma NowImpDp:
 "\<turnstile> f \<longrightarrow> dp f"
proof -
 have 1: "\<turnstile>  (skip \<longrightarrow> #True)"
  by simp
 have 2: "\<turnstile> ba(skip \<longrightarrow> #True)"
  using "1" by (simp add: BaGen)
 have 3: "\<turnstile> ba(skip \<longrightarrow> #True) \<longrightarrow> (skip \<triangle> f \<longrightarrow> #True \<triangle> f)"
  using PJ8 by blast 
 have 4: "\<turnstile> (skip \<triangle> f \<longrightarrow> #True \<triangle> f)" 
  using "2" "3" MP by blast
 show ?thesis 
 by (metis "4" PJ6 dp_d_def inteq_reflection)
qed 

lemma BpElim:
 "\<turnstile> bp f \<longrightarrow> f"
proof -
 have 1: "\<turnstile> \<not> f \<longrightarrow> dp (\<not> f)"
   by (simp add: NowImpDp) 
 hence 2: "\<turnstile> \<not>(dp (\<not> f)) \<longrightarrow> f"
   by auto
 from 2 show ?thesis 
 by (simp add: bp_d_def dp_d_def uprojection_d_def)
qed

lemma BpImpDpImpDp:
  "\<turnstile> bp (f \<longrightarrow> g) \<longrightarrow>  dp  f \<longrightarrow>  dp  g "
proof -
 have 1: "\<turnstile> bp (f\<longrightarrow> g)  \<longrightarrow> (#True \<triangle> f) \<longrightarrow> (#True \<triangle> g) "  
   by (simp add: PJ9 bp_d_def)
 from 1 show ?thesis   by (simp add: dp_d_def)
qed



lemma BpContraPosImpDist:
  "\<turnstile> bp (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> (bp  f) \<longrightarrow> (bp  g) "
proof -
  have  1: "\<turnstile> bp (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> ( dp (\<not>  g)) \<longrightarrow> ( dp (\<not>  f)) " 
     by (rule BpImpDpImpDp)
  hence 2: "\<turnstile> bp (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> (\<not> ( dp (\<not>  f))) \<longrightarrow> (\<not> ( dp (\<not>  g))) " by auto
  from 2 show ?thesis 
  by (simp add: bp_d_def dp_d_def uprojection_d_def)
qed


lemma BpImpDist:
 "\<turnstile> bp (f\<longrightarrow> g) \<longrightarrow> (bp  f) \<longrightarrow> (bp  g) "
proof -
 have  1: "\<turnstile> (f \<longrightarrow> g) \<longrightarrow> (\<not>  g \<longrightarrow> \<not>  f) " by auto
 hence 2: "\<turnstile> \<not> (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> \<not> (f \<longrightarrow> g) " by auto
 hence 3: "\<turnstile> bp (\<not> (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> \<not> (f \<longrightarrow> g)) " by (rule BpGen)
 have  4: "\<turnstile> bp (\<not> (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> \<not> (f \<longrightarrow> g))
             \<longrightarrow>  
             bp (f \<longrightarrow> g) \<longrightarrow> bp (\<not>  g \<longrightarrow> \<not>  f) " by (rule BpContraPosImpDist)
 have  5: "\<turnstile> bp (f \<longrightarrow> g) \<longrightarrow> bp (\<not>  g \<longrightarrow> \<not>  f) " using 3 4 MP by blast
 have  6: "\<turnstile> bp (\<not>  g \<longrightarrow> \<not>  f) \<longrightarrow> (bp  f) \<longrightarrow> (bp  g) " by (rule BpContraPosImpDist)
 from 5 6 show ?thesis  using lift_imp_trans by blast 
qed

lemma DpImpDpRule:
 assumes "\<turnstile> f\<longrightarrow> g "
 shows   "\<turnstile>  dp  f \<longrightarrow>  dp  g "
proof -
 have  1: "\<turnstile> f\<longrightarrow> g " using assms by auto
 hence 2: "\<turnstile> #True \<triangle> f \<longrightarrow> #True \<triangle> g " 
   by (metis BpGen MP PJ9 bp_d_def)
 from 2 show ?thesis by (simp add: dp_d_def)
qed

lemma BpImpBpRule:
 assumes "\<turnstile> f\<longrightarrow> g"
 shows   "\<turnstile> bp  f \<longrightarrow> bp  g "
proof -
 have  1: "\<turnstile> f\<longrightarrow> g" using assms by auto
 hence 2: "\<turnstile> \<not>  g \<longrightarrow> \<not>  f " by auto
 hence 3: "\<turnstile>  dp (\<not>  g) \<longrightarrow>  dp (\<not>  f) " by (rule DpImpDpRule)
 hence 4: "\<turnstile> \<not> ( dp (\<not>  f)) \<longrightarrow> \<not> ( dp (\<not>  g)) " by auto
 from 4 show ?thesis 
   by (meson BpGen BpImpDist MP assms)
qed

lemma DpEqvDpRule:
 assumes "\<turnstile> f = g "
 shows   "\<turnstile>  dp f =  dp g "
proof -
 have  1: "\<turnstile> f = g " using assms by auto
 hence 2: "\<turnstile> #True \<triangle> f =  #True \<triangle> g " 
    using RightProjEqvProj by blast
 from 2 show ?thesis by (simp add: dp_d_def)
qed

lemma BpEqvBpRule:
 assumes "\<turnstile> f = g "
 shows   "\<turnstile>  bp f =  bp  g "
proof -
 have  1: "\<turnstile> f = g " using assms by auto
 hence 2: "\<turnstile> (\<not>  f) = (\<not>  g) " by auto
 hence 3: "\<turnstile>  dp (\<not>  f) =  dp (\<not>  g) " by (rule DpEqvDpRule)
 hence 4: "\<turnstile> (\<not>  (dp (\<not>  f))) = (\<not> ( dp (\<not>  g))) " by auto
 from 4 show ?thesis 
 by (metis BpImpBpRule assms int_iffD1 int_iffI inteq_reflection)
qed


lemma DpState:
 "\<turnstile> dp (init w) = (init w) "
by (metis NowImpDp PJ5 dp_d_def int_iffI)

lemma StateEqvBp:
 "\<turnstile>   (init w) = bp  (init w) "
proof -
  have 1: "\<turnstile> (init w) \<longrightarrow> bp  (init w) " 
   by (metis (no_types, lifting) DiState DpState Initprop(2) StateEqvBi bi_d_def bp_d_def dp_d_def 
       int_iffD1 inteq_reflection uprojection_d_def)
  have 2: "\<turnstile> bp  (init w) \<longrightarrow> (init w) " by (rule BpElim)
  from 1 2 show ?thesis by fastforce
qed

lemma DpDpEqvDp:
 "\<turnstile> dp (dp f) = dp f"
proof -
 have 2: "\<turnstile> #True \<triangle> ( #True \<triangle> f) = (#True \<triangle> #True) \<triangle> f" 
   by (simp add: PJ7)
 have 3: "\<turnstile> (#True \<triangle> #True) = #True"
    by (metis DpState Initprop(4) dp_d_def int_eq_true inteq_reflection) 
 show ?thesis by (metis "2" "3" dp_d_def inteq_reflection)
qed


lemma BpBpEqvBp:
 "\<turnstile> bp (bp f) = bp f"
proof -
 have 1: "\<turnstile> dp (dp (\<not> f)) = dp (\<not> f)"
   using DpDpEqvDp by blast 
 have 2: "\<turnstile> (\<not> (dp (dp (\<not> f)))) = (\<not> (dp (\<not> f)))"
   using "1" by auto
 have 3: "\<turnstile> (\<not> (dp (\<not> f))) = bp f"
    by (simp add: bp_d_def dp_d_def uprojection_d_def)
 have 4: "\<turnstile> (\<not> (dp (dp (\<not> f)))) = bp (bp f)"
    by (simp add: bp_d_def dp_d_def uprojection_d_def) 
 from 2 3 4 show ?thesis 
 by fastforce
qed



lemma DpOrEqv:
 "\<turnstile>    dp (f \<or>  g) =  (dp  f \<or>   dp  g) " 
proof -
 have 1: "\<turnstile> #True \<triangle> (f\<or>  g) = ( #True \<triangle> f  \<or>  #True \<triangle> g) " 
  using ProjOrDist by auto
 from 1 show ?thesis by (simp add: dp_d_def)
qed

lemma BpAndEqv:
 "\<turnstile> bp(f \<and> g) = (bp f \<and> bp g)"
proof -
 have 1: "\<turnstile> dp ((\<not> f) \<or>  (\<not>g)) =  (dp  (\<not>f) \<or>   dp  (\<not>g))"
   using DpOrEqv by auto
 hence 2: "\<turnstile> (\<not> (dp ((\<not> f) \<or>  (\<not>g)))) =  (\<not>(dp  (\<not>f) \<or>   dp  (\<not>g)))  "
   by auto
 have 3: "\<turnstile> (\<not> (dp ((\<not> f) \<or>  (\<not>g)))) = bp ( \<not>((\<not> f) \<or>  (\<not>g)))"
       using NotDpEqvBpNot by blast
 have 4: "\<turnstile> (\<not>((\<not> f) \<or>  (\<not>g))) =(f \<and> g)"
   by auto
 hence 5: "\<turnstile> bp(\<not>((\<not> f) \<or>  (\<not>g))) = bp(f \<and> g)"
    by (simp add: BpEqvBpRule)
 have 6: "\<turnstile> (\<not>(dp  (\<not>f) \<or>   dp  (\<not>g))) = ((\<not>(dp (\<not> f))) \<and> (\<not>(dp (\<not> g))))"
   by auto
 have 7: "\<turnstile> ((\<not>(dp (\<not> f))) \<and> (\<not>(dp (\<not> g)))) = (bp f \<and> bp g)"
    by (simp add: bp_d_def dp_d_def uprojection_d_def)
  show ?thesis 
  by (metis "2" "3" "4" "6" "7" inteq_reflection)
qed

lemma DpAndA:
 "\<turnstile>    dp (f \<and> g) \<longrightarrow>  dp  f "
proof -
 have 1: "\<turnstile> #True \<triangle> (f \<and> g) \<longrightarrow> #True \<triangle> f " 
   by (meson Prop12 RightProjImpProj int_iffD1 lift_and_com)
 from 1 show ?thesis by (simp add: dp_d_def)
qed

lemma BpOrA:
 "\<turnstile> bp f \<longrightarrow>   bp(f \<or> g)"
by (simp add: BpImpBpRule intI)

lemma BpOrB:
 "\<turnstile> bp g \<longrightarrow>   bp(f \<or> g)"
by (simp add: BpImpBpRule intI)

lemma BpOrImpOr:
 "\<turnstile> bp f \<or> bp g \<longrightarrow> bp(f \<or> g)"
using BpOrA BpOrB by fastforce


lemma DpAndB:
 "\<turnstile>    dp (f \<and> g) \<longrightarrow>  dp  g "
proof -
 have 1: "\<turnstile> #True \<triangle> (f \<and> g) \<longrightarrow> #True \<triangle> g " 
    by (meson Prop12 RightProjImpProj int_iffD2 lift_and_com)
 from 1 show ?thesis by (simp add: dp_d_def)
qed

lemma DpAndImpAnd:
 "\<turnstile>    dp (f \<and> g) \<longrightarrow>  dp  f \<and>  dp  g "  
proof -
 have 1: "\<turnstile>  dp (f \<and> g) \<longrightarrow>  dp  f " by (rule DpAndA)
 have 2: "\<turnstile>  dp (f \<and> g) \<longrightarrow>  dp  g " by (rule DpAndB)
 from 1 2 show ?thesis by fastforce
qed

lemma DpSkipEqvMore:
 "\<turnstile>    dp  skip  =  more "
proof -
 have 1: "\<turnstile> dp skip = #True \<triangle> skip"
  by (simp add: dp_d_def)
 have 2: "\<turnstile> #True \<triangle> skip = (#True \<and> more)"
 using PJ3 by blast
 have 3: "\<turnstile> (#True \<and> more) = more"
   by auto
 from 1 2 3 show ?thesis by fastforce
qed 

lemma DpMoreEqvMore:
 "\<turnstile>    dp  more  =  more "
by (metis DpDpEqvDp DpSkipEqvMore inteq_reflection)



lemma BpEmptyEqvEmpty:
 "\<turnstile> bp empty = empty"
by (metis DpMoreEqvMore NotDpEqvBpNot empty_d_def inteq_reflection)

lemma DpEmptyEqvEmpty:
 "\<turnstile>    dp  empty = empty"
proof -
 have 1: "\<turnstile> dp empty = #True \<triangle> empty"
  by (simp add: dp_d_def)
 have 2: "\<turnstile> #True \<triangle> empty = empty"
   by (simp add: PJ2)
 from 1 2 show ?thesis by fastforce
qed


lemma BpMoreEqvMore:
 "\<turnstile> bp more =more"
by (metis DpEmptyEqvEmpty NotDpEqvBpNot NotEmptyEqvMore inteq_reflection)

lemma NextDpImpDpNext:
 "\<turnstile>  \<circle> (dp f) \<longrightarrow> dp (\<circle> f)"
proof -
 have 1: "\<turnstile> dp(\<circle> f) = #True \<triangle> (skip;f) "
  by (simp add: dp_d_def next_d_def) 
 have 2: "\<turnstile> #True \<triangle> (skip;f) = (#True \<triangle> skip);(#True \<triangle> f)"
  by (simp add: PJ4)
 have 3: "\<turnstile> (#True \<triangle> skip) = (#True \<and> more)"
  using PJ3 by blast
 have 4: "\<turnstile> (#True \<and> more) = more"
   by auto
 have 5: "\<turnstile> skip;(#True \<triangle> f) \<longrightarrow> more;(#True \<triangle> f) "
  by (metis DpSkipEqvMore LeftChopImpChop NowImpDp inteq_reflection)
 show ?thesis  
 by (metis "2" "3" "4" "5" dp_d_def inteq_reflection next_d_def)
qed

lemma BoxStateImportBp:
 "\<turnstile> \<box>(init w)  \<longrightarrow> bp (\<box>(init w))"
by (simp add: Valid_def always_defs init_defs projection_d_def bp_d_def uprojection_d_def 
       powerinterval_def)
    (metis (mono_tags, lifting) filt_expand idx_bound_1 suffix_zero)

lemma BoxStateEqvBpBoxState:
 "\<turnstile> \<box> (init w) = bp(\<box> (init w))"
proof -
 have 1: "\<turnstile>  bp(\<box> (init w)) \<longrightarrow> \<box> (init w)" 
  by (simp add: BpElim)
 have 2: "\<turnstile>  bp(\<box> (init w)) = (\<not>(#True \<triangle> (\<not> \<box> (init w))))"
   by (simp add: bp_d_def uprojection_d_def)
 have 2: "\<turnstile> \<box>(init w) \<longrightarrow> dp (\<box>(init w))"
    by (metis NowImpDp)
 have 2: "\<turnstile> \<box> (init w) \<longrightarrow> bp(\<box> (init w))"
 using BoxStateImportBp by auto
 from 1 2 show ?thesis by fastforce
qed



end
